 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: ChoicEff                                            */
 /*   TITLE: Efficient Experimental Choice Designs               */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs, choice modeling        */
 /*   PROCS: IML, TRANSREG                                       */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: This macro creates global macro variables           */
 /*          BEST _TRGIND _TRGINDN _MktVer                       */
 /*                                                              */
 /*          The following SAS products are required to run      */
 /*          this macro: BASE, STAT, IML.                        */
 /****************************************************************/

 /*---------------------------------------------------------------------

  This macro is used to find efficient experimental designs for choice
  experiments.  You supply a candidate set of either alternatives or
  sets of alternatives.  The macro searches the candidates for an
  efficient experimental design -- a design in which the variances
  of the parameter estimates are minimized.

  There are two primary ways to use the macro:

  1) You can create a candidate set of alternatives.
     Then the macro builds a design consisting of choice sets built
     from the alternatives you supplied.  You must designate for each
     candidate alternative the design alternative(s) for which it is
     a candidate.  For a branded study with (say) four brands, you must
     create four lists of candidate alternatives -- one for each brand.

  2) You can create a candidate set of choice sets.
     Then the macro builds a design from the choice sets you supplied.
     You can also use this approach to evaluate a choice design before
     you collect any data.

  Typically, you use designs created with the MktEx macro as candidate
  sets, such as full-factorial designs, orthogonal arrays, or other designs.

  The macro either constructs a random initial design from the
  candidates, or it uses an initial design that you specify.
  Then it considers swapping out every design alternative/set
  and replacing it with each candidate alternative/set.
  Swaps that increase efficiency are performed.
  Swapping continues until efficiency stabilizes.
  Then the process is repeated with a different initial design.
  The best design is output for use.

  The macro uses a modified Federov algorithm, just like PROC OPTEX.
  The key difference between this macro and PROC OPTEX is this macro
  allows you to specify the true (or assumed true) parameters and
  optimizes the variance matrix for a discrete choice model.  PROC
  OPTEX uses a linear model, which does not require you to specify
  a parameter vector.

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

------------------------------------------------------------------------

* Sample usages;

title1 'Generic Model with 3 Three-Level Factors';
title2 'Candidate Set: X1-X3 are the factors';
%mktex(3 ** 3, n = 3**3)

title2 'Find Efficient Design for Unbranded, Purely Generic Model';
%choiceff(data=design, model=class(x1-x3 / sta), nsets=3, options=relative,
          maxiter=10, seed=104, flags=3, beta=zero)

proc print; id set; by set; var x1-x3; run;




title1 'Combinatorialy Constructed Optimal Generic Design';
%mktex(3 ** 4, n = 9)
%mktlab(data=design, vars=Set x1-x3)

title2 'Evaluate Efficiency of the Optimal Design';
%choiceff(data=final, model=class(x1-x3 / sta), nsets=3, nalts=3,
          options=relative, beta=zero, init=final(keep=set), intiter=0)

proc print; by set; id set; var x1-x3; run;




title 'Candidate Set for Set-Swapping Algorithm';
%mktex(3 ** 9, n=2187)

%mktkey(3 3)

%mktroll(design=randomized, key=key, out=frac)

proc print data=frac(obs=30); run;

* Each block of 3 observations is a candidate choice set;

%choiceff(data=frac, model=class(x1-x3 / sta), seed=104, maxiter=10,
          options=relative, nsets=3, nalts=3, beta=zero)




title 'Brand Effects';
%mktex(3 ** 4, n=3**4)

data full(drop=i);
   set randomized(rename=(x4=b));
   array f[3];
   do i = 1 to 3; f[i] = (b eq i); end;
   label b = 'Brand';
   run;

proc print; run;

* X1 - X3 are the attributes, B is the brand,
  F1 flags the Brand 1 alternatives, F2 flags the Brand 2 alternatives,
  F3 flags the Brand 3 alternatives.

  The ZERO= option in CLASS(factors / ZERO=list) is used to specify the
  reference levels to be zeroed.  By default, when ZERO= is not
  specified, dummy variables are created for all but one level, which is
  the reference level.  For example, one dummy variable is created with
  a two-level factor and two dummy variables are created with a
  three-level factor.  ZERO=NONE specifies that all levels be used and
  none zeroed.  Hypothetically, CLASS(X Y Z / ZERO='a' '1' 'c')
  specifies zero the 'a' level of X, the '1' level of Y, and the 'c'
  level of Z.  Specifying a nonexistent level (like ' ') is allowed
  and has the same effect as ZERO=NONE, but for only the designated
  factors.  In CLASS(B B*X1 B*X2 B*X3 / ZERO=' ') the value ZERO=' '
  applies to the first variable B.  Since no levels were specified for
  the other variables, the default reference levels are chosen.  This
  has the effect of using all B levels but not all X1-X3 levels.
  Specifying a nonexistent level is detected and "WARNING: Reference
  level ZERO=' ' was not found for variable B." is printed, which can be
  safely ignored.

  This is a less-than-full-rank (LTFR) coding, so all efficiencies are
  zero, and the iteration table has an extra line per iteration which is
  the ridged efficiency used internally.  The diagonal of the variance
  table shows that the brand 3 effect (B3) is a structural zero.;

title2 'Find Efficient Choice Designs with Brand Effects';
%choiceff(data=full,
          model=class(b b*x1 b*x2 b*x3 / zero=' '),
          nsets=15, flags=f1-f3, beta=zero, converge=1e-12)

title3 'Switch to Full-Rank Coding, Use Best Design as Starting Point';
%choiceff(data=full, init=best(keep=index), drop=b3,
          model=class(b b*x1 b*x2 b*x3 / zero=' '),
          nsets=15, flags=f1-f3, beta=zero, converge=1e-12)



title3 'Set Swapping with Brand Effects';
%mktex(3 ** 9, n=2187)

data key;
  input b (x1-x3) ($);
  datalines;
1 x1 x2 x3
2 x4 x5 x6
3 x7 x8 x9
;

%mktroll(design=randomized, key=key, out=brand, alt=b)

* Each block of 3 observations is a candidate choice set;

proc print data=brand(obs=30); run;

* Just see what the coding is first (omit beta= specification).
  The EFFECTS option specifies -1, 0, 1 coding.;
%choiceff(data=brand, nsets=15, nalts=3,
          model=class(b) class(b*x1 b*x2 b*x3 / effects zero=' '))

* Now specify assumed betas;
%choiceff(data=brand, nsets=15, nalts=3,
          beta=1 2 -0.5 0.5 -0.75 0.75 -1 1
               -0.5 0.5 -0.75 0.75 -1 1 -0.5 0.5 -0.75 0.75 -1 1,
          model=class(b) class(b*x1 b*x2 b*x3 / effects zero=' '))



title 'Cross-Effects Model, Five Brands at 3 Prices';
%mktex(3 ** 5, n=3**5)

%mktlab(values=1.49 1.99 2.49)

data key;
  input (b p) ($);
  datalines;
1 x1
2 x2
3 x3
4 x4
5 x5
. .
;

%mktroll(design=final, key=key, out=crosscan, alt=b, keep=x1-x5)

data crosscan;
   set crosscan;
   label b  = 'Brand' p = 'Price' x1 = 'Brand 1 Price'
         x2 = 'Brand 2 Price' x3 = 'Brand 3 Price'
         x4 = 'Brand 4 Price' x5 = 'Brand 5 Price';
   run;

* Each choice set consists of 6 alternatives:
  5 brands and a constant alternative.;

proc print data=crosscan(obs=30); by set; id set b; run;

* Note only the choice set swapping algorithm can handle
  cross-effects;

title2 'Specify LTFR Coding';
title3 'Seeing Full List of Terms Can be Helpful';
%choiceff(data=crosscan,
          model=class(b b*p / zero=none)
                class(b / zero=none) * identity(x1-x5),
          nsets=20, nalts=6, beta=zero)

* Run again, with a full-rank coding.  The list of dropped names was
  created by cutting from the redundant variable list.;

title2 'Full-Rank Coding';
%choiceff(data=crosscan,
          model=class(b b*p / zero=' ')
                class(b / zero=none) * identity(x1-x5),
          drop=B1X1 B2X2 B3X3 B4X4 B5X5,
          nsets=20, nalts=6, beta=zero)

title2 'One of the Coded Candidate Choice Sets';
proc print data=tmp_cand(firstobs=265 obs=270 keep=&_trgind b p);
   id b p;
   run;



title 'Combinatorially Optimal Partial Profile Design';

%mktex(3 6 3 ** 3, n=18)

data design; set design; drop x2; run;

%mktbibd(b=20, t=16, k=4, seed=104, out=b)

%mktppro(ibd=b, print=f p)

* Since 4 of 16 or 25% of attributes vary, maximum D-efficiency = 25;

title2 'Evaluate Design';
%choiceff(data=chdes, model=class(x1-x16 / sta), nsets=120, nalts=3,
          options=relative, beta=zero, init=chdes, initvars=x1-x16)

proc print; id set; by set; var x1-x16; run;



title 'Prevent Dominated Alternatives';
%mktex(4 ** 6, n=32, seed=104)

%macro res;
   do i = 1 to nalts;
      do k = i + 1 to nalts;
         if all(x[i,] >= x[k,]) then bad = bad + 1;
         if all(x[k,] >= x[i,]) then bad = bad + 1;
         end;
      end;
   %mend;

%choiceff(data=randomized,
          model=class(x1-x6 / sta),
          nsets=8,
          flags=4,
          seed=104,
          options=relative
                  resrep,
          restrictions=res,
          resvars=x1-x6,
          maxiter=1,
          bestout=desres1,
          beta=zero)

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

%macro choiceff(  /*------These next two parameters must be set-------*/
                  /*                                                  */
help,             /* nonblank, list options                           */
                  /*                                                  */
model=,           /* TRANSREG model statement specification.          */
                  /* There are many potential forms for the model     */
                  /* specification and a number of options.           */
                  /* See the SAS/STAT PROC TRANSREG documentation.    */
                  /*                                                  */
                  /* Generic effects example:                         */
                  /* model=class(x1-x3),                              */
                  /*                                                  */
                  /* Brand and alternative specific effects example:  */
                  /* model=class(b)                                   */
                  /*       class(b*x1 b*x2 b*x3 / effects zero=' '),  */
                  /*                                                  */
                  /* Brand, alternative specific, and cross effects:  */
                  /* model=class(b b*p / zero=' ')                    */
                  /*       class(b / zero=none) * identity(x1-x5),    */
                  /*                                                  */
nsets=,           /* Number of choice sets desired.                   */
                  /*                                                  */
                  /*---------Set exactly one of these next two--------*/
                  /*                                                  */
                  /* When the candidate set consists of alternatives  */
                  /* to be swapped, specify FLAGS=.  When the         */
                  /* candidate set consists of sets of entire         */
                  /* choice sets to be swapped, specify NALTS=.       */
                  /*                                                  */
flags=,           /* Flags the alternative(s) for which each          */
                  /* candidate may be used.  There must be one flag   */
                  /* variable per alternative.  For example with      */
                  /* three alternatives, specify FLAGS=f1-f3, and     */
                  /* create a candidate set where:                    */
                  /* alternative 1 candidates are indicated by        */
                  /* f1=1 f2=0 f3=0,                                  */
                  /* alternative 2 candidates are indicated by        */
                  /* f1=0 f2=1 f3=0,                                  */
                  /* alternative 3 candidates are indicated by        */
                  /* f1=0 f2=0 f3=1.                                  */
                  /*                                                  */
                  /* If every candidate can be used in all            */
                  /* alternatives, then the flags are constant:       */
                  /* f1=1 f2=1 f3=1.   When the flags are all         */
                  /* constant (in a purely generic design), you can   */
                  /* have the macro create these flag variables for   */
                  /* you by specifying the number of alternatives     */
                  /* rather than a list of flag variables.            */
                  /* Example: flags=3.                                */
                  /*                                                  */
nalts=,           /* Number of alternatives in each choice set.       */
                  /*                                                  */
                  /*-----The rest of the parameters are optional------*/
                  /*                                                  */
data=_last_,      /* Input choice candidate set.                      */
                  /*                                                  */
out=results,      /* Output data set with final designs.              */
                  /*                                                  */
cov=cov,          /* Output data set with final covariance matrices.  */
                  /*                                                  */
bestout=best,     /* Output best design.                              */
                  /*                                                  */
bestcov=bestcov,  /* Output covariance matrix for best design.        */
                  /*                                                  */
init=,            /* Input initial design data set.  Null means       */
                  /* random start.  One usage is to specify the       */
                  /* BESTOUT= data set for an initial start.          */
                  /* When FLAGS= is specified, INIT= must contain the */
                  /* INDEX variable.  Example: INIT=best(keep=index). */
                  /* When NALTS= is specified, INIT= must contain the */
                  /* choice SET variable.  Example:                   */
                  /* INIT=best(keep=set)                              */
                  /*                                                  */
                  /* Alternatively, the INIT= data set can contain    */
                  /* an arbitrary design, potentially created outside */
                  /* this macro.  Then you must also specify          */
                  /* INITVARS=factors, where factors are the factors  */
                  /* in the design, for example INITVARS=x1-x3.       */
                  /* When alternatives are swapped, this data set     */
                  /* must also contain the FLAGS= variables.          */
                  /* When INIT= is specified with INITVARS=, the      */
                  /* data set may also contain a variable specified   */
                  /* on the FIXED= option, which indicates which      */
                  /* alternatives are fixed, and which ones can be    */
                  /* swapped in and out.                              */
                  /*                                                  */
fixed=,           /* When FIXED=variable is specified, the INIT= data */
                  /* set must contain the named variable, which       */
                  /* indicates which alternatives are fixed (cannot   */
                  /* be swapped out) and which ones may be changed.   */
                  /* Example: FIXED=fixed, INIT=init, INITVARS=x1-x3  */
                  /*                                                  */
                  /* 1 - means this alternative may never be swapped  */
                  /*     out.                                         */
                  /* 0 - means this alternative is used in the        */
                  /*     initial design, but it may be swapped out.   */
                  /* . - means this alternative should be randomly    */
                  /*     initialized, and it may be swapped out.      */
                  /*                                                  */
                  /* FIXED= may be specified only when both INIT=     */
                  /* and INITVARS= are specified.                     */
                  /*                                                  */
beta=,            /* True parameters, default: just report on coding. */
                  /* beta=zero may be specified for all zeros.        */
                  /*                                                  */
n=1,              /* Number of observations to use in variance matrix */
                  /* formula.                                         */
                  /*                                                  */
maxiter=2,        /* Maximum iterations (designs to create).          */
iter=2,           /* Alias for MAXITER=.                              */
                  /*                                                  */
intiter=10,       /* Maximum internal iterations.  Specify INTITER=0, */
                  /* just evaluate efficiency of an existing design.  */
                  /*                                                  */
converge=0.005,   /* D-efficiency convergence criterion.              */
                  /*                                                  */
initvars=,        /* Specifies the factor variables in the INIT= data */
                  /* set that must match up with the variables in the */
                  /* DATA= data set.  See INIT=.  All of these        */
                  /* variables must be of the same type.              */
                  /*                                                  */
morevars=,        /* More variables to add to the model.  This option */
                  /* gives you the ability to specify a list of       */
                  /* variables to copy along as is, through the       */
                  /* TRANSREG coding, then add them to the model.     */
                  /*                                                  */
drop=,            /* If you specified a less-than-full-rank           */
                  /* MODEL= specification, you can use DROP= to       */
                  /* produce a full rank coding.                      */
                  /*                                                  */
submat=,          /* Specifies a submatrix for which efficiency       */
                  /* calculations are desired.  Specify an index      */
                  /* vector.  For example with 3 three-level factors, */
                  /* A, B, and C, and the model CLASS(A B C A*B),     */
                  /* Specify SUBMAT=1:6, to see the efficiency of     */
                  /* just the 6 x 6 matrix of main effects.           */
                  /* Specify SUBMAT=3:6, to see the efficiency of     */
                  /* just the 4 x 4 matrix of B and C main effects.   */
                  /*                                                  */
weight=,          /* Optional weight variable.  Typical usage is      */
                  /* with an availability design.  Give unavailable   */
                  /* alternatives a weight of zero and available      */
                  /* alternatives a weight of one.                    */
                  /*                                                  */
types=,           /* This option is used when you have multiple types */
                  /* of choice sets and you want the design to        */
                  /* consist of only certain numbers of each type.    */
                  /* This option can be specified with the set        */
                  /* swapping algorithm.  The argument is an integer  */
                  /* list.  When you specify types=, you must also    */
                  /* specify typevar=.  Say you are creating a design */
                  /* with 30 choice sets and you want the first 10 to */
                  /* consist of sets whose typevar= variable in the   */
                  /* candidate set are type 1 and the rest type 2.    */
                  /* You would specify types=10 20.                   */
                  /*                                                  */
typevar=,         /* Specifies a variable in the candidate data set   */
                  /* that contains choice set types.  The types must  */
                  /* be integers starting with 1.  This option can    */
                  /* only be specified with the set swapping          */
                  /* algorithm.  When you specify typevar=, you must  */
                  /* also specify types=.                             */
                  /*                                                  */
options=,         /* List options, for example OPTIONS=coded nodups,  */
                  /*                                                  */
                  /*       coded - print coded candidate set.         */
                  /*                                                  */
                  /*      detail - print details of the swaps.        */
                  /*      resrep - alias for detail.                  */
                  /*                                                  */
                  /*     notests - suppresses printing the diagonal   */
                  /* of covariance matrix, hypothesis tests for this  */
                  /* n and beta.  When beta is not zero, the results  */
                  /* include a Wald test statistic which is normally  */
                  /* distributed (Beta / Standard Error) and the      */
                  /* probability of a larger squared Wald statistic.  */
                  /*                                                  */
                  /*     orthcan - orthogonalize the candidate set.   */
                  /*                                                  */
                  /*  nobeststar - do not print an asterisk when      */
                  /* a better design is found.                        */
                  /*                                                  */
                  /*      nocode - skip TRANSREG coding stage,        */
                  /* assumes WORK.TMP_CAND was created by a previous  */
                  /* step.  This is most useful with set swapping     */
                  /* when the candidate set can be big.  It is        */
                  /* important with nocode to note that the effect    */
                  /* of MOREVARS= and DROP= in previous runs has      */
                  /* already been taken care of, so do not specify    */
                  /* them (unless for instance you want to drop       */
                  /* still more variables).                           */
                  /*                                                  */
                  /*      nodups - prevents the same choice set from  */
                  /* coming out more than once.  This option does not */
                  /* affect the initialization, so the random initial */
                  /* design may have duplicates.  This options forces */
                  /* duplicates out during the iterations, so do not  */
                  /* set INTITER= to a small value.  It may take      */
                  /* several iterations to eliminate all duplicates.  */
                  /* It is possible that efficiency will decrease as  */
                  /* duplicates are forced out.  With set swapping,   */
                  /* this macro checks the candidate choice set       */
                  /* numbers to avoid duplicates.  With alternative   */
                  /* swapping, this macro checks the candidate        */
                  /* alternative index to avoid duplicates.  The      */
                  /* macro does not look at the actual factors.       */
                  /* This makes the checks faster, but if the         */
                  /* candidate set contains duplicate choice sets     */
                  /* or alternatives, the macro may not succeed in    */
                  /* eliminating all duplicates.  Run the MktDups     */
                  /* macro (which looks at the actual factors) on     */
                  /* the design to check and make sure all duplicates */
                  /* are eliminated.  If you are using set swapping   */
                  /* to make a generic design make sure you run the   */
                  /* MktDups macro on the candidate set to eliminate  */
                  /* duplicate choice sets in advance.                */
                  /*                                                  */
                  /*    relative - prints the relative D-efficiency   */
                  /* for the final design, whioh by default is 100 *  */
                  /* D-efficiency divided by the number of choice     */
                  /* sets.  In other words, this option causes the    */
                  /* D-efficiency relative to a (perhaps              */
                  /* hypothetical) design with D-efficiency equal to  */
                  /* the number of choice sets to be printed.  When   */
                  /* BETA=zero is specified along with the            */
                  /* standardized orthogonal coding in the MODEL      */
                  /* specification and a generic choice design is     */
                  /* requested, this scales D-efficiency to a 0 to    */
                  /* 100 scale.  Optimal generic choice designs       */
                  /* constructed through combinatorial methods will   */
                  /* have a relative D-efficiency of 100.  While you  */
                  /* can print this value for any other type of       */
                  /* design and specification, it will not generally  */
                  /* be on a 0 to 100 scale except in certain special */
                  /* cases.  You can specify the rscale= option if    */
                  /* you would like relative D-efficiency scaled      */
                  /* relative to any value other than the number of   */
                  /* choice sets.  Here is an example of where it     */
                  /* would make sense to specify OPTIONS=relative.    */
                  /*                                                  */
                  /* %mktex(4 ** 5, n=16)                             */
                  /* %mktlab(data=design, vars=Set x1-x4)             */
                  /* %choiceff(model=class(x1-x4 / sta), beta=zero,   */
                  /*           nalts=4, options=relative, init=final, */
                  /*           initvars=x1-x4, nsets=4, intiter=0)    */
                  /*                                                  */
                  /* The standardized orthogonal coding is requested  */
                  /* with the STA option in the CLASS specification.  */
                  /* Note that this option was first available with   */
                  /* the 9.2 release of SAS.  The final results table */
                  /* contains the relative D-efficiency in addition   */
                  /* to all of the other usual results.  In this      */
                  /* case, since an optimal generic design is being   */
                  /* evaluated, relative D-efficiency is 100.         */
                  /*                                                  */
                  /*           Design                  1              */
                  /*           Choice Sets             4              */
                  /*           Alternatives            4              */
                  /*           Parameters             12              */
                  /*           Maximum Parameters     12              */
                  /*           D-Efficiency       4.0000              */
                  /*           Relative D-Eff   100.0000              */
                  /*           D-Error            0.2500              */
                  /*           1 / Choice Sets    0.2500              */
                  /*                                                  */
                  /*   outputall - output all designs to the OUT= and */
                  /* COV= data sets.  By default, only designs that   */
                  /* improve on the previous best design are output   */
                  /* when MAXITER > 100.  This is a change from       */
                  /* previous releases.                               */
                  /*                                                  */
rscale=,          /* specifies the scaling factor to use for relative */
                  /* D-efficiency computations. In some cases,        */
                  /* you can know what the maximum D-efficiency is    */
                  /* and specify a scaling factor so that you get     */
                  /* a relative D-efficiency in the 0 to 100 range.   */
                  /* When you specify rscale=, options=relative is    */
                  /* implied.  By default, when nothing is specified, */
                  /* the number of choice sets is used when           */
                  /* options=relative is specified.  If you specify   */
                  /* rscale=r, where r is some number, then relative  */
                  /* D-efficiency equals: 100 times D-efficiency / r. */
                  /* If you want relative D-efficiency, and you know  */
                  /* that the number of choice sets is not the right  */
                  /* scaling factor (perhaps because you are not      */
                  /* making a generic design) and if you know the     */
                  /* D-efficiency of an optimal design, you can       */
                  /* specify it to get relative D-efficiency.         */
                  /* There are several short-cut specifications       */
                  /* available for you to use:                        */
                  /*                                                  */
                  /* RSCALE=, (with OPTIONS=RELATIVE) or              */
                  /* RSCALE=GENERIC sets RSCALE= to the number of     */
                  /* choice sets.                                     */
                  /*                                                  */
                  /* RSCALE=PARTIAL=p OF q, (where p and q are        */
                  /* integers) is used in the partial-profile context */
                  /* (where p of q attributes vary) or in the case of */
                  /* a choice design with a constant alternative      */
                  /* (where p of q alternatives vary).  It sets       */
                  /* RSCALE= to: nsets * p / q.  For example, with 16 */
                  /* choice sets and 4 or 8 attributes varying,       */
                  /* RSCALE=16 * 4 / 8, or RSCALE=8.                  */
                  /*                                                  */
                  /* RSCALE=ALT sets RSCALE= to a value that works    */
                  /* for certain basic alternative-specific designs.  */
                  /* Assume there are p parmameters.  The RSCALE=     */
                  /* value is:                                        */
                  /* ((nsets ** (nalts - 1)) *                        */
                  /* ((nsets * ((nalts - 1) / (nalts ** 2))) **       */
                  /* (p - nalts + 1))) ** (1 / p)                     */
                  /*                                                  */
                  /* If a design has brand (or alternative label)     */
                  /* effects such that brand i always occurs in       */
                  /* alternative i, and all other effects are         */
                  /* alternative-specific, then there are nalts - 1   */
                  /* parameters with a maximum determinant of nsets,  */
                  /* and the remaining p - nalts + 1 have one of      */
                  /* nalts alternatives contributing information to   */
                  /* each set, and nalts - 1 of nalts alternatives    */
                  /* contribute information so the maximum            */
                  /* D-efficiency is (nsets * ((nalts - 1) / (nalts   */
                  /* ** 2))) for the p - nalts - 1 parameters in that */
                  /* part of the design.  This formula will not       */
                  /* provide a true maximum for more complicated      */
                  /* designs such as designs with constant            */
                  /* alternatives.                                    */
                  /*                                                  */
chunks=,          /* specifies the number of observations to process  */
                  /* at one time with the coding step and PROC        */
                  /* TRANSREG.  By default, the entire data set is    */
                  /* processed at once.  You can specify a value, say */
                  /* 1/2 or 1/3 of the number of choice sets times    */
                  /* the number of alternatives to break up the       */
                  /* coding into smaller chunks if you run out of     */
                  /* memory.  Ideally, make the value a multiple of   */
                  /* the number of choice sets.  Be sure that you do  */
                  /* not leave one or a few extra observations in the */
                  /* last chunk, particularly if you are using one of */
                  /* the orthogonal codings (e.g.  STA) or you will   */
                  /* get an error.  Usually, you will not need to     */
                  /* specify this option.                             */
                  /*                                                  */
seed=0,           /* specifies the random number generation seed,     */
                  /* 0 - clock time.                                  */
                  /*--------------------------------------------------*/
restrictions=,    /* name of restrictions macro.  By default, there   */
                  /* are no restrictions.  When a restrictions macro  */
                  /* is specified, then RESVARS= must be specified as */
                  /* well.                                            */
                  /*                                                  */
resvars=,         /* variables for restrictions.  These variables     */
                  /* must all be numeric.  All restrictions must be   */
                  /* posed in terms of the values of these variables  */
                  /* along with the following:                        */
                  /* nsets  - number of choice sets in the design     */
                  /* nalts  - number of alternatives in the design    */
                  /* setnum - number of current choice set            */
                  /* altnum - number of current alternative (only     */
                  /*          available with alternative swapping)    */
                  /*                                                  */
                  /* The resvars variables are available for your     */
                  /* restrictions macro in the following matrices:    */
                  /* xmat   - the current design                      */
                  /* x      - the choice set that is being considered */
                  /*          for the setnum position in xmat         */
                  /* The first RESVARS= variable is in the first      */
                  /* column of x and xmat (x[,1] and xmat[,1]), ...,  */
                  /* and the last RESVARS= variable is in the last    */
                  /* column of x and xmat (x[,m] and xmat[,m]) where  */
                  /* m = ncol(xmat).                                  */
);                /*--------------------------------------------------*/
                  /* This macro creates three global variables:       */
                  /*                                                  */
                  /*     BEST - the number of the most efficient      */
                  /* design, which corresponds to the DESIGN variable */
                  /* in the output data sets.                         */
                  /*                                                  */
                  /*  _TRGIND - list of coded variables.              */
                  /*                                                  */
                  /* _TRGINDN - number of coded variables.            */
                  /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global best _trgind _trgindn _mktver;
%local  abort added adj altswap bestcovopts bestoutopts betalist
        betazero coded codeopt covopts detail fixedflg force i initdata
        j jj k localver maxlen ncands ninits nmores nodups notes nparms
        nvars orthcan outopts printseed ps relative resmac savedsnf
        savenote saveopts setnote setswap sing standorth termlen tests
        time timeopt makeflags outputall beststar;

%let time      = %sysfunc(datetime());
%let timeopt   = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote  = %sysfunc(getoption(notes));
%let saveopts  = &savenote missing="%sysfunc(getoption(missing))"
                 %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                 %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                 %sysfunc(getoption(serror));
%let setnote   = nonotes;
%let printseed = 1;
%let abort     = 0;    /* Abort when 1.                                   */
%let sing      = 1e-8; /* Singularity criterion, nearly zero.             */
%let added     = 0;    /* With INIT=, alternatives added to candidates.   */
%let maxlen    = 32700;
%let termlen   = 500;
%let localver  = 25Jul2010;
%let _mktver   = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES) %then %let setnote = notes;
   %if %index(&k, VERSION) %then %put ChoicEff macro version &localver.;
   %let printseed = (%index(&k, NOSEED) eq 0);
   %end;
%if %index(%upcase(%nrbquote(x &options x)), %str( NOTES ))
   %then %let setnote = notes;
options serror;

options &setnote generic missing='.';

%if %nrbquote(&help) ne %then %do; %mkthelp(choiceff) %goto endit; %end;

*------allow expressions in RSCALE=------;
data _null_;
   length c $ 32;
   c = lowcase(scan(symget('rscale'), 1, ' ='));
   call symput('jj', put((c in ('alt', 'generic', 'partial', ' ')), 1.));
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1;

%if not &jj %then %do;
   data _null_;
      length str $ 32 rscale 8;
      rscale = &rscale;
      if nmiss(rscale) then do;
         str = symget('rscale');
         put 'ERROR: RSCALE=' str 'is invalid.';
         call symput('abort', '1');
         end;
      call symput('rscale', put(rscale, best20.));

      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1;
   %end;

*------basic parameter checking------;
data _null_;
   ok = 1;
   length __beta $ &maxlen str $ &termlen;
   __beta = lowcase(left(symget('beta')));
   call symput('nparms', '0');
   call symput('betalist', put(__beta ne ' '   , 1.));
   call symput('betazero', put(__beta eq 'zero', 1.));

   *------STANDORTH specified in early release?------;
   __beta = lowcase(symget('model'));
   call symput('standorth', put(index(__beta, 'sta') and
               index(__beta, 'cla') and index(__beta, '/') and
               &sysver lt 9.2, 1.));

   *------restrictions?------;
   str = symget('restrictions');
   call symput('resmac', put(str ne ' ', 1.0));
   if str ne ' ' and symget('resvars') eq ' ' then do;
      put 'ERROR: RESVARS= option must be specified with RESTRICTIONS='
          str +(-1) '.';
      ok = 0;
      end;
   else if str eq ' ' and symget('resvars') ne ' ' then do;
      put 'ERROR: RESTRICTIONS= option must be specified with RESVARS='
          +(-1) '.';
      ok = 0;
      end;

   *------check OPTIONS=------;
   str = lowcase(symget('options'));
   i = index(str, 'relative'); if i then substr(str, i, 8) = ' ';
   call symput('relative', put(i gt 0, 1.));
   i = index(str, 'coded'); if i then substr(str, i, 5) = ' ';
   call symput('coded', put(i gt 0, 1.));
   i = index(str, 'detail'); if i then substr(str, i, 6) = ' ';
   j = index(str, 'resrep'); if j then substr(str, j, 6) = ' ';
   call symput('detail', put(i or j, 1.));
   i = index(str, 'notest'); if i then substr(str, i, 7) = ' ';
   call symput('tests', put(i eq 0, 1.));
   i = index(str, 'orthcan'); if i then substr(str, i, 7) = ' ';
   call symput('orthcan', put(i gt 0, 1.));
   i = index(str, 'nocode'); if i then substr(str, i, 6) = ' ';
   call symput('codeopt', put(i eq 0, 1.));
   i = index(str, 'nodup'); if i then substr(str, i, 6) = ' ';
   call symput('nodups', put(i gt 0, 1.));
   i = index(str, 'outputall'); if i then substr(str, i, 10) = ' ';
   call symput('outputall', put(i gt 0, 1.));
   i = index(str, 'nobeststar'); if i then substr(str, i, 10) = ' ';
   call symput('beststar', put(i eq 0, 1.));
   i = index(str, 'notes'); if i then substr(str, i, 5) = ' ';
   if i then
      put 'WARNING: OPTIONS=NOTES is obsolete.' /
          +9 'Submit "%let mktopts = notes;" before running the macro.';
   str = compbl(str);

   if str ne ' ' then do;
      put 'ERROR: OPTIONS=' str 'is invalid.';
      ok = 0;
      end;

   i = input(getoption('linesize'), ?? 12.);
   if i < 65 && symget('detail') eq '1' then i = 65;
   call symput('adj', put((i - 46) / 2, 8.0 -L));

   str = symget('nsets');
   nsets = input(str, ?? 32.);
   if nmiss(nsets) or nsets < 1 or nsets ne round(nsets) then do;
      put 'ERROR: NSETS=' str 'is invalid.';
      ok = 0;
      end;

   str = symget('nalts');
   if str ne ' ' then do;
      nalts = input(str, ?? 32.);
      if nmiss(nalts) or nalts < 2 or nalts ne round(nalts) then do;
         put 'ERROR: NALTS=' str 'is invalid.';
         ok = 0;
         end;
      end;

   *------rscale= parsing------;
   okr = 1;
   str = lowcase(symget('rscale'));
   if str ne ' ' then call symput('relative', '1');
   if str eq ' ' or str eq 'generic' then str = symget('nsets');
   else if str eq 'alt'              then str = 'a';
   else if str =: 'partial' then do;
      str = compress(left(substr(str,8)), '=');
      i = input(scan(str, 1, ' '), ?? 32.);
      j = input(scan(str, 3, ' '), ?? 32.);
      if n(i,j) eq 2 and i eq round(i) and j eq round(j) and
         scan(str, 2, ' =') eq 'of' then do;
         str = put(nsets * i / j, best16.);
         end;
      else do;
         str = symget('rscale');
         put 'ERROR: RSCALE=' str 'is invalid.';
         okr = 0;
         end;
      end;
   else if nmiss(input(str, ?? 32.)) then do;
      put 'ERROR: RSCALE=' str 'is invalid.';
      okr = 0;
      end;

   missing a;
   j = input(str, ?? 32.);
   if okr and j ne .A and (nmiss(j) or j < 0) then do;
      put 'ERROR: RSCALE=' str 'is invalid.';
      ok = 0;
      end;
   call symput('rscale', put(j, best16.));
   if okr = 0 then ok = 0;

   *------chunks=------;
   str = symget('chunks');
   if str ne ' ' then do;
      j = input(str, ?? 32.);
      if nmiss(j) or j < 2 or j ne round(j) then do;
         put 'ERROR: CHUNKS=' str 'is invalid.';
         ok = 0;
         end;
      end;

   str = symget('n');
   j = input(str, ?? 32.);
   if nmiss(j) or j < 1 then do;
      put 'ERROR: N=' str 'is invalid.';
      ok = 0;
      end;

   str = symget('iter');
   if str ne symget('maxiter') and
      str ne '2' then call symput('maxiter', trim(str));

   str = symget('maxiter');
   j = input(str, ?? 32.);
   if nmiss(j) or j < 1 or j ne round(j) then do;
      put 'ERROR: MAXITER=' str 'is invalid.';
      ok = 0;
      end;
   if j le 100 then call symput('outputall', '1');

   str = symget('intiter');
   j = input(str, ?? 32.);
   if nmiss(j) or j < 0 or j ne round(j) then do;
      put 'ERROR: INTITER=' str 'is invalid.';
      ok = 0;
      end;

   str = symget('converge');
   j = input(str, ?? 32.);
   if nmiss(j) or j < 0 then do;
      put 'ERROR: CONVERGE=' str 'is invalid.';
      ok = 0;
      end;

   *------parse and process data set names------;
   call symput('initdata', put(symget('init') ne ' ', 1.));

   str = symget('out');
   if str = ' ' then str = '__tmpcout';
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('out', substr(str, 1, j - 1));
   if str ne ' ' then call symput('outopts', substr(str, j));

   str = symget('bestout');
   if str = ' ' then str = 'best';
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('bestout', substr(str, 1, j - 1));
   if str ne ' ' then call symput('bestoutopts', substr(str, j));

   str = symget('cov');
   if str = ' ' then str = '__tmpccov';
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('cov', substr(str, 1, j - 1));
   if str ne ' ' then call symput('covopts', substr(str, j));

   str = symget('bestcov');
   if str = ' ' then str = '__tmpcbestcov';
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('bestcov', substr(str, 1, j - 1));
   if str ne ' ' then call symput('bestcovopts', substr(str, j));

   if (sysprod('iml') + sysprod('stat')) ne 2 then do;
      put 'ERROR: SAS/STAT and SAS/IML must be licensed '
          'to run this macro.';
      ok = 0;
      end;

   *------do we need to make a flag variables?------;
   i = max(input(symget('flags'), ?? 32.), 0);
   call symput('makeflags', trim(put(i, 12. -L)));

   if not ok or _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------copy input and initial data, apply any data set options------;
data __tmpcindat;
   set &data;
   %if &makeflags %then %do;
      if _n_ = 1 then call symput('flags', "_f1-_f&makeflags");
      %do i = 1 %to &makeflags; _f&i = 1; %end;
      %end;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
%let data = __tmpcindat;

%if &initdata %then %do;
   data __tmpcinitdat; set &init; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %let init = __tmpcinitdat;
   %end;

*------mention data sets, if they are invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&cov) ne %then %do;
   data _null_; set &cov; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&bestout) ne %then %do;
   data _null_; set &bestout; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&bestcov) ne %then %do;
   data _null_; set &bestcov; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )data=&data;
   %put %str(       )out=&out;
   %put %str(       )cov=&cov;
   %put %str(       )bestout=&bestout;
   %put %str(       )bestcov=&bestcov;
   %put %str(       )init=&init;
   %end;

%if &abort %then %goto endit;

data _null_;
   if 0 then set &data nobs=ncands;
   call symput('ncands', compress(put(ncands, 12.)));
   if _error_ then call symput('abort', '1');
   stop;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------count titles and footnotes------;
proc sql; create view __tmpctf as select * from dictionary.titles; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get effective page size for printing------;
data _null_;
   if _n_ = 1 then do;
      ps = input(getoption('pagesize'), ?? 23.);
      call symput('ps', compress(put(ps - 1, 12.)));
      end;
   set __tmpctf end=eof;
   titles + type eq 'T';
   foots  + type eq 'F';
   if eof then do;
      ps = input(getoption('pagesize'), ?? 23.);
      ps = ps - titles - foots - (titles gt 0) - (foots gt 0) - 1;
      if ps < 16 then ps = 1e10;
      call symput('ps', compress(put(ps, 12.)));
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data tmp_type(keep=i);
   ok = 1;
   length setstr $ &termlen typevar $ 66;

   if symget('model') = ' ' and &codeopt then do;
      put 'ERROR: MODEL= must be specified.';
      ok = 0;
      end;

   setstr  = symget('types');
   typevar = symget('typevar');
   if setstr ne ' ' then do;
      do i = 1 to &nsets until(nmiss(j));
         j = input(scan(setstr, i, ' '), ?? 32.);
         if n(j) then do m = 1 to j; output; end;
         sum = sum(sum, j);
         end;
      if sum ne input(symget('nsets'), ?? 32.) then do;
         put "ERROR: TYPES= values to not sum to NSETS=%nrbquote(&nsets).";
         ok = 0;
         end;
      end;
   if not ((setstr eq ' ' and typevar eq ' ') or
           (setstr ne ' ' and typevar ne ' ')) then do;
      put "ERROR: If either TYPES= or TYPEVAR= is specified, both must be.";
      ok = 0;
      end;
   call symput('types', put(setstr ne ' ', 1.));

   f1 = symget('flags') ne ' '; call symput('altswap', put(f1, 1.));
   f2 = symget('nalts') ne ' '; call symput('setswap', put(f2, 1.));
   if (f1 + f2) ne 1 then do;
      put 'ERROR: FLAGS= or NALTS= but not both must be specified.';
      ok = 0;
      end;

   if f1 and setstr ne ' ' then do;
      put "ERROR: TYPES= and TYPEVAR= are only valid with set swapping.";
      ok = 0;
      end;

   %if %nrbquote(&nalts) ne %then %do;
      ncands = &ncands / &nalts;
      if ncands ne round(ncands) then do;
         put 'ERROR: Number of alternatives problem:' / +7
             "%nrbquote(&nalts) alternatives and " ncands 'candidate sets.';
         ok = 0;
         end;
      %end;

   *------check INITVARS=------;
   if symget('initvars') ne ' ' and symget('init') eq ' ' then do;
      put 'ERROR: You must specify INIT= when you specify INITVARS=.';
      ok = 0;
      end;

   *------check FIXED=------;
   f1 = symget('fixed') ne ' ';
   f2 = &initdata and (symget('initvars') ne ' ');
   call symput('fixedflg', put(f1 and f2, 1.));
   if f1 and not f2 then do;
      put 'ERROR: When FIXED= is specified, both INIT= and INITVARS= must'
             / +7 'also be specified.';
      ok = 0;
      end;

   *------count MOREVARS=------;
   %if %nrbquote(&morevars) ne %then %do;
      array __m[*] &morevars;
      call symput('nmores', compress(put(dim(__m), 12.)));
      %end;

   *------check SEED=------;
   seed = input(symget('seed'), ?? 32.);
   if nmiss(seed) or seed ne round(seed) then do;
      put "ERROR: SEED=%nrbquote(&seed) is invalid.";
      ok = 0;
      end;

   *------special case random number seeds------;
   else if symget('seed') = '00' then call symput('seed', '0');

   *------run generically, with constant seed,
          eliminates using a clock generated seed------;
   else if seed eq 0 and (not index(symget('saveopts'), 'NOGENERIC')) and
      getoption('stimer' ) eq 'NOSTIMER' then call symput('seed', '072555');

   *------set seed using clock time when negative------;
   else if seed le 0 then do;
      seed = ceil(1e7 * uniform(input(reverse(put(1e2 * datetime(), 13.0)),
                                      ?? 8.)));
      call symput('seed', compress(put(seed, 8.)));
      if &printseed then put 'Seed = ' seed;
      end;

   if _error_ or not ok then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------process initial INIT= data sets------;
%if &initdata and %nrbquote(&initvars) ne %then %do;
   %if &altswap %then %let initvars = &initvars &flags;

   *------count initial variables, observations------;
   data _null_;
      if 0 then set &init(keep=&initvars) nobs=ninits;
      array __x[*] &initvars;
      %if &altswap %then %do; array __f[*] &flags; %end;
      call symput('nvars',  compress(put(dim(__x), 12.)));
      call symput('ninits', compress(put(ninits, 12.)));

      nalts = %if &altswap %then dim(__f); %else &nalts;;
      if ninits / nalts ne &nsets then do;
         ninits = ninits / nalts;
         put "ERROR: Expecting &nsets initial choice sets." /
             '       There are ' ninits 'initial choice sets.';
         call symput('abort', '1');
         end;

      if _error_ then call symput('abort', '1');
      stop;
      run;

   %if &syserr > 4 %then %do;
      %if %nrbquote(&initvars) ne %then %do;
         %let i = You may want to use INIT=initdata(KEEP=SET) approach instead.;
         %put WARNING: The problem may be due to character vars in INITVARS=.;
         %put %str(        ) &i;
         %end;
      %let abort = 1;
      %goto endit;
      %end;

   %if &fixedflg and not &altswap %then %do;
      data _null_;
         set &init(keep=&fixed);
         retain __which;
         if mod(_n_, &nalts) = 1 then __which = &fixed;
         if not (&fixed in (1, 0, .) and __which eq &fixed) then do;
            put 'ERROR: Problem in the INIT= data set at observation '
                _n_ +(-1) '.';
            if __which ne &fixed then
               put +7 "%nrbquote(&fixed)=" &fixed
                   "must equal %nrbquote(&fixed)=" __which +(-1) '.';
            if not (&fixed in (1, 0, .)) then
               put +7 "%nrbquote(&fixed)=" &fixed
                   'must be: 0, 1, or . (missing).';
            put 'ERROR: The FIXED= variable must contain blocks of'
                ' NALTS= values,' / +7
                'and all values within each block must be the same.' / +7
                "Each block must consist of %nrbquote(&nalts) "
                'missings, zeros, or ones.';
            call symput('abort', '1');
            stop;
            end;
         run;
      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   *------process initial data set------;
   data __tmpcinit(keep=__index)
        __tmpcdata(drop=__: %if &fixedflg %then &fixed;)
        %if &fixedflg %then __tmpcfixed(keep=&fixed);;

      array __x[&nvars] &initvars;
      array __t[&nvars] __t1-__t&nvars;

      *------copy original candidate set------;
      if _n_ = 1 then do;
         do __obs = 1 to &ncands;
            set &data point=__obs;
            output __tmpcdata;
            end;
         end;

      %if &altswap %then %do;

         array __f[*] &flags;

         *------find matches between alternatives and candidates------;
         set &init(keep=&initvars &fixed &weight);

         %if &fixedflg %then %do;
            if nmiss(&fixed) then do;
               __index = .;
               output __tmpcinit __tmpcfixed;
               return;
               end;
            %end;

         do __i = 1 to &nvars; __t[__i] = __x[__i]; end;
         do __i = 1 to &ncands until(not __differ);
            __k + 1; if __k > &ncands then __k = 1;
            set &data point=__k;
            do __j = 1 to &nvars until(__differ);
               __differ = (__t[__j] ne __x[__j]);
               end;
            if not __differ then do;
               __index = __k;
               output __tmpcinit %if &fixedflg %then __tmpcfixed;;
               end;
            end;

         *------augment candidates if necessary------;
         if __differ then do;
            do __i = 1 to &nvars;   __x[__i] = __t[__i]; end;
            do __i = 1 to dim(__f); __f[__i] = 0;        end;
            output __tmpcdata;
            __index = __obs;
            output __tmpcinit %if &fixedflg %then __tmpcfixed;;
            __obs + 1; __added + 1;
            call symput('added', compress(put(__added, best6.)));
            end;
         %end;

      %else /* set swapping */ %do;

         *------find matches between sets and candidates------;
         __obs = 1 + (__obs - 1) / &nalts;
         do __i = 1 to %eval(&ninits / &nalts);
            do __c = 1 to &ncands / &nalts
               until(__match %if &fixedflg %then or nmiss(&fixed););
               __count = 0;
               do __a = 1 to &nalts;
                  __p = (__i - 1) * &nalts + __a;

                  set &init(keep=&initvars &fixed &weight) point=__p;

                  %if &fixedflg %then %do;
                     if nmiss(&fixed) then do;
                        __index = .; __match = 0;
                        output __tmpcinit __tmpcfixed;
                        end;
                     %end;

                  %if &fixedflg %then else do;;
                     do __j = 1 to &nvars; __t[__j] = __x[__j]; end;
                     __k + 1; if __k > &ncands then __k = 1;
                     set &data(keep=&initvars) point=__k;
                     if __count = (__a - 1) * &nvars
                        then do __j = 1 to &nvars;
                        __count + (__t[__j] eq __x[__j]);
                        end;
                     __match = (__count eq %eval(&nvars*&nalts));
                     %if &fixedflg %then end;;

                  if __match then do __j = 1 to &nalts;
                     __index = floor((__k - 1) / &nalts) + 1;
                     output __tmpcinit %if &fixedflg %then __tmpcfixed;;
                     end;
                  end;
               end;

            *------augment candidates if necessary------;
            if not __match %if &fixedflg %then and n(&fixed); then do;
               do __a = 1 to &nalts;
                  __p = (__i - 1) * &nalts + __a;
                  set &init(keep=&initvars &fixed &weight) point=__p;
                  output __tmpcdata;
                  __index = __obs;
                  output __tmpcinit %if &fixedflg %then __tmpcfixed;;
                  end;
                __obs + 1; __added + &nalts;
               call symput('added', compress(put(__added, best6.)));
               end;
            end;
         stop;
         %end;

      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------updated data sets, warn if candidates were augmented------;
   %let init = __tmpcinit;
   %let data = __tmpcdata;
   %if &added %then
      %put WARNING: The candidate set was augmented &added times.;
   %end;

*------code the candidate set------;
%if &codeopt %then %do;
   %let standorth = %eval(&standorth + 1);
   proc transreg design%if &chunks ne %then =&chunks; data=&data
                 nozeroconstant norestoremissing;
      model &model;
      output out=tmp_cand;
      %if %nrbquote(&morevars) ne or
          %nrbquote(&weight) ne %then %do; id &morevars &weight; %end;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %let standorth = %eval(&standorth - 1);
   %if &coded %then %do; proc print label; run; %end;
   %end;

%if %nrbquote(&morevars) ne %then %do;
   %let _trgind  = &_trgind &morevars;
   %let _trgindn = %eval(&_trgindn + &nmores);
   %end;

%if &resmac %then %do;

   *------copy of the vars for restrictions, in order------;
   data __tmpcresdat(keep=&resvars);
      length &resvars 8;
      set &data;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------check restrictions macro for syntax errors------;
   filename __dummy dummy;
   proc printto log=__dummy; run;
   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc iml;
      call symput('jj', '0');
      use __tmpcresdat; read all into rawcand;
      %if &altswap %then %do;
         use &data(keep=&flags); read all into flags;
         nalts = ncol(flags);
         %end;
      %else %do; nalts = &nalts; %end;
      _nalts = nalts;
      nsets  = &nsets;
      nattrs = ncol(rawcand);
      setnum = 1;
      altnum = 1;
      xmat   = j(nsets # nalts, nattrs, 1);
      x      = xmat[1:nalts,];
      start badness(x) global(xmat, %if &altswap %then altnum,;
                              setnum, nalts, nsets);
         bad = 0; %&restrictions; return(bad);
         finish;
      bad = badness(x);
      if setnum ^= 1 | altnum ^= 1 | ^all(x = 1) | ^all(xmat = 1) |
         nsets ^= &nsets | nalts ^= _nalts then call symput('jj', '1');
      quit;
   %if &syserr > 4 %then %let abort = 1;

   proc printto log=log; run;
   %if &syserr > 4 %then %let abort = 2;

   %if &jj %then %do;
       %let i = setnum, altnum, nsets, nalts, x, or xmat.;
       %put WARNING: Restrictions macro is changing &i.;
       %put %str(        ) This might be a serious problem.  Check your macro.;
       %end;

   %if &abort eq 1 %then %do;
      %let j = %sysfunc(getoption(mprint));
      %put ERROR: Invalid RESTRICTIONS=&restrictions macro.;
      %put %str(      ) Here is your macro embedded in a simple test program.;
      options mprint;
      proc iml;
         use __tmpcresdat; read all into rawcand;
         %if &altswap %then %do;
            use &data(keep=&flags); read all into flags;
            nalts = ncol(flags);
            %end;
         %else %do; nalts = &nalts; %end;
         nsets  = &nsets;
         nattrs = ncol(rawcand);
         setnum = .;
         altnum = .;
         xmat   = j(nsets # nalts, nattrs, 1);
         x      = xmat[1:nalts,];
         start badness(x) global(xmat, %if &altswap %then altnum,;
                                 setnum, nalts, nsets);
            bad = 0; %&restrictions; return(bad);
            finish;
         bad = badness(x);
         quit;
      options &j;
      %end;

   %if &abort %then %goto endit;
   %end;

*------get variable names------;
proc contents data=tmp_cand(keep=&_trgind drop=&drop)
   noprint out=tmp_con;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc sort; by varnum; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------drop variables if necessary------;
%if %nrbquote(&drop) ne %then %do;
   data _null_;
       set tmp_con;
       n = put(_n_, 5.);
       call symput(compress('__v' || n), compress(name));
       call symput('_trgindn', compress(n));
       run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %macro _vlist; %do i = 1 %to &_trgindn; &&__v&i %end; %mend;
   %let _trgind = %_vlist;
   %end;

*------expand zero variable list------;
%if &betazero %then %do;
   %macro _zlist; %do i = 1 %to &_trgindn; 0 %end; %mend;
   %let beta = %_zlist;
   %end;

*------print coding summary------;
data tmp_con;
   set tmp_con(keep=name label) end=eof;
   n = _n_;
   %if &betalist %then %do;
      array __beta[%eval(&_trgindn+1)] _temporary_ (&beta .E);
      b = __beta[_n_];
      if b = .E then b = .;
      retain ok 1;
      output;

      if eof then do;
         e = __beta[&_trgindn+1];
         ok = (e eq .E);
         if _error_ or nmiss(b) or not ok then do;
            call symput('abort', '1');
            name = ' ';
            if n(e) then do;
               label = "ERROR: Extra beta(s).";
               put label;
               n + 1;
               b = e;
               output;
               end;
            else if nmiss(b) then do;
               label = "ERROR: Not enough betas.";
               put label;
               output;
               end;
            end;
         end;
      %end;
   %else %do; b = .; output; %end;

   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1;

proc print uniform label;
   id n name;
   var b label;
   label n = 'n' name = 'Name' b = 'Beta' label = 'Label';
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%if not &betalist %then %goto endit;

%if &initdata and not &fixedflg %then %let maxiter = 1;

*-----forcing points out due to constraints?-----;
%let force = %eval(&types or &nodups or &resmac);

*------begin main algorithm------;
proc iml;

   *------index vector for on choice set------;
   start ind(i) global(nalts); return((i-1)#nalts+1:i#nalts); finish;

   *------compute probabilities, probability center, and update x------;
   start center(x, p) global(nalts);
      do i = 1 to nrow(x) / nalts;
         k = ind(i);
         q = p[k,];
         q = q / sum(q);
         z = x[k,];
         x[k,] = (z - j(nalts, 1, 1) * q` * z) # sqrt(q);
         end;
      finish;

   *------print iteration history------;
   start printit(header, iter) global(desnum, fineff, err, lnum, maxeffp);
      if header | lnum + 3 > &ps then do;
         if lnum + 13 > &ps then do; print / ' '; lnum = 2; end;
         else do; put //; lnum = lnum + 3; end;
         lnum = lnum + 2;
         put +&adj 'Design   Iteration  D-Efficiency       D-Error' /
             +&adj '----------------------------------------------';
         end;
      if header then put +&adj desnum 6. @;
      else put +&adj +6 @;
      put iter  8. +7 fineff d11. @;
      lnum = lnum + 1;
      if fineff > maxeffp & &beststar then do;
         maxeffp = fineff;
         put ' * ' err d11.;
         end;
      else put +3 err d11.;
      finish;

   *------restrictions macro------;
   %if &resmac %then %do;
      nsets = &nsets;
      start badness(x) global(xmat, %if &altswap %then altnum,;
                              setnum, nalts, nsets);
         bad = 0; %&restrictions; return(bad);
         finish;

      *------raw candidates for restrictions------;
      use __tmpcresdat; read all into rawcand;
      %end;

   file print;
   do; /* enable goto's */

   *------read candidate set------;
   use tmp_cand(keep=&_trgind); read all into cand;

   anyforce = 0;
   ncands = nrow(cand);   /* number of candidates                     */
   npars  = ncol(cand);   /* number of parameters                     */
   np     = 1 / npars;    /* efficiency exponent                      */
   lnum   = -2;

   *-----alternative flags-------;
   %if &altswap %then %do;
      use &data(keep=&flags); read all into flags;
      nalts = ncol(flags);/* number of alternatives                   */
      call symput('nalts', char(nalts, 12, 0));
      %end;
   %else %do;
      nalts = &nalts; ncands = (ncands - &added) / nalts;
      if &nodups && ncands < &nsets then do;
         put 'ERROR: Too few candidates with NODUPS.';
         call symput('abort', '1');
         goto done;
         end;
      %end;

   *------read type variable------;
   if &types then do;
      use tmp_type; read all into types;
      use &data(keep=&typevar); read all into candtypes;
      candtypes = shape(candtypes, ncands, nalts)[,1];
      end;

   *------zero missings, compute utilities, orthogonalize------;
   cand  = cand # (cand ^= .);
   utils = exp(cand * {&beta}`);

   %if &orthcan %then %do;
      cand = j(nrow(cand), 1, 1) || cand;
      cand = (cand * sweep(root(cand` * cand)))[,2:(1+npars)];
      %end;

   imat = i(nalts);          /* identity matrix                       */
   subvec = j(nalts, 2, 1);  /* matrix of index vector indices        */
   nobs = &nsets # nalts;    /* rows in output design matrix          */
   changed = {1 1};          /* indicates where design last changed   */
   fuzz = i(npars) # &sing;  /* X`X ridge factor, avoid singularity   */

   *------weight if necessary------;
   %if %nrbquote(&weight) ne %then %do;
      use tmp_cand(keep=&weight); read all into weight;
      if any(shape(weight, nrow(weight) / nalts, nalts)[,+] = 0) then do;
         put 'ERROR: Invalid weights.';
         put '       Some alternatives in each choice set must have weight.';
         call symput('abort', '1');
         goto done;
         end;

      utils = utils # weight;
      %end;

   *------indexes into candidate set for each alternative------;
   %if &altswap %then %do;
      do i = 1 to nalts;
         j = loc(flags[,i]);
         subvec[i, 2] = ncol(j);
         if i > 1 then do;
            subvec[i, 1] = subvec[i-1, 2] + 1;
            subvec[i, 2] = subvec[i, 2] + subvec[i-1, 2];
            end;
         bigindv = bigindv || j;
         end;
      %end;

   sets = (0:(&nsets-1)) * nalts; /* index vector component           */
   ones = j(1, nalts, 1);         /* vector of ones                   */
   indvec = j(1, nobs, 0);        /* main index vector into design    */
   des = j(nobs, npars, 0);       /* initial empty design             */

   *------set up for set swapping------;
   if &setswap then do;
      run center(cand, utils);    /* center candidates                */
      dif = ((-nalts + 1) : 0);   /* index vector components          */
      con = j(1, nalts, nalts);
      adj = j(1, &nsets, 1) @ dif;
      end;

   *------set up output data sets------;
   create &cov var({"Design" "Effic" &_trgind});
   create &out var({"Design" "Efficiency" %if &force %then "Dups";
                    "Index" "Set" "Prob"});
   maxeffo = 0;    maxeffp = 0;   /* p - printing, o - output         */

   *------create &maxiter designs------;
   do desnum = 1 to &maxiter;

      *------create a random initial design------;
      %if &setswap %then %do;
         indvec = ceil(uniform(j(1, &nsets, &seed))#ncands);
         %end;

      %else %do;
         do j = 1 to nalts;
            anind = bigindv[,subvec[j,1]:subvec[j,2]];
            indvec[sets+j] = anind[ceil(uniform(j(1, &nsets, &seed))
                             * ncol(anind))];
            end;
         %end;

      *------read initial design if specified------;
      %if &initdata %then %do;
         use &init; read all into ind;
         if ncol(ind) > 1 then do;
            put // 'ERROR: Either the INIT= data set must have one'
                ' variable or' / +7 'INITVARS= must be specified.' //;
            call symput('abort', '1');
            goto done;
            end;

         %if &setswap %then %do;
            n = nrow(ind);
            if n / &nsets ^= round(n / &nsets) then do;
               put 'ERROR: INIT= problem: '
                   (compress(char(n))) " rows and %nrbquote(&nsets) sets.";
               call symput('abort', '1');
               goto done;
               end;
            ind = shape(ind, &nsets)[,1]`;
            %end;
         %else %do;
            ind = ind`;
            %end;

         anind = loc(ind ^= .);
         if nrow(anind) then indvec[anind] = ind[anind];
         %end;

      *------pull initial design out of candidate set------;
      %if &setswap %then %do;
         des = cand[indvec @ con + adj,];
         %if &resmac %then %do; xmat = rawcand[indvec @ con + adj,]; %end;
         %end;
      %else %do;
         des = cand[indvec,];
         run center(des, utils[indvec,]);
         %if &resmac %then %do; xmat = rawcand[indvec,]; %end;
         %end;

      *------flag the alternatives that are free to vary------;
      %if &fixedflg %then %do;
         fixed = j(1, ncol(indvec), 0);
         use __tmpcfixed; read all into ind;
         if ncol(ind) > 1 then do;
            put 'ERROR: The FIXED= data set must have one variable.';
            call symput('abort', '1');
            goto done;
            end;
         ind = %if &setswap %then shape(ind, nalts)[,1];
               %else ind;`;
         fixed[1 : ncol(ind)] = ind;
         %end;

      *------compute, report initial determinant, efficiency------;
      xpx    = (des` * des);    currdet = det(xpx); currdet = max(0, currdet);
      maxdet = currdet;         oldeff = currdet ## np;
      fineff = &n # oldeff;
      if fineff <= 0 then err = .; else err = 1 / fineff;

      run printit(1, 0);

      *------internal iterations, improve current design------;
      do iter = 1 to &intiter
         %if &force %then %do;
            until(convrgd & ^anyforce);
            anyforce = 0;
            %end;
         %else %do; until(convrgd); %end;

         *------process each alternative/set------;
         do desi = 1 to %if &altswap %then nobs; %else &nsets;
            until(convrgd %if &force %then & ^anyforce;);

            %if &fixedflg %then if ^ fixed[desi] then do;;

               *------store and remove current alternative/set------;
               ind = %if &setswap %then desi # con + dif;
                     %else ind(ceil(desi / nalts));;

               besttry = des[ind,];
               %if &resmac %then %do;
                  rawbesttry = xmat[ind,];
                  xset = rawbesttry;
                  %end;
               des[ind,] = 0;

               *------compute determinant, if 0 then ridge X`X------;
               xpx = des` * des; d = det(xpx); d = max(d, 0);
               do i = 1 to 100 while(d ## np <= &sing);
                  xpx = xpx + (i # i) * fuzz;
                  d   = det(xpx); d = max(d, 0);
                  end;

               xpxinv = inv(xpx);

               *------set up indices for alternative swapping------;
               %if &altswap %then %do;
                  indcan = indvec[,ind];
                  altnum = mod(desi - 1, nalts) + 1;
                  anind  = bigindv[,subvec[altnum,1]:subvec[altnum,2]];
                  %if &resmac %then %do;
                     setnum = floor((desi - 1) / nalts) + 1;
                     %end;
                  %if &nodups %then %do;
                     sorted = indcan;
                     sorted[rank(indcan)] = indcan;
                     indcheck = shape(indvec, &nsets);
                     do i = 1 to &nsets;
                        indcheck[i,rank(indcheck[i,])] = indcheck[i,];
                        end;
                     %end;
                  %end;
               %else %if &resmac %then %do; setnum = desi; %end;

               %if &force %then %do; force = 0; %end;
               %if &types %then %do;  /* force wrong type out? */
                  if candtypes[indvec[desi]] ^= types[desi] then force = 1;
                  %end;
               %if &nodups %then %do; /* force duplicates out? */
                  %if &setswap %then %do;
                     if sum(indvec[desi] = indvec) > 1 then force = 1;
                     %end;
                  %else %do;
                     if sum((indcheck = repeat(sorted, &nsets))[,+] = nalts) > 1
                        then force = 1;
                     if sum(sorted = indvec[desi]) > 1 then force = 2;
                     * if force = 2 then put 'outside' desi;
                     %end;
                  %end;

               *------try swapping in each candidate------;
               do candi = 1 to  %if &setswap %then ncands;
                                %else ncol(anind);;

                  *------set up index vector to pull out candidate------;
                  %if &altswap %then %do;
                     indcan[,altnum] = anind[,candi];
                     %end;

                  *------ensure no duplicate choice sets------;
                  %if &nodups %then %do;
                     %if &setswap %then %do;
                        if sum(candi = indvec) = 0 then do;
                        %end;
                     %else %do;
                        sorted[rank(indcan)] = indcan;
                        if ^any((indcheck = repeat(sorted, &nsets))[,+] = nalts)
                           | force = 2 then do;
                        %end;
                     %end;

                  %if &types %then %do;
                     if candtypes[candi] = types[desi] then do;
                     %end;

                  *------pull out candidate------;
                  tryit = cand[%if &altswap %then indcan;
                               %else candi#con+dif;,];
                  %if &resmac %then %do;
                     xset = rawcand[%if &altswap %then indcan;
                                    %else candi#con+dif;,];
                     %end;

                  *------center if necessary------;
                  %if &altswap %then %do;
                     run center(tryit, utils[indcan,]);
                     %end;

                  *------update determinant using this candidate------;
                  deter = det(imat + tryit * xpxinv * tryit`);
                  deter = max(deter, 0);
                  currdet = d * deter;

                  %if &resmac %then %do;
                     bad1     = badness(xset);
                     bad2     = badness(rawbesttry);
                     force    = (bad1 < bad2) | force;
                     oktoswap = bad1 <= bad2;
                     if bad2 then anyforce = 1;
                     %end;

                  *------keep change if determinant is bigger------;
                  if %if &resmac %then
                        (currdet >= maxdet | force) & oktoswap;
                     %else %do;
                        (currdet > maxdet
                        %if &force %then | force;)
                        %end;
                     then do;
                     %if &force %then %do;
                        if force then do; anyforce = 1; force = 0; end;
                        %end;
                     maxdet = currdet;  /* best determinant so far */
                     indvec[desi] = %if &altswap %then anind[,candi];
                                    %else candi;;
                     %if &nodups and &altswap %then %do;
                        indcheck[ceil(desi / nalts),] = sorted;
                        if sum(sorted = indvec[desi]) > 1 then force = 2;
                        %end;
                     besttry = tryit;
                     %if &resmac %then %do; rawbesttry = xset; %end;

                     changed = iter || desi;

                     %if &detail %then %do;
                        put +%eval(&adj-10) "at "
                           %if &setswap %then desi 4. +4;
                           %else (floor((desi - 1) / nalts) + 1) 4.
                                 (  mod((desi - 1),  nalts) + 1) 4.;
                           " swapped in" candi 6. +3 (currdet##np) d11.
                           %if &resmac %then " bad =" bad2 best8.;;
                        %end;
                     end;
                  %if &nodups %then %do; end; %end;
                  %if &types  %then %do; end; %end;
                  end;

               *------store best alternative/set------;
               des[ind,] = besttry;
               %if &resmac %then %do; xmat[ind,] = rawbesttry; %end;
               %if &fixedflg %then end;;

            convrgd = (iter > changed[1] & desi > changed[2]);
            end;

         *------evaluate efficiency, convergence, report results------;
         neweff  = maxdet ## np; /* newest efficiency */
         convrgd = (convrgd | (((neweff - oldeff) /
                                max(oldeff, &sing)) < &converge));

         oldeff = neweff; xpx = (des` * des);
         fineff = det(xpx);
         fineff = max(fineff, 0) ## np;
         fineff = &n # fineff;
         if fineff <= 0 then err = .; else err = 1 / fineff;
         run printit(0, iter);
         if fineff = 0 then put +%eval(&adj+21) neweff d11. ' (Ridged)';
         end;

      dups = 0;
      %if &force %then %do;
         if anyforce then do;
            put / "WARNING: Design " (compress(char(desnum, 6)))
                " has TYPES=, OPTIONS=NODUP, or restrictions problems.";
            dups = 1;
            end;
         %end;

      %if &setswap %then %do;

         *------Check Desgign for TYPES= OPTIONS=NODUP------;
         %if &force %then %do;
            flagit = 0;
            do i = 1 to &nsets;
               %if &types %then %do;
                  if types[i] ^= candtypes[indvec[i]] then flagit = 1;
                  %end;
               %if &nodups %then %do;
                  if sum(indvec[i] = indvec) > 1 then flagit = 1;
                  %end;
               end;
            if flagit then do;
               put / "WARNING: Design " (compress(char(desnum, 6)))
                   " has TYPES= or OPTIONS=NODUP problems.";
               dups = 1;
               end;
            %end;

         *------indices for output data set------;
         set = indvec @ ones;
         indvec = indvec @ con + adj;
         %end;

      %else %do; set = (1:&nsets) @ ones; %end;

      *------choice probabilities------;
      p = utils[indvec,];
      do i = 1 to &nsets; k = ind(i); p[k,] = p[k,] / sum(p[k,]); end;

      *------store final design indices, covariances------;
      if fineff > maxeffo | &outputall then do;
         if fineff > maxeffo then maxeffo = fineff;
         final = (shape(desnum || fineff  %if &force %then || dups;,
                                          nobs, 2 + &force) || indvec` || set` || p);
         cov   = (shape(desnum || fineff, npars, 2) || sweep(xpx / nobs) / (nobs # &n));
         setout &cov;   append from cov;
         setout &out;   append from final;
         end;
      end;

   done: end;
   quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------create final designs, put most efficient first------;
data &out;
   set &out; i = index; n = _n_;
   set &data point=i;
   set &out; n = _n_; /* get back any vars wiped out by &data */
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc sort data=&out;
   by %if &force %then dups; descending efficiency design n;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------number of most efficient design------;
data _null_;
   set &out;
   call symput('best', trim(put(design, 8. -L)));
   if _error_ then call symput('abort', '1');
   stop;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get row names, variable names right in cov matrix------;
data &cov;
   length Design Effic 8 __name $ 32 &_trgind 8;
   label __name = 'Name';
   array __vars[*] &_trgind;
   call vname(__vars[mod(_n_ - 1, &_trgindn) + 1], __name);
   set &cov;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------store best covariance matrix, diagonal, tests------;
data &bestcov;
   if 0 then set tmp_cand(keep=&_trgind);
   set &cov(where=(design = &best));

   array __vars[*] &_trgind;
   array __betas[%eval(&_trgindn)] _temporary_ (&beta);

   length __label $ 40;
   call label(__vars[_n_], __label);
   __beta = __betas[_n_]; __diag = __vars[_n_]; __n + 1;
   if __diag = 0 then do;
      __diag = .; __df = 0; __diag0 + 1;
      if __diag0 = 1 then put // 'Redundant Variables:' /;
      put __name @@;
      end;
   else if __diag < 0 then
      put // 'ERROR: Numerical errors in inverse.  Increase the number of'
          / +7 'choice sets, or delete redundant terms.' //;
   else do;
      __stderr = sqrt(__diag); __df = 1; __nparms + 1;
      __wald   = __beta / __stderr; __waldc = __wald * __wald;
      __prob   = max(1 - probchi(__waldc, 1), 0.0001);
      call symput('nparms', put(__nparms, 12.));
      end;
   keep  __diag __df __stderr __beta __wald __waldc
         __prob __label __name __n &_trgind;
   label __name = 'Variable Name' __label  = 'Label' __df = 'DF'
         __beta = 'Assumed Beta' __diag = 'Variance' __n = 'n'
         __stderr = 'Standard Error' __wald = 'Wald'
         __waldc = 'Wald Chi-Square' __prob = 'Prob > Squared Wald';
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------report on most efficient design------;
data _null_;
   file print;
   set &out;
   m = (input(getoption('linesize'), ?? 12.) - 25) / 2;
   if efficiency > 0 then err = (1 / efficiency); else err = .;
   nsets  = &nsets;   nalts = &nalts;
   rscale = %if &rscale eq A %then
            ((nsets ** (nalts - 1)) *
            ((nsets * ((nalts - 1) / (nalts ** 2))) **
            (&_trgindn - nalts + 1))) ** (1 / &_trgindn);
            %else &rscale;;
   call symput('rscale', put(rscale, best16.));

   releff   = 100 * efficiency / rscale;
   invnsets = 1 / nsets;
   maxparm  = nsets * (nalts - 1); nparms = &nparms;
   put / +m '      Final Results' /
       / +m 'Design       '  design       best12.
       / +m 'Choice Sets  '  nsets        best12.
       / +m 'Alternatives '  nalts        best12.
       / +m 'Parameters   '  nparms       best12.
       / +m 'Maximum Parameters'  maxparm best7.
       / +m 'D-Efficiency    ' efficiency d9.
       %if &relative %then
       / +m 'Relative D-Eff  ' releff     d9.;
       / +m 'D-Error         ' err        d9.
       %if &relative %then
       / +m '1 / Choice Sets ' invnsets   d9.;;
   %if &force %then %do;
      if dups then do;
         if      &nodups and not (&types  or &resmac)
            then put +m 'WARNING: Duplicates.';
         else if &resmac and not (&types  or &nodups)
            then put +m 'WARNING: Restriction violations.';
         else if &types  and not (&resmac or &nodups)
            then put +m 'WARNING: Type violations.';
            else put +m 'WARNING: Violations.';
         end;
      %end;
   if _error_ then call symput('abort', '1');
   stop;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------report tests------;
%if &tests %then %do;
   proc print uniform label data=&bestcov;
      id __n __name __label;
      var __diag
          %if not &betazero %then __beta;
          __df __stderr
          %if not &betazero %then __wald __prob;;
      format __prob 6.4;
      sum __df;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------optionally report on submatrix efficiency------;
%if %nrbquote(&submat) ne %then %do;
   proc iml; reset fuzz; file print;
      use &bestcov; read all into c[rowname=__name];
      c = c[&submat, &submat];
      err = det(c);
      err = max(err, 0) ## (1 / ncol(c));
      if err <= 0 then do; err = .; eff = 0; end;
      else eff = 1 / err;
      releff = 100 * eff / &rscale;
      print "Submatrix Report";
      print c[rowname=__name colname=__name];
      put / +17 'Efficiency     = ' eff d9.
          / +17 'D-Error        = ' err d9.
          %if &relative %then
          / +17 'Relative D-Eff = ' releff d9.;;
      quit;
   %end;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------clean up temporary data sets------;
proc datasets nolist;
   delete __tmpcinitdat __tmpcdata __tmpcfixed __tmpcinit __tmpcindat
          __tmpcout __tmpccov __tmpcbestcov __tmpctf(memtype=view)
          __tmpcresdat;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------print number of obs message------;
options &savenote no&timeopt;

data &out &outopts; set &out; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data &cov &covopts; set &cov; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data &bestcov &bestcovopts; set &bestcov; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data &bestout &bestoutopts; set &out(where=(design = &best)); run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%endit: options &saveopts;
quit;

%if %nrbquote(&help) eq %then
   %if not &betalist %then
      %put NOTE: The CHOICEFF macro ended due to no BETA= specification.;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then %do;
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
      %if &standorth eq 2 %then %do;
         %let i = is not available in this release.;
         %put NOTE: The STA or STANDORTH option &i;
         %put %str(      )It is first available in SAS 9.2.;
         %end;
      %end;
   %put ERROR: The CHOICEFF macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The CHOICEFF macro used &time seconds.;
   %end;

%mendit: %mend;
