 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktAllo                                             */
 /*   TITLE: Process a Discrete Choice Allocation Study Data Set */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs, choice modeling        */
 /*   PROCS:                                                     */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: The following SAS products are required to run      */
 /*          this macro: BASE.                                   */
 /*                                                              */
 /*          This macro creates global macro variables: _MktVer  */
 /****************************************************************/

 /*---------------------------------------------------------------------

This macro is used for manipulating data for a choice experiment.
It takes as input a data set with one row for each alternative of each
choice set.  For example in a study with 10 brands plus a constant
alternative and 27 choice sets, there are 27 * 11 = 297 observations.
Here is an example input data set.  It contains a choice set variable,
product attributes (brand and price) and a frequency variable (count)
that contains the total number of times that each alternative was
chosen.

                    Obs    Set     Brand      Price    Count

                      1     1                             0
                      2     1     Brand  1     $50      103
                      3     1     Brand  2     $75       58
                      4     1     Brand  3     $50      318
                      5     1     Brand  4    $100       99
                      6     1     Brand  5    $100       54
                      7     1     Brand  6    $100       83
                      8     1     Brand  7     $75       71
                      9     1     Brand  8     $75       58
                     10     1     Brand  9     $75      100
                     11     1     Brand 10     $50       56
                      .
                      .
                      .
                    296    27     Brand  9    $100       94
                    297    27     Brand 10     $50       65

The end result is a data set with twice as many observations that
contains the number of times each alternative was chosen and the number
of times it was not chosen.  This data set also contains a variable c
with values 1 for first choice and 2 for second or subsequent choice.

                    Obs    Set     Brand      Price   Count    c

                      1     1                             0    1
                      2     1                          1000    2
                      3     1     Brand  1     $50      103    1
                      4     1     Brand  1     $50      897    2
                      5     1     Brand  2     $75       58    1
                      6     1     Brand  2     $75      942    2
                      7     1     Brand  3     $50      318    1
                      8     1     Brand  3     $50      682    2
                      .
                      .
                      .
                    593    27     Brand 10     $50       65    1
                    594    27     Brand 10     $50      935    2

Here is a sample usage:

data allocs2;
   input Brand $ 1-8 Price $ Count Set;
   datalines;
.              .         0 1
Brand  1     $50       103 1
Brand  2     $75        58 1
Brand  3     $50       318 1
Brand  4    $100        99 1
Brand  5    $100        54 1
Brand  6    $100        83 1
Brand  7     $75        71 1
Brand  8     $75        58 1
Brand  9     $75       100 1
Brand 10     $50        56 1
.              .        10 2
Brand  1    $100        73 2
Brand  2    $100        76 2
Brand  3    $100       342 2
Brand  4     $50        55 2
Brand  5     $75        50 2
Brand  6    $100        77 2
Brand  7     $75        95 2
Brand  8    $100        71 2
Brand  9     $50        72 2
Brand 10    $100        79 2
;

%mktallo(data=allocs2, out=allocs3, nalts=11,
         vars=set brand price, freq=Count)

The option DATA= names the input data set, OUT= names the output data
set, NALTS= specifies the number of alternatives, VARS= names the
variables in the data set that will be used in the analysis excluding
the FREQ= variable, and FREQ= names the frequency variable.

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

                  /*--------------------------------------------------*/
%macro mktallo(   /* This macro is used for processing data from an   */
                  /* allocation study.                                */
                  /*--------------------------------------------------*/
help,             /* nonblank, list options                           */
                  /*--------------------------------------------------*/
data=_last_,      /* The DATA= option names the input SAS data set.   */
                  /*--------------------------------------------------*/
out=allocs,       /* The OUT= option names the output SAS data set.   */
                  /*--------------------------------------------------*/
nalts=,           /* The NALTS= option names the number of            */
                  /* alternatives (including if appropriate the       */
                  /* constant alternative).  This option must be      */
                  /* specified.                                       */
                  /*--------------------------------------------------*/
freq=,            /* The FREQ= option names the frequency variable -  */
                  /* the number of times this alternative was chosen. */
                  /* This option must be specified.                   */
                  /*--------------------------------------------------*/
vars=,            /* The VARS= option names the variables in the data */
                  /* set that will be used in the analysis but not    */
                  /* the FREQ= variable.  This option must be         */
                  /* specified.                                       */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _mktver;
%local  abort savedsnf savenote saveopts setnote time localver k timeopt;

%let time     = %sysfunc(datetime());
%let timeopt  = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote = %sysfunc(getoption(notes));
%let saveopts = &savenote missing="%sysfunc(getoption(missing))"
                %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                %sysfunc(getoption(serror));
%let setnote  = nonotes;
%let abort    = 0;
%let localver = 25Jul2010;
%let _mktver  = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES)   %then %let setnote = notes;
   %if %index(&k, VERSION) %then %put MktAllo macro version &localver.;
   %end;
options serror;

options &setnote generic missing='.';

%if %nrbquote(&help) ne %then %do; %mkthelp(mktallo) %goto endit; %end;

*------rudimentary error checking------;
data _null_;
   n = input(symget('nalts'), ?? 32.);
   if nmiss(n) or n <= 0 then do;
      error + 1;
      put 'ERROR: NALTS= must be specified.';
      end;
   if symget('freq') = ' ' then do;
      error + 1;
      put 'ERROR: FREQ= must be specified.';
      end;
   if symget('vars') = ' ' then do;
      error + 1;
      put 'ERROR: VARS= must be specified.';
      end;
   if error or _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&data) ne %then %do;
   data _null_; set &data; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )data=&data;
   %put %str(       )out=&out;
   %end;

%if &abort %then %goto endit;

*------process the allocation data set------;
data &out;

   retain __printed 0;
   * For each alternative in the data set, read how often it was
     chosen.  Keep only the count.;

   set &data(keep=&freq);

   * Create the index of the first alternative in this choice set.;

   __k = floor((_n_ - 1) / &nalts) * &nalts + 1;

   * Create __b, the alternative number within the choice set.;

   __b = _n_ - __k + 1;

   * Now we need to loop over all of the alternatives in this choice
     set.  This means reading the data set with a second set statement,
     to get the design variables.;

   do __j = 1 to &nalts;
      __was = _error_;
      set &data(keep=&vars) point=__k nobs=__n;
      if _error_ and not __was then do;
         _error_ = 0;
         if not __printed then do;
            put "ERROR: Probable error in NALTS=%nrbquote(&nalts).";
            if __k > __n then put +7
               "Accessing DATA=%nrbquote(&data) observation "
               __k 'out of ' __n +(-1) '.';
            __printed = 1;
            call symput('abort', '1');
            end;
         end;
      c = 2 - (__b eq __j);
      output;
      __k + 1;
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

options &savenote no&timeopt;

*------aggregate the results------;
proc summary data=&out nway missing;
   class &vars c;
   output sum(&freq)=&freq out=&out(drop=_type_ _freq_);
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTALLO macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTALLO macro used &time seconds.;
   %end;

%mendit: %mend;
