 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktBal                                              */
 /*   TITLE: Create a Perfectly Balanced Experimental Design     */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs                         */
 /*   PROCS: OPTEX                                               */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: This macro creates global macro variables           */
 /*             _MKTLIST _MKTVER                                 */
 /*          and calls the following autocall macros, which must */
 /*          be installed:                                       */
 /*             MktRuns MktDes                                   */
 /*                                                              */
 /*          The following SAS products are required to run      */
 /*          this macro: BASE, QC.                               */
 /****************************************************************/

 /*---------------------------------------------------------------------

This macro creates linear experimental designs using an algorithm that
ensures that the design is perfectly balanced, or in the case when the
number of levels of a factor does not divide the number of runs, as
close to perfectly balanced as possible.  In contrast, the MktDes and
MktEx macros will usually find nearly balanced but not perfectly
balanced designs.  However, designs created with the MktBal macro may
sometimes be less efficient overall, and for large problems, this macro
can be slow.  This macro has several options that can make it run faster
for large problems, but at a price of usually slightly decreased
D-efficiency.

This is not a full-featured experimental design generator.  For example,
you cannot specify interactions that you want to estimate or specify
restrictions such as which levels may or may not appear together.
You must use the MktEx macro for that.  This macro builds a design
by creating a balanced first factor, optimally (or nearly optimally)
blocking it to create the second factor, then blocking the first two
factors to create the third, and so on.  Then once it creates all
factors, it refines each factor.  Each factor is in turn removed from
the design, and the design is reblocked, replacing the initial factor
if the new design is more D-efficient.

Here is a simple example of creating a design with 2 two-level factors
and 3 three-level factors in 18 runs.  The MktEval macro evaluates the
results.  This design is in fact optimal.

   %mktbal(2 2 3 3 3, n=18, seed=151)
   %mkteval;

In all cases, the factors are named x1, x2, x3, ... and so on.

This call creates a design with 10 two-level factors and 10 three-level
factors.  The specification OPTIONS=PROGRESS tells the macro to print a
message every time it starts on a new factor.

   %mktbal(2**10 3**10, n=36, options=progress)
   %mkteval;

This next problem, at 120 runs and with factor levels greater than 5, is
starting to get big and hence, by default, will run slowly.  You can use
the MAXSTARTS= options to make the macro run more quickly.  For example,
the second call below runs much faster than the first.

   %mktbal(2 3 4 5 6 7 8 9 10, n=120, options=progress, seed=17)
   %mktbal(2 3 4 5 6 7 8 9 10, n=120, options=progress, seed=17,
           maxstarts=1, maxiter=1)

------------------------------------------------------------------------

                           References

See http://support.sas.com/techsup/technote/ts723.html and
http://support.sas.com/techsup/technote/ts723.pdf for the reference
list.

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;


%macro mktbal(    /*--------------------------------------------------*/
                  /*                                                  */
list,             /* List of numbers of levels of all the factors.    */
                  /* Lists of numbers, like 2 2 3 3 4 4 or a          */
                  /* "levels**number of factors" syntax like: 2**2    */
                  /* 3**2 4**2 or both can be combined: 2 2 3**4 5 6  */
                  /*                                                  */
                  /* ? or help - lists options                        */
                  /*                                                  */
n=,               /* Number of runs in the final design.              */
                  /*                                                  */
out=Design,       /* Output experimental design.                      */
                  /*                                                  */
init=,            /* Initial design.  You can specify just the first  */
                  /* few columns of the design, in this data set with */
                  /* column names x1, x2, ..., and these columns will */
                  /* never change.  This is different from the init=  */
                  /* data set in MktEx.  It is often the case that    */
                  /* the first few columns can be constructed         */
                  /* combinatorially, and it is known that the        */
                  /* optimal design will just add new columns to the  */
                  /* initial few orthogonal columns.  This option     */
                  /* makes that process more efficient.  By default,  */
                  /* when INIT= is not specified, all columns are     */
                  /* iteratively found and improved.                  */
                  /*                                                  */
                  /* These next few options control how much time the */
                  /* macro spends searching for better designs.       */
                  /*                                                  */
maxiter=5,        /* specifies the number of times to try to make a   */
                  /* balanced design.  By default, the macro tries 5  */
                  /* times.  You should probably specify a smaller    */
                  /* value for large problems.   This is the outer    */
                  /* most set of iterations.                          */
iter=5,           /* Alias for MAXITER=.                              */
                  /*                                                  */
maxstarts=10,     /* Maximum number of random starts for each factor. */
                  /* This is the number of iterations that PROC OPTEX */
                  /* performs for each factor.  With larger values,   */
                  /* the macro tends to find slightly better designs  */
                  /* at a cost of slower run times.                   */
                  /*                                                  */
maxinititer=,     /* Maximum number of random starts for each factor  */
                  /* during the initialization stage.  This is the    */
                  /* number of iterations that PROC OPTEX performs    */
                  /* for each factor during the initialization.  With */
                  /* larger values, the macro tends to find slightly  */
                  /* better designs at a cost of slower run times.    */
                  /* The default is the value of MAXSTARTS=.          */
                  /*                                                  */
maxtries=10,      /* Maximum number of times to try refining each     */
                  /* factor after the initialization stage.           */
                  /* Increasing the value of this option usually has  */
                  /* little effect since the macro stops refining     */
                  /* each design when the efficiency stabilizes.      */
                  /*                                                  */
seed=0,           /* specifies the random number generation seed,     */
                  /* 0 - clock time.                                  */
                  /*                                                  */
options=,         /* progress - report on progress.  For large        */
                  /* numbers of factors, a large number or runs, or   */
                  /* when the number of levels is large, this macro   */
                  /* is slow.  OPTIONS=PROGRESS gives you information */
                  /* about which step is being executed.              */
                  /*                                                  */
                  /* nosort - do not sort the final design.           */
                  /*                                                  */
                  /* sequential - only sequentially add factors to    */
                  /* the design.  Do not go back and refine them.     */
                  /*                                                  */
                  /* oa - specifies that an orthogonal array is       */
                  /* sought.  Factors are added sequentially and they */
                  /* are acceptable only when D-efficiency reaches    */
                  /* 100.  This is basically the algorithm that Xu    */
                  /* (2002) used to find new OAs and Kuhfeld (2005)   */
                  /* used to find new difference schemes.             */
                  /*                                                  */
                  /* noprint - do not print the final D-efficiency.   */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _mktver;
%local  t i m abort savenote setnote time iternum oplist dcrit try
        lastchange _optseed maxdcrit bestdcrit progress savedsnf print
        printseed localver k firstcol nextcol dosort saveopts sequential
        oa eps100 showcode timeopt;

%let time      = %sysfunc(datetime());
%let timeopt   = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote  = %sysfunc(getoption(notes));
%let saveopts  = &savenote missing="%sysfunc(getoption(missing))"
                 %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                 %sysfunc(getoption(spool))
                 %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                 %sysfunc(getoption(serror));
%let setnote   = nonotes;
%let showcode  = nocode;
%let printseed = 1;
%let abort     = 0;
%let localver  = 25Jul2010;
%let _mktver   = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES)   %then %let setnote = notes;
   %if %index(&k, VERSION) %then %put MktBal macro version &localver.;
   %if %index(&k, CODE)    %then %let showcode = ;
   %let printseed = (%index(&k, NOSEED) eq 0);
   %end;
options serror;

options &setnote generic missing='.' nospool;
ods noresults;

data _null_;
   length list $ 32;
   list = left(lowcase(symget('list')));
   if list eq 'help' or list eq '?' then call symput('abort', '-1');
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort = 1 %then %goto endit;

%if &abort = -1 %then %do; %let abort = 0; %mkthelp(mktbal) %goto endit; %end;

data _null_;
   length list $ 2100 w $ 24 ss $ 2;

   abort = 0;
   call symput('maxdcrit' , '0');
   call symput('bestdcrit', '0');
   call symput('firstcol',  '1');
   call symput('eps100',  compress(put(100 - 1e-8, 15.10)));

   *------check SEED=------;
   seed = input(symget('seed'), ?? 23.);
   if nmiss(seed) or seed ne round(seed) then do;
      put "ERROR: SEED=%nrbquote(&seed) is invalid.";
      abort = 1;
      end;

   *------special case random number seeds------;
   else if symget('seed') eq '00' then call symput('seed', '0');

   *------run generically, with constant seed,
          eliminates using a clock generated seed------;
   else if seed = 0 and (not index(symget('saveopts'), 'NOGENERIC')) and
      getoption('stimer' ) eq 'NOSTIMER' then call symput('seed', '072555');

   *------set seed using clock time when negative------;
   else if seed le 0 then do;
      seed = ceil(1e7 * uniform(input(reverse(put(1e2 * datetime(), 13.0)),
                                      ?? 8.)));
      call symput('seed', compress(put(seed, 8.)));
      if &printseed then put 'Seed = ' seed;
      end;

   *------check OPTIONS=------;
   list = lowcase(symget('options'));
   i = index(list, 'progress'); if i then substr(list, i, 8) = ' ';
   call symput('progress', put(i gt 0, 1.));
   i = index(list, 'noprint'); if i then substr(list, i, 7) = ' ';
   call symput('print',  put(i eq 0, 1.));
   list = compbl(list);
   i = index(list, 'nosort'); if i then substr(list, i, 6) = ' ';
   call symput('dosort', put(i eq 0, 1.));
   j = index(list, 'oa'); if j then substr(list, j, 2) = ' ';
   call symput('oa', put(j gt 0, 1.));
   i = index(list, 'sequential'); if i then substr(list, i, 10) = ' ';
   call symput('sequential', put(i gt 0, 1.));
   if i or j then call symput('maxtries', '0');

   if list ne ' ' then do;
      put 'ERROR: OPTIONS=' list 'is invalid.';
      abort = 1;
      end;

   *------check N=------;
   n = input(symget('n'), ?? 32.);
   if nmiss(n) or n < 2 or n ne round(n) then do;
      put "ERROR: N=%nrbquote(&n) is invalid.";
      abort = 1;
      end;

   *------check MAXITER=------;
   w = symget('iter');
   if w ne symget('maxiter') and w ne '5' then call symput('maxiter', trim(w));
   maxiter = input(symget('maxiter'), ?? 32.);
   if nmiss(maxiter) or maxiter < 1 or maxiter ne round(maxiter) then do;
      w = symget('maxiter');
      put 'ERROR: MAXITER=' w 'is invalid.';
      abort = 1;
      end;

   *------check MAXSTARTS=------;
   maxstarts = input(symget('maxstarts'), ?? 32.);
   if nmiss(maxstarts) or maxstarts < 1 or
      maxstarts ne round(maxstarts) then do;
      w = symget('maxstarts');
      put 'ERROR: MAXSTARTS=' w 'is invalid.';
      abort = 1;
      end;

   *------check MAXINITITER=------;
   maxinititer = input(symget('maxinititer'), ?? 32.);
   if nmiss(maxinititer) then do;
      w = symget('maxstarts');
      maxinititer = input(w, ?? 32.);
      call symput('maxinititer', compress(w));
      end;
   if nmiss(maxinititer) or maxinititer < 1 or
      maxinititer ne round(maxinititer) then do;
      w = symget('maxinititer');
      put 'ERROR: MAXINITITER=' w 'is invalid.';
      abort = 1;
      end;

   *------check MAXTRIES=------;
   maxtries = input(symget('maxtries'), ?? 32.);
   if nmiss(maxtries) or maxtries < 0 or maxtries ne round(maxtries) then do;
      put "ERROR: MAXTRIES=%nrbquote(&maxtries) is invalid.";
      abort = 1;
      end;

   if (sysprod('qc') + sysprod('stat')) ne 2 then do;
      put 'ERROR: SAS/STAT and SAS/QC must be licensed to run this macro.';
      abort = 1;
      end;

   %if %nrbquote(&init) ne %then %do;
      if 0 then set &init;
      array x[*] x:;
      call symput('firstcol', compress(put(dim(x) + 1, 5.)));
      %end;

   endit: if _error_ or abort then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------expand out list to _mktlist, a list of integers------;
%let _mktver = obsolete;
%mktruns(&list, options=justparse);

%if &_mktver ne &localver %then %do;
   %put ERROR: Incompatible MktRuns macro.;
   %put %str(      ) You must have the same version of all macros.;
   %put %str(      ) Versions: MktBal = &localver, MktRuns = &_mktver..;
   %put %str(      ) Get the latest tools and documentation from:;
   %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
   %let abort = 1;
   %end;

%if &syserr > 4 or %nrbquote(&_mktlist) eq ABORT %then %let abort = 1;
%if &abort %then %goto endit;

*------count parameters, factors------;
data _null_;
   length list $ 5000;
   list = symget('_mktlist');
   parms = 1;

   do i = 1 to 1024 until(n eq .);
      n = input(scan(list, i), 12.);
      if n(n) then parms + n - 1;
      end;

   call symput('m', trim(put(i - 1, 12. -L)));

   if parms gt &n then do;
      put 'ERROR: More parameters than runs.';
      call symput('abort', '1');
      end;

   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data set, if it is invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %put ERROR: OUT=%nrbquote(&out) is invalid.;

%if &abort %then %goto endit;

%do iternum = 1 %to &maxiter;

   %if &firstcol eq 1 %then %do;
      %if &progress %then %do;
         data _null_;
            it = &iternum;
            put // it best3. ' Initializing Variable   1';
            run;
         %end;

      data __tmpbal1(keep=x1);
         do j = 1 to &n; x1 = mod(j, %scan(&_mktlist, 1)) + 1; output; end;
         if _error_ then call symput('abort', '1');
         run;
      %let nextcol = 2;
      %end;

   %else %do;

      %if &progress %then %do;
         data _null_;
            it = &iternum;
            put // it best3. " Initializing Variables  1 - %eval(&firstcol-1)";
            run;
         %end;

      data __tmpbal1;
         set &init;
         if _error_ then call symput('abort', '1');
         run;

      %let nextcol = &firstcol;
      %end;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %do i = &nextcol %to &m;

      %if &progress %then %do;
         data _null_;
            i = &i; it = &iternum;
            put it best3. ' Initializing Variable' i best4.;
            run;
         %end;

      data __tmpbal2;
         set __tmpbal1;
         x&i = mod(_n_, %scan(&_mktlist, &i)) + 1;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      %if &showcode ne nocode %then %do;
         %let k = %sysfunc(getoption(mprint));
         options mprint;
         %end;

      %let _optseed = 0;

      ods select none;
      proc optex data=__tmpbal2 %if &seed %then seed=&seed;;
         ods output BlockDesignEfficiencies=__tmpdeff;
         class x&i;
         model x&i;
         generate initdesign=__tmpbal2 method=sequential;
         block design=__tmpbal1 noexchange keep=1
               iter=&maxinititer init=random
               %if %sysevalf(&sysver > 9) %then target=99.9999999999;;
         class x1-x%eval(&i-1);
         model x1-x%eval(&i-1);
         output out=__tmpbal3;
         run; quit;
      ods select all;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      %if &showcode ne nocode %then %do;
         options &k;
         %let showcode = nocode;
         %end;

      data __tmpbal1(drop=__seed DCriterion);
         if _n_ eq 1 then do;
            length __seed $ 12;
            __seed = symget('_optseed');
            if __seed = '0' then __seed = put(1e8 * uniform(&seed), 9.0);
            call symput('seed', compress(__seed));
            set __tmpdeff(keep=DCriterion);
            call symput('dcrit', compress(put(DCriterion, 15.10)));
            end;

         set __tmpbal3;

         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %if &oa and %sysevalf(&dcrit lt &eps100) %then %do;
        %let bestdcrit = &dcrit;
        %goto next;
        %end;
      %end;

   ods select none;
   proc optex data=__tmpbal1;
      ods output Efficiencies=__tmpdeff;
      class x1-x&m;
      model x1-x&m;
      generate initdesign=__tmpbal1 method=sequential;
      run; quit;
   ods select all;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data _null_; set __tmpdeff;
      call symput('dcrit'   , compress(put(DCriterion, 15.10)));
      call symput('maxdcrit', compress(put(DCriterion, 15.10)));
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %if &progress %then %do;
      data _null_;
         it = &iternum; d = &dcrit; m = &maxdcrit; b = &bestdcrit;
         m = max(d, m); b = max(m, b);
         put +31 'Current      Max     Best' /
             it best3. ' Initial Design' +11 d 9.4 m 9.4 b 9.4 ' ';
         run;
      %end;

   %if %sysevalf(&dcrit > &bestdcrit) %then %do;
      data &out __tmpbal1;
         length x1-x&m 8;
         set __tmpbal1;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      %let maxdcrit  = &dcrit;
      %let bestdcrit = &dcrit;
      %end;

   %else %if %sysevalf(&dcrit > &maxdcrit) %then %do;
      data __tmpbal1;
         length x1-x&m 8;
         set __tmpbal1;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      %let maxdcrit = &dcrit;
      %end;

   %if %sysevalf(&bestdcrit ge &eps100) %then %goto done;

   %let try = 0; %let lastchange = 0;
   %do t = 1 %to &maxtries;
      %do i = &nextcol %to &m;
         %let try = %eval(&try + 1);

         data __tmpbal2;
            length x1-x&m 8;
            set __tmpbal1(drop=x&i);
            x&i = mod(_n_, %scan(&_mktlist, &i)) + 1;
            if _error_ then call symput('abort', '1');
            run;

         %if       &i = 1  %then %let oplist = x2-x&m;
         %else %if &i = &m %then %let oplist = x1-x%eval(&m - 1);
         %else                   %let oplist = x1-x%eval(&i-1) x%eval(&i+1)-x&m;

         %let _optseed = 0;

         proc optex data=__tmpbal2 noprint %if &seed %then seed=&seed;;
            class x&i;
            model x&i;
            generate initdesign=__tmpbal2 method=sequential;
            block design=__tmpbal1 noexchange keep=1
                  iter=&maxstarts init=random
                  %if %sysevalf(&sysver > 9) %then target=99.9999999999;;
            class &oplist;
            model &oplist;
            output out=__tmpbal3;
            run; quit;

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

         data __tmpbal2(drop=__seed);
            if _n_ eq 1 then do;
               length __seed $ 12;
               __seed = symget('_optseed');
               if __seed = '0' then __seed = put(1e8 * uniform(&seed), 9.0);
               call symput('seed', compress(__seed));
               end;

            set __tmpbal3;

            if _error_ then call symput('abort', '1');
            run;

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

         ods select none;
         proc optex data=__tmpbal2;
            ods output Efficiencies=__tmpdeff;
            class x1-x&m;
            model x1-x&m;
            generate initdesign=__tmpbal2 method=sequential;
            run; quit;
         ods select all;

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

         data _null_; set __tmpdeff;
            call symput('dcrit', compress(put(DCriterion, 15.10)));
            if _error_ then call symput('abort', '1');
            run;

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

         %if &progress %then %do;
            data _null_;
               i = &i; it = &iternum; d = &dcrit; m = &maxdcrit; b = &bestdcrit;
               if d gt m then s1 = put(d, 9.4);
               if d gt b then s2 = put(d, 9.4);
               put it best3. ' Refining     variable' i best4.
                   d 9.4 s1 $char9. s2 $char9.;
               run;
            %end;

         %if %sysevalf(&dcrit > &bestdcrit) %then %do;
            data &out __tmpbal1;
               length x1-x&m 8;
               set __tmpbal2;
               if _error_ then call symput('abort', '1');
               run;

            %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

            %let maxdcrit  = &dcrit;
            %let bestdcrit = &dcrit;
            %let lastchange = &try;
            %end;

         %else %if %sysevalf(&dcrit > &maxdcrit) %then %do;
            data __tmpbal1;
               length x1-x&m 8;
               set __tmpbal2;
               if _error_ then call symput('abort', '1');
               run;

            %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

            %let maxdcrit  = &dcrit;
            %let lastchange = &try;
            %end;

         %if %sysevalf(&bestdcrit ge &eps100) %then %goto done;
         %if &try ge &lastchange + &m %then %goto next;
         %end;
      %end;

   %next:;

   %end;

%done:

%if &oa and %sysevalf(&bestdcrit lt &eps100) %then %do;
   %put ERROR: Orthogonal array not found.;
   data &out; run;
   %let dosort = 0;
   %let abort = 1;
   %end;

%if &print and not &abort %then %do;
   %let _mktver = obsolete;
   %mktdes(cand=&out, factors=x1-x&m, options=check nocode)

   %if &_mktver ne &localver %then %do;
      %put ERROR: Incompatible MktDes macro.;
      %put %str(      ) You must have the same version of all macros.;
      %put %str(      ) Versions: MktBal = &localver, MktDes = &_mktver..;
      %put %str(      ) Get the latest tools and documentation from:;
      %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
      %let abort = 1;
      %end;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

proc datasets nolist;
   delete __tmpbal1 __tmpbal2 __tmpbal3 __tmpdeff
   run; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

options &savenote no&timeopt;

%if &dosort %then %do;
   proc sort data=&out; by x:; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%else %if not &abort %then %do;
   data &out; set &out; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%endit: options &saveopts;
ods results;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTBAL macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTBAL macro used &time seconds.;
   %end;

%mendit: %mend;
