 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktBIBD                                             */
 /*   TITLE: Balanced (or Unbalanced) Incomplete Block Design    */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs                         */
 /*   PROCS: OPTEX                                               */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: The following SAS products are required to run      */
 /*          this macro: BASE, IML, QC.                          */
 /*                                                              */
 /*          This macro creates global macro variables: _MktVer  */
 /****************************************************************/

 /*---------------------------------------------------------------------

The macro MktBIBD finds balanced incomplete block designs and
unbalanced block designs.  The parameters are:

b specifies the number of blocks.  In the partial profile context, this
  is the number of profiles.  In the MaxDiff context, this is the number
  of sets.

t specifies the number of treatments.  In the partial profile or MaxDiff
  setting, this is the total number of attributes or messages.

k specifies the block size, which is the number of treatments in each block.
  In the partial profile or MaxDiff setting, this is the number of attributes
  or messages shown at one time.

When r = b * k / t and l = r * (k - 1) / (t - 1) are integers,
and k = t and b >= t, then a complete block design may be possible.

When r = b * k / t and l = r * (k - 1) / (t - 1) are integers,
and k < t and b >= t, then a balanced incomplete block design may be
possible.

You can use the macro MktBSize to find parameters in which BIB
designs may exist.  This example asks for a design with 10 blocks, 5
treatments, and 3 treatments per block.

   %mktbibd(b=10, t=5, k=3, seed=104)

The first piece of output shows the block design efficiency criterion,
which is 100, and shows that a BIBD was found.

               Block Design Efficiency Criterion       100.0000
               Number of Treatments, t                        5
               Block Size, k                                  3
               Number of Blocks, b                           10
               Treatment Frequency                            6
               Pairwise Frequency                             3
               Total Sample Size                             30
               Positional Frequencies Optimized?            Yes

The parameters of the design are also printed along with the efficiency.
The next piece of output shows how often each treatment occurs with
every other treatment.

                       Treatment by Treatment Frequencies

                                   1  2  3  4  5

                                1  6  3  3  3  3
                                2     6  3  3  3
                                3        6  3  3
                                4           6  3
                                5              6

The diagonal shows that each treatment occurs 6 times in the design,
and every treatment occurs in a block with every other treatment 3
times.  When the off-diagonal entries are all constant, then a BIBD was
found and efficiency is 100%.  When one or more off-diagonal
frequencies are not constant, then the design is not a BIBD.

Next are the positional frequencies, showing in this case that each of
the five treatments occured in each of the three within-block positions
exactly twice.  These frequencies apply to the OUT=BIBD design.

                       Treatment by Position Frequencies

                                      1  2  3

                                   1  2  2  2
                                   2  2  2  2
                                   3  2  2  2
                                   4  2  2  2
                                   5  2  2  2

Next, the BIBD is shown.  This design is available in the OUT=BIBD
data set.

                                x1    x2    x3

                                 3     5     2
                                 1     5     2
                                 3     2     4
                                 4     3     1
                                 5     3     1
                                 1     2     3
                                 4     1     5
                                 5     4     3
                                 2     1     4
                                 2     4     5

You can view the OUTI=Incidence matrix as follows:

   proc print data=incidence; run;

                      Obs    b1    b2    b3    b4    b5

                        1     0     1     1     0     1
                        2     1     1     0     0     1
                        3     0     1     1     1     0
                        4     1     0     1     1     0
                        5     1     0     1     0     1
                        6     1     1     1     0     0
                        7     1     0     0     1     1
                        8     0     0     1     1     1
                        9     1     1     0     1     0
                       10     0     1     0     1     1

This shows that the first block consists of treatments 2, 3, and 5, just
as the OUT=BIBD design does, but in a different format.

You can view the BIBD arrayed as a block by treatment factorial design.

   proc print data=factorial; run;

                          Obs    Block    Treatment

                            1       1         3
                            2       1         5
                            3       1         2
                            4       2         1
                            5       2         5
                            6       2         2
                            7       3         3
                            8       3         2
                            9       3         4
                           10       4         4
                           11       4         3
                           12       4         1
                           13       5         5
                           14       5         3
                           15       5         1
                           16       6         1
                           17       6         2
                           18       6         3
                           19       7         4
                           20       7         1
                           21       7         5
                           22       8         5
                           23       8         4
                           24       8         3
                           25       9         2
                           26       9         1
                           27       9         4
                           28      10         2
                           29      10         4
                           30      10         5

You can view the OUTS=Sorted sorted design as follows:

   proc print data=sorted; run;

                            Obs    x1    x2    x3

                              1     1     2     3
                              2     1     5     2
                              3     2     1     4
                              4     2     4     5
                              5     3     2     4
                              6     3     5     2
                              7     4     1     5
                              8     4     3     1
                              9     5     3     1
                             10     5     4     3

Note that in this design, the treatments are sorted, so it
is typically the OUT=BIBD or OUTF=Factorial design that you want to
actually use.

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%macro mktbibd(   /*--------------------------------------------------*/
                  /*                                                  */
help,             /* nonblank, list options                           */
                  /*                                                  */
b=,               /* specifies the number of blocks.  In the partial  */
                  /* profile context, this is the number of profiles. */
                  /* In the MaxDiff context, this is the number of    */
                  /* sets.                                            */
nsets=,           /* alias for B= with a name more relevant to the    */
                  /* MaxDiff and partial-profile context.             */
                  /*                                                  */
t=,               /* specifies the number of treatments.  In the      */
                  /* partial profile or MaxDiff setting, this is      */
                  /* the total number of attributes or messages.      */
nattrs=,          /* alias for T= with a name more relevant to the    */
                  /* MaxDiff and partial-profile context.  When the   */
                  /* nattrs= option is specified, the output will use */
                  /* the word "Attribute" rather than "Treatment" and */
                  /* "Set" rather than "Block".  For example, it will */
                  /* print "Attribute by Attribute Frequencies"       */
                  /* rather than "Treatment by Treatment Frequencies".*/
                  /*                                                  */
k=,               /* specifies the block size, which is the number    */
                  /* of treatments in each block.  In the partial     */
                  /* profile or MaxDiff setting, this is the number   */
                  /* of attributes or messages shown at one time.     */
setsize=,         /* alias for K= with a name more relevant to the    */
                  /* MaxDiff and partial-profile context.             */
                  /*                                                  */
out=BIBD,         /* specifies the output data set name for the       */
                  /* b x k BIBD (or unbalanced block design).         */
                  /*                                                  */
outs=Sorted,      /* specifies the output data set name for the       */
                  /* b x k sorted design.                             */
                  /*                                                  */
outi=Incidence,   /* specifies the output data set name for the       */
                  /* b x t incidence matrix.                          */
                  /*                                                  */
outf=Factorial,   /* specifies the output data set name for the       */
                  /* bk x 2 factorial design matrix.                  */
                  /*                                                  */
group=,           /* specifies the number of groups into which the    */
                  /* design is to be divided.  This could be useful   */
                  /* with MaxDiff and partial profiles.  By default,  */
                  /* the design is not divided into groups.           */
                  /*                                                  */
seed=0,           /* specifies the random number generation seed,     */
                  /* 0 - clock time.                                  */
                  /*                                                  */
options=position, /* position - the default, optimizes position       */
                  /* frequencies.                                     */
                  /* neighbor - optimizes row-neighbor balance.       */
                  /* This is a nondirectional form of balance.  The   */
                  /* order of the treatments does not matter in       */
                  /* evaluating the row-neighbor balance.  That is,   */
                  /* treatment 1 appearing before treatment 2 is      */
                  /* counted as the same as 2 appearing before 1.     */
                  /* serial - optimizes directional row-neighbor      */
                  /* balance.  In contrast to neighbor, treatment 2   */
                  /* appearing before treatment 1 is treated as       */
                  /* different from 1 appearing before 2.             */
                  /*                                                  */
optiter=500 . .,  /* specifies the number of PROC OPTEX iterations.   */
                  /* If one number is specified, it is just the       */
                  /* number of iterations.  With two numbers, n1      */
                  /* and n2, n1 iterations are performed and then     */
                  /* efficiency is checked.  If the block design      */
                  /* efficiency criterion is 100, the iterations      */
                  /* stop.  Otherwise, this process is repeated up    */
                  /* to n2 times for a maximum of n1 * n2 iterations. */
                  /* The default for n2, when it is missing, is 1000  */
                  /* when the parameters conform to the necessary     */
                  /* conditions for a BIBD and 5 otherwise.  The      */
                  /* third value is the maximum amount of time in     */
                  /* minutes to spend in PROC OPTEX.  The default for */
                  /* the time value, when it is missing, is 5 minutes */
                  /* when the parameters conform to the necessary     */
                  /* conditions for a BIBD and 0.5 minutes otherwise. */
                  /*                                                  */
positer=200 5E3 2,/* specifies the number of iterations for the       */
                  /* algorithm that attempts to optimize the          */
                  /* treatment by position frequencies.  The first    */
                  /* value specifies the number of times to try to    */
                  /* refine the design.  The second value specifies   */
                  /* the number of times to start from scratch with   */
                  /* a different random start.  Larger values         */
                  /* increase the chances of finding better treatment */
                  /* by position frequencies at a cost of slower run  */
                  /* times.  The third value is the maximum amount    */
                  /* of time in minutes to spend optimizing the       */
                  /* positional frequencies.  Specify POSITER= or     */
                  /* POSITER=0 to just request a design without       */
                  /* optimizing position.                             */
                  /*                                                  */
weights=1 2,      /* weights for position balance and neighbor        */
                  /* balance.  Specify two numeric values.  The total */
                  /* badness is a weighted sum of the position        */
                  /* badness and the neighbor badness.                */
);                /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _mktver;
%local  abort savenote saveopts setnote time _optseed printseed localver
        tnum optrep i j eff gotone sysranend ispossible outopts outiopts
        outsopts outfopts found posrep postime opttime flow savedsnf
        bibdfound keep timeopt rnbalance neighbor serial position
        nfunction oldkeep pwt nwt optexmemory attrortr;

%let time      = %sysfunc(datetime());
%let timeopt   = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote  = %sysfunc(getoption(notes));
%let saveopts  = &savenote missing="%sysfunc(getoption(missing))"
                 %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                 %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                 %sysfunc(getoption(serror));
%let abort     = 0;
%let setnote   = nonotes;
%let printseed = 1;
%let flow      = 0;
%let localver  = 25Jul2010;
%let _mktver   = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let i = %upcase(&mktopts);
   %if %index(&i, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&i, NOTES)   %then %let setnote = notes;
   %if %index(&i, VERSION) %then %put MktBIBD macro version &localver.;
   %let printseed = (%index(&i, NOSEED) eq 0);
   %let flow      = %eval(%index(&i, FLOW) gt 0);
   %end;
options serror;

options &setnote generic missing=' ';

%if %nrbquote(&help) ne %then %do; %mkthelp(mktbibd) %goto endit; %end;

*------check options------;
data _null_;

   length w $ 24 list $ 200;
   abort = 0;

   call symput('optexmemory', '0');
   call symput('eff', '0');
   call symput('bibdfound', '0');

   *------check OPTIONS=------;
   list = lowcase(symget('options'));
   i = index(list, 'serial'   ); if i then substr(list, i, 6) = ' ';
   j = index(list, 'neighbor' ); if j then substr(list, j, 8) = ' ';
   k = index(list, 'position' ); if k then substr(list, k, 8) = ' ';
   call symput('serial',    put(i > 0, 1.));
   call symput('neighbor',  put(j > 0 and not i, 1.));
   call symput('rnbalance', put(i or j, 1.));
   call symput('nfunction', scan('neighbor serial', 1 + (i > 0)));
   call symput('position',  put(i or j or k, 1.));
   if list ne ' ' then do;
      list = compbl(list);
      put 'ERROR: OPTIONS=' list 'is invalid.';
      abort = 1;
      end;

   *------check SEED=------;
   seed = input(symget('seed'), ?? 23.);
   if nmiss(seed) or seed ne round(seed) then do;
      put "ERROR: SEED=%nrbquote(&seed) is invalid.";
      abort = 1;
      end;

   *------special case random number seeds------;
   else if symget('seed') eq '00' then call symput('seed', '0');

   *------run generically, with constant seed,
          eliminates using a clock generated seed------;
   else if seed = 0 and (not index(symget('saveopts'), 'NOGENERIC')) and
      getoption('stimer' ) eq 'NOSTIMER' then call symput('seed', '072555');

   *------set seed using clock time when negative------;
   else if seed le 0 then do;
      seed = ceil(1e7 * uniform(input(reverse(put(1e2 * datetime(), 13.0)),
                                      ?? 8.)));
      call symput('seed', compress(put(seed, 8.)));
      if &printseed then put 'Seed = ' seed;
      end;

   *------check B=------;
   o1 = (symget('b') ne ' ');
   o2 = (symget('nsets') ne ' ');
   if (o1 + o2) ne 1 then do;
      put 'ERROR: You must specify exactly one of the following: '
          'B= or NSETS=.';
      abort = 1;
      end;

   else do;
      call symput('b', symget('b') || symget('nsets'));
      b = input(symget('b'), ?? 32.);
      if nmiss(b) or b < 2 or b ne round(b) or b > 5000 then do;
         if o1 then put "ERROR: B=%nrbquote(&b) is invalid.";
         if o2 then put "ERROR: NSETS=%nrbquote(&nsets) is invalid.";
         abort = 1;
         end;
      end;
   call symput('b', trim(put(b, 12. -L)));

   *------check T=------;
   o1 = (symget('t') ne ' ');
   o2 = (symget('nattrs') ne ' ');
   if (o1 + o2) ne 1 then do;
      put 'ERROR: You must specify exactly one of the following: '
          'T= or NATTRS=.';
      abort = 1;
      end;

   else do;
      call symput('t', symget('t') || symget('nattrs'));
      t = input(symget('t'), ?? 32.);
      if nmiss(t) or t < 2 or t ne round(t) or t > 5000 then do;
         if o1 then put "ERROR: T=%nrbquote(&t) is invalid.";
         if o2 then put "ERROR: NATTRS=%nrbquote(&nattrs) is invalid.";
         abort = 1;
         end;
      call symput('attrortr', scan("Treatment Attribute", 1 + o2));
      end;
   call symput('t', trim(put(t, 12. -L)));

   *------check K=------;
   o1 = (symget('k') ne ' ');
   o2 = (symget('setsize') ne ' ');
   if (o1 + o2) ne 1 then do;
      put 'ERROR: You must specify exactly one of the following: '
          'K= or SETSIZE=.';
      abort = 1;
      end;

   else do;
      call symput('k', symget('k') || symget('setsize'));
      k = input(symget('k'), ?? 32.);
      if nmiss(k) or k < 2 or k ne round(k) or k > 5000 then do;
         if o1 then put "ERROR: K=%nrbquote(&k) is invalid.";
         if o2 then put "ERROR: SETSIZE=%nrbquote(&setsize) is invalid.";
         abort = 1;
         end;
      end;
   call symput('k', trim(put(k, 12. -L)));

   *------check GROUP=------;
   w = symget('group');
   if w ne ' ' then do;
      g = input(symget('group'), ?? 32.);
      if nmiss(g) or g < 2 or g ne round(g) or g gt b or g > 5000 then do;
         put "ERROR: GROUP=%nrbquote(&group) is invalid.";
         abort = 1;
         end;
      call symput('group', trim(put(g, 12. -L)));
      end;

   *------check WEIGHTS=------;
   j = input(scan(symget('weights'), 1, ' '), ?? 32.);
   k = input(scan(symget('weights'), 2, ' '), ?? 32.);
   if nmiss(j,k) or j < 0 or k < 0 then do;
      put "ERROR: WEIGHTS=%nrbquote(&weights) is invalid.";
      abort = 1;
      end;
   call symput('pwt', trim(put(j, best16. -L)));
   call symput('nwt', trim(put(k, best16. -L)));

   if abort or _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data _null_;

   length w $ 24 str $ 500;
   abort = 0;

   b = input(symget('b'), ?? 32.);
   t = input(symget('t'), ?? 32.);
   k = input(symget('k'), ?? 32.);

   *------check necessary but not sufficient conditions for BIBD------;
   r = 0;
   if abort eq 0 and t gt 1 then do;
      r = b * k / t;
      l = r * (k - 1) / (t - 1);
      r = (abs(l - round(l)) < 1e-6 and abs(r - round(r)) < 1e-6 and
           2 le k and k le t and b ge t);
      end;
   call symput('ispossible', put(r, 1.));

   if k gt t then
      put "WARNING: k=&k > t=&t..  Typically, k should be less than t.";

   if b * k lt t then do;
      put "ERROR: b=&b * k=&k < t=&t..  Every treatment "
          "must appear at least once.";
      abort = 1;
      end;

   if not abort and (b + t - 1) gt (b * k) then do;
      put "ERROR: Invalid parameters: "
          "b + t - 1 = %eval(&b+&t-1) > b * k = %eval(&b*&k).";
      abort = 1;
      end;

   *------check OPTITER=------;
   w = symget('optiter');
   w1 = input(scan(w, 1, ' '), ?? 32.);
   if scan(w, 2, ' ') = ' ' then w2 = 1;
   else w2 = input(scan(w, 2, ' '), ?? 32.);
   if scan(w, 3, ' ') = ' ' then w3 = 0.1;
   else w3 = input(scan(w, 3, ' '), ?? 32.);
   if nmiss(w2) then do;
      if r then w2 = 1000;
      else w2 = 5;
      end;
   if nmiss(w3) then do;
      if r then w3 = 5;
      else w3 = 1;
      end;

   if nmiss(w1) or w1 < 1 or w1 ne round(w1) or
      nmiss(w2) or w2 < 1 or w2 ne round(w2) or
      nmiss(w3) or w3 < 0 then do;
      put "ERROR: OPTITER=%nrbquote(&optiter) is invalid.";
      abort = 1;
      end;

   call symput('optiter', compress(put(w1, 12.)));
   call symput('optrep' , compress(put(w2, 12.)));
   call symput('opttime', compress(put(w3 * 60, 12.)));
   w1 = min(w1, 500);
   call symput('keep',    compress(put(w1, 12.)));
   call symput('oldkeep', compress(put(w1, 12.)));

   *------check POSITER=------;
   w = symget('positer') || ' 0 0 0';
   w1 = input(scan(w, 1, ' '), ?? 32.);
   w2 = input(scan(w, 2, ' '), ?? 32.);
   w3 = input(scan(w, 3, ' '), ?? 32.);

   if nmiss(w1) or w1 < 0 or w1 ne round(w1) or
      nmiss(w2) or w2 < 0 or w2 ne round(w2) or
      nmiss(w3) or w3 < 0 then do;
      put "ERROR: POSITER=%nrbquote(&positer) is invalid.";
      abort = 1;
      end;

   call symput('positer', compress(put(w1, 12.)));
   call symput('posrep' , compress(put(w2, 12.)));
   call symput('postime', compress(put(w3 * 60, 12.)));

   *------separate out data set options------;
   str = symget('outs'); if str eq ' ' then str = 'Sorted';
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('outs', substr(str, 1, j - 1));
   call symput('outsopts', substr(str, j));

   str = symget('out'); if str eq ' ' then str = 'BIBD';
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('out', substr(str, 1, j - 1));
   call symput('outopts', substr(str, j));

   str = symget('outi'); if str eq ' ' then str = 'Incidence';
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('outi', substr(str, 1, j - 1));
   call symput('outiopts', substr(str, j));

   str = symget('outf'); if str eq ' ' then str = 'Factorial';
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('outf', substr(str, 1, j - 1));
   call symput('outfopts', substr(str, j));

   if abort or _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;

filename __dummy dummy;

proc printto log=__dummy; run;

%if &syserr > 4 %then %let abort = 2;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outi) ne %then %do;
   data _null_; set &outi; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outf) ne %then %do;
   data _null_; set &outf; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outs) ne %then %do;
   data _null_; set &outs; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;

%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )out=&out;
   %put %str(       )outf=&outf;
   %put %str(       )outs=&outs;
   %put %str(       )outi=&outi;
   %end;

%if &abort %then %goto endit;

*------count titles------;
proc sql;
   create view __tmpeview as select * from dictionary.titles;
   quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data _null_;
   set __tmpeview end=eof;
   if type = 'T' then ntitles + 1;
   if eof then call symput('tnum', compress(put(ntitles + 1, 2.)));
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------candidate set------;
data __tmpican;
   do x2 = 1 to &t; output; end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------find bibd------;
ods select none;
%let _optseed = 0;
data __tmpiheff; bcriterion = ._; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data __tmpihdes; __abort = 1; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto outofloop;

proc printto log=__dummy; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto outofloop;

%do i = 1 %to &optrep;
   proc optex data=__tmpican seed=&seed;
      ods output blockdesignefficiencies=__tmpieff;
      class x2;
      model x2;
      blocks structure=(&b)&k iter=&optiter noexchange keep=&keep;
      %do j = 1 %to &keep;
         output out=__tmpides&j(rename=(block=x1)) number=&j;
         %end;
      run;

   %if &syserr = 1016 or &syserr = 116 %then %let optexmemory = &syserr;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto outofloop;

   *------check results------;
   data _null_;
      set __tmpieff;
      length seed $ 20;
      if &flow then put "i=&i" @8 "seed=&seed " @25 bcriterion=;
      if bcriterion > &eff then do;
         call symput('gotone', '1');
         call symput('eff', put(bcriterion, 8.4));
         if bcriterion > 100 - 1e-6 then do;
            call symput('i', "&optrep");
            call symput('bibdfound', '1');
            end;
         end;
      else call symput('gotone', '0');

      if (datetime() - &time) gt &opttime then call symput('i', "&optrep");

      seed = symget('sysranend');
      if seed = ' ' then seed = symget('_optseed');
      if seed = '0' then seed = put(1e8 * uniform(&seed), 9.0);
      call symput('seed', compress(seed));

      if _error_ then call symput('abort', '1');
      stop;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto outofloop;

   %if &gotone %then %do;
      data __tmpihdes;
         set __tmpides1;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto outofloop;

     data __tmpiheff;
         set __tmpieff;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto outofloop;
      %end;

   %if &bibdfound %then %do;
      *------how many did we find?------;
      data _null_;
         set __tmpieff;
         if bcriterion > 100 - 1e-6 then call symput('keep', put(_n_, 3.));
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto outofloop;
      %end;

   %else %let keep = 1;
   %end;

%outofloop: proc printto log=log; run;
ods select all;

%if &syserr > 4 %then %let abort = 1;  %if &abort %then %goto endit;

data __tmpihdes;
   __abort = .;
   %if &keep gt 1 %then %do;
      set %do i = 1 %to &keep; __tmpides&i(in=i&i) %end;;
      keep = %do i = 1 %to &keep; &i * i&i + %end; 0;
      %end;
   %else %do;
      set __tmpihdes;
      keep = 1;
      if n(__abort) then do;
         put 'ERROR: Design not found.  Check your specification.';
         call symput('abort', '1');
         end;
      %end;
   if _error_ then call symput('abort', '1');
   drop __abort;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------go from two-factor design to more typical format------;
proc sort data=__tmpihdes; by keep x1 x2; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc transpose data=__tmpihdes out=__tmpihdes(drop=_: x1) prefix=b;
   var x2;
   by keep x1;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data __tmpihdes;
   set __tmpihdes;
   r = uniform(&seed);
   if _n_ = 1 then call symput('seed', compress(put(&seed + 1, 12.0)));
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc sort data=__tmpihdes out=__tmpihdes(drop=r); by keep r; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------optimize positional frequencies, display results------;
proc iml;

   *------compute positional frequencies------;
   start posfreq(x);
      p = j(&t, &k, 0);
      do i = 1 to &b;
         do j = 1 to &k;
            p[x[i,j],j] = p[x[i,j],j] + 1;
            end;
         end;
      return(p);
      finish;

   *------evaluate row-neighbor badness (nondirectional)------;
   start neighbor(x) global(init);
      nf = init;
      do i = 1 to &b;
         do j = 1 to %eval(&k - 1);
            p1 = min(x[i, j], x[i, j + 1]);
            p2 = max(x[i, j], x[i, j + 1]);
            nf[p1, p2] = sum(nf[p1, p2], 1);
            end;
         end;
      return(nf);
      finish;

   *------evaluate row-neighbor badness (looking both directions)------;
   start serial(x) global(init);
      nf = init;
      do i = 1 to &b;
         do j = 1 to %eval(&k - 1);
            p1 = x[i, j];
            p2 = x[i, j + 1];
            nf[p1, p2] = sum(nf[p1, p2], 1);
            end;
         end;
      return(nf);
      finish;

   *------randomize position within blocks------;
   start randomize(x);
      do i = 1 to &b;
         k = rank(uniform(j(1, &k, &seed)));
         t = x[i,]; x[i,] = t[,k];
         end;
      return(x);
      finish;

   *------read design and efficiency------;
   use __tmpihdes(drop=keep);       read all into allx; close __tmpihdes;
   use __tmpiheff(keep=bcriterion); read all into eff;  close __tmpiheff;
   eff = eff[1,1];

   *------notes about what is found and what is possible------;
   file log;
   call symput('found', 'Design');
   if eff > 100 - 1e-6 then do;
      put 'NOTE: A BIBD was found.';
      if &k = &t then call symput('found', 'Complete Block Design');
      else if &k < &t then
         call symput('found', 'Balanced Incomplete Block Design');
      end;
   else if &ispossible & &k > &t then put 'NOTE: A BIBD was not found.';
   else if &ispossible then
      put 'NOTE: A BIBD was not found, although one may exist.';
   else if ^&ispossible then
      put "NOTE: A BIBD is not possible with b=&b, t=&t, k=&k..";
   file print;

   %if &flow %then %do;
      file log;
      put "Keep = &keep (the number actually kept)";
      file print;
      %end;

   *------optimize the positions and neighbors------;
   which = 0;         w = 0;  z = 0;                pwt = &pwt;   nwt = &nwt;
   posmin = 1e34;                                   bestposmin = posmin;
   neimin = %if &rnbalance %then 1e34; %else 0;;    bestneimin = neimin;
   nmean  = ((&k - 1) * &b) / (&t * (&t - 1) / 2);
   nmn    = .; nmx = .;
   %if &serial %then %do;
      nmean    = nmean / 2;
      init     = j(&t, &t, 0) + diag(j(1, &t, .));
      holdinit = init;
      %end;
   %if &neighbor %then %do;
      init     = toeplitz(j(1,&t-1,.)||j(1,&t-1,0))[1:&t,(&t-1):2#&t-2];
      holdinit = init;
      %end;

   if &k <= 1 | ^(&position | &rnbalance) | &posrep = 0 then do;
      x = allx[1 : &b,];
      bestx = x;
      end;

   else do;
      time = datetime();
      do r = 1 to &posrep until(((pwt # posmin + nwt # neimin) <= z)
                              | ((datetime() - time) > &postime));
         which = mod(which, &keep) + 1;

         x = allx[(((which - 1) # &b + 1) : (which # &b)),];
         x = randomize(x);
         p = posfreq(x);
         %if &rnbalance %then %do; nf = &nfunction(x); %end;
         if r = 1 then do;
            pmean = p[:];
            pmx = ceil(pmean) + 1e-8;
            pmn = floor(pmean) - 1e-8;
            posrange = (pmx - pmn) > 1e-6;

            %if &rnbalance %then %do;
               nmx = ceil(nmean) + 1e-8;
               nmn = floor(nmean) - 1e-8;
               neirange = (nmx - nmn) > 1e-6;
               %end;

            %if &flow %then %do;
               file log;
               put "Pos Mean = " pmean best5.
                   " Min = " pmn best5. " Max = " pmx best5. /
                   "Nei Mean = " nmean best5.
                   " Min = " nmn best5. " Max = " nmx best5. //
                   "                           Best     Best" /
                   "  Rep  Pos Min  Nei Min  Pos Min  Nei Min  Which" /;
               file print;
               %end;
            end;
         posmin = (p - pmean)[##];
         %if &rnbalance %then %do; neimin = abs(nf - nmean)[##]; %end;
         do l = 1 to &positer until(((pwt # posmin + nwt # neimin) <= z) |
                                 (posmin = lastposmin & neimin = lastneimin));
            lastneimin = neimin;
            lastposmin = posmin;
            do pass = 1 to 2;
               do i = 1 to &b;
                  do j = 1 to &k;
                     do until(k ^= j); k = ceil(&k # uniform(&seed)); end;
                     t1 = x[i,j]; x[i,j] = x[i,k]; x[i,k] = t1;
                     p = posfreq(x);
                     v = (p - pmean)[##];

                     %if &rnbalance %then %do;
                        nf = &nfunction(x);
                        w  = abs(nf - nmean)[##];
                        %end;

                     if (pwt # v      + nwt # w) <=
                        (pwt # posmin + nwt # neimin) then do;
                        posmin = v;
                        %if &rnbalance %then %do; neimin = w; %end;
                        if (pwt # posmin + nwt # neimin) <= z then goto done;
                        end;
                     else do; x[i,k] = x[i,j]; x[i,j] = t1; end;
                     end;
                  end;
               end;
            end;
         done:
         if posrange then do;
            p = posfreq(x);
            if all((pmn <= p) & (p <= pmx)) then posmin = 0;
            end;
         %if &rnbalance %then %do;
            if neirange then do;
               init = j(&t, &t, 0);
               %if &neighbor %then %do;
                  nf = symsqr(&nfunction(x)` + nmn * i(&t));
                  %end;
               %if &serial %then %do; nf = &nfunction(x) + nmn * i(&t); %end;
               if all((nmn <= nf) & (nf <= nmx)) then neimin = 0;
               init = holdinit;
               end;
            %end;
         if (pwt # posmin + nwt # neimin) <
            (pwt # bestposmin + nwt # bestneimin) then do;
            bestx = x;
            bestposmin = posmin;
            bestneimin = neimin;
            end;
         %if &flow %then %do;
            file log;
            put r best5. posmin best9. neimin best9.
                bestposmin best9. bestneimin best9. which best7.
                ({" " " *"}[(posmin = bestposmin & neimin = bestneimin) + 1]);
            file print;
            %end;
         end;
      x = bestx;
      end;

   p = posfreq(x);
   %if &rnbalance %then %do; nf = &nfunction(x); %end;

   *------get the treatment by treatment frequencies matrix------;
   f = j(&t, &t, 0);
   do i = 1 to &b;
      do j = 1 to &k;
         do k = j to &k;
            a = min(x[i,j],x[i,k]);
            b = max(x[i,j],x[i,k]);
            f[a,b] = f[a,b] + 1;
            end;
         end;
      end;

   *------post-process the treatment frequencies to upper triangle------;
   l = 0;
   do j = 1 to &t;
      do k = j + 1 to &t;
         l = l + f[j,k];
         f[k,j] = .;
         end;
      end;

   *------store the design------;
   vname = 'x1' : rowcatc("x&k");
   create &out from x[colname=vname]; append from x;

   *------basic specification, efficiency, and results factoid------;
   n = &b * &k;
   %if ^&bibdfound %then %do;
      r = vecdiag(f)[:];
      l = l / ((&t * (&t - 1)) / 2);
      %end;
   %else %do;
      r = n / &t;
      l = r * (&k - 1) / (&t - 1);
      %end;
   posok = {"     Yes" "      No"}[(bestposmin > 0) + 1];
   neiok = {"     Yes" "      No"}[(bestneimin > 0) + 1];
   print 'Block Design Efficiency Criterion      ' eff[format=8.4 label=''],
         %if &attrortr eq Treatment %then %do;
         'Number of Treatments, t                ' (&t)[format=8.0 label=''],
         'Block Size, k                          ' (&k)[format=8.0 label=''],
         'Number of Blocks, b                    ' (&b)[format=8.0 label=''],
         %end;
         %else %do;
         'Number of Attributes, t                ' (&t)[format=8.0 label=''],
         'Set Size, k                            ' (&k)[format=8.0 label=''],
         'Number of Sets, b                      ' (&b)[format=8.0 label=''],
         %end;
         %if &bibdfound %then %do;
         "&attrortr Frequency                    " r[format=best8. label=''],
         'Pairwise Frequency                     ' l[format=best8. label=''],
         %end;
         %else %do;
         "Average &attrortr Frequency            " r[format=best8. label=''],
         'Average Pairwise Frequency             ' l[format=best8. label=''],
         %end;
         'Total Sample Size                      ' n[format=best8. label=''],
         'Positional Frequencies Optimized?      ' posok[label='']
         %if &rnbalance %then , 'Row-Neighbor Frequencies Optimized?    '
                                                   neiok[label=''];;

   *------display the frequency matrices------;
   w = ceil(log10(&t + 1)); t = right(char(1, w, 0) : char(&t, w, 0));
   w = ceil(log10(&k + 1)); k = right(char(1, w, 0) : char(&k, w, 0));
   mf = max(f);           mp = max(p);
   if mf < 100 then
      print "&attrortr by &attrortr Frequencies",,
            f[format=2. label='' rowname=t colname=t];
   else if mf < 1000 then
      print "&attrortr by &attrortr Frequencies",,
            f[format=3. label='' rowname=t colname=t];
   else
      print "&attrortr by &attrortr Frequencies",,
            f[format=best4. label='' rowname=t colname=t];
   if mp < 100 then
      print "&attrortr by Position Frequencies",,
            p[format=2. label='' rowname=t colname=k];
   else if mp < 1000 then
      print "&attrortr by Position Frequencies",,
            p[format=3. label='' rowname=t colname=k];
   else
      print "&attrortr by Position Frequencies",,
            p[format=best4. label='' rowname=t colname=k];
   %if &rnbalance %then %do;
      mp = max(nf);
      if mp < 100 then
         print 'Row-Neighbor Frequencies',,
               nf[format=2. label='' rowname=t colname=t];
      else if mp < 1000 then
         print 'Row-Neighbor Frequencies',,
               nf[format=3. label='' rowname=t colname=t];
      else
         print 'Row-Neighbor Frequencies',,
               nf[format=best4. label='' rowname=t colname=t];
      %end;
   quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------Incidence design------;
data __tmpibin(keep=%if %nrbquote(&group) ne %then Group; b:);
   %if %nrbquote(&group) ne %then %do; Group = mod(_n_ - 1, &group) + 1; %end;
   set &out;
   array b[&t];
   array x[&k];
   do i = 1 to &t; b[i] = 0; end;
   do i = 1 to &k; b[x[i]] + 1; end;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------block the BIBD (which might make sense in partial profiles)------;
%if %nrbquote(&group) ne %then %do;
   data _null_;
      call symput('seed', compress(put(&seed + 1, 12.0)));
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc optex data=__tmpibin noprint seed=&seed;
      class group / param=effect;
      model group;
      generate initdesign=__tmpibin method=sequential;
      block design=__tmpibin noexchange keep=1 iter=&optiter init=random;
      class b1-b&t / param=effect;
      model b1-b&t;
      output out=&outi;
      run; quit;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%else %do;
   data &outi; set __tmpibin; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------if grouped, sort everything------;
%if %nrbquote(&group) ne %then %do;
   data &outi; length Group 8; set &outi; n + 1; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data &outs; merge &outi(keep=group) &out; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc sort data=&outs out=&outs; by group x1-x&k; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data &out; merge &outi(keep=group n) &out; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc sort data=&out out=&out(drop=n); by group n; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc sort data=&outi out=&outi(drop=n); by group n; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%else %do;
   proc sort data=&out out=&outs; by x1-x&k; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------rearrange the design as a factorial design------;
data &outf;
   set &out;
   Block = _n_;
   array x[&k];
   do i = 1 to &k; Treatment = x[i]; output; end;
   drop i x1-x&k;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------start clean up------;
proc datasets nolist;
   delete __tmpican __tmpihdes __tmpiheff __tmpieff __tmpibin
          %do j = 1 %to &oldkeep; __tmpides&j %end;;
   run; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------final results, the BIBD------;
proc print data=&out noobs;
   title&tnum "&found";
   %if %nrbquote(&group) ne %then %do; id group; by group; %end;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

title&tnum;

*------final data sets, data set options, size notes------;
options &savenote no&timeopt;

data &outs &outsopts; set &outs; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data &outi &outiopts; set &outi; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data &outf &outfopts; set &outf; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data &out &outopts; set &out; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------final clean up, errors, notes, and quit ------;
%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116 or &optexmemory = 1016 or
       &optexmemory = 116 %then %put ERROR: Insufficient memory.;
   %else %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %else %if &optexmemory %then %put ERROR: PROC OPTEX resource problem.;
   %put ERROR: The MKTBIBD macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTBIBD macro used &time seconds.;
   %end;

%mendit: %mend;
