 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktBlock                                            */
 /*   TITLE: Block a Choice or Linear Design                     */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs, choice modeling        */
 /*   PROCS: IML                                                 */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: This macro creates global macro variables           */
 /*             _MKTSEED _MKTDETER _MKTVER                       */
 /*          and calls the following autocall macro, which must  */
 /*          be installed:                                       */
 /*             MKEVAL                                           */
 /*                                                              */
 /*          The following SAS products are required to run      */
 /*          this macro: BASE, STAT, IML.                        */
 /****************************************************************/

 /*---------------------------------------------------------------------

This macro is used to block a choice design or an ordinary linear
experimental design.  When a choice design is too large to show all
choice sets to each subject, the design is blocked and a block of choice
sets is shown to each subject.  For example if there are 36 choice sets,
instead of showing each subject 36 sets, you could instead create 2
blocks and show 2 groups of subjects 18 sets each.  You could also
create 3 blocks of 12 choice sets or 4 blocks of 9 choice sets.  You can
also request just one block if you want to see the correlations and
frequencies among all of the attributes of all of the alternatives.

The design can be in one of two formats.  Typically, a choice design has
one row for each alternative of each choice set and one column for each
of the attributes.  Typically, this kind of design is produced by either
the ChoicEff or MktRoll macro.  Alternatively, a "linear" design is an
intermediate step in preparing some choice designs.  The linear design
has one row for each choice set and one column for each attribute of
each alternative.  Typically, the linear design is produced by the MktEx
macro.  This shows part of a choice design and its corresponding linear
design.

             Linear Design                  Choice Design

             x1 x2 x3 x4                   Set  Alt  x1 x2

              1  1  1  2                    1    1    1  1
                                                 2    1  2

              1  2  1  1                    2    1    1  2
                                                 2    1  1

              2  1  2  2                    3    1    2  1
                                                 2    2  2

              2  2  2  1                    4    1    2  2
                                                 2    2  1

The output from the macro is a data set containing the design, with
the blocking variable added, and hence not in the original order,
with choice sets numbered within blocks.

              Linear Design                Choice Design

        Block  Set  x1 x2 x3 x4       Block  Set  Alt  x1 x2

          1     1    1  2  1  1         1     1    1    1  2
                                                   2    1  1

          1     2    2  1  2  2         1     2    1    2  1
                                                   2    2  2

          2     1    1  1  1  2         2     1    1    1  1
                                                   2    1  2

          2     2    2  2  2  1         2     2    1    2  2
                                                   2    2  1

The macro tries to create a blocking factor that is uncorrelated with
every attribute of every alternative.  Let X be the coded "linear"
version of the choice design.  When a choice design is input, the macro
creates the linear design from the choice design.  Let B be the coded
blocking variable.  Then the macro seeks to optimize the determinant of
the partitioned matrix.

   | [B X]'[B X] | = | B'B   B'X |
                     | X'B   X'X |

In other words, the macro is simply trying to optimally add one
additional factor, a blocking factor, to the linear design.  Hence
it is trying to make B orthogonal to all of the attributes of all
of the alternatives.

Example: Create and block some designs.

   %mktex(3 ** 6, n=27)

   * Block the linear design.  Ask for three blocks;
   %mktblock(data=randomized, nblocks=3, factors=x1-x6, seed=377)

The output shows that the blocking factor is uncorrelated with all of
the factors in the design.  This output comes from the MktEval macro,
which is called by MktBlock.

                  Canonical Correlations Between the Factors
            There are 0 Canonical Correlations Greater Than 0.316

                 Block    x1      x2      x3      x4      x5      x6

        Block    1        0       0       0       0       0       0
        x1       0        1       0       0       0       0       0
        x2       0        0       1       0       0       0       0
        x3       0        0       0       1       0       0       0
        x4       0        0       0       0       1       0       0
        x5       0        0       0       0       0       1       0
        x6       0        0       0       0       0       0       1

Furthermore, each level of each factor occurs equally often in each block.

                             Frequencies

                 Block       9 9 9
                 x1          9 9 9
                 x2          9 9 9
                 x3          9 9 9
                 x4          9 9 9
                 x5          9 9 9
                 x6          9 9 9
                 Block x1    3 3 3 3 3 3 3 3 3
                 Block x2    3 3 3 3 3 3 3 3 3
                 Block x3    3 3 3 3 3 3 3 3 3
                 Block x4    3 3 3 3 3 3 3 3 3
                 Block x5    3 3 3 3 3 3 3 3 3
                 Block x6    3 3 3 3 3 3 3 3 3
                 x1 x2       3 3 3 3 3 3 3 3 3
                 x1 x3       3 3 3 3 3 3 3 3 3
                 x1 x4       3 3 3 3 3 3 3 3 3
                 x1 x5       3 3 3 3 3 3 3 3 3
                 x1 x6       3 3 3 3 3 3 3 3 3
                 x2 x3       3 3 3 3 3 3 3 3 3
                 x2 x4       3 3 3 3 3 3 3 3 3
                 x2 x5       3 3 3 3 3 3 3 3 3
                 x2 x6       3 3 3 3 3 3 3 3 3
                 x3 x4       3 3 3 3 3 3 3 3 3
                 x3 x5       3 3 3 3 3 3 3 3 3
                 x3 x6       3 3 3 3 3 3 3 3 3
                 x4 x5       3 3 3 3 3 3 3 3 3
                 x4 x6       3 3 3 3 3 3 3 3 3
                 x5 x6       3 3 3 3 3 3 3 3 3
                 N-Way       1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
                             1 1 1 1 1 1 1 1

Here is the blocked linear design (3 blocks of nine choice sets).
Note that in the linear version of the design there is one row for each
choice set and all of the attributes of all of the alternatives are in
the same row.

               Block    Run    x1    x2    x3    x4    x5    x6

                 1       1      3     1     3     3     3     3
                         2      2     3     2     1     2     1
                         3      3     3     3     1     1     2
                         4      1     3     2     3     2     2
                         5      1     1     1     3     2     1
                         6      2     1     3     2     3     1
                         7      2     2     2     2     3     3
                         8      1     2     1     2     1     2
                         9      3     2     1     1     1     3

               Block    Run    x1    x2    x3    x4    x5    x6

                 2       1      2     1     2     3     1     2
                         2      1     2     2     1     3     1
                         3      3     3     1     3     3     1
                         4      1     1     2     2     1     3
                         5      3     1     1     2     2     2
                         6      1     3     1     1     3     3
                         7      2     3     3     3     1     3
                         8      3     2     3     2     2     1
                         9      2     2     3     1     2     2

               Block    Run    x1    x2    x3    x4    x5    x6

                 3       1      1     1     3     1     3     2
                         2      3     1     2     1     1     1
                         3      1     2     3     3     2     3
                         4      1     3     3     2     1     1
                         5      2     2     1     3     1     1
                         6      3     3     2     2     2     3
                         7      2     3     1     2     3     2
                         8      2     1     1     1     2     3
                         9      3     2     2     3     3     2

Next, we'll create a smaller choice design with two blocks of nine sets
instead of three blocks of nine sets.

   %mktex(3 ** 6, n=729)

   %mktkey(2 3)

   %mktroll(design=randomized, key=key, out=out)

   * Create an efficient choice design;
   %choiceff(data=out, model=class(x1-x3), nsets=18, nalts=2,
             beta=zero, options=nodups, seed=151)

   * Block the choice design.  Ask for 2 blocks;
   %mktblock(data=best, nalts=2, nblocks=2, factors=x1-x3, seed=472)

(Note that if this had been a branded example, and if you were running
SAS version 8.2 or an earlier release, specify ID=BRAND; do not add your
brand variable to the factors list.  For version 9 and later SAS
versions, adding your brand variable to the factors list is fine.)

Both the design and the blocking are not as good this time.  The
variable names in the output are composed of "Alt', the alternative
number, and the factor name.  Since there are two alternatives each
composed of three factors plus one blocking variable (2x3+1=7), a 7x7
correlation matrix is reported.

                  Canonical Correlations Between the Factors
            There are 10 Canonical Correlations Greater Than 0.316

           Block   Alt1_x1   Alt1_x2   Alt1_x3   Alt2_x1   Alt2_x2   Alt2_x3

 Block     1        0.13      0         0.14      0.15      0         0.14
 Alt1_x1   0.13     1         0.38      0.19      0.63      0.38      0.28
 Alt1_x2   0        0.38      1         0.47      0.18      0.58      0.34
 Alt1_x3   0.14     0.19      0.47      1         0.40      0.34      0.55
 Alt2_x1   0.15     0.63      0.18      0.40      1         0.18      0.21
 Alt2_x2   0        0.38      0.58      0.34      0.18      1         0.46
 Alt2_x3   0.14     0.28      0.34      0.55      0.21      0.46      1


                            Summary of Frequencies
            There are 10 Canonical Correlations Greater Than 0.316
                      * - Indicates Unequal Frequencies

                                 Frequencies

              Block              9 9
         *    Alt1_x1            7 4 7
              Alt1_x2            6 6 6
         *    Alt1_x3            7 5 6
         *    Alt2_x1            5 8 5
              Alt2_x2            6 6 6
         *    Alt2_x3            5 7 6
         *    Block Alt1_x1      3 2 4 4 2 3
              Block Alt1_x2      3 3 3 3 3 3
         *    Block Alt1_x3      3 3 3 4 2 3
         *    Block Alt2_x1      3 4 2 2 4 3
              Block Alt2_x2      3 3 3 3 3 3
         *    Block Alt2_x3      3 3 3 2 4 3
         *    Alt1_x1 Alt1_x2    2 2 3 2 2 0 2 2 3
         *    Alt1_x1 Alt1_x3    2 2 3 2 1 1 3 2 2
         *    Alt1_x1 Alt2_x1    0 5 2 1 0 3 4 3 0
         *    Alt1_x1 Alt2_x2    3 2 2 0 2 2 3 2 2
         *    Alt1_x1 Alt2_x3    3 2 2 1 2 1 1 3 3
         *    Alt1_x2 Alt1_x3    2 3 1 2 2 2 3 0 3
         *    Alt1_x2 Alt2_x1    2 3 1 1 3 2 2 2 2
         *    Alt1_x2 Alt2_x2    0 4 2 2 0 4 4 2 0
         *    Alt1_x2 Alt2_x3    2 3 1 1 3 2 2 1 3
         *    Alt1_x3 Alt2_x1    1 3 3 2 3 0 2 2 2
         *    Alt1_x3 Alt2_x2    3 3 1 1 2 2 2 1 3
         *    Alt1_x3 Alt2_x3    0 3 4 3 0 2 2 4 0
         *    Alt2_x1 Alt2_x2    2 2 1 2 3 3 2 1 2
         *    Alt2_x1 Alt2_x3    2 2 1 2 3 3 1 2 2
         *    Alt2_x2 Alt2_x3    3 1 2 2 3 1 0 3 3
              N-Way              1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1

Note that in this example, the input is a choice design (as opposed
to the linear version of a choice design) so the results are in choice
design format.  There is one row for each alternative of each choice
set.

                    Block    Set    Alt    x1    x2    x3

                      1       1      1      3     1     1
                                     2      2     2     2

                      1       2      1      1     2     3
                                     2      2     3     2

                      1       3      1      3     3     3
                                     2      1     2     2

                      1       4      1      3     1     2
                                     2      1     3     3

                      1       5      1      2     2     1
                                     2      3     3     3

                      1       6      1      1     2     2
                                     2      2     1     1

                      1       7      1      1     3     3
                                     2      3     1     1

                      1       8      1      2     1     2
                                     2      1     2     1

                      1       9      1      3     3     1
                                     2      2     1     3


                    Block    Set    Alt    x1    x2    x3

                      2       1      1      2     2     3
                                     2      3     3     2

                      2       2      1      1     1     3
                                     2      2     3     2

                      2       3      1      3     2     1
                                     2      1     1     2

                      2       4      1      1     3     1
                                     2      2     2     3

                      2       5      1      3     3     3
                                     2      1     1     1

                      2       6      1      1     3     1
                                     2      3     1     3

                      2       7      1      2     1     1
                                     2      3     2     2

                      2       8      1      1     1     2
                                     2      2     2     1

                      2       9      1      3     2     2
                                     2      2     3     3

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

%macro mktblock(  /*--------------------------------------------------*/
                  /*                                                  */
help,             /* nonblank, list options                           */
                  /*                                                  */
                  /*-------You must specify this first option---------*/
                  /*                                                  */
nblocks=,         /* specifies the number of blocks to create.        */
                  /*  1 - just report information on the design.      */
                  /* >1 - block and report information.               */
                  /*                                                  */
vars=,            /* specifies the factors in the design.             */
                  /* By default, all numeric variables are used,      */
                  /* except variables with names matching those in    */
                  /* the BLOCK=, SET=, ALT=, and ID= options.  (By    */
                  /* default, the variables named Block, Set, Run,    */
                  /* and Alt are excluded.)  If you are using version */
                  /* 8.2 or an earlier SAS release with a branded     */
                  /* choice design (assuming the brand factor is      */
                  /* called Brand), specify ID=Brand.  Do not add     */
                  /* Brand to the factors list unless you are using   */
                  /* version 9 or a later SAS release.                */
factors=,         /* alias for VARS=.                                 */
                  /*                                                  */
data=_last_,      /* specifies either the choice or linear design.    */
                  /* The choice design has one row for each           */
                  /* alternative of each choice set and one column    */
                  /* for each of the attributes.  Typically this      */
                  /* design is produced by either the ChoicEff or     */
                  /* MktRoll macro.  For choice designs, you must     */
                  /* also specify NALTS=.                             */
                  /*                                                  */
                  /* The linear design has one row for each choice    */
                  /* set and one column for each attribute of each    */
                  /* alternative.  Typically this design is produced  */
                  /* by either the MktEx macro.  This is the design   */
                  /* that is input into the MktRoll macro.            */
                  /*                                                  */
                  /*--------------Output data set options-------------*/
                  /*                                                  */
out=blocked,      /* specifies the output data set with block numbers.*/
                  /*                                                  */
outr=,            /* specifies the randomized output data set if you  */
                  /* would like the design randomly sorted within     */
                  /* block.                                           */
                  /*                                                  */
block=Block,      /* specifies the variable to contain the block      */
                  /* number.  If this variable is in the input data   */
                  /* set, it is excluded.                             */
                  /*                                                  */
set=,             /* specifies the variable to contain the choice set */
                  /* number.  When nalts= is specified, the default   */
                  /* is Set, otherwise the default is Run.  If this   */
                  /* variable is in the input data set, it is         */
                  /* excluded.                                        */
                  /*                                                  */
alt=Alt,          /* specifies the variable to contain the            */
                  /* alternative number.  If this variable is in the  */
                  /* input data set, it is excluded.                  */
                  /*                                                  */
id=,              /* specifies variables in the DATA= data set to     */
                  /* copy to the output data set.   If you are using  */
                  /* version 8.2 or an earlier SAS release with a     */
                  /* branded choice design (assuming the brand factor */
                  /* is called Brand), specify ID=Brand.              */
                  /*                                                  */
                  /*------------------Other options-------------------*/
                  /*                                                  */
nalts=,           /* specifies the number of alternatives in each     */
                  /* choice set.  If you are inputting a choice       */
                  /* design, you must specify NALTS=, otherwise the   */
                  /* macro assumes you are inputting a linear design. */
                  /*                                                  */
ridge=0.01,       /* specifies the value to add to the diagonal of    */
                  /* X'X to make it nonsingular.  Usually, you will   */
                  /* not need to change this value.  If you do, you   */
                  /* probably will not notice any effect.  RIDGE=0    */
                  /* uses a generalized inverse instead of ridging.   */
                  /*                                                  */
maxiter=5,        /* specifies the number of times to try to block    */
                  /* the design starting with a different random      */
                  /* blocking.  By default, the macro tries five      */
                  /* random starts, and iteratively refines each      */
                  /* until D-efficiency quits improving, then in the  */
                  /* end selects the blocking with the best           */
                  /* D-efficiency.                                    */
iter=5,           /* Alias for MAXITER=.                              */
                  /*                                                  */
initblock=,       /* specifies the name of a variable in the data set */
                  /* that is to be used as the initial blocking       */
                  /* variable for the first iteration.                */
                  /*                                                  */
print=normal,     /* specifies both the MktBlock and the MktEval      */
                  /* macro printing options, which control the        */
                  /* printing of the results.  Values include:        */
                  /* corr    - canonical correlations.                */
                  /* list    - list of big canonical correlations.    */
                  /* ordered - list ordered by variable names.        */
                  /* nonzero - list ordered by variable names and     */
                  /*           sets LIST=1e-6 to print nonzero        */
                  /*           canonical correlations.                */
                  /* freqs   - long frequencies list.                 */
                  /* summ    - frequency summaries.                   */
                  /* block   - canonical correlations within blocks.  */
                  /* design  - blocked design.                        */
                  /* note    - blocking note.                         */
                  /* all     - all of the above.                      */
                  /* noprint - no printed output.                     */
                  /* normal  - corr list summ design note.            */
                  /* short   - corr summ note.                        */
                  /*                                                  */
list=0.316,       /* List any correlations bigger than this number.   */
                  /* The default is the square root of 0.1.  Typical  */
                  /* usage is something like LIST=0.3 or LIST=0.5.    */
                  /*                                                  */
next=1,           /* specifies how far into the design to go to look  */
                  /* for the next swap.  NEXT=1 specifies that the    */
                  /* macro should try swapping the level for each run */
                  /* with the level for the next run and all other    */
                  /* runs.  NEXT=2 considers swaps with half of the   */
                  /* other runs, which makes the algorithm run more   */
                  /* quickly.  The macro considers swapping the level */
                  /* for run i with run i + 1 then uses the NEXT=     */
                  /* value to find the next potential swaps.  Other   */
                  /* values, including nonintegers can be specified   */
                  /* as well.  For example NEXT=1.5 considers         */
                  /* swapping observation 1 with observations 2, 4,   */
                  /* 5, 7, 8, 10, 11, and so on.  With smaller        */
                  /* values, the macro tends to find a slightly       */
                  /* better blocking variable at a cost of much       */
                  /* slower run time.                                 */
                  /*                                                  */
options=,         /* nosort - do not sort the design into blocks.     */
                  /*                                                  */
seed=0,           /* specifies the random number generation seed,     */
                  /* 0 - clock time.                                  */
);                /*--------------------------------------------------*/
                  /* Global Macro Variable:                           */
                  /* _mktdeter - ridged determinant.                  */
                  /* _mktseed  - random number seed at the end of the */
                  /*             random number computations.          */
                  /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _mktdeter _mktseed _mktver;
%local  abort chdes i nobs printdes printnote savenote time savedsnf maxlen
        rename setnote excnote nfacts outopts outropts outname saveopts
        printseed localver nosort idflag timeopt;

%let time      = %sysfunc(datetime());
%let timeopt   = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote  = %sysfunc(getoption(notes));
%let saveopts  = &savenote missing="%sysfunc(getoption(missing))"
                 %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                 %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                 %sysfunc(getoption(serror));
%let setnote   = nonotes;
%let printseed = 1;
%let abort     = 0;
%let maxlen    = 32700;
%let localver  = 25Jul2010;
%let _mktver   = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let i = %upcase(&mktopts);
   %if %index(&i, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&i, NOTES) %then %let setnote = notes;
   %if %index(&i, VERSION) %then %put MktBlock macro version &localver.;
   %let printseed = (%index(&i, NOSEED) eq 0);
   %end;
options serror;

options &setnote generic missing='.';

%if %nrbquote(&help) ne %then %do; %mkthelp(mktblock) %goto endit; %end;

*------basic parameter checking------;
data _null_;

   length str1 str2 options print $ 400 w1 w2 $ 20;
   ok = 1;

   *------check OPTIONS=------;
   list = lowcase(symget('options'));
   i = index(list, 'nosort'); if i then substr(list, i, 6) = ' ';
   call symput('nosort',  put(i ne 0, 1.));
   list = compbl(list);

   if list ne ' ' then do;
      put 'ERROR: OPTIONS=' list 'is invalid.';
      ok = 0;
      end;

   *------parse PRINT= string------;
   options = lowcase(symget('print'));
   if index(options, 'short')
      then options = trim(options) || ' corr summ note';
   if index(options, 'normal')
      then options = trim(options) || ' corr list summ design note';
   if index(options, 'all')
      then options = trim(options) || ' corr list freqs summ block design note';
   do i = 1 to 400 until(w1 eq ' ');
      w1 = scan(options, i, ' ');
      if trim(w1) in ('short', 'normal', 'all') then do;
         substr(options, index(options, trim(w1)), length(w1)) = ' ';
         i = i - 1;
         end;
      end;
   do i = 1 to 400 until(w1 eq ' ');
      w1 = scan(options, i, ' ');
      if w1 ne ' ' then do j = i + 1 to 400 until(w2 eq ' ');
         w2 = scan(options, j, ' ');
         if w1 = w2 then do;
            substr(options, index(options, trim(w1)), length(w1)) = ' ';
            i  = i - 1;
            w2 = ' ';
            end;
         end;
      end;
   options = compbl(options);

   j = index(options, 'corr');
   if j then do;
      substr(options, j, 4) = ' ';
      print = trim(print) || ' corr';
      end;
   j = index(options, 'list');
   if j then do;
      substr(options, j, 4) = ' ';
      print = trim(print) || ' list';
      end;
   j = index(options, 'ordered');
   if j then do;
      substr(options, j, 7) = ' ';
      print = trim(print) || ' ordered';
      end;
   j = index(options, 'nonzero');
   if j then do;
      substr(options, j, 7) = ' ';
      print = trim(print) || ' nonzero';
      call symput('list', '1e-6');
      end;
   j = index(options, 'freqs');
   if j then do;
      substr(options, j, 5) = ' ';
      print = trim(print) || ' freqs';
      end;
   j = index(options, 'summ');
   if j then do;
      substr(options, j, 4) = ' ';
      print = trim(print) || ' summ';
      end;
   j = index(options, 'block');
   if j then do;
      substr(options, j, 5) = ' ';
      print = trim(print) || ' block';
      end;
   j = index(options, 'design');
   if j then substr(options, j, 6) = ' ';
   call symput('printdes',  put(j ne 0, 1.));
   j = index(options, 'note');
   if j then substr(options, j, 4) = ' ';
   call symput('printnote', put(j ne 0, 1.));
   j = index(options, 'noprint');
   if j then do;
      substr(options, j, 7) = ' ';
      if print eq ' ' then print = 'noprint';
      end;
   if options ne ' ' then do;
      put "ERROR: PRINT=%nrbquote(&print) is invalid.";
      call symput('abort', '1');
      end;
   if print = ' ' then print = 'noprint';
   call symput('print', trim(left(print)));

   *------check SEED=------;
   seed = input(symget('seed'), ?? 32.);
   if nmiss(seed) or seed ne round(seed) then do;
      put "ERROR: SEED=%nrbquote(&seed) is invalid.";
      ok = 0;
      end;

   *------special case random number seeds------;
   else if symget('seed') = '00' then call symput('seed', '0');

   *------run generically, with constant seed,
          eliminates using a clock generated seed------;
   else if seed = 0 and (not index(symget('saveopts'), 'NOGENERIC')) and
      getoption('stimer' ) eq 'NOSTIMER' then call symput('seed', '072555');

   *------set seed using clock time when negative------;
   else if seed le 0 then do;
      seed = ceil(1e7 * uniform(input(reverse(put(1e2 * datetime(), 13.0)),
                                      ?? 8.)));
      call symput('seed', compress(put(seed, 8.)));
      if &printseed then put 'Seed = ' seed;
      end;

   *------make sure there is an input data set------;
   data = lowcase(symget('data'));
   if data in (' ' '_last_') then do;
      data = symget('syslast'); call symput('data', trim(data));
      end;
   if lowcase(data) = '_null_' then do;
      put 'ERROR: No input data set.';
      ok = 0;
      end;

   nalts = input(symget('nalts'), ?? 32.);

   call symput('chdes', put(n(nalts), 1.));

   *------check SET=------;
   if symget('set') eq ' ' then do;
      if n(nalts) then call symput('set', 'Set');
      else             call symput('set', 'Run');
      end;

   *------check FACTORS=------;
   if symget('factors') ne ' ' and symget('vars') ne ' ' then do;
      put 'ERROR: You must specify FACTORS= or VARS=, but not both.';
      ok = 0;
      end;
   if symget('factors') eq ' ' then call symput('factors', symget('vars'));
   if symget('factors') eq ' ' then call symput('factors', '_numeric_');

   *------ID specified?------;
   call symput('idlist', put(symget('id') ne ' ', 1.));

   *------check NALTS=------;
   if (n(nalts) and (nalts < 2 or nalts ne round(nalts))) or
      (symget('nalts') ne ' ' and nmiss(nalts)) then do;
      put "ERROR: NALTS=%nrbquote(&nalts) is invalid with choice designs."
          / +7   'You must specify NALTS=number.';
      ok = 0;
      end;

   *------check NBLOCKS=------;
   nblocks = input(symget('nblocks'), ?? 32.);
   if nmiss(nblocks) or nblocks < 1 or nblocks ne round(nblocks) then do;
      put "ERROR: NBLOCKS=%nrbquote(&nblocks) is invalid.";
      ok = 0;
      end;

   *------check RIDGE=------;
   ridge = input(symget('ridge'), ?? 32.);
   if nmiss(ridge) or ridge < 0 then do;
      put "ERROR: RIDGE=%nrbquote(&ridge) is invalid.";
      ok = 0;
      end;

   *------check NEXT=------;
   next = input(symget('next'), ?? 32.);
   if nmiss(next) or next < 1 then do;
      put "ERROR: NEXT=%nrbquote(&next) is invalid.";
      ok = 0;
      end;

   *------check MAXITER=------;
   str1 = symget('iter');
   if str1 ne symget('maxiter') and
      str1 ne '5' then call symput('maxiter', trim(str1));
   maxiter = input(symget('maxiter'), ?? 32.);
   if nmiss(maxiter) or maxiter < 1 then do;
      str1 = symget('maxiter');
      put 'ERROR: MAXITER=' str1 'is invalid.';
      ok = 0;
      end;

   *------check output data sets------;
   str1 = symget('out');
   str2 = symget('outr');
   if str1 eq ' ' and str2 eq ' ' then do;
      put 'ERROR: Either OUT= or OUTR= must be specified.';
      ok = 0;
      end;

   *------separate out data set options------;
   j = index(str1, '('); if j = 0 then j = length(str1) + 1;
   call symput('out', substr(str1, 1, j - 1));
   if str1 ne ' ' then do;
      call symput('outopts', substr(str1, j));
      call symput('outname', substr(str1, 1, j - 1));
      end;

   j = index(str2, '('); if j = 0 then j = length(str2) + 1;
   call symput('outr', substr(str2, 1, j - 1));
   if str2 ne ' ' then do;
      call symput('outropts', substr(str2, j));
      if str1 eq ' ' then call symput('outname', substr(str1, 1, j - 1));
      end;

   if (sysprod('iml') + sysprod('stat')) ne 2 then do;
      put 'ERROR: SAS/STAT and SAS/IML must be licensed to run this macro.';
      ok = 0;
      end;

   if _error_ or not ok then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;

*------copy input data set------;
data __tmpkdat; set &data; run;
%let data = __tmpkdat;
%if &syserr > 4 %then %let abort = 1;

%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;

%if %nrbquote(&outr) ne %then %do;
   data _null_; set &outr; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )data=&data;
   %put %str(       )out=&out;
   %put %str(       )outr=&outr;
   %end;

%if &abort %then %goto endit;

*------check some variable names------;
data _null_;
   keep &block &set &alt &id &initblock;
   if _error_ then call symput('abort', '1');
   stop;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get variable list------;
proc contents data=&data(keep=&factors %if &idlist %then drop=&id;)
   noprint out=__tmpk1(keep=varnum name);
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc sort; by varnum; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------expand list, get rid of block, set, alt variables------;
data _null_;
   length l $ &maxlen e $ 200;
   retain l e;
   set __tmpk1 end=eof;
   n = upcase(name);
   if n in (%upcase("&block"), %upcase("&alt"), %upcase("&Set"))
      then e = trim(e) || ' ' || name;
      else l = trim(l) || ' ' || name;
   if eof then do;
      call symput('factors', trim(left(l)));
      if e ne ' ' then call symput('excnote', trim(left(e)));
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get rid of character variables------;
data __tmpk1;
   set &data(keep=&factors);
   __j = _n_;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc transreg data=__tmpk1 maxiter=0 method=uni noz;
   model opscore(&factors) = ide(__j);
   output replace out=__tmpk1(drop=intercept __j _name_ _type_);
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------prepare to make linear design, some error checking------;
data _null_;
   set __tmpk1 nobs=__nobs;
   array __n[*] &factors;
   length __b __s __a __v $ 67 %if &chdes %then __l $ &maxlen;;
   call symput('nfacts', compress(put(dim(__n), 8.)));

   %if &chdes %then %do;
      if mod(__nobs, &nalts) then do;
         put 'ERROR: At least one choice set is incomplete.';
         call symput('abort', '1');
         end;
      __nobs = __nobs / &nalts;
      call symput('nobs',   compress(put(__nobs  , 8.)));

      do __i = 1 to &nalts;
         do __j = 1 to dim(__n);
            __k + 1;
            call vname(__n[__j], __v);
            __l = trim(__l) || ' __zn' || trim(put(__k, 5. -L)) || '=' ||
                  'Alt' || trim(put(__i, 3. -L)) || '_' || __v;
            end;
         end;
      call symput('rename', trim(__l));
      %end;
   __b = lowcase(dequote(symget('block')));
   __s = lowcase(dequote(symget('set')));
   __a = lowcase(dequote(symget('alt')));
   __ok = 1;
   do __i = 1 to dim(__n);
      call vname(__n[__i], __v);
      __v = lowcase(__v);
      if __v eq __b or __v eq __s or __v eq __a then __ok = 0;
      end;
   if not __ok then
      put "ERROR: BLOCK=%nrbquote(&block), SET=%nrbquote(&set), "
          "and ALT=%nrbquote(&alt) are not "
          'permitted ' / +7 'in the FACTORS= variable list.';
   if _error_ or not __ok then call symput('abort', '1');
   stop;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------for choice design input, make linear design------;
%if &chdes %then %do;

   *------process choice design - convert to "linear"
          design with one row per choice set         ------;
   data __tmpk1(keep=__z: rename=(&rename));

      if 0 then set __tmpk1;
      array __n[&nfacts] &factors;
      array __zn[%eval(&nalts * &nfacts)];

      *------read nalts obs for one choice set, write 1 obs------;
      do __set = 1 to &nobs;
         __nn = 1; __cc = 1;
         do __j = 1 to &nalts;
            __l + 1;
            set __tmpk1 point=__l;
            do __k = 1 to &nfacts;
               __zn[__nn] = __n[__k];
               __nn + 1;
               end;
            end;
         output;
         end;
      if _error_ then call symput('abort', '1');
      stop;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------code the design------;
proc transreg design data=__tmpk1 nozeroconstant norestoremissing;
   model class(%if &chdes %then Alt:; %else &factors; / effects);
   output out=__tmpk2;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------main algorithm, block the design------;
proc iml;
   use __tmpk2(keep=intercept &_trgind); read all into x; close __tmpk2;
   n = nrow(x);
   file log;
   if &printnote then do;
      put "NOTE: Blocking &nfacts factors and "
          (compress(char(n)))
          %if &chdes %then ' choice sets'; %else ' linear design runs';
          " into %nrbquote(&nblocks) blocks of size "
          (compress(char(n / &nblocks)))
          %if &chdes %then ' sets.'; %else '.';;
     if symget('excnote') ^= ' ' then
         put 'NOTE: These variables were excluded '
             "from the input data set: &excnote..";
      end;

   block = j(n, 1, 0); bestblock = block; bestdet = -1; seed = &seed;
   x = x * root(sweep(x` * x / n + &ridge * i(ncol(x)), 1:ncol(x)))` / n;

   *------try MAXITER= different random starts------;
   do iter = 1 to %if &nblocks gt 1 %then &maxiter; %else 0;;

      *------create random blocking factor------;
      call ranuni(seed, block);
      block[rank(block)] = shape(1 : &nblocks, n, 1);

      %if %nrbquote(&initblock) ne %then %do;
         if iter = 1 then do;
            use &data(keep=&initblock);
            read all into block;
            %if &chdes %then %do; block = shape(block, &nobs, &nalts)[,1]; %end;
            end;
         %end;

      *------get set to iterate------;
      changed = 1;   b = designf(block);    bx = b` * x;
      if iter = 1 then bpb = b` * b / n;
      maxdet = det(bpb - bx * bx`);

      *------start iterating------;
      do k = 1 to 10 * n while((k - changed) < n);
         i = mod(k - 1, n) + 1;

         do %if &next = 1 %then j = i + 1 to n;
            %else %do;
               jj = i + mod(ceil(k / n) - 1, &next) + 1 to n by &next;
               j  = round(jj) %end;;

            if block[i] ^= block[j] then do;

               *------try swapping this set into a different block------;
               p = i || j;   q = j || i;

               *------Add the blocking variable as a factor in
                      the design.  Keep the swap if a ridged
                      efficiency criterion shows improvement. ------;

               up  = bx + (b[q,] - b[p,])` * x[p,];
               det = det(bpb - up * up`);
               if det > maxdet then do;
                  block[p] = block[q];  b[p,]   = b[q,];   bx = up;
                  maxdet   = det;       changed = k;
                  end;

               end;
            end;
         end;

      *------keep the best design------;
      if maxdet > bestdet then do;
         bestdet   = maxdet;
         bestblock = block;
         end;

      end;

   use __tmpk2(keep=intercept &_trgind); read all into x; close __tmpk2;
   x = x || designf(bestblock);
   bestdet = 100 # (det(x` * x / n) ## (1 / ncol(x)));

   call symput('_mktdeter', compress(char(bestdet)));
   call symput('_mktseed',  compress(char(seed)));
   block = bestblock || (1:n)`;
   create __tmpk3 var({"&block" '__set'}); append from block;
   quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------if we are printing blocking evaluation statistics------;
%if %nrbquote(&print) ne noprint %then %do;

   *------set up linear design for blocking evaluation------;
   data __tmpk2;
      merge __tmpk3 __tmpk1;
      drop __set;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------evaluate correlations, frequencies------;
   %let _mktver = obsolete;
   %mkteval(%if &nblocks gt 1 %then blocks=&block,;
            list=&list, print=&print);

   %if &_mktver ne &localver %then %do;
      %put ERROR: Incompatible MktEval macro.;
      %put %str(      ) You must have the same version of all macros.;
      %put %str(      ) Versions: MktBlock = &localver, MktEval = &_mktver..;
      %put %str(      ) Get the latest tools and documentation from:;
      %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
      %let abort = 1;
      %end;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------add blocking information to the design------;
data __tmpk2;
   %if &chdes %then %do;
      set __tmpk3;
      do __alt = 1 to &nalts; set &data(keep=&factors &id); output; end;
      %end;
   %else %do;
      merge __tmpk3 &data(keep=&factors &id);
      %end;
   %if &nosort %then %do; __n__ = _n_; %end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------sort sets into blocks------;
proc sort; by &block __set  %if &chdes %then __alt;; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------add set, alternative variables------;
data &outname;
   length &block &set %if &chdes %then &alt; 8;
   set __tmpk2(drop=__set %if &chdes %then __alt;);
   by &block;
   if first.&block then &set = 0;
   %if &chdes %then %do;
      &alt = mod(_n_ - 1, &nalts) + 1;
      if mod(_n_ - 1, &nalts) = 0 then
      %end; &set + 1;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------print the final blocked design------;
%if &printdes %then %do;
   proc print data=&outname %if &nosort %then (drop=__n__);;
      id &block &set %if &chdes %then &alt;;
      by &block %if &chdes %then &set;; pageby &block;
      run;
   %end;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------clean up temporary, intermediate data sets------;
proc datasets nolist; delete __tmpk1 __tmpk2 __tmpk3 __tmpkdat; run; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------make randomized design------;
%if %nrbquote(&outr) ne %then %do;
   data &outr(drop=__seed);
      retain __seed &_mktseed __r 0;
      set &outname end=__eof;
      %if &chdes %then %do; by &block &set; %end;
      %if &chdes %then if first.&set then; call ranuni(__seed, __r);
      if __eof   then call symput('_mktseed', compress(put(__seed, best16.)));
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc sort data=&outr out=&outr(drop=__r &set);
      by &block __r %if &chdes %then &set &alt;;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

options &savenote no&timeopt;

*------final randomized design------;
%if %nrbquote(&outr) ne %then %do;
   data &outr &outropts;
      length &block &set %if &chdes %then &alt; 8;
      set &outr;
      by &block;
      if first.&block then &set = 0;
      %if &chdes %then if mod(_n_ - 1, &nalts) = 0 then; &set + 1;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------final design------;
%if %nrbquote(&out) ne %then %do;
   data &out &outopts; set &outname; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------undo final sort with options=nosort------;
%if &nosort %then %do;
   proc sort data=&out out=&out(drop=__n__); by __n__; run;

   %if %nrbquote(&outr) ne %then %do;
      proc sort data=&outr out=&outr(drop=__n__); by __n__; run;
      %end;
   %end;

%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTBLOCK macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTBLOCK macro used &time seconds.;
   %end;

%mendit: %mend;
