 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktBSize                                            */
 /*   TITLE: Balanced Incomplete Block Designs Sizes             */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs                         */
 /*   PROCS:                                                     */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: The following SAS products are required to run      */
 /*          this macro: BASE.                                   */
 /*                                                              */
 /*          This macro creates global macro variables: _MktVer  */
 /****************************************************************/

 /*---------------------------------------------------------------------

This macro suggests sizes for balanced incomplete block designs.  The
sizes that it reports are sizes that meet necessary but not sufficient
conditions for the existence of a BIBD.  So a BIBD may not exist for
every size reported.  In this example, designs with 12 treatments,
between 4 and 8 treatments per block, and between 12 and 30 blocks
are requested.

   %mktbsize(t=12, k=4 to 8, b=12 to 30)

The results are as follows:

                                                                        n
          t          k          b            r           Lambda      Total
     Number of     Block    Number       Treatment    Pairwise       Sample
     Treatments    Size     of Blocks    Frequency    Frequencies    Size

         12          6          22           11            5           132

You can use this information to create a BIBD with the MktBIBD macro.

   %mktbibd(t=12, k=6, b=22, seed=104)

There is no guarantee that MktBIBD will find a BIBD for any
specification, but in this case it does, and it finds the design:

                        Balanced Incomplete Block Design

                        x1    x2    x3    x4    x5    x6

                         9     3    10     7    12     2
                         5     2     7     6    10     4
                         5    11    12    10     9     8
                         7     9    12    11     4     1
                        12    10     9     5     3     4
                        10     9    11     2     6     1
                         9     8     7     3     1     6
                         6    12     8     1     2    10
                         8     7     6     4     9     5
                         1    11     6     5    12     7
                         3     1     5     2     7    12
                         4     8     1    10     7    12
                        12     4     2     8     6     9
                         4     6    11     3    10     7
                        11     5     4     9     1     2
                        11     7     2     8     3     9
                         6     3    11    12     8     5
                         3     1    10     9     5     6
                         2    12     4     6    11     3
                         1    10     3    11     4     8
                         2     4     5     1     8     3
                        10     5     8     7     2    11

The design has b=22 blocks (rows), k=6 treatments per block (columns),
and t=12 treatments (the entries are the integers 1 to 12).  Each of
the t=12 treatments occurs r=11 times, and each treatment occurs in a
block with every other treatment lambda=5 times.

You can use b, t, and k in the arguments for the other options.  Note
however, that you cannot use their aliases, nsets=, nattrs=, or
setsize=.  Running the following step creates a list of over 50 designs:

   %mktbsize(t=5 to 20, k=3 to t - 1, b=t to 30)

Here are the first few:

                                                                        n
          t          k          b            r           Lambda      Total
     Number of     Block    Number       Treatment    Pairwise       Sample
     Treatments    Size     of Blocks    Frequency    Frequencies    Size

          5           3         10            6             3           30
          5           4          5            4             3           20
          6           3         10            5             2           30
          6           4         15           10             6           60
          6           5          6            5             4           30

Note that by default, MAXREPS=1 (the maximum number of replications is
1), so for example, the design t=5, k=3, b=20 is not listed since it
consists of two replications of t=5, k=3, b=10, which is listed.  Also
note that b, the number of blocks, was specified so that it is never
less than the number of treatments.  Furthermore, k, the block size
(number of treatments per block), is set to always be less than the
number of treatments.  Even more complicated expressions are allowed.
For example, to limit the number of treatments per block to no more
than half of the number of treatments, you could specify:

   %mktbsize(t = 2 to 10, k = 2 to 0.5 * t, b=t to 10)

which produces:

                                                                        n
          t          k          b            r           Lambda      Total
     Number of     Block    Number       Treatment    Pairwise       Sample
     Treatments    Size     of Blocks    Frequency    Frequencies    Size

          4          2           6           3             1           12
          5          2          10           4             1           20
          6          3          10           5             2           30
          7          3           7           3             1           21

To limit the number of blocks as a function of the number of
treatments, you could specify:

   %mktbsize(t = 2 to 10, k = 2 to t - 1, b = t to 2 * t)

However, if you want to limit the number of treatments as a function of
the number of blocks, you need to use the ORDER= option to ensure that
the number of blocks loop comes first.

   %mktbsize(b = 2 to 10, t = 2 to 0.5 * b, k = 2 to t - 1, order=btk)

The macro reports sizes in which r = b * k / t and l = r * (k - 1) / (t - 1)
are integers, 2 <= k < t, and b >= t.

When r = b * k / t and l = r * (k - 1) / (t - 1) are integers,
and k = t and b >= t, then a complete block design may be possible.

When r = b * k / t and l = r * (k - 1) / (t - 1) are integers,
and k < t and b >= t, then a balanced incomplete block design may be
possible.

When you specify options=ubd and r = b * k / t is an integer, then
unbalanced block design sizes are reported.  For example, if you want a
design with t=20 treatments and a block size of 6, you can run the
following to find out how many blocks you need.

   %mktbsize(t=20, k=6, options=ubd)

This step reports:

                                                                        n
          t          k          b            r          Lambda       Total
     Number of     Block    Number       Treatment    Pairwise       Sample
     Treatments    Size     of Blocks    Frequency    Frequencies    Size

         20          6          10           3           0.79          60

Then MktBIBD can be used to find a design where each treatment occurs 3
times, but the treatments do not appear together an equal number of
times.

   %mktbibd(t=20, k=6, b=10, seed=104)

                       Treatment by Treatment Frequencies

             1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20

         1   3  1  0  0  2  1  1  1  0  1  1  1  0  2  0  1  0  1  1  1
         2      3  0  1  1  1  1  1  1  1  0  1  1  0  1  0  1  0  2  1
         3         3  1  1  1  1  0  1  0  1  0  1  1  1  2  2  1  1  0
         4            3  0  1  1  1  0  1  1  1  1  0  1  1  1  1  2  0
         5               3  1  1  1  1  1  0  1  1  1  0  2  0  0  1  0
         6                  3  0  0  1  1  1  0  1  1  1  1  0  1  1  1
         7                     3  1  0  0  2  2  1  1  1  0  1  0  1  0
         8                        3  1  1  1  1  2  1  0  1  1  0  0  1
         9                           3  1  0  1  2  0  1  1  1  1  0  2
         10                             3  0  1  0  1  2  1  0  1  1  1
         11                                3  1  1  2  1  0  1  1  0  1
         12                                   3  1  0  1  0  0  1  1  1
         13                                      3  0  0  1  1  0  0  1
         14                                         3  1  1  1  1  0  1
         15                                            3  0  1  1  1  1
         16                                               3  1  1  1  0
         17                                                  3  1  1  1
         18                                                     3  1  2
         19                                                        3  0
         20                                                           3

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%macro mktbsize(  /*--------------------------------------------------*/
                  /*                                                  */
help,             /* nonblank, list options                           */
                  /*                                                  */
out=bibd,         /* specifies the output data set with the list of   */
                  /* potential design sizes.                          */
                  /*                                                  */
options=,         /* nocheck - by default, certain checks are         */
                  /* performed on b, t, and k.  Specify               */
                  /* options=nocheck to turn them off.  This would    */
                  /* allow you to do make some creative expressions   */
                  /* that otherwise would not be allowed.             */
                  /*                                                  */
                  /* ubd - results are reported when r = b * k / t is */
                  /* in integer but l = r * (k - 1) / (t - 1) may or  */
                  /* may not be an integer.  The design will not      */
                  /* necessarily be balanced. You can alternatively   */
                  /* specify options=ibd.  Use this option when you   */
                  /* want to see sizes where every treatment can      */
                  /* occur equally often, but the pair-wise           */
                  /* frequencies are allowed to be unequal.  The      */
                  /* listing may contain both sizes where a BIBD may  */
                  /* be possible (lambda is an integer) and sizes     */
                  /* where a BIBD is not possible (lambda is not an   */
                  /* integer).  This option can be useful when the    */
                  /* block design is being used to make a partial     */
                  /* profile design.                                  */
                  /*                                                  */
order=tkb,        /* specifies the order of the loops, the default is */
                  /* t then k then b.  If you specify expressions in  */
                  /* t=, b=, or k=, you may need some other ordering. */
                  /* Alternatively, you may want to specify this      */
                  /* option just to change the default ordering of    */
                  /* the results.                                     */
                  /*                                                  */
maxreps=1,        /* maximum number of replications.                  */
                  /*                                                  */
                  /*--------------------------------------------------*/
                  /* For these next arguments, specify either an      */
                  /* integer or a list of integers in the SAS DO list */
                  /* syntax.  Example:                                */
                  /*                                                  */
                  /* %mktbsize(b=t to 20, t=5 to 10, k=5)             */
                  /*                                                  */
b=,               /* specifies the number of blocks.  In the partial  */
                  /* profile context, this is the number of profiles. */
                  /* In the MaxDiff context, this is the number of    */
                  /* sets.  The default is b=2 to 500.                */
nsets=,           /* alias for B= with a name more relevant to the    */
                  /* MaxDiff and partial-profile context.             */
                  /*                                                  */
t=,               /* specifies the number of treatments.  In the      */
                  /* partial profile or MaxDiff setting, this is the  */
                  /* total number of attributes or messages.          */
nattrs=,          /* alias for T= with a name more relevant to the    */
                  /* MaxDiff and partial-profile context.  When the   */
                  /* nattrs= option is specified, the output will use */
                  /* the word "Attribute" rather than "Treatment" and */
                  /* "Set" rather than "Block".                       */
                  /*                                                  */
k=,               /* specifies the block size, or the number of       */
                  /* treatments in each block.  In the partial        */
                  /* profile or MaxDiff setting, this is the number   */
                  /* of attributes or messages shown at one time.     */
setsize=,         /* alias for K= with a name more relevant to the    */
                  /* MaxDiff and partial-profile context.             */
);                /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _mktver;
%local  i abort savenote saveopts setnote time localver l1 l2 l3 outopts
        ubd savedsnf timeopt attrortr;

%let time      = %sysfunc(datetime());
%let timeopt   = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote  = %sysfunc(getoption(notes));
%let saveopts  = &savenote %sysfunc(getoption(generic))
                 %sysfunc(getoption(&timeopt))
                 %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                 %sysfunc(getoption(serror));
%let abort     = 0;
%let setnote   = nonotes;
%let localver  = 25Jul2010;
%let _mktver   = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let i = %upcase(&mktopts);
   %if %index(&i, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&i, NOTES) %then %let setnote = notes;
   %if %index(&i, VERSION) %then %put MktBSize macro version &localver.;
   %end;
options serror;

options &setnote generic;

%if %nrbquote(&help) ne %then %do; %mkthelp(mktbsize) %goto endit; %end;

*------check options------;
data _null_;
   length w $ 80;
   abort = 0;

   *------resolve B= and alias NSETS=, set default------;
   o1 = (symget('b') ne ' ');
   o2 = (symget('nsets') ne ' ');
   if (o1 + o2) eq 0 then call symput('b', '2 to 500');
   else if (o1 + o2) ne 1 then do;
      put 'ERROR: You must specify exactly one of the following: '
          'B= or NSETS=.';
      abort = 1;
      end;

   else if o2 then call symput('b', symget('nsets'));

   *------resolve T= and alias NATTRS=------;
   o1 = (symget('t') ne ' ');
   o2 = (symget('nattrs') ne ' ');
   call symput('attrortr', scan("Treatment Attribute", 1 + o2));
   if (o1 + o2) ne 1 then do;
      put 'ERROR: You must specify exactly one of the following: '
          'T= or NATTRS=.';
      abort = 1;
      end;

   else if o2 then call symput('t', symget('nattrs'));

   *------resolve K= and alias SETSIZE=------;
   o1 = (symget('k') ne ' ');
   o2 = (symget('setsize') ne ' ');
   if (o1 + o2) ne 1 then do;
      put 'ERROR: You must specify exactly one of the following: '
          'K= or SETSIZE=.';
      abort = 1;
      end;

   else if o2 then call symput('k', symget('setsize'));

   w = lowcase(symget('options'));
   i = index(w, 'ubd'); if i then substr(w, i, 3) = ' ';
   call symput('ubd', put(i gt 0, 1.));
   i = index(w, 'ibd'); if i then substr(w, i, 3) = ' ';
   if i then call symput('ubd', '1');
   i = index(w, 'nocheck'); if i then substr(w, i, 7) = ' ';

   if w ne ' ' then do;
      w = compbl(w);
      put 'ERROR: OPTIONS=' w 'is invalid.';
      abort = 1;
      end;

   if not i then do;
      w = symget('b'); link check;
      w = symget('t'); link check;
      w = symget('k'); link check;
      if abort then put "ERROR: B=&b, T=&t, K=&k is invalid.";
      end;

   t = input(symget('maxreps'), ?? 32.);
   if nmiss(t) or t < 1 or t ne round(t) then do;
      put "ERROR: MAXREPS=%nrbquote(&maxreps) is invalid.";
      abort = 1;
      end;

   w = compress(lowcase(symget('order')));
   if length(w) ne 3 or compress(w, 'btk') ne ' ' then do;
      put "ERROR: ORDER=&order is invalid.";
      abort = 1;
      end;

   else do;
      c = substr(w, 1, 1); call symput('l1', c || ' = ' || symget(c));
      c = substr(w, 2, 1); call symput('l2', c || ' = ' || symget(c));
      c = substr(w, 3, 1); call symput('l3', c || ' = ' || symget(c));
      call symput('order', substr(w, 1, 1) || ' ' ||
                           substr(w, 2, 1) || ' ' || substr(w, 3, 1));
      end;

   if abort or _error_ then call symput('abort', '1');

   *------separate out data set options------;
   w = symget('out'); if w = ' ' then w = 'bibd';
   j = index(w, '('); if j = 0 then j = length(w) + 1;
   call symput('out', substr(w, 1, j - 1));
   call symput('outopts', substr(w, j));

   return;

   check:;
   w = lowcase(w);
   if w = ' ' then abort = 1;
   i = index(w, ' by ');
   if i then substr(w, i, 4) = ' ';
   i = index(w, ' to ');
   if i then substr(w, i, 4) = ' ';
   w = compress(w, '0123456789tbs.*+-/');
   if w ne ' ' then abort = 1;
   return;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: The data set name is invalid:;
   %put %str(       )out=&out;
   %end;

%if &abort %then %goto endit;

*------generate list------;
data &out(keep=b t k r l n);
   do &l1;
      do &l2;
         do &l3;
            if n(b, t, k) eq 3 and t gt 1 then do;
               r = b * k / t;
               l = r * (k - 1) / (t - 1);
               n = b * k;
               if b eq round(b) and t eq round(t) and k eq round(k) and
                  b gt 1 and k gt 1 and abs(r - round(r)) < 1e-6
                  %if not &ubd %then
                     and abs(l - round(l)) < 1e-6 and k lt t and b ge t;
                  then do;
                  i + 1;
                  output;
                  end;
               end;
            end;
         end;
      end;
   if i = 0 then do;
      put 'WARNING: ' %if &ubd %then 'A UBD'; %else 'A BIBD';
          " is not possible with b=&b, t=&t, k=&k..";
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%if %nrbquote(&maxreps) ne %then %do;
   proc sort data=&out; by t k; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data &out;
      set &out;
      by t k;
      if first.k then rep = 0;
      rep + 1;
      if rep le &maxreps then output;
      %if &maxreps = 1 %then %do; drop rep; %end;
      %else %do; label rep = '-Number of   -Replications'; %end;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc sort data=&out; by &order; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------display results------;
proc print noobs label split='-';
   label t = "-t         -Number of -&attrortr.s"
         %if &attrortr eq Treatment %then
         b = '-b        -Number   -of Blocks'
         k = '-k    -Block-Size -';
         %else
         b = '-b      -Number -of Sets'
         k = '-k   -Set -Size-';
         r = "-r        -&attrortr-Frequency-"
         l = '-Lambda     -Pairwise   -Frequencies'
         n = '-n     -Total -Sample-Size  -';
   format l best4.;
   run;

*------final data sets------;
options &savenote no&timeopt;

data &out &outopts; set &out; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTBSIZE macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTBSIZE macro used &time seconds.;
   %end;

%mendit: %mend;
