 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktDes                                              */
 /*   TITLE: Efficient Experimental Designs                      */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs                         */
 /*   PROCS: FACTEX, OPTEX, PLAN                                 */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: This file contains three macros.                    */
 /*          * MktDes   - finds efficient designs for linear     */
 /*                       models.                                */
 /*          * MktDes6  - used in conjunction with MktDes for    */
 /*                       six-level factors (obsolete).          */
 /*          * MktDes10 - used in conjunction with MktDes for    */
 /*                       ten-level factors (obsolete).          */
 /*                                                              */
 /*          The MktDes macro must be called, either directly    */
 /*          or indirectly through the MktEx macro, before the   */
 /*          obsolete MktDes6 or MktDes10 macros can be called.  */
 /*                                                              */
 /*          This macro creates global macro variables           */
 /*             &&_MKTCLASS&STEP &&_MKTINTER&STEP _MKTVER        */
 /*          For example, for STEP=1 it creates                  */
 /*             _MKTCLASS1 _MKTINTER1                            */
 /*                                                              */
 /*          The following SAS products are required to run      */
 /*          this macro: BASE, STAT, QC.                         */
 /****************************************************************/

 /*---------------------------------------------------------------------

This macro creates efficient experimental designs.  Usually you would
call the MktEx macro instead of calling the MktDes macro directly.  The
MktDes macro is used by the MktEx macro as one of its tools.  If you do
call the MktDes macro, you specify the names of the factors and the
number of levels for each factor.  You also specify the number of runs
you want in your final design.  Here for example is how you can create a
design in 18 runs with 2 two-level factors (x1 and x2) and 3 three-level
factors (x3, x4, and x5).

   %mktdes(factors=x1-x2=2 x3-x5=3, n=18)

You can also optionally specify interactions that you want to be
estimable.  The macro creates a candidate design in which every effect
you want to be estimable is estimable, but the candidate design is
bigger than you want.  By default, the candidate set is stored in a
SAS data set called Cand1.  The macro then uses PROC OPTEX to search the
candidate design for an efficient final design.  By default, the final
experimental design is stored in a SAS data set called Design.

When the full-factorial is small (under 5000 runs is small, although by
default any problem larger than the default big=2188 is considered
large), the experimental design problem is straight-forward.  First, the
macro uses PROC PLAN to create a full-factorial candidate set.  Then
PROC OPTEX searches the full-factorial candidate set and will almost
certainly find the optimal design or a very nearly optimal design given
sufficient iteration (for example, specify ITER=100).  Run time will
typically be a few seconds or a few minutes, but it could run up to an
hour.

When the full-factorial design is larger, the design problem is less
straight-forward and requires more thought.  Then the macro uses PROC
FACTEX to create a fractional-factorial candidate set.  The default
SIZE=, NLEV=, number of pseudo-factors, and so on will often give a good
design.  However, changing these values may increase or decrease
efficiency, 1, 2, 5, 10, perhaps up to 15%.  Usually you are better off
using the MktEx macro when the full factorial is larger than a few
thousand runs.

Here are some typical examples of usage:

   *------Six three-level factors in 18 runs------;
   %mktdes(factors=x1-x6=3, n=18)

   *------Two two-level factors and 3 three-level factors in 18 runs------;
   %mktdes(factors=x1-x2=2 x3-x5=3, n=18)

   *------Mixed 2-, 3-, 5-level factors, done in one step------;
   %mktdes(factors=x1-x3=2 x4-x6=3 x7-x9=5, n=30)

   *---Mixed 2-, 3-, 5-level factors, done in three steps------;
   %mktdes(factors=x1-x3=2, n=30, run=factex,       step=1)
   %mktdes(factors=x4-x6=3, n=30, run=factex,       step=2)
   %mktdes(factors=x7-x9=5, n=30, run=factex optex, step=3)

   *------A design with interactions------;
   %mktdes(factors=x1-x3=2 x4-x6=3(3) x7-x9=4, n=32, size=1024,
           interact=x1|x2|x3@2 x4*x7)

   *------Look at one-way and two-way frequencies------;
   %mkteval;

   *------evaluate the efficiency of a given design using options=check------;
   %mktdes(cand=design, factors=x1-x3=2 x4-x6=3(3) x7-x9=4,
           interact=x1|x2|x3@2 x4*x7, options=check)

   *------evaluate the efficiency for main effects model------;
   %mktdes(cand=design, factors=x1-x9, options=check)

If you will subsequently be using the MktRoll macro to make a choice
design from the linear design that MktDes produces, you can call
MktKey with a variable list to get all of the variable names.  For
example,

   %mktkey(x1-x15)

produces:

   x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15

which can be cut and pasted into a program like this to make the choice design:

   data key;
      input (Brand Price Size Color Shape Form) ($);
      datalines;
   A     x1  x2  x3  x4  x5
   B     x6  x7  x8  x9 x10
   C    x11 x12 x13 x14 x15
   None   .   .   .   .   .
   ;

   %mktdes(factors=x1-x15=3, n=36, out=des4)
   %mktroll(key=key, design=des4, out=choicedes, alt=brand)

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

%macro mktdes(    /*--------------------------------------------------*/
                  /* This macro is a front end to PROC FACTEX, PROC   */
                  /* PLAN, and PROC OPTEX, and is used to create      */
                  /* experimental designs, potentially with mixed     */
                  /* numbers of factor levels.                        */
                  /*                                                  */
help,             /* nonblank, list options                           */
                  /*                                                  */
                  /* The FACTORS= option must be specified.           */
                  /* All other options are optional.                  */
                  /*--------------------------------------------------*/
factors=,         /* Specifies the factors and the number of levels   */
                  /* for each factor.  Optionally, the number of      */
                  /* pseudo-factors can also be specified.  Here is a */
                  /* simple example -- create a design with 10        */
                  /* two-level factors.                               */
                  /*                                                  */
                  /*    FACTORS=x1-x10=2                              */
                  /*                                                  */
                  /* First a factor list, which is a valid SAS        */
                  /* variable list, is specified.  The factor list    */
                  /* MUST be followed by an equal sign and an         */
                  /* integer, which gives the number of levels.       */
                  /* Multiple lists may be specified.  For example,   */
                  /* to create 5 two-level factors, 5 three-level     */
                  /* factors, and 5 five-level factors, specify:      */
                  /*                                                  */
                  /*    FACTORS=x1-x5=2 x6-x10=3 x11-x15=5            */
                  /*                                                  */
                  /* By default, this macro creates each factor from  */
                  /* a minimum number of pseudo-factors.              */
                  /* Pseudo-factors are used to create factors of     */
                  /* interest but are not themselves output.  So for  */
                  /* example, with NLEV=2, a three-level factor x1 is */
                  /* created from 2 two-level pseudo-factors (_1 and  */
                  /* _2) and their interaction by coding down:        */
                  /* (_1=1,_2=1) -> x1=1  (_1=1,_2=2) -> x1=2         */
                  /* (_1=2,_2=1) -> x1=3  (_1=2,_2=2) -> x1=1         */
                  /* This creates imbalance -- the 1 level appears    */
                  /* twice as often as 2 and 3.  Somewhat better      */
                  /* balance can be obtained by instead using three   */
                  /* pseudo-factors.  The number of pseudo-factors is */
                  /* specified in parentheses after the number of     */
                  /* levels.  Example:                                */
                  /*                                                  */
                  /*    FACTORS=x1-x5=2 x6-x10=3(3).                  */
                  /*                                                  */
                  /* Then the levels 1 to 8 are coded down to 1 to 3, */
                  /* 1 to 3, 1, 3, which is a little better balanced. */
                  /* The cost is candidate set size may increase and  */
                  /* efficiency may actually decrease.                */
                  /*                                                  */
                  /* With OPTIONS=CHECK, only the variable list is    */
                  /* required:                                        */
                  /*                                                  */
                  /*    FACTORS=x1-x20.                               */
                  /*--------------------------------------------------*/
otherfac=,        /* Other terms to mention on the FACTORS statement  */
                  /* of PROC FACTEX.  These terms are not guaranteed  */
                  /* to be estimable.  The value of OTHERFAC= is a    */
                  /* variable list.                                   */
                  /*--------------------------------------------------*/
step=1,           /* Step number.  By default, each run through the   */
                  /* macro is the first step in a one step process.   */
                  /* However, sometimes, a better design can be found */
                  /* using a multi-step approach.  Do not specify     */
                  /* CAND= on any step of a multistep run.  Consider  */
                  /* the problem of making a design with 3 two-level  */
                  /* factors, 3 three-level factors, and 3 five-level */
                  /* factors.  The simplest approach is to do         */
                  /* something like this -- create a design from      */
                  /* two-level factors using pseudo-factors and       */
                  /* coding down.                                     */
                  /*                                                  */
                  /*    %mktdes(factors=x1-x3=2 x4-x6=3               */
                  /*         x7-x9=5, n=30)                           */
                  /*                                                  */
                  /* However, for small problems like this, the       */
                  /* following three-step approach will usually be    */
                  /* better.                                          */
                  /*                                                  */
                  /*    %mktdes(step=1, factors=x1-x3=2,              */
                  /*         n=30, run=factex)                        */
                  /*    %mktdes(step=2, factors=x4-x6=3,              */
                  /*         n=30, run=factex)                        */
                  /*    %mktdes(step=3, factors=x7-x9=5,              */
                  /*         n=30, run=factex optex)                  */
                  /*                                                  */
                  /* The first step uses PROC FACTEX to create a      */
                  /* fractional-factorial design for the two-level    */
                  /* factors.  The second step uses PROC FACTEX to    */
                  /* create a fractional-factorial design for the     */
                  /* three-level factors and cross it with the        */
                  /* two-level factors.  The third step uses PROC     */
                  /* FACTEX to create a fractional-factorial design   */
                  /* for the five-level factors and cross it with the */
                  /* design for the two and three-level factors and   */
                  /* then run PROC OPTEX.                             */
                  /*                                                  */
                  /* You cannot use PROC PLAN in second and           */
                  /* subsequent steps; you must use PROC FACTEX.      */
                  /*                                                  */
                  /* Each step stores globally two macro variables    */
                  /* &&_mktclass&step and &&_mktinter&step that are   */
                  /* used to construct the PROC OPTEX CLASS and MODEL */
                  /* statements.  When STEP > 1, variables from the   */
                  /* previous steps are used on the CLASS and MODEL   */
                  /* statements.  In the example above, the following */
                  /* OPTEX code is created by step 3:                 */
                  /*                                                  */
                  /*    proc optex data=Cand3;                        */
                  /*       class x1-x3 x4-x6 x7-x9;                   */
                  /*       model x1-x3 x4-x6 x7-x9;                   */
                  /*       generate n=30 iter=10 keep=5               */
                  /*                method=m_federov;                 */
                  /*       output out=Design;                         */
                  /*       run; quit;                                 */
                  /*                                                  */
                  /* This step uses the previously stored macro       */
                  /* variables CLASS1=x1=x3 and CLASS2=X4-x6.         */
                  /*--------------------------------------------------*/
size=min,         /* Candidate set size.  Start with the default      */
                  /* SIZE=min and see how big that is.  If you want,  */
                  /* subsequently you can specify larger values that  */
                  /* are NLEV=n multiples of the minimum size.  This  */
                  /* option is used with FACTEX:                      */
                  /*                                                  */
                  /*    SIZE DESIGN=&size;                            */
                  /*                                                  */
                  /* Say you specified NLEV=2 or the macro defaulted  */
                  /* to NLEV=2.  Increase the SIZE= value by a factor */
                  /* of two each time.  For example if SIZE=min       */
                  /* implies SIZE=128, then 256, 512, 1024, and 2048  */
                  /* are reasonable sizes to try.  You do not have to */
                  /* multiply out the numbers.  Instead, you can      */
                  /* specify a product.  For example, these are       */
                  /* equivalent:                                      */
                  /* SIZE=256 or SIZE=128*2 & SIZE=1024 or SIZE=128*8 */
                  /*--------------------------------------------------*/
nlev=,            /* Number of levels from which factors are          */
                  /* constructed through pseudo-factors and coding    */
                  /* down.  The value must be a prime or a power of   */
                  /* a prime: 2, 3, 4, 5, 7, 8, 9, 11 ...             */
                  /* This option is used with FACTEX:                 */
                  /*                                                  */
                  /*    FACTORS factors / NLEV=&nlev;                 */
                  /*                                                  */
                  /* By default, the macro uses the minimum prime or  */
                  /* power of a prime from the FACTORS= list or 2 of  */
                  /* no suitable value is found.                      */
                  /*--------------------------------------------------*/
cand=,            /* Output data set with candidate design (from      */
                  /* FACTEX or PLAN).  The default name is Cand&step, */
                  /* for example, Cand1 for step 1, Cand2 for step 2, */
                  /* and so on.  You should only use this option when */
                  /* you are reading an external candidate set.  When */
                  /* you specify STEP= values greater than 1, the     */
                  /* macro assumes the default candidate set names,   */
                  /* CAND1, CAND2, and so on, were used in previous   */
                  /* steps.  Specify just a data set name, no         */
                  /* data set options.                                */
                  /*--------------------------------------------------*/
where=,           /* Where clause for candidate design, used to       */
                  /* restrict the candidates.                         */
                  /*--------------------------------------------------*/
out=Design,       /* Output experimental design (from OPTEX).         */
                  /*--------------------------------------------------*/
procopts=,        /* PROC OPTEX statement options.                    */
                  /*--------------------------------------------------*/
facopts=,         /* PROC FACTEX statement options.                   */
                  /*--------------------------------------------------*/
coding=,          /* OPTEX CODING= option.  If you are using version  */
                  /* 7 or an earlier release of the SAS System and if */
                  /* you have a balanced and orthogonal candidate set */
                  /* you may want to specify CODING=orthcan.  With    */
                  /* version 8 of the SAS System, this option is      */
                  /* usually not needed.                              */
                  /*--------------------------------------------------*/
generate=,        /* OPTEX GENERATE statement options.                */
                  /*--------------------------------------------------*/
examine=,         /* OPTEX EXAMINE statement options.  Typical values */
                  /* are I and V for the information and variance     */
                  /* matrices.                                        */
                  /*--------------------------------------------------*/
method=,          /* OPTEX METHOD= option.  Search method.  Modified  */
                  /* Federov is the default.                          */
                  /*--------------------------------------------------*/
keep=5,           /* OPTEX KEEP= option.  Keep n designs.             */
                  /*--------------------------------------------------*/
maxiter=10,       /* OPTEX ITER= option.  Create n designs.           */
iter=10,          /* Alias for MAXITER=.                              */
                  /*--------------------------------------------------*/
n=,               /* OPTEX N= option.  Number of runs in the final    */
                  /* OPTEX-created design.  The default is the PROC   */
                  /* OPTEX default and depends on the problem.        */
                  /* Example: N=32                                    */
                  /*--------------------------------------------------*/
interact=,        /* Interactions that must be estimable.             */
                  /* Examples:                                        */
                  /*    INTERACT=x1*x2                                */
                  /*    INTERACT=x1*x2 x3*x4*x5                       */
                  /*    INTERACT=x1|x2|x3|x4|x5@2                     */
                  /*    INTERACT=@2                                   */
                  /*                                                  */
                  /* The interaction syntax is like GLM's and many of */
                  /* the other modeling procedures.  It uses "*" for  */
                  /* simple interactions (x1*x2, the interaction      */
                  /* between x1 and x2), "|" for main effects and     */
                  /* interactions (x1|x2|x3 is the same as x1 x2      */
                  /* x1*x2 x3 x1*x3 x2*x3 x1*x2*x3) and "@" to        */
                  /* eliminate higher-order interactions (x1|x2|x3@2  */
                  /* is the same as x1 x2 x1*x2 x3 x1*x3 x2*x3).      */
                  /* Specifying "@2" with a list like x1|x2|x3@2      */
                  /* allows only main effects and two-way             */
                  /* interactions without higher-order interactions.  */
                  /* If you specify "@2" by itself, a resolution V    */
                  /* design is requested.  Only "@" values of 2 or 3  */
                  /* are allowed.  By default, no interactions are    */
                  /* guaranteed to be estimable.  For the factor      */
                  /* names, you must specify either the actual        */
                  /* variable names (eg x1 x2 ...) or you can just    */
                  /* guaranteed to be estimable.                      */
                  /*--------------------------------------------------*/
otherint=,        /* Interaction terms that will only be specified    */
                  /* with PROC OPTEX for multi-step macro             */
                  /* invocations.  Normally, interactions that are    */
                  /* specified via the INTERACT= option affect both   */
                  /* the PROC FACTEX and the PROC OPTEX MODEL         */
                  /* statements.  In multi-step problems, part of an  */
                  /* interaction may not be in a particular PROC      */
                  /* FACTEX step.  In that case, the interaction term */
                  /* must only appear on the PROC OPTEX step.  For    */
                  /* example if x1 is created in one step and x4 is   */
                  /* created in another, and if the x1*x4 interaction */
                  /* must be estimable, specify OTHERINT=x1*x4 on the */
                  /* final step, the one that runs PROC OPTEX.        */
                  /*    %mktdes(step=1, factors=x1-x3=2, n=30,        */
                  /*            run=factex)                           */
                  /*    %mktdes(step=2, factors=x4-x6=3, n=30,        */
                  /*            run=factex)                           */
                  /*    %mktdes(step=3, factors=x7-x9=5, n=30,        */
                  /*            run=factex optex, otherint=x1*x4)     */
                  /*--------------------------------------------------*/
run=plan factex   /* Normally, the macro runs either FACTEX or PLAN   */
    optex,        /* and then OPTEX.  You can omit steps -- if for    */
                  /* example you only want the code listing -- by     */
                  /* omitting proc names from this list.  When both   */
                  /* PLAN and FACTEX are in the list, the macro       */
                  /* chooses between them based on the size of the    */
                  /* full factorial and the value of BIG=.  When PLAN */
                  /* is not in the list, the macro generates code for */
                  /* FACTEX.                                          */
                  /*--------------------------------------------------*/
big=2188,         /* If the size of the full factorial is less than   */
                  /* this size, and if PROC PLAN is in the RUN= list, */
                  /* the macro uses PROC PLAN instead of PROC FACTEX. */
                  /* (2188 = max(2 ** 11, 3 ** 7) + 1)                */
                  /*--------------------------------------------------*/
classopts=        /* Class statement options.  The default,           */
param=orthref,    /* param=orthref, uses an orthogonal coding in PROC */
                  /* OPTEX.  You probably never want to change this   */
                  /* option.                                          */
                  /*--------------------------------------------------*/
                  /* General boolean options:                         */
options=,         /* eval - ignored after V8.  Used to specify        */
                  /* CLASSOPTS=param=orthref, which is now the        */
                  /* default.                                         */
                  /*                                                  */
                  /* check - checks the efficiency of a given design, */
                  /* specified in CAND=.  This option sets ITER=1,    */
                  /* KEEP=1, METHOD=SEQUENTIAL, OUT=, and adds        */
                  /* INITDESIGN=&CAND to the GENERATE statement in    */
                  /* PROC OPTEX.  Also, when the RUN= option          */
                  /* specifies both FACTEX and OPTEX (the default),   */
                  /* then OPTIONS=CHECK sets RUN=OPTEX.  You can      */
                  /* specify RUN=FACTEX along with OPTIONS=CHECK to   */
                  /* run FACTEX and then report on the efficiency of  */
                  /* the FACTEX-generated design.  When OPTIONS=CHECK */
                  /* is specified, FACTORS= is allowed to contain     */
                  /* only a variable list.  The number of levels of   */
                  /* each factor is not required.                     */
                  /*                                                  */
                  /* nocode - don't print the plan/factex and optex   */
                  /* code generated.                                  */
                  /*                                                  */
                  /* allcode - show all code, even code that will not */
                  /* be run.  This used to be the default.            */
                  /*--------------------------------------------------*/
seed=0,           /* specifies the random number generation seed,     */
                  /* 0 - clock time.                                  */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _mktver;

%local abort class fullfac generic generic2 i t maxlen justcheck
       makeout maxp neffs nlists note1 note2 nnotes nterms primepow
       savenote setnote s2 s5 size time usedef savedsnf code outopts
       printseed checkeff saveopts r5 timeopt localver;

%let time      = %sysfunc(datetime());
%let timeopt   = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote  = %sysfunc(getoption(notes));
%let saveopts  = &savenote missing="%sysfunc(getoption(missing))"
                 %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                 %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                 %sysfunc(getoption(serror));
%let setnote   = nonotes;
%let printseed = 1;
%let abort     = 0;
%let localver  = 25Jul2010;
%let _mktver   = &localver;

                /* powers of primes */
%let primepow = 2 3 4 5 7 8 9 11 13 16 17 19 23 25 27 29 31 32 37
                41 43 47 49 53 59 61 64 67 71 73 79 81 83 89 97;

options noserror;
%if %nrquote(&&mktdesopts) eq %nrstr(&)mktdesopts %then %let mktdesopts = ;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let i = %upcase(&mktopts);
   %if %index(&i, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&i, NOTES)   %then %let setnote = notes;
   %if %index(&i, VERSION) %then %put MktDes macro version &localver.;
   %let printseed = (%index(&i, NOSEED) eq 0);
   %end;
options serror;

options &setnote generic missing='.';

%if %nrbquote(&help) ne %then %do; %mkthelp(mktdes) %goto endit; %end;

*------maximum length of character variables------;
%let maxlen = 32700;

*------override parameters?-------;
data _null_;
   length name $ 16 value $ 500 glob $ &maxlen;

   glob = left(symget('mktdesopts'));
   if glob ne ' ' then put 'Overridden Parameters:';
   do while(glob ne ' ');
      i = index(glob, '=');
      name = substr(glob, 1, i - 1);
      glob = left(substr(glob, i + 1));
      i = index(glob, ',');
      if i = 0 then i = length(glob) + 1;
      value = substr(glob, 1, i);
      value = translate(value, ' ', ',');
      glob  = left(substr(glob, i + 1));
      put +3 name +(-1) '=' value;
      call symput(name, trim(value));
      end;

   if not sysprod('qc') then do;
      put 'ERROR: SAS/QC must be licensed to run this macro.';
      call symput('abort', '1');
      end;

   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------check lengths, make sure they not too long------;
%if %length(%nrbquote(&factors)) > &maxlen or
    %length(%nrbquote(&interact)) > &maxlen %then %do;
   %put ERROR: FACTOR= or INTERACT= value length > &maxlen.;
   %let abort = 1;
   %goto endit;
   %end;

*------start getting set up: error checking, start parsing------;
data _null_;
   length str varlist factors class $ &maxlen levnpse part $ 40
          levs charind $ 8 q $ 2;
   abort = 0;

   if symget('facopts') ne ' '
      then call symput('facopts', ' ' || symget('facopts'));

   str = left(symget('classopts'));
   if str ne ' ' then call symput('classopts', trim(' / ' || str));

   *------resolve ITER=, MAXITER=------;
   part = symget('iter');
   if part ne symget('maxiter') and
      part ne '10' then call symput('maxiter', trim(part));

   *------check SIZE=------;
   part = left(symget('size'));

   if upcase(part) ne 'MIN' then do;
      min = 0; i = &size +0; if _error_ then goto endit;
      if n(i) and i eq round(i) and i > 1
         then call symput('size', trim(put(i, best12. -L)));
      else do;
         put 'ERROR: SIZE=' part 'is invalid.';
         abort = 1;
         goto endit;
         end;
      end;

   *------check seed out of procopts------;
   noseed = index(lowcase(symget('procopts')), 'seed');

   *------make sure factors are specified------;
   factors = symget('factors');
   if factors = ' ' then do;
      put 'ERROR: FACTORS= must be specified.';
      abort = 1;
      goto endit;
      end;

   *------quit if names appear to be quoted------;
   q = '"' || "'";
   interact = symget('interact');
   if indexc(factors, q) or indexc(interact, q) then do;
      put 'ERROR: n-literals are not supported.';
      abort = 1;
      goto endit;
      end;

   call symput('r5', put(interact eq '@2', 1.));

   *------upper case proc names------;
   part = symget('run');
   str  = upcase(part);
   call symput('run', trim(compbl(str)));
   runo = index(str, 'OPTEX' ); if runo then substr(str, runo, 5) = ' ';
   runf = index(str, 'FACTEX'); if runf then substr(str, runf, 6) = ' ';
   runp = index(str, 'PLAN'  ); if runp then substr(str, runp, 4) = ' ';
   if str ne ' ' then do;
      put 'ERROR: RUN=' part 'is invalid.';
      abort = 1;
      goto endit;
      end;

   *------upper case options------;
   str = upcase(symget('options'));
   call symput('options', trim(str));
   i = index(str, 'CHECK' );  if i then substr(str, i, 5) = ' ';
   i = index(str, 'EVAL'  );  if i then substr(str, i, 4) = ' ';
   i = index(str, 'NOCODE');  if i then substr(str, i, 6) = ' ';
   i = index(str, 'ALLCODE'); if i then substr(str, i, 7) = ' ';
   str = compbl(str);

   if str ne ' ' then do;
      put 'ERROR: OPTIONS=' str 'is invalid.';
      abort = 1;
      goto endit;
      end;

   *------check STEP=------;
   m = input(left(symget('step')), ?? 32.);
   if nmiss(m) or m ne round(m) or m < 0 then do;
      put 'ERROR: STEP= specification is invalid.  '
          'Positive integer expected.';
      abort = 1;
      goto endit;
      end;

   *------create default candidate data set name------;
   part = symget('cand');
   if part eq ' ' then part = compress('Cand' || symget('step'));
   call symput('cand', trim(part));

   *------set options for checking efficiency of a given design------;
   check = index(symget('options'), 'NOCODE');
   i     = index(symget('options'), 'ALLCODE');
   if check  then check = 0;  /* no code            */
   else if i then check = 2;  /* all code           */
   else           check = 1;  /* just executed code */
   if getoption('mprint') eq 'MPRINT' then check = 2;
   call symput('code', put(check, 1.));

   check = index(symget('options'), 'CHECK');
   call symput('justcheck', put(check ne 0, 1.));
   if check then do;
      call symput('maxiter', '1');
      call symput('keep', '1');
      call symput('method', 'sequential');
      call symput('out', '');
      call symput('generate', trim('initdesign=' || symget('cand') || ' ' ||
                              symget('generate')));

      if runf and runp then str = 'OPTEX';
      else do;
         if runf or runp then call symput('justcheck', '0');
         str = left(symget('run') || ' OPTEX');
         end;
      call symput('run', trim(str));
      end;

   *------separate out data set options------;
   str = symget('out');
   call symput('makeout', put(str ne ' ', 1.));
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('out', substr(str, 1, j - 1));
   call symput('outopts', trim(left(substr(str, j))));

   *------get rid of extra blanks in where clause------;
   call symput('where', compbl(symget('where')));

   *------generate PROC OPTEX statement option list------;
   str = 'data=' || trim(part) || ' ' || symget('procopts');
   part = left(symget('coding'));
   if part ne ' ' then str = trim(str) || ' coding=' || part;
   link compress;
   checkeff = not index(lowcase(symget('procopts')), 'noprint');
   call symput('procopts', trim(str));

   *------strip extra blanks from interact list------;
   str = left(symget('interact'));
   if str ne ' ' then do;
      link compress;
      call symput('interact', trim(str));
      end;

   part = left(symget('n'));
   i = input(part, ?? 32.);

   if upcase(part) ne 'SATURATED' and part ne ' ' and
      (nmiss(i) or i ne round(i) or i < 1) then do;
      put 'ERROR: N=' part 'is invalid.';
      abort = 1;
      goto endit;
      end;

   str = ' ';
   if n(i) or upcase(part) = 'SATURATED'
      then str = trim(str) || ' n=' || part;

   part = left(symget('maxiter'));
   i = input(part, ?? 32.);

   if part ne ' ' and (nmiss(i) or i ne round(i) or i < 1) then do;
      put 'ERROR: ITER=' part 'is invalid.';
      abort = 1;
      goto endit;
      end;

   if n(i) then str = trim(str) || ' iter=' || part;

   part = left(symget('keep'));
   i = input(part, ?? 32.);
   if part ne ' ' and (nmiss(i) or i ne round(i) or i < 1) then do;
      put 'ERROR: KEEP=' part 'is invalid.';
      abort = 1;
      goto endit;
      end;

   if n(i) then str = trim(str) || ' keep=' || part;
   part = left(symget('method'));
   if part = ' ' then part = 'm_federov';
   str = trim(left(str)) || ' method=' || part ||
         ' ' || left(symget('generate'));
   link compress;
   call symput('generate', trim(str));

   *------strip extra blanks from factors list------;
   str = left(factors);
   link compress;
   factors = str;

   *------default number of levels, try min levels from FACTORS=------;
   str = factors; nlev = .;
   do until(i = 0);
      i = index(str, '=');
      if i then do;
         str  = substr(str, i + 1);
         levs = scan(str, 1, ' (');
         m    = input(levs, ?? 32.);

         if nmiss(m) or m ne round(m) or m < 1 then do;
            put 'ERROR: FACTORS= level specification is invalid: "'
                 levs +(-1) '".';
            abort = 1;
            goto endit;
            end;

         if m in (&primepow) then nlev = min(nlev, m);
         end;
      end;

   *------if min is not prime power, just use 2------;
   if not (nlev in (&primepow)) then nlev = 2;

   *---if user specified NLEV= use it, otherwise use default------;
   usedef = 0;
   levs = symget('nlev');
   if levs ne ' ' then nlev = input(levs, ?? 32.);
   else usedef = 1;
   call symput('usedef', put(usedef, 1.));

   if nmiss(nlev) or nlev ne round(nlev) or nlev < 2 then do;
      put 'ERROR: NLEV=' levs 'is invalid.';
      abort = 1;
      goto endit;
      end;

   call symput('nlev', compress(put(nlev, 8.)));

   *------parse FACTORS= specification
          three macro variable lists are generated with
          one entry for each list in FACTORS=
          __list1, __list2, ... - variable lists like x1-x4
          __lev1 , __lev2 , ... - number of levels in this list
          __pseu1, __pseu2, ... - number of pseudo-factors to use

          also, a list of CLASS vars is stored in class macro var
          the number of lists is stored in nlists                ------;

   do until(varlist eq ' ');
      varlist = scan(factors, 1, '=');
      if varlist ne ' ' then do;
         m = index(factors, '=');
         if m = 0 and check then do;
            factors = trim(factors) || '=2';
             m = index(factors, '=');
            end;
         if m = 0 then do;
            put 'ERROR: Equal sign expected after "'
                varlist +(-1) '" in FACTORS=.';
            abort = 1;
            goto endit;
            end;

         nlists  + 1;
         factors = substr(factors, m + 1);
         levnpse = scan(factors, 1, ' ');
         levs    = scan(levnpse, 1, '()');
         pseudo  = input(scan(levnpse, 2, '()'), ?? 32.);
         if (indexc(levnpse, '()') and nmiss(pseudo)) or
            (n(pseudo) and pseudo ne round(pseudo)) then do;
            put 'ERROR: Invalid syntax "' levnpse +(-1) '" in FACTORS=.';
            abort = 1;
            goto endit;
            end;

         factors = left(substr(factors, length(levnpse) + 1));
         charind = compress(put(nlists, ?? 8.));

         m = input(levs, ?? 32.);
         if nmiss(pseudo) then pseudo = max(1, ceil(log(m) / log(nlev)));

         if (not (1 <= pseudo <= 5)) and symget('code') ne '0' then
            put 'WARNING: Number of pseudo-factors '
                pseudo 'is suspect.';

         call symput(compress('__list' || charind), trim(varlist));
         call symput(compress('__lev'  || charind), trim(levs));
         call symput(compress('__pseu' || charind),
                     compress(put(pseudo, 8.)));
         class = trim(class) || ' ' || varlist;
         end;
      end;

   call symput('nlists' , compress(charind));
   call symput('class'  , trim(left(class)));

   *------check SEED=------;
   generic = ((not index(symget('saveopts'), 'NOGENERIC'))  and
              getoption('stimer' ) eq 'NOSTIMER');
   if generic then call symput('generic', 'seed=072555');
   seed = input(symget('seed'), ?? 32.);
   if nmiss(seed) or seed ne round(seed) then do;
      put "ERROR: SEED=%nrbquote(&seed) is invalid.";
      abort = 1;
      end;

   *------run generically, with a constant seed,
          eliminates using a clock generated seed------;
   else if seed = 0 then do;
      if generic then call symput('generic2', 'seed=072555');
      end;

   *------set seed using clock time when negative------;
   else if not noseed then do;
      if seed < 0 then do;
         seed = ceil(1e7 * uniform(input(reverse(put(1e2 * datetime(), 13.0)),
                                         ?? 8.)));
         if &printseed then put 'Seed = ' seed;
         end;
      call symput('procopts', trim(symget('procopts') || ' seed=' ||
                                   put(seed, 12.0 -L)));
      end;

   *------check for no efficiency table------;
   checkeff = (checkeff and symget('justcheck') eq '1' and
               symget('code') eq '0');
   call symput('checkeff', put(checkeff, 1.));

   endit: if abort or _error_ then call symput('abort','1');
   return;

compress:  *------get rid of extra, leading, trailing blanks------;

   str = compbl(str);

   *------get rid of extra leading blanks------;
   do chars = ' *', ' |', ' @', ' -', ' (', ' )', ' =';
      do j = 1 to &maxlen until(done);
         done = 1;
         i = index(str, chars);
         if i then do;
            done = 0;
            str = substr(str, 1, i-1) || substr(str, i+1);
            end;
         end;
      if j >= &maxlen then do;
         put 'ERROR: Invalid specification: ' str '.';
         call symput('abort', '1');
         goto endit;
         end;
      end;

   *------get rid of extra trailing blanks------;
   do chars = '* ', '| ', '@ ', '- ', '( ', '= ';
      do j = 1 to &maxlen until(done);
         done = 1;
         i = index(str, chars);
         if i then do;
            done = 0;
            str = substr(str, 1, i) || substr(str, i+2);
            end;
         end;
      if j >= &maxlen then do;
         put 'ERROR: Invalid specification: "' str +(-1) '".';
         call symput('abort', '1');
         goto endit;
         end;
      end;

   return;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&cand) ne %then %do;
   data _null_; set &cand; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )cand=&cand;
   %put %str(       )out=&out;
   %end;

%if &abort %then %goto endit;

*------continue parsing, generate statement fragments------;
data _null_;
   length term newterm inlist factors maineff nvals
          interact inteff $ &maxlen termpart onevar $ 32;

   *------expand variable lists: x1-x5 -> x1 x2 x3 x4 x5------;
   fullfac = 1;
   %do i = 1 %to &nlists;
      levs = &&__lev&i;

      array list&i[*] &&&__list&i;
      inlist = ' ';
      do j = 1 to dim(list&i);
         call vname(list&i[j], onevar);
         inlist = trim(inlist) || ' ' || onevar;
         fullfac = fullfac * levs;
         end;
      maineff = trim(maineff) || ' ' || left(inlist);
      call symput("__list&i", trim(inlist));
      %end;

   *------is this a big full-factorial?------;
   termpart = left(symget('big'));
   big      = input(termpart, ?? 32.);
   if nmiss(big) or big ne round(big) or big < 0 then do;
      put 'ERROR: BIG=' termpart 'is invalid.';
      call symput('big', '1');
      call symput('abort', '1');
      stop;
      end;

   big = (fullfac >= big or not index(symget('run'), 'PLAN'));
   call symput('big', put(big, 1.));

   if big and index(symget('run'), 'PLAN') and
      not index(symget('run'), 'FAC') then do;
      put 'ERROR: This problem is too big for PROC PLAN.'
          / +7   'Increase the value of BIG= or change your approach.';
      call symput('abort', '1');
      end;

   if fullfac < 1e15
      then termpart = put(fullfac, comma32.);
      else termpart = put(fullfac, best8. );
   call symput('note1', 'NOTE: The full-factorial design has ' ||
               compress(termpart) || ' runs.');
   if &usedef and big then do;
      call symput('note2', "NOTE: NLEV=%nrbquote(&nlev) assumed.");
      call symput('nnotes', '2');
      end;
   else call symput('nnotes', '1');

   *------generate FACTEX MODEL, OUTPUT statement syntax
          inlist - expanded variable list
          n      - number of pseudo-factors to base factor on
          levs   - number of levels in this factor
          the factex link generates the factex statement fragments-----;

   %do i = 1 %to &nlists;
      inlist = symget("__list&i");
      n      = &&__pseu&i;
      levs   = &&__lev&i;
      if big then link factex;
      else        link plan;
      %end;

   *------handle interactions------;
   interact = symget('interact');
   neffs    = nterms;

   if &r5 then do;
      interact = ' ';
      %do i = 1 %to &nlists;
         interact = trim(interact) || ' ' || symget("__list&i");
         %end;
      interact = left(compbl(interact));
      interact = translate(trim(interact),'|',' ') || '@2';
      end;

   if interact ne ' ' and big then do;

      *------replace '@n' specification with component parts------;
      do i = 1 to &maxlen until(term eq ' ');
         term = scan(interact, i, ' ');
         if term ne ' ' then do;
            m = index(term,'@');
            if m then do;
               degree = input(substr(term, m + 1), ?? 5.);
               m = 0;
               do j = 1 to &maxlen until(termpart eq ' ');
                  termpart = scan(term, j, '*|');
                  if termpart ne ' ' then m + 1;
                  end;
               if degree = 2 then do;
                  do j = 2 to m;
                     do k = 1 to j - 1;
                        newterm = trim(scan(term, k, '*|@')) || '|' ||
                                  trim(scan(term, j, '*|@'));
                        link maketerm;
                        end;
                     end;
                  end;
               else if degree = 3 then do;
                  do j = 2 to m;
                     do k = 1 to j - 1;
                        do l = 1 to k - 1;
                           newterm = trim(scan(term, l, '*|@')) || '|'
                                  || trim(scan(term, k, '*|@')) || '|'
                                  || trim(scan(term, j, '*|@'));
                           link maketerm;
                           end;
                        end;
                     end;
                  end;
               else do;
                  put "ERROR: Either '@2' or '@3' expected.";
                  call symput('abort', '1');
                  stop;
                  end;
               end;
            else do;
               newterm = term;
               link maketerm;
               end;
            end;
         end;
      end;

   *------store numbers of macro variables generated------;
   call symput('interact', trim(left(interact)));
   call symput('neffs'   , compress(put(neffs, 5.)));
   call symput('nterms'  , compress(put(nterms, 5.)));
   call symput('maxp'    , compress(put(p, 5.)));
   call symput('fullfac' , compress(put(fullfac, best32.)));

   if upcase(symget('size')) ne 'MIN' then do;
      min = 0;
      toomany = ceil(((1 + nparms * (&nlev - 1)) - (&size +0)) / (&nlev - 1));
      if toomany gt 0 then call symput('abort', '2');
      end;

   if _error_ then call symput('abort','1');
   stop;

plan: *------generate PLAN effects, OUTPUT terms/coding------;

   *------generate syntax fragments------;
   efflist = ' '; outlist = ' ';
   if inlist ne ' ' then do i = 1 to &maxlen until(onevar eq ' ');
      onevar = scan(inlist, i, ' ');
      if onevar ne ' ' then do;
         factors = ' '; nterms + 1;
         factors = trim(onevar) || '=' || compress(put(levs,5.));
         call symput(compress('__eff' || put(nterms,5.)),
                     trim(factors));
         end;
      end;
   return;

factex: *------generate FACTEX nvals, effects, OUTPUT terms/coding-----;

   vals  = &nlev ** n; /* number of values to code down */
   nvals = ' ';

   do while(vals >= (levs - 1e-8));
      if levs <= 3 then nvals = trim(nvals) || substr(' 1 2 3', 1, 2 * levs);
      else nvals = trim(nvals) || ' 1 to ' || compress(put(levs, 8.));
      vals + -levs;
      end;

   if vals > 1e-8 then do;
      if vals < (1+1e-8) then nvals = trim(nvals) || ' 1';
      else do k = 1 to levs by (levs - 1) / (vals - 1);
         nvals = trim(nvals) || ' ' || compress(put(k, 8.));
         end;
      end;

   nvals = left(nvals);

   *------generate syntax fragments------;
   efflist = ' '; outlist = ' ';
   if inlist ne ' ' then do i = 1 to &maxlen until(onevar eq ' ');
      onevar = scan(inlist, i, ' ');
      if onevar ne ' ' then do;
         factors = ' '; nterms + 1;
         do j = 1 to n;
            p + 1;
            factors = trim(factors) || ' _' || compress(put(p, 5.));
            end;
         nparms + n;
         factors = left(factors);
         call symput('__out' || compress(put(nterms,5.)), '[' ||
                     trim(factors) || ']=' || trim(onevar) ||
                     ' nvals=(' || trim(nvals) || ')');
         call symput(compress('__eff' ||put(nterms,5.)),
                     translate(trim(factors), '|', ' '));
         end;
      end;
   return;

maketerm: *------generate extra FACTEX MODEL interaction effects------;
   neffs + 1; inteff = ' ';
   do mj = 1 to &maxlen until(termpart eq ' ');
      termpart = scan(newterm, mj, '*|');
      if termpart ne ' ' then do;
         do mk = 1 to &maxlen until(onevar eq ' ');
            onevar = scan(maineff, mk, ' ');
            if upcase(onevar) = upcase(termpart) then do;
               if mj = 1 then sep = ' '; else sep = '|';
               inteff = trim(inteff) || sep ||
                        symget(compress('__eff' || put(mk, 5.)));
               onevar = ' ';
               end;
            end;
         end;
      end;
   call symput('__eff' || compress(put(neffs,5.)),
               trim(left(inteff)));
   return;
   run;

%if &syserr > 4 %then %let abort = 1;

%if %index(%nrbquote(&run), FAC) and &big and &abort eq 2 %then %do;
   data &cand; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;

options &savenote;

%if %index(%nrbquote(&run), FAC) and &big and &abort eq 2 %then %do;
   %if &code %then %do;
      %let i = ERROR;
      %let abort = 1;
      %end;
   %else %do;
      %let i = NOTE;
      %let abort = 0;
      %end;
   %put &i: Invalid size specification.;
   %put &i: Either increase the size or drop some factors.;
   %goto endit;
   %end;

%if not &justcheck %then %do;
   %put &note1;
   %if &nnotes eq 2 %then %put &note2;
   %end;

%if &abort %then %goto endit;

*------store OPTEX syntax results from this step------;
%global _mktclass&step _mktinter&step;
%let _mktclass&step = &class;
%let _mktinter&step = &interact &otherint;

*------fix up WHERE clause, OTHERFAC------;
%if %nrbquote(&where) eq %then %let where = ;
%else %if &big %then %let where = %str( )where=(%nrbquote(&where));
%else                %let where = (where=(%nrbquote(&where)));
%if %nrbquote(&otherfac) eq
   %then %let otherfac = ;
   %else %let otherfac = %trim(&otherfac)%str( );

%let s2 = %str( )%str( );
%let s5 = &s2 &s2;

*------check for incorrect use of PROC PLAN------;
%if &step > 1 and not &big %then %do;
   %put ERROR: PROC PLAN may not be used in second and subsequent steps.;
   %put NOTE: Specify "run=factex" for intermediate steps.;
   %put &s5 Specify "run=factex optex" for the final step.;
   %let abort = 1;
   %goto endit;
   %end;

*------report generated code------;
%if &code %then %do;
   %put NOTE: The MKTDES macro generated the following code:;
   %put &s2;
   %if &big and not &justcheck %then %do;
      %if &code eq 2 or %index(%nrbquote(&run), FAC) %then %do;
         %put proc factex&facopts%str(;);
         %put &s2 factors &otherfac._1-_&maxp / nlev=&nlev%str(;);
         %put &s2 size design=&size%str(;);
         %if &r5 %then %put &s2 model resolution=5%str(;);
         %else %do;
            %put &s2 model estimate=(;
            %do i = 1 %to &neffs; %put &s5 &&__eff&i; %end;
            %put &s5 )%str(;);
            %end;
         %put &s2 output out=&cand(drop=_:&where);
         %if &step > 1 %then %put &s5 pointrep=Cand%eval(&step - 1);
         %do i = 1 %to &nterms; %put &s5 &&__out&i; %end;
         %put &s5 %str(;);
         %put &s2 %str(run; quit;);
         %put &s2;
         %end;
      %end;

   %else %if not &justcheck %then %do;
      %if &code = 2 or %index(%nrbquote(&run), PLAN) %then %do;
         %put proc plan ordered%str(;);
         %put &s2 factors;
         %do i = 1 %to &neffs; %put &s5 &&__eff&i; %end;
         %put &s5 / noprint%str(;);
         %put &s2 output out=&cand&where%str(;);
         %put &s2 %str(run; quit;);
         %put &s2;
         %end;
      %end;

   %if &code = 2 or %index(%nrbquote(&run), OPT) %then %do;
      %put proc optex &procopts%str(;);
      %if &step = 1 %then %do;
         %put &s2 class &_mktclass1&classopts%str(;);
         %if %nrbquote(&_mktinter1) ne
            %then %put &s2 model &_mktclass1 &_mktinter1%str(;);
            %else %put &s2 model &_mktclass1%str(;);
         %end;
      %else %do;
         %put &s2 class;
         %do i = 1 %to &step;
            %put &s5 &&_mktclass&i;
            %end;
         %put &s5&classopts%str(;);
         %put &s2 model;
         %do i = 1 %to &step;
            %put &s5 &&_mktclass&i &&_mktinter&i;
            %end;
         %put &s5 %str(;);
         %end;
      %put &s2 generate &generate%str(;);
      %if %nrbquote(&outopts) eq %then %let outopts = ;
      %if &makeout %then %put &s2 output out=&out&outopts%str(;);
      %if %nrbquote(&examine) ne %then %put &s2 examine &examine%str(;);
      %put &s2 %str(run; quit;);
      %put &s2;
      %end;
   %end;

*------make the candidate set------;
%if %index(%nrbquote(&run), FAC) and &big %then %do;
   %put NOTE: Starting the PROC FACTEX step.;
   options &setnote;
   data &cand; stop; run;
   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   options &savenote;
   run;

   *------suppress errors when called from mktex------;
   %if not &code %then %do;
      filename __dummy dummy;
      proc printto log=__dummy; run;
      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   proc factex &facopts;
      factors &otherfac._1-_&maxp / nlev=&nlev;
      size    design=&size;
      %if &r5 %then %do; model resolution=5; %end;
      %else %do;
         model estimate=(%do i = 1 %to &neffs; &&__eff&i %end;);
         %end;
      output  out=&cand(drop=_:&where)
              %if &step > 1 %then pointrep=Cand%eval(&step - 1);
              %do i = 1 %to &nterms; &&__out&i %end;;
      quit;

   %if &syserr > 4 %then %let abort = 1;

   %if not &code %then %do;
      proc printto log=log; run;
      %if &syserr > 4 %then %let abort = 1;
      %end;

   options &savenote;

   %if &abort %then %goto endit;
   %end;

%else %if %index(%nrbquote(&run), PLAN) and not &big %then %do;
   options &setnote;
   data &cand; stop; run;
   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   options &savenote;

   %put NOTE: Starting the PROC PLAN step.;

   *------suppress errors when called from mktex------;
   %if not &code %then %do;
      filename __dummy dummy;
      proc printto log=__dummy; run;
      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %if %sysfunc(sysprod(stat)) %then %do;
      proc plan ordered &generic;
         factors %do i = 1 %to &neffs; &&__eff&i %end; / noprint;
         output out=&cand&where;
         quit;
      %end;

   %else %do;
      data &cand&where;
         %do i = 1 %to &neffs;
            %let t = %index(&&__eff&i,=);
            do %substr(&&__eff&i,1,&t) 1 to %substr(&&__eff&i,&t+1);
            %end;
         output;
         %do i = 1 %to &neffs; end; %end;
         if _error_ then call symput('abort','1');
         run;
      %end;
   %if &syserr > 4 %then %let abort = 1;

   %if not &code %then %do;
      proc printto log=log; run;
      %if &syserr > 4 %then %let abort = 1;
      %end;
   %if &abort %then %goto endit;
   %end;

*------find a good design, evaluate its efficiency------;
%if %index(%nrbquote(&run), OPT) %then %do;
   %if %nrbquote(&out) ne %then %do;
      options &setnote;
      data &out; stop; run;
      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      options &savenote;
      %end;

   *------find a good design------;
   %put NOTE: Starting the PROC OPTEX step.;

   *------suppress errors when called from mktex------;
   %if not &code %then %do;
      filename __dummy dummy;
      proc printto log=__dummy; run;
      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %if &checkeff %then %do;
      options &setnote;
      data __tmpdeff; if 0 then output; run;

      %if &syserr > 4 %then %let abort = 1;
      options &savenote;

      ods output efficiencies=__tmpdeff;
      %end;

   proc optex &generic2 &procopts;
      class %do i = 1 %to &step; &&_mktclass&i %end; &classopts;
      model %do i = 1 %to &step; &&_mktclass&i &&_mktinter&i %end;;
      generate &generate;
      %if &makeout %then %do; output out=&out&outopts; %end;
      %if %nrbquote(&examine) ne %then %do; examine &examine; %end;
      quit;

   %if &syserr > 4 %then %let abort = 1;

   %if not &code %then %do;
      proc printto log=log; run;
      %if &syserr > 4 %then %let abort = 1;
      %end;

   %if &abort %then %goto endit;

   %if &checkeff %then %do;
      options &setnote;
      data _null_;
         if 0 then set __tmpdeff nobs=n;
         if n = 0 then do;
            designnumber = 1; Seed = .; DCriterion = .; ACriterion = .;
            GCriterion = .; AvePredStdErr = .;
            file print ods=(template='QC.OPTEX.Efficiencies');
            put _ods_;
            end;
         if _error_ then call symput('abort', '1');
         stop;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      data _null_;
         if 0 then set __tmpdeff nobs=n;
         if n = 0 then do;
            file log;
            put 'WARNING: PROC OPTEX probably '
                   'cannot estimate the model parameters.' /
                +9 'The efficiency is probably zero.' /
                +9 'Efficiency values in the iteration history are ridged.' /
                +9 'That is why they typically will be larger than zero.';
            end;
         if _error_ then call symput('abort', '1');
         stop;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      proc datasets nolist; delete __tmpdeff; run; quit;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      options &savenote;
      %end;

   %end;

%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTDES macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTDES macro used &time seconds.;
   %end;

%mendit: %mend;

*------Make a six-level factor from a two-level and a three-level------;
%macro mktdes6(two=, three=);
   array _2_6[*] _i &two;
   array _3_6[*] _i &three;
   if dim(_2_6) ne dim(_3_6) then do;
      put 'ERROR: TWO= and THREE= list must contain '
          'the same number of variables.';
      stop;
      end;
   do _i = 2 to dim(_2_6);
      if not (1 le _2_6[_i] le 2 and 1 le _3_6[_i] le 3 and
              _2_6[_i] eq round(_2_6[_i]) and
              _3_6[_i] eq round(_3_6[_i])) then do;
         put 'ERROR: Consecutive integers starting with 1 are required.';
         _error_ = 1;
         abort;
         end;
      _2_6[_i] = 3 * (_2_6[_i] - 1) + _3_6[_i];
      end;
   drop &three _i;
   %mend;

*------Make a ten-level factor from a two-level and a five-level------;
%macro mktdes10(two=, five=);
   array _2_10[*] _i &two;
   array _5_10[*] _i &five;
   if dim(_2_10) ne dim(_5_10) then do;
      put 'ERROR: TWO= and FIVE= list must contain '
          'the same number of variables.';
      stop;
      end;
   do _i = 2 to dim(_2_10);
      if not (1 le _2_10[_i] le 2 and 1 le _5_10[_i] le 5 and
              _2_10[_i] eq round(_2_10[_i]) and
              _5_10[_i] eq round(_5_10[_i])) then do;
         put 'ERROR: Consecutive integers starting with 1 are required.';
         _error_ = 1;
         abort;
         end;
      _2_10[_i] = 5 * (_2_10[_i] - 1) + _5_10[_i];
      end;
   drop &five _i;
   %mend;
