 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktDups                                             */
 /*   TITLE: Check Experimental Design for Duplicate Runs or Sets*/
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs, choice modeling        */
 /*   PROCS:                                                     */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: This macro creates global macro variables           */
 /*             _MKTDUPSSETS _MKTDUPSALTS _MktVer                */
 /*                                                              */
 /*          The following SAS products are required to run      */
 /*          this macro: BASE.                                   */
 /****************************************************************/

 /*---------------------------------------------------------------------

This macro detects duplicate choice sets and duplicate alternatives
within generic choice sets.  For example, consider a simple experiment
with these two choice sets:

   a b c       a b c
   1 2 1       1 1 1
   2 1 2       2 2 2
   1 1 2       2 2 1
   2 1 1       1 2 2

These choice sets are completely different and are not duplicates.
Now consider these two choice sets:

   a b c       a b c
   1 2 1       2 1 2
   2 1 2       1 1 2
   1 1 2       2 1 1
   2 1 1       1 2 1

They are the same for a generic study because all of the same alternatives
are there, they are just in a different order.  However, for a branded
study they are different, because for a branded study there would be a
different brand for each alternative so the choice sets would be the
same only if all the same alternatives appeared in the same order.
For both a branded and generic study, these choice sets are duplicates:

   a b c       a b c
   1 2 1       1 2 1
   2 1 2       2 1 2
   1 1 2       1 1 2
   2 1 1       2 1 1

Now consider these choice sets for a generic study.

   a b c       a b c
   1 2 1       1 2 1
   2 1 1       1 2 1
   1 1 2       1 1 2
   2 1 1       2 1 1

First, each of these choice sets has duplicate alternatives (2 1 1 in
the first and 1 2 1 in the second).  Second, these two choice sets are
flagged as duplicates, even though they are not exactly the same.  They
are flagged as duplicates because every alternative in choice set one is
also in choice set two, and every alternative in choice set two is also
in choice set one.  In generic studies, choice sets are considered
duplicates unless either choice set has at least one alternative that is
not in the other choice set.

Sample Usage:

Create a design with the choice set swapping algorithm for a branded
study.  Check for and eliminate duplicate choice sets.

   %mktex(3 ** 9, n=27)

   data key;
      input (Brand x1-x3) ($);
      datalines;
   Acme   x1 x2 x3
   Ajax   x4 x5 x6
   Widgit x7 x8 x9
   ;

   %mktroll(design=randomized, key=key, alt=brand, out=cand);

   %choiceff(data=cand, model=class(brand x1-x3), seed=420,
             nsets=18, nalts=3, beta=zero);

   proc freq; tables set; run;

   %mktdups(branded, data=best, factors=brand x1-x3, nalts=3, out=out)

   proc freq; tables set; run;

The first PROC FREQ output shows us that candidate choice set 25 occurs
twice in the design.

                              The FREQ Procedure

                                           Cumulative    Cumulative
           Set    Frequency     Percent     Frequency      Percent
           --------------------------------------------------------
             2           3        5.56             3         5.56
             3           3        5.56             6        11.11
             4           3        5.56             9        16.67
             5           3        5.56            12        22.22
             6           3        5.56            15        27.78
             7           3        5.56            18        33.33
            12           3        5.56            21        38.89
            13           3        5.56            24        44.44
            14           3        5.56            27        50.00
            16           3        5.56            30        55.56
            17           3        5.56            33        61.11
            18           3        5.56            36        66.67
            22           3        5.56            39        72.22
            23           3        5.56            42        77.78
            25           6       11.11            48        88.89
            26           3        5.56            51        94.44
            27           3        5.56            54       100.00

The output contains the following tables:

Design:          Branded
Factors:         brand x1-x3
                 Brand
                 x1 x2 x3
Duplicate Sets:  1

                                        Duplicate
                            Choice    Choice Sets
                               Set      To Delete

                                 2              9


The first line of the first table tells us that this is a branded design
as opposed to generic.  The second line tells us the factors as
specified on the FACTORS= option.  These are followed by the
actual variable names for the factors.  The last line reports the number
of duplicates. The second table tells us that the second choice set is
the same as the ninth (which we know from above is candidate choice set
25).  The OUT= data set will contain the design with the duplicate
choice set eliminated.

Now consider an example with purely generic alternatives.

   %mktex(2 ** 5, n=2**5)
   %mktlab(int=f1-f4, values=1 2)

   %choiceff(data=final, model=class(x1-x5), seed=109,
             nsets=42, flags=f1-f4, beta=zero);

   %mktdups(generic, data=best, factors=x1-x5, nalts=4, out=out)

The macro produces the following tables:

Design:          Generic
Factors:         x1-x5
                 x1 x2 x3 x4 x5
Sets w Dup Alts: 2
Duplicate Sets:  3


                                   Duplicate
                       Choice    Choice Sets
                          Set      To Delete

                            3             36

                            5                   Alternatives

                           10             26

                           24                   Alternatives

                           31             38

For each choice set listed in the choice set column, either the other
choice sets it duplicates are listed or the word 'Alternatives' is
printed if the problem is with duplicate alternatives.

Here are just the choice sets with duplication problems:

   Set   Index   x1 x2 x3 x4 x5

     3      4     1  1  1  2  2
            7     1  1  2  2  1
           25     2  2  1  1  1
           30     2  2  2  1  2

     5      7     1  1  2  2  1
            7     1  1  2  2  1
           26     2  2  1  1  2
           26     2  2  1  1  2

    10     25     2  2  1  1  1
           12     1  2  1  2  2
           23     2  1  2  2  1
            6     1  1  2  1  2

    24     17     2  1  1  1  1
           17     2  1  1  1  1
           16     1  2  2  2  2
           16     1  2  2  2  2

    26     12     1  2  1  2  2
           23     2  1  2  2  1
           25     2  2  1  1  1
            6     1  1  2  1  2

    31     13     1  2  2  1  1
           11     1  2  1  2  1
           24     2  1  2  2  2
           18     2  1  1  1  2

    36      4     1  1  1  2  2
            7     1  1  2  2  1
           25     2  2  1  1  1
           30     2  2  2  1  2

    38     18     2  1  1  1  2
           13     1  2  2  1  1
           24     2  1  2  2  2
           11     1  2  1  2  1

You can see that the macro detects duplicates even though the
alternatives do not always appear in the same order in the different
choice sets.

Now consider another example.

   %mktex(2 ** 6, n=2**6)

   data key;
      input (x1-x2) ($) @@;
      datalines;
   x1 x2 x3 x4 x5 x6
   ;

   %mktroll(design=randomized, key=key, out=cand);

   %mktdups(generic, data=cand, factors=x1-x2, nalts=3, out=out)

proc print; by set; id set; run;

Here is some of the output.  The output lists for each set of duplicates
the choice set that will be kept (in the first column) and all the
matching choice sets that will be deleted in the second column.

Design:          Generic
Factors:         x1-x2
                 x1 x2
Sets w Dup Alts: 40
Duplicate Sets:  50



                                   Duplicate
                       Choice    Choice Sets
                          Set      To Delete

                            1                   Alternatives

                            2                   Alternatives
                                           5
                                           6
                                          17
                                          18
                                          21

                            3                   Alternatives
                                           9
                                          11
                                          33
                                          35
                                          41

                            4                   Alternatives
                                          13
                                          16
                                          49
                                          52
                                          61

                            5                   Alternatives

                            6                   Alternatives

                            7             10
                                          19
                                          25
                                          34
                                          37


Here are the nonduplicate choice sets.

                           Set    _Alt_    x1    x2

                             7      1       1     1
                                    2       1     2
                                    3       2     1

                             8      1       1     1
                                    2       1     2
                                    3       2     2

                            12      1       1     1
                                    2       2     1
                                    3       2     2

                            28      1       1     2
                                    2       2     1
                                    3       2     2


This next example creates a conjoint design and tests it for duplicates.

 %mktex(3 ** 3 2 ** 2, n=19, seed=513)
 %mktdups(linear, factors=x1-x5);

Design:          Linear
Factors:         x1-x5
                 x1 x2 x3 x4 x5
Duplicate Runs:  1


                                        Duplicate
                                             Runs
                               Run      To Delete

                                 9             10

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

%macro mktdups(   /*--------------------------------------------------*/
                  /*                                                  */
                  /*---------------Positional Parameter---------------*/
                  /*                                                  */
options,          /* Positional parameter for binary options.         */
                  /* Specify either generic or branded but not both.  */
                  /*                                                  */
                  /*     generic - (default) generic design.  This    */
                  /* means that there are no brands, so options are   */
                  /* interchangeable, so the macro needs to compare   */
                  /* each alternative with every other alternative in */
                  /* every choice set.                                */
                  /*                                                  */
                  /*     branded - Since one of the factors is brand, */
                  /* the macro only needs to compare corresponding    */
                  /* alternatives in each choice set.                 */
                  /*                                                  */
                  /*      linear - not a choice design.  Specify      */
                  /* linear for a full-profile conjoint design, for   */
                  /* an ANOVA design, or for the linear version of a  */
                  /* branded choice design.                           */
                  /*                                                  */
                  /*     noprint - Do not print anything.  This       */
                  /* option will be used when you are only interested */
                  /* in the output data set or macro variable.        */
                  /*                                                  */
                  /* Example: %mktdups(branded noprint, nalts=3)      */
                  /*                                                  */
                  /* If you do not specify options, you must still    */
                  /* leave a space for it: %mktdups( , nalts=3).      */
                  /*                                                  */
                  /* ? or help - lists options                        */
                  /*                                                  */
                  /*-------(Usually) Mandatory Keyword Parameter------*/
                  /*                                                  */
nalts=,           /* specifies the number of alternatives.  This      */
                  /* option must be specified with generic or branded */
                  /* designs.  It is ignored with linear designs.     */
                  /* For generic or branded designs, the DATA= data   */
                  /* set must contain NALTS= observations for the     */
                  /* first choice set, NALTS= observations for the    */
                  /* second choice set, and so on.                    */
                  /*                                                  */
                  /*------------Optional Keyword Parameters-----------*/
                  /*                                                  */
data=_last_,      /* specifies the input choice design.               */
                  /*                                                  */
out=,             /* specifies an output data set that contains the   */
                  /* design with duplicate choice sets excluded.  By  */
                  /* default, no data set is created, and the macro   */
                  /* just reports on duplicates.                      */
                  /*                                                  */
outlist=outdups,  /* output data set with the list of duplicates.     */
                  /*                                                  */
vars=,            /* specifies the factors in the design.             */
                  /* By default, all numeric variables are used.      */
factors=,         /* alias for VARS=.                                 */
                  /*                                                  */
                  /*--------------Output Macro Variables--------------*/
                  /*                                                  */
                  /* This macro creates two global macro variables.   */
                  /*                                                  */
                  /* _MKTDUPSSETS - the number of duplicate           */
                  /* observations.  The value is .  if the macro ends */
                  /* prematurely.  Otherwise it is the number of      */
                  /* duplicates.                                      */
                  /*                                                  */
                  /* _MKTDUPSALTS - the number of sets with duplicate */
                  /* alternatives.  The value is .  if the macro ends */
                  /* prematurely.  Otherwise it is the number of      */
                  /* duplicates.                                      */
                  /*                                                  */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _mktver _mktdupssets _mktdupsalts;
%local  abort char generic ifcfact ifnfact linear nfact cfact maxlen
        localver nsets num outfile put savenote setnote savedsnf time
        saveopts k timeopt;

%let time     = %sysfunc(datetime());
%let timeopt  = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote = %sysfunc(getoption(notes));
%let saveopts = &savenote missing="%sysfunc(getoption(missing))"
                %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                %sysfunc(getoption(serror));
%let setnote  = nonotes;
%let maxlen   = 32700;
%let localver = 25Jul2010;
%let _mktver  = &localver;
%let abort    = 0;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES)   %then %let setnote = notes;
   %if %index(&k, VERSION) %then %put MktDups macro version &localver.;
   %end;
options serror;

options &setnote generic missing='.';

data _null_;
   length list $ 32;
   list = left(lowcase(symget('options')));
   if list eq 'help' or list eq '?' then call symput('abort', '-1');
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort = 1 %then %goto endit;

%if &abort = -1 %then %do; %let abort = 0; %mkthelp(mktdups) %goto endit; %end;

*------basic set up and error checking------;
data _null_;
   ok = 1;

   length data options $ 200;

   call symput('_mktdupssets', '.');
   call symput('_mktdupsalts', '.');

   options = lowcase(symget('options'));
   generic = index(options, 'generic');
   branded = index(options, 'brand'  );
   linear  = index(options, 'linear' );
   noprint = index(options, 'noprint');
   if generic then substr(options, generic, 7) = ' ';
   if branded then substr(options, branded, 7) = ' ';
   if linear  then substr(options, linear , 6) = ' ';
   if noprint then substr(options, noprint, 7) = ' ';

   if (generic + branded + linear) > 1 then do;
      put 'ERROR: Specify only one of the following options:' /
          +7     'GENERIC, BRANDED, LINEAR.';
      ok = 0;
      end;

   if not (generic or branded or linear) then generic = 1;

   if options ne ' ' then do;
      put 'ERROR: Options not recognized: ' options +(-1) '.';
      ok = 0;
      end;

   call symput('linear' , put(linear  ne 0, 1.));
   call symput('generic', put(generic ne 0, 1.));
   call symput('put', substr('*put', 1 + not noprint));
   if linear then call symput('nalts', '1');

   nalts = input(symget('nalts'), ?? 32.);
   n = n(nalts);
   if n then do; if nalts ne floor(nalts) or nalts < 1 then n = 0; end;
   if not linear and not n then do;
      put "ERROR: NALTS=%nrbquote(&nalts) is invalid.  You "
          "must specify NALTS=n " "where n is an integer.";
      ok = 0;
      end;

   *------check FACTORS=------;
   if symget('factors') ne ' ' and symget('vars') ne ' ' then do;
      put 'ERROR: You must specify FACTORS= or VARS=, but not both.';
      ok = 0;
      end;
   if symget('factors') eq ' ' then call symput('factors', symget('vars'));
   if symget('factors') eq ' ' then call symput('factors', '_numeric_');

   call symput('outfile', put(symget('out') ne ' ', 1.));

   if symget('outlist') = ' ' then call symput('outlist', 'outdups');

   data = left(lowcase(symget('data')));
   if data in (' ' '_last_') then do;
      data = symget('syslast');
      call symput('data', trim(data));
      end;
   if lowcase(data) = '_null_' then do;
      put 'ERROR: No input data set.';
      ok = 0;
      end;

   if _error_ or not ok then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&data) ne %then %do;
   data _null_; set &data; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outlist) ne %then %do;
   data _null_; set &outlist; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;

options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )data=&data;
   %put %str(       )out=&out;
   %put %str(       )outlist=&outlist;
   %end;

%if &abort %then %goto endit;

*------get variable names------;
proc contents data=&data(keep=&factors)
              out=&outlist(keep=name type) noprint; run;

%if &syserr > 4 %then %let abort = 1;
%if &abort %then %do;
   %put ERROR: FACTORS=&factors is invalid with DATA=&data..;
   %goto endit;
   %end;

*------separate numeric, character variables------;
data _null_;
   set &outlist end=eof;
   length c n $ &maxlen;
   retain c n ' ';
   %if %sysevalf(&sysver >= 9) %then %do; name = nliteral(name); %end;
   if type = 1 then n = trim(n) || ' ' || name;
   else             c = trim(c) || ' ' || name;
   if eof then do;
      call symput('ifnfact', put(n ne ' ', 1.));
      call symput('ifcfact', put(c ne ' ', 1.));
      call symput('nfact'  , trim(left(n)));
      call symput('cfact'  , trim(left(c)));
      end;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------set up variable list, number of sets, number of alts------;
data _null_;
   if 0 then set &data nobs=nobs;
   %if &ifnfact %then %do; array __n[*] &nfact; %end;
   %if &ifcfact %then %do; array __c[*] &cfact; %end;

   __ok = 1;
   length str $ &maxlen;
   &put / 'Design:' +10
          %if       &generic %then 'Generic';
          %else %if &linear  %then 'Linear';
          %else                    'Branded';;
   str = symget('factors');
   &put 'Factors:' +9 str;

   %if &ifcfact %then %do;
      call symput('char', compress(put(dim(__c), best12.)));
      str = symget('cfact');
      &put @18 str;
      %end;

   %if &ifnfact %then %do;
      call symput('num' , compress(put(dim(__n), best12.)));
      str = symget('nfact');
      &put @18 str;
      %end;

   nsets = nobs / &nalts;
   if nsets < 1 or nsets ne floor(nsets) then do;
      put 'ERROR: The number of observations must be a '
          "multiple of %nrbquote(&nalts).";
      __ok = 0;
      end;

   call symput('nsets', compress(put(nsets, best12.)));

   if _error_ or not __ok then call symput('abort', '1');
   stop;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------process data, hunt for duplicates------;
data &outlist(keep=__i __match rename=(__i=Set __match=Dup));

   if 0 then set &data;

   %if &ifnfact %then %do; array __n[&num] &nfact; %end;
   %if &ifcfact %then %do;
      array __c[&char] &cfact;
      length __tc1 - __tc&char $ 200;
      array __tc[&char];
      %end;
   %if &ifnfact %then %do; array __tn[&num]; %end;
   %if &generic %then %do; array __m[&nalts]; array __mm[&nalts]; %end;
   __dupalts = 0; __dup = 0;

   *------loop over each pair of choice sets------;
   do __i = 1 to &nsets;
      do __j = __i %if not &generic %then + 1; to &nsets;

         __ok = 0;

         %if &generic %then %do;
            *------generic design, check all alternative pairs------;
            do __k = 1 to &nalts; __m[__k] = 0; __mm[__k] = 0; end;
            do __k = 1 to &nalts;
               link copyit;
               do __q = 1 to &nalts;
                  __ok = 0;
                  link compareit;
                  if __ok eq 0 then do; __m[__q] + 1; __mm[__k] + 1; end;
                  end;
               end;

            *------ok if any alternative is different------;
            __ok = 0; __dup = 0;
            do __k = 1 to &nalts;
               if __m[__k] eq 0 or __mm[__k] eq 0 then __ok = 1;
               if __i eq __j then do;
                  __ok = 1;
                  if __m[__k] > 1 then __dup = 1;
                  end;
               end;

            __dupalts + __dup;
            %end;

         %else %do;
            *------branded design, just check corresponding alt pairs------;
            do __k = 1 to &nalts;
               link copyit;
               link compareit;
               end;
            %end;

         *------output list of duplicates------;
         if not __ok then do;
            __count + 1;
            __match = __j;
            output;
            end;

         if __dup  then do;
            __count + 1;
            __match = ._;
            output;
            end;
         end;
      end;

   call symput('_mktdupssets', trim(put(__count,   best12. -L)));
   call symput('_mktdupsalts', trim(put(__dupalts, best12. -L)));
   if _error_ then call symput('abort', '1');
   stop;
   return;

   *------copy factors to temporary storage------;
   copyit:
      __point = (__i - 1) * &nalts + __k;
      set &data point=__point;
      %if &ifnfact %then %do;
         do __l = 1 to &num; __tn[__l] = __n[__l]; end;
         %end;
      %if &ifcfact %then %do;
         do __l = 1 to &char; __tc[__l] = __c[__l]; end;
         %end;
      return;

   *------compare factors to temporary copy from other choice set------;
   compareit:
      %if not &generic %then %do; __q = __k; %end;
      __point = (__j - 1) * &nalts + __q;
      set &data point=__point;
      %if &ifnfact %then %do;
         do __l = 1 to &num;  if __tn[__l] ne __n[__l] then __ok = 1; end;
         %end;
      %if &ifcfact %then %do;
         do __l = 1 to &char; if __tc[__l] ne __c[__l] then __ok = 1; end;
         %end;
      return;

   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get rid of duplicate mentions of duplicates------;
%if &_mktdupssets or &_mktdupsalts %then %do;
   proc sort data=&outlist; by dup set; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc format; value dupform ._ = ' '; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data &outlist(drop=counts counta n);
      set &outlist end=eof;
      by dup;
      n = n(dup);
      if first.dup or not n then do;
         if not n then a = 'Alternatives';
         output;
         counts + n;
         counta + not n;
         end;
      if eof then do;
         if &generic then do; &put 'Sets w Dup Alts:' +1 counta; end;
         if &linear  then do; &put 'Duplicate Runs:'  +2 counts; end;
         else             do; &put 'Duplicate Sets:'  +2 counts; end;
         call symput('_mktdupssets', trim(put(counts, best12. -L)));
         end;
      format dup dupform11.;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc sort data=&outlist out=&outlist; by set dup; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------print information about duplicate choice sets------;
   %if %length(%nrbquote(&put)) > 1 %then %do;
      proc print data=&outlist label split='+';
         label %if &linear %then
                  set = '+Run'
                  dup = '  Duplicate+       Runs+  To Delete';
               %else
                  set = '+Choice+   Set'
                  dup = '  Duplicate+Choice Sets+  To Delete';
               a = '00'x;
         id set; by set;
         format set 11.;
         run;
      %end;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;


%else %do;
   data _null_;
      if &generic then do; &put 'Sets w Dup Alts:' +1 '0'; end;
      if &linear  then do; &put 'Duplicate Runs:'  +2 '0'; end;
      else             do; &put 'Duplicate Sets:'  +2 '0'; end;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

options &savenote no&timeopt;

*------output data set with duplicates excluded------;
%if &outfile and (&_mktdupssets or &_mktdupsalts) %then %do;
   data &out;
      set &data;
      __set = ceil(_n_ / &nalts);
      __ok = 1;
      do __k = 1 to __nobs;
         set &outlist(keep=set dup rename=(dup=__match set = __s))
             point=__k nobs=__nobs;
         if (nmiss(__match) and __s eq __set) or
            (__match eq __set) then do;
            __ok = 0;
            leave;
            end;
         end;
      if __ok then output;
      drop __:;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%else %if &outfile and not &_mktdupssets %then %do;
   data &out; set &data; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTDUPS macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTDUPS macro used &time seconds.;
   %end;

%mendit: %mend;
