 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktMDiff                                            */
 /*   TITLE: Process and Analyze MaxDiff Data and Designs        */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs                         */
 /*   PROCS: TRANSREG PHREG                                      */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: This macro creates global macro variable: _MktVer   */
 /*                                                              */
 /*          The following SAS products are required to run      */
 /*          this macro: BASE, STAT.                             */
 /****************************************************************/

 /*---------------------------------------------------------------------

In a MaxDiff or Best-Worst study, subjects are shown sets of messages or
product attributes and are asked to choose the best (or most important)
from each set as well as the worst (or least important).  A set of t
attributes are shown in b sets (or blocks) of size k.  These are the
parameters of a Balanced Incomplete Blocks Design or BIBD, which can
be constructed from the MktBIBD macro.  Note, however, that you can
use designs that are produced by the macro but do not meet the strict
requirements for a BIBD.  The macro reads the experimental design and
the data, combines them, arranges them in the right form for analysis,
and performs the analysis using a multinomial logit model.  The data
are arrayed so that each set forms two choice sets in the analysis: a
positively weighted set for the best choice and a negatively weighted
set for the worst choice.  The input data can come in one of eight
forms:

   bw       - best then worst (e.g. b1-b18 w1-w18), and the data are
              attribute numbers (range from 1 to t).

   wb       - worst then best (e.g. w1-w18 b1-b18), and the data are
              attribute numbers (range from 1 to t).

   bwalt    - best then worst and alternating
              (e.g. b1 w1 b2 w2 ... b18 w18), and the data are
              attribute numbers (range from 1 to t).

   wbalt    - worst then best and alternating
              (e.g. w1 b1 w2 b2 ... w18 b18), and the data are
              attribute numbers (range from 1 to t).

   bwpos    - best then worst (e.g. b1-b18 w1-w18),
              and the data are positions (range from 1 to k).

   wbpos    - worst then best (e.g. w1-w18 b1-b18),
              and the data are positions (range from 1 to k).

   bwaltpos - best then worst and alternating
              (e.g. b1 w1 b2 w2 ... b18 w18),
              and the data are positions (range from 1 to k).

   wbaltpos - worst then best and alternating
              (e.g. w1 b1 w2 b2 ... w18 b18),
              and the data are positions (range from 1 to k).

The following is an example:

%mktbibd(t=9, k=5, b=18, seed=377, out=sasuser.bibd)

title 'Best Worst Example with Cell Phone Attributes';

data bestworst;
   input Sub $ @4 (b1-b18 w1-w18) (1.);
   datalines;
 1 188661884399349653941955342212935494
 2 765358873891388493922673644336595554
 3 782126282892848564995993447213935655
 4 481363264246399187162125415351281453
 5 787168863811878175225995382352235293
 6 787658867891878667495965442313345453
 7 788171867736888187465395344393344453
 8 788771867711888687445353445353335254
 9 188778887896878687425323443242344454
10 788778887816888687445353444343344454
11 787778877816878667442353343343235453
12 787778387711898667425321443253544594
13 767668285791988687441951364232234194
14 187168877741878687445323445216334453
15 788176487116988675492393314339579457
16 267665615733884677442193342349545564
17 481191867813938266147956244139584594
18 725778814832585185141193643296944467
19 188678863811279263445123344253934596
20 728698612719281483265755285851944597
;

%let attributes=Camera,Flip,Hands Free,Games,Internet
,Free Replacement,Battery Life,Large Letters,Applications;

%phchoice( on )

%mktmdiff(bw, b=18, t=9, k=5, attrs=attributes,
          data=bestworst, design=sasuser.bibd)

The MktBIBD macro makes a balanced incomplete design for this
specification.  The MktMDiff macro begins by printing a summary of the
input as follows:

Var Order:   Best then Worst
Alternating? Variables Do Not Alternate
Data?        Attribute Numbers (Not Positions)
Best Vars:   b1 b2 b3 b4 b5 b6 b7 b8 b9 b10 b11 b12 b13 b14 b15 b16 b17 b18
Worst Vars:  w1 w2 w3 w4 w5 w6 w7 w8 w9 w10 w11 w12 w13 w14 w15 w16 w17 w18
Attributes:  Camera
             Flip
             Hands Free
             Games
             Internet
             Free Replacement
             Battery Life
             Large Letters
             Applications

Because of the initial "bw" specification, the data are best then worst.
The variables do not alternate.  The data are attribute numbers not
positions.  The best variables are b1-b18, and the worst variables are
w1-w18.  Finally, the descriptions of each of the t attributes are
listed.  The rest of the output comes from PROC PHREG which fits choice
models.

The following table is of interest as a check of the integrity of the
input data:

         Summary of Subjects, Sets, and Chosen and Unchosen Alternatives

                   Number of       Number of          Chosen         Not
        Pattern      Choices    Alternatives    Alternatives      Chosen

              1           36             100              20          80


In the aggregate data set, there was one pattern of input and it occured
36 times (18 best choices per subject plus 18 worst choices).  There
were 100 alternatives (5 attributes in a set examined by 20
individuals), each of the 20 subjects chose 1 (20 x 1 = 20) and did not
choose 4 (20 x 4 = 80).

The final table of results is as follows:

                 Best Worst Example with Cell Phone Attributes

                              Parameter      Standard                    Pr >
                       DF      Estimate         Error    Chi-Square     ChiSq

   Large Letters        0             0             .         .         .
   Battery Life         1      -0.52009       0.17088        9.2635    0.0023
   Free Replacement     1      -1.01229       0.17686       32.7598    <.0001
   Camera               1      -1.30851       0.18232       51.5096    <.0001
   Applications         1      -1.93107       0.18741      106.1680    <.0001
   Flip                 1      -2.00892       0.18788      114.3334    <.0001
   Internet             1      -2.44592       0.18731      170.5156    <.0001
   Hands Free           1      -2.47063       0.18781      173.0569    <.0001
   Games                1      -2.86329       0.18802      231.9065    <.0001

The parameter estimates are arranged from most preferred to least
preferred.  These results are also available in a SAS data set called ParmEst.

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

%macro mktmdiff(  /*--------------------------------------------------*/
layout,           /* Layout is a positional parameter that must be    */
                  /* specified to indicate the structure of the       */
                  /* choice data.  Just specify a value, not layout=. */
                  /* The value has several components, you can        */
                  /* use any case, and spaces between the pieces are  */
                  /* optional.  Values include "b", "w", "alt", and   */
                  /* "pos".  Both a "b" and a "w" must be specified.  */
                  /* When "b" comes before "w", the best variables    */
                  /* come before the worst variables.  Otherwise,     */
                  /* when "w" comes before "b", the worst variables   */
                  /* come before the best variables.  You may         */
                  /* optionally specify "alt" as well which means     */
                  /* that the variables alternate (e.g. b1 w1 b2 w2   */
                  /* b3 w3 where the "b" variables are best and the   */
                  /* "w" variables are worst variables).  Otherwise   */
                  /* it is expected that all of one type appear then  */
                  /* all of the other type.  By default, the data are */
                  /* assumed to be the numbers of the attributes      */
                  /* that were chosen (e.g. 1 to 6 when there are a   */
                  /* total of t=6 attributes).  If instead (say       */
                  /* with k=3), the data are 1 to 3 indicating the    */
                  /* position of the chosen attribute, then you       */
                  /* must specify "pos" as well.                      */
                  /*                                                  */
                  /* Examples:                                        */
                  /* bw (or "b w") - best then worst, e.g. b1-b6 w1-w6*/
                  /* or even x1-x12 (if 2 * b = 12).  While the       */
                  /* values need to be best then worst, the variable  */
                  /* names do not need to reflect this.  The data are */
                  /* attribute numbers.                               */
                  /*                                                  */
                  /* wb (or "w b") - worst then best, e.g. w1-w6 b1-b6*/
                  /* or even x1-x12 (if 2 * b = 12).  While the       */
                  /* values need to be worst then best, the variable  */
                  /* names do not need to reflect this.  The data are */
                  /* attribute numbers.                               */
                  /*                                                  */
                  /* bwalt (or "b w alt" or "alt w b" and so on) -    */
                  /* best then worst and alternating, e.g.  b1 w1 b2  */
                  /* w2 ... or even x1-x12 (if 2 * b = 12).  While    */
                  /* the values need to alternate best then worst,    */
                  /* the variable names do not need to reflect this.  */
                  /* The data are attribute numbers.                  */
                  /*                                                  */
                  /* wbalt (or "w b alt" or "alt b w" and so on) -    */
                  /* worst then best and alternating, e.g.  w1 b1 w2  */
                  /* b2 ... or even x1-x12 (if 2 * b = 12).  While    */
                  /* the values need to alternate worst then best,    */
                  /* the variable names do not need to reflect this.  */
                  /* The data are attribute numbers.                  */
                  /*                                                  */
                  /* bwpos (or "b w alt" or "w pos b" and so on) -    */
                  /* best then worst, and the data are positions.     */
                  /*                                                  */
                  /* wbpos (or "w b pos" or "b w pos" and so on) -    */
                  /* worst then best, and the data are positions.     */
                  /*                                                  */
                  /* bwaltpos (or "b w alt pos" or "alt w pos b" and  */
                  /* so on) - best then worst and alternating, and    */
                  /* the data are positions.                          */
                  /*                                                  */
                  /* wbaltpos (or "w b alt pos" or "alt b w pos" and  */
                  /* so on) - worst then best and alternating, and    */
                  /* the data are positions.                          */
                  /*                                                  */
                  /* ? or help - lists options                        */
                  /*                                                  */
b=,               /* specifies the number of sets.  The "b" in b=     */
                  /* stands for blocks and matches the option name in */
                  /* the MktBIBD macro since the design is typically  */
                  /* a Balanced Incomplete Block Design or BIBD.      */
                  /* This option must be specified (or its alias      */
                  /* NSETS=).                                         */
nsets=,           /* alias for B= with a MaxDiff name.                */
                  /*                                                  */
t=,               /* specifies the number of attributes or messages.  */
                  /* The "t" in t= stands for treatments and matches  */
                  /* the option name in the MktBIBD macro since the   */
                  /* design is typically a BIBD.  This option must be */
                  /* specified (or its alias NATTRS=).                */
nattrs=,          /* alias for T= with a MaxDiff name.                */
                  /*                                                  */
k=,               /* specifies the number of attributes or messages   */
                  /* shown at one time (in a set).  The k= option is  */
                  /* named to match the same option in the MktBIBD    */
                  /* macro since the design is typically a BIBD.      */
                  /* This option must be specified (or its alias      */
                  /* SETSIZE=).                                       */
setsize=,         /* alias for K= with a MaxDiff name.                */
                  /*                                                  */
group=1,          /* specifies the number of blocks of choice sets.   */
                  /* By default when GROUP= is not specified, it is   */
                  /* assumed that the design consists of one big      */
                  /* group.  Otherwise, with n subjects in each group */
                  /* and GROUP=g there are n x g rows of data each    */
                  /* consisting of b / g best values and b / g worst  */
                  /* values.  The number of subjects in each group    */
                  /* must be the same.  Furthermore, it is assumed    */
                  /* that the design is sorted by the group variable. */
                  /* If the design is made by the MktBIBD macro, this */
                  /* will be the case.  If the DESIGN= data set has k */
                  /* + 1 columns and either the first column or the   */
                  /* last column is called "Group" (case is ignored), */
                  /* then that variable is ignored and not treated    */
                  /* like part of the design.                         */
                  /*                                                  */
options=,         /* nocode - just create the OUT= data set but do    */
                  /* not code or do the analysis.                     */
                  /* noanalysis - just create the OUT= and OUTCODED=  */
                  /* data sets but do not do the analysis.            */
                  /* nosort - do not sort the parameter estimates     */
                  /* table by the parameter estimates.                */
                  /* nodrop - do not drop the variable or parameter   */
                  /* column from the parameter estimates table.       */
                  /* nosum - do not print sums on the rescaled        */
                  /* parameters table.                                */
                  /* rescale - force rescale= option in the context   */
                  /* of dubious other options.                        */
                  /*                                                  */
rescale=,         /* rescale the parameter estimates as follows:      */
                  /* center - center.                                 */
                  /* p - exponentiate, divide by sum of exponentiated */
                  /* utilities.                                       */
                  /* p100 - exponentiate, divide by sum of            */
                  /* exponentiated utilities, multiply by 100.        */
                  /* adjusted (or adj) - center, exp(Beta) /          */
                  /* (exp(Beta) + k - 1), rescale to sum to 1.        */
                  /* adjusted100 (or adj100) - center, exp(Beta) /    */
                  /* (exp(Beta) + k - 1), rescale to sum to 100.      */
                  /* default - report raw parameter estimates.        */
                  /* all - report all rescaled values.                */
                  /* Note that RESCALE= and RESCALE=DEFAULT are       */
                  /* equivalent.  When this option is specified (or   */
                  /* any option besides RESCALE=DEFAULT is            */
                  /* specified), an additional table is displayed     */
                  /* with the rescaled parameter estimates.  You can  */
                  /* specify multiple values and get multiple columns */
                  /* in the results table.  Example: RESCALE=DEFAULT  */
                  /* ADJ100.  All specified rescalings are added to   */
                  /* the OUTPARM= data set.                           */
                  /*                                                  */
classopts=        /* specifies options to apply to the class variable,*/
zero=none,        /* for example: CLASSOPTS=effects.  The default is  */
                  /* CLASSOPTS=zero=none.  A reference cell coding is */
                  /* used, and the reference level is displayed with  */
                  /* a parameter estimate of zero.  If you specify a  */
                  /* null value (CLASSOPTS=,) then a reference cell   */
                  /* coding is used and the reference level is not    */
                  /* displayed. You can control the reference level   */
                  /* by specifying CLASSOPTS=zero='reference-level'   */
                  /* and naming the appropriate reference level.      */
                  /*                                                  */
vars=_numeric_,   /* specifies the variables in the DATA= data set    */
                  /* that contain the data.  There must be 2 * b      */
                  /* variables in this list.  The default is all      */
                  /* numeric variables in the data set.               */
                  /*                                                  */
attrs=,           /* specifies the name of a macro variable that      */
                  /* contains a comma delimited list of attribute     */
                  /* descriptions.  Alternatively, after the name     */
                  /* you can specify a single non-blank delimiter     */
                  /* character.  Examples: ATTRS=myattributes,        */
                  /* ATTRS=myattrs \, (using a back slash as a        */
                  /* delimiter).  If you do not specify this option,  */
                  /* the default attribute names are A1, A2, and so   */
                  /* on.                                              */
                  /*                                                  */
data=_last_,      /* specifies the input data set.  By default, the   */
                  /* last data set created is used.                   */
design=,          /* specifies the data set with the design.  This is */
                  /* typically a BIBD in the format produced by the   */
                  /* OUT= option of the BIBD macro.  You must specify */
                  /* this option.  All numeric columns are assumed    */
                  /* to contain the BIBD, unless there is exactly one */
                  /* extra column, and the first or last column is    */
                  /* called "Group".  In that case the group variable */
                  /* is ignored.                                      */
out=maxdiff,      /* specifies the output data set in a form ready    */
                  /* for coding.  The default is MAXDIFF.             */
outcoded=coded,   /* specifies the output data set that has been      */
                  /* coded by PROC TRANSREG.  The default is CODED.   */
outparm=parmest); /* specifies the output data set with the parameter */
                  /* estimates.  The default is PARMEST.              */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _MktVer;
%local  i abort time savenote setnote saveopts localver timeopt
        drop bestfirst alt pos code analysis bestlist worstlist
        delim outopts outcopts outpopts localatrs atrlen sort
        dodrop rscenter rsp rsp100 rsadj rsadj100 rsdefault summit;

%let time     = %sysfunc(datetime());
%let timeopt  = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote = %sysfunc(getoption(notes));
%let saveopts = &savenote %sysfunc(getoption(generic))
                %sysfunc(getoption(&timeopt)) %sysfunc(getoption(dsnferr))
                %sysfunc(getoption(fmterr)) %sysfunc(getoption(serror));
%let setnote  = nonotes;
%let abort    = 0;
%let localver = 25Jul2010;
%let _mktver  = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let i = %upcase(&mktopts);
   %if %index(&i, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&i, NOTES)   %then %let setnote = notes;
   %if %index(&i, VERSION) %then %put MktMDiff macro version &localver.;
   %end;
options serror;

options &setnote generic;

data _null_;
   length list $ 32;
   list = left(lowcase(symget('layout')));
   if list eq 'help' or list eq '?' then call symput('abort', '-1');
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort = 1 %then %goto endit;

%if &abort = -1 %then %do; %let abort = 0; %mkthelp(mktmdiff) %goto endit; %end;

*------rudimentary option processing and error checking------;
data _null_;
   abort = 0;
   length w $ 50 list $ 500 str $ 32767;

   *------check LAYOUT------;
   w = lowcase(symget('layout'));
   if w eq ' ' then do;
      put 'ERROR: You must specify a layout as a first positional parameter.';
      put '       Example: %mktmdiff(B W ALT, ...)';
      abort = 1;
      end;
   else do;
      bi = index(w, 'b');
      wi = index(w, 'w');
      if bi eq 0 or wi eq 0 then do;
         put "ERROR: The layout '%nrbquote(&layout)' is invalid.";
         put 'ERROR: The layout positional parameter must contain '
             'both a "b" and a "w".';
         abort = 1;
         end;
      call symput('bestfirst', put(bi lt wi, 1.));
      call symput('pos', put(index(w, 'pos') gt 0, 1.));
      call symput('alt', put(index(w, 'alt') gt 0, 1.));
      end;

   *------check OPTIONS=------;
   list = lowcase(symget('options'));
   i = index(list, 'nocode'); if i then substr(list, i, 6) = ' ';
   call symput('code', put(i eq 0, 1.));
   i = index(list, 'noanalysis'); if i then substr(list, i, 10) = ' ';
   call symput('analysis', put(i eq 0, 1.));
   i = index(list, 'nosort'); if i then substr(list, i, 6) = ' ';
   call symput('sort', put(i eq 0, 1.));
   i = index(list, 'nodrop'); if i then substr(list, i, 6) = ' ';
   call symput('dodrop', put(i eq 0, 1.));
   i = index(list, 'nosum'); if i then substr(list, i, 5) = ' ';
   call symput('summit', put(i eq 0, 1.));
   rs = index(list, 'rescale'); if rs then substr(list, rs, 7) = ' ';
   if list ne ' ' then do;
      list = compbl(list);
      put 'ERROR: OPTIONS=' list 'is invalid.';
      abort = 1;
      end;

   *------OK to rescale?------;
   list  = lowcase(compress(symget('classopts'), ' '));
   cldef = list eq ' ' or list eq 'zero=none';
   list  = lowcase(symget('rescale'));
   rsdef = list eq ' ' or list eq 'default';

   if not rs and not rsdef and not cldef then do;
      put 'ERROR: You cannot specify the RESCALE= option with any '
          'CLASSOPTS= value ' 'other than ' 'the default of '
          'CLASSOPTS=zero=none ' 'or a null CLASSOPTS=.  '
          'You can use ' 'RESCALE= with ' 'CLASSOPTS= if you also '
          'specify OPTIONS=RESCALE.';
      abort = 1;
      end;
   if not rsdef and not cldef then
      put 'Note: When you specify ' 'OPTIONS=RESCALE, you are '
          'specifying that ' 'you are not ' 'doing anything with '
          'CLASSOPTS= besides ' 'controlling the ' 'reference level.  '
          'Any use of CLASSOPTS= ' 'and OPTIONS=RESCALE that '
          'changes the coding ' 'from reference-cell '
          'coding is not ' 'guaranteed to ' 'produce correct numbers.';

   *------check RESCALE=------;
   list = lowcase(symget('rescale'));
   call symput('rsdefault', put(list eq ' ', 1.));
   i = index(list, 'center'); if i then substr(list, i, 6) = ' ';
   call symput('rscenter', put(i gt 0, 1.));
   i = index(list, 'p100'); if i then substr(list, i, 4) = ' ';
   call symput('rsp100', put(i gt 0, 1.));
   i = index(list, 'default'); if i then substr(list, i, 7) = ' ';
   if i then call symput('rsdefault', '1');
   i = index(list, 'adjusted100'); if i then substr(list, i, 11) = ' ';
   k = index(list, 'adj100'); if k then substr(list, k, 6) = ' ';
   call symput('rsadj100', put(i or k, 1.));
   i = index(list, 'adjusted'); if i then substr(list, i, 8) = ' ';
   k = index(list, 'adj'); if k then substr(list, k, 3) = ' ';
   call symput('rsadj', put(i or k, 1.));
   i = index(list, 'p'); if i then substr(list, i, 1) = ' ';
   call symput('rsp', put(i gt 0, 1.));
   i = index(list, 'all'); if i then substr(list, i, 3) = ' ';
   if i then do;
      call symput('rsdefault', '1');  call symput('rsp100', '1');
      call symput('rsadj100', '1');   call symput('rscenter', '1');
      call symput('rsp', '1');        call symput('rsadj', '1');
      end;
   if list ne ' ' then do;
      list = compbl(list);
      put 'ERROR: RESCALE=' list 'is invalid.';
      abort = 1;
      end;

   *------check B=------;
   o1 = (symget('b') ne ' ');
   o2 = (symget('nsets') ne ' ');
   if (o1 + o2) ne 1 then do;
      put 'ERROR: You must specify exactly one of the following: '
          'B= or NSETS=.';
      abort = 1;
      end;

   else do;
      call symput('b', symget('b') || symget('nsets'));
      b = input(symget('b'), ?? 32.);
      if nmiss(b) or b < 2 or b ne round(b) or b > 5000 then do;
         if o1 then put "ERROR: B=%nrbquote(&b) is invalid.";
         if o2 then put "ERROR: NSETS=%nrbquote(&nsets) is invalid.";
         abort = 1;
         end;

      %*------check GROUP=------;
      else do;
         g = symget('group');
         if g eq ' ' then call symput('group', '1');
         else do;
            ngroups = input(symget('group'), ?? 32.);
            if ngroups < 1 or ngroups ne round(ngroups) then do;
               put "ERROR: GROUP=%nrbquote(&group) is invalid.";
               abort = 1;
               end;
            else do;
               l = b / ngroups;
               if l ne round(l) then do;
                  put "ERROR: GROUP=%nrbquote(&group) is invalid.";
                  abort = 1;
                  end;
               end;
            end;
         end;
      end;

   *------check K=------;
   o1 = (symget('k') ne ' ');
   o2 = (symget('setsize') ne ' ');
   if (o1 + o2) ne 1 then do;
      put 'ERROR: You must specify exactly one of the following: '
          'K= or SETSIZE=.';
      abort = 1;
      end;

   else do;
      call symput('k', symget('k') || symget('setsize'));
      k = input(symget('k'), ?? 32.);
      if nmiss(k) or k < 2 or k ne round(k) or k > 5000 then do;
         if o1 then put "ERROR: K=%nrbquote(&k) is invalid.";
         if o2 then put "ERROR: SETSIZE=%nrbquote(&setsize) is invalid.";
         abort = 1;
         end;
      end;

   *------check T=------;
   o1 = (symget('t') ne ' ');
   o2 = (symget('nattrs') ne ' ');
   t  = 0;
   if (o1 + o2) ne 1 then do;
      put 'ERROR: You must specify exactly one of the following: '
          'T= or NATTRS=.';
      abort = 1;
      end;

   else do;
      call symput('t', symget('t') || symget('nattrs'));
      t = input(symget('t'), ?? 32.);
      if nmiss(t) or t < 2 or t ne round(t) or t > 5000 then do;
         if o1 then put "ERROR: T=%nrbquote(&t) is invalid.";
         if o2 then put "ERROR: NATTRS=%nrbquote(&nattrs) is invalid.";
         abort = 1;
         end;
      end;

   *------check ATTRS=------;
   if abort ne 2 then do;
      w = symget('attrs');
      d = scan(w, 2, ' ');
      if d eq ' ' then d = ',';
      call symput('delim', trim(d));
      call symput('attrs', scan(w, 1, ' '));
      str = symget('attrs');
      if str eq ' ' then do;
         call symput('attrs', 'localatrs');
         do i = 1 to t;
            str = compress(trim(str) || 'A' || put(i, 5.) || ',');
            end;
         substr(str, length(str)) = ' ';
         call symput('localatrs', trim(str));
         end;

      %if %sysevalf(&sysver gt 9) %then %do;
         else do;
            if not symglobl(str) then do;
               put 'ERROR: ATTRS=' str 'is not a global macro variable.';
               abort = 1;
               end;
            if symlocal(str) then do;
               put 'ERROR: You cannot specify ATTRS=' str
                   +(-1) '.  Use some other name.';
               abort = 1;
               end;
            end;
         %end;
      end;

   *------check DESIGN=------;
   if symget('design') eq ' ' then do;
      put "ERROR: You must specify DESIGN= to specify the design.";
      abort = 1;
      end;

   *------store data set names------;
   i = 0;
   w = lowcase(symget('data'));
   i + index(w, '(');
   if w eq '_last_' or w eq ' ' then do;
      w = symget('syslast');
      if w eq ' ' or lowcase(w) eq '_null_' then do;
         put 'ERROR: No input DATA= data set.';
         abort = 1;
         end;
      call symput('data', trim(w));
      end;
   w = symget('out');
   i + index(w, '(');
   if w eq ' ' then call symput('out', 'maxdiff');
   w = symget('outcoded');
   i + index(w, '(');
   if w eq ' ' then call symput('outcoded', 'coded');
   w = symget('outparm');
   i + index(w, '(');
   if w eq ' ' then call symput('outparm', 'parmest');
   if i then do;
      put 'ERROR: Data set options are not supported in this macro.';
      abort = 1;
      end;

   if abort or _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;
filename _dummy dummy;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outcoded) ne %then %do;
   data _null_; set &outcoded; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outparm) ne %then %do;
   data _null_; set &outparm; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&data) ne %then %do;
   data _null_; set &data; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&design) ne %then %do;
   data _null_; set &design; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )data=&data;
   %put %str(       )data=&design;
   %put %str(       )out=&out;
   %put %str(       )out=&outcoded;
   %put %str(       )out=&outparm;
   %end;

%if &abort %then %goto endit;

*------get variable names from DATA= data set (in order)------;
proc contents data=&data(keep=&vars) out=__tmpmvars(keep=npos name) noprint;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc sort; by npos; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*-----get lists of best and worst vars and make initial report------;
%if %sysevalf(&sysver ge 9) %then %let i = nliteral; %else %let i = ;

data _null_;
   length best worst str $  32767 w $ 500;
   retain best worst ' ' bw &bestfirst abort 0;
   set __tmpmvars nobs=nobs;

   if _n_ eq 1 and nobs ne 2 * &b / &group then do;
      put "ERROR: There must be %eval(2 * &b / &group) numeric variables.  "
          nobs "were found in DATA=&data data set.";
      abort = 1;
      end;

   if bw then best  = trim(best)  || ' ' || &i(name);
   else       worst = trim(worst) || ' ' || &i(name);

   %if not &alt %then %do; if _n_ eq &b / &group then %end; bw = not bw;

   if not (abort or _error_) and _n_ eq 2 * &b / &group then do;
      call symput('bestlist' , left(trim(best)));
      call symput('worstlist', left(trim(worst)));
      file print;
      put 'Var Order:   ' %if &bestfirst %then 'Best then Worst';
                          %else 'Worst then Best';;
      put 'Alternating: ' %if &alt %then 'Variables Alternate';
                          %else 'Variables Do Not Alternate';;
      put 'Data:        ' %if &pos %then 'Positions (Not Attribute Numbers)';
                          %else 'Attribute Numbers (Not Positions)';;
      %if &group ne 1 %then %do;
         put 'Grouping:    ' "&group Groups of Size %eval(&b / &group)";
         %end;

      put 'Best Vars:   ' @;
      do i = 1 to 32767 until(w eq ' ');
         w = scan(best, i, ' ');
         if w ne ' ' then put w @;
         end;
      put;
      put 'Worst Vars:  ' @;
      do i = 1 to 32767 until(w eq ' ');
         w = scan(worst, i, ' ');
         if w ne ' ' then put w @;
         end;
      put;
      put 'Attributes:  ' @;
      str = symget("&attrs");
      atrlen = 0;
      printed = 0;
      do i = 1 to &t;
         w = scan(str, i, "&delim");
         atrlen = max(atrlen, length(w));
         if not printed and w eq ' ' then do;
            file log;
            put "ERROR: Each of the &t attributes must have a label.";
            file print;
            printed = 1;
            abort = 1;
            end;
         put @14 w;
         end;
      call symput('atrlen', put(atrlen, 5.));
      end;

   if _error_ or abort then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------if data are grouped, rearrange accordingly------;
%if &group ne 1 %then %do;

   data __tmpmdat;
      set &data(keep=&vars) nobs=__nobs;
      if _n_ eq 1 then do;
         if (__nobs / &group) ne round(__nobs / &group) then do;
            put 'ERROR: The number of observations must be a '
                'multiple of the number of groups.';
            call symput('abort', '1');
            stop;
            end;
         end;

      __row = _n_;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc transpose data=__tmpmdat out=__tmpmdat(keep=col1); by __row; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data __tmpmdat;
      retain bw &bestfirst;
      set __tmpmdat;

      sub = 1 + floor((_n_ - 1) / (&b * 2));
      %if not &alt %then if mod(_n_ - 1, &b / &group) eq 0 then; bw = not bw;
      i = _n_;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc sort data=__tmpmdat; by sub bw i; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc transpose data=__tmpmdat out=__tmpmdat(keep=x:) prefix=x;
      by sub; var col1;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %let bestlist  = x1-x&b;
   %let worstlist = x%eval(&b + 1)-x%eval(2 * &b);
   %let data      = __tmpmdat;
   %end;

*------check data------;
data _null_;
   set &data;

   array _x[*] &bestlist &worstlist;
   do _i = 1 to dim(_x);
      _z = _x[_i];
      if _z ne round(_z) or _z lt 1 or _z gt
         %if &pos %then &k; %else &t; then do;
         put 'ERROR: Data value ' _z 'is invalid.';
         error + 1;
         call symput('abort', '1');
         if error > 5 then do;
            put 'WARNING: No more of these error messages will be printed.';
            stop;
            end;
         end;
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------see if the design has a group variable that we need to ignore------;
data _null_;
   length __name $ 32;
   set &design;
   array __x[*] _numeric_;
   __t = dim(__x);
   __drop = 0;
   if __t eq &k + 1 then do;
      call vname(__x[1], __name);
      if lowcase(__name) eq 'group' then __drop = 1;
      call vname(__x[__t], __name);
      if lowcase(__name) eq 'group' then __drop = 1;
      end;
   call symput('drop', put(__drop, 1.));
   stop;
   run;

*------check design------;
data _null_;
   set &design %if &drop %then (drop=group); nobs=nobs;
   array _x[*] _numeric_;
   if _n_ eq 1 then do;
      if dim(_x) ne &k then do;
         put "ERROR: The DESIGN= data set must have &k variables.";
         call symput('abort', '1');
         end;
      if nobs ne &b then do;
         put "ERROR: The DESIGN= data set must have &b observations.";
         call symput('abort', '1');
         end;
      end;
   do _i = 1 to dim(_x);
      _z = _x[_i];
      if _z ne round(_z) or _z lt 1 or _z gt &t then do;
         put 'ERROR: Design value ' _z 'is invalid.';
         error + 1;
         call symput('abort', '1');
         if error > 5 then do;
            put 'WARNING: No more of these error messages will be printed.';
            stop;
            end;
         end;
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------convert from positions to indexes------;
%if &pos %then %do;
   data __tmpmbibd(keep=_x:);
      set &design %if &drop %then (drop=group);;
      array __x[&k] _numeric_;
      array _x[&k];
      do _i = 1 to &k; _x[_i] = __x[_i]; end;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data __tmpmdat;
      set &data;
      array _b[&b] &bestlist;
      array _w[&b] &worstlist;
      array _x[&k];
      do _i = 1 to &b;
         set __tmpmbibd point=_i;
         _b[_i] = _x[_b[_i]];
         _w[_i] = _x[_w[_i]];
         end;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %let data = __tmpmdat;
   %end;

*------roll out design for analysis------;
data __tmpmchdes(keep=_alt _set _attr _bw _type
     rename=( _alt = Alt _set = Set _attr = Attr _bw = BW _type = Type));
   set &design %if &drop %then (drop=group);;
   length _BW $ 5 _attr $ &atrlen _attrlist $ 32767;
   retain _attrlist;
   array _x[&k] _numeric_;
   array _z[&t];
   if _n_ eq 1 then _attrlist = symget("&attrs");
   do _j = 1 to &k; _z[_x[_j]] = 1; end;
   _set = _n_;
   do _alt = 1 to &t;
      _attr = scan(_attrlist, _alt, "&delim");
      _bw = 'Best';  _type = sum(0, _z[_alt]); output;
      _bw = 'Worst'; _type = -_type;           output;
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc sort; by bw set alt; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------roll out data set for analysis------;
data __tmpm2(keep=_bw _set _alt rename=(_bw = BW _set = Set _alt = Alt));
   set &data;
   array _b[&b] &bestlist;
   array _w[&b] &worstlist;
   _BW = 'Best '; do _set = 1 to &b; _alt = _b[_set]; output; end;
   _BW = 'Worst'; do _set = 1 to &b; _alt = _w[_set]; output; end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc summary data=__tmpm2 nway completetypes;
   class bw set alt;
   output out=__tmpm3(drop=_type_ rename=_freq_ = Count);
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------merge the data set and the design------;
data __tmpm4;
   merge __tmpm3 __tmpmchdes;
   by bw set alt;
   if count = . then count = 0;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------create and flag chosen and unchosen alts------;
data __tmpm5;
   retain first 1;
   set __tmpm4(keep=count alt);
   if alt = 1 then first = _n_;
   do j = 1 to &t;
      k = first + j - 1;
      set __tmpm4(drop=count alt) point = k;
      c = 2 - (alt eq j);
      output;
      end;
   if _error_ then call symput('abort', '1');
  run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------aggregate the data------;
proc summary data=__tmpm5 nway missing;
   class bw set attr c type;
   output sum(count)=Count out=&out(drop=_type_ _freq_
                                    where=(type ne 0 and count ne 0));
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------do the coding but suppress transreg warnings------;
%if &code %then %do;

   proc printto log=__dummy print=_dummy; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %let i = 0;
   %tryagain:
   proc transreg data=&out design noz nor;
      model ide(type) * class(attr / &classopts);
      output out=&outcoded(drop=_type_ _name_ intercept type)
             separators=' ' ' ' cprefix=0 lprefix=0;
      label type = 'x';
      where type ne 0;
      id bw set count c;
      run;

   %if &syserr > 4 %then %let abort = 1;

   %if &i eq 0 %then %do;
      proc printto log=log print=print; run;

      %if &syserr > 4 %then %let abort = 2;

      %if &abort eq 1 %then %do; %let i = 1; %goto tryagain; %end;
      %end;

   %if &abort %then %goto endit;
   %end;

*------do the analysis, but show sorted parameter estimates------;
%if &code and &analysis %then %do;
   proc phreg data=&outcoded brief;
      ods output parameterestimates=&outparm;
      ods exclude parameterestimates;
      strata bw set;
      where count ne 0;
      freq count;
      model c*c(2) = &_trgind / ties=breslow;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data &outparm;
      set &outparm;
      %if &dodrop %then %do;
         drop %if %sysevalf(&sysver lt 9.2) %then variable; %else parameter;;
         %end;
      label = substr(label, 3);
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %if &sort %then %do;
      proc sort; by descending estimate; run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   data _null_;
      set &outparm;
      file print ods=(template='Stat.Phreg.ParameterEstimates');
      put _ods_;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %if &rscenter or &rsadj or &rsadj100 %then %do;
      proc means data=&outparm noprint;
         var estimate;
         output mean=mean out=__tmpm2;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %if &rsp or &rsp100 %then %do;
      data __tmpm3;
         set &outparm;
         ebeta = exp(estimate);
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      proc means data=__tmpm3 noprint;
         var ebeta;
         output sum=sum out=__tmpm3;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %if &rscenter or &rsadj or &rsadj100 or &rsp or &rsp100 %then %do;
      data &outparm;
         set &outparm;

         %if &rscenter or &rsadj or &rsadj100 %then %do;
            if _n_ eq 1 then set __tmpm2(keep=mean);
            drop mean;
            %end;

         %if &rsp or &rsp100 %then %do;
            if _n_ eq 1 then set __tmpm3(keep=sum);
            drop sum;
            %end;

         %if &rscenter or &rsadj or &rsadj100 %then %do;
            Center = estimate - mean;
            label center = 'Centered Parameter Estimate';
            %end;

         %if &rsp %then %do;
            P = exp(estimate) / sum;
            label p = 'Probability Scaled Parameter Estimate';
            format p 6.4;
            %end;

         %if &rsp100 %then %do;
            P100 = 100 * exp(estimate) / sum;
            label p100 = 'Probability Scaled Parameter Estimate';
            format p100 6.2;
            %end;

         %if &rsadj or &rsadj100 %then %do;
            Adj = exp(center) / (exp(center) + &k - 1);
            label adj = 'Adjusted Parameter Estimate';
            format adj 6.4;
            %end;

         %if &rsadj100 %then %do;
            Adj100 = exp(center) / (exp(center) + &k - 1);
            label adj100 = 'Adjusted Parameter Estimate';
            format adj100 6.2;
            %end;

         label label = '00'x;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      %if &rsadj or &rsadj100 %then %do;
         proc means data=&outparm noprint;
            var adj;
            output sum=sum out=__tmpm3;
            run;

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

         data &outparm;
            set &outparm;

            if _n_ eq 1 then set __tmpm3(keep=sum);
            drop sum;

            %if &rsadj %then adj = adj / sum;
            %else drop adj;;

            %if &rsadj100 %then %do; adj100 = 100 * adj100 / sum; %end;

            if _error_ then call symput('abort', '1');
            run;

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

         %end;

      proc print label data=&outparm(drop=df stderr chisq probchisq hazardratio
                                     %if not &rsdefault %then estimate;);
         id label;
         %if &summit and
             (&rscenter or &rsp or &rsp100 or &rsadj or &rsadj100) %then %do;
            sum %if &rscenter %then center;   %if &rsp %then p;
                %if &rsp100 %then p100;       %if &rsadj %then adj;
                %if &rsadj100 %then adj100;;
            %end;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;
   %end;

proc datasets nolist;
   delete __tmpm2 __tmpm3 __tmpm4 __tmpm5 __tmpmres
          __tmpmbibd __tmpmchdes __tmpmdat __tmpmvars;
   run; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------final data sets, data set options, size notes------;
options &savenote no&timeopt;

data &out; set &out; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%if &code %then %do;
   data &outcoded; set &outcoded; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%if &code and &analysis %then %do;
   data &outparm; set &outparm; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTMDIFF macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTMDIFF macro used &time seconds.;
   %end;

%mend;
