 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktMerge                                            */
 /*   TITLE: Merge a Choice Experimental Design with Choice Data */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs, choice modeling        */
 /*   PROCS:                                                     */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: The following SAS products are required to run      */
 /*          this macro: BASE.                                   */
 /*                                                              */
 /*          This macro creates global macro variables: _MktVer  */
 /****************************************************************/

 /*---------------------------------------------------------------------

This macro merges a data set containing an experimental design for
a choice model with the data for the choice model.  Here is a typical
usage of the macro.

   %mktex(3 ** 4, n=18, seed=7654321)

   data results;
      input Subj (choose1-choose18) (1.) @@;
      datalines;
     1 333542334333314443  2 333212344333333345  3 333212333333313333
     4 133242144334414453  5 335242134333513443  6 333242234333314443
     7 333432334332323443  8 333242234334414443  9 333432331352313343
   ;

   data key;
      input Brand $ & Price $;
      datalines;
   Brand 1  x1
   Brand 2  x2
   Brand 3  x3
   Brand 4  x4
   Another  .
   ;

   %mktroll(design=randomized, key=key, alt=brand, out=rolled)

   %mktmerge(design=rolled, data=results, out=res2,
             nsets=18, nalts=5, setvars=choose1-choose18)

   proc print data=res2(obs=10); by subj set; id subj set; run;

The DESIGN= data set comes from the MktRoll macro.  The DATA= data
set contains the data, and the SETVARS= variables in the DATA= data
set contain the numbers of the chosen alternatives for each of the
18 choice sets.  The NSETS= option specifies the number of choice
sets, and the NALTS= option specifies the number of alternatives.
The OUT= option names the output SAS data set that contains the
experimental design and a variable c that contains 1 for the chosen
alternatives (first choice) and 2 for unchosen alternatives (second
or subsequent choice).

                     Subj    Set     Brand     Price    c

                       1      1     Brand 1      3      2
                                    Brand 2      2      2
                                    Brand 3      2      1
                                    Brand 4      3      2
                                    Another      .      2

                       1      2     Brand 1      1      2
                                    Brand 2      3      2
                                    Brand 3      2      1
                                    Brand 4      2      2
                                    Another      .      2

When the DATA= data set contains a blocking variable, name it on
the BLOCKS= option.  When there is blocking, it is assumed that the
DESIGN= data set contains blocks of NALTS * NSETS observations, one
set per block.  The BLOCKS= variable must contain values 1, 2, ..., n
for n blocks.  Here is an example of using the MktMerge macro with
blocking.

%mktmerge(design=rolled, data=results, out=res2, blocks=form,
          nsets=18, nalts=5, setvars=choose1-choose18)

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

                  /*--------------------------------------------------*/
%macro mktmerge(  /* This macro merges a data set with an             */
                  /* experimental design for a choice model with the  */
                  /* data for a choice model.                         */
                  /*--------------------------------------------------*/
help,             /* nonblank, list options                           */
                  /*--------------------------------------------------*/
design=,          /* The DESIGN= option names an input SAS data set   */
                  /* for a choice model.  This data set may have been */
                  /* created for example with the MktRoll macro.      */
                  /* This option must be specified.                   */
                  /*--------------------------------------------------*/
data=_last_,      /* The DATA= option names an input SAS data set     */
                  /* with data for the choice model.  By default the  */
                  /* DATA= data set is the last data set created.     */
                  /*--------------------------------------------------*/
out=,             /* The OUT= option names the output SAS data set.   */
                  /* If OUT= is not specified, the DATAn convention   */
                  /* is used.  This data set contains the             */
                  /* experimental design and a variable c that        */
                  /* contains 1 for the chosen alternatives (first    */
                  /* choice) and 2 for unchosen alternatives (second  */
                  /* or subsequent choice).                           */
                  /*--------------------------------------------------*/
nsets=,           /* NSETS= is the number of choice sets.             */
                  /* This option must be specified.                   */
                  /*--------------------------------------------------*/
nalts=,           /* NALTS= is the number of alternatives.            */
                  /* This option must be specified.                   */
                  /*--------------------------------------------------*/
setvars=,         /* SETVARS= is a list of variables, one per choice  */
                  /* set, in the DATA= data set that contain the      */
                  /* numbers of the chosen alternatives.  It is       */
                  /* assumed that the values of these variables range */
                  /* from 1 to NALTS.  This option must be specified. */
                  /*--------------------------------------------------*/
statements=,      /* Additional statements like format and label      */
stmts=,           /* statements.                                      */
                  /*--------------------------------------------------*/
blocks=1,         /* BLOCKS= either contains a 1 if there is no       */
                  /* blocking or the name of a variable in the DATA=  */
                  /* data set that contains the block number.  When   */
                  /* there is blocking, it is assumed that the        */
                  /* DESIGN= data set contains blocks of NALTS *      */
                  /* NSETS observations, one set per block.  The      */
                  /* BLOCKS= variable must contain values 1, 2, ...,  */
                  /* n for n blocks.                                  */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _MktVer;
%local  abort time savedsnf savenote setnote saveopts localver timeopt;

%let time     = %sysfunc(datetime());
%let timeopt  = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote = %sysfunc(getoption(notes));
%let saveopts = &savenote missing="%sysfunc(getoption(missing))"
                %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                %sysfunc(getoption(serror));
%let setnote  = nonotes;
%let abort    = 0;
%let localver = 25Jul2010;
%let _mktver  = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES)   %then %let setnote = notes;
   %if %index(&k, VERSION) %then %put MktMerge macro version &localver.;
   %end;
options serror;

options &setnote generic missing='.';

%if %nrbquote(&help) ne %then %do; %mkthelp(mktmerge) %goto endit; %end;

*------rudimentary error checking------;
data _null_;
   n = input(symget('nalts'), ?? 32.);
   if nmiss(n) or n < 1 or n ne round(n) then do;
      error + 1;
      put "ERROR: NALTS=%nrbquote(&nalts) is invalid.";
      end;
   n = input(symget('nsets'), ?? 32.);
   if nmiss(n) or n < 1 or n ne round(n) then do;
      error + 1;
      put "ERROR: NSETS=%nrbquote(&nsets) is invalid.";
      end;
   if symget('design') = ' ' then do;
      error + 1;
      put 'ERROR: DESIGN= must be specified.';
      end;
   if symget('setvars') = ' ' then do;
      error + 1;
      put 'ERROR: SETVARS= must be specified.';
      end;

   if error or _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&data) ne %then %do;
   data _null_; set &data; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )data=&data;
   %put %str(       )out=&out;
   %end;

%if &abort %then %goto endit;

options &savenote no&timeopt;

*------merge data sets------;
data &out;
   set &data;
   retain __printed 0;
   array __c[*] &setvars;
   drop &setvars __:;
   __k = (&blocks - 1) * &nsets * &nalts;
   do __i = 1 to dim(__c);
      do __j = 1 to &nalts;
         __k + 1;
         __was = _error_;
         set &design point=__k nobs=__n;
         if _error_ and not __was then do;
            _error_ = 0;
            if not __printed then do;
               put "ERROR: Probable error in NALTS=%nrbquote(&nalts), "
                   "NSETS=%nrbquote(&nsets), or BLOCKS=%nrbquote(&blocks).";
               if __k > __n then put +7
                  "Accessing DESIGN=%nrbquote(&design) observation "
                  __k 'out of ' __n +(-1) '.';
               __printed = 1;
               call symput('abort', '1');
               end;
            end;
         c = 2 - (__c[__i] eq __j);
         &statements; &stmts;
         output;
         end;
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTMERGE macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTMERGE macro used &time seconds.;
   %end;

%mendit: %mend;
