 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktOrth                                             */
 /*   TITLE: List Orthogonal Designs that MktEx can Produce      */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs, choice modeling        */
 /*   PROCS:                                                     */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: The following SAS products are required to run      */
 /*          this macro: BASE.                                   */
 /*                                                              */
 /*                                                              */
 /*          This macro creates global macro variables: _MktVer  */
 /****************************************************************/

 /*---------------------------------------------------------------------

The MktOrth macro lists some of the 100% orthogonal main-effects plans
that the MktEx macro can generate.  Here is a typical usage.

   %mktorth(range=n eq 12 or n eq 18)

The macro creates data sets and no printed output.  Here are the
results.

   proc print; run;

           Obs     n             Design              Reference

            1     12    2 ** 11                      Hadamard
            2     12    2 **  4  3 **  1             Orthogonal Array
            3     12    2 **  2           6 **  1    Orthogonal Array
            4     12             3 **  1  4 **  1    Full-Factorial
            5     18    2 **  1  3 **  7             Orthogonal Array
            6     18    2 **  1           9 **  1    Full-Factorial
            7     18             3 **  6  6 **  1    Orthogonal Array

You can use the OUTLEV=MktDesLev data set to select potential designs.
Say you are interested in 2 ** 5 3 ** 5 4 ** 1 6 ** 1 in 72 runs.

   %mktorth(range=n=72)
   proc print data=mktdeslev(where=(x2 ge 5 and x3 ge 5 and x4 and x6));
      var n design reference;
      run;

     Obs     n                   Design                      Reference

      22    72    2 ** 43  3 **  8   4 **  1   6 **  1    Orthogonal Array
      38    72    2 ** 36  3 **  9   4 **  1   6 **  1    Orthogonal Array
      42    72    2 ** 35  3 ** 12   4 **  1   6 **  1    Orthogonal Array
      43    72    2 ** 35  3 **  5   4 **  1   6 **  2    Orthogonal Array
      45    72    2 ** 34  3 **  8   4 **  1   6 **  2    Orthogonal Array
      58    72    2 ** 19  3 ** 20   4 **  1   6 **  1    Orthogonal Array
      60    72    2 ** 18  3 ** 16   4 **  1   6 **  2    Orthogonal Array
      62    72    2 ** 17  3 ** 12   4 **  1   6 **  3    Orthogonal Array
      65    72    2 ** 16  3 **  8   4 **  1   6 **  4    Orthogonal Array
      73    72    2 ** 12  3 ** 21   4 **  1   6 **  1    Orthogonal Array
      76    72    2 ** 11  3 ** 24   4 **  1   6 **  1    Orthogonal Array
      78    72    2 ** 11  3 ** 17   4 **  1   6 **  2    Orthogonal Array
      81    72    2 ** 10  3 ** 20   4 **  1   6 **  2    Orthogonal Array
      83    72    2 ** 10  3 ** 13   4 **  1   6 **  3    Orthogonal Array
      85    72    2 **  9  3 ** 16   4 **  1   6 **  3    Orthogonal Array
      87    72    2 **  9  3 **  9   4 **  1   6 **  4    Orthogonal Array
      89    72    2 **  8  3 ** 12   4 **  1   6 **  4    Orthogonal Array
      92    72    2 **  7  3 **  8   4 **  1   6 **  5    Orthogonal Array

------------------------------------------------------------------------

The orthogonal arrays are constructed using methods and arrays from a
variety of sources.  Most helpful were Dey (1985), Hedayat, Sloane, and
Stufken (1999), Sloane (2004), and Wang and Wu (1991).  Some of the
36-run designs (and hence 72, 144, ...) with six-level factors came from
Zhang, Pang, and Wang (2001) and Xu (2002).  Some of the 60, 84, and
100-run designs came from De Cock and Stufken (2000).  Many of the
designs were created using difference schemes and Generalized Hadamard
matrices (de Launey, 1986, 1987ab) and the methods of Wang and Wu (1991)
and Wang (1996).  Many of the newer designs were first developed by Suen
(2003).  Don Anderson, Warwick de Launey, Chung-yi Suen, Neil Sloane,
Randy Tobias, and J.C. Wang all provided valuable help in making the
orthogonal array catalog.

A previous edition of the macro attempted to give specific references
for each design.  This was always problematic.  The macro can generate
over one-hundred thousand orthogonal arrays, many in multiple ways.
Since many of the designs are algorithmically generated using parent
arrays and smaller designs as "children," it is extremely difficult for
the macro to generate an accurate and complete reference list.  For
example, the L36 design, 2 ** 11 3 ** 12, is often associated with
Taguchi (1987).  However, it is generated with a difference scheme
(Bose, 1947; Wang and Wu, 1991) and a Hadamard matrix (Hadamard, 1893;
Paley, 1933) and through recoding (Addelman, 1962).  There are other
relevant references as well.  The reference "Orthogonal Array" refers to
designs created using methods outlined in one or more of the sources
below.  Reference lists for many specific designs can be found in
Hedayat, Sloane, and Stufken (1999) and Sloane (2004).
The reference list can be found in mktex.sas.

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*==================== Macro Argument List Start ======================;

%macro mktorth(   /*--------------------------------------------------*/
                  /*                                                  */
help,             /* nonblank, list options                           */
                  /*                                                  */
outcat=mktdescat, /* Output data set with the list of designs that    */
                  /* the MktEx macro can create.                      */
                  /*                                                  */
outlev=mktdeslev, /* Output data set with the list of designs and 50  */
                  /* (by default) more variables:                     */
                  /* x2 - number of two-level factors,                */
                  /* x3 - number of three-level factors and so on.    */
                  /* The number of x variables is determined by the   */
                  /* MAXLEV= option.                                  */
                  /*                                                  */
outall=,          /* Output data set with all designs.  This is like  */
                  /* the OUTLEV= data set, except larger.  For        */
                  /* example, the OUTLEV= data set has x2=2 x3=2, and */
                  /* the OUTALL= data set has that design and also    */
                  /* x1=2 x3=1, x1=1 x3=2, x1=1 x2=1.  This data set  */
                  /* is not created by default.  Unless you also      */
                  /* specify a reasonably small RANGE= or MAXN=       */
                  /* specification, the OUTALL= specification takes   */
                  /* a *LONG* time and can create a *HUGE* data set.  */
                  /*                                                  */
options=,         /* mktex - macro is being called from the MktEx     */
                  /* macro and just the OUTLEV= data set is needed.   */
                  /* The macro takes short cuts to make it run faster */
                  /* doing only what MktEx needs.                     */
                  /*                                                  */
                  /* ADX - macro is being called by ADX.              */
                  /*                                                  */
                  /* mktruns - macro is being called from the MktRuns */
                  /* macro and just the OUTLEV= data set is needed.   */
                  /* The macro takes short cuts to make it run faster */
                  /* doing only what MktRuns needs.                   */
                  /*                                                  */
                  /* lineage - output design lineage.                 */
                  /*                                                  */
                  /* parent - just output parent designs.             */
                  /*                                                  */
                  /* dups - do not filter out duplicate and inferior  */
                  /* designs from the catalog.  This can be useful    */
                  /* when you are creating a data set for the CAT=    */
                  /* option in the MktEx macro.                       */
                  /*                                                  */
                  /* 512 - adds some larger designs in 512 runs with  */
                  /* mixes of 16, 8, 4, and 2-level factors to the    */
                  /* catalog, which gives added flexibility in 512    */
                  /* runs at a cost of much slower run time.  This    */
                  /* option replaces the default 4^160 32^1 parent    */
                  /* with 16^32 32^1.                                 */
                  /*                                                  */
range=,           /* specifies the number of runs of interest.        */
                  /* Specify a range involving "n", the number of     */
                  /* runs.  Examples:                                 */
                  /* range=n=36,                                      */
                  /* range=18 le n le 36,                             */
                  /* range=n eq 18 or n eq 36,                        */
                  /*                                                  */
maxn=,            /* similar to range, specifies the maximum number   */
                  /* of runs of interest.  Specifying small numbers   */
                  /* (e.g. n le 200) will make the macro run faster.  */
                  /*                                                  */
maxlev=50,        /* specifies the maximum number of levels to        */
                  /* consider.  This option controls the number of x  */
                  /* variables in the OUTLEV= data set.  It also      */
                  /* excludes from consideration designs with factors */
                  /* of more than MAXLEV= levels so it affects the    */
                  /* number of rows in the output data sets.  Specify */
                  /* a value 2 <= maxlev <= 144.  Note that           */
                  /* specifying MAXLEV=n does not preclude designs    */
                  /* with more than n-level factors from being used   */
                  /* as parents for other designs, it just precludes  */
                  /* the larger designs from being output.  For       */
                  /* example, with MAXLEV=3, the design               */
                  /* 3 ** 12 12 ** 1 in 36 runs is used to make       */
                  /* 2 ** 11 3 ** 12 before the 3 ** 12 12 ** 1       */
                  /* design is discarded.                             */
                  /*                                                  */
filter=0,         /* specifies extra design catalog filtering.        */
                  /* Usually, you will never have to use this option. */
                  /* By default, MktOrth filters out inferior         */
                  /* designs.  On the one hand, this process is       */
                  /* expensive, but it also saves some time and       */
                  /* resources by limiting the information that must  */
                  /* be processed.  Care is taken do to only the      */
                  /* minimum amount of work to filter.  However, this */
                  /* gets complicated.  If you specify MAXLEV=144,    */
                  /* the maximum, you will get perfect filtering of   */
                  /* duplicates in a reasonable amount of time.       */
                  /* MktOrth uses internal and optimized constants to */
                  /* avoid doing extra work.  Unfortunately, these    */
                  /* constants may not be optimal for any other       */
                  /* MAXLEV= value.  This is almost never going to be */
                  /* a real issue.  If however, you want to specify   */
                  /* both MAXLEV= and ensure you get better           */
                  /* filtering, specify FILTER=n (for some n) and you */
                  /* may get a few more inferior designs filtered     */
                  /* out.  The n value increases how deep into the    */
                  /* catalog MktOrth searches for inferior designs.   */
                  /* Larger values find more designs to exclude at a  */
                  /* cost of greater run time.  For example, if you   */
                  /* specify both of the following,                   */
                  /*    %mktorth(maxlev=20)                           */
                  /*    %mktorth(maxlev=20, filter=1000)              */
                  /* you will find that the latter specification      */
                  /* removes on the order of thirty more designs but  */
                  /* at cost of much slower run time.  Actually,      */
                  /* FILTER=27 is large enough for this example,      */
                  /* and it has a negligible effect on run time, but  */
                  /* you have to run the macro multiple times to      */
                  /* figure that out.  Values of a couple hundred or  */
                  /* so are probably always going to be sufficient.   */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _mktver;
%local  abort time saveopts savenote setnote mktex mktruns rangeflag
        flow adx k parents gotsome lineage primes h tempmaxlev needlin
        nfilter bylist nodups bigkid localver makelist fflist do512 timeopt;

%let time     = %sysfunc(datetime());
%let timeopt  = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote = %sysfunc(getoption(notes));
%let saveopts = &savenote missing="%sysfunc(getoption(missing))"
                %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                %sysfunc(getoption(serror));
%let setnote  = nonotes no&timeopt;
%let abort    = 0;
%let flow     = 0;
%let primes   = 2,  3,  5,  7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53,
59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139;

%let localver = 25Jul2010;
%let _mktver  = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES)   %then %let setnote = notes &timeopt;
   %if %index(&k, VERSION) %then %put MktOrth macro version &localver.;
   %let flow = (%index(&k, FLOW) gt 0);
   %end;
options serror;

options &setnote generic missing='.';

%if %nrbquote(&help) ne %then %do; %mkthelp(mktorth) %goto endit; %end;

%if &flow %then
   %put ** MktOrth: starting, maxlev=&maxlev, range=&range, maxn=&maxn..;

*------some basic option processing------;
data _null_;
   abort = 0;
   length list $ 100 by $ 1000;
   call symput('gotsome', '0');
   call symput('rangeflag', put(symget('range') ne ' ', 1.));
   list = lowcase(symget('options'));
   i = index(list, 'adx');    if i then substr(list, i, 3) = ' ';
   call symput('adx', put(i ne 0, 1.));
   i = index(list, 'dups');    if i then substr(list, i, 4) = ' ';
   call symput('nodups', put(i eq 0, 1.));
   i = index(list, 'mktex');   if i then substr(list, i, 5) = ' ';
   mktex = (i gt 0);
   call symput('mktex', put(mktex, 1.));
   i = index(list, 'mktruns'); if i then substr(list, i, 7) = ' ';
   call symput('mktruns', put(i gt 0, 1.));
   i = index(list, '512'); if i then substr(list, i, 3) = ' ';
   call symput('do512', put(i gt 0, 1.));
   i = index(list, 'parents'); if i then substr(list, i, 7) = ' ';
   j = index(list, 'parent' ); if j then substr(list, j, 6) = ' ';
   k = index(list, 'lineage'); if k then substr(list, k, 7) = ' ';
   call symput('parents', put(i or j, 1.));
   call symput('lineage', put(k gt 0, 1.));
   call symput('needlin', put(i or j or k, 1.));
   if i or j or k then call symput('h', 'h');
   list = compbl(list);

   if list ne ' ' then do;
      put 'ERROR: OPTIONS=' list 'is invalid.';
      abort = 1;
      end;

   f = input(symget('filter'), ?? 23.);
   if nmiss(f) or f lt 0 or f ne round(f) then do;
      put "ERROR: Invalid filter=&filter..";
      abort = 1;
      end;

   call symput('bigkid', '144');
   call symput('tempmaxlev', symget('bigkid'));
   maxlev = input(symget('maxlev'), ?? 23.);
   ok = n(maxlev);
   i = input(symget('bigkid'), 23.);
   if ok then ok = (maxlev eq floor(maxlev) and 2 le maxlev le i);
   if not ok then do;
      if not mktex then put "ERROR: Invalid maxlev=&maxlev..";
      abort = 1;
      end;

   else do;
      by = '';
      do i = maxlev to 1 by -1; by = trim(by) || ' x' || put(i, 3. -L); end;
      call symput('bylist', left(compbl(by)));
      end;

   if symget('outcat') eq ' ' then call symput('outcat', 'MktDesCat');
   if symget('outlev') eq ' ' then call symput('outlev', 'MktDesLev');

   if _error_ or abort then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*======================= List the Parent Designs ======================;
data &outlev(drop=i p q s d4 nh hadsize gotsome);
   length n 8 s $ 60 p q $ 12 x1-x&tempmaxlev 4
          %if &needlin %then Lineage $ %eval((1 + &do512) * 440);;
   array x[&tempmaxlev];

   %if &nodups %then %do;
      %let nfilter = 23; /* set makelist to 1 below, to regenerate these */
      array _c[&nfilter] _temporary_ (2 3 12 5 4 14 6 19 6 11 43 8
                                      52 4 113 2 3 187 22 70 5 2 2);
      array _m[&nfilter] _temporary_ ( 36  72  80 100 108 112 144 160
                                      200 216 240 288 320 324 336 360
                                      392 400 432 448 500 504 576);
      %end;

   *------Hadamards, not otherwise made by OAs------;
   do n = 148 to 1000 by 4;
      if not (n in (256, 452, 476, 508, 512, 532, 652, 668, 716, 764, 772, 836,
          852, 876, 892, 904, 932, 940, 952, 956, 964, 980, 988, 996)) then do;

         hadsize = n;
         do nh = 0 to 20 while(mod(hadsize, 8) eq 0);
            hadsize = hadsize / 2;
            end;

         if      nh ge 6 then s = '2 ** ' || put(n - 128, 4.0 -L) || '128 ** 1';
         else if nh eq 5 then s = '2 ** ' || put(n -  64, 4.0 -L) || ' 64 ** 1';
         else if nh eq 4 then s = '2 ** ' || put(n -  32, 4.0 -L) || ' 32 ** 1';
         else if nh eq 3 then s = '2 ** ' || put(n -  16, 4.0 -L) || ' 16 ** 1';
         else if nh eq 2 then s = '2 ** ' || put(n -   8, 4.0 -L) || '  8 ** 1';
         else if nh eq 1 then s = '2 ** ' || put(n -   4, 4.0 -L) || '  4 ** 1';
         else                 s = '2 ** ' || put(n -   1, 4.0 -L);
         link expand;

         if not (n in (192, 224, 288, 432, 448)) then do;
            d4 = min(2 ** nh - 1, hadsize - 1);
            if d4 gt 1 and mod(n, 128) then do;
               s  = '2 ** ' || put(n - 1 - 3 * d4, 4.0 -L) ||
                    ' 4 ** ' || put(d4, 4.0 -L);
               link expand;
               end;
            end;
         end;
      end;

   *------small parents, n ** 1, handled entirely by expansion------;
   do n = 4 to &bigkid;
      if not (n in (&primes, 128)) then do;
         s    = put(n, 3.0 -L) || ' ** 1';
         link expand;
         end;
      end;

   *------do these with FACTEX but get them in the catalog------;
   %if not &mktex %then %do;
      s = '2 ** 64                  64 ** 1'; n = 128; link expand;
      s = '2 **  6 4 **  5  8 ** 13 16 ** 1';          link expand;
      s = '2 **  4 4 **  8  8 ** 12 16 ** 1';          link expand;
      s = '2 **  5 4 ** 10  8 ** 11 16 ** 1';          link expand;
      s = '2 **  6 4 ** 12  8 ** 10 16 ** 1';          link expand;
      s = '2 **  4 4 ** 15  8 **  9 16 ** 1';          link expand;
      s = '2 **  5 4 ** 17  8 **  8 16 ** 1';          link expand;
      s = '2 **  6 4 ** 19  8 **  7 16 ** 1';          link expand;
      s = '2 **  4 4 ** 22  8 **  6 16 ** 1';          link expand;
      s = '2 **  5 4 ** 24  8 **  5 16 ** 1';          link expand;
      s = '2 **  6 4 ** 26  8 **  4 16 ** 1';          link expand;
      s = '2 **  4 4 ** 29  8 **  3 16 ** 1';          link expand;
      s = '2 **  5 4 ** 31  8 **  2 16 ** 1';          link expand;
      s = '2 **  6 4 ** 33  8 **  1 16 ** 1';          link expand;
      s = '2 **  4 4 ** 36          16 ** 1';          link expand;
      s = '2 **  5 4 **  8  8 ** 14        ';          link expand;
      s = '2 **  3 4 ** 11  8 ** 13        ';          link expand;
      s = '2 **  3 4 ** 18  8 ** 10        ';          link expand;
      s = '2 **  3 4 ** 25  8 **  7        ';          link expand;
      %end;

   *--------------------------------------------------------------------------
   | When adding new designs, be sure to check fast filtering below.         |
   | Use the makelist flag to make a new list of numbers for filtering.      |
   --------------------------------------------------------------------------;

   *------large parent designs------;
   s = '         4 **  32            32 ** 1        '; n = 128; link expand;
   s = '                   8 **  16  16 ** 1        ';          link expand;
   s = '                   7 **   9  21 ** 1        '; n = 147; link expand;
   s = '                   5 **  11  30 ** 1        '; n = 150; link expand;
   s = '2 **  76                     76 ** 1        '; n = 152; link expand;
   s = '         3 **  25            17 ** 1        '; n = 153; link expand;
   s = '2 ** 133           5 **   1  16 ** 1        '; n = 160; link expand;
   s = '2 ** 127                     10 ** 1 16 ** 1';          link expand;
   s = '2 **  80                     80 ** 1        ';          link expand;
   s = '                   4 **  16  40 ** 1        ';          link expand;
   s = '         3 **  65  6 **   1  27 ** 1        '; n = 162; link expand;
   s = '         3 **  54            54 ** 1        ';          link expand;
   s = '                   9 **  18  18 ** 1        ';          link expand;
   s = '                             13 ** 14       '; n = 169; link expand;
   s = '2 **  84                     84 ** 1        '; n = 168; link expand;
   s = '         3 **  28            19 ** 1        '; n = 171; link expand;
   s = '                   5 **  10  35 ** 1        '; n = 175; link expand;
   s = '2 **  88                     88 ** 1        '; n = 176; link expand;
   s = '                   4 **  12  44 ** 1        ';          link expand;
   s = '         3 **  30            60 ** 1        '; n = 180; link expand;
   s = '                   6 **   2  30 ** 1        ';          link expand;
   s = '2 **  92                     92 ** 1        '; n = 184; link expand;
   s = '         3 **  36            63 ** 1        '; n = 189; link expand;
   s = '2 **  96                     96 ** 1        '; n = 192; link expand;
   s = '                   4 **  48  48 ** 1        ';          link expand;
   s = '                   8 **   8  24 ** 1        ';          link expand;
   s = '                   7 **  28  28 ** 1        '; n = 196; link expand;
   s = '                  14 **   5                 ';          link expand;
   s = '         3 **  30            66 ** 1        '; n = 198; link expand;
   s = '2 ** 100                    100 ** 1        '; n = 200; link expand;
   s = '                   5 **  20  40 ** 1        ';          link expand;
   s = '                  10 **   5  20 ** 1        ';          link expand;
   s = '         3 **  25            23 ** 1        '; n = 207; link expand;
   s = '2 ** 104                    104 ** 1        '; n = 208; link expand;
   s = '                   4 **  16  52 ** 1        ';          link expand;
   s = '2 ** 108                    108 ** 1        '; n = 216; link expand;
   s = '2 **  11 3 **  77            12 ** 1 18 ** 1';          link expand;
   s = '         3 **  72            72 ** 1        ';          link expand;
   s = '         3 **  66  6 **   5  12 ** 1 18 ** 1';          link expand;
   s = '                   6 **   7  36 ** 1        ';          link expand;
   s = '2 ** 193           7 **   1  16 ** 1        '; n = 224; link expand;
   s = '2 ** 183                     14 ** 1 16 ** 1';          link expand;
   s = '2 ** 112                    112 ** 1        ';          link expand;
   s = '                   4 **  56  56 ** 1        ';          link expand;
   s = '         3 **  27            75 ** 1        '; n = 225; link expand;
   s = '                  15 **   6                 ';          link expand;
   s = '                   5 **  20  45 ** 1        ';          link expand;
   s = '2 ** 116                    116 ** 1        '; n = 232; link expand;
   s = '         3 **  30            78 ** 1        '; n = 234; link expand;
   s = '2 ** 205           5 **   1  24 ** 1        '; n = 240; link expand;
   s = '2 ** 199                     10 ** 1 24 ** 1';          link expand;
   s = '2 ** 120                    120 ** 1        ';          link expand;
   s = '                   4 **  20  60 ** 1        ';          link expand;
   s = '                  11 **  22  22 ** 1        '; n = 242; link expand;
   s = '         3 **  81            81 ** 1        '; n = 243; link expand;
   s = '                   9 **  27  27 ** 1        ';          link expand;
   s = '                   7 **  10  35 ** 1        '; n = 245; link expand;
   s = '2 ** 124                    124 ** 1        '; n = 248; link expand;
   s = '                   5 **  50  50 ** 1        '; n = 250; link expand;
   s = '         3 **  42            84 ** 1        '; n = 252; link expand;
   s = '                   6 **   2  42 ** 1        ';          link expand;
   s = '                   8 **  32  32 ** 1        '; n = 256; link expand;
   s = '                  16 **  17                 ';          link expand;
   s = '         3 **  27            87 ** 1        '; n = 261; link expand;
   s = '2 ** 132                    132 ** 1        '; n = 264; link expand;
   s = '         3 **  90            90 ** 1        '; n = 270; link expand;
   s = '2 ** 136                    136 ** 1        '; n = 272; link expand;
   s = '                   4 **  32  68 ** 1        ';          link expand;
   s = '                   5 **  11  55 ** 1        '; n = 275; link expand;
   s = '         3 **  30            93 ** 1        '; n = 279; link expand;
   s = '2 ** 140                    140 ** 1        '; n = 280; link expand;
   s = '2 ** 253           9 **   1  16 ** 1        '; n = 288; link expand;
   s = '2 ** 239                     16 ** 1 18 ** 1';          link expand;
   s = '2 ** 144                    144 ** 1        ';          link expand;
   s = '         3 **  96            96 ** 1        ';          link expand;
   s = '                   4 **  36  72 ** 1        ';          link expand;
   s = '                   6 **  10  48 ** 1        ';          link expand;
   s = '                             12 ** 6 24 ** 1';          link expand;
   s = '                             17 ** 18       '; n = 289; link expand;
   s = '                   7 **  18  42 ** 1        '; n = 294; link expand;
   s = '         3 **  39            99 ** 1        '; n = 297; link expand;
   s = '                   5 **  20  60 ** 1        '; n = 300; link expand;
   s = '                  10 **   2  30 ** 1        ';          link expand;
   s = '2 ** 228                     76 ** 1        '; n = 304; link expand;
   s = '                   4 **  16  76 ** 1        ';          link expand;
   s = '         3 **  48           102 ** 1        '; n = 306; link expand;
   s = '         3 **  29           105 ** 1        '; n = 315; link expand;
   s = '2 ** 240                     80 ** 1        '; n = 320; link expand;
   s = '                   4 **  40  80 ** 1        ';          link expand;
   s = '                   8 **  10  40 ** 1        ';          link expand;
   s = '         3 ** 143            12 ** 1 27 ** 1'; n = 324; link expand;
   s = '         3 ** 108           108 ** 1        ';          link expand;
   s = '                   6 **   2  54 ** 1        ';          link expand;
   s = '                   9 **  36  36 ** 1        ';          link expand;
   s = '                  18 **   3                 ';          link expand;
   s = '                   5 **  20  65 ** 1        '; n = 325; link expand;
   s = '         3 **  36           111 ** 1        '; n = 333; link expand;
   s = '2 ** 297           7 **   1  24 ** 1        '; n = 336; link expand;
   s = '2 ** 287                     14 ** 1 24 ** 1';          link expand;
   s = '2 ** 252                     84 ** 1        ';          link expand;
   s = '                   4 **  36  84 ** 1        ';          link expand;
   s = '                  13 **  26  26 ** 1        '; n = 338; link expand;
   s = '         3 **  30           114 ** 1        '; n = 342; link expand;
   s = '                   7 **  49  49 ** 1        '; n = 343; link expand;
   s = '                   5 **  20  70 ** 1        '; n = 350; link expand;
   s = '         3 **  39           117 ** 1        '; n = 351; link expand;
   s = '2 ** 264                     88 ** 1        '; n = 352; link expand;
   s = '                   4 **  32  88 ** 1        ';          link expand;
   s = '         3 **  48           120 ** 1        '; n = 360; link expand;
   s = '                   6 **   8  60 ** 1        ';          link expand;
   s = '                             19 ** 20       '; n = 361; link expand;
   s = '                  11 **  11  33 ** 1        '; n = 363; link expand;
   s = '2 ** 276                     92 ** 1        '; n = 368; link expand;
   s = '                   4 **  36  92 ** 1        ';          link expand;
   s = '         3 **  30           123 ** 1        '; n = 369; link expand;
   s = '                   5 **  40  75 ** 1        '; n = 375; link expand;
   s = '         3 **  72           126 ** 1        '; n = 378; link expand;
   s = '                   4 **  96  96 ** 1        '; n = 384; link expand;
   s = '                   8 **  16  48 ** 1        ';          link expand;
   s = '         3 **  48           129 ** 1        '; n = 387; link expand;
   s = '2 ** 196           7 **  28  28 ** 1        '; n = 392; link expand;
   s = '                   7 **  28  56 ** 1        ';          link expand;
   s = '                  14 **   5  28 ** 1        ';          link expand;
   s = '         3 ** 132           132 ** 1        '; n = 396; link expand;
   s = '                   6 **   2  66 ** 1        ';          link expand;
   s = '2 ** 300                    100 ** 1        '; n = 400; link expand;
   s = '                   4 **  36 100 ** 1        ';          link expand;
   s = '                   5 **  80  80 ** 1        ';          link expand;
   s = '                  10 **   6  40 ** 1        ';          link expand;
   s = '                  20 **   5                 ';          link expand;
   s = '         3 **  81           135 ** 1        '; n = 405; link expand;
   s = '                   9 **  18  45 ** 1        ';          link expand;
   s = '         3 **  48           138 ** 1        '; n = 414; link expand;
   s = '2 ** 312                    104 ** 1        '; n = 416; link expand;
   s = '                   4 **  48 104 ** 1        ';          link expand;
   s = '         3 **  30           141 ** 1        '; n = 423; link expand;
   s = '                   5 **  20  85 ** 1        '; n = 425; link expand;
   s = '2 ** 389           9 **   1  24 ** 1        '; n = 432; link expand;
   s = '2 ** 375                     18 ** 1 24 ** 1';          link expand;
   s = '         3 ** 144           144 ** 1        ';          link expand;
   s = '                   4 ** 108 108 ** 1        ';          link expand;
   s = '                   6 **  12  72 ** 1        ';          link expand;
   s = '                             12 ** 6 36 ** 1';          link expand;
   s = '         3 **  42  7 **   7  21 ** 1        '; n = 441; link expand;
   s = '                   7 **  14  63 ** 1        ';          link expand;
   s = '                  21 **   7                 ';          link expand;
   s = '2 ** 336                    112 ** 1        '; n = 448; link expand;
   s = '                   4 **  56 112 ** 1        ';          link expand;
   s = '                   8 **  56  56 ** 1        ';          link expand;
   s = '         3 ** 150  5 **  11  30 ** 1        '; n = 450; link expand;
   s = '                   5 **  90  90 ** 1        ';          link expand;
   s = '                  15 **   5  30 ** 1        ';          link expand;
   s = '         3 **  72  9 **   1  17 ** 1        '; n = 459; link expand;
   s = '2 ** 348                    116 ** 1        '; n = 464; link expand;
   s = '                   4 **  36 116 ** 1        ';          link expand;
   s = '         3 **  49            52 ** 1        '; n = 468; link expand;
   s = '                   6 **   2  78 ** 1        ';          link expand;
   s = '                   5 **  20  95 ** 1        '; n = 475; link expand;
   s = '         3 **  37            53 ** 1        '; n = 477; link expand;
   s = '2 ** 360                    120 ** 1        '; n = 480; link expand;
   s = '                   4 **  56 120 ** 1        ';          link expand;
   s = '                  22 **   3                 '; n = 484; link expand;
   s = '                  11 **  44  44 ** 1        ';          link expand;
   s = '                   9 **  54  54 ** 1        '; n = 486; link expand;
   s = '                   7 **  18  70 ** 1        '; n = 490; link expand;
   s = '         3 **  42  5 **   1  33 ** 1        '; n = 495; link expand;
   s = '                   4 **  18 124 ** 1        '; n = 496; link expand;
   s = '2 ** 372                    124 ** 1        ';          link expand;
   s = '                   5 ** 100 100 ** 1        '; n = 500; link expand;
   s = '                  10 **   2  50 ** 1        ';          link expand;
   s = '2 **  84 3 **  84            84 ** 1        '; n = 504; link expand;
   s = '                   6 **   8  84 ** 1        ';          link expand;
   s = '                   8 **  64  64 ** 1        '; n = 512; link expand;
   s = %if &do512 %then ' 16 **  32  32 ** 1        ';
                  %else '  4 ** 160  32 ** 1        ';;         link expand;
   s = '         3 **  81  9 **   1  19 ** 1        '; n = 513; link expand;
   s = '                  24 **   8                 '; n = 576; link expand;
   if gotsome then call symput('gotsome', '1');
   if _error_ then call symput('abort', '1');
   return;

   expand: *------expand string to x1 - x&tempmaxlev------;
      %if &nodups %then %do;
         count = &filter;
         do i = 1 to &nfilter;
            if _m[i] eq n then count = _c[i] + &filter;
            if _m[i] ge n then i = %eval(&nfilter + 1);
            end;
         %end;

      %if &maxn ne %then %do; if n le &maxn then do; %end;
      %if &rangeflag %then %do; if &range then do; %end;

         do i = 1 to &tempmaxlev; x[i] = 0; end;
         do i = 1 to 40 by 2 until(p eq ' ');
            p = scan(s, i,     ' *');
            q = scan(s, i + 1, ' *');
            if p ne ' ' then x[input(p, 12.)] = input(q, 12.);
            end;
         %if &needlin %then %do;
            lineage = left(compbl(s));
            %end;
         gotsome + 1;
         output;
         %if &rangeflag %then %do; end; %end;
         %if &maxn ne %then %do; end; %end;

      return;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%if &maxn eq %then %let maxn = 1000;

*===================== Start Procreating Children =====================;
%if &gotsome %then %do;

   *------expand all the higher-level factors------;
   %macro mktmakeit(f, m1, n1, m2, n2, m3, n3, m4, n4, m5, n5);
      %if &needlin %then %do;
         lineage = trim(lineage) || " : &f ** 1 > &m1 ** &n1"
                   %if &n2 %then || " &m2 ** &n2";
                   %if &n3 %then || " &m3 ** &n3";
                   %if &n4 %then || " &m4 ** &n4";
                   %if &n5 %then || " &m5 ** &n5";;
         %end;
      do while(x&f);
                            x&m1 + &n1;
         %if &m2 %then %do; x&m2 + &n2; %end;
         %if &m3 %then %do; x&m3 + &n3; %end;
         %if &m4 %then %do; x&m4 + &n4; %end;
         %if &m5 %then %do; x&m5 + &n5; %end;
         x&f + -1;
         %if &parents %then if scan(lineage, 3, ':') eq ' ' then; output;
         %if &needlin and (&f eq 12 or &f eq 16) %then %do;
            lineage = trim(lineage) || " : &f ** 1 > &m1 ** &n1"
                      %if &n2 %then || " &m2 ** &n2";
                      %if &n3 %then || " &m3 ** &n3";
                      %if &n4 %then || " &m4 ** &n4";
                      %if &n5 %then || " &m5 ** &n5";;
            %end;
         end;
      do i = 1 to &tempmaxlev; x[i] = z[i]; end;
      %if &needlin %then %do; lineage = h; %end;
      %mend;

   *------expand parents > 72.  72 is done in next step------;
   %if &maxn ge 74 %then %do;
      data &outlev(drop=z: i &h); set &outlev; output;
         %if &needlin %then %do; h = lineage; %end;
         array x[&tempmaxlev]; array z[&tempmaxlev];
         do i = 1 to &tempmaxlev; z[i] = x[i]; end;

         %mktmakeit(144, 2,136,8, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(144, 2,117,8, 1, 9, 1, 0, 0, 0, 0)
         %mktmakeit(144, 2,113,3, 1,24, 1, 0, 0, 0, 0)
         %mktmakeit(144, 2,111,6, 1,24, 1, 0, 0, 0, 0)
         %mktmakeit(144, 2,103,8, 1,18, 1, 0, 0, 0, 0)
         %mktmakeit(144, 2,76, 3, 7, 4, 1, 6, 5,12, 1)
         %mktmakeit(144, 2,76, 3,12, 6, 4, 8, 1, 0, 0)
         %mktmakeit(144, 2,75, 3, 3, 4, 1, 6, 6,12, 1)
         %mktmakeit(144, 2,74, 3, 4, 6, 6, 8, 1, 0, 0)
         %mktmakeit(144, 2,72,72, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(144, 2,44, 3,11,12, 2, 0, 0, 0, 0)
         %mktmakeit(144, 2,16, 3, 3, 6, 6,24, 1, 0, 0)
         %mktmakeit(144, 3,48,48, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(144, 4,36,36, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(144, 4,11,12, 2, 0, 0, 0, 0, 0, 0)
         %mktmakeit(144,12, 7, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(143,11, 1,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(142, 2, 1,71, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(141, 3, 1,47, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(140, 2,139,0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(140, 2,38, 7, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(140, 2,36,10, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(140, 2,34,14, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(140, 2,27, 5, 1, 7, 1, 0, 0, 0, 0)
         %mktmakeit(140, 2,25, 5, 1,14, 1, 0, 0, 0, 0)
         %mktmakeit(140, 2,21, 7, 1,10, 1, 0, 0, 0, 0)
         %mktmakeit(140, 2,22,35, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(140, 2,17,10, 1,14, 1, 0, 0, 0, 0)
         %mktmakeit(140, 2, 2,70, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(140, 4, 1,35, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(140, 5, 1,28, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(140, 7, 1,20, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(138, 2, 1,69, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(138, 3, 1,46, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(138, 6, 1,23, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(136, 2,132,4, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(136, 2,83, 4, 1,17, 1, 0, 0, 0, 0)
         %mktmakeit(136, 2,68,68, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(136, 2,67, 4, 1,34, 1, 0, 0, 0, 0)
         %mktmakeit(136, 8, 1,17, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(135, 3,32, 9, 1,15, 1, 0, 0, 0, 0)
         %mktmakeit(135, 3,27,45, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(135, 5, 1,27, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(134, 2, 1,67, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(133, 7, 1,19, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(132, 2,131,0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(132, 2,42, 6, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(132, 2,27,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(132, 2,22,33, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(132, 2,18, 3, 1,22, 1, 0, 0, 0, 0)
         %mktmakeit(132, 2,18, 6, 1,11, 1, 0, 0, 0, 0)
         %mktmakeit(132, 2,15, 6, 1,22, 1, 0, 0, 0, 0)
         %mktmakeit(132, 2, 2,66, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(132, 3, 1,44, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(132, 4, 1,33, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(132,11, 1,12, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(130, 2, 1,65, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(130, 5, 1,26, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(130,10, 1,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(129, 3, 1,43, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(128, 2,64,64, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(126, 2, 1,63, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(126, 3,24,14, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(126, 3,23, 6, 1, 7, 1, 0, 0, 0, 0)
         %mktmakeit(126, 3,20, 6, 1,21, 1, 0, 0, 0, 0)
         %mktmakeit(126, 3,21,42, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(126, 7, 1,18, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(126, 9, 1,14, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(125, 5,25,25, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(124, 2,123,0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(124, 2,22,31, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(124, 2, 2,62, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(124, 4, 1,31, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(123, 3, 1,41, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(122, 2, 1,61, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(121,11,12, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(120, 2,116,4, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(120, 2,87, 3, 1, 4, 1, 0, 0, 0, 0)
         %mktmakeit(120, 2,79, 4, 1, 5, 1, 0, 0, 0, 0)
         %mktmakeit(120, 2,75, 4, 1, 6, 1, 0, 0, 0, 0)
         %mktmakeit(120, 2,75, 4, 1,10, 1, 0, 0, 0, 0)
         %mktmakeit(120, 2,74, 4, 1,15, 1, 0, 0, 0, 0)
         %mktmakeit(120, 2,70, 3, 1, 4, 1,10, 1, 0, 0)
         %mktmakeit(120, 2,70, 4, 1, 5, 1, 6, 1, 0, 0)
         %mktmakeit(120, 2,68, 4, 1, 6, 1,10, 1, 0, 0)
         %mktmakeit(120, 2,60,60, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(120, 2,59, 4, 1,30, 1, 0, 0, 0, 0)
         %mktmakeit(120, 2,30, 6, 1,20, 1, 0, 0, 0, 0)
         %mktmakeit(120, 2,28,10, 1,12, 1, 0, 0, 0, 0)
         %mktmakeit(120, 3, 1,40, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(120, 5, 1,24, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(120, 8, 1,15, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(119, 7, 1,17, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(118, 2, 1,59, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(117, 3,13,39, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(117, 9, 1,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(116, 2,115,0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(116, 2,23,29, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(116, 2, 2,58, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(116, 4, 1,29, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(115, 5, 1,23, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(114, 2, 1,57, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(114, 3, 1,38, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(114, 6, 1,19, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(112, 2,104,8, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(112, 2,89, 7, 1, 8, 1, 0, 0, 0, 0)
         %mktmakeit(112, 2,79, 8, 1,14, 1, 0, 0, 0, 0)
         %mktmakeit(112, 2,75, 4, 3,28, 1, 0, 0, 0, 0)
         %mktmakeit(112, 2,56,56, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(112, 4,12,28, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(112, 7, 1,16, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(111, 3, 1,37, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(110, 2, 1,55, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(110, 5, 1,22, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(110,10, 1,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,107,0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,34, 3,29, 6, 1, 0, 0, 0, 0)
         %mktmakeit(108, 3,44, 9, 1,12, 1, 0, 0, 0, 0)
         %mktmakeit(108, 3,36,36, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,40, 6, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,27, 3,33, 9, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,22,27, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,20, 3,34, 9, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,18, 3,33, 6, 1, 9, 1, 0, 0)
         %mktmakeit(108, 2,18, 3,31,18, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,21, 3, 1, 6, 2, 0, 0, 0, 0)
         %mktmakeit(108, 2,17, 3,29, 6, 2, 0, 0, 0, 0)
         %mktmakeit(108, 2,15, 6, 1,18, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,13, 3,30, 6, 1,18, 1, 0, 0)
         %mktmakeit(108, 2,13, 6, 3, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,12, 3,29, 6, 3, 0, 0, 0, 0)
         %mktmakeit(108, 2,10, 3,40, 6, 1, 9, 1, 0, 0)
         %mktmakeit(108, 2,10, 3,33, 6, 2, 9, 1, 0, 0)
         %mktmakeit(108, 2,10, 3,31, 6, 1,18, 1, 0, 0)
         %mktmakeit(108, 2, 9, 3,36, 6, 2, 9, 1, 0, 0)
         %mktmakeit(108, 2, 9, 3,34, 6, 1,18, 1, 0, 0)
         %mktmakeit(108, 2, 8, 3,30, 6, 2,18, 1, 0, 0)
         %mktmakeit(108, 2, 4, 3,33, 6, 3, 9, 1, 0, 0)
         %mktmakeit(108, 2, 4, 3,31, 6, 2,18, 1, 0, 0)
         %mktmakeit(108, 2, 3, 3,41, 6, 1, 9, 1, 0, 0)
         %mktmakeit(108, 2, 3, 3,39,18, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2, 3, 3,34, 6, 3, 9, 1, 0, 0)
         %mktmakeit(108, 2, 3, 3,32, 6, 2,18, 1, 0, 0)
         %mktmakeit(108, 2, 3, 3,16, 6, 8, 0, 0, 0, 0)
         %mktmakeit(108, 2, 2,54, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2, 2, 3,42,18, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2, 2, 3,37, 6, 2, 9, 1, 0, 0)
         %mktmakeit(108, 2, 2, 3,35, 6, 1,18, 1, 0, 0)
         %mktmakeit(108, 2, 1, 3,35, 6, 3, 9, 1, 0, 0)
         %mktmakeit(108, 2, 1, 3,33, 6, 2,18, 1, 0, 0)
         %mktmakeit(108, 3,39, 6, 3, 9, 1, 0, 0, 0, 0)
         %mktmakeit(108, 3,37, 6, 2,18, 1, 0, 0, 0, 0)
         %mktmakeit(108, 3, 4, 6,11, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 4, 1,27, 1, 0, 0, 0, 0, 0, 0)

         /* Pang and Zhang constructions (pangzhang.sas)
         %mktmakeit(108, 2,76, 3, 4, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,68, 3, 5, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,66, 3, 4, 6, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,61, 3, 6, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,59, 3,15, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,59, 3, 5, 6, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,58, 3,11, 6, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,57, 3, 7, 6, 2, 0, 0, 0, 0)
         %mktmakeit(108, 2,49, 3, 4, 6, 3, 0, 0, 0, 0)
         %mktmakeit(108, 2,41, 3, 5, 6, 3, 0, 0, 0, 0)
         %mktmakeit(108, 2,39, 3, 4, 6, 4, 0, 0, 0, 0)
         %mktmakeit(108, 2,38, 3,18, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,37, 3,14, 6, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,36, 3,10, 6, 2, 0, 0, 0, 0)
         %mktmakeit(108, 2,34, 3, 6, 6, 3, 0, 0, 0, 0)
         %mktmakeit(108, 2,32, 3,15, 6, 3, 0, 0, 0, 0)
         %mktmakeit(108, 2,32, 3, 5, 6, 4, 0, 0, 0, 0)
         %mktmakeit(108, 2,31, 3,11, 6, 4, 0, 0, 0, 0)
         %mktmakeit(108, 2,30, 3, 7, 6, 5, 0, 0, 0, 0)
         %mktmakeit(108, 2,28, 3,37, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,25, 3,25, 6, 3, 0, 0, 0, 0)
         %mktmakeit(108, 2,22, 3,13, 6, 6, 0, 0, 0, 0)
         %mktmakeit(108, 2,20, 3,38, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,19, 3, 1, 6, 9, 0, 0, 0, 0)
         %mktmakeit(108, 2,18, 3,37, 6, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,17, 3,26, 6, 3, 0, 0, 0, 0)
         %mktmakeit(108, 2,15, 3,25, 6, 4, 0, 0, 0, 0)
         %mktmakeit(108, 2,14, 3,14, 6, 6, 0, 0, 0, 0)
         %mktmakeit(108, 2,12, 3,13, 6, 7, 0, 0, 0, 0)
         %mktmakeit(108, 2,11, 3,48, 0, 0, 0, 0, 0, 0)
         %mktmakeit(108, 2,11, 3,38, 6, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2,11, 3, 2, 6, 9, 0, 0, 0, 0)
         %mktmakeit(108, 2,10, 3,44, 6, 1, 0, 0, 0, 0)
         %mktmakeit(108, 2, 9, 3,40, 6, 2, 0, 0, 0, 0)
         %mktmakeit(108, 2, 9, 3, 1, 6,10, 0, 0, 0, 0)
         %mktmakeit(108, 2, 8, 3,36, 6, 3, 0, 0, 0, 0)
         %mktmakeit(108, 2, 8, 3,26, 6, 4, 0, 0, 0, 0)
         %mktmakeit(108, 2, 7, 3,32, 6, 4, 0, 0, 0, 0)
         %mktmakeit(108, 2, 7, 3,15, 6, 6, 0, 0, 0, 0)
         %mktmakeit(108, 2, 6, 3,28, 6, 5, 0, 0, 0, 0)
         %mktmakeit(108, 2, 5, 3,24, 6, 6, 0, 0, 0, 0)
         %mktmakeit(108, 2, 5, 3,14, 6, 7, 0, 0, 0, 0)
         %mktmakeit(108, 2, 4, 3,20, 6, 7, 0, 0, 0, 0)
         %mktmakeit(108, 2, 4, 3, 3, 6, 9, 0, 0, 0, 0)
         %mktmakeit(108, 2, 3, 3, 7, 6, 8, 0, 0, 0, 0)
         %mktmakeit(108, 2, 2, 3,12, 6, 9, 0, 0, 0, 0)
         %mktmakeit(108, 2, 2, 3, 2, 6,10, 0, 0, 0, 0)
         %mktmakeit(108, 2, 1, 3, 8, 6,10, 0, 0, 0, 0)
         */

         %mktmakeit(106, 2, 1,53, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(105, 3, 1,35, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(105, 5, 1,21, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(105, 7, 1,15, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(104, 2,100,4, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(104, 2,65, 4, 1,13, 1, 0, 0, 0, 0)
         %mktmakeit(104, 2,52,52, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(104, 2,51, 4, 1,26, 1, 0, 0, 0, 0)
         %mktmakeit(104, 8, 1,13, 1, 0, 1, 0, 0, 0, 0)
         %mktmakeit(102, 2, 1,51, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(102, 3, 1,34, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(102, 6, 1,17, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 2,99, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 2,51, 5, 3, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 2,40, 5, 4, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 2,34, 5, 3,10, 1, 0, 0, 0, 0)
         %mktmakeit(100, 2,29, 5, 5, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 2,22,25, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 2,18, 5, 9,10, 1, 0, 0, 0, 0)
         %mktmakeit(100, 2,16, 5, 3,10, 3, 0, 0, 0, 0)
         %mktmakeit(100, 2, 7, 5,10,10, 1, 0, 0, 0, 0)
         %mktmakeit(100, 2, 5, 5, 4,10, 3, 0, 0, 0, 0)
         %mktmakeit(100, 2, 4,10, 4, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 2, 2,50, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 4, 1,25, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 5,20,20, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(100, 5, 8,10, 3, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 99, 9, 1,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 99, 3,13,33, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 98, 2, 1,49, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 98, 7,14,14, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 96, 2,80,16, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 96, 2,73, 3, 1,16, 1, 0, 0, 0, 0)
         %mktmakeit( 96, 2,71, 6, 1,16, 1, 0, 0, 0, 0)
         %mktmakeit( 96, 2,48,48, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 96, 2,44, 4,11, 8, 1,12, 1, 0, 0)
         %mktmakeit( 96, 2,43, 4,15, 8, 1, 0, 0, 0, 0)
         %mktmakeit( 96, 2,43, 4,12, 6, 1, 8, 1, 0, 0)
         %mktmakeit( 96, 2,39, 3, 1, 4,14, 8, 1, 0, 0)
         %mktmakeit( 96, 2,26, 4,23, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 96, 2,19, 3, 1, 4,23, 0, 0, 0, 0)
         %mktmakeit( 96, 2,18, 4,22,12, 1, 0, 0, 0, 0)
         %mktmakeit( 96, 2,17, 4,23, 6, 1, 0, 0, 0, 0)
         %mktmakeit( 96, 2,12, 4,20,24, 1, 0, 0, 0, 0)
         %mktmakeit( 96, 3, 1,32, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 95, 5, 1,19, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 94, 2, 1,47, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 93, 3, 1,31, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 92, 2,91, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 92, 2,21,23, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 92, 2, 2,46, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 92, 4, 1,23, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 91, 7, 1,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 90, 2, 1,45, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 90, 3,30,30, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 90, 3,26, 6, 1,15, 1, 0, 0, 0, 0)
         %mktmakeit( 90, 5, 1,18, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 90, 9, 1,10, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 88, 2,84, 4, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 88, 2,56, 4, 1,11, 1, 0, 0, 0, 0)
         %mktmakeit( 88, 2,44,44, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 88, 2,43, 4, 1,22, 1, 0, 0, 0, 0)
         %mktmakeit( 88, 8, 1,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 87, 3, 1,29, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 86, 2, 1,43, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 85, 5, 1,17, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 84, 2,83, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 84, 2,33, 3, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 84, 2,28, 7, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 84, 2,27, 6, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 84, 2,20,21, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 84, 2,22, 6, 1, 7, 1, 0, 0, 0, 0)
         %mktmakeit( 84, 2,20, 3, 1,14, 1, 0, 0, 0, 0)
         %mktmakeit( 84, 2,14, 6, 1,14, 1, 0, 0, 0, 0)
         %mktmakeit( 84, 2, 2,42, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 84, 3, 1,28, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 84, 4, 1,21, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 84, 7, 1,12, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 82, 2, 1,41, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 81, 3,27,27, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 81, 9,10, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 80, 2,72, 8, 1, 0, 1, 0, 0, 0, 0)
         %mktmakeit( 80, 2,61, 5, 1, 8, 1, 0, 0, 0, 0)
         %mktmakeit( 80, 2,55, 8, 1,10, 1, 0, 0, 0, 0)
         %mktmakeit( 80, 2,51, 4, 3,20, 1, 0, 0, 0, 0)
         %mktmakeit( 80, 2,40,40, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 80, 4,10,20, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 80, 5, 1,16, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 78, 2, 1,39, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 78, 3, 1,26, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 78, 6, 1,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 76, 2,75, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 76, 2,19,19, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 76, 2, 2,38, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 76, 4, 1,19, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 75, 3, 1,25, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 75, 5, 8,15, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 77, 7, 1,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit( 74, 2, 1,37, 1, 0, 0, 0, 0, 0, 0)
         %if &maxlev lt &tempmaxlev %then %do;
            drop x%sysfunc(max(73, &maxlev + 1)) - x&tempmaxlev;
            %let tempmaxlev = 72;
            %end;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   *------expand 72 - 37.  half of 72 is done in next step------;
   %if &maxn ge 37 %then %do;
      data &outlev(drop=z: i &h); set &outlev; output;
         %if &needlin %then %do; h = lineage; %end;
         array x[&tempmaxlev]; array z[&tempmaxlev];
         do i = 1 to &tempmaxlev; z[i] = x[i]; end;

         %mktmakeit(72, 2,68, 4, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(72, 2,60, 3, 1, 4, 1, 0, 0, 0, 0)
         %mktmakeit(72, 2,53, 3, 2, 4, 1, 0, 0, 0, 0)
         %mktmakeit(72, 2,51, 3, 1, 4, 1, 6, 1, 0, 0)
         %mktmakeit(72, 2,49, 4, 1, 9, 1, 0, 0, 0, 0)
         %mktmakeit(72, 2,46, 3, 2, 4, 1, 6, 1, 0, 0)
         %mktmakeit(72, 2,46, 4, 1, 6, 2, 0, 0, 0, 0)
         %mktmakeit(72, 2,44, 3,12, 4, 1, 0, 0, 0, 0)
         %mktmakeit(72, 2,43, 3, 8, 4, 1, 6, 1, 0, 0)
         %mktmakeit(72, 2,43, 3, 1, 4, 1, 6, 2, 0, 0)
         %mktmakeit(72, 2,42, 3, 4, 4, 1, 6, 2, 0, 0)
         %mktmakeit(72, 2,41, 4, 1, 6, 3, 0, 0, 0, 0)
         %mktmakeit(72, 2,37, 3,13, 4, 1, 0, 0, 0, 0)
         %mktmakeit(72, 2,37, 3, 1, 4, 1, 6, 3, 0, 0)
         %mktmakeit(72, 2,36, 3, 9, 4, 1, 6, 1, 0, 0)
         %mktmakeit(72, 2,36, 3, 2, 4, 1, 6, 3, 0, 0)
         %mktmakeit(72, 2,36,36, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(72, 2,35, 3,12, 4, 1, 6, 1, 0, 0)
         %mktmakeit(72, 2,35, 3, 5, 4, 1, 6, 2, 0, 0)
         %mktmakeit(72, 2,35, 4, 1,18, 1, 0, 0, 0, 0)
         %mktmakeit(72, 2,34, 3, 8, 4, 1, 6, 2, 0, 0)
         %mktmakeit(72, 2,34, 3, 3, 4, 1, 6, 3, 0, 0)
         %mktmakeit(72, 2,31, 6, 4, 0, 0, 0, 0, 0, 0)
         %mktmakeit(72, 2,30, 3, 1, 6, 4, 0, 0, 0, 0)
         %mktmakeit(72, 2,28, 3, 2, 6, 4, 0, 0, 0, 0)
         %mktmakeit(72, 2,27, 3,11, 6, 1,12, 1, 0, 0)
         %mktmakeit(72, 2,27, 3, 6, 6, 4, 0, 0, 0, 0)
         %mktmakeit(72, 2,19, 3,20, 4, 1, 6, 1, 0, 0)
         %mktmakeit(72, 2,18, 3,16, 4, 1, 6, 2, 0, 0)
         %mktmakeit(72, 2,17, 3,12, 4, 1, 6, 3, 0, 0)
         %mktmakeit(72, 2,15, 3, 7, 4, 1, 6, 5, 0, 0)
         %mktmakeit(72, 2,14, 3, 3, 4, 1, 6, 6, 0, 0)
         %mktmakeit(72, 2,12, 3,21, 4, 1, 6, 1, 0, 0)
         %mktmakeit(72, 2,11, 3,20, 6, 1,12, 1, 0, 0)
         %mktmakeit(72, 2,11, 3,17, 4, 1, 6, 2, 0, 0)
         %mktmakeit(72, 2,10, 3,20, 4, 1, 6, 2, 0, 0)
         %mktmakeit(72, 2,10, 3,16, 6, 2,12, 1, 0, 0)
         %mktmakeit(72, 2,10, 3,13, 4, 1, 6, 3, 0, 0)
         %mktmakeit(72, 2, 9, 3,16, 4, 1, 6, 3, 0, 0)
         %mktmakeit(72, 2, 9, 3,12, 6, 3,12, 1, 0, 0)
         %mktmakeit(72, 2, 8, 3,12, 4, 1, 6, 4, 0, 0)
         %mktmakeit(72, 2, 8, 3, 8, 4, 1, 6, 5, 0, 0)
         %mktmakeit(72, 2, 7, 3, 7, 6, 5,12, 1, 0, 0)
         %mktmakeit(72, 2, 7, 3, 4, 4, 1, 6, 6, 0, 0)
         %mktmakeit(72, 2, 6, 3, 7, 4, 1, 6, 6, 0, 0)
         %mktmakeit(72, 2, 6, 3, 3, 6, 6,12, 1, 0, 0)
         %mktmakeit(72, 2, 5, 3, 3, 4, 1, 6, 7, 0, 0)
         %mktmakeit(72, 3,24,24, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(72, 8, 1, 9, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(70, 2, 1,35, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(70, 5, 1,14, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(70, 7, 1,10, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(69, 3, 1,23, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(68, 2,67, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(68, 2,18,17, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(68, 2, 2,34, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(68, 4, 1,17, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(66, 2, 1,33, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(66, 3, 1,22, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(66, 6, 1,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(65, 5, 1,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(64, 2,32,32, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(64, 2, 5, 4,17, 8, 1, 0, 0, 0, 0)
         %mktmakeit(64, 2, 5, 4,10, 8, 4, 0, 0, 0, 0)
         %mktmakeit(64, 4,16,16, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(64, 4,14, 8, 3, 0, 0, 0, 0, 0, 0)
         %mktmakeit(64, 4, 7, 8, 6, 0, 0, 0, 0, 0, 0)
         %mktmakeit(64, 8, 9, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(63, 7, 1, 9, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(63, 3,12,21, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(62, 2, 1,31, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 2,59, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 2,30, 3, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 2,23, 5, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 2,24, 6, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 2,21,10, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 2,17,15, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 2,15, 6, 1,10, 1, 0, 0, 0, 0)
         %mktmakeit(60, 2, 2,30, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 4, 1,15, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 3, 1,20, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(60, 5, 1,12, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(58, 2, 1,29, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(57, 3, 1,19, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(56, 2,52, 4, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(56, 2,28,28, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(56, 2,27, 4, 1,14, 1, 0, 0, 0, 0)
         %mktmakeit(56, 2,37, 4, 1, 7, 1, 0, 0, 0, 0)
         %mktmakeit(56, 7, 1, 8, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(55, 5, 1,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(54, 3,20, 6, 1, 9, 1, 0, 0, 0, 0)
         %mktmakeit(54, 3,18,18, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(54, 2, 1,27, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(52, 2,16,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(52, 2, 2,26, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(52, 2,51, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(52, 4, 1,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(51, 3, 1,17, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(50, 2, 1,25, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(50, 5,10,10, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(49, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(48, 2,24,24, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(48, 2,31, 6, 1, 8, 1, 0, 0, 0, 0)
         %mktmakeit(48, 2,33, 3, 1, 8, 1, 0, 0, 0, 0)
         %mktmakeit(48, 2,40, 8, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(48, 4,12,12, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(48, 3, 1,16, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(46, 2, 1,23, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(45, 5, 1, 9, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(45, 3, 9,15, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(44, 2,43, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(44, 2,15,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(44, 2, 2,22, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(44, 4, 1,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(42, 6, 1, 7, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(42, 2, 1,21, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(42, 3, 1,14, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(40, 2,19, 4, 1,10, 1, 0, 0, 0, 0)
         %mktmakeit(40, 2,20,20, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(40, 2,25, 4, 1, 5, 1, 0, 0, 0, 0)
         %mktmakeit(40, 2,36, 4, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(40, 5, 1, 8, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(39, 3, 1,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(38, 2, 1,19, 1, 0, 0, 0, 0, 0, 0)

         %if &maxlev lt &tempmaxlev %then %do;
            drop x%sysfunc(max(37, &maxlev + 1)) - x&tempmaxlev;
            %let tempmaxlev = 36;
            %end;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   *------expand 36 - 19.  half of 36 is done in next step------;
   %if &maxn ge 19 %then %do;
      data &outlev(drop=z: i &h); set &outlev; output;
         %if &needlin %then %do; h = lineage; %end;
         array x[&tempmaxlev]; array z[&tempmaxlev];
         do i = 1 to &tempmaxlev; z[i] = x[i]; end;
         %mktmakeit(36, 2, 1, 3, 3, 6, 3, 0, 0, 0, 0)
         %mktmakeit(36, 2, 2, 3, 5, 6, 2, 0, 0, 0, 0)
         %mktmakeit(36, 2, 2,18, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(36, 2, 3, 3, 2, 6, 3, 0, 0, 0, 0)
         %mktmakeit(36, 2, 3, 3, 9, 6, 1, 0, 0, 0, 0)
         %mktmakeit(36, 2, 4, 3, 1, 6, 3, 0, 0, 0, 0)
         %mktmakeit(36, 2, 8, 6, 3, 0, 0, 0, 0, 0, 0)
         %mktmakeit(36, 2, 9, 3, 4, 6, 2, 0, 0, 0, 0)
         %mktmakeit(36, 2,10, 3, 1, 6, 2, 0, 0, 0, 0)
         %mktmakeit(36, 2,10, 3, 8, 6, 1, 0, 0, 0, 0)
         %mktmakeit(36, 2,13, 3, 2, 6, 1, 0, 0, 0, 0)
         %mktmakeit(36, 2,13, 6, 2, 0, 0, 0, 0, 0, 0)
         %mktmakeit(36, 2,16, 9, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(36, 2,18, 3, 1, 6, 1, 0, 0, 0, 0)
         %mktmakeit(36, 2,20, 3, 2, 0, 0, 0, 0, 0, 0)
         %mktmakeit(36, 2,27, 3, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(36, 2,35, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(36, 3, 7, 6, 3, 0, 0, 0, 0, 0, 0)
         %mktmakeit(36, 3,12,12, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(36, 4, 1, 9, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(35, 5, 1, 7, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(34, 2, 1,17, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(33, 3, 1,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(32, 2,16,16, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(32, 4, 8, 8, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(30, 2, 1,15, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(30, 3, 1,10, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(30, 5, 1, 6, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(28, 2, 2,14, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(28, 2,12, 7, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(28, 2,27, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(28, 4, 1, 7, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(27, 3, 9, 9, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(26, 2, 1,13, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(25, 5, 6, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(24, 2,11, 4, 1, 6, 1, 0, 0, 0, 0)
         %mktmakeit(24, 2,12,12, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(24, 2,13, 3, 1, 4, 1, 0, 0, 0, 0)
         %mktmakeit(24, 2,20, 4, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(24, 3, 1, 8, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(22, 2, 1,11, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(21, 3, 1, 7, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(20, 2,19, 0, 0, 0, 0, 0, 0, 0, 0)
         %mktmakeit(20, 2, 2,10, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(20, 2, 8, 5, 1, 0, 0, 0, 0, 0, 0)
         %mktmakeit(20, 4, 1, 5, 1, 0, 0, 0, 0, 0, 0)

         %if &maxlev lt &tempmaxlev %then %do;
            drop x%sysfunc(max(19, &maxlev + 1)) - x&tempmaxlev;
            %let tempmaxlev = 18;
            %end;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   *------expand 18 - 14.  12 and 16 are done in next steps------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(18, 2, 1, 9, 1, 0, 0, 0, 0, 0, 0)
      %mktmakeit(18, 3, 6, 6, 1, 0, 0, 0, 0, 0, 0)
      %mktmakeit(15, 3, 1, 5, 1, 0, 0, 0, 0, 0, 0)
      %mktmakeit(14, 2, 1, 7, 1, 0, 0, 0, 0, 0, 0)

      %if &maxlev lt &tempmaxlev %then %do;
         drop x%sysfunc(max(17, &maxlev + 1)) - x&tempmaxlev;
         %let tempmaxlev = 16;
         %end;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------we can expand multiple 16s and 12s - 4s, need separate steps------;
   *------CAUTION: this relies on 12s not appearing with 16s          ------;

   *------expand 12 and 16------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(16, 4, 5, 0, 0, 0, 0, 0, 0, 0, 0)
      %mktmakeit(12, 3, 1, 4, 1, 0, 0, 0, 0, 0, 0)
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------expand 12 and 16------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(16, 2, 8, 8, 1, 0, 0, 0, 0, 0, 0)
      %mktmakeit(12, 2, 4, 3, 1, 0, 0, 0, 0, 0, 0)
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------expand 12------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(12, 2, 2, 6, 1, 0, 0, 0, 0, 0, 0)
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------expand 12------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(12, 2,11, 0, 0, 0, 0, 0, 0, 0, 0)
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------expand 10------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(10, 2, 1, 5, 1, 0, 0, 0, 0, 0, 0)

      %if &maxlev lt &tempmaxlev %then %do;
         drop x%sysfunc(max(10, &maxlev + 1)) - x&tempmaxlev;
         %let tempmaxlev = 9;
         %end;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------expand 9------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(9, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0)
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------expand 8------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(8, 2, 4, 4, 1, 0, 0, 0, 0, 0, 0)
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------expand 6------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(6, 2, 1, 3, 1, 0, 0, 0, 0, 0, 0)
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------get rid of duplicates so the next step consumes fewer resources------;
   %if &nodups %then %do;
      proc sort data=&outlev nodupkeys; by n x:; run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   *------expand 4------;
   data &outlev(drop=z: i &h); set &outlev; output;
      %if &needlin %then %do; h = lineage; %end;
      array x[&tempmaxlev]; array z[&tempmaxlev];
      do i = 1 to &tempmaxlev; z[i] = x[i]; end;
      %mktmakeit(4, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0)
      %if &maxlev lt &tempmaxlev %then %do;
         drop x%eval(&maxlev+1) - x&tempmaxlev;
         %end;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %if &flow %then %put ** MktOrth: catalog expanded.;

   *======= Filter Inferior Designs, Get Rid of Duplicates First ======;
   %if &nodups %then %do;

      %if &flow %then %put ** MktOrth: filtering.;

      %let makelist = 0; /* set to 1 to get new list */

      proc sort data=&outlev nodupkeys; by n &bylist; run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      *------most inferior designs can be eliminated by just
             checking the next few designs.  Throw those out right away.------;
      data &outlev;
         drop __: i j f;
         set &outlev nobs=nobs;
         array   x[&maxlev];
         array __x[&maxlev];
         f = 1;
         do i = 1 to 2 while(f);
            p = _n_ + i;
            if p le nobs then do;
               set &outlev(keep=n x1-x&maxlev
                           rename=( n = __n
                           %do i = 1 %to &maxlev; x&i = __x&i %end; ))
                           point=p;
               if n eq __n then do;
                  f = 0;
                  do j = 1 to &maxlev until(f);
                     if __x[j] < x[j] then f = 1;
                     end;
                  end;
               end;
            end;

         /* uncomment here and below to check for redundant parents */
         * if not f and not index(substr(lineage, 12), ':')
            then put 'Redundant: ' n= lineage;

         if f then output &outlev;

         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      data &outlev %if &makelist %then freqs(keep=l n);;
         drop count __: i j f;
         set &outlev nobs=nobs;
         array   x[&maxlev];
         array __x[&maxlev];
         f = 1;
         %if &makelist %then %do; count = 1000; %end;
         do i = 1 to count while(f);
            p = _n_ + i;
            if p le nobs then do;
               set &outlev(keep=n x1-x&maxlev
                           rename=( n = __n
                           %do i = 1 %to &maxlev; x&i = __x&i %end; ))
                           point=p;
               if n eq __n then do;
                  f = 0;
                  do j = 1 to &maxlev until(f);
                     if __x[j] < x[j] then f = 1;
                     end;
                  %if &makelist %then %do;
                     if f eq 0 then do; l = i + 1; output freqs; end;
                     %end;
                  end;
               end;
            end;

         /* uncomment here and above to check for redundant parents */
         * if not f and not index(substr(lineage, 12), ':')
            then put 'Redundant: ' n= lineage;

         if f then output &outlev;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      %if &makelist %then %do;
         proc means data=freqs noprint;
            var l;
            class n;
            output out=freq2(where=(max gt 0)) max(l) = max;
            run;

         proc print; where n(n); var n max; run;
         %end;
      %end;
   %end;

*========================== Final Processing ==========================;

%if &flow %then %put ** MktOrth: finishing.;

*------build design string, references------;
data &outlev(drop=i l sum p powerof lognruns ff rename=(s=Design));
   length n 8 s $ 60 Reference $ 20;
   set &outlev;

   array x[&maxlev];
   sum = sum(of x1-x&maxlev);
   if sum gt 1;
   s = ' ';
   if x2 then s = '2 **' || put(x2, 3.);
   %if &maxlev gt 2
      %then %do; if x3 then substr(s, 10) = '3 **' || put(x3, 3.); %end;

   do i = 4 to &maxlev;
      l = max(16, length(s)) + 2;
      if x[i] then substr(s, l) = put(i, 3.) || ' **' || put(x[i], 3.);
      end;

   p = 0;
   do powerof = &primes;
      lognruns = log(n) / log(powerof);
      if abs(lognruns - round(lognruns)) lt 1e-8 then do; p = 1; leave; end;
      end;
   if not p then powerof = 0;

   reference = 'Orthogonal Array';
   if n - 1 eq x2               then reference = 'Hadamard';
   else if powerof in (&primes) then reference = 'Fractional-Factorial';
   else do;
      ff = 0;
      do i = 1 to &maxlev;
         if x[i] then ff = ff + log10(i) * x[i];
         if ff gt log10(n + 1) then leave;
         end;
      if ff lt 32 then do;
         if abs(10 ** ff - n) lt 0.5 then reference = 'Full-Factorial';
         end;
      end;

   %if &needlin %then %do;
      if index(lineage, ':') eq 0 or (scan(lineage, 3, ' ') eq '1' and
         scan(lineage, 4, ' ') eq ':' and scan(lineage, 3, ':') eq ' ')
         then lineage = trim(lineage) || ' (parent)';
      %if &parents %then %do; else delete; %end;
      %end;
   %if &needlin and not &lineage %then %do; drop lineage; %end;

   %if &adx %then %do;
      array ls[5];
      array ns[5];

      do j = 1 to 5; ls[j] = 0; ns[j] = 0; end;

      i = 0;
      do j = 2 to &maxlev;
         if x[j] ne 0 then do;
            i + 1;
            ns[i] = x[j];
            ls[i] = j;
            end;
         end;

      drop j;
      %end;

   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%if &adx %then %do;
   proc means data=&outlev noprint;
      output out=__tmpolist(where=(_stat_ eq 'MAX'));
      var ns: ls:;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data &outlev(drop=i j k ls: ns: _:);
      set &outlev(drop=ls: ns:);
      array ls[5];
      array ns[5];
      array x[&maxlev];
      if _n_ eq 1 then do;
         set __tmpolist;
         do i = 1 to 5;
            if ls[i] then ls[i] = ceil(log10(ls[i] + 1));
            if ns[i] then ns[i] = ceil(log10(ns[i] + 1));
            end;
         end;
      j = 1; k = 1;
      nf = sum(of x:);
      design = ' ';
      do i = 1 to &maxlev;
         if x[i] then do;
            substr(design, j, ls[k]) = putn(i, ' ', ls[k]);
            j + ls[k];
            substr(design, j, 1) = '^'; j + 1;
            substr(design, j, ns[k]) = left(putn(x[i], ' ', ns[k]));
            j + ns[k] + 1;
            k + 1;
            end;
         end;

      if _error_ then call symput('abort', '1');
      run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %end;

*------generate all smaller designs too------;
%if %nrbquote(&outall) ne %then %do;
   data &outall;
      set &outlev;
      array x[&maxlev];
      array p[10];
      array q[10];
      array i[10];
      r = 0;
      do k = 1 to 10; p[k] = 0; q[k] = 0; end;
      do k = 1 to &maxlev;
         if x[k] then do;
            r + 1;
            p[r] = k;
            q[r] = x[k];
            end;
         end;
      %do i = 1 %to 10; do i&i = 0 to q&i; %end;
         prod = 1;
         do k = 1 to r;
            x[p[k]] = i[k];
            if prod le n and i[k] then prod = prod * i[k];
            end;
         if prod le n and sum(of i1-i4) > 1 then output;
         %do i = 1 %to 10; end; %end;
      keep x1-x&maxlev n reference design;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc sort nodupkeys; by n &bylist; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

proc datasets nolist;
   delete __tmpolist;
   run; quit;

options &savenote no&timeopt;

*------output the data sets------;
%if %nrbquote(&outall) ne %then %do;
   proc sort data=&outall;
      by n %do i = 1 %to %sysfunc(min(50, &maxlev)); descending x&i %end;
         descending design;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%if not &mktex %then %do;
   proc sort data=&outlev;
      by n %do i = 1 %to %sysfunc(min(50, &maxlev)); descending x&i %end;
         descending design;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%if not (&mktex or &mktruns) %then %do;
   data &outcat; set &outlev(drop=x:); run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%endit: options &saveopts;
quit;

%if &flow %then %put ** MktOrth: done.;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %if not &mktex %then %put ERROR: The MKTORTH macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTORTH macro used &time seconds.;
   %end;

%mendit: %mend;

  /*

%let maxlev = 144;
%mktorth(options=lineage, maxlev=&maxlev)

%let maxlev = 144;
options nogeneric;
data test(keep=n c depth r line);
   set mktdeslev;
   by n;
   if first.n then c = 0;
   length line $ 80;
   array x[&maxlev];
   depth = 0;
   if index(lineage, '(parent)') eq 0 then do i = 1 to 10;
      depth = depth + (scan(lineage, i, '>') ne ' ');
      end;
   link makeit;
   nflag = (n lt 144 and n ne 128) and uniform(7) < 0.5;
   if nflag or uniform(7) le 0.05 then do;
      do i = 1 to &maxlev; x[i] = floor(x[i] * 0.6); end;
      c + 10000;
      link makeit;
      c + -10000;
      end;
   return;
makeit:
   c + 1;
   line = '%mktex(';
   do i = 1 to &maxlev;
      if x[i] then line = compbl(trim(line) || ' ' || put(i, 4.) ||
                          ' ** ' || put(x[i], 4.));
      end;
   line = compbl(line || ', n=' || trim(put(n, 5. -L)) || ')');
   r = uniform(7);
   if index(line, '**') then output;
   return;
   run;

proc sort nodupkeys; by n line; run;
proc sort; by depth c n r; run;

data test2(keep=n line depth);
   set test;
   array ns[1000];
   ns[n] + 1;
   if depth le 2 or (depth gt 6 and n ne 256)
      or ns[n] le 100 or (n lt 144 and n ne 128) or
      ((not (n in (432, 448, 512))) and uniform(7) le 0.05) or
      uniform(7) le 0.01;
   run;

proc freq; tables n; run;

data _null_; set;
   file = compress('mktexb' || put(ceil(_n_ / 100), z3.) || '.sas');
   file a filevar=file;
   if mod(_n_, 100) = 1 then do;
      put ' /'
          '***************************************************************'
        / '  *                     SAS/STAT TEST LIBRARY'
        / '  *'
        / '  *       NAME:' file
        / '  *  COMPONENT: STAT.AUTOCALL'
        / '  *  REFERENCE: Marketing Research, Discrete Choice, Conjoint,'
        / '  *             Experimental Design, and Graphics Reports TS-722A'
        / '  *             through TS-722L, Warren F. Kuhfeld'
        / '  *   COMMENTS:'
        / '  *    DEFECTS:'
        / '  *    HISTORY: 05May2009 saswfk created'
        / '  *    PURPOSE: Use autocall macro %mktex to generate orthogonal'
        / '  *             designs that are 100% efficient, depth=' depth
        / '  ****************************************************************'
          '/' /;
      put '%let mktexopts = target=1e-6, options=nofinal nosort;';
      put '%let mktopts = silent title2;';
      end;
   put line;
   run;

*---big design test------;
%mktorth(options=mktex 512, maxlev=32, range=n=512)

data _null_;
   set mktdeslev;
   file 'mktexb512.sas';
   if _n_ = 1 then do;
      put ' /'
          '***************************************************************'
        / '  *                     SAS/STAT TEST LIBRARY'
        / '  *'
        / '  *       NAME: mktexb512.sas'
        / '  *  COMPONENT: STAT.AUTOCALL'
        / '  *  REFERENCE: Marketing Research, Discrete Choice, Conjoint,'
        / '  *             Experimental Design, and Graphics Reports TS-722A'
        / '  *             through TS-722L, Warren F. Kuhfeld'
        / '  *   COMMENTS:'
        / '  *    DEFECTS:'
        / '  *    HISTORY: 05May2009 saswfk created'
        / '  *    PURPOSE: Use autocall macro %mktex to generate orthogonal'
        / '  *             designs that are 100% efficient'
        / '  ****************************************************************'
          '/' /;
      put '%let mktexopts = maxdesigns=1, '
          'options=nofinal nosort, target=0.001;';
      put '%let mktopts = silent title2;';
      put '%mktorth(options=lineage 512 mktex, maxlev=32, range=n=512)' /;
      put '%mktex(16 **  32 32 ** 1, n=512, cat=mktdeslev)';
      put '%mktex( 8 **  64 32 ** 1 2 ** 32, n=512, cat=mktdeslev)';
      put '%mktex( 4 ** 160 32 ** 1, n=512, cat=mktdeslev)';
      end;
   if uniform(7) < 0.0012 then do;
      design = compbl(design);
      put '%mktex(' design +(-1) ', n=512, cat=mktdeslev)';
      end;
   run;


options generic;







* Begin redundant parent check
  Another redundant parent check is buried in the macro itself.
  This one won't get them all;
%mktorth(options=lineage parent, maxlev=144, outcat=parent)
%mktorth(options=lineage dups,   maxlev=144, outcat=all)

proc sort data=all;    by n design; run;
proc sort data=parent; by n design; run;

data x; merge all parent(in=p); by n design; if p; run;
data y; set x; by n design; if not (first.design and last.design); run;

proc print; var n design; run;
* End redundant parent check;



* Uncomment lines above and run this for second redundant parent check.
* Note that false positives involving:
  14 **, 15 **, 18 **, 20 **, 21 **, 22 **
  are expected and can be igored.;
%mktorth(options=lineage, maxlev=144, outcat=all)





* Count designs that can be made from deleting factors.;
* Ignore the fact that discarding leads to duplicate designs.;
data lev;
   set mktdeslev end=eof;
   array x[144];
   c = 1; do i = 1 to 144; c = c * (1 + x[i]); end;
   s + (c - 1);
   if eof then put s= comma15.;
   run;




* List of all designs;
%mktorth(maxlev=144)

data all;
   set mktdescat;
   by n;
   file 'new' termstr=nl;
   if _n_ = 1 then do;
      put '%let mktexopts = maxdesigns=1, '
          'options=nofinal nosort, target=0.001;';
      put '%let mktopts = silent title2;';
      end;
   put '%mktex(' design +(-1) ', n=' n +(-1) ')';
   run;

proc freq; tables n; run;

  */
