 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktPPro                                             */
 /*   TITLE: Create Optimal Choice Designs using Blocks and OA's */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs                         */
 /*   PROCS: IML                                                 */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: The following SAS products are required to run      */
 /*          this macro: BASE, IML.                              */
 /*                                                              */
 /*          This macro creates global macro variables: _MktVer  */
 /****************************************************************/

 /*---------------------------------------------------------------------

The MktPPro macro is used to make a partial-profile choice designs from
a balanced incomplete blocks design (BIBD) and an orthogonal array.  A
block design is balanced when each attribute pair appears equally often.
For partial profiles, the BIBD is used to determine which attributes to
vary in each choice set.  This example makes an optimal partial-profile
choice design with attrs=16 binary attributes, setsize=4 of which vary
in each set, and it does so in b x k = 80 choice sets.

   %mktex(4 2 ** 4, n=8, seed=306)

   proc sort data=randomized out=randes(drop=x1);
      by x2 x1;
      run;

   proc print noobs data=randes; run;

   %mktbibd(b=20, nattrs=16, setsize=4, seed=104)

   %mktppro(design=randes, ibd=bibd, print=f p)

   %choiceff(data=chdes, init=chdes, initvars=x1-x16,
             model=class(x1-x16 / sta), nsets=80, nalts=2,
             beta=zero, options=relative)

This design has a relative D-efficiency of 25.00, which corresponds to
the proportion 4 / 16, which is the proportion of attributes that vary.
In a partial profile design where 25% of the attributes vary, the
optimal relative D-efficiency is 25, showing that this design is
optimal for a partial profile design and 25% efficient relative to an
optimal generic design with all attributes varying.

Two input designs are input to the MktPPro macro.  The first is an
orthogonal array.  The orthogonal array must be a p^q subset of an
array p^q k^1 in p x k runs with q <= k.  Often, q = k.  Examples:

2^4 in  8 runs, selected from 2^4 4^1 in  8 runs
3^3 in  9 runs, selected from 3^3 3^1 in  9 runs
4^4 in 16 runs, selected from 4^4 4^1 in 16 runs
3^6 in 18 runs, selected from 3^6 6^1 in 18 runs

You can use the MktOrth or MktRuns macros to see if the orthogonal
array of interest exists.  These steps list arrays that work.

   %mktorth(options=parent, maxlev=144)

   data x(keep=n design);
      set mktdeslev;
      array x[144];
      c = 0; one = 0; q = 0;
      do i = 1 to 144;
         c + (x[i] > 0); * how many differing numbers of levels;
         if x[i] > 1 then do; p = i; q = x[i]; end; * p^q ;
         if x[i] = 1 then do; one + 1; k = i;  end; * k^1 ;
         end;
      if c = 1 then do; c = 2; one = 1; k = p; q = p - 1; end;
      if c = 2 and one = 1 and q > 2 and k * p = n;
      design = compbl(left(design));
      run;

   proc print; run;

The rows of the orthogonal array must be sorted into the right order.
Each block of k rows in the orthogonal array is called a difference
scheme, and blocks 2 through p are obtained from the preceding block by
adding 1 (in the appropriate field).  You will not get the right
results if you stick in any other kind of orthogonal array.

The easiest way to create the orthogonal array with the rows in the
right order is to first request one of the p-level factors, then request
the k-level factor, then request the remaining (q-1) p-level factors.
Then *after* the array is created, discard x2, the k-level factor.  See
the example above.  Note that you cannot drop the second factor in the
MktEx step by specifying OUT=DESIGN(DROP=X2), because MktEx will drop
the variable and then sort, and your rows will be in the wrong order.
Here is the orthogonal array.


                              x1    x3    x4    x5

                               1     1     1     1
                               1     2     1     2
                               1     1     2     2
                               1     2     2     1
                               2     2     2     2
                               2     1     2     1
                               2     2     1     1
                               2     1     1     2

In this array, if you take the first block of k=4 observations and
change 1 to 2 and 2 to 1, you get the second block.  This shows for
two-level factors that the design is in the right order.  More
generally, when p is prime, the i + 1 block can be made from the ith
block by adding 1 in a field with p elements.  For example with p = 3
and elements 1, 2, and 3, then 1 + 1 = 2, 2 + 1 = 3, 3 + 1 = 1.  You do
not really need to worry about any of this though, because when you
create the orthogonal array using the method outlined above, your array
will always be in the right order.

The second input data set contains the block design.  In the example
above, the design is the following BIBD:

                       Balanced Incomplete Blocks Design

                              x1    x2    x3    x4

                              15    16     5     6
                               4     1    16    12
                               5    13     4     7
                               4     6     9     8
                              14     8    13    16
                              12    14     5    10
                               3     9    12    13
                               7    10     1     8
                              16     7     3    11
                               8    12    11    15
                              10     4    15     3
                               1     3     6    14
                               2    11    14     4
                               7    15    14     9
                               6    12     7     2
                              11     5     9     1
                              13    15     2     1
                               9    10    16     2
                              11     6    10    13
                               3     2     8     5

This says that in the first block of k=4 choice sets, attributes 15,
16, 5, and 4 will vary; in the second block of 4 choice sets,
attributes 4, 1, 16, and 12 will vary; and so on.  Alternative i = 1,
... p, in each block of k choice sets will be made from a block of k
runs in the orthogonal array.  For example, alternative 1 of the first
k choice sets will be made from the first k runs in the orthogonal
array, and alternative 2 will be made from the next k runs in the
orthogonal array, and so on.  Each attribute should appear the same
number of times in the design.  The t=16 by t=16 attribute frequencies
show how often each attribute appears with every other attribute in our
BIBD.

                             Attribute Frequencies

                   1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16

                1  5  1  1  1  1  1  1  1  1  1  1  1  1  1  1  1
                2     5  1  1  1  1  1  1  1  1  1  1  1  1  1  1
                3        5  1  1  1  1  1  1  1  1  1  1  1  1  1
                4           5  1  1  1  1  1  1  1  1  1  1  1  1
                5              5  1  1  1  1  1  1  1  1  1  1  1
                6                 5  1  1  1  1  1  1  1  1  1  1
                7                    5  1  1  1  1  1  1  1  1  1
                8                       5  1  1  1  1  1  1  1  1
                9                          5  1  1  1  1  1  1  1
               10                             5  1  1  1  1  1  1
               11                                5  1  1  1  1  1
               12                                   5  1  1  1  1
               13                                      5  1  1  1
               14                                         5  1  1
               15                                            5  1
               16                                               5

Each attribute appears 5 times and is paired with every other attribute
exactly once.  The fact that values above the diagonal are constant, in
this case all ones, makes this design a BIBD.  We always want the diagonal
elements to all be constant.  The number of choice sets is k times the
number of blocks (rows in the design), in this case, 4 x 20 = 80.  The
number of attributes is the maximum value in the BIBD (in this case,
t=16).  The number of attributes that vary at any one time is q <= k.
All factors have p levels, and all choice sets have p alternatives.

You can make an optimal partial-profile choice design even without a
BIBD.  This example makes a design with 6 three-level factors in 60
choice sets with three alternatives each.  The MktBSize macro is used to
suggest the size of an unbalanced block design.  Each attribute will
appear three times in the design and each pair will occur in the range 0
to 2 times (average 0.79).  The design is optimal for this
specification.

   %mktex(3 6 3 ** 5, n=18)

   data design; set design; drop x2; run;

   proc print noobs; run;

   %mktbsize(nattrs=20, setsize=6, options=ibd)

   %mktbibd(b=10, nattrs=20, setsize=6, seed=104, out=b)

   %mktppro(ibd=b, print=f p)

   %choiceff(data=chdes, model=class(x1-x20 / sta), nsets=60, nalts=3,
             options=relative, beta=zero, init=chdes, initvars=x1-x20)

This design has a relative D-efficiency of 30.00, which corresponds to
the proportion 6 / 20, which is the proportion of attributes that vary.
In a partial profile design where 30% of the attributes vary, the
optimal relative D-efficiency is 30, showing that this design is
optimal for a partial profile design and 30% efficient relative to an
optimal generic design with all attributes varying.

You could also specify B=20 in MktBIBD and NSETS=120 in ChoicEff to make
a larger design.  Any integer multiple of B=10 will work.

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

%macro mktppro(   /*--------------------------------------------------*/
                  /*                                                  */
help,             /* nonblank, list options                           */
                  /*                                                  */
                  /* You must specify either IBD= or X= but not both. */
                  /*                                                  */
design=design,    /* specifies the orthogonal array design from       */
                  /* MktEx.                                           */
                  /*                                                  */
ibd=,             /* specifies an (possibly balanced) incomplete      */
                  /* block design.                                    */
                  /*                                                  */
x=,               /* incidence matrix for the block design which is a */
                  /* binary coding of the design.                     */
                  /*                                                  */
out=chdes,        /* specifies the output choice design.              */
                  /*                                                  */
print=f,          /* printed output:                                  */
                  /* f - frequencies matrix                           */
                  /* i - incomplete blocks design                     */
                  /* p - partial profile design                       */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _MktVer;
%local  abort time savenote setnote xflag sets alts savedsnf k
        printb printf printp localver saveopts timeopt;

%let time     = %sysfunc(datetime());
%let timeopt  = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote = %sysfunc(getoption(notes));
%let saveopts = &savenote missing="%sysfunc(getoption(missing))"
                %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                %sysfunc(getoption(serror));
%let abort    = 0;
%let setnote  = nonotes;
%let localver = 25Jul2010;
%let _mktver  = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES) %then %let setnote = notes;
   %if %index(&k, VERSION) %then %put MktPPro macro version &localver.;
   %end;
options serror;

options &setnote generic missing=' ';

%if %nrbquote(&help) ne %then %do; %mkthelp(mktppro) %goto endit; %end;

*------check and process options------;
data _null_;
   abort = 0;
   length list $ 512;

   *------either X= or IBD= but not both------;
   x = symget('x');
   b = symget('ibd');
   if ((x eq ' ') + (b eq ' ')) ne 1 then do;
      put "ERROR: X=&x, IBD=&ibd is invalid.";
      put 'ERROR: Specify X= or IBD= but not both.';
      abort = 1;
      end;
   call symput('xflag', put(x ne ' ', 1.));

   *------need an output data set------;
   if symget('out') eq ' ' then call symput('out', 'chdes');

   *------check OPTIONS=------;
   list = lowcase(symget('print'));
   i = index(list, 'f'); if i then substr(list, i, 1) = ' ';
   call symput('printf', put(i gt 0, 1.));
   i = index(list, 'p'); if i then substr(list, i, 1) = ' ';
   call symput('printp', put(i gt 0, 1.));
   i = index(list, 'i'); if i then substr(list, i, 1) = ' ';
   call symput('printb', put(i gt 0, 1.));
   if list ne ' ' then do;
      put 'ERROR: PRINT=' list 'is invalid.';
      abort = 1;
      end;

   if _error_ or abort then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;

filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
data __tmppdes; set &design; if _error_ then call symput('abort', '1'); run;
%if &syserr > 4 %then %let abort = 1;

%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&ibd) ne %then %do;
   data _null_; set &ibd; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&x) ne %then %do;
   data _null_; set &x; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )design=&design;
   %put %str(       )ibd=&ibd;
   %put %str(       )x=&x;
   %put %str(       )out=&out;
   %end;

%if &abort %then %goto endit;

*------process design------;
proc iml;
   file print;
   use __tmppdes; read all into d; close __tmpdes;
   %if &xflag %then %do;
      use &x; read all into x; close &x;
      b = nrow(x);  k = ncol(d);
      bibd = j(b, k, 0);
      do i = 1 to b; bibd[i,] = loc(x[i,] = 2); end;
      free x;
      %end;
   %else %do; use &ibd; read all into bibd; close &ibd; %end;
   t = max(bibd);            /* number of treatments                   */
   b = nrow(bibd);           /* number of blocks                       */
   k = ncol(bibd);           /* block size, number of cols in designs  */
   ds = nrow(d) / max(d);    /* number of rows in difference scheme    */
   nlevs = nrow(d) / ds;     /* number of levels in orthogonal array   */
   sets = b # ds;            /* number of choice sets                  */
   f = j(t, t, 0);           /* treatment frequency matrix             */

   do;
   if nlevs ^= max(d) then do;
      file log;
      put 'ERROR: Corrupt orthogonal array.';
      call symput('abort', '1');
      goto endit;
      end;

   if k ^= ncol(d) then do;
      file log;
      put 'ERROR: The orthogogonal array and IBD must '
          'have the same number of columns.';
      call symput('abort', '1');
      goto endit;
      end;

   if any(bibd < 1) | any(bibd ^= round(bibd)) |
      any(d < 1) | any(d ^= round(d)) then do;
      put 'ERROR: The orthogogonal array and IBD must '
          'consist of positive integers.';
      call symput('abort', '1');
      goto endit;
      end;

   %if &printb %then %do;
      w = 1 + (t > 9) + (t > 99);
      if      w = 1 then print bibd[format=1. label='Blocks Design'];
      else if w = 2 then print bibd[format=2. label='Blocks Design'];
      else               print bibd[          label='Blocks Design'];
      %end;

   do i = 1 to b;
      do j = 1 to k;
         do l = j to k;
            a = min(bibd[i,j], bibd[i,l]);
            c = max(bibd[i,j], bibd[i,l]);
            f[a,c] = f[a,c] + 1;
            end;
         end;
      end;
   w = 0;
   do i = 1 to t;
      w = max(w,f[i,i]);
      do j = 1 to i - 1; w = max(w,f[j,i]); f[i,j] = .; end;
      end;
   %if &printf %then %do;
      w = ceil(log10(max(w,t) + 1)); l = right(char(1, w, 0) : char(t, w, 0));
      if w = 1 then print
         'Attribute Frequencies',, f[format=1. label='' rowname=l colname=l];
      else if w = 2 then print
         'Attribute Frequencies',, f[format=2. label='' rowname=l colname=l];
      else print 'Attribute Frequencies',, f[label='' rowname=l colname=l];
      %end;
   if ^all(f[1,1] = vecdiag(f)) then do;
      file log;
      put 'WARNING: Unequal attribute frequencies.';
      file print;
      end;
   free f;

   x = j(sets, nlevs # t, 1);  /* partial-profile design               */
   do i = 1 to b;
      j = ((i-1) # ds : i # ds - 1) + 1;
      do l = 0 to nlevs - 1;
        x[j, bibd[i,] + l # t] = d[(1:ds) + l # ds,];
        end;
      end;
   x = (1:sets)` @ j(nlevs, 1, 1) || shape(x, nlevs # sets);
   vname = 'Set' || ('x1' : rowcatc('x' || char(t)));
   create &out from x[colname=vname]; append from x;
   call symput('sets', compress(char(sets, 12, 0)));
   call symput('alts', compress(char(nlevs, 12, 0)));
   endit: end;
   quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%if &printp %then %do;
   proc print data=&out; id set; by set; var x:; run;
   %end;

proc datasets nolist;
   delete __tmppdes;
   run; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

options &savenote no&timeopt;

data &out; set &out; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let alts = and &alts alternatives.;
%put NOTE: The partial-profile design has &sets choice sets &alts.;

%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTPPRO macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTPPRO macro used &time seconds.;
   %end;

%mendit: %mend;
