 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktRoll                                             */
 /*   TITLE: Roll Out a Linear Design Into a Choice Design       */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs, choice modeling        */
 /*   PROCS:                                                     */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: The following SAS products are required to run      */
 /*          this macro: BASE.                                   */
 /*                                                              */
 /*          This macro creates global macro variables: _MktVer  */
 /****************************************************************/

 /*---------------------------------------------------------------------

This macro is used for manipulating the experimental design for choice
experiments.  It takes as input a SAS data set containing an
experimental design with one row per choice set, for example a design
created by the MktEx macro.  This data set is specified in the DESIGN=
option.  This data set has one variable for each attribute of each
alternative in the choice experiment.

The output from this macro is an OUT= SAS data set containing the
experimental design with one row per alternative per choice set.
There is one column for each different attribute.  For example,
in a simple branded study, DESIGN= may contain the variables X1-X5
which contain the prices of each of five alternative brands.  The
output data set would have one factor, Price, that contains the
price of each of the five alternatives.  In addition, it would
have the number (or optionally the name) of each alternative.

The rules for determining the mapping between factors in the DESIGN=
data set and the OUT= data set are contained in the KEY= data set.

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

------------------------------------------------------------------------

For example, assume that the DESIGN= data set contains the variables
X1-X5, which contain the prices of each of five alternative brands: Brand
A, B, C, D, and E.  Here is how you would create the KEY= data set.  The
choice design has two factors, Brand and Price.  Brand A price is made
from X1, Brand B price is made from X2, ..., and Brand E price is made
from X5.,

   data key;
      input (Brand Price) ($);
      datalines;
   A x1
   B x2
   C x3
   D x4
   E x5
   ;

This data set has two variables, Brand contain the brand names and Price
contains the names of the factors that are used to make the price
effects for each of the alternatives.  The OUT= data set will contain
the variables with the same names as the variables in the KEY= data set.

Here is how you can create the design with one row per choice set:

   %mktex(3 ** 5, n=12)

Here is how you can create the design with one row per
alternative per choice set:

   %mktroll(design=randomized, key=key, out=sasuser.design, alt=brand)

For example if the data set RANDOMIZED contains the row:


   Obs    x1    x2    x3    x4    x5

     9     1     3     2     3     3

Then the data set SASUSER.DESIGN contains the rows:

   Obs    Set    Brand    Price

    41      9      A        1
    42      9      B        3
    43      9      C        2
    44      9      D        3
    45      9      E        3

The price for Brand A is made from x1=1, ..., and the price for Brand E
is made from x5=3.

------------------------------------------------------------------------

Now assume that there are three alternatives, each composed of four
factors: Brand Price Size Color Shape.  In addition, there is a constant
alternative.  First, the MktEx macro is used to create a design with 12
factors, one for each attribute of each alternative.

   %mktex(2 ** 12, n=16)

Then the KEY= data set is created.  It will contain the names of all of
the factors, in the linear design.  You can use the MktKey macro to make
the names. You can then copy and paste into the key data set.  For example:

   %mktkey(x1-x12)

displays:

   x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12

Alternatively:

   %mktkey(3 4)

produces a data set called KEY with 3 rows and 4 columns which is as
follows:

                            x1    x2     x3     x4

                            x1    x2     x3     x4
                            x5    x6     x7     x8
                            x9    x10    x11    x12

Either of these can be used as an aid in constructing the key data set
that you need. (In some cases, you can use this KEY data set directly if
it matches what you need.)  The KEY= data set below shows that there are
three brands, A, B, and C, plus None.

   data key;
      input (Brand Price Size Color Shape) ($);
      datalines;
   A     x1  x2  x3  x4
   B     x5  x6  x7  x8
   C     x9 x10 x11 x12
   None   .   .   .   .
   ;

Brand A is created from Brand = 'A',
Price = x1, Size = x2, Color = x3, Shape = x4.

Brand B is created from Brand = 'B',
Price = x5, Size = x6, Color = x7, Shape = x8.

Brand C is created from Brand = 'C',
Price = x9, Size = x10, Color = x11, Shape = x12.

The constant alternative is created from Brand = 'None' and none of the
attributes.  The "." notation is used to indicate missing values in
input data sets.  The actual values in the SAS data set will be blank
(character missing).

Here is how you create the design with one row per alternative per
choice set:

   %mktroll(key=key, design=randomized, out=sasuser.design, alt=brand)

For example if the data set DESIGN contains the row:

Obs  x1  x2  x3  x4  x5  x6  x7  x8  x9  x10  x11  x12

  8   2   1   1   1   1   1   2   2   2   2    1    2

Then the data set SASUSER.DESIGN contains the rows:

    Obs    Set    Brand    Price    Size    Color    Shape

     29     8     A          2        1       1        1
     30     8     B          1        1       2        2
     31     8     C          2        2       1        2
     32     8     None       .        .       .        .

------------------------------------------------------------------------

Now assume like before that there are three alternatives, each composed
of four factors: Brand Price Size Color Shape.  In addition there is a
constant alternative.  Also, there is an alternative-specific factor,
Pattern, that only applies to Brand A and Brand C.  First, the MktEx
macro is used to create a design with 14 factors, one for each attribute
of each alternative.

   %mktex(2 ** 14, n=16)

Then the KEY= data set is created.  It shows that there are three
brands, A, B, and C, plus None.

   data key;
      input (Brand Price Size Color Shape Pattern) ($);
      datalines;
   A     x1  x2  x3  x4 x13
   B     x5  x6  x7  x8   .
   C     x9 x10 x11 x12 x14
   None   .   .   .   .   .
   ;

Brand A is created from Brand = 'A',
Price = x1, Size = x2, Color = x3, Shape = x4, Pattern = x13.

Brand B is created from Brand = 'B',
Price = x5, Size = x6, Color = x7, Shape = x8.

Brand C is created from Brand = 'C',
Price = x9, Size = x10, Color = x11, Shape = x12, Pattern = x14.

The constant alternative is Brand = 'None' and none of the attributes.

Here is how you can create the design with one row per
alternative per choice set:

%mktroll(key=key, design=randomized, out=sasuser.design, alt=brand)

For example if the data set DESIGN contains the row:

Obs  x1  x2  x3  x4  x5  x6  x7  x8  x9  x10  x11  x12  x13  x14

  8   2   1   1   1   1   1   2   2   2   2    1    2    2    1

Then the data set SASUSER.DESIGN contains the rows:

Obs   Set   Brand   Price   Size   Color   Shape   Pattern

 29    8    A         2       1      1       1        2
 30    8    B         1       1      2       2        .
 31    8    C         2       2      1       2        1
 32    8    None      .       .      .       .        .

------------------------------------------------------------------------

Now assume we are going to fit a model with price cross effects so we
need x1, x5, and x9 (the three price effects) available in the OUT= data
set.

   %mktroll(key=key, design=randomized, out=sasuser.design, alt=brand,
            keep=x1 x5 x9)

Now the data set also contains the three original price variables.

 Obs  Set  Brand  Price  Size  Color  Shape  Pattern  x1  x5  x9

  29   8   A        2      1     1      1       2      2   1   2
  30   8   B        1      1     2      2       .      2   1   2
  31   8   C        2      2     1      2       1      2   1   2
  32   8   None     .      .     .      .       .      2   1   2

------------------------------------------------------------------------

The macro performs some rudimentary error checking.  Every value in the
KEY= data set must appear as a variable in the DESIGN= data set.  The
macro prints a warning if it encounters a variable name in the DESIGN=
data set that does not appear as a value in the KEY= data set.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

                  /*--------------------------------------------------*/
%macro mktroll(   /* This macro creates an experimental design with   */
                  /* one row per alternative per choice set from a    */
                  /* design with one row per choice set.              */
                  /*--------------------------------------------------*/
help,             /* nonblank, list options                           */
                  /*--------------------------------------------------*/
design=,          /* The DESIGN= option names an input SAS data set   */
data=,            /* with one row per choice set.  The DESIGN= option */
                  /* must be specified.  DATA= is an alias for        */
                  /* DESIGN=. Specify DESIGN= or DATA= but not both.  */
                  /*--------------------------------------------------*/
key=,             /* The KEY= option must be specified.  It has one   */
                  /* of two forms.                                    */
                  /* 1) The KEY= option names an input SAS data set   */
                  /* containing the rules for mapping the DESIGN=     */
                  /* data set to the OUT= data set.  The structure of */
                  /* this data set is described in detail in the      */
                  /* header above.                                    */
                  /* 2) When you want the KEY= data set to exactly    */
                  /* match the data set that comes out of the MktKey  */
                  /* macro, you can specify the argument to the       */
                  /* MktKey macro directly in the KEY= option, and    */
                  /* the MktRoll macro will make the KEY=KEY data     */
                  /* set for you.  In other words, the following two  */
                  /* specifications are equivalent:                   */
                  /*                                                  */
                  /* %mktkey(3 3 t)                                   */
                  /* %mktroll(design=design, key=key,   out=frac)     */
                  /*                                                  */
                  /* %mktroll(design=design, key=3 3 t, out=frac)     */
                  /*                                                  */
                  /*--------------------------------------------------*/
out=,             /* The OUT= option names the output SAS data set.   */
                  /* If OUT= is not specified, the DATAn convention   */
                  /* is used.                                         */
                  /*--------------------------------------------------*/
alt=,             /* ALT= names the variable in the KEY= data set     */
                  /* that contains the name of each alternative.      */
                  /* Often this will be something like ALT=Brand.     */
                  /* When ALT= is not specified, the macro creates    */
                  /* a variable _Alt_ that contains the alternative   */
                  /* number.                                          */
                  /*--------------------------------------------------*/
keep=,            /* KEEP= names factors from the DESIGN= data set    */
                  /* that should also be kept in the OUT= data set.   */
                  /* This option is useful for blocking variables     */
                  /* and to keep terms that will be used to create    */
                  /* cross effects.                                   */
                  /*--------------------------------------------------*/
set=Set,          /* SET= names the variable in the OUT= data set     */
                  /* that will contain the choice set number.  By     */
                  /* default this variable is named Set.              */
                  /*--------------------------------------------------*/
options=,         /* nowarn - do not print a warning when the DESIGN= */
                  /*          data set contains variables not         */
                  /*          mentioned in the KEY= data set.         */
                  /*          Sometimes this is perfectly fine.       */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _MktVer;
%local  abort altlen nalts nfact rename savenote j k maxlen localver
        setnote time savedsnf warn scanq saveopts timeopt dokey header;

%let time     = %sysfunc(datetime());
%let timeopt  = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote = %sysfunc(getoption(notes));
%let saveopts = &savenote missing="%sysfunc(getoption(missing))"
                %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                %sysfunc(getoption(serror));
%let setnote  = nonotes;
%let maxlen   = 32700;
%let abort    = 0;
%let localver = 25Jul2010;
%let _mktver  = &localver;

%if %index(%nrbquote(&options), notes) %then %let setnote = notes;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES)   %then %let setnote = notes;
   %if %index(&k, VERSION) %then %put MktRoll macro version &localver.;
   %end;
options serror;

options &setnote generic missing='.';

%if %nrbquote(&help) ne %then %do; %mkthelp(mktroll) %goto endit; %end;

*------rudimentary error checking------;
data _null_;
   length list $ 200 n1-n2 $ 12;

   list = symget('key');
   if list = ' ' then do;
      put 'ERROR: KEY= must be specified.';
      error + 1;
      end;

   if symget('design') ne ' ' and symget('data') ne ' ' then do;
      put 'ERROR: Specify DESIGN= or DATA= but not both.';
      error + 1;
      end;

   else if symget('design') eq ' ' and symget('data') ne ' '
      then call symput('design', symget('data'));

   if symget('design') = ' ' then do;
      put 'ERROR: DESIGN= must be specified.';
      error + 1;
      end;

   n1 = scan(list, 3, ' ');
   if lowcase(n1) = 't' then n1 = scan(list, 4, ' ');
   n2 = scan(list, 2, ' ');
   call symput('dokey', '0');
   if n2 ne ' ' and n1 eq ' ' then do;
      n1 = scan(list, 1, ' ');
      i = input(n1, ?? 12.);
      j = input(n2, ?? 12.);
      if n(i,j) = 2 and i = round(i) and j = round(j) then
         call symput('dokey', '1');
      end;

   *------parsing option------;
   list = lowcase(symget('options'));
   i    = index(list, 'nowarn'); if i then substr(list, i, 6) = ' ';
   call symput('warn', put(i eq 0, 1.));
   i    = index(list, 'notes'); if i then substr(list, i, 5) = ' ';
   if i then
      put 'WARNING: OPTIONS=NOTES is obsolete.' /
          +9 'Submit "%let mktopts = notes;" before running the macro.';
   list = compbl(list);

   if list ne ' ' then do;
      put 'ERROR: OPTIONS=' list 'is invalid.';
      error + 1;
      end;

   if error or _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------optionally make key data set------;
%if &dokey %then %do;
   %let header = title1;
   proc sql; create view __tmpltf as select * from dictionary.titles; quit;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data _null_;
      set __tmpltf end=eof;
      titles + type eq 'T';
      if eof then call symput('header', 'title' || put(titles + 1, 2.0 -L));
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %let _mktver = obsolete;
   &header 'Automatically Generated KEY=KEY Data Set';
   %mktkey(&key)
   &header;

   %if &_mktver ne &localver %then %do;
      %put ERROR: Incompatible MktKey macro.;
      %put %str(      ) You must have the same version of all macros.;
      %put %str(      ) Versions: MktRoll = &localver, MktKey = &_mktver..;
      %put %str(      ) Get the latest tools and documentation from:;
      %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
      %let abort = 1;
      %end;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %let key = key;
   %end;

*------mention data sets, if they are invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&design) ne %then %do;
   data _null_; set &design; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&key) ne %then %do;
   data _null_; set &key; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )design=&design;
   %put %str(       )out=&out;
   %put %str(       )key=&key;
   %end;

%if &abort %then %goto endit;

*------count factors, alternatives------;
*------output factor names in macro variables (fac:)------;
*------output names from key to check------;
data __tmplknam(keep=__name);

   set &key %if %nrbquote(&alt) ne %then (drop=&alt); nobs=__nalts;
   array __x[*] _character_;
   length __name $ 32;
   if _n_ = 1 then do;
      call symput('nfact', compress(put(dim(__x), 5.)));
      call symput('nalts', compress(put(__nalts, 5.)));
      do __i = 1 to dim(__x);
         call vname(__x[__i], __name);
         call symput(compress('__fac' || put(__i, 5.)), trim(__name));
         end;
      end;
   do __i = 1 to dim(__x);
      __name = upcase(__x[__i]);
      if __name ne ' ' then output;
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------store the key data set, rename vars (__fac:)------;
data __tmplkey;
   keep __fac:;
   set &key %if %nrbquote(&alt) ne %then (drop=&alt);;
   array __x[*] _character_;
   array __fac[&nfact] $;
   do __i = 1 to &nfact;
      __fac[__i] = upcase(__x[__i]);
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------store alternative names in macro variables: (alt:)------;
data _null_;
   set &key;
   length __name $ 32;
   retain altlen 8;
   __name = compress('alt' || put(_n_, 5.));
   %if %nrbquote(&alt) ne %then %do;
      call symput(__name, trim(&alt));
      altlen = max(length(&alt), altlen);
      call symput('altlen', compress(put(altlen, 5.)));
      if _n_ = 1 then do;
         call vname(&alt,  __name);
         call symput('alt', trim(__name));
         end;
      %end;
   %else %do;
      %let alt = _Alt_;
      %let altlen = 8;
      call symput(__name, compress(put(_n_, 5.)));
      %end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get names of factors from linear design------;
proc contents data=&design out=__tmplcon noprint;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------check variable names for one to one match------;
data __tmplcon2;
   set __tmplcon(keep=name);
   name = upcase(name);
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc sort; by name; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

proc sort data=__tmplknam(rename=(__name=name)) out=__tmplknam;
   by name;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%if %sysevalf(&sysver ge 9) %then %let scanq = scanq;
                            %else %let scanq = scan;

data _null_;
   merge __tmplcon2(in=design) __tmplknam(in=key);
   by name;
   length w $ 32 list $ 32000;
   retain list;
   keep = 0;
   if _n_ = 1 then list = symget('keep');
   if list ne ' ' then do;
      do i = 1 to 32000 until(keep or w eq ' ');
         w = left(dequote(&scanq(list, i)));
         if upcase(w) = upcase(name) then keep = 1;
         end;
      end;
   if &warn and design and not key and not keep then do;
      put 'WARNING: The variable ' name 'is in the DESIGN= data set '
          'but ' 'not the ' 'KEY= data set.';
      end;
   if key and not design then do;
      put 'ERROR: The variable ' name 'is in the KEY= data set '
          'but ' 'not the ' 'DESIGN= data set.';
      error + 1;
      end;
   if name eq lag(name) then do;
      put 'ERROR: The variable ' name 'is used more than once.';
      error + 1;
      end;

   if error or _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------restore original order of names------;
proc sort data=__tmplcon out=__tmplcon(keep=name); by varnum; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------rename design variables, eliminate possible name collisions------;
data _null_;
   length rename $ &maxlen; retain rename;
   set __tmplcon end=__eof;
   rename = trim(rename) || ' ' || trim(name) ||
            compress('=__des' || put(_n_, 5.));
   if __eof then call symput('rename', trim(rename));
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------create a single row of names------;
proc transpose out=__tmplcon(keep=__:) prefix=__nam;
   var name;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------convert names to indexes------;
data __tmplkey2(keep=__num:);
   set __tmplkey;
   array __fac[&nfact];
   array __num[&nfact];
   if _n_ = 1 then do;
      set __tmplcon;
      array __nam[*] __nam:;
      do __i = 1 to dim(__nam);
         __nam[__i] = upcase(__nam[__i]);
         end;
      end;
   do __j = 1 to &nfact;
      __num[__j] = .;
      do __i = 1 to dim(__nam);
         if __nam[__i] = __fac[__j] then do;
            __num[__j] = __i;
            leave;
            end;
         end;
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

options &savenote no&timeopt;

*------roll out the design------;
data &out;
   set &design(rename=(&rename));
   array __des[*] __des:;
   &set = _n_;
   length &alt $ &altlen;
   do __i = 1 to &nalts;
      set __tmplkey2 point=__i;
      array __num[&nfact];
      &alt = symget(compress("alt" || put(__i, 5.)));
      %do j = 1 %to &nfact;
         &&__fac&j = .;
         if __num[&j] ne . then &&__fac&j = __des[__num[&j]];
         %end;
      %if %nrbquote(&keep) ne %then %do;
         set &design(keep=&keep) point=_n_;
         %end;
      output;
      end;
   keep &set &alt %do j = 1 %to &nfact; &&__fac&j %end; &keep;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

options &setnote;
proc datasets nolist;
   delete __tmplcon __tmplcon2 __tmplkey __tmplkey2 __tmplknam
          __tmpltf(memtype=view);
   run; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%endit: options &saveopts;
quit;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTROLL macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTROLL macro used &time seconds.;
   %end;

%mendit: %mend;
