 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktRuns                                             */
 /*   TITLE: Calculate Reasonable Experimental Designs Sizes     */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs                         */
 /*   PROCS:                                                     */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: This macro creates global macro variables           */
 /*             _MktList _MktSat _MktLogFF _MktFF _MktVer        */
 /*             _MktIntList _MktOrth                             */
 /*                                                              */
 /*          The following SAS products are required to run      */
 /*          this macro: BASE.                                   */
 /****************************************************************/

 /*---------------------------------------------------------------------

This macro calculates reasonable sizes for main-effects experimental
designs and designs with two-way interactions.  It tries to find design
sizes in which perfect balance and orthogonality can occur or at least
sizes in which violations of orthogonality and balance are minimized.
Typically, the macro takes one argument, a list of the number of levels
of each factor.

For example, with 3 two-level and 4 three-level factors, specify the
macro with three 2's and four 3's, which are the numbers of levels for
all of the factors.

   %mktruns(2 2 2 3 3 3 3)

The output of the macro in this example is:

                            Design Summary

                        Number of
                        Levels       Frequency

                             2           3
                             3           4

            Saturated      = 12
            Full Factorial = 648

            Some Reasonable                      Cannot Be
               Design Sizes       Violations     Divided By

                         36 *              0
                         72 *              0
                         18                3     4
                         54                3     4
                         12 S              6     9
                         24                6     9
                         48                6     9
                         60                6     9
                         30                9     4 9
                         42                9     4 9

      * - 100% Efficient design can be made with the MktEx macro.
      S - Saturated Design - The smallest design that can be made.

       n    Design                                   Reference

      36    2 ** 13  3 **  4                      Orthogonal Array
      36    2 ** 11  3 ** 12                      Orthogonal Array
      36    2 ** 10  3 **  8  6 **  1             Orthogonal Array
      36    2 **  9  3 **  4  6 **  2             Orthogonal Array
      36    2 **  4  3 ** 13                      Orthogonal Array
      36    2 **  3  3 **  9  6 **  1             Orthogonal Array
      72    2 ** 49  3 **  4                      Orthogonal Array
      72    2 ** 47  3 ** 12                      Orthogonal Array
      72    2 ** 46  3 **  8  6 **  1             Orthogonal Array
       .
       .
       .

The macro reports that the saturated design has 12 runs and that 36 and
72 are optimal design sizes.  The macro picks 36 because it is the
smallest integer >= 12 that can be divided by 2, 3, 2 * 2, 2 * 3, and 3
* 3.  Since 72 = 2x36, it can be divided by the same numbers.
The macro also shows you that 100% efficient designs in 36 or 72 runs
can be created by the MktEx macro.  To see every size the macro
considered, simply run proc print after the macro finishes.

   proc print label data=nums split='-';
      label s = '00'x;
      id n;
      run;

The macro also reports 18 as a reasonable size.  There are three
violations with 18 because 18 cannot be divided by each of the three
pairs of 2 * 2, so perfect orthogonality in the two-level factors will
not be possible.  The macro also reports larger sizes.

For 2 two-level factors, 2 three-level factors, 2 four-level factors,
and 2 five-level factors specify:

   %mktruns(2 2 3 3 4 4 5 5)

Here are the results:

                        Design Summary

                    Number of
                    Levels       Frequency

                         2           2
                         3           2
                         4           2
                         5           2

   Saturated      = 21
   Full Factorial = 14,400

   Some Reasonable                      Cannot Be
      Design Sizes       Violations     Divided By

               120                3     9 16 25
               180                6     8 16 25
                60                7     8  9 16 25
               144               15     5 10 15 20 25
                48               16     5  9 10 15 20 25
                72               16     5 10 15 16 20 25
                80               16     3  6  9 12 15 25
                96               16     5  9 10 15 20 25
               160               16     3  6  9 12 15 25
               192               16     5  9 10 15 20 25
                21 S             34     2  4  5  6  8  9 10 12 15 16 20 25

          S - Saturated Design - The smallest design that can be made.
              Note that the saturated design is not one of the
              recommended designs for this problem.  It is shown
              to provide some context for the recommended sizes.

Among the smaller design sizes, 60 or 48 look like good possibilities.

The macro has a second, optional, keyword parameter: MAX=.  It specifies
the maximum number of sizes to try.  The smallest design considered is
the saturated design.  For example, this specification tries 5000 sizes
and reports that a perfect design can be found with 3600 runs.

   %mktruns(2 2 3 3 4 4 5 5, max=5000)

Usually you will not need to specify the MAX= option.

Now consider again the problem with 3 two-level and 4 three-level
factors, but this time we want to be estimable the interaction of
two of all of the two-level factors.

   %mktruns(2 2 2 3 3 3 3, interact=1*2 1*3 2*3, options=source)

                                             Failing to
                                             Divide by
                                             the Divisor
                                             Contributes
                                             this to the
                                             Violations
      To Achieve Both    The Source of       Count
      Orthogonality      the Divisor                          The Divisor
      and Balance,                           The Number of    Comes
      N Must Be          Number of           Sources of       from these
      Divided By         Levels or Cells     the Divisor      Factors

              2               2                     3            1
                                                                 2
                                                                 3

              3               3                     4            4
                                                                 5
                                                                 6
                                                                 7

              4               2*2                   3            1 2
                                                                 1 3
                                                                 2 3

              6               2*3                  12            1 4
                                                                 1 5
                                                                 1 6
                                                                 1 7
                                                                 2 4
                                                                 2 5
                                                                 2 6
                                                                 2 7
                                                                 3 4
                                                                 3 5
                                                                 3 6
                                                                 3 7

              8               2*2*2                 1            1 2 3

              9               3*3                   6            4 5
                                                                 4 6
                                                                 4 7
                                                                 5 6
                                                                 5 7
                                                                 6 7

             12               2*2*3                12            1 2 4
                                                                 1 2 5
                                                                 1 2 6
                                                                 1 2 7
                                                                 1 3 4
                                                                 1 3 5
                                                                 1 3 6
                                                                 1 3 7
                                                                 2 3 4
                                                                 2 3 5
                                                                 2 3 6
                                                                 2 3 7


                                Design Summary

                            Number of
                            Levels       Frequency

                                 2           3
                                 3           4


            Saturated      = 15
            Full Factorial = 648

            Some Reasonable                      Cannot Be
               Design Sizes       Violations     Divided By

                         72                0
                        144                0
                         36                1     8
                        108                1     8
                         24                6     9
                         48                6     9
                         96                6     9
                        120                6     9
                         60                7     8  9
                         84                7     8  9
                         15 S             37     2  4  6  8  9 12

          S - Saturated Design - The smallest design that can be made.
              Note that the saturated design is not one of the
              recommended designs for this problem.  It is shown
              to provide some context for the recommended sizes.

Now we need 72 runs for perfect balance and orthogonality and there are
six violations in 24 runs.

You can also specify levels using a standard design notation with
exponentials.  For example, for 10 two-level factors and 5 three-level
factors, you can specify:

   %mktruns(2**10 3**5)

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*===================== Macro Argument List Start =====================;

%macro mktruns(   /*--------------------------------------------------*/
list,             /* List of numbers of levels of all the factors.    */
                  /* The list can be a list of numbers, like:         */
                  /*    2 2 3 3 4 4                                   */
                  /* or a "levels**number of factors" list like:      */
                  /*    2**2 3**2 4**2                                */
                  /* or any mix of the two:                           */
                  /*    2 2 3**4 5 6                                  */
                  /*                                                  */
                  /* ? or help - lists options                        */
                  /*--------------------------------------------------*/
n=,               /* when specified, prints information for just that */
                  /* design size.                                     */
                  /*--------------------------------------------------*/
max=200 2 f,      /* When one number is specified, it is the number   */
                  /* of sizes to try starting with the saturated      */
                  /* design.  A second optional value, m, can be      */
                  /* specified.  This indicates that the macro should */
                  /* stop if zero violations are found using a design */
                  /* size that is m times the size of the smallest    */
                  /* design with zero violations.  For example, if    */
                  /* there are zero violations in 36 runs, then with  */
                  /* the default MAX=200 2 f, the macro stops when it */
                  /* reaches zero violations at 2 * 36 = 72 runs, and */
                  /* it does not consider larger sizes.  A third      */
                  /* optional value of 'f' or 'F' is specified by     */
                  /* default, which instructs the MktRuns macro to    */
                  /* not consider designs larger than                 */
                  /* full-factorials.  If you specify MAX= without    */
                  /* this option, larger designs might be considered. */
                  /*--------------------------------------------------*/
options=,         /* source - prints the source of all of the numbers */
                  /* in the table of reasonable design sizes.         */
                  /*                                                  */
                  /* ADX - macro is being called by ADX.              */
                  /*                                                  */
                  /* multiple - allows a term that is required for    */
                  /* orthogonality to be counted multiple times when  */
                  /* counting orthogonality violations.  For example, */
                  /* combinations of levels for the pair of variable  */
                  /* 1 and variable 2 must have equal frequencies for */
                  /* orthogonality in the main effects, and if        */
                  /* two-way interactions are required as well, then  */
                  /* the (1, 2) pair must have equal frequencies      */
                  /* again.  By default, each combination of          */
                  /* variables is only counted once.                  */
                  /*                                                  */
                  /* multiple2 - is like multiple, but with more      */
                  /* detailed output: the reason each term appears is */
                  /* added to the source table when OPTIONS=SOURCE    */
                  /* MULTIPLE2 is specified.                          */
                  /*                                                  */
                  /* The difference between the single and multiple   */
                  /* sources is the single source method just counts  */
                  /* the places in the design where equal frequencies */
                  /* must occur.  The multiple source method weights  */
                  /* this count by the number of times each of these  */
                  /* terms is important for achieving orthogonality.  */
                  /* The results for the two methods are often highly */
                  /* correlated, but they can be different.           */
                  /*                                                  */
                  /* justparse - option used by other Mkt macros to   */
                  /* have this macro just parse the list argument     */
                  /* and return it as a simple list of integers.      */
                  /* This option is also used to parse INTERACT=.     */
                  /* The results of parsing are returned in global    */
                  /* macro variables:                                 */
                  /* _MktList - list converted to a simple integer    */
                  /* list, eg if a list of 2**5 is specified, then    */
                  /* _MktList=2 2 2 2 2.                              */
                  /* _MktSat - number of runs in the saturated        */
                  /* design.                                          */
                  /* _MktLogFF - log10 of the number of runs in the   */
                  /* full-factorial design.                           */
                  /* _MktFF - number of runs in the full-factorial    */
                  /* design.  For large problems, this number may be  */
                  /* too big to assign to a numeric variable.         */
                  /* _MktIntList - Interaction terms as an integer    */
                  /* list.                                            */
                  /*                                                  */
                  /* nosat - by default, the saturated design is      */
                  /* reported along with the reasonable design sizes, */
                  /* even if it is not a very reasonable size in      */
                  /* order to provide some context for the other      */
                  /* numbers. If you specify the NOSAT option, then   */
                  /* the saturated design will only be added to the   */
                  /* list if it meets the usual criteria for being a  */
                  /* reasonable size.                                 */
                  /*                                                  */
                  /* 512 - adds some larger designs in 512 runs with  */
                  /* mixes of 16, 8, 4, and 2-level factors to the    */
                  /* catalog, which gives added flexibility in 512    */
                  /* runs at a cost of much slower run time.  This    */
                  /* option replaces the default 4^160 32^1 parent    */
                  /* with 16^32 32^1.                                 */
                  /*                                                  */
                  /* noprint - just make output data sets, don't      */
                  /* print anything.                                  */
                  /*--------------------------------------------------*/
interact=,        /* Interactions that must be estimable.             */
                  /* Examples:                                        */
                  /*    INTERACT=x1*x2                                */
                  /*    INTERACT=x1*x2 x3*x4*x5                       */
                  /*    INTERACT=x1|x2|x3|x4|x5@2                     */
                  /*                                                  */
                  /* The interaction syntax is like GLM's and many of */
                  /* the other modeling procedures.  It uses "*" for  */
                  /* simple interactions (x1*x2, the interaction      */
                  /* between x1 and x2), "|" for main effects and     */
                  /* interactions (x1|x2|x3 is the same as x1 x2      */
                  /* x1*x2 x3 x1*x3 x2*x3 x1*x2*x3) and "@" to        */
                  /* eliminate higher-order interactions (x1|x2|x3@2  */
                  /* is the same as x1 x2 x1*x2 x3 x1*x3 x2*x3).      */
                  /* "@2" allows only main effects and two-way        */
                  /* interactions.  By default, no interactions are   */
                  /* guaranteed to be estimable.  For the factor      */
                  /* names, you must specify either the actual        */
                  /* variable names (eg x1 x2 ...) or you can just    */
                  /* specify the number without the "x" (eg x1*x2 is  */
                  /* equivalent to 1*2).                              */
                  /*                                                  */
                  /* You can also specify INTERACT=@2 for all main    */
                  /* effects and two-way interactions.  For example,  */
                  /* these two specifications are equivalent:         */
                  /*    %mktruns(2 ** 5, interact=@2)                 */
                  /*    %mktruns(2 ** 5, interact=1|2|3|4|5@2)        */
                  /*                                                  */
out=nums,         /* output data set with suggested sizes.  Full      */
                  /* data set with all sizes that were considered.    */
                  /*                                                  */
outtab=,          /* output data set with printed table of suggested  */
                  /* sizes.  By default this data set is not created. */
                  /*                                                  */
outorth=,         /* output data set with orthogonal array list.      */
                  /* By default this data set is not created.         */
                  /*                                                  */
maxlev=50,        /* specifies the maximum number of levels to        */
                  /* consider when generating the orthogonal array    */
                  /* list.  The actual maximum is the max of the      */
                  /* specified MAXLEV= value and the maximum number   */
                  /* of levels in the factor list.  Specify a value   */
                  /* 2 <= maxlev <= 144.                              */
                  /*                                                  */
maxoa=,           /* specifies the maximum number of orthogonal       */
                  /* arrays to print in the list of orthogonal arrays */
                  /* that the MktEx macro knows how to make.  By      */
                  /* default, when no value is specified, the entire  */
                  /* list is printed.  You can specify for example    */
                  /* MAXOA=0, OUTORTH=OA, to get a data set instead   */
                  /* of printing this table.  You could instead       */
                  /* specify MAXOA=10 to just see the first ten       */
                  /* arrays.                                          */
                  /*                                                  */
toobig=11400,     /* used to flag problems that are too big and take  */
                  /* too long before they chew up tons of resources.  */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* the notes, submit the statement                  */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=========================== Macro Start =============================;

*------initial macro stuff------;
%global _MktList _MktSat _MktLogFF _MktFF _MktVer _MktIntList _MktExists;
%local  abort ff levformat maxlen maxstrlen savenote needlist nvars nints k
        foot setnote stop time parse adj range maxn saveopts localver source
        multiple multiple2 d0512 fullfacstop foot_s foot_a showsat print
        savedsnf adx maxterms timeopt;

%let time     = %sysfunc(datetime());
%let timeopt  = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote = %sysfunc(getoption(notes));
%let saveopts = &savenote missing="%sysfunc(getoption(missing))"
                %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                %sysfunc(getoption(serror));
%let setnote  = nonotes;
%let abort    = 0;
%let parse    = 0;
%let maxlen   = 32700;
%let localver = 25Jul2010;
%let _mktver  = &localver;

options noserror;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES)   %then %let setnote = notes;
   %if %index(&k, VERSION) %then %put MktRuns macro version &localver.;
   %end;
options serror;

options &setnote generic missing='.';

data _null_;
   length list $ 32;
   list = left(lowcase(symget('list')));
   if list eq 'help' or list eq '?' then call symput('abort', '-1');
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort = 1 %then %goto endit;

%if &abort = -1 %then %do; %let abort = 0; %mkthelp(mktruns) %goto endit; %end;


*------integrity checks and initial processing------;
data _null_;
   length intlist list hold term $ &maxlen w w2 $ 24;
   array ci[10]; ci0 = 0;

   abort = 0;

   *------initialize null out=------;
   if symget('out') = ' ' then call symput('out', '__tmprout');

   *------initialize flags------;
   call symput('foot'  , '0');
   call symput('foot_s', '0');
   call symput('foot_a', '0');
   call symput('_MktExists', '0');

   *------parsing option------;
   list = lowcase(symget('options'));
   i = index(list, 'adx');    if i then substr(list, i, 3) = ' ';
   call symput('adx', put(i ne 0, 1.));
   i = index(list, 'noprint'); if i then substr(list, i, 7) = ' ';
   call symput('print', put(i eq 0, 1.));
   i = index(list, 'justparse'); if i then substr(list, i, 9) = ' ';
   call symput('parse', put(i ne 0, 1.));
   i = index(list, 'source'); if i then substr(list, i, 6) = ' ';
   call symput('source', put(i ne 0, 1.));
   i = index(list, 'multiple2'); if i then substr(list, i, 9) = ' ';
   call symput('multiple2', put(i ne 0, 1.));
   i = index(list, 'multiple'); if i then substr(list, i, 8) = ' ';
   call symput('multiple', put(i ne 0, 1.));
   i = index(list, '512'); if i then substr(list, i, 3) = ' ';
   call symput('do512', put(i gt 0, 1.));
   i = index(list, 'nosat'); if i then substr(list, i, 5) = ' ';
   call symput('showsat', put(i eq 0, 1.));
   list = compbl(list);

   if list ne ' ' then do;
      put 'ERROR: OPTIONS=' list 'is invalid.';
      abort = 1;
      goto endit;
      end;

   *------check n=------;
   w     = symget('n');
   check = input(w, ?? 24.);
   if w ne ' ' and (nmiss(check) or check < 1 or check ne round(check)) then do;
      put "ERROR: N=%nrbquote(&n) is invalid.";
      abort = 1;
      goto endit;
      end;

   *------check maxoa=------;
   w     = symget('maxoa');
   check = input(w, ?? 24.);
   if w ne ' ' and (nmiss(check) or check < 0 or check ne round(check)) then do;
      put "ERROR: MAXOA=%nrbquote(&maxoa) is invalid.";
      abort = 1;
      goto endit;
      end;

   *------check maxlev=------;
   maxlev = input(symget('maxlev'), ?? 23.);
   ok = n(maxlev);
   if ok then ok = (maxlev eq floor(maxlev) and 2 le maxlev le 144);
   if not ok then do;
      put "ERROR: Invalid maxlev=&maxlev..";
      abort = 1;
      end;

   *------integrity checks on MAX= option------;
   list  = symget('max');
   i     = index(lowcase(list), 'f');
   call symput('fullfacstop', put(i gt 0, 1.));
   if i then substr(list, i, 1) = ' ';

   w     = scan(list, 1, ' ');
   check = input(w, ?? 12.);
   if nmiss(check) or check < 1 then do;
      put 'ERROR: MAX=' list 'is invalid.';
      abort = 1;
      goto endit;
      end;

   call symput('max', compress(w));
   list = scan(list, 2, ' ');
   if list = ' ' then list = w;
   check = input(list, ?? 12.);
   if not abort and (nmiss(check) or check < 1) then do;
      put 'ERROR: MAX= second value of ' list 'is invalid.';
      abort = 1;
      goto endit;
      end;
   call symput('stop', compress(list));

   *------start processing factor level list-----;
   hold = symget('list');
   if length(hold) > (&maxlen - 100) then do;
      put 'ERROR: The factor-level list is too long.';
      abort = 1;
      goto endit;
      end;

   if hold = ' ' then do;
      put 'ERROR: The factor-level list must be specified.';
      abort = 1;
      goto endit;
      end;

   *------compress out extra blanks, some integrity checks------;
   list = compbl(hold);
   do i = 1 to length(list);
      ss = substr(list, i, 2);
      if ss eq ' *' or ss = '* '
         then substr(list, i) = '*' || left(substr(list, i + 2));
      end;
   do i = 1 to length(list);
      if not (scan(list, i, ' 0123456789') in (' ', '**')) then do;
         put 'ERROR: Invalid factor-level specification: ' hold +(-1) '.';
         abort = 1;
         goto endit;
         end;
      end;

   if substr(list, length(list), 1) = '*' then do;
      put 'ERROR: The number of factor-levels list may not end with "*".';
      abort = 1;
      goto endit;
      end;

   *------expand out 'n**m' to n-levels list------;
   do i = 1 to &maxlen until(w eq ' ');
      w = scan(list, i, ' ');
      if w ne ' ' then do;
         if index(w, '*') then do;
            p1 = index(list, trim(w));
            p2 = p1 + index(substr(list, p1), ' ');
            r  = input(scan(w, 2, ' *'), ?? 12.);
            if nmiss(r) or r < 1 or r > (&maxlen / 2) or
               r ne round(r) or scan(w, 3, ' *') ne ' ' then do;
               put 'ERROR: Invalid factor-level specification: ' w +(-1) '.';
               abort = 1;
               goto endit;
               end;
            w = scan(w, 1, ' *');
            hold = substr(list, p2);
            if p1 > 1 then list = substr(list, 1, p1 - 1);
            else list = ' ';
            do j = 1 to r;
               list = trim(left(list)) || ' ' || compress(w);
               end;
            list = trim(list) || ' ' || hold;
            i = i - 1;
            end;
         else do;
            a = input(w, ?? best12.);
            if nmiss(a) or a < 2 or a > 1e5 or a ne round(a) then do;
               put 'ERROR: Invalid number of factor levels: ' w +(-1) '.';
               abort = 1;
               goto endit;
               end;
            end;
         end;
      end;

   if length(list) > (&maxlen - 100) then do;
      put 'ERROR: The factor-level list is too long.';
      abort = 1;
      goto endit;
      end;

   call symput('_mktlist', trim(list));

   maxlev = &maxlev;
   do nvars = 0 to &maxlen until(w eq ' ');
      w = scan(list, nvars + 1);
      if w ne ' ' then maxlev = max(maxlev, input(w, ?? 23.));
      end;
   maxlev = max(2, min(maxlev, 144));
   call symput('nvars' , compress(put(nvars , 12.)));
   call symput('maxlev', compress(put(maxlev, 12.)));

   *------interaction list------;
   list = left(compbl(translate(symget('interact'), '  ', 'xX')));

   if list =: '@' and length(list) le 3 and
      n(input(scan(list, 1, '@ '), ?? 23.)) then do;
      hold = '1';
      do i = 2 to nvars; hold = trim(hold) || '|' || put(i, 5. -L); end;
      list = trim(hold) || list;
      end;

   l = length(list);

   if l gt &maxlen * 0.9 or compress(list, '0123456789*@|') ne ' ' then do;
      abort = 2;
      goto endit;
      end;

   *------get rid of extra blanks so we can grab terms------;
   do i = 1 to l;
      if substr(list, i, 2) = ' |' then substr(list, i)   = substr(list, i + 1);
      if substr(list, i, 2) = '| ' then substr(list, i+1) = substr(list, i + 2);
      if substr(list, i, 2) = ' *' then substr(list, i)   = substr(list, i + 1);
      if substr(list, i, 2) = '* ' then substr(list, i+1) = substr(list, i + 2);
      if substr(list, i, 2) = ' @' then substr(list, i)   = substr(list, i + 1);
      if substr(list, i, 2) = '@ ' then substr(list, i+1) = substr(list, i + 2);
      end;

   gt1 = (list ne ' ' and index(trim(list), ' '));

   *------do some error checking and expand out "|" and "@"------;
   intlist = ' ';   invalid = 0;
   do j = 1 to &maxlen until(term eq ' ');
      invalid = 0;

      *------pull out one interaction term------;
      term = scan(list, j, ' ');
      hold = term;
      if term ne ' ' then do;

         *------put blanks on each side of "|*@"------;
         l = min(&maxlen, 3 * length(term));
         do i = 1 to l;
            if substr(term, i, 1) = '|' then do;
               substr(term, i) = ' | ' || substr(term, i + 1);
               i + 2;
               end;
            if substr(term, i, 1) = '*' then do;
               substr(term, i) = ' * ' || substr(term, i + 1);
               i + 2;
               end;
            if substr(term, i, 1) = '@' then do;
               substr(term, i) = ' @ ' || substr(term, i + 1);
               i + 2;
               end;
            end;

         *------see what kind of list this is------;
         star = index(term, '*');
         bar  = index(term, '|');
         at   = index(term, '@');
         if star and (bar or at) then invalid = 1;
         if scan(term, 3, ' ') = ' ' then invalid = 1;
         w = substr(term, length(term), 1);
         if w eq '*' or w eq '|' or w eq '@' then invalid = 1;
         atnum = 0;

         *------star list make sure n * m ...------;
         l = length(term);
         if star then do i = 1 to l until(w eq ' ');
            w = scan(term, i, ' ');
            if w ne ' ' then do;
               if mod(i, 2) then do;
                  if nmiss(input(w, ?? 8.)) then invalid = 1;
                  end;
                  else if w ne '*' then invalid = 1;
               end;
            end;

         %*------bar list, check sequence, find @ if any------;
         else if bar then do;
            do i = 1 to l until(w eq ' ');
               w = scan(term, i, ' ');
               if w ne ' ' then do;
                  if mod(i, 2) then do;
                     if nmiss(input(w, ?? 8.)) then invalid = 1;
                     end;
                  else do;
                     if w eq '@' then do;
                        if atnum then invalid = 1;
                        else atnum = i;
                        end;
                     end;
                  end;
               end;
            l = i - 1;
            if atnum and atnum + 1 ne l then invalid = 1;
            end;

         *------check for duplicate var nums------;
         do ii = 2 to l until(w eq ' ' or w eq '@');
            w = scan(term, ii, ' ');
            n = input(w, ?? 23.);
            if n gt nvars then do;
               put 'ERROR: ' term +(-1) ' is invalid with ' nvars 'factors.';
               abort = 2;
               end;
            do jj = 1 to ii - 1 until(w2 eq ' ' or w2 eq '@');
               w2 = scan(term, jj, ' ');
               n2 = input(w2, ?? 23.);
               if n(n, n2) eq 2 and n eq n2 then invalid = 1;
               end;
            end;

         *------issue error if error checks failed------;
         if invalid then do;
            put 'ERROR: Invalid INTERACT= term: ' term +(-1) '.';
            abort = 2;
            end;

         %*------otherwise start processing list------;
         else do;
            *------star lists just get added to interacton list------;
            if star then intlist = trim(intlist) || ',' ||
                                   translate(hold, ' ', '*');

            %*------bar lists need more processing------;
            else do;

               *------figure out @ value, remove from list------;
               if atnum then do;
                  atnum = input(scan(term, atnum + 1, ' '), ?? 12.);
                  substr(term, at) = ' ';
                  l = ceil((l - 2) / 2);
                  end;

               %*------implicit @ is number of vars in term------;
               else do;
                  l = ceil(l / 2);
                  atnum = l;
                  end;

               %*------10 way interactions max------;
               if atnum le 1 or atnum gt 10 then do;
                  put 'ERROR: ' atnum +(-1)
                      '-way ineractions are not supported.';
                  abort = 2;
                  end;

               *------now generate interaction terms as matrix list------;
               term = compbl(translate(term, ' ', '|'));
               if not abort then do a = 2 to atnum;
                  %do i = 1 %to 10;
                     do ci&i = (ci%eval(&i-1) + 1) * (a ge &i) to l * (a ge &i);
                     %end;
                  intlist = trim(intlist) || ',' ||
                            trim(scan(term, ci[1], ' '));;
                  do b = 2 to a;
                     intlist = trim(intlist) || ' ' ||
                               trim(scan(term, ci[b], ' '));
                     end;
                  %do i = 1 %to 11; end; %end;
               end;
            end;
         end;
      end;

   l = length(intlist);
   do i = 1 to l until(nmiss(j));
      j = input(scan(intlist, i), ?? 24.);
      if n(j) then if j lt 1 or j gt nvars then abort = 2;
      end;

   if invalid or abort = 2 or compress(intlist, '0123456789,') ne ' ' then do;
      abort = 2;
      goto endit;
      end;

   *------count interactions, discard duplicate terms------;
   intlist = substr(intlist, 3);
   hold = compress(intlist, '0123456789 ');
   if intlist = ' ' then nints = 0;
   else if hold = ' ' then nints = 1;
   else nints = length(hold) + 1;

   *------sort terms in order within each interaction------;
   maxterms = 0;
   intlist = ',' || trim(intlist) || ',';
   l = length(intlist) - 1;
   do i = 1 to l;
      j = index(substr(intlist, i+1), ',') + i;
      term = substr(intlist, i + 1, j - i - 1);
      array _n[12] _temporary_;
      do k = 1 to length(term) until(w = ' ');
         w = scan(term, k, ' ');
         if w ne ' ' then do;
            if k gt 11 then do; abort = 2; goto endit; end;
            _n[k] = input(w, ?? 24.);
            end;
         end;
      k + -1;
      do l = 1 to k;
         do m = l + 1 to k;
            if _n[l] gt _n[m] then do; t = _n[l]; _n[l] = _n[m]; _n[m] = t; end;
            end;
         end;

      maxterms = max(maxterms, k);
      term = ' ';
      do l = 1 to k; term = compbl(term || put(_n[l], 5.)); end;
      substr(intlist, i + 1, j - i - 1) = trim(left(term));
      i = j - 1;
      end;

   call symput('maxterms', put(max(2, min(nvars, maxterms ** 2)), 12. -L));

   if gt1 then do;

      do i = 1 to nints while(i le nints);
         w = scan(intlist, i, ',');
         do j = i + 1 to nints while(j le nints);
            w2 = scan(intlist, j, ',');
            if w eq w2 then do; * locate and delete second term;
               nints + -1;
               w2 = ',' || trim(w2) || ',';
               l = length(w2);
               k = index(intlist, trim(w2));
               substr(intlist, k, l - 1) = ' ';
               m = index(intlist, trim(w2));
               substr(intlist, k, l - 1) = substr(w2, 1, l-1);
               intlist = substr(intlist, 1, m - 1) || substr(intlist, m + l - 1);
               j + -1;
               end;
            end;
         end;

      end;

   intlist = substr(intlist, 2, length(intlist) - 2);
   call symput('nints', compress(put(nints, 12.)));
   call symput('_mktintlist', trim(intlist));

   if nints gt 200 then
      put 'WARNING: ' nints 'interaction terms.  Expect slow processing.';

   if length(intlist) gt &maxlen * 0.9 or
      (intlist eq ' ' and symget('interact') ne ' ') then abort = 2;

   endit:
      if abort eq 2 then do;
         hold = symget('interact');
         put 'ERROR: INTERACT=' hold 'is invalid.';
         end;

      if _error_ or abort then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1;
%if &abort %then %let _mktlist = ABORT;
%if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;
filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outorth) ne %then %do;
   data _null_; set &outorth; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outtab) ne %then %do;
   data _null_; set &outtab; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )out=&out;
   %put %str(       )outorth=&outorth;
   %put %str(       )outtab=&outtab;
   %end;

%if &abort %then %goto endit;

data __tmprnums(drop=list term intlist w levs1-levs&maxterms
                i j __n nterms d a);
   length list term intlist $ &maxlen w $ 12 Reason $ 40;
   array terms[&maxterms];
   array levs[&maxterms];

   list    = symget('_mktlist');
   intlist = symget('_mktintlist');

   do j = 1 to &maxterms; terms[j] = .; end;

   *------Output each number of levels and the product of each level pair.
          Count the parameters in the saturated main-effects design.------;
   do i = 1 to &nvars;
      a = input(scan(list, i, ' '), ?? best12.);
      terms[1] = i; terms[2] = .;
      flag = 1;
      way = 0;
      reason = 'Balance';
      output; __n + 1;
      if __n gt &toobig then do; call symput('abort', '2'); stop; end;
      flag = 0;
      sat + a - 1;
      %if not &parse %then %do;
         do j = 1 to i - 1;
            terms[2] = j;
            way = 1;
            reason = 'Main-Effect';
            output; __n + 1;
            if __n gt &toobig then do; call symput('abort', '2'); stop; end;
            end;
         %end;
      end;

   *------take care of interactions------;
   do i = 1 to &nints;
      do j = 1 to &maxterms; terms[j] = .; end;
      term = scan(intlist, i, ',');

      do nterms = 0 to &maxterms until(w eq ' ');
         w = scan(term, nterms + 1, ' ');
         if w ne ' ' then do;
            terms[nterms + 1] = input(w, ?? best12.);
            levs [nterms + 1] = input(scan(list, terms[nterms + 1], ' '),
                                      ?? best12.);
            end;
         end;

      *------add interactions to list------;
      d = 1;
      do j = 1 to nterms; d = d * (levs[j] - 1); end;
      sat + d;
      way = nterms;
      reason = 'Interaction';
      output; __n + 1;
      if __n gt &toobig then do; call symput('abort', '2'); stop; end;

      %if not &parse %then %do;
         length hold $ &maxlen;
         drop hold ok k l;

         *------interactions times other single vars------;
         way = nterms + 1;
         do j = 1 to &nvars;
            ok = 1;
            reason = trim(put(j, 12. -L)) || ' x ' ||
                     translate(trim(term), '*', ' ');
            do k = 1 to nterms; if j eq terms[k] then ok = 0; end;
            if ok then do;
               terms[nterms + 1] = j;
               output; __n + 1;
               if __n gt &toobig then do; call symput('abort', '2'); stop; end;
               end;
            end;

         *------interactions with other interactions------;
         do j = 1 to i - 1;
            hold = scan(intlist, j, ',');
            extra = 0;
            do k = nterms + 1 to &maxterms; terms[k] = .; end;
            do k = 0 to &nvars until(w eq ' ');
               w = scan(hold, k + 1, ' ');
               if w ne ' ' then do;
                  a =  input(w, ?? best12.);
                  ok = 1;
                  do l = 1 to nterms + extra; if a eq terms[l] then ok = 0; end;
                  if ok then do;
                     extra + 1;
                     terms[nterms + extra] = a;
                     end;
                  end;
               end;

            if extra then do;
               way = nterms + k;
               reason = translate(trim(term), '*', ' ') ||
                        ' x ' || translate(trim(hold), '*', ' ');
               output; __n + 1;
               if __n gt &toobig then do; call symput('abort', '2'); stop; end;
               end;
            end;
         %end;
      end;

   call symput('_mktsat', compress(put(sat + 1, best12.)));
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1;
%if &abort = 2 %then %do;
   %put ERROR: Problem too big.;
   %put %str(       )Specify a larger TOOBIG= if you really want to do this.;
   %let abort = 1;
   %end;
%if &abort %then %goto endit;

*------post-process------;
data __tmprnums;
   length list $ &maxlen Variables Source $ 40 w tc z1-z&maxterms $ 24;
   retain maxx 2 list;
   array terms[&maxterms];
   array z[&maxterms];

   if _n_ = 1 then list = symget('_mktlist');

   set __tmprnums end=eof;

   *------sort variable nums------;
   do i = 1 to &maxterms;
      if n(terms[i]) then do j = 1 to i - 1;
         if terms[j] gt terms[i] then do;
            t = terms[i]; terms[i] = terms[j]; terms[j] = t;
            end;
         end;
      end;

   *-------manufacture sources of violations------;
   source    = ' ';
   variables = ' ';
   x = 1;
   nterms = 0;
   do i = 1 to &maxterms;
      if n(terms[i]) then do;
         nterms + 1;
         w = scan(list, terms[i], ' ');
         x = x * input(w, ?? 12.);
         z[i] = left(w);
         variables = trim(variables) || ' ' || put(terms[i], 12. -L);
         end;
      else leave;
      end;

   *------sort levels------;
   do i = 1 to nterms;
      if n(z[i]) then do j = 1 to i - 1;
         if z[j] gt z[i] then do;
            tc = z[i]; z[i] = z[j]; z[j] = tc;
            end;
         end;
      end;

   *------manufacture source------;
   do i = 1 to nterms;
      if n(z[i]) then source = trim(source) || '*' || z[i];
      end;

   source    = substr(source   , 3);
   variables = substr(variables, 2);
   maxx      = max(maxx, x);

   if eof then call symput('levformat', compress(put(ceil(log10(maxx + 1)) + 1,
                                                     best12.)));
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get rid of duplicate terms------;
%if not (&multiple or &multiple2) %then %do;
   proc sort data=__tmprnums
      out=__tmprnums(keep=x flag source variables) nodupkeys;
      by x nterms z1-z&maxterms terms1-terms&maxterms;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------report of sources of nonorthogonality------;
%if &source and not (&multiple or &multiple2) %then %do;
   data __tmprlist1(keep=n);
      set __tmprnums;
      by x notsorted source;
      if first.source then n = 0;
      n + 1;
      if last.source then output;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data __tmprnums;
      set __tmprnums;
      by x notsorted source;
      if first.source then do; k + 1; set __tmprlist1 point=k; end;
      else n = .;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %if &print %then %do;
      options missing=' ';
      proc print label noobs split='-' data=__tmprnums;
         label x = 'To Achieve Both-Orthogonality  -and Balance,   -
N Must Be      -Divided By     -'
               source = 'The Source of   -the Divisor     - -
Number of       -Levels or Cells -'
               variables = 'The Divisor-Comes     -from these-Factors    -'
               n = 'Failing to   -Divide by    -the Divisor  -Contributes  -
this to the  -Violations   -Count        - -The Number of-Sources of   -
the Divisor  -';
         var n variables;
         by x notsorted source;
         id x source;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      options missing='.';
      %end;
   %end;

%else %if &source and &multiple %then %do;
   proc summary data=__tmprnums noprint nway missing;
      class x terms1-terms&maxterms source variables;
      output out=__tmprnum2;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %if &print %then %do;
      proc print label noobs split='-';
         label x = 'N Must Be -Divided By' _freq_ = 'Frequency';
         var variables _freq_;
         by x notsorted source;
         id x source;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;
   %end;

%else %if &source and &multiple2 %then %do;
   proc sort data=__tmprnums out=__tmprnum2;
      by x terms1-terms&maxterms way reason;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc print label noobs split='-';
      label x = 'N Must Be -Divided By';
      var reason;
      by x notsorted source notsorted variables;
      id x source variables;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------report on numbers of levels------;
proc summary data=__tmprnums(where=(flag=1)) noprint;
   class x;
   output out=&out(where=(_type_=1));
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data _null_;
   file print;
   retain ind 24;
   set &out end=eof;
   logff + log10(x) * _freq_;

   %if &print and not &parse %then %do;
      if _n_ = 1 then do;
         ind = (input(getoption('linesize'), ?? 12.) - 22) / 2;
         %if &print %then %do;
            put                                /
                @ind '    Design Summary    '  //
                @ind 'Number of'               /
                @ind 'Levels       Frequency'  /;
            %end;
         end;
      %if &print %then %do;
         put    @ind x 6. _freq_ 12.;
         %end;
      %end;
   if eof then do;
      if      logff < 15 then ffstr = put(10 ** logff, comma20.0 -L);
      else if logff < 38 then ffstr = put(10 ** logff, best10.   -L);
      else                    ffstr = '1E' || put(logff, 6. -L);
      call symput('_mktff', trim(ffstr));
      call symput('_mktlogff', trim(put(logff, best16. -L)));
      end;

   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1;
%if &abort or &parse %then %goto endit;

*------sort and count values that must divide n------;
proc freq data=__tmprnums noprint;
   tables x / out=__tmprnums(drop=percent);
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------count and list sources of orthogonality and balance violations------;

data &out(keep=v n l);
   length n v 8 l range $ 200;
   stopit = &_mktsat + &max - 1; startit = &_mktsat; maxn = 0;
   %if %nrbquote(&n) ne %then %do; stopit = &n; startit = &n; %end;
   do n = startit to stopit while(n <= stopit);
      v = 0; c = 0; l = ' ';
      do i = 1 to nobs;
         set __tmprnums point=i nobs=nobs;
         r = n / x;
         if r ne round(r) then do;
            v + count; c + 1;
            if c * &levformat + 5 < 200
               then l = trim(l) || put(x, &levformat..);
            else if c < 200 then do;
               substr(l, c * &levformat - 1) = '...';
               c = 200;
               end;
            end;
         end;
      if v eq 0 and vcount lt 10 then do;
         vcount + 1;
         stopit = min(&stop * n, stopit);
         %if &fullfacstop %then %do; if log10(n) le &_mktlogff then do; %end;
         if range eq ' ' then range = 'n = ' || put(n, 12. -L);
         else range = trim(range) || ' or n = ' || put(n, 12. -L);
         maxn = max(maxn, n);
         %if &fullfacstop %then %do; end; %end;
         end;
      l = substr(l, 2);
      output;
      end;

   if maxn gt 100 then maxn = 0;
   call symput('maxn', put(maxn, 12.0 -L));
   call symput('range', trim(range));
   call symput('needlist', put(range ne ' ' and symget('interact') eq ' ', 1.));
   label v = 'Violations'
         n = '-Some Reasonable-Design Sizes'
         s = '00'x
         l = 'Cannot Be-Divided By';
   if _error_ then call symput('abort', '1');
   stop;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get a list of orthogonal designs MktEx can make------;
%if &needlist %then %do;
   %let _mktver = obsolete;

   %mktorth(outlev=__tmprlist1,
            options=mktruns %if &adx %then adx lineage; %if &do512 %then 512;,
            range=&range %if &maxn %then , maxn=&maxn; , maxlev=&maxlev);

   %if &_mktver ne &localver %then %do;
      %put ERROR: Incompatible MktOrth macro.;
      %put %str(      ) You must have the same version of all macros.;
      %put %str(      ) Versions: MktRuns = &localver, MktOrth = &_mktver..;
      %put %str(      ) Get the latest tools and documentation from:;
      %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
      %let abort = 1;
      %end;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------make list of designs with zero violations------;
   data __tmprlist1(keep=n design reference %if &adx %then x: lineage;);
      retain list;
      length list $ &maxlen w $ 12;
      array x[&maxlev];
      array t[&maxlev];
      if _n_ = 1 then list = symget('_mktlist');
      merge &out(where=(v = 0) in=a) __tmprlist1(in=b);
      by n;
      if a and b;
      do i = 1 to &maxlev; t[i] = x[i]; end;
      ok = 1;
      do i = 1 to &maxlen until(w eq ' ');
         w = scan(list, i, ' ');
         if w ne ' ' then do;
            a = input(w, ?? best12.);
            if a le &maxlev then do;
               if t[a] le 0 then do;
                  ok = 0;
                  end;
               t[a] + -1;
               end;
            end;
         end;
      if ok then output;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc sort nodupkeys out=__tmprnums; by n; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------add asterisk, footer for designs MktEx can make------;
data &out(keep=n s v l);
   length n 8 s $ 2 v 8 l $ 200;
   %if &needlist %then %do; merge &out __tmprnums(in=a); by n; %end;
   %else               %do; set &out; %end;

   %if %sysevalf(&_mktlogff < 15) %then %do;
      if abs(n - 10 ** &_mktlogff) < 0.99 then a = 1;
      %end;

   if a then do;
      call symput('foot_a', '1');
      s = '*';
      call symput('_mktexists', '1');
      end;
   if n eq &_mktsat then do;
      call symput('foot_s', '1');
      s = left(compress(s || 'S'));
      end;
   if s ne ' ' then do;
      __f = input(symget('foot'), 1.);
      __f = max(1 + length(s), __f);
      call symput('foot', put(__f, 1.));
      end;

   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------find the best few sizes------;
proc sort data=&out; by v n; run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

data __tmprout(drop=__n __o);
   set &out;
   __o = 0;
   if &fullfacstop and log10(n) gt &_mktlogff + 1e-8 then return;
   if __n lt 10 then __o = 1;
   if __o %if &showsat %then or n eq &_mktsat; then output;
   if __o then __n + 1;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get ready to print results with line wrapping if necessary------;
data %if %nrbquote(&outtab) eq %then _null_; %else &outtab;;
   set __tmprout end=eof;
   retain maxstrlen minlead 10 maxn 0;
   maxn = max(maxn, n);
   if l ne ' ' then do;
      len       = length(l);
      maxstrlen = max(maxstrlen, len);
      minlead   = min(minlead, len - length(left(l)));
      end;
   if eof then do;
      call symput('maxn',      trim(put(maxn,        best12. -L)));
      call symput('maxstrlen', trim(put(maxstrlen,   best12. -L)));
      call symput('adj',       trim(put(minlead - 1, best12. -L)));
      end;
   %if %nrbquote(&outtab) ne %then %do;
      keep n s v l;
      label v = 'Violations'
            n = '-Some Reasonable-Design Sizes'
            s = '00'x
            l = 'Cannot Be-Divided By';
      %end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------print the violations table------;
data _null_;
   length str1 $ &maxlen;
   set __tmprout end=eof;
   file print;

   if _n_ eq 1 then do;
      ls    + input(getoption('linesize'), ?? 12.);
      sp    + 2;
      extra + ls - 35 - &foot - 2 * sp;
      llen  + min(max(&maxstrlen, 10), 10 + extra);
      llen  = ceil(llen / &levformat) * &levformat;
      do until(extra ge 0);
         extra = ls - 25 - &foot - llen - 2 * sp;
         if extra lt 0 then llen + -&levformat;
         end;
      do k = 1 to 3; if extra ge 6 then do; extra + -2; sp + 1; end; end;
      extra = floor(extra / 2);
      adj   + (&maxstrlen le llen) * &adj + 1 - sp;

      %if &print %then %do;
         put / +extra "Saturated      = &_mktsat"
             / +extra "Full Factorial = &_mktff";
         k = &foot + sp;
         m = &foot + 10 + 2 * sp;
         put / +extra 'Some Reasonable' +m                  'Cannot Be'
             / +extra '   Design Sizes' +k 'Violations' +sp 'Divided By' /;
         %end;
      end;

   str1 = substr(l, 1, llen);
   m    = length(str1) - length(left(str1)) - adj;
   if length(s) = 2 then sp + -1;
   %if &print %then %do;
      put +extra n 15. %if &foot %then +1 s +(-1); +sp v 10. +m str1;
      %end;
   if length(s) = 2 then sp + 1;
   do while(length(l) gt llen);
      l    = substr(l, llen + 1);
      str1 = substr(l, 1, llen);
      m    = length(str1) - length(left(str1)) - adj + extra + 25 + sp + &foot;
      %if &print %then %do;
         put +m str1;
         %end;
      end;

   if eof then do;
      m = (ls - 59) / 2;
      if &foot_a or &foot_s then put;
      if &foot_a and &print then do;
         put +m '* - 100% Efficient design can be made with the MktEx macro.';
         end;
      if &foot_s and &print then do;
         put +m 'S - Saturated Design - The smallest design that can be made.';
         if s eq 'S' then
            put +m '    Note that the saturated design is not one of the    ' /
                +m '    recommended designs for this problem.  It is shown  ' /
                +m '    to provide some context for the recommended sizes.  ';
         end;
      end;

   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%if &needlist and &print %then %do;
   proc print data=__tmprlist1(%if %nrbquote(&maxoa) ne %then obs=&maxoa;
                               where=(n le &maxn)) noobs uniform;
      var n design reference;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;


%if &needlist and %nrbquote(&outorth) ne %then %do;
   data &outorth;
      set __tmprlist1;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %end;

proc datasets nolist; delete __tmprnums __tmprlist1 __tmprnum2 __tmprout; run; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

options &savenote no&timeopt;

%if &source eq 0 %then %put
NOTE: Specify OPTIONS=SOURCE for an explanation of the design sizes table.;

%if %nrbquote(&outorth) ne %then %do;
   %if &needlist %then %do;
      data &outorth;
         set &outorth;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;
   %else %do; data &outorth; stop; run; %end;
   %end;

%if %nrbquote(&outtab) ne %then %do;
  data &outtab; set &outtab; if _error_ then call symput('abort', '1'); run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%if %nrbquote(&out) ne __tmprout %then %do;
  data &out; set &out; if _error_ then call symput('abort', '1'); run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%endit: options &saveopts;
quit;

%if &abort and not &parse %then %do;
   %if &syserr = 1016 or &syserr = 116  %then %put ERROR: Insufficient memory.;
   %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %put ERROR: The MKTRUNS macro ended abnormally.;
   %end;

%if %sysfunc(getoption(&timeopt)) eq &timeopt and
   %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %put NOTE: The MKTRUNS macro used &time seconds.;
   %end;

%mendit: %mend;
