 /****************************************************************/
 /*          S A S   A U T O C A L L   L I B R A R Y             */
 /*                                                              */
 /*    NAME: MktEx                                               */
 /*   TITLE: Create Factorial Experimental Designs               */
 /* PRODUCT: STAT                                                */
 /*  SYSTEM: ALL                                                 */
 /*    KEYS: marketing research, designs                         */
 /*   PROCS: IML FACTEX PLAN OPTEX                               */
 /*    DATA:                                                     */
 /* SUPPORT: saswfk                      UPDATE:  25Jul2010      */
 /*     REF: See "Marketing Research Methods in SAS" at the      */
 /*          following URL for more information on this macro.   */
 /*                                                              */
 /* http://support.sas.com/techsup/tnote/tnote_stat.html#market  */
 /*                                                              */
 /*    MISC: This macro creates global macro variables           */
 /*          _MktList _MktEff _MktSat _MktLogFF _MktFF _MktVer   */
 /*          _MktIntList _MktTime1                               */
 /*          and calls the following autocall macros, which must */
 /*          be installed:                                       */
 /*             MktRuns MktDes MktOrth                           */
 /*          The _MktEff macro contains the efficiency of the    */
 /*          final design.                                       */
 /*                                                              */
 /*          The following SAS products are required to run      */
 /*          this macro: BASE, STAT, QC, IML.                    */
 /****************************************************************/

 /*---------------------------------------------------------------------

The MktEx macro (pronounced MarkTex) generates factorial experimental
designs.  While it is part of the marketing research family of macros,
it was developed for any one who wants to make efficient factorial
designs.  The MktEx family includes MktDes, MktEval, MktRuns, MktBIBD,
MktBlock, MktDups, MktBal, MktOrth, and ChoicEff, which help you find
and evaluate experimental designs.  This macro is designed to be very
simple to use, and to run in seconds for trivial problems, minutes for
small problems, and in under an hour for larger and difficult problems.
The MktTex macro is a full-featured linear designer that can handle
simple problems like main-effects designs and more complicated problems
including designs with interactions and restrictions on which levels can
appear together.  The macro is particularly designed to easily create
the kinds of linear designs that marketing researchers need for conjoint
and choice experiments.  It does not create for example Latin Squares,
randomized blocks, response surface, resolution IV, or other specialized
designs that are not as widely used in marketing research.

Here is an example of using the MktEx macro to create a design with 5
two-level factors, 4 three-level factors, 3 five-level factors, 2
six-level factors, all in 60 runs (60 rows or conjoint profiles or
choice sets).

   %mktex( 2 ** 5  3 ** 4  5 5 5  6 6,  n=60 )

The notation "m ** n" means n m-level factors.  For example "2 ** 5"
means 5 two-level factors.  For almost all linear design problems, you
should be able to simply run the macro once, specifying only the levels
of all the factors and the number of runs and get a good design - no
more "try this, try that, now try this" iterations on your part.  The
macro does all of that for you.

The MktEx macro creates efficient linear experimental designs using
several approaches.  Depending on the problem, the macro may try to
create an orthogonal array (OA) design, it will search a set of
candidate runs (rows of the design), it will use a coordinate exchange
algorithm using both random initial designs, and if appropriate, a
partial OA design initialization.  The coordinate-exchange algorithm is
based on Meyer and Nachtsheim (1995).  This first phase is the algorithm
search phase.  In it, the macro determines which approach is working
best for this problem.  It stops if at any time it finds a perfect, 100%
efficient, orthogonal and balanced design.  Otherwise, at the end of
this phase, the macro chooses the method that has produced the best
design and performs another set of iterations using exclusively the
chosen approach.  Finally, the macro performs a third set of iterations
where it takes the best design found so far, and tries to improve it.
It can also directly create many 100% efficient designs.

The candidate set search approach has two phases.  First, either PROC
PLAN is run to create a full-factorial design for small problems, or
PROC FACTEX is run to create a fractional-factorial design for large
problems.  Either way, this design is a candidate set that is searched
by PROC OPTEX using the modified Federov algorithm, and a final design
is built from a selection of the rows of the candidate set (Fedorov
(1972), Cook and Nachtsheim (1980)).

In all methods, the macro attempts to optimize D-efficiency (also known
as D-optimality), which is a standard measure of the goodness of the
experimental design.  As D-efficiency increases, the standard errors of
the parameter estimates in the linear model decrease.  A perfect design
is orthogonal and balanced and has 100% efficiency.  A design is
orthogonal when all of the parameter estimates are uncorrelated.  A
design is balanced when all of the levels within each of the factors
occur equally often.  A design is orthogonal and balanced when the
variance matrix, which is proportional to inv(X`X) is diagonal, where
X is a suitable orthogonal coding of the design matrix.  See Kuhfeld,
Tobias, and Garratt (1994), available from
http://support.sas.com/techsup/tnote/tnote_stat.html#market
for more information on efficient experimental designs.

For most problems, you only need to specify the levels of all the
factors and the number of runs.  For more complicated problems, you may
need to also specify the interactions that you want to be estimable or
restrictions on which levels may not appear together.  Other than that,
you should not need any other options.  For advanced users, there are
options that control the details of what the macro does.  Note however,
that this macro is not like other design tools that you have to tell
what to do.  With this macro, you just tell it what you want, and it
figures out a good way to do it.  For some problems, the sophisticated
user, with a lot of work, will be able to adjust the options to come up
with a better design.  However, this macro should always produce a very
good design for even the most unsophisticated users, with minimal
effort.

For certain designs, you can just specify N=, which specifies the number
of runs or rows in the design, and the macro will give you a 100%
efficient design.  Certain OA's, fractional and full-factorial designs,
and Hadamard designs can be requested just by specifying N=.  Here are
some of the designs up through n=100, and a few larger OA's that can be
obtained by just specifying N=.

         N    Design

         4    2 ** 3             Hadamard
         8    2 ** 7             Hadamard
         9    3 ** 4             Fractional-factorial
        12    2 ** 11            Hadamard
        16    2 ** 15            Hadamard
        18    2 3 ** 7           OA
        20    2 ** 19            Hadamard
        24    2 ** 23            Hadamard
        25    5 ** 6             Fractional-factorial
        27    3 ** 13            Fractional-factorial
        28    2 ** 27            Hadamard
        32    2 ** 31            Hadamard
        36    2 ** 11 3 ** 12    OA
        40    2 ** 39            Hadamard
        44    2 ** 43            Hadamard
        48    2 ** 11 4 ** 12    OA
        49    7 ** 8             Fractional-factorial
        50    2 5 ** 11          OA
        52    2 ** 51            Hadamard
        54    2 3 ** 25          OA
        56    2 ** 55            Hadamard
        60    2 ** 59            Hadamard
        64    2 ** 63            Hadamard
        68    2 ** 67            Hadamard
        72    2 ** 23 3 ** 24    OA
        76    2 ** 75            Hadamard
        80    2 ** 79            Hadamard
        81    3 ** 40            Fractional-factorial
        84    2 ** 83            Hadamard
        88    2 ** 87            Hadamard
        92    2 ** 91            Hadamard
        96    2 ** 95            Hadamard
       100    2 ** 99            Hadamard
       108    2 ** 35 3 ** 36    OA
       144    2 ** 88 3 ** 25    OA
       162    2 3 ** 79          OA
       192    2 **144 3 4 ** 13, OA
       216    2 ** 71 3 ** 72    OA
       288    2 ** 232 3 ** 25   OA
       324    2 ** 35 3 ** 144   OA
       432    2 ** 143 3 ** 144  OA
       486    2 3 ** 241         OA

The macro will also output certain full-factorial designs with just a
specification of N=.  For example N=75 implies 3 5 ** 2.  The N= value
is factored into a list of prime factors, and the full-factorial for
that list is created.

To get a list of all of the 100% efficient designs that the macro
can directly create, run the following.

   %mktorth(maxlev=144)

Running the above step can take several minutes, and the full data set
produced by the MktOrth macro is large, so you typically do not want to
generate and print the entire thing.  Here is an example of how to print
designs with 72 or fewer runs and no more than 6-level factors.
This step should run in a few seconds.

   %mktorth(range=n le 72, maxlev=6)
   proc print; run;

The orthogonal arrays are constructed using methods and arrays from a
variety of sources.  See http://support.sas.com/techsup/technote/ts723.html
and more specifically the reference list at the end of
http://support.sas.com/techsup/technote/ts723.pdf for more information.

Often, when N=n is a multiple of 4, the MktEx macro can construct
orthogonal designs with up n-1 two-level factors.  The two-level designs
are constructed from Hadamard matrices.  The catalog contains a Hamard
matrix for every n up through 448, with good coverage beyond 448.  For
some of these sizes, the macro can create orthogonal designs with a
small number (say m) four-level factors in place of 3*m of the two-level
factors, or an eight-level factor in place of 7 two-level factors, or a
sixteen-level factor in place of 15 two-level factors, and so on.

------------------------------------------------------------------------

This example requests the Taguchi (1987) L36 design, which has 11
two-level factors and 12 three-level factors.

   %mktex( n=36 )

No iterations are needed, and the macro immediately creates the L36,
which is 100% efficient.  This example runs in a few seconds.  Here is
the last part of the output.

                                                                  Average
                                                                 Prediction
   Design                                                         Standard
   Number     D-Efficiency     A-Efficiency     G-Efficiency       Error
   ------------------------------------------------------------------------
        1       100.0000         100.0000         100.0000          1.0000

Here are the first few rows and columns of the design.  The factors are
always named x1, x2, ... and the levels are always integers starting
with 1.

 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15 x16 x17 x18

  1  1  1  2  1  1  2  1  2  2   2   1   2   2   3   2   1   3
  1  1  1  2  1  1  2  1  2  2   2   2   3   3   1   3   2   1
  1  1  1  2  1  1  2  1  2  2   2   3   1   1   2   1   3   2
  1  1  2  1  1  2  1  2  2  2   1   1   2   1   2   3   3   1  . . .
  1  1  2  1  1  2  1  2  2  2   1   2   3   2   3   1   1   2
  1  1  2  1  1  2  1  2  2  2   1   3   1   3   1   2   2   3
  1  1  2  1  2  2  2  1  1  1   2   1   1   2   1   3   2   3
  1  1  2  1  2  2  2  1  1  1   2   2   2   3   2   1   3   1
                             .
                             .
                             .

By default the macro creates two output data sets.

   OUT=Design      - the experimental design, sorted by the factor levels
   OUTR=Randomized - the randomized experimental design

The OUT=Design data set is shown above.  The designs are equivalent.  At
least for small problems, the sorted design is usually nicer to look
at, however the randomized design is the better design to use.  The
randomized design has the rows sorted into a random order, and all of
the factor levels are randomly reassigned.  So for example with
two-level factors, approximately half of the original (1, 2) mappings
will be reassigned (2, 1).  Similarly, with three level factors, the
mappings (1, 2, 3) will be changed to one of the following: (1, 2, 3),
(1, 3, 2), (2, 1, 3), (2, 3, 1), (3, 1, 2), or (3, 2, 1).  The
reassignment of levels is usually not critical for the iteratively
derived designs, but it can be very important for some of the OA's have
all 1's in the first row.

We could run the MktEval macro to see canonical correlations between the
factors and frequency information.

   %mkteval;

Here are some of the results.  All factors are uncorrelated, all levels
occur equally often within each factor, all two-way frequencies are
equal within each pair of factors, and each run occurs exactly once.

                  Canonical Correlations Between the Factors
            There are 0 Canonical Correlations Greater Than 0.316

                      x1   x2   x3   x4   x5   x6   x7   x8

                  x1  1    0    0    0    0    0    0    0
                  x2  0    1    0    0    0    0    0    0
                  x3  0    0    1    0    0    0    0    0
                  x4  0    0    0    1    0    0    0    0  . . .
                  x5  0    0    0    0    1    0    0    0
                  x6  0    0    0    0    0    1    0    0
                  x7  0    0    0    0    0    0    1    0
                  x8  0    0    0    0    0    0    0    1
                                          .
                                          .
                                          .

                            Summary of Frequencies
            There are 0 Canonical Correlations Greater Than 0.316

                            Frequencies

                 x1         18 18
                 x2         18 18
                 x3         18 18
                 x4         18 18
                 x5         18 18
                 x6         18 18
                 x7         18 18
                 x8         18 18
                 x9         18 18
                 x10        18 18
                 x11        18 18
                 x12        12 12 12
                 x13        12 12 12
                 x14        12 12 12
                 x15        12 12 12
                 x16        12 12 12
                 x17        12 12 12
                 x18        12 12 12
                 x19        12 12 12
                 x20        12 12 12
                 x21        12 12 12
                 x22        12 12 12
                 x23        12 12 12
                 x1 x2      9 9 9 9
                 x1 x3      9 9 9 9
                 .
                 .
                 .
                 x1 x12     6 6 6 6 6 6
                 x1 x13     6 6 6 6 6 6
                 .
                 .
                 .
                 x12 x13    4 4 4 4 4 4 4 4 4
                 x12 x14    4 4 4 4 4 4 4 4 4
                 .
                 .
                 .
                 N-Way      1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
                            1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1

------------------------------------------------------------------------

This call creates an orthogonal design with 103 two-level factors in 104
runs.

   %mktex( n=104 )

This example runs in a few seconds.  The design is based on a Hadamard
matrix.  If you want to actually create a Hadamard matrix, you will need
to reassign levels from {1 2} to {-1 1} (using LEVELS=i, where "i" means
integer levels that are centered) and add an intercept column using
OPTIONS=int.  Here is an example in 12 runs

   %mktex(n=12, options=nosort int render +-, levels=i)

Here is the Hadamard matrix, rendered as a series of plusses and
minuses.  The actual OUT= data set contains ones and minus ones.

   ++++++++++++
   +-+-+++---+-
   +--+-+++---+
   ++--+-+++---
   +-+--+-+++--
   +--+--+-+++-
   +---+--+-+++
   ++---+--+-++
   +++---+--+-+
   ++++---+--+-
   +-+++---+--+
   ++-+++---+--

------------------------------------------------------------------------

This next example illustrates some of the things that the macro tries to
find a good design.

   %mktex( 2 2 2 3 3 3 4, n=36, seed=151 )

This example requests 3 two-level factors, 3 three-level factors, and a
four-level factor.  Perfect balance and orthogonality are not both
possible since we have both two-level and four-level factors (36 is not
divided by 2x4=8), hence 100% efficiency is not possible (we could run
the macro %mktruns( 2 2 2 3 3 3 4 ) to see this).

Some of the results appear next.  Here is part of first part of the
iteration history.

                            Algorithm Search History

                                   Current          Best
           Design    Row,Col  D-Efficiency  D-Efficiency  Notes
           ----------------------------------------------------------
                1      Start       99.5674       99.5674  Can
                1        End       99.5674

                2      Start       98.2043                Tab,Unb
                2        End       99.1797

                3      Start       98.2043                Tab,Unb
                3        End       99.1797

                4      Start       94.5823                Tab,Ran
                4        End       99.4210

                .
                .
                .

               13      Start       89.6474                Ran,Mut,Ann
               13      6   1       99.6149       99.6149
               13     14   6       99.6149       99.6149
               13     35   7       99.6149       99.6149
               13        End       99.5055

                .
                .
                .

               21      Start       86.6033                Ran,Mut,Ann
               21        End       99.0585

We see in the Notes column for the first design "Can" which means that
this design was created by searching a candidate set with PROC OPTEX.
No level-swapping improvements are tried on OPTEX designs during the
algorithm search phase.

We see in the Notes column for the second design "Tab,Unb" which means
that this design was created by using an OA (or tabled design 'Tab') to
initialize some of the factors and an unbalanced initialization of the
remaining factors.  In this case, the two-level and three-level factors
are initialized with the orthogonal array and the four-level factor is
initialized by collapsing levels of a six-level factor into a four-level
factor.  The macro tries to improve the four-level factor with level
swaps.

We see in the Notes column for the fourth design "Tab,Ran" which means
that this design was created by using an OA (or tabled design 'Tab') to
initialize some of the factors and a random initialization of the
remaining factors.  In this case, the two-level and three-level factors
are initialized with the orthogonal array and the four-level factor is
randomly initialized.  The macro tries to improve the four-level factor
with level swaps.

The twelfth through twenty-first designs have a fully random
initialization (Ran) and use random mutations of the initial design
(Mut) and simulated annealing (Ann).  Information is only added to the
table when at the start and end of each design and when a better design
is found.  The Best column only has values when a better designs is
found.  The best design is found in iterations 2 and 3, using an
unbalance OA initialization.

Level swapping works like this.  Imagine we have 2 two-level factors and
4 runs.  Normally, the macro would directly generate the full-factorial
for this problem without any iteration, but for pedagogical purposes, we
will look at what the level exchange algorithm does with this problem.
Imagine the initial design is as shown in the left partition of the table.

         Initial     After        After
         Design      Step 1       Step 2
         1  1         2* 1         2  2*
         1  2         1  2         1  2
         2  1         2  1         2  1
         1  1         1  1         1  1

         * - changed in this iteration

The optimal design for this situation has all four pairs of 1 and 2.
You can immediately see that the initial design is nonoptimal since it
has two pairs of (1, 1) and no (2, 2) combination.  The initial design
has D-efficiency 79.3701.  The first step considers swapping the level
of the first factor in the first row from 1 to 2.  This has no effect on
D-efficiency.  The duplication of (1, 1) is replaced by the duplication
of (2, 1).  Ignoring annealing for a moment (described below), the macro
will do a swap even if it does not help efficiency, as long as it does
not hurt it.  The table above shows the result after the first swap.
Next, the macro step considers swapping the level of the second factor
in the first row from 1 to 2.  This changes efficiency to 100.000 and
creates the optimal design.  All other swaps (first factor second row,
second factor second row, first factor third row, ...) will decrease
efficiency so will not be done.  Once the macro has processed all n
rows, it stops iterating and outputs the 100% efficient design.

Random mutations add random noise to the initial design before
iterations start.  This may eliminate the perfect balance that will
often be in the initial random design.  Simulated annealing allows
within each design step, that the design may get worse occasionally but
with decreasing probability as the number of swaps increases.  So for
design 1, for the first level of the first factor, by default, the macro
may execute a swap (say change a 2 to a 1) with probability 0.05 that
makes the design worse.  As more and more swaps occur, this probability
decreases so at the end of the processing of design 1, swaps that
decrease efficiency are hardly ever done.  Then for design 2, this same
process is repeated, again starting by default with an annealing
probability of 0.05.  This often helps the algorithm overcome local
efficiency maxima.  To envision, this, imagine that you are standing on
top of a pebble that sits six feet from the base of Mount Everest.  The
only way you can start going up the mountain is to first step down off
the pebble.  Simulated annealing, by occasionally stepping down the
efficiency function, often allows the macro to go farther up than it
would otherwise.  Consider again the problem of 2 two-level factors and
4 runs, with the same initial design as before, but with the rows in a
different order.

         Initial
         Design
         1  2
         1  1
         2  1
         1  1

Without annealing, the algorithm will never find the optimal design from
this starting point.   It just keeps swapping from one locally optimal
design to another.  With annealing, it finds it very quickly.

Here is part of the next part of the output for the problem:

   %mktex( 2 2 2 3 3 3 4, n=36, seed=151 )

It consists of more iterations using the best method as determined by
the algorithm search step.

                             Design Search History

                                   Current          Best
           Design    Row,Col  D-Efficiency  D-Efficiency  Notes
           ----------------------------------------------------------
                0    Initial       99.6149       99.6149  Ini

                1      Start       83.2691                Ran,Mut,Ann
                1        End       99.3236

                .
                .
                .

                5      Start       85.3893                Ran,Mut,Ann
                5      1   1       99.6161       99.6161
                5        End       99.2064

                .
                .
                .

               25      Start       86.2724                Ran,Mut,Ann
               25        End       99.4917

Next, the macro takes the best design it found in the previous step, and
tries to improve it using random mutations and simulated annealing.
Design 0 is the initial design (Ini), and its efficiency matches the
maximum printed above.  Then we have ten sets of iterations trying to
improve this initial design.  Each of these iterations starts with a
mutated version of the previous best design (Pre).  The starting designs
are often worse than the initial designs due to the random mutations.

                           Design Refinement History

                                   Current          Best
           Design    Row,Col  D-Efficiency  D-Efficiency  Notes
           ----------------------------------------------------------
                0    Initial       99.6161       99.6161  Ini

                1      Start       99.6161       99.6161  Pre,Mut,Ann
                1     10   3       99.6161       99.6161
                1        End       99.3445

                2      Start       99.6161       99.6161  Pre,Mut,Ann
                2      3   6       99.6161       99.6161
                2        End       99.3245

                3      Start       99.6161       99.6161  Pre,Mut,Ann
                3     27   1       99.7125       99.7125
                3     28   3       99.7125       99.7125
                3     30   5       99.7125       99.7125
                3     13   4       99.7125       99.7125
                3     20   7       99.7125       99.7125
                3      8   4       99.7125       99.7125
                3        End       99.3636

                4      Start       97.7482                Pre,Mut,Ann
                4        End       99.3531

                5      Start       98.4235                Pre,Mut,Ann
                5        End       99.3477

                6      Start       99.7125       99.7125  Pre,Mut,Ann
                6     25   3       99.7125       99.7125
                6        End       99.1837

                .
                .
                .

In this case, MktEx was able to slightly improve the previous best
design.  Here is the last bit of output.  It is a PROC OPTEX report on
the factor names and levels and the final D-efficiency.

                              The OPTEX Procedure

                            Class Level Information

                             Class  Levels  Values

                             x1       2     1 2
                             x2       2     1 2
                             x3       2     1 2
                             x4       3     1 2 3
                             x5       3     1 2 3
                             x6       3     1 2 3
                             x7       4     1 2 3 4

                                                                   Average
                                                                  Prediction
    Design                                                         Standard
    Number     D-Efficiency     A-Efficiency     G-Efficiency       Error
    ------------------------------------------------------------------------
         1        99.7125          99.4216          97.1228          0.6009

The macro found a very good design with perfect balance and almost
perfect orthogonality (we could run the MktEval macro to see this).

------------------------------------------------------------------------

This next example is more complicated in two ways.  It is the consumer
food product example from Kuhfeld, Tobias, and Garratt, (1994).  This
example has interactions.  The interactions between factors 2 and 3, 2
and 4, 3 and 4, and 6 and 7 must be estimable.  Furthermore, this is an
availability design where the maximum level of the price factor flags
the unavailable alternative.  The design has restrictions such that the
number of available brands (the number of brands whose price is less
than the maximum) must be between 2 and 4.  This means choice sets with
0, 1, or 5 available brands must be excluded from consideration.  This
is accomplished by specifying RESTRICTIONS=macro-name in the MktEx macro
and writing a macro that quantifies the badness of each row in the
design.  This macro first counts the number of available brands using
addition and Boolean algebra.  So when the first brand is available, its
price factor x1 has a level less than the maximum (or 4 for the first
price factor).  The Boolean expression (x1 < 4) is either true (1) or
false (0).  The sum of the Boolean expressions is the number of
available brands.  When the number of available brands is between two
and four, the Boolean expression ((navail < 2) | (navail > 4)) is false
(0).  Otherwise, bad is set to a nonzero value indicating that this
choice set is to be excluded.  With 5 brands available, bad = 2; with 1
brand available, bad = 2; and with zero brands available, bad = 3.  Bad
could simply be set to 0 for OK choice sets and 1 for unacceptable
choice sets, but the fact that bad was set up to distinguish between one
available (bad) and zero available (worse) helps the macro go in the
right direction.

   %macro bad;
     navail = (x1 < 4) + (x2 < 4) + (x5 < 3) + (x6 < 3) + (x8 < 3);
     bad = abs(navail - 3) * ((navail < 2) | (navail > 4));
     %mend;

   %mktex( 4 4 2 2 3 3 2 3, n=26, interact=x2*x3 x2*x4 x3*x4 x6*x7,
           restrictions=bad, seed=377 )

Here is the last part of the output.

                              The OPTEX Procedure

                            Class Level Information

                             Class  Levels  Values

                             x1       4     1 2 3 4
                             x2       4     1 2 3 4
                             x3       2     1 2
                             x4       2     1 2
                             x5       3     1 2 3
                             x6       3     1 2 3
                             x7       2     1 2
                             x8       3     1 2 3

                                                                   Average
                                                                  Prediction
    Design                                                         Standard
    Number     D-Efficiency     A-Efficiency     G-Efficiency       Error
    ------------------------------------------------------------------------
         1        84.7548          71.1686          98.0583          0.9806

In this example, the number of possible choice sets that meet the
restrictions is 2776, which is number small enough to search using the
MktDes macro.  In this case the MktDes macro using PROC OPTEX almost
always out-performs the other approaches in the MktEx macro.

This step assigns the final variable names and actual levels to the
OUTR=Randomized design.  The data set KEY contains the variable names
and levels that you ultimately want.  The variable Client with 4 levels
will be made from x1, ClientLineExtension with 4 levels will be made
from x2, ClientMicro with 2 levels will be made from x3, and so on.  The
N (for not available) is treated as a special missing value.  The key
data set has four rows because the maximum number of levels is four.
Factors with fewer than four levels are filled in with ordinary missing
values.  The macro MktLab takes by default the DATA=RANDOMIZED data set
from MktEx and the KEY=KEY data set, and combines the information to
create the OUT=Final data set, which is shown below, sorted by the
shelf talker variable.

   data key;
     missing N;
     input Client ClientLineExtension ClientMicro $ ShelfTalker $
           Regional Private PrivateMicro $ NationalLabel;
     format _numeric_ dollar5.2;
     datalines;
   1.29 1.39 micro Yes 1.99 1.49 micro 1.99
   1.69 1.89 stove No  2.49 2.29 stove 2.39
   2.09 2.39 .     .   N     N   .     N
   N    N    .     .   .     .   .     .
   ;

   %mktlab(key=key);

   proc sort; by shelftalker; run;

   proc print; by shelftalker; run;

Here is the design:

-------------------------------- ShelfTalker=No --------------------------------

                  Client
                   Line      Client                        Private   National
  OBS   Client   Extension   Micro    Regional   Private    Micro     Label

    1   $2.09      $1.89     micro       N          N       stove       N
    2   $1.69      $1.39     stove       N        $1.49     stove     $1.99
    3   $1.69      $2.39     stove       N        $2.29     stove     $2.39
    4   $1.29        N       stove     $2.49      $1.49     micro     $1.99
    5   $2.09        N       stove     $1.99        N       micro       N
    6     N        $1.39     stove     $2.49        N       stove     $2.39
    7   $1.29      $1.39     micro     $1.99      $2.29     stove       N
    8   $1.29      $1.89     stove     $1.99      $2.29     micro       N
    9     N        $2.39     micro     $1.99      $1.49     stove       N
   10   $1.69        N       micro       N        $2.29     micro     $2.39
   11     N        $1.89     stove     $2.49      $2.29     stove     $2.39
   12   $2.09      $2.39     micro     $2.49        N       micro     $1.99
   13   $1.69      $1.89     micro     $1.99        N       micro     $2.39

------------------------------- ShelfTalker=Yes --------------------------------

                  Client
                   Line      Client                        Private   National
  OBS   Client   Extension   Micro    Regional   Private    Micro     Label

   14   $1.69      $1.89     stove     $1.99      $1.49     micro       N
   15   $1.69      $1.39     micro     $2.49        N       stove     $1.99
   16   $1.29      $1.89     stove       N        $1.49     stove     $2.39
   17   $1.29      $1.39     stove     $1.99      $2.29     stove       N
   18   $2.09      $2.39     stove     $2.49      $1.49     stove       N
   19   $1.29        N       micro     $2.49        N       micro       N
   20   $1.69        N       stove       N        $2.29     micro     $2.39
   21   $1.29      $2.39     micro     $1.99        N       stove     $2.39
   22   $2.09        N       micro     $1.99      $2.29     stove     $1.99
   23     N        $2.39     stove       N          N       micro     $1.99
   24     N        $1.89     micro     $2.49      $2.29     micro     $1.99
   25   $1.69      $2.39     micro       N        $2.29     stove       N
   26   $2.09      $1.39     micro       N        $1.49     micro     $2.39

Here again is the "bad" macro that was used in this example.

   %macro bad;
     navail = (x1 < 4) + (x2 < 4) + (x5 < 3) + (x6 < 3) + (x8 < 3);
     bad = abs(navail - 3) * ((navail < 2) | (navail > 4));
     %mend;

While it has some rather involved boolean algebra, it is simple in that
badness is purely a function of the levels within one row of the design,
and that information is stored in the vector x1-x8.  We will see below that
the macro for imposing restrictions can be quite a bit more complicated
and use additional information.

------------------------------------------------------------------------

This next example creates a partial-profile design (Chrzan and Elrod,
1995).  This design has 12 five-level factors, and only 4 of them are
allowed to have levels other than the minimum level in each run, except
in the first run, where no levels are allowed to vary.  This first
constant observation is needed because for all other observations, the
sum of the dummy variables for the first levels of all of the other
factors is a constant (in this case 12 - 4 = 8).  Without this extra
observation there is a linear dependency (and zero efficiency).  Partial
profile designs are slow to construct, so MAXDESIGNS=1 was specified to
make the process faster.

   %mktex( 5 ** 12, n=49, partial=4, seed=368, maxdesigns=1 )

Here is a small part of the iteration history.

                            Algorithm Search History

                                   Current          Best
           Design    Row,Col  D-Efficiency  D-Efficiency  Notes
           ----------------------------------------------------------
                1      Start       35.5045                Ran,Mut,Ann
                1      1   1       22.7342       22.7342  Conforms
                1      2   7       22.7342       22.7342
                .
                .
                .
                1     41   3       43.6568       43.6568
                1     42   1       43.6568       43.6568
                1     37   6       43.6568       43.6568
                1     45   1       43.6568       43.6568
                1        End       43.6568

Notice that the D-efficiency for the initial design is 35.5045, and the
first entry in the Best D-efficiency column is 22.7342.  This is normal
with partial profile designs.  The initial D-efficiency is for the
initial random design, which does not conform to the requirements of a
partial profile design.  At the end, the final design is better than the
nonconforming random initial design, but this will not always happen.

Here is the last part of the results:

                                                                   Average
                                                                  Prediction
    Design                                                         Standard
    Number     D-Efficiency     A-Efficiency     G-Efficiency       Error
    ------------------------------------------------------------------------
         1        43.6567          18.9365         100.0000          1.0000

The final D-Efficiency reported by PROC OPTEX differs from the final
value in the iteration history table in the last decimal place.  This
happens fairly regularly and is due to the different ridging strategies
uses by the two different algorithms.

Here is part of the design:

    x1    x2    x3    x4    x5    x6    x7    x8    x9    x10    x11    x12

     1     1     1     1     1     1     1     1     1     1      1      1
     1     1     1     1     1     1     2     1     1     3      5      5
     1     1     1     1     1     1     4     4     5     1      1      3
     .
     .
     .
     5     1     2     1     1     1     1     1     1     5      1      2
     5     3     1     1     3     1     2     1     1     1      1      1
     5     4     1     1     1     1     1     1     4     1      1      5

Notice that in the first observation, no levels vary, and in all others,
exactly 4 vary.

------------------------------------------------------------------------

This next design is a bit crazy: one two-level factor, one three-level
factor, ...,  and one eleven-level factor, all in 168 runs.

   %mktex( 2 3 4 5 6 7 8 9 10 11, n=168, seed=151 )

                             The OPTEX Procedure

                           Class Level Information

                    Class  Levels  Values

                    x1        2    1 2
                    x2        3    1 2 3
                    x3        4    1 2 3 4
                    x4        5    1 2 3 4 5
                    x5        6    1 2 3 4 5 6
                    x6        7    1 2 3 4 5 6 7
                    x7        8    1 2 3 4 5 6 7 8
                    x8        9    1 2 3 4 5 6 7 8 9
                    x9       10    1 2 3 4 5 6 7 8 9 10
                    x10      11    1 2 3 4 5 6 7 8 9 10 11

                                                                  Average
                                                                 Prediction
   Design                                                         Standard
   Number     D-Efficiency     A-Efficiency     G-Efficiency       Error
   ------------------------------------------------------------------------
        1        98.5290          97.0629          96.2612          0.5774

This is the sort of problem where the MktEx macro excels compared to
other methods, because the extreme asymmetry does not pose a problem for
this macro.

------------------------------------------------------------------------

The rest of these examples shows some of the details about how the
RESTRICTIONS= option is used.

Partial profiles could instead be handled with the RESTRICTIONS= option
and a user written macro.

   %macro partprof;
      nvary = sum(x ^= 1);
      if i = 1 then bad = nvary;
      else          bad = abs(nvary - 4);
      %mend;

   %mktex( 5 ** 12, n=49, restrictions=partprof, seed=368, maxdesigns=1,
           optiter=0, tabiter=0, order=random, exchange=2)

The PARTPROF macro first counts the number of factors in a row with
values not equal to one and stores this result in NVARY.  Then it
creates a badness value that increases as the number of values gets
farther away from 4 (BAD = ABS(NVARY - 4);).  Notice that the macro is a
little more complicated than that.  This is because in a partial profile
design we need one observation that contains the minimum value for all
levels.  So we set up the restrictions so that when i = 1 (the first run
of the design) we produce a constant run where no factors vary, and for
all others, exactly 4 are allowed to vary.

Your macro can look at several things in quantifying badness, and must
store its results in "bad".

  i      - is a scalar that contains the number of the row currently
           being changed.
  try    - is like i, but it is only reset when a new design starts
  j1     - is a scalar that contains the number of the column currently
           being changed.
  j2     - is a scalar that contains the number of the second column
           currently being changed with exchange=2.
  j3     - is a scalar that contains the number of the third column
           currently being changed with exchange=3.
  x      - is a vector of factor levels, always containing values
           beginning with 1 and continuing on to the number of levels
           for each factors.
  x1     - x[1],  x2 - x[2], ..., and so on
  xmat   - is the entire x matrix.  Note that the ith row of xmat may
           not be x since x may contain information on swaps being
           considered.
  bad    - results, 0 - fine, or the number of violations of restrictions
  __pbad - is the badness from the partial= option.  When partial= is not
           specified, this scalar does not exist. Your macro can weight
           this value, typically by multiplying it times a constant, to
           differentially weight the contributors to badness, e.g.:
           __pbad = __pbad * 10.
  __bbad - is the badness from the balance= option.  When balance= is not
           specified, this scalar does not exist. Your macro can weight
           this value, typically by multiplying it times a constant, to
           differentially weight the contributors to badness, e.g.:
           __bbad = __bbad * 100.

For most problems, you will only need to look at x and sometimes i.
DO NOT USE THESE NAMES FOR INTERMEDIATE VALUES!!!
You can use any other name for intermediate values.

Let's look at the restrictions macro for this example in more detail.

   %macro partprof;
      nvary = sum(x ^= 1);
      if i = 1 then bad = nvary;
      else          bad = abs(nvary - 4);
      %mend;

The macro that imposes the restrictions is defined and called in four
distinct places in the MktEx macro.  First, the restrictions macro is
called in a separate, preliminary IML step, just to catch some syntax
errors you might have made.  Then it is called in between calling PROC
PLAN or FACTEX and calling PROC OPTEX.  Here, the restrictions macro is
used to impose restrictions on the candidate set.  Next it is used in
the obvious way during design creation and the coordinate exchange
algorithm.  Finally, when OPTIONS=ACCEPT is specified which means that
restriction violations are acceptable, the macro is called after all of
the iterations have completed to report on restriction violations in the
final design.  For some advanced restrictions, we will not want exactly
the same code running in all four places.  When the restrictions are
purely written in terms of restrictions on x, which is the ith row of
the design matrix, there is no problem.  The same macro will work fine
for all uses.  However, when xmat (the full x matrix) or i or j1 (the
row or column number) are used, the same code typically cannot be used
for all applications, although sometimes it does not matter.

Syntax Check:
In this phase, the macro is defined and called just to check for syntax
errors.  Your restrictions macro can recognize when it is in this phase
because the macro variable &main is set to 0 and the macro variable
&pass is set to null.

   * execute in syntax check;
   %if &main eq 0 and &pass eq %then %do;

Candidate Check:
In this phase, the macro is used to impose restrictions on the candidate
set created by PROC PLAN or PROC FACTEX before it is searched by PROC
OPTEX.  Your restrictions macro can recognize when it is in this phase
because the macro variable &main is set to 0 and the macro variable
&pass is set to 1 or 2.

   * execute on candidates;
   %if &main eq 0 and &pass ge 1 and &pass le 2 %then %do;

Main Coordinate Exchange Algorithm:
In this phase, the macro is used to impose restrictions on the design as
it is being built in the coordinate exchange algorithm.  Your
restrictions macro can recognize when it is in this phase because the
macro variable &main is set to 1 and the macro variable &pass is set to
1, 2, or 3.

   * execute on coordinate exchange;
   %if &main eq 1 and &pass ge 1 and &pass le 3 %then %do;

Restrictions Violations Check:
In this phase, the macro is used to check the design when there are
restrictions and OPTIONS=ACCEPT.  Your restrictions macro can recognize
when it is in this phase because the macro variable &main is set to 1
and the macro variable &pass is greater than 3.

   * execute on final check;
   %if      &main eq 1 and &pass gt 3 %then %do;

------------------------------------------------------------------------

Here is another example.  If you want to ensure that the design has no
duplicate runs, you could specify OPTIONS=NODUPS.

   %mktex( 2 2 2 2 2, n=20, seed=114, options=nodups )

You could instead program the no duplicates restriction yourself as
follows.

   %macro nodups;
      %if &main %then %do;
         l = (1:20); l[i] = 0; l = loc(l);
         bad = sum(((shape(x, 19, 5) = xmat[l,])[,+] = 5));
         %end;
      %mend;

   %mktex( 2 2 2 2 2, n=20, seed=114, restrictions=nodups )

The %IF &MAIN specification ensures that the restrictions are imposed on
the design and not on the candidate set.  This is important because if
these restrictions were imposed on the candidate set, there would be a
nonconforming matrices error since xmat does not have 20 rows and 5
columns in the candidate set.  Incidentally, bad is initialized to zero
automatically before the restrictions macro is called, so in the
candidate restrictions, when the macro does nothing, a bad value of zero
is always returned.

------------------------------------------------------------------------

Here is another example with a pretty complicated set of restrictions.
This time the restrictions macro BAL ensures that the design is at least
nearly balanced.  In the candidate set restrictions phase, &MAIN is
zero, and all candidates are considered fine.  (Recall the default
initialization bad = 0;) is always performed.  When the macro is in the
main algorithm but before the iterations, j1 = 0, so bad = 1, and the
algorithm is told that all initial designs have restriction violations.
When j1, the column number is set in the main algorithm, the restrictions
count the number of times each level occurs within each factor and
ensures that the counts do not differ by more than one, ensuring good
balance.   The statement ACOL = XMAT[,J1]; pulls out the current column.
The statement ACOL[I,] = X[,J1]; stores in it the level being considered.
The statement ACOL = DESIGN(ACOL)[+,]; counts the number of levels and
stores the counts in a row vectors.  The DESIGN function creates a set
of binary dummy variables, which are summed with the [+,] specification.
Finally, the statement BAD = MAX(0, MAX(ACOL) - MIN(ACOL) - 1); finds
the frequency range, subtracts one, and computes the maximum of the
result and zero.  This allows the minimum and maximum to be no more than
one different.

   %macro bal;
      %if &main %then %do;
         bad = 1;
         if j1 then do;
            acol = xmat[,j1];
            acol[i,] = x[,j1];
            acol = design(acol)[+,];
            bad = max(0, max(acol) - min(acol) - 1);
            end;
         %end;
      %mend;

   %mktex(3 3 2 ** 4, n=16, restrictions=bal, seed=205)
   %mkteval;

------------------------------------------------------------------------

                           References

See http://support.sas.com/techsup/technote/ts723.html and
http://support.sas.com/techsup/technote/ts723.pdf for the reference
list.

------------------------------------------------------------------------

 DISCLAIMER:

       THIS INFORMATION IS PROVIDED BY SAS INSTITUTE INC. AS A SERVICE
 TO ITS USERS.  IT IS PROVIDED "AS IS".  THERE ARE NO WARRANTIES,
 EXPRESSED OR IMPLIED, AS TO MERCHANTABILITY OR FITNESS FOR A
 PARTICULAR PURPOSE REGARDING THE ACCURACY OF THE MATERIALS OR CODE
 CONTAINED HEREIN.

----------------------------------------------------------------------*/

%*========================= Macro Argument List Start =========================;

%macro mktex(     /*--------------------------------------------------*/
                  /* Most usages of this macro are not very           */
                  /* complicated and are similar to these, but with   */
                  /* different numbers:                               */
                  /*                                                  */
                  /*    %mktex( n=18 )                                */
                  /*    %mktex( 2 2 3 3 4 4, n=24 )                   */
                  /*    %mktex( 4 2 ** 5 3 ** 5, n=36 )               */
                  /*                                                  */
                  /* You always need to specify n=, and you usually   */
                  /* need to specify a list of factor levels.         */
                  /*                                                  */
list,             /* The first parameter is a list of the numbers of  */
                  /* levels of all the factors.  For example, for 3   */
                  /* two-level factors specify either "2 2 2" or      */
                  /* "2 ** 3".  Lists of numbers, like 2 2 3 3 4 4 or */
                  /* a "levels**number of factors" syntax like: 2**2  */
                  /* 3**2 4**2 can be used, or both can be combined:  */
                  /* 2 2 3**4 5 6.  A specification like '3**4' means */
                  /* four three-level factors.  Usually, you have to  */
                  /* specify a list, however, in some cases, you can  */
                  /* just specify N=, and omit the list, and a        */
                  /* default list is implied.  For example, n=18      */
                  /* implies a list of 2 3 ** 7.  See the comments    */
                  /* above for more examples.  Note that the factor   */
                  /* list is a positional parameter.  This means it   */
                  /* must come first, and unlike all other            */
                  /* parameters, it is not specified after an equal   */
                  /* sign.                                            */
                  /*                                                  */
                  /* When the list is omitted, and if there are no    */
                  /* interactions, restrictions, or duplicate         */
                  /* exclusions, then by default there are no OPTEX   */
                  /* iterations (OPTITER=0).                          */
                  /*                                                  */
                  /* ? or help - lists options                        */
                  /*                                                  */
n=,               /* specifies the number of runs in the final        */
                  /* design.  You must specify n=.  You can specify   */
                  /* the same factor level list as above with the     */
                  /* MktRuns macro to get suggestions for values of   */
                  /* N=.                                              */
                  /*                                                  */
                  /*    %mktruns( 4 2 ** 5 3 ** 5 )                   */
                  /*                                                  */
                  /*--------------------------------------------------*/
                  /*                                                  */
                  /* This next group of options contains some of the  */
                  /* more commonly used options.                      */
                  /*                                                  */
interact=,        /* specifies interactions that must be estimable.   */
                  /* Examples:                                        */
                  /*    INTERACT=x1*x2                                */
                  /*    INTERACT=x1*x2 x3*x4*x5                       */
                  /*    INTERACT=x1|x2|x3|x4|x5@2                     */
                  /*                                                  */
                  /* The interaction syntax is like GLM's and many of */
                  /* the other modeling procedures.  It uses "*" for  */
                  /* simple interactions (x1*x2, the interaction      */
                  /* between x1 and x2), "|" for main effects and     */
                  /* interactions (x1|x2|x3 is the same as x1 x2      */
                  /* x1*x2 x3 x1*x3 x2*x3 x1*x2*x3) and "@" to        */
                  /* eliminate higher-order interactions (x1|x2|x3@2  */
                  /* is the same as x1 x2 x1*x2 x3 x1*x3 x2*x3).      */
                  /* "@2" allows only main effects and two-way        */
                  /* interactions.  By default, no interactions are   */
                  /* guaranteed to be estimable.  For the factor      */
                  /* names, you must specify either the actual        */
                  /* variable names (e.g. x1 x2 ...) or you can just  */
                  /* specify the number without the "x" (e.g. x1*x2   */
                  /* is equivalent to 1*2).                           */
                  /*                                                  */
                  /* You can also specify INTERACT=@2 for all main    */
                  /* effects and two-way interactions.  For example,  */
                  /* these two specifications are equivalent:         */
                  /*    %mktex(2 ** 5, interact=@2, n=32)             */
                  /*    %mktex(2 ** 5, interact=1|2|3|4|5@2, n=32)    */
                  /*                                                  */
restrictions=,    /* specifies the name of a macro that places        */
                  /* restrictions on the design.  If you have         */
                  /* restrictions on the design, what combinations    */
                  /* can appear with what other combinations, then    */
                  /* you must create a macro that creates a variable  */
                  /* called "bad" that contains a numerical summary   */
                  /* of how bad the row of the design is.  When       */
                  /* everything is fine, set bad to zero.  Otherwise  */
                  /* set bad to a larger value.  Ideally, set bad to  */
                  /* the number of violations so that the macro knows */
                  /* if changes to factor levels are moving in the    */
                  /* right direction.  The macro must consist of IML  */
                  /* statements that create the variable bad.         */
                  /*                                                  */
                  /* Be sure to check the log when you specify        */
                  /* RESTRICTIONS=.  The macro usually cannot ensure  */
                  /* that your statements are syntax-error free and   */
                  /* stop if they aren't.                             */
                  /*                                                  */
                  /* Your macro can look at several things in         */
                  /* quantifying badness, and must store its results  */
                  /* in "bad".                                        */
                  /* i    - is a scalar that contains the number of   */
                  /*        the row currently being changed.          */
                  /*        If you are writing restrictions that use  */
                  /*        the variable i, you almost certainly      */
                  /*        should specify OPTIONS=NOSORT.            */
                  /* try  - is a scalar similar to i that contains    */
                  /*        the number of the row currently being     */
                  /*        changed.  However, try, starts at zero    */
                  /*        and is incremented for each row, but it   */
                  /*        is only set back to zero when a new       */
                  /*        design starts, not when MktEx reaches the */
                  /*        last row.  Use i as a matrix index and    */
                  /*        try to evaluate how far MKtEx is into the */
                  /*        process of constructing the design.  In   */
                  /*        previous releases, try was not            */
                  /*        automatically available.                  */
                  /* x    - is a vector of factor levels, always      */
                  /*        containing values beginning with 1 and    */
                  /*        continuing on to the number of levels for */
                  /*        each factor.  These values are one-based  */
                  /*        even if LEVELS= is specified.             */
                  /* x1   - x[1]                                      */
                  /* x2   - x[2] and so on.                           */
                  /* j1   - is a scalar that contains the number of   */
                  /*        the column currently being changed.       */
                  /*        (With exchange=, there is a j2 and j3 as  */
                  /*        well.)                                    */
                  /* xmat - is the entire x matrix.  Note that the    */
                  /*        ith row of xmat may not be x since x may  */
                  /*        contain information on swaps being        */
                  /*        considered.                               */
                  /* bad  - results, 0 - fine, or the number of       */
                  /*        violations of restrictions                */
                  /*                                                  */
                  /* DO NOT USE THESE NAMES FOR INTERMEDIATE VALUES!! */
                  /* Other than that, you can create intermediate     */
                  /* variables without worrying about conflicting     */
                  /* with the names in the macro.  The levels of the  */
                  /* factors for one row of the experimental design   */
                  /* are stored in a vector x, and the first level is */
                  /* always 1, the second always 2, and so on.        */
                  /* All restrictions must be defined in terms of     */
                  /* x[j] (or alternatively, x1, x2, ..., and perhaps */
                  /* the other matrices).  For example, if there are  */
                  /* five three-level factors and if it is bad if the */
                  /* level for a factor equals the level for the      */
                  /* following factor, create a macro restrict as     */
                  /* follows:                                         */
                  /*                                                  */
                  /* %macro restrict;                                 */
                  /*    bad = (x1 = x2) +                             */
                  /*          (x2 = x3) +                             */
                  /*          (x3 = x4) +                             */
                  /*          (x4 = x5);                              */
                  /*    %mend;                                        */
                  /*                                                  */
                  /* and specify RESTRICTIONS=restrict.  Note that    */
                  /* you specify just the macro name, no percents on  */
                  /* the RESTRICTIONS= option.  Also note that IML    */
                  /* does not have the full set of boolean operators  */
                  /* that the DATA step and other parts of SAS have.  */
                  /* For example, these are NOT available:            */
                  /* OR AND NOT GT LT GE LE EQ NE.                    */
                  /* Here are the operators you can use along with    */
                  /* their meaning.                                   */
                  /*  = - equals                                      */
                  /* ^= - not equals                                  */
                  /* <  - less than                                   */
                  /* <= - less than or equal to                       */
                  /* >  - greater than                                */
                  /* >= - greater than or equal to                    */
                  /* &  - and                                         */
                  /* |  - or                                          */
                  /* ^  - not                                         */
                  /*                                                  */
                  /* Restrictions seriously slow down the algorithm.  */
                  /*                                                  */
                  /* With restrictions, the Current D-Efficiency      */
                  /* column of the iteration history table may        */
                  /* contain values larger than the Best D-Efficiency */
                  /* column.  This is because the design              */
                  /* corresponding to the Current D-Efficiency may    */
                  /* have restriction violations.  Values are only    */
                  /* reported in the Best D-Efficiency column after   */
                  /* all of the restriction violations have been      */
                  /* removed.                                         */
                  /*                                                  */
                  /* For certain advanced designs, you can impose a   */
                  /* different set of restrictions on the PROC OPTEX  */
                  /* candidate set and on the design.  See the        */
                  /* example above in the header for more             */
                  /* information.                                     */
                  /*                                                  */
                  /* You can specify OPTIONS=ACCEPT with              */
                  /* RESTRICTIONS= when it is OK if the restrictions  */
                  /* are not met.                                     */
                  /*                                                  */
                  /* The RESLIST= and RESMAC= options can be used     */
                  /* jointly for certain complicated restrictions to  */
                  /* set up some constant matrices that the           */
                  /* restrictions macro can use.  Since the           */
                  /* restrictions macro is called a lot.  Anything    */
                  /* you can do only once helps speed up the          */
                  /* algorithm.                                       */
                  /*                                                  */
reslist=,         /* specifies a list of constant matrices.  Begin    */
                  /* all names with an underscore to ensure that they */
                  /* do not conflict with any of MktEx's names.  If   */
                  /* you specify more than one name, then names must  */
                  /* be separated by commas. Example:                 */
                  /* reslist=%str(_a, _b)                             */
                  /*                                                  */
resmac=,          /* specifies the name of a macro that creates the   */
                  /* matrices named on the RESLIST= option.  Begin    */
                  /* all names including all intermediate matrix      */
                  /* names with an underscore to ensure that they do  */
                  /* not conflict with any of MktEx's names.          */
                  /*                                                  */
partial=,         /* specifies a partial profile design.  Specify for */
                  /* example partial=4 if you only want 4 attributes  */
                  /* to vary in each row of the design (except the    */
                  /* first run, in which none vary).  This option     */
                  /* works by adding restrictions to the design (see  */
                  /* RESTRICTIONS=), by specifying ORDER=RANDOM,      */
                  /* and by specifying EXCHANGE=2.  Because of the    */
                  /* default EXCHANGE=2 with partial profile designs, */
                  /* the construction is slow, so you may want to     */
                  /* specify MAXDESIGNS=1 or other options to make it */
                  /* run faster.  For large problems, you may get     */
                  /* faster but less good results by specifying       */
                  /* ORDER=SEQRAN.  Specifying OPTIONS=ACCEPT or      */
                  /* BALANCE= with PARTIAL= is usually not a good     */
                  /* idea.                                            */
                  /*                                                  */
balance=,         /* specifies the maximum allowable level-frequency  */
                  /* range.  This option allows you to tell the macro */
                  /* that it should make an extra effort to ensure    */
                  /* that the design is nearly balanced.  Specify a   */
                  /* positive integer, usually 1 or 2, that specifies */
                  /* the degree of imbalance that is acceptable.      */
                  /* BALANCE=n specifies that for each factor, a      */
                  /* difference between the frequencies of the most   */
                  /* and least frequently occuring levels should be   */
                  /* no larger than n.  You may specify BALANCE=0,    */
                  /* however, this usually is not a good idea because */
                  /* the macro needs the flexibility to have          */
                  /* imbalance as it refines the design.  Often, the  */
                  /* design actually found will be better balanced    */
                  /* than your BALANCE=n specificaion would require.  */
                  /* For this reason, it is good to start by          */
                  /* specifying a value larger than the minimum       */
                  /* acceptable value.  The larger the value, the     */
                  /* more freedom the algorithm has to optimize both  */
                  /* balance and efficiency.                          */
                  /*                                                  */
                  /* The macro usually does a good job of producing   */
                  /* nearly balanced designs, but if balance is       */
                  /* critically important, and your designs are not   */
                  /* balanced enough, you can sometimes achieve       */
                  /* better balance by specifying BALANCE=, but       */
                  /* usually at the price of worse efficiency,        */
                  /* sometimes much worse.  The BALANCE= option       */
                  /* specifies additional restrictions (see           */
                  /* RESTRICTIONS=) that helps achieve better         */
                  /* balance.  By default, no additional restrictions */
                  /* are added.  When you specify BALANCE=, you       */
                  /* should also specify MINTRY= (perhaps something   */
                  /* like MINTRY=5 * n, or MINTRY=10 * n).  This can  */
                  /* often help the BALANCE= option quite a bit.      */
                  /* Another approach is to instead use the MktBal    */
                  /* macro, which for main effects plans with no      */
                  /* restrictions, produces designs that are          */
                  /* guaranteed to have optimal balance.              */
                  /*                                                  */
mintry=n,         /* specifies the minimum number of rows to process  */
                  /* before giving up for each design.  For example,  */
                  /* to ensure that the macro passes through each row */
                  /* of the design at least five times, you can       */
                  /* specify MINTRY=5 * n.  You can specify a number  */
                  /* or a DATA step language expression involving n   */
                  /* (rows) and m (columns).  By default, the macro   */
                  /* will always consider at least n rows.  This      */
                  /* option can be useful with certain restrictions,  */
                  /* particularly with BALANCE=.  When BALANCE= and   */
                  /* MINTRY=mt are both specified, then the balance   */
                  /* restrictions are ignored for the first mt - 3 *  */
                  /* n / 2 passes through the design.  During this    */
                  /* period, the badness function for the balance     */
                  /* restrictions is set to 1 so that MktEx knows     */
                  /* that the design does not conform.  After that,   */
                  /* all restrictions are considered.  The BALANCE=   */
                  /* option works best when its restrictions are      */
                  /* imposed on a reasonably efficient design not an  */
                  /* inefficient initial design.                      */
                  /*                                                  */
                  /* MktEx sometimes prints:                          */
                  /*    WARNING: It may be impossible to meet all     */
                  /*             restrictions.                        */
                  /* In previous releases, it did this when MktEx     */
                  /* passed through the entire design (n rows)        */
                  /* without succeeding in minimizing the badness in  */
                  /* any of the rows.  Now, the message is printed    */
                  /* after MINTRY= rows are passed without any        */
                  /* success.  Sometimes, it is premature to expect   */
                  /* any success during the first pass.  When you     */
                  /* know this, you can specify this option to        */
                  /* prevent that warning from coming out.            */
                  /*                                                  */
examine=,         /* specify:                                         */
                  /* I to examine the information matrix,             */
                  /* V to examine the variance matrix,                */
                  /* ALIASING=n to examine a condensed representation */
                  /* of the design aliaising structure.  If you       */
                  /* specify ALIASING=2, MktEx will print the terms   */
                  /* in the model and how they are aliased with up to */
                  /* two-factor interactions.  More generally, with   */
                  /* ALIASING=n, up to n-factor interactions will be  */
                  /* displayed.   You can also specify FULL (e.g.     */
                  /* EXAMINE=ALIASING=2 FULL) to see the full (and    */
                  /* often much more complicated) aliasing structure  */
                  /* that PROC GLM produces directly.  You can also   */
                  /* specify MAIN to see only the estimable functions */
                  /* that begin with main effects, and not the ones   */
                  /* that begin with interactions.  Interactions will */
                  /* still be used with ALIASING=2 and larger values. */
                  /* This option just removes some of the output.     */
                  /*                                                  */
options=,         /* specifies binary options.  Specify one or more   */
                  /* of the following values after OPTIONS=.          */
                  /*                                                  */
                  /* nox - suppresses the creation of x1, x2, x3, ... */
                  /* with restrictions.  If you are not using these   */
                  /* names in your restrictions macro, specifying     */
                  /* OPTIONS=NOX can make the macro run somewhat more */
                  /* efficiently.                                     */
                  /*                                                  */
                  /* nodups  - eliminates duplicate runs.             */
                  /*                                                  */
                  /* nooadups - for orthogonal array construction,    */
                  /* checks to see if a design with duplicate runs    */
                  /* is created.  If so, it tries using other factors */
                  /* from the larger OA to see if that helps avoid    */
                  /* duplicates.  There is no guarantee that this     */
                  /* option will work.  If you select only a small    */
                  /* subset of the columns of an OA, duplicates may   */
                  /* be unavoidable.  If you really wish to ensure    */
                  /* no duplicates, even at the expense of            */
                  /* nonorthogonality, you must also specify          */
                  /* OPTIONS=NODUPS.                                  */
                  /*                                                  */
                  /* nofinal - do not call PROC OPTEX to print the    */
                  /* efficiency of the final experimental design.     */
                  /*                                                  */
                  /* nosort - do not sort the design.                 */
                  /*                                                  */
                  /* nohistory - do not print the iteration history.  */
                  /*                                                  */
                  /* noqc - specifies that you do not have the SAS/QC */
                  /* product, so the \mktex macro should try to get   */
                  /* by without it.  This means it tries to get by    */
                  /* using the coordinate exchange and orthogonal     */
                  /* array code without using PROC FACTEX and PROC    */
                  /* OPTEX.  The MktEx macro will skip generating a   */
                  /* candidate set, searching the candidate set, and  */
                  /* printing the final efficiency values.            */
                  /*                                                  */
                  /* resrep - report on the progress of the           */
                  /* restrictions.  You should always use this option */
                  /* with a new set of restrictions.  It can help you */
                  /* figure out what is happening if your             */
                  /* restrictions are not being met.                  */
                  /*                                                  */
                  /* int - add an intercept to the design.            */
                  /*                                                  */
                  /* lineage - print design lineage.                  */
                  /*                                                  */
                  /* render - print the design compactly.             */
                  /*                                                  */
                  /* file - render to a generated file name, e.g.:    */
                  /* OA(36,2^11,3^12)                                 */
                  /*                                                  */
                  /* +- - with render, print -1 as '-' and 1 as '+'   */
                  /* in two-level factors.  This option is typically  */
                  /* used with LEVELS=i for printing Hadamard         */
                  /* matrices.                                        */
                  /*                                                  */
                  /* 3 - modifies +- to apply to three-level factors  */
                  /* as well: -1 as '-', 0 as '0', and 1 as '+'.      */
                  /*                                                  */
                  /* justinit - the macro should stop processing as   */
                  /* soon as it is done making the initial design,    */
                  /* even if that design would not normally be the    */
                  /* final design.  Usually, this design will be an   */
                  /* orthogonal array or some function of an          */
                  /* orthogonal array (e.g. some three-level factors  */
                  /* could be recoded into two-level factors), but    */
                  /* there are no guarantees.  Use this option when   */
                  /* you want to output the initial tabled design,    */
                  /* for example, if you want to see an orthogonal    */
                  /* but unbalanced design that MktEx sometimes uses  */
                  /* as an initial design.  OPTIONS=JUSTINIT implies  */
                  /* OPTITER=0 and OUTR=.  OPTIONS=JUSTINIT NOFINAL   */
                  /* both stops processing and prevents the final     */
                  /* design from being evaluated.  Particularly when  */
                  /* you specify OPTIONS=NOFINAL, you must ensure     */
                  /* that this design has a suitable efficiency.      */
                  /*                                                  */
                  /* check - checks the efficiency of a given design, */
                  /* specified in INIT=, and disables the OUT=,       */
                  /* OUTR=, and OUTALL= options.  If INIT= is not     */
                  /* specified, OPTIONS=CHECK is ignored.             */
                  /*                                                  */
                  /* accept - allows the macro to output designs that */
                  /* violate restrictions imposed by RESTRICTIONS=,   */
                  /* BALANCE=, or PARTIAL=, or have duplicates with   */
                  /* OPTIONS=NODUPS.  Normally the macro will not     */
                  /* output such designs.  With OPTIONS=ACCEPT, a     */
                  /* design becomes eligible for output when the      */
                  /* macro can no longer improve on the restrictions  */
                  /* or eliminate duplicates.  Without                */
                  /* OPTIONS=ACCEPT, a design is only eligible when   */
                  /* all restrictions are met and all duplicates are  */
                  /* eliminated.                                      */
                  /*                                                  */
                  /* largedesign - allows the macro to stop after     */
                  /* MAXTIME= minutes have elapsed in the coordinate  */
                  /* exchange algorithm.  Typically you would use     */
                  /* this with MAXSTAGES=1 and other options that     */
                  /* make the algorithm run faster.  By default, the  */
                  /* macro checks time after it finishes with a       */
                  /* design.  With this option, the macro checks the  */
                  /* time at the end of each row, after it has        */
                  /* completed the first full pass through the        */
                  /* design, and after any restrictions have been     */
                  /* met, so the macro may stop before D-efficiency   */
                  /* has converged.  For really large problems and    */
                  /* problems with restrictions, this option may make */
                  /* the macro run much faster but at a price of      */
                  /* lower D-efficiency.  For example, for large      */
                  /* problems with restrictions, you might just want  */
                  /* to try one run through the coordinate exchange   */
                  /* algorithm with no candidate set search, OA's or  */
                  /* mutations.                                       */
                  /*                                                  */
                  /* %mktex(2**4 3**19 4**22 5**7 6**10 8**2,         */
                  /*        n=240, options=largedesign, maxtime=60,   */
                  /*        restrictions=bad, tabiter=0, optiter=0,   */
                  /*        maxstages=1, mutate=)                     */
                  /*                                                  */
                  /* quick - sets: OPTITER=0, MAXDESIGNS=2,           */
                  /*               UNBALANCED=0, TABITER=1,           */
                  /* provides a quick run with one tabled design and  */
                  /* if necessary, one coordinate exchange design     */
                  /* with random initialization.                      */
                  /*                                                  */
                  /* quickr - sets: OPTITER=0, MAXDESIGNS=1,          */
                  /*                UNBALANCED=0, TABITER=0,          */
                  /* provides an even quicker run with one coordinate */
                  /* exchange design with random initialization for   */
                  /* when you think using a tabled design will not    */
                  /* help.                                            */
                  /*                                                  */
                  /* quickt - sets: OPTITER=0, MAXDESIGNS=1,          */
                  /*                UNBALANCED=0, TABITER=1,          */
                  /* provides an even quicker run with one coordinate */
                  /* exchange design with a tabled design             */
                  /* initialization.                                  */
                  /*                                                  */
                  /* 512 - adds some larger designs in 512 runs with  */
                  /* mixes of 16, 8, 4, and 2-level factors to the    */
                  /* catalog, which gives added flexibility in 512    */
                  /* runs at a cost of much slower run time.  This    */
                  /* option replaces the default 4^160 32^1 parent    */
                  /* with 16^32 32^1.  Many of these designs are      */
                  /* automatically available with FACTEX, so don't    */
                  /* use this option unless you have first tried and  */
                  /* failed to find the design without it.            */
                  /*                                                  */
                  /* refine - specifies that with an INIT= design     */
                  /* data set with at least one nonpositive entry,    */
                  /* each successive design iteration tries to refine */
                  /* the best design from before.  By default, the    */
                  /* part of the design that is not fixed is randomly */
                  /* reinitialized each time.  The default strategy   */
                  /* is usually superior.                             */
                  /*                                                  */
levels=1,         /* specifies the method of assigning the final      */
                  /* factor levels.  This recoding occurs after the   */
                  /* design is created, so all restrictions must be   */
                  /* expressed in terms of one-based factors,         */
                  /* regardless of what is specified in the LEVELS=   */
                  /* option.  Values:                                 */
                  /* 1 - default, the levels are 1, 2, ...            */
                  /* 0 - the levels are 0, 1, ...                     */
                  /* c - centered, possibly resulting in nonintegers  */
                  /* {1 2} -> {-0.5 0.5}, {1 2 3} -> {-1 0 1}.        */
                  /* i - centered and scaled to integers.             */
                  /* {1 2} -> {-1 1}, {1 2 3} -> {-1 0 1}.            */
                  /* You can also specify separate values for two-    */
                  /* and three-level factors by preceding a value by  */
                  /* "2" or "3".  For example, LEVELS=2 i 3 0 c means */
                  /* two-level factors are coded -1, 1 and            */
                  /* three-level factors are coded 0, 1, 2.  The      */
                  /* remaining factors are centered.  Note that the   */
                  /* centering is based on centering the level values */
                  /* not on centering the (potentially unbalanced)    */
                  /* factor.  So for example the centered levels for  */
                  /* a two-level factor in five runs (1 2 1 2 1) are  */
                  /* (-0.5 0.5 -0.5 0.5 -0.5) not (-0.4 0.6 -0.4 0.6  */
                  /* -0.4).  If you want the latter form of           */
                  /* centering, use PROC STANDARD M=0.  See the       */
                  /* MktLab macro for more general level setting.     */
                  /* You can also specify three other values:         */
                  /* first - means the first row of the design        */
                  /* should consist entirely of the first level.      */
                  /* last - means the first row of the design should  */
                  /* consist entirely of the last level, which is     */
                  /* useful for Hadamard matrices.                    */
                  /* int - adds an intercept column to the design.    */
                  /*                                                  */
seed=0,           /* specifies the random number generation seed,     */
                  /* 0 - clock time.                                  */
                  /*                                                  */
                  /*--------------------------------------------------*/
                  /*                                                  */
                  /* These next options specify the names of the      */
                  /* input and output data sets.                      */
                  /*                                                  */
init=,            /* specifies the initial (input) experimental       */
                  /* design.  If all values in the initial design are */
                  /* positive, then a first step evaluates the        */
                  /* design, the next step tries to improve it, and   */
                  /* subsequent steps try to improve the best design  */
                  /* found.  However, if any values in the initial    */
                  /* design are nonpositive or missing then a         */
                  /* different approach is used.  The initial design  */
                  /* can have three types of values:                  */
                  /* * positive integers are fixed and constant and   */
                  /*   will not change throughout the course of the   */
                  /*   iterations.                                    */
                  /* * zero and missing values are replaced by random */
                  /*   values at the start of each new design search  */
                  /*   and can change throughout the course of the    */
                  /*   iterations.                                    */
                  /* * negative values are replaced by their absolute */
                  /*   value at the start of each new design attempt  */
                  /*   and can change throughout the course of the    */
                  /*   iterations.                                    */
                  /* When absolute orthogonality and balance are      */
                  /* required in a few factors, you can fix them in   */
                  /* advance.  This example illustrates how.          */
                  /*                                                  */
                  /* * Get the first four factors;                    */
                  /* %mktex(8 6 2 2, n=48)                            */
                  /*                                                  */
                  /* * Flag the first four as fixed and set up to     */
                  /*   search for the next six;                       */
                  /* data init;                                       */
                  /*    set design;                                   */
                  /*    retain x5-x10 .;                              */
                  /*    run;                                          */
                  /*                                                  */
                  /* * Get the last factors holding the first 4 fixed;*/
                  /* %mktex(8 6 2 2 4 ** 6, n=48, init=init,          */
                  /*        maxiter=100)                              */
                  /*                                                  */
                  /* %mkteval;                                        */
                  /*                                                  */
                  /* Alternatively you can use HOLDOUTS= or FIXED=    */
                  /* to fix just certain rows.                        */
                  /*                                                  */
out=Design,       /* specifies the output experimental design.  By    */
                  /* default this design is sorted unless you specify */
                  /* OPTIONS=NOSORT.  This is the output data set to  */
                  /* look at in evaluating the design.  See the OUTR= */
                  /* option (next) for a randomized version of the    */
                  /* same design, which is generally more suitable    */
                  /* for actual use.  Specify a null value for OUT=   */
                  /* if you do not want this data set created.        */
                  /*                                                  */
outr=Randomized,  /* specifies the randomized output experimental     */
                  /* design.  Random levels are assigned within       */
                  /* factors, and the runs are sorted into a random   */
                  /* order.  When RESTRICTIONS= or PARTIAL= is        */
                  /* specified, only the random sort is performed.    */
                  /* Specify a null value for OUTR= if you do not     */
                  /* want a randomized design created.                */
                  /*                                                  */
outeff=,          /* specifies the output data set with the final     */
                  /* efficiencies, the method used to find the        */
                  /* design, and the initial random number seed.      */
                  /* By default, this data set is not created.        */
                  /*                                                  */
outall=,          /* specifies the output data set containing all     */
                  /* designs found.  By default, this data set is not */
                  /* created.  This data set contains the design      */
                  /* number, efficiency, and the design.  You could   */
                  /* use this option to allow you to break while the  */
                  /* macro is running and still recover the best      */
                  /* design found so far.  Note, however, that        */
                  /* designs are only stored in this data set when    */
                  /* they are done being processed.  For example,     */
                  /* design 1 is stored once at the end, not every    */
                  /* time an improvement is found along the way.      */
                  /* Make sure you store the data set in a permanent  */
                  /* SAS data set.  If for example, your macro is     */
                  /* currently working on the tenth design, with this */
                  /* option, you could break and get access to        */
                  /* designs 1 through 9.  Example:                   */
                  /*                                                  */
                  /* %mktex(2 2 2 3 3 3, n=18, outall=sasuser.a)      */
                  /*                                                  */
                  /* proc means data=sasuser.a noprint;               */
                  /*    output out=m max(efficiency)=m;               */
                  /*    run;                                          */
                  /*                                                  */
                  /* data best;                                       */
                  /*    retain id .;                                  */
                  /*    set sasuser.a;                                */
                  /*    if _n_ = 1 then set m;                        */
                  /*    if nmiss(id) and abs(m - efficiency) < 1e-12  */
                  /*       then id = design;                          */
                  /*    if id = design;                               */
                  /*    keep design efficiency x:;                    */
                  /*    run;                                          */
                  /*                                                  */
                  /* proc print; run;                                 */
                  /*                                                  */
cat=,             /* specifies the input design catalog.  By default, */
                  /* the MktEx macro automatically runs the MktOrth   */
                  /* macro to get this catalog.  However, many        */
                  /* designs can be made in multiple ways, so you can */
                  /* instead run MktOrth yourself, and select the     */
                  /* exact design you want and specify the resulting  */
                  /* data set on the CAT= option.  Be sure to specify */
                  /* OPTIONS=dups lineage.  For example, the design   */
                  /* 2 ** 71, n=72, can be made from either           */
                  /* 2 ** 36 36 ** 1 or 2 ** 68 4 ** 1.  This example */
                  /* selects the 2 ** 36 36 ** 1 method:              */
                  /*                                                  */
                  /*    %mktorth(range=n=72, options=dups lineage)    */
                  /*                                                  */
                  /*    proc print data=mktdeslev; var lineage; run;  */
                  /*                                                  */
                  /*    data lev;                                     */
                  /*       set mktdeslev(where=(x2 = 71 and           */
                  /*           index(lineage, '2 ** 36 36 ** 1')));   */
                  /*       run;                                       */
                  /*                                                  */
                  /*    %mktex(2 ** 71, n=72, cat=lev, out=b)         */
                  /*                                                  */
                  /* Running these next steps shows that you are in   */
                  /* fact getting a design different from the         */
                  /* default.                                         */
                  /*                                                  */
                  /*    %mktex(2 ** 71, n=72, out=a)                  */
                  /*                                                  */
                  /*    proc compare data=a compare=b noprint note;   */
                  /*       run;                                       */
                  /*                                                  */
                  /*--------------------------------------------------*/
                  /*                                                  */
                  /* These next options control some of the details   */
                  /* of the iterations.  The macro can perform three  */
                  /* sets of iterations.  The "Algorithm Search" set  */
                  /* of iterations looks for efficient designs using  */
                  /* three different approaches.  It then determines  */
                  /* which approach appears to be most fruitful and   */
                  /* uses that approach exclusively in the second set */
                  /* of "Design Search" iterations.  The third set or */
                  /* "Design Refinement" iterations tries to refine   */
                  /* the best design found so far using level swaps   */
                  /* that optimize efficiency combined with random    */
                  /* mutations and simulated annealing.               */
                  /*                                                  */
                  /* The first set of iterations can have up to three */
                  /* parts.  The first part uses either PROC PLAN or  */
                  /* PROC FACTEX followed by PROC OPTEX, called       */
                  /* through the MktDes macro, to create and search a */
                  /* candidate set for an optimal initial design.     */
                  /* The second part may use an OA or fractional      */
                  /* factorial design as an initial design.  The next */
                  /* part consists of level exchanges starting with   */
                  /* random initial designs.                          */
                  /*                                                  */
                  /* In the first part, if the full-factorial design  */
                  /* is manageable (arbitrarily defined as < 5186     */
                  /* runs), it is used as a candidate set, otherwise  */
                  /* a fractional-factorial candidate set is used.    */
                  /* The macro tries OPTITER= iterations to make an   */
                  /* optimal design using MktDes and OPTEX.           */
                  /*                                                  */
                  /* In the second part, if the number of runs and    */
                  /* the levels of the factors match certain          */
                  /* patterns, then the macro will try to improve a   */
                  /* standard OA or fractional factorial design.      */
                  /* Sometimes, this can lead immediately to an       */
                  /* optimal design, for example with                 */
                  /* 2 ** 11 3 ** 12, n=36.  In other cases, when     */
                  /* only part of the desired design matches some     */
                  /* standard design, only part of the design is      */
                  /* initialized with the standard design and         */
                  /* multiple iterations are run using the standard   */
                  /* design as a partial initialization with the rest */
                  /* of the design randomly initialized.              */
                  /*                                                  */
                  /* In the third part, the macro goes into the level */
                  /* exchange algorithm.                              */
                  /*                                                  */
mutate=.05 .05    /* specifies the probability at which each value in */
       .01,       /* an initial design may mutate or be assigned a    */
                  /* different random value before the level-exchange */
                  /* iterations begin.  Specify a zero or null for no */
                  /* mutation.  The first value can be optionally     */
                  /* followed by a second and a third if you would    */
                  /* like to use a different value for the algorithm  */
                  /* search, design search, and design refinement     */
                  /* iterations.                                      */
                  /* Examples:                                        */
                  /* MUTATE= or MUTATE=0 specifies no random          */
                  /* mutations.                                       */
                  /* MUTATE=0.1 specifies a mutation probability of   */
                  /* 0.1 during all three sets of iterations.         */
                  /* MUTATE=0 0.1 0.05 specifies no mutations during  */
                  /* the first iterations, a mutation probability of  */
                  /* 0.1 during the search iterations, and a mutation */
                  /* probability of 0.05 during the refinement        */
                  /* iterations.                                      */
                  /*                                                  */
anneal=.05 .05    /* specifies the starting probability for simulated */
       .01,       /* annealing in the coordinate exchange algorithm.  */
                  /* Specify a zero or null for no annealing.  The    */
                  /* first value can optionally be followed by a      */
                  /* second and a third if you would like to use a    */
                  /* different value for the algorithm search, design */
                  /* search, and design refinement iterations.        */
                  /* Specifying a value (greater than zero and less   */
                  /* than one, for example 0.1) allows the design to  */
                  /* get worse with decreasing probability as the     */
                  /* number of iterations increases.  This often      */
                  /* helps the algorithm overcome local efficiency    */
                  /* maxima.  For example in the function below,      */
                  /* allowing efficiency to decrease can help get     */
                  /* past the bumps on the efficiency function.       */
                  /*                           . .                    */
                  /*                    .   .                         */
                  /*                  .   .                           */
                  /*           .   .                                  */
                  /*        .    .                                    */
                  /*      .                                           */
                  /*     .                                            */
                  /*                                                  */
                  /* Examples:                                        */
                  /* ANNEAL= or ANNEAL=0 specifies no annealing.      */
                  /* ANNEAL=0.1 specifies an annealing probability of */
                  /* 0.1 during all three sets of iterations.         */
                  /* MUTATE=0 0.1 0.05 specifies no annealing during  */
                  /* the initial iterations, an annealing probability */
                  /* of 0.1 during the search iterations, and an      */
                  /* annealing probability of 0.05 during the         */
                  /* refinement iterations.                           */
                  /*                                                  */
                  /* If you have factors with many levels with the    */
                  /* coordinate exchange algorithm, you should        */
                  /* specify smaller annealing probabilities.  For    */
                  /* example, a probability of 0.05 means 1 out of    */
                  /* 20 times the algorithm can make a swap that      */
                  /* decreases efficiency, which is too many with say */
                  /* a 25-level factor.                               */
                  /*                                                  */
maxiter=21 25 10, /* specifies the maximum number of iterations or    */
                  /* designs to generate.  With larger values, the    */
                  /* macro tends to find better designs at a cost of  */
                  /* much slower run times.  The first value can      */
                  /* optionally be followed by a second and a third   */
                  /* if you would like to use a different value for   */
                  /* the algorithm search, design search, and design  */
                  /* refinement iterations.  The second value is only */
                  /* used if the second set or iterations consists of */
                  /* level exchange iterations.  Otherwise, the       */
                  /* number of iterations for the second set is       */
                  /* specified with the TABITER=, or CANDITER= and    */
                  /* OPTITER= options.  If you want more iterations,  */
                  /* be sure to set the MAXTIME= option as well,      */
                  /* because iteration stops when the maximum number  */
                  /* of iterations is reached or the maximum amount   */
                  /* of time, whichever comes first.                  */
                  /* Examples:                                        */
                  /* MAXITER=10 specifies 10 iterations for the       */
                  /* initial, search, and refinement iterations.      */
                  /* MAXITER=10 10 5 specifies 10 initial iterations, */
                  /* followed by 10 search iterations, followed by 5  */
                  /* refinement iterations.                           */
iter=21 25 10,    /* Alias for MAXITER=.                              */
                  /*                                                  */
tabiter=10 200,   /* specifies the number of times to try to improve  */
                  /* a tabled (OA, fractional-factorial, or Hadamard  */
                  /* initial design.  The default is 10 tries in the  */
                  /* algorithm search iterations and 200 tries in the */
                  /* design search iterations.                        */
                  /*                                                  */
unbalanced=.2 .1, /* specifies the proportion of the TABITER=         */
                  /* iterations to consider using unbalanced factors  */
                  /* in the initial design.  The default is 0.2 0.1.  */
                  /*                                                  */
                  /* One way that unbalanced factors occur is through */
                  /* coding down.  Coding down for example creates a  */
                  /* three-level factor from a four-level factor:     */
                  /* (1 2 3 4) -> (1 2 3 3)                           */
                  /* or a two-level factor from a three-level factor: */
                  /* (1 2 3) -> (1 2 2).                              */
                  /* For any particular problem, chances are this     */
                  /* strategy is either going to work really well or  */
                  /* not well at all, without a lot of variability in */
                  /* the results, so it is not tried very often by    */
                  /* default.  This option will try to create         */
                  /* two-level through five-level factors from        */
                  /* three-level through six-level factors.  It will  */
                  /* not attempt for example to code down a           */
                  /* twenty-level factor into a nineteen-level factor */
                  /* (although the macro is often capable of in       */
                  /* effect doing this through level swaps).          */
                  /*                                                  */
canditer=1 1,     /* specifies the number of times to try to improve  */
                  /* a candidate-set based, OPTEX-generated initial   */
                  /* design (if one was created).  The default is 1   */
                  /* try in the initial iterations and 1 try in the   */
                  /* search iterations.  Note that OPTITER= controls  */
                  /* the number of OPTEX iterations.  This option     */
                  /* controls the number of times level swapping is   */
                  /* tried as a means of improving the OPTEX design.  */
                  /* Unless you are using annealing, mutation, or     */
                  /* OPTIONS=NODUPS, there is no reason to change     */
                  /* theses values.                                   */
                  /*                                                  */
mutiter=. . .,    /* specifies the first iteration to consider        */
                  /* mutating the design.  The default is the first   */
                  /* iteration that uses a random initial design in   */
                  /* each of the three sets of iterations.            */
                  /*                                                  */
anniter=. . .,    /* specifies the first iteration to consider using  */
                  /* annealing on the design.  The default is the     */
                  /* first iteration that uses a random initial       */
                  /* design in each of the three sets of iterations.  */
                  /*                                                  */
optiter=. .,      /* specifies the number of iterations to use in the */
                  /* OPTEX candidate set based searches in the        */
                  /* initial and search iterations.  By default, the  */
                  /* macro chooses values to use.  When the first     */
                  /* value is "." (missing), the macro will choose a  */
                  /* value usually no smaller than 20 for larger      */
                  /* problems and usually no larger than 200 for      */
                  /* smaller problems.  However, MAXTIME= values      */
                  /* other than the defaults can make the macro       */
                  /* choose values outside this range.  When the      */
                  /* second value is missing, the macro will choose a */
                  /* value based on how long the first OPTEX run took */
                  /* and the value of MAXTIME=, but no larger than    */
                  /* 5000.  When a missing value is specified for the */
                  /* first OPTITER= value, the default, the macro may */
                  /* choose to not perform any OPTEX iterations to    */
                  /* save time if it thinks it can find a perfect     */
                  /* design without them.                             */
                  /*                                                  */
maxtime=10 20 5,  /* specifies the approximate maximum amount of time */
                  /* in minutes to run each phase.  When an iteration */
                  /* completes (a design is completed), if more than  */
                  /* the specified amount of time has elapsed, the    */
                  /* macro quits iterating in that phase.  By         */
                  /* default, run time can be quite a bit larger than */
                  /* the value specified in MAXTIME=, particularly    */
                  /* for large problems, with restrictions, and with  */
                  /* EXCHANGE= values other than 1.  Alternatively, to*/
                  /* make large problems run more quickly, you can    */
                  /* specify OPTIONS=LARGEDESIGN, and the macro will  */
                  /* not wait until each iteration ends before        */
                  /* checking run time.  The first                    */
                  /* value can optionally be followed by a second and */
                  /* a third if you would like to use a different     */
                  /* value for the different sets of iterations.  By  */
                  /* default, the macro spends up to 10 minutes on    */
                  /* the algorithm search iterations, 20 minutes on   */
                  /* the design search iterations, and 5 in the       */
                  /* refinement stage.  Most problems run in much     */
                  /* less time than this.  Note that the second value */
                  /* is ignored for OPTEX iterations since OPTEX does */
                  /* not have any timing options.  This option also   */
                  /* affects when in the algorithm search iterations, */
                  /* the macro switches between using an OA initial   */
                  /* design to using a random initial design.  If the */
                  /* macro is not done using table design             */
                  /* initializations, and one half of the first time  */
                  /* value has passed, it switches.  Examples:        */
                  /* MAXTIME=60 specifies up to one hour for each     */
                  /* phase.  MAXTIME=20 30 10 specifies 20 minutes    */
                  /* for the first phase and 30 minutes for the       */
                  /* second, and 10 for the third.                    */
                  /*                                                  */
                  /* If you specify really large time values          */
                  /* (anything more than hours), you probably need to */
                  /* also specify OPTITER= since the default values   */
                  /* depend on MAXTIME=.                              */
                  /*                                                  */
                  /* MAXTIME=0 provides a way to get a quick run,     */
                  /* with no more than one iteration in each phase.   */
                  /* However, even with MAXTIME=0, run time can be    */
                  /* several minutes for large problems.              */
                  /*                                                  */
maxdesigns=,      /* specifies that the macro should stop after       */
                  /* MAXDESIGNS= designs have been created.  This     */
                  /* option may be useful for big, slow problems with */
                  /* restrictions.  You could specify MAXDESIGNS=3    */
                  /* and MAXTIME=0 and the macro would perform one    */
                  /* candidate set based iteration, one OA            */
                  /* initialization iteration, and one random         */
                  /* initialization iteration and then stop.  By      */
                  /* default, this option is ignored and stopping is  */
                  /* based on the other iteration options.            */
                  /*                                                  */
maxstages=3,      /* specifies that the macro should stop after       */
                  /* MAXSTAGES= algorithm stages have been completed. */
                  /* This option may be useful for big, slow problems */
                  /* with restrictions.  You could specify            */
                  /* MAXSTAGES=1 and the macro will stop after the    */
                  /* algorithm search stage, or MAXSTAGES=2 and the   */
                  /* macro will stop after the design search stage.   */
                  /* The default is MAXSTAGES=3, which means the      */
                  /* macro will stop after the design refinement      */
                  /* stage.                                           */
                  /*                                                  */
                  /*--------------------------------------------------*/
                  /*                                                  */
                  /* This set of options contains some miscellaneous  */
                  /* options that occasionally users may find useful. */
                  /*                                                  */
exchange=,        /* specifies the number of factors to consider at a */
                  /* time when exchanging levels.  You can specify    */
                  /* EXCHANGE=2 to do pair-wise exchanges.  Pair-wise */
                  /* exchanges are MUCH slower, but may produce       */
                  /* better designs.  The EXCHANGE= option interacts  */
                  /* with ORDER=.  ORDER=SEQRAN is faster with        */
                  /* EXCHANGE=2 than ORDER=SEQUENTIAL or              */
                  /* ORDER=RANDOM.  The default is EXCHANGE=2 when    */
                  /* PARTIAL= is specified.  With ORDER=MATRIX, the   */
                  /* EXCHANGE= value is the number of matrix columns. */
                  /* Otherwise, the default is EXCHANGE=1.            */
                  /*                                                  */
big=5185 choose,  /* specifies the MktDes BIG= option.  When the full */
                  /* factorial is this big or smaller, the MktEx      */
                  /* macro uses the MktDes macro to search a          */
                  /* full-factorial candidate set.  Otherwise it      */
                  /* searches a fractional-factorial candidate set.   */
                  /* The default is 5185.  When "choose" is specified */
                  /* as well, the default, the macro is allowed to    */
                  /* choose to use a fractional- factorial even if    */
                  /* the full-factorial is not too big, if it appears */
                  /* that the final design can be created from the    */
                  /* fractional factorial.                            */
                  /*                                                  */
target=100,       /* specifies the target efficiency criterion.  The  */
                  /* macro stops when it finds an efficiency value    */
                  /* greater than or equal to this number.  If you    */
                  /* know what the maximum efficiency criterion is,   */
                  /* or you know how big is big enough, you can       */
                  /* sometimes make the macro run faster by allowing  */
                  /* it to stop when it reaches the specified         */
                  /* efficiency.  You can also use this option if     */
                  /* you just want to see the initial design that     */
                  /* MktEx is using: TARGET=1, OPTITER=0              */
                  /* By specifying TARGET=1, the macro will stop      */
                  /* after the initialization as long as the initial  */
                  /* efficiency is GE 1.                              */
                  /*                                                  */
stopearly=5,      /* allows the macro to stop early when it keeps     */
                  /* finding the same maximum D-efficiency over and   */
                  /* over again in different designs.  By default,    */
                  /* during the design search iterations and          */
                  /* refinement iterations, the macro will stop early */
                  /* if 5 times, the macro finds a D-efficiency       */
                  /* essentially equal to the maximum but not greater */
                  /* than the maximum.  This almost always means that */
                  /* the macro has found the optimal design.  This    */
                  /* will only usually happen for small problems.     */
                  /* When the macro stops for this reason, the macro  */
                  /* will print                                       */
                  /*    NOTE: Stopping since it appears that no       */
                  /*          improvement is possible.                */
                  /* Specify either 0 or a very large value to turn   */
                  /* off the stop early checking.                     */
                  /*                                                  */
tabsize=,         /* allows you some control on which tabled (OA,     */
                  /* FACTEX or Hadamard) design is used for the       */
                  /* partial initialization when an exact match to a  */
                  /* tabled design is not found.  Specify the number  */
                  /* of runs in the tabled design.  See the CAT=      */
                  /* option for more detailed control.                */
                  /*                                                  */
fixed=,           /* When FIXED=variable is specified, the INIT= data */
                  /* set must contain the named variable, which       */
                  /* indicates which runs are fixed (cannot be        */
                  /* changed) and which ones may be changed.          */
                  /*                                                  */
                  /* 1 - (or any nonmissing) means this run may never */
                  /*     change.                                      */
                  /* 0 - means this run is used in the                */
                  /*     initial design, but it may be swapped out.   */
                  /* . - means this run should be randomly            */
                  /*     initialized, and it may be swapped out.      */
                  /*                                                  */
                  /* To more generally fix parts of the design, see   */
                  /* the INIT= option.                                */
                  /*                                                  */
holdouts=,        /* adds holdout observations to the INIT= data set. */
                  /* This option augments an initial design.          */
                  /* Specifying HOLDOUTS=n  adds n runs to the INIT=  */
                  /* design.  The option HOLDOUTS=n works by adding a */
                  /* FIXED= variable and extra runs to the INIT= data */
                  /* set.  Do not specify both FIXED= and HOLDOUTS=.  */
                  /* The number of rows in the INIT= design, plus the */
                  /* value specified in HOLDOUTS= must equal the N=   */
                  /* value.                                           */
                  /*                                                  */
order=,           /* specifies the order in which the columns are     */
                  /* looped over in the coordinate exchange           */
                  /* algorithm.  Valid values include:                */
                  /*                                                  */
                  /* sequential - 1, 2, 3, ...                        */
                  /*                                                  */
                  /* random - random order                            */
                  /*                                                  */
                  /* random=n - is like random, but with an           */
                  /* adaptation that is particularly useful for       */
                  /* partial-profile choice designs.  Use this option */
                  /* with EXCHANGE=2.  Say you are making a partial   */
                  /* profile design with 10 attributes and 3          */
                  /* alternatives.  Then attribute 1 is made from x1, */
                  /* x11, and x21; attribute 2 is made from x2, x12,  */
                  /* and x22; and so on.  Specifying ORDER=RANDOM=10  */
                  /* means that the columns, as shown by column index */
                  /* j1, are traversed in a random order.  A second   */
                  /* loop (with variable j2) traverses all of the     */
                  /* factors in the current attribute.  So for        */
                  /* example when j1 is 13, then j2 = 3, 13, 23.      */
                  /* This allows pair-wise exchanges within choice    */
                  /* attributes.                                      */
                  /*                                                  */
                  /* ranseq (or seqran) - sequential from a random    */
                  /* first column                                     */
                  /*                                                  */
                  /* col=n - picks n random columns to work on for    */
                  /* each row.  Specify an integer for n, for example */
                  /* ORDER=COL=2.  This should only be used for HUGE  */
                  /* problems where you do not care if you hit every  */
                  /* column.  Typically, this option will be used in  */
                  /* conjunction with OPTIONS=LARGEDESIGN,            */
                  /* MAXDESIGNS=1, OPTITER=0, TABITER=0.  You would   */
                  /* use it when you have a large problem and you do  */
                  /* not have enough time for one complete pass       */
                  /* through the design.  You just want to iterate    */
                  /* for approximately the MAXTIME= amount of time    */
                  /* then stop.  You should not use COL= with         */
                  /* restrictions.                                    */
                  /*                                                  */
                  /* matrix=SAS-data-set - allows you to specify      */
                  /* exactly what columns are worked on in what order */
                  /* and in what groups.  To use this option, specify */
                  /* ORDER=MATRIX followed by the name of a SAS data  */
                  /* set, e.g.  ORDER=MATRIX=MYDATA.  In the data set */
                  /* MYDATA provide one row for every column          */
                  /* grouping.  Say you want to work on columns in    */
                  /* pairs.  Then the data set would have two         */
                  /* columns.  The first variable would contain the   */
                  /* number of a design column and the second         */
                  /* variable would contain the number of a second    */
                  /* column that is to be exchanged with the first.   */
                  /* The EXCHANGE= value is ignored and is set to two */
                  /* because the data set has two columns.  The       */
                  /* values must be integers between 1 and m, where m */
                  /* is the number of factors.  The values may also   */
                  /* be missing, except in the first column.  Missing */
                  /* values are replaced by a random column           */
                  /* (potentially a different random column each      */
                  /* time).  This can sometimes be enourmously        */
                  /* helpful.                                         */
                  /*                                                  */
                  /* null, the default - random when there are        */
                  /* partial profile restrictions, ranseq when there  */
                  /* are other restrictions, and sequential           */
                  /* otherwise.                                       */
                  /*                                                  */
                  /* ORDER= interacts with EXCHANGE=.  With a random  */
                  /* order and EXCHANGE=2, the variable j1 loops over */
                  /* the columns of the design in a random order and  */
                  /* for each j1, j2 loops over the columns greater   */
                  /* than j1 in a random order.  With a sequential    */
                  /* order and EXCHANGE=2, the variable j1 loops over */
                  /* the columns in 1, 2, 3 order and for each j1, j2 */
                  /* loops over the columns greater than j1 in a      */
                  /* j1+1, j1+2, j1+3 order.  ORDER=RANSEQ is a bit   */
                  /* different.  With EXCHANGE=2, the variable j1     */
                  /* loops over the columns in an order r, r+1, r+2,  */
                  /* ..., m, 1, 2, ..., r-1 (for random r), and for   */
                  /* each j1 there is a single random j2.  Hence,     */
                  /* ORDER=RANSEQ is the fastest option with          */
                  /* EXCHANGE=2 since it does not consider all pairs, */
                  /* just one pair.  ORDER=RANSEQ is the only         */
                  /* situation where you might try EXCHANGE=3.        */
                  /*                                                  */
                  /*--------------------------------------------------*/
                  /*                                                  */
                  /* This last set of options contains all of the     */
                  /* other miscellaneous options.  Most of the time,  */
                  /* most people should not specify options from      */
                  /* this list.                                       */
                  /*                                                  */
annealfun=        /* specifies the function that controls how the     */
anneal * 0.85,    /* simulated annealing probability changes with     */
                  /* each pass through the design.  Most people       */
                  /* should not use this capability most of the time. */
                  /*                                                  */
imlopts=,         /* specifies IML PROC statement options.  For       */
                  /* example, for very large problems, you can use    */
                  /* this option to specify the IML SYMSIZE= or       */
                  /* WORKSIZE= options: IMLOPTS=SYMSIZE=n WORKSIZE=m, */
                  /* substituting numeric values for n and m.  The    */
                  /* defaults for these options is host dependent.    */
                  /* This option is rarely needed.                    */
                  /*                                                  */
ridge=1e-7,       /* specifies the value to add to the diagonal of    */
                  /* X'X to make it nonsingular.  Usually, for normal */
                  /* problems, you will not need to change this       */
                  /* value.  The default is 1e-7.  If you want the    */
                  /* macro to create designs with more parameters     */
                  /* than runs, you must specify some other value,    */
                  /* usually something like 0.01.  By default, the    */
                  /* macro will quit when there are more parameters   */
                  /* than runs.  Specifying a RIDGE= value other than */
                  /* the default (even if you just upper case the     */
                  /* "e") allows the macro to create a design with    */
                  /* more parameters than runs.  Most people should   */
                  /* not use this capability most of the time.        */
                  /*                                                  */
refreshat=47,     /* specifies how often to refresh the determinant   */
                  /* update calculations.  Larger values make the     */
                  /* algorithm run faster with the risk of an         */
                  /* increased accumulation of rounding error.  Very  */
                  /* rarely it may be the case that you should        */
                  /* specify a smaller value to get more accurate     */
                  /* results.                                         */
                  /*                                                  */
detfuzz=1e-8,     /* specifies the value used to determine if         */
                  /* determinants are changing.                       */
                  /* If newdeter > olddeter * (1 + detfuzz)           */
                  /* then the new determinant is larger.              */
                  /* If newdeter > olddeter * (1 - detfuzz)           */
                  /* then the new determinant is the same.            */
                  /* Otherwise the new determinant is smaller.        */
                  /* Most people should not change this value most of */
                  /* the time.                                        */
                  /*                                                  */
repeat=25 . .,    /* specifies the maximum number of times to         */
                  /* repeatedly work on a row to eliminate            */
                  /* restriction violations.  The default value of    */
                  /* REPEAT=25 . ., specifies that a row should be    */
                  /* worked on up to 25 times to eliminate            */
                  /* violations.  The second value is the place in    */
                  /* the design refinement where this processesing    */
                  /* starts.  This is based on a zero-based total     */
                  /* number of rows processed so far.  This is like a */
                  /* zero-based row index, but it never resets within */
                  /* a design.  The third value is the place where    */
                  /* this extra repeated processing stops.  Let m be  */
                  /* the MINTRY=m value, which by default is n, the   */
                  /* number of rows.  By default, when the second and */
                  /* third values are missing, the processs starts    */
                  /* after m rows have been processed (the second     */
                  /* complete pass through the design) and stops      */
                  /* after m + 10 * n rows have been processed.       */
                  /*                                                  */
);                /*--------------------------------------------------*/
                  /* This macro specifies OPTIONS NONOTES throughout  */
                  /* most of its execution.  If you want to see all   */
                  /* of the notes, submit the statement               */
                  /*    %let mktopts = notes;                         */
                  /* before running the macro.                        */
                  /*--------------------------------------------------*/

;*=============================== Macro Start =================================;

*------initial macro stuff------;
%global _MktEff _MktTime1 _MktLoc1 _MktVer _MktIntList
        _MktList _MktSat _MktFF _MktLogFF;
%local  _optseed abort accept allfactex almost1 anneal2 anneal3 annealflag
        anniter2 anniter3 anyneg balancetry big1 bigkid canditer2 candnlev
        debugextra destype difflag do512 dosort dropvars dupflag elefixed ex
        examineinfo examinevar exchangedat facdone factexnlevs factorslist
        finaleff first firstpass flow flowinit full generic genfile hadflag
        hadsize header history i initdata initrand int interactmat j jj ind
        justinit k largedesign last levels2 levels3 lineage localver ls ltfr
        m main maxiter2 maxiter3 maxlen maxlev maxlevfound maxlevneed maxtime2
        maxtime3 method missex mutate2 mutate3 mutateflag mutiter2
        mutiter3 ncands nd2 nm1 nointeract nostimer npars nrunsintab optabort
        optflag optiter2 opttime ordcoln order outallnum outallopts outopts
        outropts over1 p partinc pass pfinal plineage plusmin2 plusmin3
        powerof primes printseed ps r5 ranseed readinit render renderdif
        renfor resetinit resflag resrep restrictflag ridgemsg rowsfixed
        savenote savedsnf setnote setoptiter showcode silent somenotes spacer
        sqlobs sqloops sqlrc sqlxobs stepnum steptime stopearly tabiter2
        tableflag tablenote termlen time tryfac unbalanced2 userrestricts
        varlist sysranend optvarlist aliasing model adhochad nooadups noqc
        catmin catmax nox outeffopts bestmethod nreps saveopts timeopt
        _trgind _trgindn fullaliasing inaliasing mainaliasing;

%let time         = %sysfunc(datetime());
%let steptime     = &time;
%let stepnum      = 0;
*mkttime(Macro Start);

%let timeopt    = %scan(STIMER STATS, 1 + (&sysscp = OS));
%let savenote   = %sysfunc(getoption(notes));
%let saveopts   = &savenote missing="%sysfunc(getoption(missing))"
                  %sysfunc(getoption(generic)) %sysfunc(getoption(&timeopt))
                  %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr))
                  %sysfunc(getoption(serror));
%let abort      = 0;
%let maxlen     = 32700;
%let termlen    = 500;
%let setnote    = nonotes;
%let primes     =  2,  3,  5,  7, 11, 13, 17, 19, 23, 29, 31, 37, 41,
                  43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97;
%let showcode   = nocode;
%let printseed  = 1;
%let renderdif  = 0;
%let flow       = 0;
%let flowinit   = 0;
%let silent     = 0;
%let debugextra = 0;
%let difflag    = 0;
%let nostimer   = 0;
%let localver   = 25Jul2010;
%let _mktver    = &localver;

options noserror;
%if %nrquote(&&mktexopts) eq %nrstr(&)mktexopts %then %let mktexopts = ;
%if %nrquote(&&mktopts) ne %nrstr(&)mktopts %then %do;
   %let k = %upcase(&mktopts);
   %if %index(&k, CHECKVERSION) %then %do; options serror; %goto mendit; %end;
   %if %index(&k, NOTES)   %then %let setnote = notes;
   %if %index(&k, VERSION) %then %put MktEx macro version &localver.;
   %if %index(&k, CODE)    %then %let showcode = ;
   %let i = %index(&k, TITLE);
   %if &i %then %do;
      %substr(&mktopts%str( ), &i, 6)
           '%' "mktex(    %sysfunc(compbl(&list%str( )))  ,  n=&n    )";
      %end;
   %let flowinit   = (%index(&k, FLOWINIT)  gt 0);
   %let flow       = (%index(&k, FLOW)      gt 0);
   %let silent     = (%index(&k, SILENT)    gt 0);
   %let debugextra = (%index(&k, EXTRA)     gt 0);
   %let printseed  = (%index(&k, NOSEED)    eq 0);
   %let renderdif  = (%index(&k, RENDERDIF) gt 0);
   %if &flowinit %then %let flow = 0;
   %end;
options serror;

options &setnote generic missing='.';

data _null_;
   length list $ 32;
   list = left(lowcase(symget('list')));
   if list eq 'help' or list eq '?' then call symput('abort', '-1');
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort = 1 %then %goto endit;

%if &abort = -1 %then %do; %let abort = 0; %mkthelp(mktex) %goto endit; %end;


*------preliminary option setting------;
data _null_;
   length name $ 16 list $ &termlen glob $ &maxlen;
   abort = 0;  powerof = 0;

   call symput('inaliasing',  '0');
   call symput('maxlev',      '1000');
   call symput('maxlevfound', '1000');
   call symput('bigkid',      '144');
   call symput('_mkteff',     '.');

   call symput('elefixed',    '0');
   call symput('anyneg',      '0');
   call symput('catmin',      '0');
   call symput('catmax',      '0');
   call symput('header',      'title1 ');

   *------allow for a quick run, without all the bells and whistles------;
   list = lowcase(symget('options'));
   if index(list, 'quick') then do;
      call symput('optiter', '0');
      call symput('maxdesigns', '1');
      call symput('unbalanced', '0');
      call symput('tabiter', '1');
      if index(list, 'quickr') then call symput('tabiter', '0');
      else if not index(list, 'quickt') then call symput('maxdesigns', '2');
      end;

   *------number of runs------;
   nruns = &n +0; /* allow expressions like 2*12 */
   if nmiss(nruns) or nruns lt 2 or nruns ne round(nruns)
      or nruns gt 268435455 then do;
      put "ERROR: N=%nrbquote(&n) is invalid.";
      abort = 1;
      end;

   *------is the number of runs a power of a prime?------;
   if not abort then do;
      call symput('n', compress(put(nruns, 12.)));
      p = 0;
      do powerof = &primes;
         lognruns = log(nruns) / log(powerof);
         if abs(lognruns - round(lognruns)) lt 1e-8 then do; p = 1; leave; end;
         end;
      if not p then powerof = 0;
      if &flow then put '** preliminary ' powerof=;
      end;

   *------default list OA designs?-------;
   list = compbl(symget('list'));
   if not abort and list eq ' ' then do;
      if powerof then list = compress(put(powerof, 2.)) || ' ** ' ||
                             put((nruns - 1) / (powerof - 1), 12. -L);
      else if nruns =  18 then list = '2 3 ** 7';
      else if nruns =  36 then list = '2 ** 11 3 ** 12';
      else if nruns =  48 then list = '2 ** 11 4 ** 12';
      else if nruns =  50 then list = '2 5 ** 11';
      else if nruns =  54 then list = '2 3 ** 25';
      else if nruns =  72 then list = '2 ** 23 3 ** 24';
      else if nruns = 108 then list = '2 ** 35 3 ** 36';
      else if nruns = 144 then list = '2 ** 88 3 ** 25';
      else if nruns = 162 then list = '2 3 ** 79';
      else if nruns = 192 then list = '2 ** 145 3 4 ** 13';
      else if nruns = 216 then list = '2 ** 71 3 ** 72';
      else if nruns = 288 then list = '2 ** 232 3 ** 25';
      else if nruns = 324 then list = '2 ** 35 3 ** 144';
      else if nruns = 432 then list = '2 ** 143 3 ** 144';
      else if nruns = 486 then list = '2 3 ** 241';
      else if mod(nruns, 4) eq 0 then
         list = '2 ** ' || put(nruns - 1, 12. -L);
      if list eq ' ' then do;
         n = nruns;
         do i = &primes;
            do while(mod(n, i) eq 0);
               list = compbl(trim(list) || put(i, 3.));
               n = n / i;
               end;
            end;
         if n gt 1 then list = ' ';
         end;
      if list eq ' ' then do;
         put 'ERROR: The factor-level list must be specified with N='
             nruns +(-1) '.';
         call symput('abort', '1');
         end;
      end;
   if not abort then do;
      call symput('list', trim(left(list)));
      if &flow then put '** ' list=;
      end;

   *------check system options------;
   notes = (symget('savenote' ) eq 'NOTES');
   call symput('somenotes', put(notes, 1.));
   call symput('generic',   put(not index(symget('saveopts'), 'NOGENERIC'), 1.));
   call symput('nostimer',  put(getoption("&timeopt" ) eq "NO&timeopt", 1.));

   *------override parameters?-------;
   glob = left(symget('mktexopts'));
   if notes and not &silent and glob ne ' ' then put 'Overridden Parameters:';
   do while(glob ne ' ');
      i = index(glob, '=');
      name = substr(glob, 1, i - 1);
      glob = left(substr(glob, i + 1));
      i = index(glob, ',');
      if i = 0 then i = length(glob) + 1;
      list = substr(glob, 1, i);
      list = translate(list, ' ', ',');
      glob = left(substr(glob, i + 1));
      if notes and not &silent then put +3 name +(-1) '=' list;
      call symput(name, trim(list));
      end;

   call symput('initdata', put(symget('init') ne ' ', 1.));

   /*---resolution 5 design?---*/
   call symput('r5', symget('interact') eq '@2');

   if index(lowcase(symget('options')), 'noqc') then do;
      if (sysprod('iml') + sysprod('stat')) ne 2 then do;
         put 'ERROR: SAS/STAT and SAS/IML must be licensed '
             'to run this macro.';
         abort = 1;
         end;
      end;
   else do;
      if (sysprod('iml') + sysprod('qc') + sysprod('stat')) ne 3 then do;
         put 'ERROR: SAS/STAT, SAS/QC, and SAS/IML must be licensed '
             'to run this macro.';
         abort = 1;
         end;
      end;

   if abort or _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------copy initial data, apply any data set options, holdouts------;
%if &initdata %then %do;

   data _null_;
      length str $ 23;
      abort = 0;
      if 0 then set &init nobs=__nobs;
      str = symget('holdouts');
      n = input(str, ?? 23.);
      if n(n) then do;

         if symget('fixed') ne ' ' then do;
            put 'ERROR: Do not specify both HOLDOUTS= and FIXED=.';
            abort = 1;
            end;

         if &n - __nobs ne n then do;
            put "ERROR: HOLDOUTS=&holdouts is invalid with N=&n and "
                __nobs 'initial runs.';
            abort = 1;
            end;

         call symput('fixed', 'w');
         end;

      if str ne ' ' and nmiss(n) then do;
         put "ERROR: HOLDOUTS=&holdouts is invalid.";
         abort = 1;
         end;

      if abort or _error_ then call symput('abort', '1');
      stop;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data __tmpxinitdat(drop=__j
                      %if %nrbquote(&fixed) ne %then rename=(__fixed=&fixed););
      set &init %if %nrbquote(&fixed) ne and
                    %nrbquote(&holdouts) eq %then (rename=(&fixed=__fixed));
          end=__eof;
      array __x[*] x:;
      do __j = 1 to dim(__x);
         if nmiss(__x[__j]) then __x[__j] = 0;
         if __x[__j] le 0 then call symput('elefixed', '1');
         if __x[__j] lt 0 then call symput('anyneg'  , '1');
         end;

      %if &holdouts ne %then %do; retain __fixed 1; %end;

      output;

      %if &holdouts ne %then %do;
         if __eof then do;
            __fixed = .;
            do __i = 1 to &holdouts; output; end;
            end;
         drop __i;
         %end;

      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %let init = __tmpxinitdat;
   %end;

*------expand out list to _mktlist, a list of integers------;
*mkttime(MktRuns Start);

%let _mktver = obsolete;
%mktruns(&list, options=justparse, out=__tmpxnums,
         interact=%nrbquote(&interact));

%if &_mktver ne &localver %then %do;
   %put ERROR: Incompatible MktRuns macro.;
   %put %str(      ) You must have the same version of all macros.;
   %put %str(      ) Versions: MktEx = &localver, MktRuns = &_mktver..;
   %put %str(      ) Get the latest tools and documentation from:;
   %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
   %let abort = 1;
   %end;

%if &syserr > 4 or %nrbquote(&_mktlist) eq ABORT %then %let abort = 1;
%if &abort %then %goto endit;

*mkttime(Printing Overhead and Options);

*------count titles and footnotes------;
proc sql; create view __tmpxtf as select * from dictionary.titles; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------get effective page size for printing------;
data _null_;
   if _n_ = 1 then do;
      ps = input(getoption('pagesize'), ?? 23.);
      call symput('ps', compress(put(ps - 1, 12.)));
      ls = input(getoption('linesize'), ?? 23.);
      ind = floor((ls - (58 + 2 * (ls le 65))) / 2);
      call symput('ind', compress(put(ind, 12.)));
      call symput('ls', compress(put(ls, 12.)));
      end;
   set __tmpxtf end=eof;
   titles + type eq 'T';
   foots  + type eq 'F';
   if eof then do;
      ps = input(getoption('pagesize'), ?? 23.);
      ps = ps - titles - foots - (titles gt 0) - (foots gt 0) - 1;
      if ps lt 10 then ps = 1e10;
      call symput('ps', compress(put(ps, 12.)));
      call symput('header', 'title' || put(titles + 1, 2.0 -L));
      end;
   if _error_ then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------check and process most options------;
data _null_;
   length varlist newlist mktlist list hold opts $ &maxlen w w2 w3 $ 46;
   length str $ &termlen t1-t5 $ 2;
   array t[5];

   abort = 0; fullrank = 1;

   call symput('main', '0');
   call symput('tablenote', '0');
   call symput('factexnlevs', '0');
   call symput('outallnum', '0');
   call symput('spacer', '**');
   call symput('destype', 'random');
   call symput('finaleff', '100.0000');
   call symput('missex', '0');
   call symput('rowsfixed', '0');
   call symput('balancetry', '&n');
   call symput('aliasing', '0');

   *------EXAMINE=------;
   w = lowcase(compbl(symget('examine')));
   i = index(w, 'aliasing');
   if i then substr(w, i + 5, 3) = ' ';
   i = index(w, 'alias');
   j = index(w, 'full');
   k = index(w, 'main');
   if j then substr(w, j, 4) = ' ';
   if k then substr(w, k, 4) = ' ';

   if (j or k) and not i then do;
      w = trim(w) || ' alias=1';
      i = index(w, 'alias');
      end;
   call symput('fullaliasing', put(j gt 0, 1.));
   call symput('mainaliasing', put(k gt 0, 1.));

   if i then do;
      substr(w, i, 5) = ' ';
      w2 = scan(substr(w, i), 1, '= ');
      n  = input(w2, ?? 23.);
      if nmiss(n) or n ne floor(n) then do;
         put "ERROR: ALIASING= specification in"
             " EXAMINE=%nrbquote(&examine) is invalid.";
         abort = 2;
         end;
      call symput('aliasing', trim(w2));
      i = index(w, '=');
      if i then substr(w, i, 1) = ' ';
      i = index(w, trim(w2));
      if i then substr(w, i, length(w2)) = ' ';
      end;

   i = index(w, 'i');
   v = index(w, 'v');
   if i then substr(w, i, 1) = ' ';
   if v then substr(w, v, 1) = ' ';
   call symput('examineinfo', put(i ne 0, 1.));
   call symput('examinevar' , put(v ne 0, 1.));

   if abort ne 2 and w ne ' ' then do;
      w = compbl(w);
      put 'ERROR: EXAMINE=' w 'is invalid.';
      abort = 1;
      end;

   w = ' ';
   if i then w = 'i';
   if v then w = trim(w) || ' v';
   call symput('examine', trim(w));

   *------refresh rate------;
   n = input(symget('refreshat'), ?? 23.);
   if nmiss(n) or n lt 1 then do;
      put "ERROR: REFRESHAT=%nrbquote(&refreshat) is invalid.";
      abort = 1;
      end;

   *------determinant improvement fudge factor------;
   n = input(symget('detfuzz'), ?? 23.);
   if nmiss(n) or n lt 0 or n gt 0.1 then do;
      put "ERROR: DETFUZZ=%nrbquote(&detfuzz) is invalid.";
      abort = 1;
      end;
   else do;
      call symput('almost1', compress(put(1 - n, best20.)));
      call symput('over1'  , compress(put(1 + n, best20.)));
      end;

   *------check for lists that are too long------;
   if scan(symget('mutate'), 4, ' ') ne ' ' then do;
      put "ERROR: MUTATE=%nrbquote(&mutate) is invalid.";
      abort = 1;
      end;
   if scan(symget('anneal'), 4, ' ') ne ' ' then do;
      put "ERROR: ANNEAL=%nrbquote(&anneal) is invalid.";
      abort = 1;
      end;
   if scan(symget('maxiter'), 4, ' ') ne ' ' then do;
      put "ERROR: MAXITER=%nrbquote(&maxiter) is invalid.";
      abort = 1;
      end;
   if scan(symget('iter'), 4, ' ') ne ' ' then do;
      put "ERROR: ITER=%nrbquote(&iter) is invalid.";
      abort = 1;
      end;
   if scan(symget('tabiter'), 3, ' ') ne ' ' then do;
      put "ERROR: TABITER=%nrbquote(&tabiter) is invalid.";
      abort = 1;
      end;
   if scan(symget('unbalanced'), 3, ' ') ne ' ' then do;
      put "ERROR: UNBALANCED=%nrbquote(&unbalanced) is invalid.";
      abort = 1;
      end;
   if scan(symget('canditer'), 3, ' ') ne ' ' then do;
      put "ERROR: CANDITER=%nrbquote(&canditer) is invalid.";
      abort = 1;
      end;
   if scan(symget('mutiter'), 4, ' ') ne ' ' then do;
      put "ERROR: MUTITER=%nrbquote(&mutiter) is invalid.";
      abort = 1;
      end;
   if scan(symget('anniter'), 4, ' ') ne ' ' then do;
      put "ERROR: ANNITER=%nrbquote(&anniter) is invalid.";
      abort = 1;
      end;
   if scan(symget('optiter'), 3, ' ') ne ' ' then do;
      put "ERROR: OPTITER=%nrbquote(&optiter) is invalid.";
      abort = 1;
      end;
   if scan(symget('maxtime'), 4, ' ') ne ' ' then do;
      put "ERROR: MAXTIME=%nrbquote(&maxtime) is invalid.";
      abort = 1;
      end;
   if scan(symget('big'), 3, ' ') ne ' ' then do;
      put "ERROR: BIG=%nrbquote(&big) is invalid.";
      abort = 1;
      end;

   *------start processing factors list------;
   mktlist = symget('_mktlist');

   *------count levels of factors------;
   start = 1; c = input(scan(mktlist, 1, ' '), ?? 23.);

   maxlevfound = 0;
   do nvars = 1 to &maxlen until(w eq ' ');
      w = scan(mktlist, nvars, ' ');
      if w ne ' ' then do;
         n = input(w, ?? 23.);
         maxlevfound = max(maxlevfound, n);
         end;
      if w eq ' ' or n ne c then do;
         if (nvars - 1) gt start
            then varlist = trim(varlist) || ' ' ||
                           compress('x' || put(start, 5.) || '-x' ||
                                    put(nvars  - 1, 5.) || '=' || put(c, 5.));
         else varlist = trim(varlist) || ' ' ||
                        compress('x'  || put(start, 5.) || '=' || put(c, 5.));
         start = nvars;
         c = n;
         end;
      end;

   *------just do level swaps for silly designs------;
   if maxlevfound gt &maxlev then do;
      call symput('optiter', '0');
      call symput('tabiter', '0');
      maxlevfound = &maxlev;
      end;

   call symput('maxlevfound', compress(put(maxlevfound, 23.)));
   call symput('maxlevneed', compress(put(max(6, maxlevfound), 23.)));

   *------set number of factors------;
   nvars + -1;     npars = &_mktsat;
   call symput('m', compress(put(nvars, 5.)));

   *------number of observations------;
   nruns = input(symget('n'), ?? 23.);

   *------check TABSIZE=------;
   w = symget('tabsize');
   if w ne ' ' then do;
      n = input(w, ?? 23.);
      if n ne round(n) or n lt 2 then do;
         put "ERROR: TABSIZE=%nrbquote(&tabsize) is invalid.";
         abort = 1;
         end;
      end;
   else call symput('tabsize', '0');

   *------silly design------;
   if nruns lt npars then fullrank = 0;

   if &flow then put '** ' nruns= nvars= npars= fullrank=;

   *------check BIG=------;
   big = input(scan(symget('big'), 1, ' '), ?? 23.);
   w   = lowcase(scan(symget('big'), 2, ' '));
   if nmiss(big) or big ne round(big) or big lt 1 or
      (w ne ' ' and w ne 'choose') then do;
      put "ERROR: BIG=%nrbquote(&big) is invalid.";
      abort = 1;
      big = 1;
      end;
   small = (&_mktlogff le log10(big));

   call symput('big', compress(put(big, best12.)));
   call symput('big1', put(w eq 'choose', 1.));

   *------use FACTEX with smaller designs with interactions------;
   call symput('tryfac', w eq 'choose' and symget('interact') ne ' ' and
               symget('restrictions') eq ' ' and symget('partial') eq ' '
               and small and symget('balance') eq ' ' and
               abs(log2(nruns) - round(log2(nruns))) le 1e-8);

   *------check MAXDESIGNS=------;
   w = symget('maxdesigns');
   if w ne ' ' then do;
      n = input(w, ?? 23.);
      if nmiss(n) or n lt 0 or n ne round(n) then do;
         put "ERROR: MAXDESIGNS=%nrbquote(&maxdesigns) is invalid.";
         abort = 1;
         end;
      end;

   *------check MAXSTAGES=------;
   w = symget('maxstages');
   if w eq ' ' then call symput('maxstages', '3');
   else do;
      n = input(w, ?? 23.);
      if nmiss(n) or n lt 1 or n ne round(n) then do;
         put "ERROR: MAXSTAGES=%nrbquote(&maxstages) is invalid.";
         abort = 1;
         end;
      end;

   *------check REPEAT=------;
   w = symget('repeat');
   if w eq ' ' then w = '25 . .';
   w2 = scan(w, 2, ' '); if w2 = ' ' then w2 = '.';
   w3 = scan(w, 3, ' '); if w3 = ' ' then w3 = '.';
   w  = scan(w, 1, ' ');
   n = input(w, ?? 23.);
   if w2 eq '.' then w2 = symget('mintry');
   if w2 eq ' ' then do; w2 = 'n'; call symput('mintry', 'n'); end;
   if w3 eq '.' then w3 = trim(w2) || ' + n10';
   call symput('repeat', trim(w) || ' $ ' || trim(w2) || ' $ ' || trim(w3));

   if nmiss(n) or n lt 1 or n ne round(n) then do;
      put "ERROR: REPEAT=%nrbquote(&repeat) is invalid.";
      abort = 1;
      end;

   *-------check RESTRICTIONS=, RESLIST=, RESMAC=------;
   w  = symget('restrictions');
   w2 = symget('reslist');
   w3 = symget('resmac');
   if (w2 ne ' ' or w3 ne ' ') and (w eq ' ' or w2 eq ' ' or w3 eq ' ') then do;
      put 'ERROR: If either RESLIST= or RESMAC= is specified,' / +8
          'then both need to be specified and so does RESTRICTIONS=.';
      abort = 1;
      end;
   call symput('resflag', put(w2 ne ' ', 1.));

   *-------check RESTRICTIONS=, BALANCED=, PARTIAL=------;
   w  = symget('restrictions');
   w2 = symget('partial');
   w3 = symget('balance');
   n  = w ne ' ' or w2 ne ' ' or w3 ne ' ';
   call symput('restrictflag' , put(n, 1.));
   call symput('userrestricts', put(w ne ' ', 1.));
   if w2 ne ' ' then do;
      call symput('tabiter', '0');
      if scan(symget('optiter'), 1, ' ') = '.' then call symput('optiter', '0');
      n = input(w2, ?? 23.);
      if nmiss(n) or n lt 1 or n ne round(n) then do;
         put "ERROR: PARTIAL=%nrbquote(&partial) is invalid.";
         abort = 1;
         end;
      else if n ge nvars then do;
         put "ERROR: PARTIAL=%nrbquote(&partial) is too large.";
         abort = 1;
         end;
      end;

   if w3 ne ' ' then do;
      n = input(w3, ?? 23.);
      if nmiss(n) or n lt 0 then do;
         put "ERROR: BALANCE=%nrbquote(&balance) is invalid.";
         abort = 1;
         end;
      else if symget('mintry') eq 'n' then
         put 'WARNING: When BALANCE= is specified, you should also'
             ' specify MINTRY=.'  ' The two options together'
             ' often work better than just BALANCE= alone.';
      end;

   *------maximum time parameters------;
   opts = symget('maxtime');
   if opts eq ' ' then opts = '10 20 5';
   w    = scan(opts, 1, ' ');     maxtime  = input(w , ?? 23.);
   w2   = scan(opts, 2, ' ');     maxtime2 = input(w2, ?? 23.);
   w3   = scan(opts, 3, ' ');     maxtime3 = input(w3, ?? 23.);
   if w2 eq ' ' then do; w2 = w;  maxtime2 = maxtime;  end;
   if w3 eq ' ' then do; w3 = w2; maxtime3 = maxtime2; end;
   if (w  ne ' ' and (nmiss(maxtime ) or maxtime  lt 0)) or
      (w2 ne ' ' and (nmiss(maxtime2) or maxtime2 lt 0)) or
      (w3 ne ' ' and (nmiss(maxtime3) or maxtime3 lt 0)) then do;
      put "ERROR: MAXTIME=%nrbquote(&maxtime) is invalid.";
      abort = 1; maxtime = 10; maxtime2 = 20;
      end;
   call symput('maxtime' , compress(w ));
   call symput('maxtime2', compress(w2));
   call symput('maxtime3', compress(w3));

   *------number of iterations to work on an OA------;
   opts = symget('tabiter');
   if opts eq ' ' then opts = '5 20';
   w  = scan(opts, 1, ' ');      tabiter  = input(w , ?? 23.);
   w2 = scan(opts, 2, ' ');      tabiter2 = input(w2, ?? 23.);
   if w2 eq ' ' then do; w2 = w; tabiter2 = tabiter;  end;
   if nmiss(tabiter ) or tabiter  lt 0 or tabiter  ne round(tabiter ) or
      nmiss(tabiter2) or tabiter2 lt 0 or tabiter2 ne round(tabiter2) then do;
      put "ERROR: TABITER=%nrbquote(&tabiter) is invalid.";
      abort = 1;
      end;
   call symput('tabiter' , compress(put(tabiter , 12.0)));
   call symput('tabiter2', compress(put(tabiter2, 12.0)));

   *------number of iterations for MktDes and OPTEX------;
   opts = left(symget('optiter'));
   if opts eq ' ' then opts = '0';
   w  = scan(opts, 1, ' ');      optiter  = input(w , ?? 23.);
   w2 = scan(opts, 2, ' ');      optiter2 = input(w2, ?? 23.);
   if w2 eq ' ' then do; w2 = w; optiter2 = optiter;  end;

   n = 2 * (w ne '.') + (w2 ne '.'); /* 0: . ., 1: . x, 2: x ., 3: x x */
   call symput('setoptiter', put(n, 1.));
   call symput('opttime', '1');

   n2 = &_mktlogff;
   if small then n = max(0.003 * (10 ** n2) + 0.07 * &n - 2.5, 0.1);

   if w = '.'  then do;
      if small then optiter = min(max(round(600 / n, 10), 20), 200);
      else          optiter = 20;
      optiter = ceil(maxtime * optiter / 10);    /* 10 - default maxtime */
      if maxtime eq 0 then optiter = 1;
      end;
   if w2 = '.' then do;
      if small then optiter2 = min(max(round(6000 / n, 100), 200), 1000);
      else          optiter2 = 200;
      optiter2 = ceil(maxtime2 * optiter2 / 20); /* 20 - default maxtime */
      optiter2 = min(optiter2, 1000);
      if maxtime2 eq 0 then optiter2 = 1;
      end;

   if nmiss(optiter ) or optiter  lt 0 or optiter  ne round(optiter ) or
      nmiss(optiter2) or optiter2 lt 0 or optiter2 ne round(optiter2) then do;
      put "ERROR: OPTITER=%nrbquote(&optiter) is invalid.";
      abort = 1;
      end;

   call symput('optiter' , compress(compress(put(optiter , 12.0))));
   call symput('optiter2', compress(compress(put(optiter2, 12.0))));
   call symput('optabort', '0');

   *------check OPTIONS=------;
   list = lowcase(symget('options'));
   i = index(list, 'quickt'); if i then substr(list, i, 6) = ' ';
   i = index(list, 'quickr'); if i then substr(list, i, 6) = ' ';
   i = index(list, 'quick');  if i then substr(list, i, 5) = ' ';
   i = index(list, 'refine'); if i then substr(list, i, 6) = ' ';
   call symput('resetinit', put(i eq 0, 1.));
   i = index(list, '512'); if i then substr(list, i, 3) = ' ';
   call symput('do512', put(i gt 0, 1.));
   i = index(list, 'nox'); if i then substr(list, i, 3) = ' ';
   call symput('nox', put(i gt 0, 1.));
   i = index(list, 'lineage'); if i then substr(list, i, 7) = ' ';
   call symput('plineage', put(i gt 0, 1.));
   i = index(list, 'nooadups'); if i then substr(list, i, 8) = ' ';
   call symput('nooadups', put(i gt 0, 1.));
   i = index(list, 'nodups'); if i then substr(list, i, 6) = ' ';
   call symput('dupflag', put(i gt 0, 1.));
   i = index(list, 'nofinal'); if i then substr(list, i, 7) = ' ';
   call symput('pfinal', put(i eq 0, 1.));
   i = index(list, 'render'); if i then substr(list, i, 6) = ' ';
   call symput('render', put(i gt 0, 1.));
   i = index(list, 'file'); if i then substr(list, i, 4) = ' ';
   call symput('genfile', put(i gt 0, 1.));
   i = index(list, '3'); if i then substr(list, i, 1) = ' ';
   call symput('plusmin3', put(i gt 0, 1.));
   i = index(list, '+-');
   if i = 0 then i = index(list, '-+'); if i then substr(list, i, 2) = ' ';
   call symput('plusmin2', put(i gt 0, 1.));
   i = index(list, 'nosort'); if i then substr(list, i, 6) = ' ';
   call symput('dosort', put(i eq 0, 1.));
   i = index(list, 'nohistory'); if i then substr(list, i, 9) = ' ';
   call symput('history', put(i eq 0, 1.));
   i = index(list, 'accept'); if i then substr(list, i, 6) = ' ';
   call symput('accept', put(i gt 0, 1.));
   i = index(list, 'check'); if i then substr(list, i, 5) = ' ';
   justcheck = i and &initdata;
   i = index(list, 'largedesign'); if i then substr(list, i, 11) = ' ';
   call symput('largedesign', put(i gt 0, 1.));
   i = index(list, 'noresrep'); if i then substr(list, i, 8) = ' ';
   i = index(list, 'resrep'); if i then substr(list, i, 6) = ' ';
   call symput('resrep', put(i gt 0, 1.));
   i = index(list, 'intercept'); if i then substr(list, i, 9) = ' ';
   j = index(list, 'int'); if j then substr(list, j, 3) = ' ';
   call symput('int', put(i or j, 1.));
   j = index(list, 'noqc'); if j then substr(list, j, 4) = ' ';
   call symput('noqc', put(j gt 0, 1.));
   if j then call symput('pfinal', '0');
   i = index(list, 'justinit'); if i then substr(list, i, 8) = ' ';
   call symput('justinit', put(i gt 0, 1.));
   if i or j then do;
      optiter = 0;
      call symput('optiter' , '0');
      call symput('optiter2', '0');
      call symput('optflag' , '0');
      call symput('outr'    , ' ');
      end;

   if list ne ' ' then do;
      list = compbl(list);
      put 'ERROR: OPTIONS=' list 'is invalid.';
      abort = 1;
      end;

   *------maximum number of iterations------;
   opts = symget('iter');
   if opts ne symget('maxiter') and
      opts ne '21 25 10' then call symput('maxiter', trim(opts));
   opts = symget('maxiter');
   w    = scan(opts, 1, ' ');     n  = input(w , ?? 23.);
   w2   = scan(opts, 2, ' ');     n2 = input(w2, ?? 23.);
   w3   = scan(opts, 3, ' ');     n3 = input(w3, ?? 23.);
   if w2 eq ' ' then do; w2 = w;  n2 = n;  end;
   if w3 eq ' ' then do; w3 = w2; n3 = n2; end;
   if nmiss(n ) or n  lt 1 or n  ne round(n ) or
      nmiss(n2) or n2 lt 0 or n2 ne round(n2) or
      nmiss(n3) or n3 lt 0 or n3 ne round(n3) then do;
      put 'ERROR: MAXITER=' opts 'is invalid.';
      abort = 1;
      end;

   *------options=check gives MAXITER=0------;
   if justcheck then w3 = '0';

   call symput('maxiter' , compress(w ));
   call symput('maxiter2', compress(w2));
   call symput('maxiter3', compress(w3));

   *------random mutations?------;
   opts = symget('mutate');
   w    = scan(opts, 1, ' ');     n  = input(w , ?? 23.);
   w2   = scan(opts, 2, ' ');     n2 = input(w2, ?? 23.);
   w3   = scan(opts, 3, ' ');     n3 = input(w3, ?? 23.);
   if w2 eq ' ' then do; w2 = w;  n2 = n;  end;
   if w3 eq ' ' then do; w3 = w2; n3 = n2; end;
   if w  eq ' ' or n  eq 0 then do; w  = '0'; n  = 0; end;
   if w2 eq ' ' or n2 eq 0 then do; w2 = '0'; n2 = 0; end;
   if w3 eq ' ' or n3 eq 0 then do; w3 = '0'; n3 = 0; end;
   if nmiss(n ) or n  lt 0 or n  gt 1 or
      nmiss(n2) or n2 lt 0 or n2 gt 1 or
      nmiss(n3) or n3 lt 0 or n3 gt 1 then do;
      put "ERROR: MUTATE=%nrbquote(&mutate) is invalid.";
      abort = 1;
      end;
   call symput('mutate' , compress(w ));
   call symput('mutate2', compress(w2));
   call symput('mutate3', compress(w3));
   call symput('mutateflag', put(n gt 0, 1.));

   *------simulated annealing?------;
   opts = symget('anneal');
   w    = scan(opts, 1, ' ');     n  = input(w , ?? 23.);
   w2   = scan(opts, 2, ' ');     n2 = input(w2, ?? 23.);
   w3   = scan(opts, 3, ' ');     n3 = input(w3, ?? 23.);
   if w2 eq ' ' then do; w2 = w;  n2 = n;  end;
   if w3 eq ' ' then do; w3 = w2; n3 = n2; end;
   if w  eq ' ' or n  eq 0 then do; w  = '0'; n  = 0; end;
   if w2 eq ' ' or n2 eq 0 then do; w2 = '0'; n2 = 0; end;
   if w3 eq ' ' or n3 eq 0 then do; w3 = '0'; n3 = 0; end;
   if nmiss(n ) or n  lt 0 or n  ge 1 or
      nmiss(n2) or n2 lt 0 or n2 ge 1 or
      nmiss(n3) or n3 lt 0 or n3 ge 1 then do;
      put "ERROR: ANNEAL=%nrbquote(&anneal) is invalid.";
      abort = 1;
      end;
   call symput('anneal' , compress(w ));
   call symput('anneal2', compress(w2));
   call symput('anneal3', compress(w3));
   call symput('annealflag', put(n gt 0, 1.));

   *------number of iterations to work on an OPTEX design------;
   opts = symget('canditer');
   if opts eq ' ' then opts = '1 1';
   w  = scan(opts, 1, ' ');      canditer  = input(w , ?? 23.);
   w2 = scan(opts, 2, ' ');      canditer2 = input(w2, ?? 23.);
   if w2 eq ' ' then do; w2 = w; canditer2 = canditer;  end;
   if nmiss(canditer ) or canditer  lt 0 or canditer  ne round(canditer ) or
      nmiss(canditer2) or canditer2 lt 0 or canditer2 ne round(canditer2)
      then do;
      put "ERROR: CANDITER=%nrbquote(&canditer) is invalid.";
      abort = 1;
      end;
   call symput('canditer' , compress(w ));
   call symput('canditer2', compress(w2));

   *------read an external initial design?------;
   call symput('readinit',  put(&initdata, 1.));
   call symput('initrand',  put(not &initdata, 1.));
   call symput('firstpass', put(1 + 2 * &initdata, 1.));

   *------check FIXED=------;
   if symget('fixed') ne ' ' then do;
      if not &initdata then do;
         put "ERROR: FIXED= may only be specified with INIT=.";
         abort = 1;
         end;
      call symput('rowsfixed', '1');
      end;

   *------check HOLDOUTS=------;
   if symget('holdouts') ne ' ' and not &initdata then do;
      put "ERROR: HOLDOUTS== may only be specified with INIT=.";
      abort = 1;
      end;

   *------must be some output data set------;
   if justcheck then do;
      call symput('out', '');
      call symput('outr', '');
      call symput('outall', '');
      call symput('outeff', '');
      end;

   else if symget('out') eq ' ' and symget('outr') eq ' ' and
      symget('outall') eq ' ' and symget('outeff') eq ' ' then do;
      put "ERROR: You must specify OUT=, OUTR=, OUTALL=, or OUTEFF=.";
      abort = 1;
      end;

   *------need temporary OUT= if none specified------;
   if symget('out') eq ' ' then do;
      call symput('out', '__tmpxout');
      call symput('dosort', '0');
      end;

   *------separate out data set options------;
   str = symget('out');
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('out', substr(str, 1, j - 1));
   if str ne ' ' then call symput('outopts', substr(str, j));

   str = symget('outr');
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('outr', substr(str, 1, j - 1));
   if str ne ' ' then call symput('outropts', substr(str, j));

   str = symget('outall');
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('outall', substr(str, 1, j - 1));
   if str ne ' ' then call symput('outallopts', substr(str, j));

   str = symget('outeff');
   j = index(str, '('); if j = 0 then j = length(str) + 1;
   call symput('outeff', substr(str, 1, j - 1));
   if str ne ' ' then call symput('outeffopts', substr(str, j));

   *------number of levels to simultaneously exchange------;
   w = symget('exchange');
   if w eq ' ' then n = 1 + (symget('partial') ne ' ');
   else             n = input(symget('exchange'), ?? 23.);
   if nmiss(n) or n lt 1 or n ne round(n) then do;
      put "ERROR: EXCHANGE=%nrbquote(&exchange) is invalid.";
      abort = 1;
      end;
   call symput('exchange', trim(put(n, 8. -L)));

   *------target efficiency criterion------;
   n = input(symget('target'), ?? 23.);
   if nmiss(n) or n le 0 then do;
      put "ERROR: TARGET=%nrbquote(&target) is invalid.";
      abort = 1;
      end;

   *------check SEED=------;
   seed = input(symget('seed'), ?? 23.);
   if nmiss(seed) or seed ne round(seed) then do;
      put "ERROR: SEED=%nrbquote(&seed) is invalid.";
      abort = 1;
      end;

   *------special case random number seeds------;
   else if symget('seed') eq '00' then call symput('seed', '0');

   *------run generically, with constant seed,
          eliminates using a clock generated seed------;
   else if seed eq 0 and getoption('generic') eq 'GENERIC' and
      getoption('stimer' ) eq 'NOSTIMER' then call symput('seed', '072555');

   *------set seed using clock time when negative------;
   else if seed le 0 then do;
      seed = ceil(1e7 * uniform(input(reverse(put(1e2 * datetime(), 13.0)),
                                      ?? 8.)));
      call symput('seed', compress(put(seed, 8.)));
      if &printseed then put 'Seed = ' seed;
      end;

   *------seed for randomization------;
   call symput('ranseed', symget('seed'));

   *------get parsed interaction list made by MktRuns------;
   newlist = symget('_mktintlist');
   call symput('nointeract', put(newlist eq ' ', 1.));

   *------do some more error checks, build MktDes syntax------;
   if newlist ne ' ' then do;
      varlist = trim(varlist) || ', interact=';

      *------add spaces around commas------;
      maxnum = 0; num = 0;
      l = min(3 * length(newlist), &maxlen);
      do i = 1 to l;
         if substr(newlist, i, 1) eq ',' then do;
            substr(newlist, i) = ' , ' || substr(newlist, i + 1);
            i + 2;
            end;
         end;
      newlist = compbl(newlist) || ' ,';

      *------check variable numbers build MktDes interaction list------;
      do i = 1 to &maxlen until(w eq ' ');
         w = scan(newlist, i, ' ');
         if w ne ' ' then do;
            if w eq ',' then num = 0;
            else do;
               n = input(w, ?? 23.);
               if nmiss(n) or n lt 1 or n gt nvars or n ne round(n) then do;
                  hold = symget('interact');
                  put 'ERROR: INTERACT=' hold 'is invalid.';
                  abort = 1;
                  goto endit;
                  end;
               num + 1;
               maxnum = max(maxnum, num);
               varlist = trim(varlist) || ' x' || w;
               if scan(newlist, i + 1, ' ') ne ','
                  then varlist = trim(varlist) || ' *';
               end;
            end;
         end;

      *------pad IML matrix with trailing 0s------;
      hold = ' '; num = 0;
      do i = 1 to &maxlen until(w eq ' ');
         w = scan(newlist, i, ' ');
         if w ne ' ' then do;
            if w eq ',' then do;
               do j = 1 to maxnum - num;
                  hold = trim(hold) || ' 0';
                  end;
               num = 0;
               end;
            else num + 1;
            hold = trim(hold) || ' ' || w;
            end;
         end;
      lh = length(hold);
      if lh gt (&maxlen - 10) then do;
         put 'ERROR: Too many interactions.';
         abort = 1;
         goto endit;
         end;

      substr(hold, lh) = ' ';
      call symput('interactmat', trim(hold));
      end;

   * simplify request for all two-way interactions to resolution V design;
   call symput('optvarlist', trim(varlist));
   if &r5 then substr(varlist, index(varlist, 'interact') + 9) = '@2';
   call symput('varlist', trim(varlist));

   *------skip stuff with silly ltfr specifications------;
   if nruns lt npars then fullrank = 0;
   optflag = (fullrank and optiter and not &initdata);
   call symput('optflag', put(optflag, 1.));
   call symput('ltfr', put(not fullrank, 1.));
   call symput('npars', compress(put(npars, 12.)));

   *-------check STOPEARLY=------;
   w = symget('stopearly');
   if w = ' ' then do;
      n = 0;
      call symput('stopearly', '0');
      end;
   else n = input(symget('stopearly'), ?? 23.);
   if nmiss(n) or n ne round(n) or n lt 0 then do;
      put "ERROR: STOPEARLY=%nrbquote(&stopearly) is invalid.";
      abort = 1;
      end;

   *------check ORDER=------;
   call symput('partinc', '0');
   w = lowcase(symget('order'));
   if w =: 'matrix' then do;
      hold = scan(w, 2, ' =');
      if hold = ' ' then do;
         put "ERROR: ORDER=%nrbquote(&order) is invalid.";
         put "ERROR: You must specify a data set name.";
         abort = 1;
         end;
      w = 'matrix';
      call symput('exchangedat', trim(hold));
      end;
   if w =: 'col' then do;
      n = input(scan(w, 2, ' ='), 23.);
      if nmiss(n) then n = 1;
      w = 'col';
      if n ne floor(n) then do;
         put "ERROR: ORDER=%nrbquote(&order) is invalid.";
         abort = 1;
         end;
      call symput('ordcoln', trim(put(n, 12. -L)));
      end;
   if w eq 'seqran' then w = 'ranseq';
   if w =: 'random' then do;
      if index(w, '=') then do;
         n = input(scan(w, 2, ' ='), 23.);
         if nmiss(n) or n ne floor(n) then do;
            put "ERROR: ORDER=%nrbquote(&order) is invalid.";
            abort = 1;
            end;
         else if symget('exchange') ne '2' then do;
            put 'ERROR: ORDER=RANDOM= is specified with EXCHANGE=2.';
            abort = 1;
            end;
         call symput('partinc', trim(put(n, 12. -L)));
         end;
      w = 'random';
      end;
   else if w ne 'sequential' and w ne 'random' and w ne 'ranseq' and
      w ne 'matrix' and w ne 'col' and w ne ' ' then do;
      put "ERROR: ORDER=%nrbquote(&order) is invalid.";
      abort = 1;
      end;
   else if w eq ' ' then do;
      if symget('partial') ne ' ' then w = 'random';
      else if symget('dupflag') eq '1' or symget('restrictflag') eq '1'
         then w = 'ranseq';
         else w = 'sequential';
      end;
   call symput('order', trim(w));

   if w eq 'col' then do;
      if symget('restrictions') ne ' ' or symget('partial') ne ' ' or
         symget('balance')      ne ' ' or symget('dupflag') eq '1' then do;
         put 'WARNING: Specifying ORDER=COL=n with RESTRICTIONS=, PARTIAL=,' /
             '         BALANCE= or DUPFLAG= is risky.' /
             '         The restrictions may not be met.';
         end;
      end;

   *------check RIDGE=------;
   w = symget('ridge');
   ridge = input(w, ?? 23.);
   if nmiss(ridge) or ridge lt 0 then do;
      put "ERROR: RIDGE=%nrbquote(&ridge) is invalid.";
      abort = 1;
      end;
   else if ridge lt 1e-10 or ridge gt 0.2 then
      put 'WARNING: Extreme RIDGE= values can cause problems.';

   *------silly rank-deficient design warning------;
   if not abort and not fullrank and w eq '1e-7' then do;
      put 'ERROR: More parameters than runs.' /
          '       If you really want to do this, specify RIDGE=.' /
          '       There are ' nruns 'runs with ' npars 'parameters.';
      abort = 1;
      end;

   call symput('ridgemsg', put(not fullrank and not abort, 1.));

   *------generated file name------;
   if symget('genfile') eq '1' then do;
      list = '"OA(' || compress(put(&n, 5.));
      hold = symget('list');
      do i = 1 to 20 by 2 until(w = ' ');
         w = scan(hold, i, ' *');
         if w ne ' ' then do;
            w2 = scan(hold, i + 1, ' *');
            list = trim(list) || ',' || trim(w) || '^' || trim(w2);
            end;
         end;
      list = trim(list) || ')"';
      call symput('genfile', trim(list));
      end;

   else call symput('genfile', 'print');

   *------parse LEVELS=------;
   list = left(lowcase(symget('levels')));

   i = index(list, 'first'); if i then substr(list, i, 5) = ' ';
   call symput('first', put(i gt 0, 1.));
   i = index(list, 'last'); if i then substr(list, i, 4) = ' ';
   call symput('last', put(i gt 0, 1.));
   i = index(list, 'intercept'); if i then substr(list, i, 9) = ' ';
   j = index(list, 'int'); if j then substr(list, j, 3) = ' ';
   if i or j then call symput('int', '1');

   call symput('levels2', ' ');
   call symput('levels3', ' ');
   if list in ('1', '0', 'c', 'i') then call symput('levels', trim(list));
   else if list eq ' ' then call symput('levels', '1');
   else do;
      if scan(list, 6, ' ') ne ' ' then do;
         put 'ERROR: LEVELS=' list 'is too long.' /
             '       Expecting no more than five values.';
         abort = 1;
         end;
      else do;
         ok = 1;
         do n = 1 to 5;
            t[n] = scan(list, n, ' ');
            if length(t[n]) gt 1 then ok = 0;
            if not (trim(t[n]) in ('1', '0', 'c', 'i', '2', '3', ' '))
               then ok = 0;
            if t[n] = ' ' then leave;
            end;
         n + -1;
         if trim(t[n]) in ('2', '3') then ok = 0;
         if ok then do;
            two = 0; three = 0;
            do i = 1 to n;
               if t[i] eq '2' then do;
                  two + 1;
                  if not (trim(t[i+1]) in ('1', '0', 'c', 'i')) then ok = 0;
                  call symput('levels2', trim(t[i+1]));
                  if ok then do; t[i] = ' '; t[i + 1] = ' '; end;
                  end;
               if t[i] eq '3' then do;
                  three + 1;
                  if not (trim(t[i+1]) in ('1', '0', 'c', 'i')) then ok = 0;
                  call symput('levels3', trim(t[i+1]));
                  if ok then do; t[i] = ' '; t[i + 1] = ' '; end;
                  end;
                end;
            end;
         if two gt 1 or three gt 1 then ok = 0;
         if ok then do;
            hold = left(t1 || t2 || t3 || t4 || t5);
            if hold eq ' ' then hold = '1';
            if not (trim(hold) in ('1', '0', 'c', 'i')) then ok = 0;
            else call symput('levels', trim(hold));
            end;
         if not ok then do;
            put 'ERROR: LEVELS=' list 'is invalid.';
            abort = 1;
            end;
         end;
      end;
   if symget('levels2') eq ' ' then call symput('levels2', symget('levels'));
   if symget('levels3') eq ' ' then call symput('levels3', symget('levels'));

   i = (symget('elefixed')  eq '1');
   j = (symget('rowsfixed') eq '1');
   k = (symget('resetinit') eq '1');
   if i and j then do;
      put 'ERROR: FIXED= and HOLDOUTS= may not be '
          'specified when ' 'the initial ' 'design has ' 'elements of '
          '0.  Put ' 'zeros in ' 'the initial ' 'design for ' 'all '
          'fixed ' 'element ' 'flagging.';
      abort = 1;
      end;

   call symput('resetinit', put(k and (i or j), 1.));
   if j then call symput('elefixed', '1');

   endit: if _error_ or abort then call symput('abort', '1');
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------mention data sets, if they are invalid, quit------;

filename __dummy dummy;
proc printto log=__dummy; run;
%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

%let savedsnf = %sysfunc(getoption(dsnferr)) %sysfunc(getoption(fmterr));
options nodsnferr nofmterr;
%if %nrbquote(&out) ne %then %do;
   data _null_; set &out; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outr) ne %then %do;
   data _null_; set &outr; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outall) ne %then %do;
   data _null_; set &outall; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if %nrbquote(&outeff) ne %then %do;
   data _null_; set &outeff; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
%if &order eq matrix %then %do;
   data _null_; set &exchangedat; stop; run;
   %if &syserr > 4 %then %let abort = 1;
   %end;
options &savedsnf;

proc printto log=log; run;
%if &syserr > 4 %then %let abort = 2;

%if &abort eq 1 %then %do;
   %put ERROR: One or more data set names are invalid:;
   %put %str(       )outall=&outall;
   %put %str(       )outall=&outeff;
   %put %str(       )out=&out;
   %put %str(       )outr=&outr;
   %end;

%if &abort %then %goto endit;

*------this copy allows drop= option in INIT= data set------;
%if &initdata %then %do;
   data __tmpxinit; set &init; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %let init = __tmpxinit;
   %end;

*------check the exchange data set------;
%if &order eq matrix %then %do;
   data _null_;
      set &exchangedat;
      array __n[*] _numeric_;
      if _n_ = 1 then do;
         call symput('exchange', compress(put(dim(__n), 5.)));
         if dim(__n) lt 1 then do;
            put 'ERROR: ORDER=MATRIX exchange data set must contain numbers.';
            call symput('abort', '1');
            end;
         end;
      do __i = 1 to dim(__n);
         __c = __n[__i];
         if n(__c) and (__c lt 1 or __c ne round(__c) or __c gt &m) then do;
            put 'ERROR: ORDER=MATRIX exchange data set value ' __c
                'is invalid.';
            call symput('abort', '1');
            end;
         if nmiss(__c) then call symput('missex', '1');
         end;
      if nmiss(__n[1]) then do;
         put 'ERROR: The first ORDER=MATRIX exchange data set variable '
             'may not have a missing value.';
         call symput('abort', '1');
         end;
      if _error_ then call symput('abort', '1');
      run;
   %end;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*============================ What Design is This? ============================;
*mkttime(MktOrth);

*------what designs can we make?------;
data __tmpxcat; run;

%if &tabiter and %nrbquote(&cat) eq %then %do;
   %let _mktver = obsolete;

   %let i = %sysfunc(min(&maxlevneed, 144));
   %mktorth(outlev=__tmpxcat, options=mktex lineage %if &do512 %then 512;,
            maxlev=&i,
            range=%if &tabsize %then n=&tabsize, maxn=&tabsize;
                 %else %eval(&n - 6) le n le %eval(&n + 6), maxn=%eval(&n + 6);)

   %if &_mktver ne &localver %then %do;
      %put ERROR: Incompatible MktOrth macro.;
      %put %str(      ) You must have the same version of all macros.;
      %put %str(      ) Versions: MktEx = &localver, MktOrth = &_mktver..;
      %put %str(      ) Get the latest tools and documentation from:;
      %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
      %let abort = 1;
      %end;
   %end;

%else %if &tabiter %then %do;
   data __tmpxcat;
      set &cat end=eof;
      if _n_ = 1 then do; min = n; max = n; end;
      min = min(min, n);
      max = max(max, n);
      if eof then do;
         call symput('catmin', compress(put(min, 23.)));
         call symput('catmax', compress(put(max, 23.)));
         end;
      drop min max;
      run;
   %end;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------with interactions, throw away designs that would bypass FACTEX------;
%if not &nointeract %then %do;
   data __tmpxcat; set __tmpxcat; if not (reference =: 'Frac'); run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*mkttime(What Design is This?);

*------see if we can use an OA or FACTEX------;
data _null_;
   length newlist mktlist $ &maxlen w $ 23 reference $ 20
          tableflag $ %eval(3*&maxlev+10) nlevs $ 3;
   array nfacts[&maxlev];
   array nrows[&maxlev];
   array x[&maxlev];
   array any[&maxlev] any01-any&maxlev;
   retain any01-any&maxlev nfacts1-nfacts&maxlev nrows1-nrows&maxlev 0;
   abort = 0; nvars = &m; nruns = &n; powerof = 0; largehad = 0;
   tabiter = &tabiter; tabiter2 = &tabiter2;
   needed  = nvars * &n; mindif  = needed + 1;

   *------allow user to override size of table to make------;
   %if &tabsize %then %do; nruns = &tabsize; %end;

   *------start processing factors list------;
   mktlist = symget('_mktlist');

   *------count levels of factors------;
   do i = 1 to &maxlen until(w eq ' ');
      w = scan(mktlist, i, ' ');
      if w ne ' ' then do;
         n = input(w, ?? 23.);
         if 2 le n le &maxlev then any[n] + 1;
         end;
      end;

   *------special OA or FACTEX initializations?------;
   tableflag = ' '; newlist = ' ';
   tabflag = 0; factexnlevs = 0; hadsize = 0; full = 0; plan = 0; nreps = 1;
   call symput('factexnlevs', '0'); call symput('allfactex', '0');
   call symput('hadsize',     '0'); call symput('tableflag', '0');
   call symput('full',        '0');

   *------(potentially replicated) full-factorial design------;
   if &tabiter and sum(of any:) eq nvars then do;
      if &_mktlogff lt 16 then do;
         nreps = nruns / round(10 ** &_mktlogff);
         if nreps gt 0.99 and abs(nreps - round(nreps)) le 1e-8 then full = 1;
         end;
      end;

   call symput('nreps', compress(put(nreps, 23.0)));

   *------number of levels to base candidate set on------;
   %if %nrbquote(&cat) eq %then %do;
      do i = &primes; nfacts[i] = any[i]; end;
      nfacts1 = any02;
      nfacts2 + any04 + any08 + any16 + any32 + any64;
      nfacts3 + any09 + any27 + any81;
      nfacts5 + any25;
      nfacts7 + any49;
      max     = max(of nfacts:);
      p       = 0;
      if max then do;
         do i = &primes; if max eq nfacts[i] then do; p = 1; leave; end; end;
         end;
      if not p then i = 2;
      call symput('candnlev', compress(put(i, 3.)));
      if &flow then put '** candlev=' i;
      %end;

   if &tabiter and not &ltfr and not &initdata then do;

      *-----sizes for Hadamard and FACTEX designs-------;
      do i = 1 to &maxlev; nrows[i] = i; end;
      do c = 0 to 100 by 4 until(nrows1); /* nearest Hadamard */
         nrows1 = 4 * ceil((nruns + c) / 4);    hadsize = nrows1;
         do while(mod(hadsize, 8) eq 0); hadsize = hadsize / 2; end;
         holdhad = hadsize;
         link paley;
         if not (%mktexishad(hadsize)) then do;
            do while(hadsize lt nrows1 and not (ispm1 or ispm2));
               hadsize = hadsize * 2;
               link paley;
               end;
            if not (ispm1 or ispm2) then hadsize = holdhad;
            end;
         if not (%mktexishad(hadsize) or ispm1 or ispm2) then nrows1 = 0;
         end;
      hadsize = 0;

      *------large hadamard?------;
      largehad = (nfacts2 eq nvars and nrows1 gt &maxlev);

      do i = &primes;
         nrows[i] = i ** round(log(nruns) / log(i));
         if nrows[i] gt 1.25 * &n
            then nrows[i] = i ** floor(log(nruns) / log(i));

         maxfacts  = (nrows[i] - 1) / (i - 1);
         nfacts[i] = min(nfacts[i], maxfacts);
         nfacts[i] + max(min(nvars - nfacts[i], maxfacts - nfacts[i]), 0) *
                     0.6; /* fuzz for unbals */
         end;

      *------count matching factors------;
      if __nobs1 then do __n = 1 to 1e8 until(__n ge __nobs1);
         set __tmpxcat nobs=__nobs1;
         link countmatch;
         nfacts[n] = max(nfacts[n], matches);
         end;

      *------see how close we are to the right size------;
      nruns = 1;
      do i = %if &tabsize %then &tabsize;
             %else %if &catmin %then &catmin to &catmax;
             %else 1 to &maxlev;;
         rows = max(0, &n - 2 * abs(nrows[i] - &n));
         sizedif = abs(nfacts[i] * rows - needed);
         if sizedif eq 0 or sizedif lt mindif then do;
            mindif = sizedif;
            nruns  = i;
            %if &flow %then %do; put '** ' i= mindif=; %end;
            end;
         if mindif eq 0 and i gt 1 then leave;
         end;

      *------use PLAN with FF design------;
      if full then do; plan = 1; nruns = n; end;

      *------what kind of design are we making?------;
      incat   = not (nruns in (1, &primes));
      frac    = not incat and nruns ne 1;
      nruns2  = 0;
      if not incat then nruns2 = nrows[nruns];

      if __nobs2 and not plan then do __n = 1 to 1e8 until(__n ge __nobs2);
         set __tmpxcat nobs=__nobs2;

         if n eq nruns or (not incat and n eq nruns2) then do;
            link countmatch;
            if abs(nfacts[nruns] - matches) le 0.0005 then do;
               if not (reference =: 'Had') then reference = lowcase(reference);
               if reference eq ' ' and frac
                  then reference = 'fractional-factorial';
               call symput('destype', trim(reference));
               call symput('lineage', trim(lineage));
               if incat then leave;
               end;
            end;

         end;

      *------pick a size, set nruns to OA design size------;
      if not incat then nruns = nruns2;
      lineage = symget('lineage');

      %if &flow or &flowinit %then %do;
         put '** ' nruns= mindif= nvars= lineage=;
         %end;
      %if &flow %then %do;
         if nruns le &maxlev then put '** ' nfacts[nruns]=;
         %end;
      if &plineage then put 'Design Lineage:' / lineage;

      *------figure out powerof, based on OA design size------;
      p = 0;
      if lineage eq ' ' or (&r5 and not full) then do powerof = &primes;
         lognruns = log(nruns) / log(powerof);
         if abs(lognruns - round(lognruns)) lt 1e-8 then do; p = 1; leave; end;
         end;
      if not p then powerof = 0;

      *------get into FACTEX with Resolution V------;
      r5 = 0;
      if &r5 then do;
         if nruns eq &n and powerof and max(of any:) eq sum(of any:) then do;
            lineage = ' ';
            call symput('lineage', ' ');
            hadsize = 0;
            r5 = 1;
            end;
         else do; /* do not ask for resolution V */
            newlist = symget('varlist');
            substr(newlist, index(newlist, '@2'), 2) = ' ';
            newlist = trim(newlist) || 'x1';
            do i = 2 to &m;
               newlist = trim(newlist) || compress('|x' || put(i, 5.));
               end;
            newlist = trim(newlist) || '@2';
            call symput('varlist', trim(newlist));
            if lineage ne ' ' then powerof = 0;
            end;
         end;
      call symput('r5', put(r5, 1.));
      if &flow then put '** ' r5=;

      *------full-factorial design with PLAN------;
      if plan then do;
         call symput('full', '1');
         call symput('tryfac', '0');
         nruns = &n; mindif = 0; lineage = ' '; newlist = ' ';

         *------build MktDes syntax for PLAN------;
         do i = 2 to &maxlevneed;
            if any[i] then do;
               if i lt 100 then nlevs = put(i, z2.);
               else             nlevs = put(i, 3.);
               newlist = trim(newlist) || ' ' ||
                         compress('x'  || nlevs || '_1-x' || nlevs || '_'  ||
                                  put(any[i], 5.) || '=' || nlevs);
               tableflag = trim(tableflag) || '_' || nlevs;
               end;
            end;
         call symput('factorslist', trim(left(newlist)));
         end;

      %*-------orthogonal arrays------;
      else if lineage ne ' ' then tableflag = 'table';

      %*-------FACTEX designs------;
      else if powerof and hadsize eq 0 then do;
         factexnlevs = powerof;
         tableflag   = 'factex';
         call symput('destype', 'fractional-factorial');
         end;

      if full and nreps lt 1.5 then call symput('destype', 'full-factorial');
      else if full then call symput('destype', 'replicated full-factorial');
      else if mod(nruns, 4) eq 0 and any02 eq nvars and not r5 then
         call symput('destype', 'Hadamard');

      *------Hadamard matrices------;
      if (any02 or any04 or any08 or any16 or any32 or any64) and
         mod(nruns, 4) eq 0 and not full and
         ((any02 eq nvars and &nointeract) or powerof ne 2) then do;

         tableflag = 'table';    hadsize = nruns;

         do while(mod(hadsize, 8) eq 0); hadsize = hadsize / 2; end;
         holdhad = hadsize;
         link paley;
         if not %mktexishad(hadsize) then do;
            do while(hadsize lt nruns and not (ispm1 or ispm2));
               hadsize = hadsize * 2;
               link paley;
               end;
            if not (ispm1 or ispm2) then hadsize = holdhad;
            end;

         if ispm1 or ispm2 then
            call symput('hadflag', 'hadamard' || put(ispm1 + 2 * ispm2, 1.));

         if %mktexishad(hadsize) or ispm1 or ispm2 then do;
            if symget('destype') eq 'random'
               then call symput('destype', 'Hadamard');
            call symput('hadsize', compress(put(hadsize, 12.)));
            end;
         else hadsize = 0;
         end;

      *------now we know our size, report------;
      if &flow then do;
         if nruns eq &n
            then put '** full row initialization ' nruns= "n=&n " powerof=;
            else put '** partial row initialization ' nruns= "n=&n " powerof=;
         end;
      end;

   if &flow then put '** ' nruns= factexnlevs= tableflag= powerof=;

   if tableflag eq ' ' then tableflag = '0';
   tabflag = (tableflag ne '0' and &tabiter);
   if not tabflag then tableflag = '0';
   call symput('powerof', compress(put(powerof, 3.)));
   skipit = 0;

   if tabiter and not &ltfr and not &initdata then do;

      call symput('tableflag'  , compress(tableflag));
      call symput('factexnlevs', compress(put(factexnlevs, 2.)));

      *------skip OPTEX if we are sure an OA will work,
             or occasionally, if we are almost sure (largehad)
             and also expect OPTEX to not be a good strategy  ------;
      if nruns eq &n and &restrictflag eq 0 and (mindif eq 0 or largehad) and
         index(lowcase(symget('options')), 'nodups') eq 0 then do;

         if &nointeract or full then do;
            call symput('optflag' , '0');
            call symput('optiter' , '0');
            call symput('optiter2', '0');
            skipit = 1;
            end;

         %*------otherwise does this look like an all-FACTEX problem?------;
         else if factexnlevs then do;
            do i = &primes; nfacts[i] = any[i]; end;
            nfacts1 = any02;
            nfacts2 + any04 + any08  + any16 + any32 + any64;
            nfacts3 + any09 + any27 + any81;
            nfacts5 + any25;
            nfacts7 + any49;
            p = 0;
            do i = &primes;
               if nfacts[i] eq nvars and powerof eq i
                  then do; p = 1; leave; end;
               end;
            if p then do; skipit = 1; call symput('allfactex', '1'); end;
            end;

         end;
      end;

   *------annealing probabilities look too big?------;
   if not skipit then do;
      p  = 1 / &maxlevfound;
      n  = input(symget('anneal') , ?? 23.);
      n2 = input(symget('anneal2'), ?? 23.);
      n3 = input(symget('anneal3'), ?? 23.);
      if p lt n or p lt n2 or p lt n3 then
         put 'WARNING: ANNEAL=' n n2 n3 'may be too large with '
             "&maxlevfound-level factors.";
      end;

   *------silly design warning------;
   if &ltfr then do;
      call symput('tableflag', '0');
      put 'WARNING: Not enough design points to estimate model parameters.';
      end;

   if &ridgemsg then
      put "WARNING: Iterations will continue with RIDGE=&ridge specified.";

   *------did the user-specified OA size work?------;
   if &tabsize and not tabflag then do;
      put "ERROR: TABSIZE=%nrbquote(&tabsize) is not allowed for this design.";
      abort = 1;
      end;

   call symput('tabsize', compress(put(nruns, 12.0)));

   *------number of iterations to use unbalanced initializations------;
   opts = symget('unbalanced');
   if opts eq ' ' then opts = '0.2 0.1';
   w  = scan(opts, 1, ' ');   w2 = scan(opts, 2, ' ');
   if w2 eq ' ' then w2 = w;
   unbalanced  = input(w , ?? 23.);
   unbalanced2 = input(w2, ?? 23.);
   if nmiss(unbalanced ) or unbalanced  lt 0 or unbalanced  gt 1 or
      nmiss(unbalanced2) or unbalanced2 lt 0 or unbalanced2 gt 1 then do;
      put "ERROR: UNBALANCED=%nrbquote(&unbalanced) is invalid.";
      abort = 1;
      end;
   else do;
      unbalanced  = ceil(unbalanced  * tabiter );
      unbalanced2 = ceil(unbalanced2 * tabiter2);
      end;
   call symput('unbalanced' , compress(put(unbalanced , 12.)));
   call symput('unbalanced2', compress(put(unbalanced2, 12.)));

   *------first iteration to mutate------;
   opts = symget('mutiter');
   if opts eq ' ' then opts = '. . .';
   w  = scan(opts, 1, ' ');    n  = input(w , ?? 23.);
   w2 = scan(opts, 2, ' ');    n2 = input(w2, ?? 23.);
   w3 = scan(opts, 3, ' ');    n3 = input(w3, ?? 23.);
   if w2 eq ' ' then do; w2 = w;  n2 = n;  end;
   if w3 eq ' ' then do; w3 = w2; n3 = n2; end;
   if (w  ne ' ' and w  ne '.' and nmiss(n ))  or
      (n(n ) and (n  lt 0 or n  ne round(n ))) or
      (w2 ne ' ' and w2 ne '.' and nmiss(n2))  or
      (n(n2) and (n2 lt 0 or n2 ne round(n2))) or
      (w3 ne ' ' and w3 ne '.' and nmiss(n3))  or
      (n(n3) and (n3 lt 0 or n3 ne round(n3)))
      then do;
      put "ERROR: MUTITER=%nrbquote(&mutiter) is invalid.";
      abort = 1;
      end;
   if nmiss(n) then do;
      if &initdata then n = 1;
      else              n = 1 + &optflag * &canditer + tabflag * tabiter;
      end;
   if nmiss(n3) then n3 = 1;
   call symput('mutiter' , compress(put(n , 12.0)));
   call symput('mutiter2', compress(put(n2, 12.0)));
   call symput('mutiter3', compress(put(n3, 12.0)));

   *------first iteration to anneal------;
   opts = symget('anniter');
   if opts eq ' ' then opts = '. . .';
   w  = scan(opts, 1, ' ');    n  = input(w , ?? 23.);
   w2 = scan(opts, 2, ' ');    n2 = input(w2, ?? 23.);
   w3 = scan(opts, 3, ' ');    n3 = input(w3, ?? 23.);
   if w2 eq ' ' then do; w2 = w;  n2 = n;  end;
   if w3 eq ' ' then do; w3 = w2; n3 = n2; end;
   if (w  ne ' ' and w  ne '.' and nmiss(n ))  or
      (n(n ) and (n  lt 0 or n  ne round(n ))) or
      (w2 ne ' ' and w2 ne '.' and nmiss(n2))  or
      (n(n2) and (n2 lt 0 or n2 ne round(n2))) or
      (w3 ne ' ' and w3 ne '.' and nmiss(n3))  or
      (n(n3) and (n3 lt 0 or n3 ne round(n3))) then do;
      put "ERROR: ANNITER=%nrbquote(&anniter) is invalid.";
      abort = 1;
      end;
   if nmiss(n) then do;
      if &initdata then n = 1;
      else              n = 1 + &optflag * &canditer + tabflag * tabiter;
      end;
   if nmiss(n3) then n3 = 1;
   call symput('anniter' , compress(put(n , 12.0)));
   call symput('anniter2', compress(put(n2, 12.0)));
   call symput('anniter3', compress(put(n3, 12.0)));

   if _error_ or abort then call symput('abort', '1');
   stop;

   paley: *------will one of the Paley Hadamard constructions work?------;
      ispm1 = 1; ispm2 = mod(hadsize / 2, 4) eq 2;
      do i = 2 to sqrt(hadsize) while(ispm1);
         if mod(hadsize - 1, i) eq 0 then ispm1 = 0;
         end;
      do i = 2 to sqrt(hadsize / 2) while(ispm2);
         if mod(hadsize / 2 - 1, i) eq 0 then ispm2 = 0;
         end;
      ispm2 = ispm2 and not ispm1;
      return;

   countmatch: *------count matching factors, fuzz for unbalanced------;
      matches = 0;
      do i = 1 to &maxlevneed; matches + min(any[i], x[i]); end;
      if x6 gt any06 and (any05 gt x5 or any04 gt x4)
         then matches + (x6 - any06) * 0.001;
      if x5 gt any05 and (any04 gt x4 or any03 gt x3)
         then matches + (x5 - any05) * 0.001;
      if x4 gt any04 and any03 gt x3 then matches + (x4 - any04) * 0.001;
      if x3 gt any03 and any02 gt x2 then matches + (x3 - any03) * 0.001;
      return;
   run;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*======================== Set Up Macros, Start Checking =======================;
*mkttime(Set Up Macros - Start Checking);

%if &flow %then
   %mktexdump(lineage optiter optiter2 maxlevfound maxlevneed m optflag
              powerof candnlev factexnlevs allfactex tableflag hadflag
              hadsize full);

*------partial profile macro------;
%if &partial ne %then %do;
   %macro __partprof;
      nvary = sum(x ^= 1);
      if i = 1 then __pbad = nvary;
      else          __pbad = abs(nvary - &partial);
      %mend;
   %end;

*------balance macro------;
%if &balance ne %then %do;
   data _null_;
      n = &n; m = &m; mintry = &mintry; n = mintry - 3 * n / 2;
      call symput('balancetry', compress(put(n, 12.0)));
      call symput('mintry',     compress(put(mintry, best12.)));
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %macro __balance;
      __bbad = 1;
      if try > &balancetry & j1 then do;
         acol = xmat[,j1];
         acol[i,] = x[,j1];
         acol = design(acol)[+,];
         %if &ridgemsg %then %do;
            if ncol(acol) < __facts[j1] then acol = acol || 0;
            %end;
         __bbad = max(0, max(acol) - min(acol) - &balance);
         end;
      %mend;
   %end;

*------check restrictions macro for syntax errors------;
%if &userrestricts %then %do;

   filename __dummy dummy;
   proc printto log=__dummy; run;
   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   proc iml;
      call symput('jj', '0');
      %if &ridgemsg %then %do; __facts = { &_mktlist }; %end;

      %if &resflag %then %&resmac;
      start badness(x) global(xmat, i, try %if &ridgemsg %then , __facts;
                              %do i = 1 %to &exchange; , j&i %end;
                              %if &resflag %then , &reslist;);
         %if not &nox %then %do i = 1 %to &m; x&i = x[&i]; %end;
         %if &balance ne %then %do; __bbad = 0; %end;
         %if &partial ne %then %do; __pbad = 0; %end;
         bad = 0; %&restrictions; return(bad);
         finish;
      %do i = 1 %to &exchange; j&i = 1; %end; i = 1; try = 1;
      xmat = j(&n, &m, 1);
      x = xmat[i,];
      bad = badness(x);
      if %do i = 1 %to &exchange; j&i ^= 1 | %end; i ^= 1 | try ^= 1 |
         ^all(xmat = 1) | ^all(x = 1) then call symput('jj', '1');
      quit;
   %if &syserr > 4 %then %let abort = 1;

   proc printto log=log; run;
   %if &syserr > 4 %then %let abort = 2;

   %if &jj %then %do;
       %let i = i, try, j1, j2, j3, x, or xmat.;
       %put WARNING: Restrictions macro is changing &i.;
       %put %str(        ) This might be a serious problem.  Check your macro.;
       %end;

   %if &abort eq 1 %then %do;
      %let j = %sysfunc(getoption(mprint));
      %put ERROR: Invalid RESTRICTIONS=&restrictions macro.;
      %put %str(      ) Here is your macro embedded in a simple test program.;
      options mprint;
      proc iml;
         %if &resflag %then %&resmac;
         start badness(x) global(xmat, i, try
                                 %do i = 1 %to &exchange; , j&i %end;
                                 %if &resflag %then , &reslist;);
            %if not &nox %then %do i = 1 %to &m; x&i = x[&i]; %end;
            %if &balance ne %then %do; __bbad = 0; %end;
            %if &partial ne %then %do; __pbad = 0; %end;
            bad = 0; %&restrictions; return(bad);
            finish;
         %do i = 1 %to &exchange; j&i = 1; %end; i = 1; try = 1;
         xmat = j(&n, &m, 1);
         x = xmat[i,];
         bad = badness(x);
         quit;
      options &j;
      %end;

   %if &abort %then %goto endit;
   %end;

*------check INIT= data set------;
%if &initdata %then %do;
   data _null_;
      set &init %if &rowsfixed %then (drop=&fixed); nobs=__nobs;
      array __x[*] _numeric_;
      length __name $ 32;
      if _n_ eq 1 then do;
         if dim(__x) ne &m then do;
            put "ERROR: INIT= data set must have %nrbquote(&m)"
                ' design variables' " with design &list..";
            call symput('abort', '1');
            stop;
            end;
         if __nobs ne &n then do;
            put "ERROR: INIT= data set must have %nrbquote(&n) observations.";
            call symput('abort', '1');
            stop;
            end;
         do __j = 1 to dim(__x);
            call vname(__x[__j], __name);
            if compress('x' || put(__j, 5.)) ne lowcase(__name) then do;
               put 'ERROR: INIT= variables must be numeric and ' /
                   +7 "must be named x1 through x%nrbquote(&m) in that order.";
               call symput('abort', '2');
               stop;
               end;
            end;
         end;
      do __j = 1 to dim(__x);
         __max = input(scan("&_mktlist", __j), ?? 12.);
         __val = __x[__j];
         if nmiss(__val) or abs(__val) gt __max or __val ne round(__val)
            then do;
            put 'ERROR: INIT= (' _n_ +(-1) ',' __j
                +(-1) ') value of ' __val 'is invalid.';
            call symput('abort', '1');
            stop;
            end;
         end;

      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 and &abort = 0 %then %let abort = 1;

   %if &abort = 2 %then %do;
      proc contents p data=&init; ods select position; run;
      %let abort = 1;
      %end;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------make FIXED= look like INIT= with 0, missing, and negatives------;
   %if &rowsfixed %then %do;

      data &init;
         set &init;
         array __x[&m] x1-x&m;

         if nmiss(&fixed) then do __j = 1 to &m; __x[__j] = 0; end;

         else if &fixed eq 0 then do;
            call symput('anyneg', '1');
            do __j = 1 to &m; __x[__j] = -__x[__j]; end;
            end;

         drop __j;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %end;

%if &flow %then %mktexdump(tableflag hadflag full factexnlevs tablenote);

*============================ Start Making Designs ============================;

%if &full or &factexnlevs %then %do;
   options &savenote;
   %put NOTE: Generating the &destype design, n=&tabsize..;
   options &setnote;
   %let tablenote = 1;
   %end;

%*------use PLAN to make a full-factorial------;
%if &full %then %do;

   *mkttime(Full Factorials - PLAN);

   %if &flow %then %put &spacer Making full-factorial, &factorslist..;
   %let _mktver = obsolete;
   %mktdes(run=plan, size=&tabsize, cand=__tmpxcand, options=&showcode,
           factors=&factorslist, big=%eval(&tabsize+1))

   %if &nreps gt 1 %then %do;
      data __tmpxcand(drop=i);
         set __tmpxcand;
         do i = 1 to &nreps; output; end;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %if &_mktver ne &localver %then %do;
      %put ERROR: Incompatible MktDes macro.;
      %put %str(      ) You must have the same version of all macros.;
      %put %str(      ) Versions: MktEx = &localver, MktDes = &_mktver..;
      %put %str(      ) Get the latest tools and documentation from:;
      %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
      %let abort = 1;
      %end;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*======================== Fractional Factorials, FACTEX =======================;
%if &factexnlevs %then %do;
   *mkttime(Fractional Factorial - FACTEX);
   %end;

*------get fractional-factorial init design, perhaps dropping variables------;
%let facdone = 0; %let dropvars = 0; %let pass = 0;
%if &factexnlevs %then %do %until(&facdone);

   %let pass = %eval(&pass + 1);
   %if &flow and &dropvars %then %put &spacer dropping &dropvars variables.;
   %else %if &flow and &pass gt 1 %then %put &spacer FACTEX pass &pass..;

   data _null_;
      length newlist mktlist $ &maxlen w $ 23
             tableflag $ %eval(3*&maxlev+10) nlevs $ 3;
      array any[&maxlev] any01-any&maxlev;
      retain any01-any&maxlev 0;
      abort = 0; nvars = &m; nruns = &tabsize; maxlevneed = &maxlevneed;
      dropvars = &dropvars;  npars = &npars;   m = &m;
      factexnlevs = &factexnlevs; powerof = &powerof;

      *------expand factors list------;
      mktlist = symget('_mktlist');

      *------count levels of factors------;
      do i = 1 to &maxlen until(w eq ' ');
         w = scan(mktlist, i, ' ');
         if w ne ' ' then do;
            n = input(w, ?? 23.);
            if 2 le n le &maxlev then any[n] + 1;
            end;
         end;

      *------give FACTEX an easier time with 2/4 mixes------;
      if powerof eq 2 and any04 then do;
         lognruns = log(nruns) / log(4);
         if any02 + any03 + any04 eq nvars and
            any02 + 3 * (any03 + any04) lt nruns and
            abs(lognruns - round(lognruns)) lt 1e-8 then do;
            any04 + any03 + ceil(any02 / 3);
            any02 = 0;   any03 = 0;   m = any04;
            end;
         end;

      *------drop variables if necessary------;
      do while(dropvars gt 0 or npars gt nruns);

         *------first drop unbalanced factors------;
         do i = maxlevneed to 2 by -1 while(dropvars);

            if any[i] then do;
               logi = log(i) / log(&factexnlevs);
               if abs(logi - round(logi)) ge 1e-8 then do;
                  if any[i] ge dropvars then do;
                     npars    + -(i - 1) * dropvars;
                     any[i]   + -dropvars;
                     m        + -dropvars;
                     dropvars = 0;
                     end;
                  else do;
                     npars    + -(i - 1) * any[i];
                     dropvars + -any[i];
                     m        + -any[i];
                     any[i]   = 0;
                     end;
                  end;
               end;
            end;

         *------then try the balanced factors------;
         do i = maxlevneed to 2 by -1 while(dropvars);
            if any[i] then do;
               if any[i] ge dropvars then do;
                  npars    + -(i - 1) * dropvars;
                  any[i]   + -dropvars;
                  m        + -dropvars;
                  dropvars = 0;
                  end;
               else do;
                  npars    + -(i - 1) * any[i];
                  dropvars + -any[i];
                  m        + -any[i];
                  any[i]   = 0;
                  end;
               end;
            end;

         if npars gt nruns then dropvars = 1;
         end;

      *------see if we can optimally tweak the number of levels------;
      if powerof eq 2 then do;
         lognruns = log(nruns) / log(4);
         if abs(lognruns - round(lognruns)) lt 1e-8 and
            (any04 + any16 + any64) eq m then factexnlevs = 4;
         lognruns = log(nruns) / log(8);
         if abs(lognruns - round(lognruns)) lt 1e-8 and
            (any08 + any64       ) eq m then factexnlevs = 8;
         lognruns = log(nruns) / log(16);
         if abs(lognruns - round(lognruns)) lt 1e-8 and
            any16                  eq m then factexnlevs = 16;
         end;
      else if powerof eq 3 then do;
         lognruns = log(nruns) / log(9);
         if abs(lognruns - round(lognruns)) lt 1e-8 and
            (any09 + any81       ) eq m then factexnlevs = 9;
         lognruns = log(nruns) / log(27);
         if abs(lognruns - round(lognruns)) lt 1e-8 and
            any27                  eq m then factexnlevs = 27;
         end;
      else if powerof eq 5 then do;
         lognruns = log(nruns) / log(25);
         if abs(lognruns - round(lognruns)) lt 1e-8 and
            any25                  eq m then factexnlevs = 25;
         end;
      else if powerof eq 7 then do;
         lognruns = log(nruns) / log(49);
         if abs(lognruns - round(lognruns)) lt 1e-8 and
            any49                  eq m then factexnlevs = 49;
         end;

      call symput('factexnlevs', compress(put(factexnlevs, 2.)));

      *------build MktDes syntax for FACTEX------;
      newlist = ' ';
      do i = maxlevneed to 2 by -1;
         if any[i] then do;
            if i lt 100 then nlevs = put(i, z2.);
            else             nlevs = put(i, 3.);
            newlist = trim(newlist) || ' ' ||
                      compress('x'  || nlevs || '_1-x' || nlevs || '_'  ||
                               put(any[i], 5.) || '=' || nlevs);
            tableflag = trim(tableflag) || '_' || nlevs;
            end;
         end;

      if &generic and dropvars ge nvars then
         put 'ERROR: Apparent bug in variable dropping.';
      call symput('factorslist', trim(left(newlist)));
      call symput('tableflag'  , compress(tableflag));
      run;

   %if &flow %then %mktexdump(factorslist tableflag);

   *------use FACTEX to make a fractional-factorial initialization------;
   %if %nrbquote(&tableflag) ne 0 %then %do;
      %let j = Making fractional-factorial, FACTEX, &factorslist.;
      %if &flow %then %put &spacer &j, nlev=&factexnlevs, size=&tabsize..;
      %let i = ;
      %if &dropvars eq 0 and &n eq &tabsize %then %do;
         %let i = , facopts=ranorder;
         %if %sysevalf(&sysver ge 9.2) %then %let i = &i(seed=&seed);
         %else %if %sysevalf(&sysver ge 9) %then %let i = &i(&seed restart);
         %if &factexnlevs eq 2 and not %sysfunc(mod(&pass, 2)) and
            (not (&pass eq 2 and &n eq 128)) %then %let i = &i check=lots;
         %end;

      %let _mktver = obsolete;
      %mktdes(run=factex, size=&tabsize, cand=__tmpxcand, options=&showcode,
              factors=&factorslist, nlev=&factexnlevs &i)

      %if &_mktver ne &localver %then %do;
         %put ERROR: Incompatible MktDes macro.;
         %put %str(      ) You must have the same version of all macros.;
         %put %str(      ) Versions: MktEx = &localver, MktDes = &_mktver..;
         %put %str(      ) Get the latest tools and documentation from:;
         %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
         %let abort = 1;
         %end;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      *------see if FACTEX worked.  If not drop a var and try again------;
      data _null_;
         length seed $ 12;
         if 0 then set __tmpxcand nobs=n;
         call symput('facdone', put(n ne 0, 1.));

         if &flow and n then put '** FACTEX worked.';

         %*------FACTEX failed------;
         if n = 0 then do;

            *------if we really should not have failed, try again------;
            if ((&pass le 4 and &allfactex) or (&pass le 1 and &n eq 128))
               then do;
               seed = '0';
               if &seed then do;
                  seed = put(1e8 * uniform(&seed), 9.0 -L);
                  call symput('seed', compress(seed));
                  end;
               if &flow or &flowinit then
                  put '** Should work, try again (seed=' seed +(-1) ').';
               end;

            %*------otherwise, drop some vars and try again------;
            else do;
               dropvars = &dropvars + 1;
               call symput('dropvars', compress(put(dropvars, 5.)));
               if dropvars ge &m or &noqc then do;
                  call symput('facdone',   '1');
                  call symput('allfactex', '0');
                  call symput('tableflag', '0');
                  put 'WARNING: PROC FACTEX step failed, '
                      'iterations continuing.';
                  end;
               end;
            end;

         %*------success!------;
         else if symget('allfactex') eq '1' and &dropvars eq 0 and
                 &nointeract then do;
            call symput('optiter' , '0'); /* Do not need OPTEX when FACTEX */
            call symput('optiter2', '0'); /* makes all the variables and   */
            call symput('optflag' , '0'); /* there are no restrictions.    */
            if &flow then put '** FACTEX worked optiter=0 optiter2=0 optflag=0';
            end;
         if _error_ then call symput('abort', '1');
         stop;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %end;

*========================== Candidate Set for OPTEX ===========================;

*-----get the candidate set------;
%if &optflag %then %do;
   *mkttime(Candidate Set);
   options &savenote;
   %put NOTE: Generating the candidate set.;
   %if &n ge 200 and &nointeract %then %do;
      %let i = results with OPTITER=0.;
      %put NOTE: With a design this large, you may get faster &i;
      %end;
   options &setnote;
   %end;

%if &optflag %then %do %until(&facdone);
   %let facdone = 1;

   *------make candidate set------;
   %if &flow %then
      %put &spacer Making candidate set, big1=&big1, &factorslist..;

      %let _mktver = obsolete;

      %if &tryfac %then %do;
         %if &factexnlevs eq 0 or not &r5 %then %let i = 2;
         %else %let i = &factexnlevs;
         %mktdes(factors=&varlist, procopts=noprint, cand=__tmpxcand2, big=1,
                 nlev=&i, facopts=check=lots, run=factex, options=&showcode)

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

         data _null_;
            set __tmpxcand2 nobs=__n;
            if __n gt &n then call symput('tryfac', '0');
            if _error_ then call symput('abort', '1');
            stop;
            run;

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
         %end;

      %if not &tryfac %then %do;
         %mktdes(factors=&varlist, procopts=noprint, cand=__tmpxcand2,
                 %if &allfactex %then %do;
                    %if &big1 %then big=1,;
                    %if &factexnlevs eq 2 %then nlev=2, facopts=check=lots,;
                    %end;
                 %else %do; nlev=&candnlev, big=&big, %end;
                 run=plan factex, options=&showcode)

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
         %end;

   %if &_mktver ne &localver %then %do;
      %put ERROR: Incompatible MktDes macro.;
      %put %str(      ) You must have the same version of all macros.;
      %put %str(      ) Versions: MktEx = &localver, MktDes = &_mktver..;
      %put %str(      ) Get the latest tools and documentation from:;
      %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
      %let abort = 1;
      %end;

   *------impose restrictions on candidate set------;
   %if &userrestricts or &partial ne %then %do;

      proc iml &imlopts;
         %if &resflag %then %&resmac;
         start badness(x) global(xmat, i, try
                                 %do i = 1 %to &exchange; , j&i %end;
                                 %if &resflag %then , &reslist;);
            %if not &nox %then %do i = 1 %to &m; x&i = x[&i]; %end;
            bad = 0;
            %if &balance ne    %then %do; __bbad = 0;     %end;
            %if &partial ne    %then %do; %__partprof   ; %end;
            %if &userrestricts %then %do; %&restrictions; %end;
            return(bad %if &partial ne %then + __pbad;);
            finish;
         %do i = 1 %to &exchange; j&i = 1; %end; i = 1;
         use __tmpxcand2;
         read all into xmat;
         close __tmpxcand2;
         n = nrow(xmat);
         f = j(n, 1, 1);
         do i = 1 to n;
            xsubi = xmat[i,];
            if badness(xsubi) then f[i,] = 0;
            end;
         locf = loc(f);
         if ncol(locf) >= &n then do;
            xmat  = xmat[locf,];
            vname = 'x1' : rowcatc('x' || char(ncol(xmat)));
            create __tmpxcand2 from xmat[colname=vname];
            append from xmat;
            end;
         else do;
            call symput('optabort', '2');
            call symput('optflag',  '0');
            end;
         quit;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %*------if in an all FACTEX or PLAN problem, skip OptEx?------;
   %else %if &allfactex or &full %then %do;
      data _null_;
         if 0 then set __tmpxcand2 nobs=n;
         if n eq 0  then call symput('allfactex', '0');
         if n eq &n then call symput('optabort' , '1');
         call symput('facdone', put(n ne 0, 1.));
         if _error_ then call symput('abort'    , '1');
         stop;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      %if &optabort %then %do;
         %if &flow %then %put &spacer optabort=1, skipping OptEx.;

         data __tmpxopt;
            set __tmpxcand2;
            if _error_ then call symput('abort', '1');
            run;

         %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
         %end;
      %end;

   *------check the results------;
   %if &optflag and not &optabort %then %do;
      %let _optseed = 0;

      data _null_;
         if 0 then set __tmpxcand2 nobs=n;
         call symput('ncands', compress(put(n, comma12.)));
         if _error_ then call symput('abort', '1');
         stop;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %if &noqc %then %let facdone = 1;
   %end;

*============ Start Main Body Of Three Sets Of Iterative Algorithms ===========;
*mkttime(Start Three Sets Of Iterations);

%let nrunsintab = &tabsize;
%if &factexnlevs %then %let tabsize = 0;

*-------pass 1 (Initial) or 2 (Search) or 3 (Refinement)------;
%do pass = &firstpass %to 3;

   %let main = 0; /* not in main IML code yet */

   *------iterate using the best method------;
   %if &pass eq 2 %then %do;

      data _null_;
         call symput('optiter',    symget('optiter2'));
         call symput('maxtime',    symget('maxtime2'));
         call symput('mutate' ,    symget('mutate2' ));
         call symput('anneal' ,    symget('anneal2' ));
         call symput('tabiter',    symget('tabiter2'));
         call symput('unbalanced', symget('unbalanced2'));
         call symput('canditer',   symget('canditer2'));
         call symput('mutiter',    symget('mutiter2'));
         call symput('anniter',    symget('anniter2'));
         call symput('init',       symget('out'));
         call symput('initdata',   '1');
         call symput('readinit',   'nt = 0');
         call symput('optflag',    '0');
         call symput('optabort',   '0');
         call symput('initrand',   '1');
         call symput('mutateflag', put(symget('mutate') ne '0', 1.));
         call symput('annealflag', put(symget('anneal') ne '0', 1.));
         if symget('method') =: 'Can' then do;
            call symput('maxiter', symget('canditer2'));
            call symput('optflag',   '1');
            call symput('tableflag', '0');
            n = 1 + input(symget('canditer'), ?? 23.);
            end;
         else if symget('method') =: 'Tab' then do;
            call symput('maxiter', symget('tabiter2'));
            n = 1 + input(symget('tabiter'), ?? 23.);
            end;
         else do;
            call symput('maxiter',   symget('maxiter2'));
            call symput('tableflag', '0');
            n = 1;
            end;
         if symget('mutiter') = '.'
            then call symput('mutiter', compress(put(n, 12.)));
         if symget('anniter') = '.'
            then call symput('anniter', compress(put(n, 12.)));
         if &setoptiter in (0, 2) and not &generic and &nostimer then do;
            n = round(60 * &maxtime / (&opttime / &optiter), 100);
            n = min(max(n, 1), 5000);
            if &maxtime2 = 0 then optiter = 1;
            call symput('optiter', compress(put(n, 12.0)));
            end;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %*------refine the best design------;
   %else %if &pass eq 3 %then %do;

      data _null_;
         call symput('maxtime',    symget('maxtime3'));
         call symput('mutate' ,    symget('mutate3' ));
         call symput('anneal' ,    symget('anneal3' ));
         call symput('mutiter',    symget('mutiter3'));
         call symput('anniter',    symget('anniter3'));
         call symput('maxiter',    symget('maxiter3'));
         call symput('mutateflag', put(symget('mutate') ne '0', 1.));
         call symput('annealflag', put(symget('anneal') ne '0', 1.));
         call symput('initdata',   '1');
         call symput('readinit',   'nt = 0');
         call symput('optflag',    '0');
         call symput('tableflag',  '0');
         call symput('initrand',   '0');
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      %end;

   *------run OPTEX------;
   %if &optflag and not &optabort %then %do;
      options &savenote;
      %let j = Performing &optiter searches of &ncands candidates;
      %if &pass eq &firstpass %then %let j = &j, full-factorial=&_mktff;
      %put NOTE: &j..;
      options &setnote;

      %if &flow %then
        %put &spacer Searching candidate set, iter=&optiter, &factorslist..;
      *mkttime(OPTEX);
      %let _mktver = obsolete;
      %let opttime = %sysfunc(datetime());
      %mktdes(factors=&optvarlist, out=__tmpxopt, n=&n, cand=__tmpxcand2,
              procopts=noprint, seed=&seed, run=optex,
              options=&showcode, keep=1, iter=&optiter);
      %let opttime = %sysevalf(%sysfunc(datetime()) - &opttime);

      %if &_mktver ne &localver %then %do;
         %put ERROR: Incompatible MktDes macro.;
         %put %str(      ) You must have the same version of all macros.;
         %put %str(      ) Versions: MktEx = &localver, MktDes = &_mktver..;
         %put %str(      ) Get the latest tools and documentation from:;
         %put http://support.sas.com/techsup/tnote/tnote_stat.html#market;
         %let abort = 1;
         %end;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

      *------check the results------;
      data _null_;
         if 0 then set __tmpxopt nobs=n;
         if n ne &n then do;
            call symput('optflag',  '0');
            call symput('optabort', '2');
            if &generic then put 'ERROR: PROC OPTEX iterations failed. ' n=;
            if &pass > 1 then do;
               put 'ERROR: PROC OPTEX iterations failed.';
               call symput('abort', '1');
               end;
            end;
         seed = symget('sysranend');
         if seed = ' ' then seed = symget('_optseed');
         if seed = '0' then seed = put(1e8 * uniform(&seed), 9.0);
         call symput('seed', compress(seed));
         if _error_ then call symput('abort', '1');
         stop;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %let main = 1;

   *===================== Proc IML Does Most of the Work ======================;
   *mkttime(Proc Iml);

   proc iml %if %sysevalf(&sysver >= 9) %then symsize=2200; &imlopts;
      file print;

      *------module that imposes restrictions on designs------;
      %if &restrictflag %then %do;
         %if &resflag %then %&resmac;
         start badness(x) global(xmat, i, try %if &ridgemsg %then , __facts;
                                 %do i = 1 %to &exchange; , j&i %end;
                                 %if &resflag %then , &reslist;);
            %if not &nox %then %do i = 1 %to &m; x&i = x[&i]; %end;
            bad = 0;
            %if &balance ne    %then %do; %__balance    ; %end;
            %if &partial ne    %then %do; %__partprof   ; %end;
            %if &userrestricts %then %do; %&restrictions; %end;

            return(bad
                   %if &balance ne %then + __bbad;
                   %if &partial ne %then + __pbad;);
            finish;
         %end;

      *------orthogonal coding for 0 to 100 efficiency values------;
      start orthogcode(x, levels);
         xstar  = shape(x, levels - 1, nrow(x))`;
         j = shape(1 : (levels - 1), nrow(x), levels - 1);
         r = sqrt(levels # (x / (x + 1))) # (j = xstar) -
             sqrt(levels / (j # (j + 1))) # (j > xstar | xstar = levels);
         return(r);
         finish;

      *------use table to initialize some of the factors------;
      %if &tableflag ne 0 %then %do;

         *============== Orthogonal Array Construction Functions ==============;

         *------1 and -1 coding to 1 and 2 coding------;
         start twolev(h);
            return((h = -1) + 1);
            finish;

         *------make binary matrix from n(0) n(1) n(0) n(1) ... ------;
         start expand(x);
            n = ncol(x);    r = j(1,sum(x),0);    k = 0;    c = (x[1] = 0);
            do i = c + 1 to n;
               r[(k+1):(k+x[i])] = c;
               k = k + x[i];
               c = ^c;
               end;
            return(r);
            finish;

         /*
          * start contract(x);
          *    n = ncol(x);
          *    c = 1;
          *    free r;
          *    if x[1] = 1 then r = 0;
          *    do i = 1 to n;
          *       do j = i + 1 to n while(x[i] = x[j]); end;
          *       r = r || (j - i);
          *       i = j - 1;
          *       end;
          *    return(r);
          *    finish;
          * options nocenter;
          * print fac[format=2. colname=' ' rowname=' '];
          * options center;
          *
          * options ps=5000;
          * n = ncol(v);
          * print n;
          * j = 1;
          * file print;
          * do i = 2 to n;
          *   if v[i] ^= v[j] then do;
          *    d = i - j;
          *    if d = 1 then put (v[j]);
          *    else put '[' d ']' (v[j]);
          *    j = i;
          *    end;
          *   end;
          * put (v[n]);
          */

         *------1 and 2 coding to 2 and 1 coding------;
         start reverse(h);
            return((h = 1) + 1);
            finish;

         *------shift a vector to the right to form a matrix------;
         start shiftr(a);
            n = ncol(a);
            a = shape(a, n, n);
            c = n || (1:(n-1));
            do i = 2 to n;
               a[i,] = a[i-1,c];
               end;
            return(a);
            finish;

         *------shift a vector to the left to form a matrix------;
         start shiftl(a);
            n = ncol(a);
            a = shape(a, n, n);
            c = (2:n) || {1};
            do i = 2 to n;
               a[i,] = a[i-1,c];
               end;
            return(a);
            finish;

         *------Hadamard, Paley method 1------;
         start hadp1(n, had);
            nm1 = n - 1;
            h = j(n, n, -1);
            h[1,1] = 1;
            do j = 1 to n - 2; h[1, mod(j * j, nm1) + 1] = 1; end;
            do i = 2 to n; h[i,1:nm1] = h[i-1, nm1] || h[i-1, 1:(nm1 - 1)]; end;
            h = h[2:n,]`;
            if had then h = j(n, 1, 1) || h;
            else        h = (h = -1) + 1;
            return(h);
            finish;

         *------Hadamard, Paley method 2 set up, n / 2 - 1 prime------;
         start hadp2(n, had);
            nd2 = n / 2;
            p   = nd2 - 1;
            chi = j(1, p, -1); chi[1] = 0;
            do i = 1 to p-1; chi[mod(i*i, p) + 1] = 1; end;
            s = j(nd2, nd2, 1); s[1,1] = 0;
            do i = 2 to nd2;
               s[i,1] = 1;
               do j = 2 to nd2;
                  d = (i-2) - (j-2);
                  do while (d < 0); d = d + p; end;
                  s[i, j] = chi[mod(d, p) + 1];
                  end;
               end;
            free chi;
            return(hadp2b(n, s, had));
            finish;

         *------Hadamard, Paley method 2, n / 2 - 1 prime power------;
         start hadp2pp(n); /* n=340, 580, 724, 1060 */
            nd2 = n / 2;
            if n = 340 then chi = expand({ 0 13 2 1 1 2 3 2 1 1 4 2 1 2 1
            2 2 2 1 1 1 1 3 1 1 1 1 1 1 1 3 4 8 3 1 3 2 1 1 1 2 2 2 2 1 2 2 2 2
            1 1 1 2 3 1 3 8 4 3 1 1 1 1 1 1 1 3 1 1 1 1 2 2 2 1 2 1 2 4 1 1 2 3
            2 1 1 1}) # 2 - 1;
            else if n = 580 then chi = expand({1 16 3 2 1 1 1 1 1 1 1 1 1
            2 4 1 2 3 2 3 2 1 1 2 4 3 1 3 8 3 1 1 2 1 1 3 2 2 2 2 1 1 3 1 1 2 1
            2 1 2 1 1 1 1 4 1 1 1 2 4 2 1 2 1 1 1 2 1 1 1 5 1 2 6 3 6 2 1 5 1 1
            1 2 1 1 1 2 1 2 4 2 1 1 1 4 1 1 1 1 2 1 2 1 2 1 1 3 1 1 2 2 2 2 3 1
            1 2 1 1 3 8 3 1 3 4 2 1 1 2 3 2 3 2 1 4 2 1 1 1 1 1 1 1 1 1 2 2})
            # 2 - 1;
            else if n = 724 then chi = expand({1 18 2 1 4 7 4 1 1 1 1 3 1
            1 6 1 1 3 1 2 2 2 1 1 2 1 2 1 1 2 3 1 3 1 1 3 2 3 1 1 3 2 3 2 1 1 1
            1 1 1 1 2 3 3 4 2 1 2 1 2 3 1 4 1 1 1 2 5 2 1 1 1 2 1 2 1 1 1 1 2 2
            2 1 1 3 4 3 1 2 3 2 1 1 1 2 3 2 1 3 4 3 1 1 2 2 2 1 1 1 1 2 1 2 1 1
            1 2 5 2 1 1 1 4 1 3 2 1 2 1 2 4 3 3 2 1 1 1 1 1 1 1 2 3 2 3 1 1 3 2
            3 1 1 3 1 3 2 1 1 2 1 2 1 1 2 2 2 1 3 1 1 6 1 1 3 1 1 1 1 4 7 4 1 1
            }) # 2 - 1;
            else if n = 1060 then chi = expand({1 22 2 2 1 1 1 1 3 3 3 1 1
            1 1 2 1 1 1 2 1 1 3 2 2 2 3 1 1 2 5 2 2 3 1 1 1 3 2 2 3 1 2 4 3 1 2
            1 3 4 2 2 5 2 1 5 1 2 6 1 3 1 1 4 4 4 1 1 3 2 3 2 2 1 1 3 1 1 2 2 1
            1 2 1 2 1 2 1 1 1 1 4 1 1 1 1 3 1 1 2 1 1 2 1 3 5 3 1 1 1 1 5 1 1 3
            1 1 1 2 1 1 1 3 3 1 2 1 3 1 1 4 1 4 1 1 1 4 1 4 1 1 3 1 2 1 3 3 1 1
            1 2 1 1 1 3 1 1 5 1 1 1 1 3 5 3 1 2 1 1 2 1 1 3 1 1 1 1 4 1 1 1 1 2
            1 2 1 2 1 1 2 2 1 1 3 1 1 2 2 3 2 3 1 1 4 4 4 1 1 3 1 6 2 1 5 1 2 5
            2 2 4 3 1 2 1 3 4 2 1 3 2 2 3 1 1 1 3 2 2 5 2 1 1 3 2 2 2 3 1 1 2 1
            1 1 2 1 1 1 1 3 3 3 1 1 1 1 2 1}) # 2 - 1;
            p   = nd2 - 1;
            s = j(nd2, nd2, 1); s[1,1] = 0;
            c2 = sqrt(p);
            neg = j(1, p, .);
            do i = 0 to p-1; do j = 0 to p-1;
               t = mod(mod(i , c2) + mod(j , c2), c2) +
                   c2 # mod(int(i / c2) + int(j / c2), c2);
               if t = 0 then neg[i + 1] = j;
               end; end;
            do i = 2 to nd2; do j = 2 to nd2;
               t = neg[j - 1];
               t = mod(mod(i - 2 , c2) + mod(t, c2), c2) +
                   c2 # mod(int((i - 2) / c2) + int(t / c2), c2);
               s[i,j] = chi[t + 1];
               end; end;
            free neg chi;
            return(hadp2b(n, s, 1));
            finish;

         *------Hadamard, Paley method 2, actual construction------;
         start hadp2b(n, s, had);
            k = 1;   nd2 = n / 2;   h = j(n, n, 1);
            do l = 1 to nd2;
               do j = 1 to -1 by -2;
                  ix = 1;
                  do i = 1 to nd2;
                     if (i ^= l) then do;
                        h[k,ix] = s[l,i]; h[k,ix+1] = j * s[l,i];
                        end;
                     else do; h[k,ix] = j; h[k,ix+1] = -1; end;
                     ix = ix + 2;
                     end;
                  h[k,2] = -h[k,2];
                  k = k + 1;
                  end;
               end;
            h = h`; h = h[,1] # h; h = h # h[1,];
            if had = 0 then h = twolev(h[,2:n]);
            return(h);
            finish;

         *------special code for Hadamards of size 12 and 20------;
         start had(n, had); /* 0 -> 1, 2   level coding, n-1 cols;
                               1 -> 1, -1  level coding, n   cols */
            if n = 12 then v = {1 0 1 0 0 0 1 1 1 0 1};
            else v = {0 1 0 0 1 1 1 1 0 1 0 1 0 0 0 0 1 1 0};
            n = ncol(v);
            h = shape(v, n, n);
            do i = 2 to n; h[i,] = h[i-1, n] || h[i-1, 1:(n-1)]; end;
            h = j(1, n, (n = 19)) // h;
            if had = 1 then h = j(n + 1, 1, 1) || (2 # h - 1);
            else            h = h + 1;
            if had = 9 then h = h[{1 2 12 3 8 5 4 7 9 6 10 11}, 2:10];
            return(h);
            finish;

         *------Hadamard, Method Ehrlrich------;
         start hadc(n);
            if n = 324 then do;
               m  = sqrt(n); mp = m + 1; mm = m - 1;
               he = {0 1 0 1 0 0 1 0 0 1 1 0 0 0 1 1 1 1 0};
               hf = {0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1};
               ha = j((m - 2), (m - 2), 0);
               hb = j((m - 2), (m - 2), 0);
               ha[1:2,] = { 0 0 0 0 1 0 1 1 0 0 1 1 1 1 0 1 ,
                            0 0 1 1 1 0 1 0 0 1 1 0 0 0 0 1 } # 2 - 1;
               hb[1:2,] = { 0 1 1 1 0 1 0 0 1 1 0 0 0 0 1 0 ,
                            1 0 0 0 0 1 0 1 1 0 0 1 1 1 1 0 } # 2 - 1;
               do i = 1 to 7;
                  ha[2 # i + 1,] = ha[2 # i - 1,(m - 3):(m - 2)] ||
                                   ha[2 # i - 1,1:(m - 4)];
                  ha[2 # i + 2,] = ha[2 # i    ,(m - 3):(m - 2)] ||
                                   ha[2 # i    ,1:(m - 4)];
                  hb[2 # i + 1,] = hb[2 # i - 1,(m - 3):(m - 2)] ||
                                   hb[2 # i - 1,1:(m - 4)];
                  hb[2 # i + 2,] = hb[2 # i    ,(m - 3):(m - 2)] ||
                                   hb[2 # i    ,1:(m - 4)];
                  end;
               ha = (ha || j(8, 1, 1) @ {-1 1}`) //
                          (j(1, 8, 1) @ {-1 1} || {-1});
               hb = (hb || j(8, 1, 1) @ {1 -1}`) //
                          (j(1, 8, 1) @ {1 -1} || {-1});

               hd = j(mp, mp, 1);
               do i = 1 to mp;
                  k = i;
                  h = he[i];
                  do j = 1 to 20 - i;
                     hd[k,j] = h;
                     h = mod(h + 1, 2);
                     hd[j,k] = h;
                     k = k + 1;
                     end;
                  end;
               hd[mp,1:m] = hf;
               hd[1:m,mp] = mod(hd[mp,1:m]` + 1, 2);

               h = j(n, n, 1);
               hc = j(mm, mm, 1);
               vi = 2:m;
               do i = 1 to mp;
                  vj = 2:m;
                  do j = 1 to mp;
                     if i = j then        h[vi,vj] = hc;
                     else if hd[i,j] then h[vi,vj] = hb;
                     else                 h[vi,vj] = ha;
                     vj = vj + mm;
                     end;
                  vi = vi + mm;
                  end;
               end;
            free ha hb hc hd he hf vi vj;
            return(h);
            finish;

         *------Hadamard, Method Turyn 2------;
         start hadd(n);
            p = n / 12;   pm1 = p - 1;
            if n = 372 then do;
               ha = shape(expand({ 0 6 3 1 1 1 1 2 2 2 1 1 1 1 3 5}) # 2 - 1,
                          p, p);
               hb = shape(expand({ 0 1 5 3 1 1 1 1 2 2 2 1 1 1 1 3 5}) # 2 - 1,
                          p, p);
               hc = shape(expand({ 1 2 3 1 1 2 1 4 2 4 1 2 1 1 3 2}) # 2 - 1,
                          p, p);
               end;
            else if n = 612 then do;
               ha = shape(expand({ 0 1 3 3 1 2 1 1 1 3 2 5 2 2 2 5 2 3 1 1 1 2 1
               3 3}) # 2 - 1, p, p);
               hb = shape(expand({ 0 4 3 1 2 1 1 1 3 2 5 2 2 2 5 2 3 1 1 1 2 1 3
               3}) # 2 - 1, p, p);
               hc = shape(expand({ 1 1 2 1 4 1 1 3 1 1 1 5 2 1 2 1 2 5 1 1 1 3 1
               1 4 1 2 1}) # 2 - 1, p, p);
               end;
            else if n = 732 then do;
               ha = shape(expand({ 0 2 2 1 2 2 2 1 1 1 1 4 2 1 5 1 6 1 5 1 2 4 1
               1 1 1 2 2 2 1 2 1}) # 2 - 1, p, p);
               hb = shape(expand({ 4 1 1 1 1 4 3 2 2 1 1 2 1 1 3 6 3 1 1 2 1 1 2
               2 3 4 1 1 1 1 3}) # 2 - 1, p, p);
               hc = shape(expand({ 1 3 1 1 1 1 4 3 2 2 1 1 2 1 1 3 6 3 1 1 2 1 1
               2 2 3 4 1 1 1 1 3}) # 2 - 1, p, p);
               end;
            else if n = 756 then do;
               ha = shape(expand({ 0 4 1 2 1 1 1 4 1 1 3 1 3 3 3 6 3 3 3 1 3 1 1
               4 1 1 1 2 1 3}) # 2 - 1, p, p);
               hb = shape(expand({ 1 1 1 3 2 2 1 2 2 1 2 1 1 2 1 1 1 3 8 3 1 1 1
               2 1 1 2 1 2 2 1 2 2 3 1 1}) # 2 - 1, p, p);
               hc = shape(expand({ 2 1 3 2 2 1 2 2 1 2 1 1 2 1 1 1 3 8 3 1 1 1 2
               1 1 2 1 2 2 1 2 2 3 1 1}) # 2 - 1, p, p);
               end;

            if n = 372 | n = 612 then
               hd = shape({4 4 4 5 1 0 6 6 5 0 6 6 4 2 5 2 1 6 0 6 1 6 6 6 4 1 2
               4 6 0 1 5 6 6 0 6 5 4 2 2 0 0 0 0 0 1 1 6 5 6 0 0 4 4 4 0 6 5 6 5
               5 0 6 0 4 2 0 2 6 0 5 1 0 6 5 1 4 6 2 4 5 0 0 6 6 6 6 6 0 4 2 2 6
               5 1 1 0 6 1 1 1 0 0 6 4 4 4 0 6 1 0 0 0 5 5 6 4 2 0 2 0 1 6 0 0 1
               6 1 4 6 2 4 6 6 6 0 6 1 5 5 0 4 2 2}, 12) - 3;
            else if n = 732 | n = 756 then
               hd = shape( { 4 4 4 4 2 0 6 5 4 0 5 5 4 2 4 2 2 5 1 6 2 5 6 6 4 2
               2 4 5 1 2 4 6 5 0 6 4 4 2 2 1 1 1 0 0 2 2 6 4 5 1 1 4 4 4 0 6 4 6
               4 4 0 5 1 4 2 0 2 5 0 4 2 1 6 4 2 4 6 2 4 4 0 1 5 6 5 6 5 0 4 2 2
               5 4 2 2 1 6 2 2 2 0 0 5 4 4 4 1 5 2 0 0 0 4 4 5 4 2 1 2 0 2 6 0 1
               2 5 2 4 5 2 4 6 5 5 0 6 2 4 4 1 4 2 2}, 12) - 3;

            do i = 2 to p;
               ha[i,] = ha[i-1,p] || ha[i-1,1:pm1];
               hb[i,] = hb[i-1,p] || hb[i-1,1:pm1];
               hc[i,] = hc[i-1,p] || hc[i-1,1:pm1];
               end;

            h = j(n, n);
            vi = 1:p;
            do i = 1 to 12;
               vj = 1:p;
               do j = 1 to 12;
                  if      hd[i,j] =  1 then h[vi,vj] =  ha;
                  else if hd[i,j] =  2 then h[vi,vj] =  hb;
                  else if hd[i,j] =  3 then h[vi,vj] =  hc;
                  else if hd[i,j] = -1 then h[vi,vj] = -ha;
                  else if hd[i,j] = -2 then h[vi,vj] = -hb;
                  else if hd[i,j] = -3 then h[vi,vj] = -hc;
                  vj = vj + p;
                  end;
               vi = vi + p;
               end;
            h = h[,1] # h; h = h # h[1,];
            free ha hb hc hd vi vj;
            return(h);
            finish;

         *------Hadamard, Yamada, order 412, 604------;
         start hady(n);
            p = (n - 4) / 4;
            pd2 = p / 2;
            a = j(p, p, 0);
            b = j(p, p, 0);
            if n = 412 then do;
               a[1,] = expand({0 2 1 2 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 1 1 1 1
               2 2 1 1 2 2 2 1 1 2 1 1 1 1 2 2 1 1 2 2 1 1 1 1 2 1 1 2 2 2 1 1 2
               2 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 1 1 1 1 1 1 2 1} ) # 2 - 1;
               b[1,] = expand({1 1 1 2 1 1 2 1 1 1 1 1 1 2 1 1 2 2 2 2 2 2 2 2 1
               1 2 1 1 2 1 1 2 2 1 1 1 1 1 1 2 2 1 1 2 1 1 2 1 1 2 2 2 2 2 2 2 2
               1 1 2 1 1 1 1 1 1 2 1 1 2 1}) # 2 - 1;
               c = sqrsym(expand({ 1 1 1 2 1 2 4 2 4 2 4 1 3 1 2 1 2 1 2 1 1 1 2
               2 2 3 3 2 2 1 4 1 1 1 3 3 1 4 3 1 1 1 2 2 4 1 2 1 2 2 1 3 6 2 1 1
               1 2 4 1 1 1 1 3 1 1 1 1 1 4 2 1 1 2 1 3 1 1 4 2 1 1 2 2 1 1 1 1 4
               3 2 1 1 1 1 2 1 3 2 4 5 1 3 1 1 1 2 2 1 3 3 3 2 1 4 1 1 4 1 1 1 1
               3 1 2 1 3 1 2 1 1 1 4 2 2 1 1 1 1 1 2 4 1 2 2 4 2 3 1 2 1 2 1 1 1
               2 2 1 3 3 2 1 1 1 1 2 2 1 2 2 1 2 2 1 2 2 1 1 1 3 3 2 2 2 2 4 2 1
               6 3 1 1 2 3 1 1 3 1 3 4 3 4 4 2 1 1 1 3 2 2 4 1 1 1 1 1 2 3 1 3 4
               1 1 2 3 3 4 1 2 3 1 2 1 2 2 1 1 1 2 4 1 1 2 1 3 1 1 1 1 1 2 2 1 2
               2 3 1 1 2 2 1 1 2 1 2 1 2 1 2 1 1 2 1 6 4 3 1 1 3 1 1 5 1 2 1 2 1
               1 1 1 3 1 8 3 1 1 3 1 4 2 1 1 5 4 1 3 1 5 1 1 2 2 2 1 2 1 3 1 2 1
               2 1 3 4 2 1 2 4 1 2 1 4 3 2 2 1 1 1 2 2 2 1 1 3 1 1 1 2 3 2 1 1 3
               3 1 6 1 1 3 1 2 2 1 4 3 1 1 1 2 2 2 1 1 1 1 1 1 1 2 2 1 2 2 6 5 1
               1 1 1 4 5 1 1 1 3 3 1 1 1 4 3 2 1 3 3 1 1 3 1 1 1 1 1 2 4 1 2 1 2
               1 2 1 1 1 2 3 4 2 1 1 2 5 3 1 1 2 4 2 2 1 2 4 1 2 2 4 4 3 5 4 3 1
               2 1 2 1 2 4 4 1 4 3 7 3 1 5 1 1 1 2 2 1 1 1 3 2 2 2 2 2 1 1 1 1 3
               5 4 1 3 3 1 2 1 3 6 1 3 5 1 1 1 1 1 2 3 2 1 3 4 2 1 1 1 1 1 3 1 3
               1 1 2 2 1 1 1 2 5 1 4 1 1 2 2 2 4 4 1 2 2 1 2 2 1 2 2 3 1 1 2 2 1
               1 6 1 1 3 1 4 3 2 3 1 3 2 1 6 4 3 2 1 2 3 2 1 2 2 2 2 1 1 1 1 3 1
               6 1 1 1 1 3 1 5 1 1 6 1 3 1 1 1 3 5 1 1 1 1 1 2 1 3 1 5 4 1 3 3 3
               2 1 1 1 2 3 1 1 1 1 1 2 3 1 1 2 2 4 1 1 1 1 1 2 3 1 1 1 1 1 2 3 2
               1 1 1 2 3 1 1 1 1 1 2 3 1 1 1 1 2 4 1 1 1 1 1 2 3 1 1 1 1 1 2 1 2
               }) # 2 - 1);
               end;
            else if n = 604 then do;
               a[1,] = expand({0 2 1 2 2 1 1 1 1 2 1 1 1 1 1 1 1 1 2 1 1 2 1 1 2
               2 2 1 1 1 1 1 1 2 2 2 2 1 1 2 2 1 1 1 1 2 2 2 2 2 1 1 2 1 1 1 1 1
               1 2 1 1 2 2 2 2 2 1 1 1 1 2 2 1 1 2 2 2 2 1 1 1 1 1 1 2 2 2 1 1 2
               1 1 2 1 1 1 1 1 1 1 1 2 1 1 1 1 2 2 1} ) # 2 - 1;
               b[1,] = expand({1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 2 1 1 2 1 1 1 1 2
               1 1 2 2 2 1 1 2 1 1 2 2 2 1 1 1 1 2 2 1 1 2 1 1 1 1 1 1 2 2 2 1 1
               1 1 1 1 2 2 2 1 1 1 1 1 1 2 1 1 2 2 1 1 1 1 2 2 2 1 1 2 1 1 2 2
               2 1 1 2 1 1 1 1 2 1 1 2 1 1 2 1 1 1 1 1 1 1 1 1 1 1}) # 2 - 1;
               c = sqrsym(expand({ 1 1 1 2 6 2 3 1 1 1 3 1 2 1 2 1 1 3 1 1 3 4 3
               1 7 1 3 1 4 1 3 1 3 3 3 1 3 1 2 2 2 1 4 3 2 1 2 3 3 1 1 2 4 2 1 1
               3 1 2 2 1 1 1 2 2 1 2 1 1 3 6 3 1 1 1 2 4 1 1 1 1 1 4 2 10 9 1 2
               4 1 2 1 1 4 2 1 2 1 1 3 2 4 3 1 1 1 2 1 2 1 2 3 1 1 2 2 1 2 3 2 2
               1 1 6 2 1 1 5 6 5 1 2 3 3 2 1 2 2 4 3 2 2 1 4 1 1 1 1 2 1 1 1 2 1
               1 2 3 3 1 2 1 2 1 1 2 1 2 1 3 1 1 4 1 3 1 1 3 1 2 1 1 3 1 1 2 7 1
               3 4 2 3 4 5 4 3 1 7 2 1 7 2 1 1 3 1 1 3 1 4 1 1 4 1 4 1 1 3 1 2 1
               2 1 3 3 2 1 1 1 1 3 3 2 1 1 2 1 1 1 4 1 2 2 3 5 1 2 2 3 4 2 3 3 2
               1 5 3 3 2 1 5 7 1 1 2 6 1 1 3 1 1 2 6 1 1 2 2 2 1 2 2 1 1 3 2 1 3
               1 2 2 1 1 3 2 1 2 1 1 1 1 3 4 2 3 1 1 1 1 1 3 4 2 3 1 1 2 2 4 1 1
               2 1 4 2 1 2 4 1 1 2 1 4 2 1 6 1 1 1 1 1 1 7 1 2 1 1 1 1 1 2 1 1 1
               1 1 2 1 1 2 3 10 7 1 1 4 3 1 1 1 1 3 1 1 1 2 2 1 1 1 1 1 3 1 3 1
               3 1 2 1 1 2 2 1 1 6 2 1 1 2 6 2 1 1 3 1 3 2 4 1 2 1 2 2 1 3 2 3 2
               1 1 1 2 1 1 1 1 3 1 4 2 1 1 1 1 2 2 1 1 3 6 1 2 4 3 1 1 1 2 2 2 1
               2 3 1 1 3 1 4 1 1 2 2 1 3 2 1 2 1 2 1 1 1 1 1 1 1 1 3 2 2 1 1 1 1
               1 2 4 1 1 1 1 4 1 2 2 3 2 1 1 2 1 4 1 2 1 2 1 3 2 1 3 1 1 1 2 1 1
               1 1 3 3 1 2 2 1 3 2 1 1 3 1 1 2 2 3 1 1 3 2 1 3 5 1 5 2 2 5 4 1 1
               1 1 3 1 1 1 1 1 1 2 1 1 2 5 1 1 1 5 2 2 5 1 2 1 1 2 1 1 1 1 1 1 3
               1 1 1 1 2 2 5 1 5 3 1 2 2 1 3 3 5 1 1 3 2 2 1 1 3 1 1 2 1 3 3 1 1
               1 1 2 1 1 2 3 2 2 1 2 1 3 1 2 3 1 2 1 2 1 2 2 3 2 2 1 4 1 1 1 1 1
               2 1 2 1 2 3 2 3 2 1 1 1 1 1 2 2 3 1 1 2 1 2 1 2 3 1 2 2 1 1 1 3 4
               2 1 3 4 1 3 1 1 3 2 1 3 1 3 4 2 1 6 1 1 1 1 2 1 1 1 2 3 1 1 2 1 1
               2 2 1 1 1 1 2 4 1 1 1 1 2 1 1 1 2 3 2 3 1 1 2 6 2 1 1 2 1 2 2 1 2
               1 4 2 4 1 1 2 6 2 1 1 2 3 1 3 1 1 1 1 1 2 2 1 1 5 1 1 2 2 1 1 2 1
               5 1 3 1 1 1 1 1 2 2 1 1 1 3 10 3 3 2 1 1 1 1 3 4 1 1 6 10 3 2 1 1
               2 1 1 1 1 1 1 1 1 1 1 1 1 2 1 1 6 1 1 2 1 1 7 6 1 1 2 1 1 6 4 1 1
               7 1 1 1 1 2 1 2 3 4 6 3 3 6 4 3 2 1 1 1 3 1 2 1 1 1 1 2 1 1 4 1 1
               1 2 2 1 1 2 1 3 1 3 2 1 3 1 2 1 1 2 2 1 1 1 4 3 2 5 2 1 2 2 3 1 1
               2 4 2 2 1 1 3 1 1 1 2 2 4 2 1 1 2 4 1 1 4 2 1 1 2 2 2 1 1 2 2 3 2
               1 1 3 1 1 1 1 4 1 1 1 3 1 1 2 3 5 1 1 2 1 2 1 2 3 2 3 1 7 1 2 7 1
               1 1 2 2 1 7 2 1 4 2 1 1 4 1 1 3 2 2 2 1 1 1 2 3 1 1 3 2 1 3 2 1 1
               1 1 1 2 2 3 1 2 3 1 2 1 1 1 1 1 2 3 3 1 2 1 2 4 2 1 3 3 1 2 2 2 1
               2 1 1 2 1 3 1 2 2 2 1 3 1 1 2 1 1 1 2 1 3 1 1 3 1 1 2 1 1 3 1 1 4
               1 3 3 1 1 1 2 1 3 3 1 1 1 3 3 1 3 3 1 4 1 1 6 1 1 1 1 4 1 1 1 1 1
               2 1 3 1 5 1 1 2 5 2 1 5 1 3 1 5 2 1 1 1 1 1 1 2 1 1 1 2 1 1 1 1 1
               4 1 1 1 1 3 1 5 2 1 1 5 1 2 1 2 5 1 3 1 5 5 1 1 1 1 6 1 1 3 1 1 2
               1 1 3 1 3 3 1 2 1 1 1 3 3 1 2 1 3 3 1 3 3 1 1 2 1 1 2 1 1 3 1 1 3
               1 2 4 2 1 2 1 3 1 1 2 1 1 2 1 2 2 2 1 2 1 1 1 2 1 1 3 1 2 2 2 1 1
               4 2 1 2 1 3 3 2 1 1 1 2 2 2 3 1 2 1 1 1 1 1 2 3 1 2 3 1 1 1 1 1 1
               2 1 3 2 1 3 3 1 1 2 2 2 3 1 1 4 1 3 2 3 2 1 4 2 1 1 1 7 2 1 3 1 2
               4 1 2 7 2 3 2 3 2 1 2 1 2 1 1 2 2 2 1 1 2 4 4 1 1 1 1 3 1 1 2 3 2
               5 3 2 1 1 3 1 1 2 2 2 2 1 1 2 4 1 1 2 2 1 2 5 2 4 3 1 1 2 2 4 2 1
               1 4 2 1 1 2 4 2 2 1 2 2 1 2 5 2 3 5 1 1 2 1 1 1 1 2 1 4 3 1 3 1 2
               1 1 2 2 1 1 2 1 1 1 2 2 1 1 2 1 3 1 1 3 1 1 2 1 1 1 1 2 1 3 1 2 1
               1 1 1 7 1 1 4 3 6 4 3 2 1 1 2 3 4 6 2 2 1 1 1 1 7 1 1 4}) # 2 -
               1)`;
               end;

            a[2,] = -a[1,];
            a[2,1:2] = a[1,1:2];
            b[2,] = -b[1,];
            do i = 1 to (pd2-1); c[i,(i+1):pd2] = -c[(i+1):pd2,i]`; end;
            c = c @ j(2,2,1);
            j = (1:(p-2));   k = ((p-1):p) || j;
            do i = 3 to p;
               a[i,] = a[i-2,k];
               b[i,] = b[mod(i - 1, 2) + 1,(i-2):(i-1)] || b[i-2,j];
               end;
            do i = 1 to p; a[,i] = a[i,]`; b[,i] = b[i,]`; c[i,i] = 1; end;
            h = (i(4) //
                 {1 1 1 1, 1 0 1 0, 1 0 0 1, 1 1 0 0} @ j(p,1,1)) # 2 - 1 ||
                ({1 0 0 0, 1 1 1 0, 1 0 1 1, 1 1 0 1} @ j(1,p,1)  # 2 - 1 //
                ( a  ||  b  ||  c  ||  c ) // (-b` ||  a` || -c` ||  c`) //
                (-c` ||  c  ||  a` || -b`) // (-c` || -c  ||  b` ||  a ));
            free a b c;
            h = h[,1] # h; h = h # h[1,];
            return(h);
            finish;

         *------Hadamard, Turyn, Hedayat and Wallis (1978), Spence ------;
         start hadthw(n);
            adj = 0;
            if n = 188 then do;
               ha = { 1 0 1 0 0 0 0 1 0 0 1 0 0 0 0 1 1 0 0 0 1 1 1 1
                      0 0 0 1 0 1 0 0 0 0 1 1 1 1 1 1 0 0 1 0 0 0 1 };
               hb = { 1 0 0 0 1 0 0 1 1 1 1 1 1 0 0 0 0 1 0 1 0 0 0 1
                      0 0 0 1 1 1 0 0 1 1 1 1 0 1 1 0 1 1 1 1 0 1 0 };
               hc = { 1 0 0 1 0 0 1 0 1 1 1 0 0 0 0 0 1 0 0 0 1 0 0 0
                      1 1 0 0 1 0 1 1 0 1 1 1 0 1 0 1 0 0 1 0 0 0 1 };
               hd = { 1 0 0 1 0 0 1 0 1 1 1 0 0 0 0 0 1 0 0 0 1 0 0 0
                      0 0 1 1 0 1 0 0 1 0 0 0 1 0 1 0 1 1 0 1 1 1 0 };
               he = { 0 0 0 1 0 0 1 0 1 0 1 1 1 0 1 1 0 1 0 0 1 1 1 1
                      1 0 1 1 1 0 1 1 1 1 1 0 0 0 1 0 1 1 0 1 1 0 1 };
               hf = { 0 0 0 1 0 0 1 0 1 0 1 1 1 0 1 1 0 1 0 0 1 1 0 0
                      0 1 0 0 0 1 0 0 0 0 0 1 1 1 0 1 0 0 1 0 0 1 1 };
               hg = { 0 1 0 0 0 0 1 0 0 1 0 0 0 0 1 1 0 0 0 1 1 1 0 1
                      1 1 0 1 0 1 1 1 1 0 0 0 0 0 0 1 1 0 1 1 1 0 1 };
               end;
            else if n = 236 then do;
               ha = { 1 1 1 1 1 0 1 1 1 1 0 1 0 0 1 1 0 0 1 1
                      1 1 0 1 0 0 0 1 0 1 1 1 1 0 1 0 0 0 1 0
                      1 1 0 0 0 1 0 0 1 1 1 1 0 1 1 1 1 0 0 };
               hb = { 0 1 1 0 1 0 0 0 1 0 1 1 0 0 0 1 0 0 1 1
                      1 1 0 1 1 1 1 0 1 1 0 1 0 0 0 1 0 1 1 1
                      1 0 0 1 1 0 0 1 0 1 1 1 1 0 1 1 1 1 0 };
               hc = { 0 0 1 1 1 1 0 1 1 1 1 0 0 1 0 0 0 1 1 0
                      1 0 0 0 1 0 1 1 1 0 1 0 1 1 1 0 1 0 0 0
                      0 1 1 0 0 1 1 0 1 0 0 0 0 1 0 0 0 0 0 };
               hd = { 1 0 0 1 0 1 1 1 0 1 0 0 1 1 1 0 1 1 0 0
                      0 0 1 0 0 0 0 1 0 1 0 1 0 0 0 1 0 1 1 1
                      1 0 0 1 1 0 0 1 0 1 1 1 1 0 1 1 1 1 0 };
               he = { 0 0 0 0 1 0 0 0 0 1 0 1 1 0 0 1 1 0 0 0
                      0 1 0 1 1 1 0 1 0 1 0 1 1 1 1 0 1 1 1 1
                      0 0 1 0 0 0 1 1 0 1 0 0 0 1 0 1 1 0 1 };
               hf = { 0 0 0 0 1 0 0 0 0 1 0 1 1 0 0 1 1 0 0 0
                      0 1 0 1 1 1 0 1 0 1 1 1 0 1 0 0 0 1 0 1
                      1 0 0 0 1 0 0 1 1 1 1 0 1 1 1 1 0 0 0 };
               hg = { 0 0 0 0 1 0 0 0 0 1 0 1 1 0 0 1 1 0 0 0
                      0 1 0 1 1 1 0 1 0 0 1 0 0 0 0 1 0 0 0 0
                      1 1 0 1 1 1 0 0 1 0 1 1 1 0 1 0 0 1 1 };
               end;
            else if n = 260 then do;
               ha = { 1 1 1 1 0 1 0 1 1 0 0 1 0 0 1 1 1 0 0 0 1 1
                      0 1 1 1 1 0 1 0 1 1 1 0 0 0 1 0 1 0 0 0 0 1
                      0 0 1 1 1 1 1 1 0 0 1 0 0 1 1 0 1 0 1 1 1 };
               hb = { 0 1 1 1 0 1 0 1 1 0 0 1 0 0 1 1 1 1 1 1 0 0
                      1 0 0 0 0 1 0 1 0 0 0 1 1 1 0 1 0 1 1 1 1 0
                      1 1 0 0 0 1 1 1 0 0 1 0 0 1 1 0 1 0 1 1 1 };
               hc = { 0 0 0 0 1 0 1 0 0 1 1 0 1 1 0 0 0 1 1 1 0 0
                      1 0 0 0 0 1 0 1 0 0 0 0 0 0 1 0 1 0 0 0 0 1
                      0 0 1 1 1 1 1 1 0 0 1 0 0 1 1 0 1 0 1 1 1 };
               hd = { 0 1 1 1 0 1 0 1 1 0 0 1 0 0 1 1 1 0 0 0 1 1
                      0 1 1 1 1 0 1 0 1 1 1 1 1 1 0 1 0 1 1 1 1 0
                      1 1 0 0 0 0 0 0 1 1 0 1 1 0 0 1 0 1 0 0 0 };
               he = { 1 1 1 1 0 1 0 1 1 0 0 1 0 0 1 1 1 1 1 1 0 0
                      1 0 0 0 0 1 0 1 0 0 0 0 0 0 1 0 1 0 0 0 0 1
                      0 0 1 1 1 0 0 0 1 1 0 1 1 0 0 1 0 1 0 0 0 };
               hf = { 0 1 1 1 0 1 0 1 1 0 0 1 0 0 1 1 1 1 1 1 0 0
                      1 0 0 0 0 1 0 1 0 0 0 0 0 0 1 0 1 0 0 0 0 1
                      0 0 1 1 1 0 0 0 1 1 0 1 1 0 0 1 0 1 0 0 0 };
               hg = { 1 0 0 0 1 0 1 0 0 1 1 0 1 1 0 0 0 1 1 1 0 0
                      1 0 0 0 0 1 0 1 0 0 0 1 1 1 0 1 0 1 1 1 1 0
                      1 1 0 0 0 0 0 0 1 1 0 1 1 0 0 1 0 1 0 0 0 };
               end;
            else if n = 268 then do;
               ha = { 1 1 0 0 0 0 1 0 0 1 1 0 0 0 0 1 1 0 1 1 0 1
                      1 1 0 0 0 0 0 1 1 0 0 1 0 1 0 1 1 0 1 1 0 1
                      0 0 0 1 0 0 1 0 0 1 0 0 1 0 0 0 1 0 1 1 1 1 1 };
               hb = { 0 1 1 0 0 1 1 1 0 0 1 0 1 1 0 1 0 1 0 0 0 0
                      0 1 1 1 1 0 0 1 0 0 1 0 1 1 0 0 1 0 1 0 1 1
                      1 0 0 1 1 1 1 1 1 1 1 0 1 0 1 0 1 0 1 1 1 1 0 };
               hc = { 0 0 0 1 1 0 1 0 0 0 1 1 0 0 1 0 1 1 0 1 1 1
                      1 1 1 0 0 1 0 1 0 1 0 1 0 1 0 1 1 0 1 0 1 1
                      0 1 0 1 1 1 0 0 1 1 1 1 1 1 1 1 1 0 1 0 0 0 0 };
               hd = { 0 1 1 1 1 0 0 0 1 1 0 0 0 1 0 0 1 1 1 1 0 0
                      0 0 1 1 1 0 0 0 1 1 0 1 0 1 0 1 1 1 1 1 1 0
                      1 1 1 0 0 1 1 1 1 1 1 0 0 1 1 0 0 1 1 0 0 1 0 };
               he = { 0 1 1 0 0 1 1 0 0 1 1 0 0 0 0 0 0 1 1 0 0 0
                      1 0 0 0 0 0 0 1 0 1 0 1 0 0 1 1 1 0 0 0 1 1
                      1 1 0 0 0 0 1 1 0 1 1 1 0 0 1 1 1 0 0 0 0 1 1 };
               hf = { 0 0 0 1 0 1 1 1 1 1 1 1 1 1 0 0 1 1 1 0 1 0
                      1 1 0 1 0 1 1 0 1 0 1 0 1 0 1 0 1 0 0 1 1 1
                      1 1 1 0 1 1 0 1 0 0 1 1 0 0 0 1 0 1 1 0 0 0 0 };
               hg = { 0 0 0 0 1 0 1 0 1 0 1 0 0 0 0 0 0 0 0 1 1 0
                      0 0 1 0 1 0 1 1 0 0 1 0 1 1 0 1 1 0 0 0 0 1
                      1 1 1 1 0 1 0 1 0 0 1 0 1 1 0 0 0 1 1 0 0 1 1 };
               end;
            else if n = 292 then do;
               ha = { 0 1 1 1 1 0 1 0 1 1 0 0 1 0 0 0 1 0 1 1 0 0 0 1
                      1 1 0 1 0 0 0 0 1 0 0 1 1 1 1 0 0 1 0 0 0 0 1 0
                      1 1 1 0 0 0 1 1 0 1 0 0 0 1 0 0 1 1 0 1 0 1 1 1 1 };
               hb = { 0 0 0 0 0 1 0 1 0 0 1 1 0 1 1 1 0 1 0 0 1 1 1 0
                      0 0 1 0 1 1 1 1 0 1 1 0 0 0 0 1 1 0 1 1 1 1 0 1
                      0 0 0 1 1 1 0 0 1 0 1 1 1 0 1 1 0 0 1 0 1 0 0 0 0 };
               hc = { 1 1 1 0 1 0 0 0 1 1 0 1 0 0 0 1 1 1 1 0 0 1 1 0
                      0 0 0 0 0 0 1 0 1 1 1 0 1 1 0 0 0 1 1 0 1 1 0 1
                      0 0 0 0 0 1 0 1 0 1 0 1 1 0 0 1 1 1 1 0 1 1 0 1 1 };
               hd = { 1 1 1 0 1 0 0 0 1 1 0 1 0 0 0 1 1 0 1 0 0 1 1 0
                      0 0 0 0 0 0 1 0 1 0 0 0 1 1 0 0 0 1 1 0 1 0 0 1
                      0 0 0 0 0 0 0 1 0 1 0 0 1 0 0 0 1 1 0 0 0 1 0 1 1 };
               he = { 0 0 0 1 0 1 1 1 0 0 1 1 1 0 1 1 0 1 0 1 1 1 1 1
                      1 1 0 1 1 0 1 0 0 1 1 1 0 0 1 1 1 0 1 0 1 1 1 1
                      1 1 1 0 0 1 1 0 1 0 0 1 1 1 0 1 0 0 1 1 1 0 1 0 0 };
               hf = { 1 1 1 0 1 1 0 1 1 1 1 0 0 1 1 0 1 0 1 0 1 0 0 0
                      0 0 1 0 1 1 0 1 1 0 0 0 1 1 0 1 1 1 0 1 0 0 0 0
                      0 0 0 1 1 0 0 1 1 1 1 0 0 0 1 0 1 1 0 0 0 1 0 1 1 };
               hg = { 1 1 1 1 1 0 1 0 1 1 0 0 1 0 0 0 1 0 1 1 0 0 0 1
                      1 1 0 1 0 0 0 0 1 0 0 1 1 1 1 0 0 1 0 0 0 0 1 0
                      1 1 1 0 0 0 1 1 0 1 0 0 0 1 0 0 1 1 0 1 0 1 1 1 1 };
               end;
            else if n = 356 then do;
               ha = { 0 1 1 1 0 1 1 0 1 0 0 1 0 0 1 1 0 1 0 1 0 1
                      0 1 1 1 1 1 1 1 0 0 1 1 1 0 0 0 0 1 1 1 0 1
                      1 0 0 0 1 0 0 1 0 1 1 0 1 1 0 0 1 0 1 0 1 0
                      1 0 0 0 0 0 0 0 1 1 0 0 0 1 1 1 1 0 0 0 1 0 };
               hb = { 0 0 1 0 1 1 0 0 0 1 1 0 0 1 0 0 0 0 0 1 0 1
                      1 1 1 1 0 1 1 0 0 1 1 1 0 0 1 0 1 1 0 1 0 1
                      0 0 1 0 1 1 0 0 0 1 1 0 0 1 0 0 0 0 0 1 0 1
                      1 1 1 1 0 1 1 0 0 1 1 1 0 0 1 0 1 1 0 1 0 1 };
               hc = { 1 0 0 0 1 1 1 1 0 0 0 1 1 0 0 0 0 0 0 0 0 0
                      1 0 1 0 1 0 0 1 1 0 1 1 0 1 0 0 1 0 0 0 1 1
                      0 1 1 1 0 0 0 0 1 1 1 0 0 1 1 1 1 1 1 1 1 1
                      0 1 0 1 0 1 1 0 0 1 0 0 1 0 1 1 0 1 1 1 0 0 };
               hd = { 0 0 1 1 0 1 0 0 0 1 1 1 1 0 1 1 1 0 1 0 0 0
                      1 0 1 1 1 0 1 1 1 1 0 0 0 1 0 1 1 0 0 0 0 0
                      0 0 1 1 0 1 0 0 0 1 1 1 1 0 1 1 1 0 1 0 0 0
                      1 0 1 1 1 0 1 1 1 1 0 0 0 1 0 1 1 0 0 0 1 0 };
               he = 1 - hd;
               hf = { 1 1 0 1 1 0 1 0 0 1 0 0 1 1 0 1 0 1 0 1 1 1
                      1 1 1 1 1 1 0 0 1 1 1 0 0 0 0 1 1 1 0 1 1 0
                      0 0 1 0 0 1 0 1 1 0 1 1 0 0 1 0 1 0 1 0 0 0
                      0 0 0 0 0 0 1 1 0 0 0 1 1 1 1 0 0 0 1 0 0 1 };
               hg = { 1 0 0 1 0 1 1 0 0 0 1 1 0 0 1 0 0 0 0 0 1 0
                      1 1 1 1 1 0 1 1 0 0 1 1 1 0 0 1 0 1 1 0 1 0
                      1 0 0 1 0 1 1 0 0 0 1 1 0 0 1 0 0 0 0 0 1 0
                      1 1 1 1 1 0 1 1 0 0 1 1 1 0 0 1 0 1 1 0 1 0 };
               adj = 4;
               n = n - adj;
               end;
            else if n = 404 then do;
               ha = { 1 1 1 0 0 1 1 1 0 0 1 1 1 1 1 1 1 1 0 1 0 0 1 0 1
                      0 0 0 0 0 1 1 1 1 0 1 0 1 0 1 0 0 1 1 0 1 1 0 0 1
                      0 0 0 1 1 0 0 0 1 1 0 0 0 0 0 0 0 0 1 0 1 1 0 1 0
                      1 1 1 1 1 0 0 0 0 1 0 1 0 1 0 1 1 0 0 1 0 0 1 1 0 };
                hb = { 0 1 0 1 1 0 1 0 0 0 0 1 0 0 1 1 0 0 0 0 1 1 0 1 0
                      0 1 1 1 1 0 0 1 1 0 1 1 1 1 0 1 0 0 1 0 1 0 1 0 1
                      0 1 0 1 1 0 1 0 0 0 0 1 0 0 1 1 0 0 0 0 1 1 0 1 0
                      0 1 1 1 1 0 0 1 1 0 1 1 1 1 0 1 0 0 1 0 1 0 1 0 1 };
               hc = { 1 1 0 0 1 0 0 1 1 0 1 0 1 0 1 0 0 0 0 1 1 1 1 0 0
                      1 0 1 1 0 1 0 0 0 0 0 0 0 0 1 1 0 0 0 1 1 0 0 0 1
                      0 0 1 1 0 1 1 0 0 1 0 1 0 1 0 1 1 1 1 0 0 0 0 1 1
                      0 1 0 0 1 0 1 1 1 1 1 1 1 1 0 0 1 1 1 0 0 1 1 1 0 };
               hd = { 1 1 1 0 1 0 0 1 1 0 0 0 0 0 1 1 0 1 1 1 0 0 0 1 0
                      0 0 1 1 1 0 1 1 0 0 0 0 0 1 1 0 0 1 0 1 1 1 1 0 1
                      1 1 1 0 1 0 0 1 1 0 0 0 0 0 1 1 0 1 1 1 0 0 0 1 0
                      0 0 1 1 1 0 1 1 0 0 0 0 0 1 1 0 0 1 0 1 1 1 1 1 1 };
               he = 1 - hd;
               hf = { 1 0 0 1 1 1 0 0 1 1 1 1 1 1 1 1 0 1 0 0 1 0 1 1 0
                      0 0 0 1 1 1 1 0 1 0 1 0 1 0 0 1 1 0 1 1 0 0 1 0 0
                      0 1 1 0 0 0 1 1 0 0 0 0 0 0 0 0 1 0 1 1 0 1 0 0 1
                      1 1 1 0 0 0 0 1 0 1 0 1 0 1 1 0 0 1 0 0 1 1 0 1 1 };
               hg = { 1 0 1 0 1 1 0 1 0 0 0 0 1 0 0 1 1 0 0 0 0 1 1 0 1
                      0 0 1 1 1 1 0 0 1 1 0 1 1 1 1 0 1 0 0 1 0 1 0 1 0
                      1 0 1 0 1 1 0 1 0 0 0 0 1 0 0 1 1 0 0 0 0 1 1 0 1
                      0 0 1 1 1 1 0 0 1 1 0 1 1 1 1 0 1 0 0 1 0 1 0 1 0 };
               adj = 4;
               n = n - adj;
               end;
            else if n = 436 then do;
               ha = { 0 1 0 1 0 1 1 0 0 0 1 0 0 0 1 0 1 0 0 1 0 0 1 0 1 1 0
                      0 1 1 0 0 0 0 1 1 1 0 0 0 0 0 1 0 0 0 1 0 0 1 1 1 1 1
                      1 0 1 0 1 0 0 1 1 1 0 1 1 1 0 1 0 1 1 0 1 1 0 1 0 0 1
                      1 0 0 1 1 1 1 0 0 0 1 1 1 1 1 0 1 1 1 0 1 1 0 0 0 0 0 };
               hb = { 0 1 0 0 1 1 0 1 0 0 0 1 0 1 0 0 1 1 1 0 1 1 1 1 1 0 0
                      0 0 0 1 0 0 0 1 1 0 1 0 1 1 1 0 1 0 0 1 1 0 1 1 1 0 0
                      0 1 0 0 1 1 0 1 0 0 0 1 0 1 0 0 1 1 1 0 1 1 1 1 1 0 0
                      0 0 0 1 0 0 0 1 1 0 1 0 1 1 1 0 1 0 0 1 1 0 1 1 1 0 0 };
               hc = { 0 0 0 0 1 1 0 1 1 1 0 1 1 1 1 1 0 0 0 1 1 1 1 0 0 0 1
                      0 0 1 0 1 1 0 1 1 0 1 0 1 1 1 0 1 1 1 0 0 1 0 1 0 1 1
                      1 1 1 1 0 0 1 0 0 0 1 0 0 0 0 0 1 1 1 0 0 0 0 1 1 1 0
                      1 1 0 1 0 0 1 0 0 1 0 1 0 0 0 1 0 0 0 1 1 0 1 0 1 0 0 };
               hd = { 1 0 0 0 0 1 0 0 1 1 1 1 0 1 0 0 1 0 1 1 1 0 0 1 1 1 1
                      1 0 0 1 1 1 0 1 0 0 1 0 1 1 1 1 0 0 1 0 0 0 0 1 0 0 0
                      1 0 0 0 0 1 0 0 1 1 1 1 0 1 0 0 1 0 1 1 1 0 0 1 1 1 1
                      1 0 0 1 1 1 0 1 0 0 1 0 1 1 1 1 0 0 1 0 0 0 0 1 0 1 0 };
               he = 1 - hd;
               hf = { 0 1 0 1 1 0 0 0 1 0 0 0 1 0 1 0 0 1 0 0 1 0 1 1 0 1 1
                      1 0 0 0 0 1 1 1 0 0 0 0 0 1 0 0 0 1 0 0 1 1 1 1 1 1 0
                      1 0 1 0 0 1 1 1 0 1 1 1 0 1 0 1 1 0 1 1 0 1 0 0 1 0 0
                      0 1 1 1 1 0 0 0 1 1 1 1 1 0 1 1 1 0 1 1 0 0 0 0 0 0 1 };
               hg = { 0 0 1 0 0 1 1 0 1 0 0 0 1 0 1 0 0 1 1 1 0 1 1 1 1 1 0
                      0 0 0 0 1 0 0 0 1 1 0 1 0 1 1 1 0 1 0 0 1 1 0 1 1 1 0
                      0 0 1 0 0 1 1 0 1 0 0 0 1 0 1 0 0 1 1 1 0 1 1 1 1 1 0
                      0 0 0 0 1 0 0 0 1 1 0 1 0 1 1 1 0 1 0 0 1 1 0 1 1 1 0 };
               adj = 4;
               n = n - adj;
               end;
            else if n = 596 then do;
               ha = expand({ 0 8 2 1 4 1 2 3 2 1 1 2 1 2 1 3 1 1 2 4 1 3 3 4 2
               1 2 3 1 1 3 2 1 1 1 1 1 1 8 2 1 4 1 2 3 2 1 1 2 1 2 1 3 1 1 2 4
               1 3 3 4 2 1 2 3 1 1 3 2 1 1 1 1 1 1});
               hb = expand({ 5 1 1 1 1 2 1 4 2 5 1 3 2 1 1 3 1 1 3 1 1 2 3 1 5
               2 4 1 2 1 1 1 1 7 7 1 1 1 1 2 1 4 2 5 1 3 2 1 1 3 1 1 3 1 1 2 3
               1 5 2 4 1 2 1 1 1 1 7 2});
               hc = expand({ 0 1 1 1 1 1 2 3 1 1 3 2 1 2 4 3 3 1 5 1 1 1 3 1 2
               1 2 1 1 2 3 2 1 4 1 2 8 1 1 1 1 1 1 2 3 1 1 3 2 1 2 4 3 3 1 5 1
               1 1 3 1 2 1 2 1 1 2 3 2 1 4 1 2 8 1});
               hd = expand({ 0 1 1 1 1 1 2 2 2 1 2 1 1 3 3 2 3 1 1 4 2 1 2 4 1
               1 3 2 3 3 1 1 2 1 2 2 2 1 1 1 1 2 1 2 1 1 1 1 2 2 2 1 2 1 1 3 3
               2 3 1 1 4 2 1 2 4 1 1 3 2 3 3 1 1 2 1 2 2 2 1 1 1 1 4});
               he = 1 - hd;
               hf = expand({ 0 6 2 1 4 1 2 3 2 1 1 2 1 2 1 3 1 1 1 5 1 3 3 4 2
               1 2 3 1 1 3 2 1 1 1 1 1 1 8 2 1 4 1 2 3 2 1 1 2 1 2 1 3 1 1 1 5
               1 3 3 4 2 1 2 3 1 1 3 2 1 1 1 1 1 1 2});
               hg = expand({ 6 1 1 1 1 2 1 4 2 5 1 3 2 1 1 3 1 1 3 1 1 2 3 1 5
               2 4 1 2 1 1 1 1 7 7 1 1 1 1 2 1 4 2 5 1 3 2 1 1 3 1 1 3 1 1 2 3
               1 5 2 4 1 2 1 1 1 1 7 1});
               adj = 4;
               n = n - adj;
               end;

            mm = n / 4;
            ha = shape(2 # ha - 1, mm, mm);   hb = shape(2 # hb - 1, mm, mm);
            hc = shape(2 # hc - 1, mm, mm);   hd = shape(2 # hd - 1, mm, mm);
            he = shape(2 # he - 1, mm, mm);   hf = shape(2 # hf - 1, mm, mm);
            hg = shape(2 # hg - 1, mm, mm);
            h = (2:mm) || 1;
            do i = 2 to mm;
               ha[i,] = ha[i - 1, mm] || ha[i - 1, 1 : (mm - 1)];
               hb[i,] = hb[i - 1, h]; hc[i,] = hc[i - 1, h];
               hd[i,] = hd[i - 1, h]; he[i,] = he[i - 1, h];
               hf[i,] = hf[i - 1, h]; hg[i,] = hg[i - 1, h];
               end;

            h  = (  ha ||  hb ||  hc ||  hd ) //
                 ( -hb ||  ha ||  he ||  hf ) //
                 ( -hc || -he ||  ha ||  hg ) //
                 ( -hd || -hf || -hg ||  ha );

            if adj then do;
               ha = {1 -1 1 1, 1 1 -1 1, -1 1 1 1, -1 -1 -1 1};
               hb = j(1, mm, 1);
               hc =  hb ||  hb ||  hb ||  hb;
               hd = -hb ||  hb || -hb ||  hb;
               he = -hb ||  hb ||  hb || -hb;
               hf = -hb || -hb ||  hb ||  hb;
               hg = -hb || -hb || -hb ||  hb;
               hh =  hb ||  hb || -hb ||  hb;
               hi =  hb || -hb || -hb || -hb;
               hj = -hb ||  hb || -hb || -hb;
               h = (ha || (hc // hd // he // hf)) //
                   (hg` || hi` || hh` || hj` || h);
               end;

            h = h[,1] # h; h = h # h[1,];
            free ha hb hc hd he hf hg hh hi hj;
            return(h);
            finish;

         *------Suen 1989 construction------;
         start suen(s) global(h2);
            v = (h2 @ h2)[,{3 2 4}]; v = twolev(v || -v);
            r = nrow(s); c = ncol(s);
            x = j(r, 1, 1) @ v[,1] || j(4 # r, c, 0);
            do i = 1 to r;
               l = ((i - 1) # 4 + 1) : (i # 4);
               do j = 1 to c; x[l, j + 1] = v[,s[i,j]]; end;
               end;
            return(x);
            finish;

         *------return index vector that sorts vector a into order of b------;
         start sortlike(a, b);
            t = a;
            n = nrow(a);
            v = (1:n)`;
            do i = 1 to n;
               do j = i to n;
                  if b[i,] = t[j,] then do;
                     if i ^= j then do;
                        temp = t[i,]; t[i,] = t[j,]; t[j,] = temp;
                        temp = v[i,]; v[i,] = v[j,]; v[j,] = temp;
                        end;
                     goto leave;
                     end;
                  end;
               leave:;
               end;
            return(v);
            finish;

         *------assembles pieces of GH(4 * f, f)------;
         start makehad4(f, alpha, beta, gamma);
            run fieldop(f, add, sub, mul, inv);
            h = j(4 # f, 4 # f, 0);
            do i = 1 to 4; do k = 1 to 4;
               do x = 1 to f; x2 = mul[x,x] + 1; do z = 1 to f;
                  xz = mul[x,z] + 1; z2 = mul[z,z] + 1;
                  t1 = mul[alpha[i,k], z2];
                  t2 = mul[ beta[i,k], xz];
                  t3 = mul[gamma[i,k], x2];
                  h[(i-1)#f+x, (k-1)#f+z] = add[add[t1+1,t2+1]+1,t3+1];
                  end; end; end; end;
            return( nrmz(h, f) );
            finish;

         *------GH(4 * f, f), prime f >= 7, Dawson (1985)------;
         start gh4(f, coef);
            run fieldop(f, add, sub, mul, inv);
            a = coef[1]; b = coef[2]; c = coef[3]; d = coef[4]; lb = 1;
            beta  = (a || d || b || c) // (b || c || a || d) //
                    (c || b || d || a) // (d || a || c || b);
            ad    = mul[a+1,d+1]; ac = mul[a+1,c+1];
            bc    = mul[b+1,c+1]; bd = mul[b+1,d+1];
            a2    = mul[a+1,a+1]; b2 = mul[b+1,b+1];
            den1  = mul[d+1,add[a2+1,b2+1]+1]; den2 = mul[c+1,add[a2+1,b2+1]+1];
            lc    = mul[lb+1,mul[mul[a+1,sub[ad+1,bc+1]+1]+1, inv[den1]+1]+1];
            ld    = mul[lb+1,mul[mul[a+1,add[ac+1,bd+1]+1]+1, inv[den2]+1]+1];
            bp    = mul[add[a2+1,b2+1]+1,inv[mul[4+1,lb+1]]+1];
            cp    = mul[bp+1,mul[mul[a+1,add[ad+1,bc+1]+1]+1,
                                 inv[mul[d+1,add[a2+1,b2+1]+1]]+1]+1];
            dp    = mul[bp+1,mul[mul[a+1,sub[ac+1,bd+1]+1]+1,
                                 inv[mul[c+1,add[a2+1,b2+1]+1]]+1]+1];
            alpha = (0  || ld || lb || lc) // (lb || lc || 0  || ld) //
                    (lc || lb || ld || 0 ) // (ld || 0  || lc || lb);
            gamma = (0  || dp || bp || cp) // (bp || cp || 0  || dp) //
                    (cp || bp || dp || 0 ) // (dp || 0  || cp || bp);
            return( makehad4(f, alpha + 1, beta + 1, gamma + 1) );
            finish;

         *------rules for add, subtract, multiply, and
                invert in field/group/ring of  order
                f < 24 | f = 27 (f = 24 add only)     ------;
         start fieldop(f, add, sub, mul, inv);
            sub = j(f, f, .); add = j(f, f, .); mul = j(f, f, 0);
            if f = 24 then do;
               do i = 0 to f - 1; do j = 0 to i;
                  a = int(j / 4);   b = int(j / 2);   c = mod(j, 2);
                  x = int(i / 4);   y = int(i / 2);   z = mod(i, 2);
                  add[i + 1, j + 1] = 4 * mod(a + x, 6) +
                                      2 * mod(b + y, 2) + mod(c + z, 2);
                  add[j + 1, i + 1] = add[i + 1, j + 1];
                  end; end;
               end;
            else if any(f = {4 8 16 27}) then do;
               p = 1 + (f = 8 | f = 27) + 2 # (f = 16);
               c2 = 2 + mod(f, 2);
               do i = 0 to f - 1; do j = 0 to f - 1;
                  t = mod(mod(i , c2) + mod(j , c2), c2);
                  do k = 1 to p;
                     c = c2 ## k;
                     t = t + c # mod(int(i / c) + int(j / c), c2);
                     end;
                  add[i + 1, j + 1] = t;
                  sub[t + 1, i + 1] = j;
                  end; end;
               end;
            else if mod(f, 4) = 0 | mod(f, 9) = 0 then do;
               c2 = 2 + mod(f, 2); c = f / c2;
               do i = 0 to f - 1; do j = 0 to f - 1;
                  t = c # mod(int(i / c) + int(j / c), c2) +
                          mod(mod(i , c) + mod(j , c), c);
                  add[i + 1, j + 1] = t;
                  sub[t + 1, i + 1] = j;
                  end; end;
               end;
            else do;
               do i = 0 to f - 1;
                  j = f - i + 1;
                  add[i + 1,] = mod((0:(f-1)) + i, f);
                  sub[,i + 1] = ((0:(f-1)) || (0:(f-1)))[,j:(j+f-1)]`;
                  end;
               end;
            if f = 4 then mul[2:4,2:4] = sqrsym({ 1 2 3 3 1 2 });
            else if f = 8 then
               mul[2:8, 2:8] = sqrsym({ 1 2 4 3 6 5 4 5 1 7 5 7 2 3
                                        6 6 1 7 2 4 3 7 3 4 6 1 5 2 });
            else if f = 9 then do;
               ff0 = {3 7 8 2 6 5 4 1}; ff1 = 1 || ff0;
               do i = 1 to 8; do j = 1 to 8;
                  mul[i+1, j+1] = ff1[mod(loc(ff0 = i) + loc(ff0 = j), 8) + 1];
                  end; end;
               end;
            else if f = 16 then do;
               mul[,2] = (0:15)`; mul[2,] = (0:15);
               mul[3:16, 3:16] = sqrsym({ 4 6 5 8 12 9 10 15 13 8 12 10 1 7 13
               14 9 5 2 11 12 9 1 11 3 2 10 15 11 2 15 6 4 13 7 14 13 7 3 9 14
               4 6 12 11 15 4 7 12 8 3 14 5 1 10 1 13 2 14 3 15 4 8 5 9 6 3 14
               6 11 5 8 12 1 15 2 10 7 5 11 10 4 15 1 13 3 8 6 7 9 2 7 8 14 1
               9 6 5 10 2 13 11 4 12 3});
               end;
            else if f = 27 then do;
               mul[,2] = (0:26)`; mul[2,] = (0:26);
               mul[3:27,3:27] = sqrsym({ 1 6 9 8 12 16 7 15 11 13 3 18 24 21 9
               5 21 19 26 15 13 4 24 23 19 12 11 16 18 11 20 2 19 1 10 17 20 14
               21 4 25 8 15 26 6 19 17 25 6 22 3 14 8 16 24 24 20 5 17 10 22 7
               25 1 13 15 26 23 6 10 16 20 3 7 11 21 18 4 25 26 1 12 13 24 2 16
               18 5 3 17 19 21 2 17 23 1 16 22 6 12 18 8 14 20 7 23 5 9 25 7 14
               18 15 22 2 11 24 4 13 20 22 8 13 18 4 9 26 24 5 10 23 1 15 19 6
               14 9 19 10 1 11 2 20 22 13 4 14 5 23 3 21 12 17 11 22 14 3 17 6
               25 4 23 12 26 15 7 9 1 20 8 24 10 25 15 8 14 4 21 13 3 23 2 19 9
               24 17 7 26 16 6 15 1 22 16 2 23 17 3 24 9 4 25 10 5 26 11 6 18 12
               7 17 4 26 9 8 18 13 12 7 20 16 2 21 11 3 25 24 10 5 19 14 16 7 18
               14 5 25 9 21 17 1 19 12 8 26 10 3 15 2 22 13 6 20 12 10 7 22 20
               17 5 14 2 26 21 9 6 4 19 16 25 13 1 8 23 11 15 14 13 2 24 26 12 1
               23 9 7 6 22 11 10 8 21 16 5 18 20 15 4 3 19 13 16 3 20 23 10 6 5
               19 15 9 8 22 25 12 2 7 21 11 14 1 24 18 17 4 });
               end;
            else mul = mod((0:(f-1))` * (0:(f-1)), f);
            inv      = (mul = 1)[<:>,2:f] - 1;
            finish;

         *------some factors that we cannot more elegantly construct------;
         start fac(n, m, l) global(h2);
            if n = 196 then do;
               if m = 3 & l = 14 then do;
                  fac = { 1 1 1 2 2 2 3 3 3 4 4 4 5 5 5 6 6 6 7 7 7 8 8 8 9 9 9
                  10 10 10 11 11 11 12 12 12 13 13 13 14 14 14 2 4 14 1 10 3 6 7
                  11 10 1 13 4 3 8 13 8 10 8 12 6 12 2 5 14 6 2 5 11 12 7 14 9 9
                  5 4 11 9 7 3 13 1 3 5 2 4 14 1 1 11 4 7 8 12 11 1 14 5 4 9 14
                  9 11 9 13 7 13 3 6 2 7 3 6 12 13 8 2 10 10 6 5 12 10 8 4 6 3
                  13 11 9 5 2 1 1 12 5 8 9 13 12 1 2 6 5 10 2 10 12 10 14 8 14 4
                  7 3 8 4 7 13 14 9 3 11 11 7 6 5 7 4 12 8 7 14 12 10 6 3 1 1 13
                  6 9 10 14 13 1 3 7 6 11 3 11 13 11 2 9 2 5 8 4 9 5 8 14 2 10 4
                  12 6 8 5 11 5 13 13 9 8 2 13 11 7 4 1 1 14 7 10 11 2 14 1 4 8
                  7 12 4 12 14 12 3 10 3 6 9 5 10 6 9 2 3 7 9 6 10 3 4 12 6 14
                  14 10 9 3 14 12 8 5 1 1 2 8 11 12 3 2 1 5 9 8 13 5 13 2 13 4
                  11 4 7 10 6 11 7 8 10 7 7 12 8 11 4 5 13 7 2 2 11 10 4 2 13 9
                  6 1 1 3 9 12 13 4 3 1 6 10 9 14 6 14 3 14 5 12 5 8 11 9 11 8 6
                  9 12 8 13 9 12 5 6 14 8 3 3 12 11 5 3 14 10 7 1 1 4 10 13 14 5
                  4 1 7 11 10 2 7 2 4 2 6 13 10 12 9 3 7 14 7 10 13 9 14 10 13 6
                  7 2 9 4 4 13 12 6 4 2 11 8 1 1 5 11 14 2 6 5 1 8 12 11 3 8 3 5
                  11 13 10 9 4 6 4 8 2 8 11 14 10 2 11 14 7 8 3 10 5 5 14 13 7 5
                  3 12 9 1 1 6 12 2 3 7 6 1 9 13 12 4 12 14 11 14 13 5 10 5 7 5
                  9 3 9 12 2 11 3 12 2 8 9 4 11 6 6 2 14 8 6 4 13 10 1 1 7 13 3
                  4 8 7 1 10 13 2 12 8 1 11 2 14 6 11 6 8 6 10 4 10 13 3 12 4 13
                  3 9 10 5 12 7 7 3 2 9 7 5 14 11 1 1 8 14 4 5 9 14 3 13 5 6 10
                  9 1 12 3 2 7 12 7 9 7 11 5 11 14 4 13 5 14 4 10 11 6 13 8 8 4
                  3 10 8 6 2 12 1 1 9 2 };
                  end;
               end;

            else if n = 144 then do;
               if m = 10 & l = 2 then
                  fac = expand({2 3 1 5 1 3 2 1 1 2 1 6 3 2 2 1 1 3 1 1 1 2 1 5
                  1 2 2 1 1 2 2 1 2 1 1 1 1 4 2 5 1 1 5 1 1 1 1 1 5 1 2 2 1 2 1
                  6 3 3 3 1 1 2 1 1 3 4 1 2 2 1 1 1 1 4 2 1 2 1 1 1 1 1 5 1 1 3
                  2 4 1 2 2 1 1 1 3 1 8 1 1 1 1 1 1 1 2 2 2 1 2 2 2 1 1 1 3 1 4
                  2 2 2 1 3 2 1 1 1 2 1 3 2 2 3 3 1 2 1 2 3 3 2 1 1 2 1 1 2 3 4
                  3 5 1 3 3 1 1 3 1 2 1 3 3 1 1 1 1 4 1 5 3 1 1 1 1 2 1 1 3 2 3
                  1 2 3 3 1 3 1 1 1 2 1 2 1 2 1 2 1 1 1 1 3 3 1 4 2 2 1 2 1 1 2
                  5 1  10 2 2 5 5 2 1 1 3 1 2 1 4 6 1 1 1 5 2 2 1 4 2 1 1 4 1 1
                  1 3 2 1 2 2 4 5 3 2 2 1 2 5 1 1 2 1 1 2 2 1 4 1 4 3 1 2 4 1 2
                  3 3 1 1 1 1 1 5 1 1 2 2 1 1 1 1 1 1 1 1 6 7 1 3 3 1 4 1 2 1 1
                  1 1 1 1 2 1 4 2 1 1 5 3 2 1 1 3 1 2 1 3 1 4 3 2 3 2 1 3 2 6 1
                  5 2 1 1 5 3 1 2 4 4 2 1 2 1 2 1 1 5 1 3 4 3 1 5 1 3 2 1 1 2 1
                  6 3 2 2 1 1 3 1 1 1 2 1 5 1 2 2 1 1 2 2 1 2 1 1 1 1 4 2 5 1 1
                  5 1 1 1 1 1 5 1 2 2 1 2 1 6 3 3 3 1 1 2 1 1 3 4 1 2 2 1 1 1 1
                  4 2 1 2 1 1 1 1 1 5 1 1 3 2 4 1 2 2 1 1 1 3 1 8 1 1 1 1 1 1 1
                  2 2 2 1 2 2 2 1 1 1 3 1 4 2 2 2 1 3 2 1 1 1 2 1 3 2 2 3 3 1 2
                  1 2 3 3 2 1 1 2 1 1 2 3 4 3 5 1 3 3 1 1 3 1 2 1 3 3 1 1 1 1 4
                  1 5 3 1 1 1 1 2 1 1 3 2 3 1 2 3 3 1 3 1 1 1 2 1 2 1 2 1 2 1 1
                  1 1 3 3 1 4 2 2 1 2 1 1 2 5 1  10 2 2 5 5 2 1 1 3 1 2 1 4 6 1
                  1 1 5 2 2 1 4 2 1 1 4 1 1 1 3 2 1 2 2 4 5 3 2 2 1 2 5 1 1 2 1
                  1 2 2 1 4 1 4 3 1 2 4 1 2 3 3 1 1 1 1 1 5 1 1 2 2 1 1 1 1 1 1
                  1 1 6 7 1 3 3 1 4 1 2 1 1 1 1 1 1 2 1 4 2 1 1 5 3 2 1 1 3 1 2
                  1 3 1 4 3 2 3 2 1 3 2 6 1 5 2 1 1 5 3 1 2 4 4 2 1 2 1 2 1 1 5
                  1 3 2}) + 1;
               end;

            else if n = 140 then do;
               if m = 10 & l = 2 then
                  fac = expand({5 1 2 1 1 5 1 2 3 1 1 1 1 2 1 3 1 1 1 1 1 1 5 1
                  1 2 1 2 2 2 4 2 1 1 3 2 1 2 4 1 3 1 1 1 1 1 1 2 1 1 1 1 1 1 1
                  1 2 3 2 1 1 2 1 1 1 4 1 4 2 1 2 3 2 2 2 3 3 4 1 2 5 4 2 5 1 2
                  2 1 1 4 1 3 3 4 2 1 1 1 1 1 1 2 1 2 1 1 2 2 7 1 1 2 3 6 2 1 1
                  1 2 1 2 2 2 1 1 3 1 3 6 2 4 1 1 3 1 5 2 2 2 1 5 3 3 2 2 1 3 1
                  1 1 3 1 1 1 1 1 3 2 6 1 1 2 6 1 3 1 1 1 1 2 1 1 2 1 4 3 1 3 1
                  2 1 1 3 1 2 1 1 2 3 5 1 2 1 2 2 3 5 2 1 1 2 9 3 6 4 1 1 1 3 3
                  2 1 3 2 2 1 1 2 1 3 2 1 1 6 1 1 3 1 1 2 2 2 3 6 2 1 1 3 3 1 2
                  1 1 3 1 1 1 1 2 6 2 3 1 2 2 1 3 3 4 1 1 3 1 1 2 1 3 1 4 2 2 2
                  1 5 4 2 2 1 4 1 5 1 1 7 2 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 3 4 2
                  1 2 1 2 7 1 1 4 2 1 3 1 1 2 1 6 1 1 1 1 1 1 2 1 2 1 1 1 2 4 2
                  1 3 1 2 1 2 1 2 2 1 1 1 1 1 1 1 2 2 4 1 3 2 4 1 1 4 5 1 3 1 1
                  1 2 1 2 1 1 2 2 1 1 3 5 8 3 4 1 1 2 3 2 2 3 3 1 1 2 1 1 2 3 3
                  1 1 3 2 1 1 1 1 1 2 2 2 3 2 1 3 2 1 2 2 1 2 1 1 1 1 1 1 1 1 3
                  2 1 2 1 6 1 2 4 1 1 5 1 2 2 2 4 2 2 1 2 1 3 4 2 1 1 1 1 3 1 1
                  1 1 1 1 1 1 6 1 3 1 1 4 1 2 1 1 2 3 2 1 1 1 2 3 1 2 1 1 1 1 2
                  3 1 1 4 2 4 1 1 3 7 1 6 1 1 2 1 2 1 3 1 5 2 2 2 1 1 1 2 1 4 3
                  3 5 1 2 4 2 1 2 3 6 5 2 2 1 1 1 2 3 1 2 1 1 1 3 2 1 1 2 2 1 1
                  2 4 1 2 1 1 1 3 1 3 1 2 3 4 2 1 1 1 1 1 4 1 1 1 1 4 1 1 4 2 1
                  3 2 2 1 3 1 4 1 1 3 1 1 5 1 1 1 1 1 1 3 4 2 1 3 2 1 4 7 1 2 1
                  3 1 4 2 1 1 4 1 4 1 2 1 1 2 2 1 2 1 2 1 3 1 2 2 1 1 4 1 1 1 1
                  2 3 3 1 1 4 2 3 1 2 1 3 1 1 1 2 2 1 1 1 1 1 2 1 1 1 1 1 1 2 1
                  3 2 2 4 1 1 1 4 4 3 1 1 3 2 1 1 1 2 2 3 1 1 1 1 1 1 1 7 1 1 1
                  2 1 3 1 8 2 8 1 2 1 1 3 2 4 1 1 2 2 5 1 3 1 3 2 2 4 1 3 1 1 4
                  2 1 4 1 2 2 2 1 2 2 1 1 1 1 3 3 4 2 2 2 5 1 3 1 1 1 1 1 2 1 2
                  3 1 4 1 1 2 3 1 1 4 2 5 2}) + 1;
               end;

            else if n = 124 then do;
               if m = 22 & l = 2 then
                  fac = suen(shape({4 1 1 1 4 1 2 5 5 1 2 4 5 5 3 5 1 2 3 6 5 4
                  4 1 1 1 4 1 6 4 3 2 3 2 3 1 6 3 5 6 4 6 3 3 3 4 3 1 4 6 6 3 4
                  1 1 3 3 2 5 3 2 3 3 2 1 1 1 2 1 5 2 1 3 4 1 5 3 1 1 4 4 1 5 2
                  1 1 6 4 3 2 6 5 5 1 4 6 6 3 1 4 1 2 4 3 5 6 1 3 3 2 6 4 3 3 1
                  2 4 5 5 5 4 3 2 2 3 3 5 5 4 4 3 4 2 3 4 2 5 4 4 4 1 1 1 2 6 6
                  3 3 1 1 3 4 4 6 6 6 4 3 4 4 1 3 2 4 6 2 6 2 5 5 2 4 5 2 3 6 2
                  3 3 1 4 4 4 4 5 6 5 3 3 6 2 3 2 1 4 6 2 5 2 2 6 2 6 5 3 2 4 2
                  3 6 4 5 5 6 6 6 4 1 4 6 5 5 2 3 2 3 4 3 6 4 5 2 6 4 2 5 5 6 6
                  5 5 1 1 2 5 1 5 6 6 1 1 4 2 6 5 4 3 1 2 2 6 2 1 6 2 3 2 2 5 6
                  6 1 5 2 3 4 2 4 3 4 2 6 5 4 2 6 6 2 3 1 1 1 2 5 1 6 6 1 1 3 5
                  4 1 6 6 5 6 1 4 5 5 2 3 2 3 2 6 6 4 6 4 6 1 6 5 1 1 1 3 6 4 2
                  6 6 2 2 6 3 6 6 1 5 3 6 2 1 6 5 5 2 4 5 4 5 2 4 3 1 3 2 5 4 2
                  4 3 3 5 2 1 6 2 2 2 2 6 5 4 6 1 4 5 3 2 1 1 2 6 2 6 3 3 3 1 2
                  6 3 5 3 3 6 4 1 3 2 4 6 2 5 5 5 3 2 6 5 4 2 5 6 2 4 1 4 2 1 3
                  1 3 3 5 5 5 5 4 4 2 6 2 3 4 5 1 2 2 3 2 2 6 4 5 2 5 6 3 6 3 4
                  1 1 3 5 1 4 4 5 4 1 6 2 1 2 2 4 2 1 4 2 3 6 2 2 4 3 3 2 1 3 6
                  3 2 1 2 4 6 3 5 3 6 4 1 4 2 4 1 3 5 1 4 2 4 3 1 3 1 4 1 4 1 5
                  2 6 2 6 5 4 1 2 5 2 6 5 6 6 1 1 2 2 3 5 6 3 2 6 1 4 3 3 4 5 6
                  3 1 1 5 3 4 4 2 6 6 1 3 6 6 2 1 3 2 4 3 4 2 5 2 6 4 5 4 1 6 3
                  1 5 5 6 3 6 6 2 1 4 3 6 3 3 1 4 1 6 4 4 1 4 4 6 6 2 2 5 6 3 4
                  3 6 6 4 3 3 6 1 6 5 6 5 6 1 5 6 5 2 1 5 4 1 5 6 2 3 3 6 2 3 4
                  1 3 2 5 6 5 1 1 2} , n / 4));
               end;

            else if n = 120 then do;
               if m = 2 & l = 2 then
                  fac = expand({0 3 2 1 1 1 1 2 3 1 2 2 1 1 2 2 1 4 1 2 3 1 2 1
                  1 2 2 1 3 5 1 2 4 2 1 1 1 1 2 1 2 2 1 2 1 1 1 1 2 1 2 1 2 1 1
                  1 1 2 4 1 2 1 2 1 2 5 3 7 7 2 1 1 1 3 1 2 1 1 5 9 5 2 3 1 3 1
                  2 1 2 1 2 5 3 7 7 2 1 1 1 3 1 2 1 1 5 9 5 2 3 1}) + 1;
               end;

            else if n = 116 then do;
               if m = 23 & l = 2 then
                  fac = suen(shape({4 1 4 1 2 1 5 2 1 2 3 2 3 5 2 3 3 2 6 5 3 1
                  4 1 2 3 5 1 6 2 6 5 6 1 5 1 2 2 6 1 3 5 3 1 4 2 1 5 4 5 1 2 2
                  5 6 2 2 2 6 2 3 3 3 1 1 2 2 2 5 1 2 6 5 4 3 3 4 1 2 4 4 2 5 6
                  1 4 6 1 1 2 4 5 1 6 6 3 6 6 2 2 5 4 1 6 6 5 4 1 6 4 4 4 2 3 4
                  4 4 3 3 5 5 3 1 5 1 2 6 2 1 2 6 1 1 2 6 1 2 1 6 1 4 1 1 3 5 1
                  2 1 2 2 3 3 5 4 2 2 2 2 1 5 5 4 6 6 4 6 6 1 1 6 4 5 4 2 4 2 1
                  2 5 5 6 6 6 1 1 3 3 2 3 5 4 5 3 1 2 3 1 1 1 2 2 5 6 4 3 3 1 2
                  5 3 6 4 5 5 6 5 6 6 6 3 1 1 2 4 2 5 6 6 6 4 5 4 1 1 6 5 3 3 5
                  6 6 2 2 3 4 3 5 5 5 3 5 4 2 6 6 2 6 5 4 4 3 3 1 6 4 2 3 2 5 3
                  4 4 2 4 2 1 5 3 3 4 1 4 4 3 1 4 2 5 4 1 3 4 3 1 1 4 6 4 6 3 5
                  4 1 2 3 4 3 4 1 2 1 5 6 1 1 4 3 1 1 5 1 4 6 2 6 4 1 3 4 6 2 6
                  1 2 1 3 6 1 4 6 3 6 1 6 5 5 3 3 5 4 1 3 2 2 1 4 2 3 3 2 5 4 3
                  2 3 1 4 6 2 1 1 6 4 4 4 2 1 6 2 4 6 5 2 1 2 5 6 3 6 5 5 2 1 4
                  6 4 2 1 4 3 4 4 1 5 4 1 6 5 3 4 3 2 5 1 2 5 6 6 2 5 3 4 3 1 4
                  5 1 1 3 2 4 2 1 4 6 1 2 6 3 2 2 2 2 6 5 4 1 5 5 6 5 3 5 5 2 3
                  3 6 5 6 2 6 2 5 5 6 2 6 2 3 4 3 1 6 4 2 1 5 2 5 1 6 3 2 2 4 2
                  3 1 6 5 6 2 2 5 1 4 3 4 3 2 3 6 1 1 4 2 2 4 6 5 2 5 5 2 5 4 4
                  4 6 6 1 1 4 6 5 2 1 2 4 2 4 3 6 4 6 2 6 3 2 6 6 1 3 4 6 3 1 6
                  5 4 1 2 3 2 5 2 4 4 5 2 3 1 5 4 3 3 2 4 2 4 2 1 4 2 2 4 3 6 3
                  6 5 2 2 4 6 6 5 6 6 4 4 4 3 2 4 5 5 5 3 2 3 5 6 6 2 3 6 4 1 4
                  2 4 2 6 6 2 3 1 3 5 1 2 1 6 5 5 4 2 3 2 5 1 4 2 1 4 1},
                  n / 4));
               end;

            else if n = 108 then do;

               if m = 22 & l = 2 then
                  fac = suen(shape({4 1 1 2 4 5 4 4 2 4 2 2 2 3 2 5 2 3 6 4 5 2
                  1 2 5 4 5 4 1 4 1 3 4 5 3 3 3 6 2 6 2 3 2 1 2 5 4 4 5 1 2 2 2
                  5 3 5 6 4 6 5 3 6 4 6 1 1 6 3 3 1 4 4 3 4 5 6 5 3 5 3 3 5 5 3
                  3 1 4 1 1 3 6 3 4 1 6 5 3 6 1 6 2 2 5 6 5 6 3 4 4 1 6 3 2 4 6
                  1 1 4 5 1 2 5 4 1 5 6 3 1 5 2 5 2 3 3 5 3 1 5 1 2 1 2 3 3 3 1
                  1 3 6 5 2 4 1 1 6 6 4 6 2 1 5 4 2 5 1 1 1 1 4 4 4 4 2 1 4 3 1
                  6 1 5 5 4 4 3 3 1 3 6 3 1 6 3 2 6 1 3 5 2 2 4 5 6 6 1 5 1 5 6
                  1 2 3 6 2 3 2 5 4 5 1 3 5 5 2 5 2 3 5 2 2 4 1 4 4 2 4 2 5 6 3
                  2 1 4 3 2 2 4 2 4 4 6 1 2 5 4 3 2 5 4 5 1 2 6 5 1 5 5 2 4 2 1
                  2 2 4 3 3 5 3 1 2 2 4 2 3 4 1 2 6 4 5 6 3 5 2 4 2 6 4 1 6 2 3
                  4 1 4 2 2 4 3 1 6 4 6 2 5 5 6 6 5 5 5 5 1 2 6 3 5 5 3 3 3 6 6
                  1 4 4 1 1 4 5 2 6 5 6 6 4 3 4 5 5 3 5 3 3 6 2 4 6 1 3 1 6 2 1
                  1 1 2 4 5 6 4 2 1 6 2 4 4 5 3 3 4 2 6 5 4 2 6 2 2 1 4 5 2 1 1
                  5 1 2 5 4 3 6 5 3 5 5 6 4 4 6 1 6 5 4 4 5 6 4 6 3 2 6 6 3 1 2
                  4 6 3 6 1 4 1 5 3 4 5 1 5 4 5 1 1 6 6 6 6 1 2 6 1 2 2 4 3 1 3
                  5 5 3 1 6 4 1 2 6 5 5 4 3 5 1 6 1 6 3 1 4 2 3 6 5 6 5 3 6 5 4
                  2 6 6 6 6 6 3 5 5 2 5 6 3 2 1 1 3 4 6 6 6 1 6 3 5 3 2 6 4 2 6
                  3 3 2 6 4 4 4 2 2 5 5 3 4 5 1 2 6 5 2 1 2 6 1 5 4 5 3 1 4 4 1
                  1 6 1 2 6 1 2 2 3 6 5 5 2 2 3 5 6 3}, n / 4));

               else if m = 7 & l = 2 then
                  fac = expand({2 1 2 2 3 1 1 1 1 1 2 1 2 1 1 2 2 2 2 2 1 3 1 3
                  1 1 1 1 2 1 1 2 3 2 5 1 2 1 1 1 1 3 2 4 3 1 2 2 1 1 1 1 2 1 1
                  3 2 4 5 1 5 1 2 1 1 4 7 1 2 1 2 2 1 2 1 1 2 6 1 3 1 1 1 1 1 1
                  1 2 3 2 1 2 1 1 2 4 1 1 1 1 1 2 2 2 1 6 1 4 2 6 2 4 9 1 2 2 2
                  2 1 1 1 1 1 1 5 3 1 2 3 1 1 3 2 3 2 1 2 1 1 1 5 1 4 1 6 1 2 4
                  1 1 1 1 2 3 1 1 1 2 1 2 2 1 3 2 1 1 1 1 3 1 2 1 1 2 2 3 2 1 1
                  1 2 1 1 1 4 2 1 1 2 4 1 3 1 1 1 1 3 4 1 4 3 1 1 1 1 3 1 1 6 2
                  1 3 4 1 1 2 4 2 1 1 1 1 2 1 2 2 1 1 1 1 1 3 1 2 1 1 1 1 1 2 3
                  1 2 2 2 6 1 3 1 1 1 4 1 1 2 2 1 2 3 2 1 3 2 1 3 7 3 3 1 2 1 2
                  2 9 1 6 1 1 6 1 2 1 1 2 3 1 2 3 2 5 5 4 2 5 1 1 3 4 2 2 1 2 1
                  1 3 7 1 2 4 1 2 1 1 1 1 1 1 3 1 1 1 3 1 2 4 1 2 1 1 1 1 2 5 1
                  1 2 1 1 2 3 1 1 4 3 5 1 1 1 1 2 5 1 6 2 1 2 3 1 8 1 1 2 1 2 1
                  1 1 1 2 1 1 1 2 4 2 1 2 3 2 1 3 1}) + 1;

               else if m = 5 & l = 2 then
                  fac = expand({1 1 1 1 1 3 1 2 2 1 1 5 1 2 1 3 3 1 2 1 1 1 2 1
                  1 1 3 2 2 4 4 1 4 1 5 1 6 1 1 5 3 1 2 2 1 2 2 1 2 1 1 1 2 2 1
                  2 1 2 2 1 1 5 1 2 2 6 2 1 3 4 3 1 3 1 1 1 3 3 4 4 2 3 1 1 1 2
                  2 2 1 1 2 3 2 1 2 5 3 1 1 6 1 1 1 1 2 6 2 1 1 1 3 1 1 1 1 1 1
                  1 3 1 1 2 1 6 2 2 1 1 3 1 6 3 3 1 1 4 5 1 1 3 2 1 3 2 2 2 1 2
                  3 3 4 3 3 1 1 1 2 1 1 1 1 1 5 1 1 1 1 3 1 2 1 1 1 3 2 2 2 2 1
                  1 1 1 1 2 1 3 6 1 1 3 2 1 1 1 1 1 3 2 2 1 2 1 2 1 1 1 6 4 1 2
                  4 3 1 1 3 1 1 2 1 1 2 1 1 1 2 4 1 6 1 4 1 4 1 3 3 1 2 2 2 1 5
                  2 1 3 2 2 1 3 6 1 1 2 1 4 3 1 2 2 1 2 2 1 6 3 1 1 1 2 2 2})
                  + 1;

               else if m = 5.1 & l = 2 then
                  fac = expand({2 3 3 1 1 1 1 1 1 2 1 3 1 3 1 1 1 1 1 1 1 2 1 2
                  2 1 1 2 2 5 2 2 3 1 1 2 1 1 2 1 4 2 5 2 1 1 3 4 3 6 2 1 1 3 2
                  1 6 4 1 2 2 1 2 2 5 1 1 2 3 1 1 1 1 1 1 1 1 1 3 4 2 2 1 1 1 3
                  2 1 1 1 1 1 3 2 1 2 1 1 1 1 2 2 1 1 8 2 3 1 5 3 3 2 5 1 2 4 1
                  1 1 3 2 2 1 1 2 2 2 9 1 1 3 2 1 4 1 1 3 1 5 2 2 1 1 1 3 4 2 3
                  1 1 1 3 2 1 1 1 1 1 2 1 2 1 2 1 2 1 3 6 1 3 1 2 5 4 1 1 1 6 3
                  3 4 1 1 1 1 2 1 3 3 6 2 1 3 1 1 4 1 1 4 3 3 2 1 2 2 2 2 2 1 5
                  3 3 1 3 1 1 2 1 2 1 1 2 6 1 1 1 1 4 4 3 1 1 2 1 3 1 3 4 1 3 1
                  2 1 3 1 1 1 2 2 1 3 2 1 1 4 5 1 2 3 1 7 1 3 1}) + 1;

               else if m = 1 & l = 2 then
                  fac = expand({1 3 2 1 2 2 1 1 3 2 4 6 1 1 1 1 1 1 2 1 1 3 1 1
                  2 2 2 1 2 1 1 1 1 1 1 1 1 1 2 2 2 2 1 1 4 1 2 2 2 4 1 1 1 2 1
                  4 2 1 6 3}) + 1;
               end;

            else if n = 100 then do;

               if m = 3 & l = 10 then
                  fac = { 0 0 2 1 1 9 0 1 4 1 0 7 2 4 2 3 5 4 2 5 9 3 4 7 4 8 2
                  5 9 4 4 9 9 5 8 7 6 2 2 7 3 4 6 3 9 7 2 7 8 6 7 9 7 4 8 7 9 9
                  6 2 2 2 4 3 3 1 2 3 6 3 2 9 4 6 4 5 7 6 4 7 1 5 6 9 6 0 4 7 1
                  6 6 1 1 7 0 9 8 4 4 9 5 6 8 5 1 9 4 9 0 8 9 1 9 6 0 9 1 1 8 4
                  4 4 6 5 5 3 4 5 8 5 4 1 6 8 6 7 9 8 6 9 3 7 8 1 8 2 6 9 3 8 8
                  3 3 9 2 1 0 6 6 1 7 8 0 7 3 1 6 1 2 0 1 3 1 8 2 1 3 3 0 6 6 6
                  8 7 7 5 6 7 0 7 6 3 8 0 8 9 1 0 8 1 5 9 0 3 0 4 8 1 5 0 0 5 5
                  1 4 3 2 8 8 3 9 0 2 9 5 3 8 3 4 2 3 5 3 0 4 3 5 5 2 8 8 8 0 9
                  9 7 8 9 2 9 8 5 0 2 0 1 3 2 0 3 7 1 2 5 2 6 0 3 7 2 2 7 7 3 6
                  5 4 0 0 5 1 2 4 1 7 5 0 5 6 4 5 7 5 2 6 5 7 7 4 0} + 1;

               else if m = 4 & l = 2 then
                  fac = expand({1 2 1 3 4 1 1 4 3 1 1 4 1 1 8 5 3 1 1 2 3 1 1 2
                  1 2 3 11 10 2 1 2 1 2 2 1 1 1 1 1 2 2 2 1 1 1 4 4 1 1 1 2 1 3
                  1 1 3 1 1 4 3 2 2 1 1 2 1 2 2 1 4 1 1 1 1 1 1 1 6 2 5 6 2 1 2
                  1 2 8 1 1 1 3 3 2 2 2 1 1 5 4 5 2 1 4 1 1 2 2 1 4 1 1 1 1 2 1
                  1 2 1 3 1 1 4 1 1 1 5 1 1 2 1 1 1 1 1 2 1 1 1 3 1 1 3 1 4 1 2
                  2 1 1 1 3 2 2 2 3 2 1 1 1 1 1 2 1 2 3 1 2 1 1 1 1 2 1 1 2 3 2
                  1 2 1 1 4 1 1 1 1 6 2 1 2 2 2 1 1 1 1 1 1 2 2 3 2 1 2}) + 1;

               else if m = 4 & l = 5 then
                  fac = shape({ 1 5 5 2 5 5 3 5 5 4 5 5 3 3 4 2 3 4 1 3 4 4 3 4
                  4 1 3 3 1 3 1 1 3 2 1 3 4 4 2 1 4 2 2 4 2 3 4 2 2 2 1 1 2 1 4
                  2 1 3 2 1 2 1 1 3 1 1 4 1 1 5 1 1 4 4 5 3 4 5 2 4 5 5 4 5 5 2
                  4 4 2 4 2 2 4 3 2 4 5 5 3 2 5 3 3 5 3 4 5 3 3 3 2 2 3 2 5 3 2
                  4 3 2 3 2 2 4 2 2 5 2 2 1 2 2 5 5 1 4 5 1 3 5 1 1 5 1 1 3 5 5
                  3 5 3 3 5 4 3 5 1 1 4 3 1 4 4 1 4 5 1 4 4 4 3 3 4 3 1 4 3 5 4
                  3 4 3 3 5 3 3 1 3 3 2 3 3 1 1 2 5 1 2 4 1 2 2 1 2 2 4 1 1 4 1
                  4 4 1 5 4 1 2 2 5 4 2 5 5 2 5 1 2 5 5 5 4 4 5 4 2 5 4 1 5 4 5
                  4 4 1 4 4 2 4 4 3 4 4 2 2 3 1 2 3 5 2 3 3 2 3 3 5 2 2 5 2 5 5
                  2 1 5 2 3 3 1 5 3 1 1 3 1 2 3 1 1 1 5 5 1 5 3 1 5 2 1 5},
                  100) || j(5,1,1) @ (1:5)` @ j(4,1,1);

               else if m = 4 & l = 10 then
                  fac = { 0 0 6 7 5 8 4 9 9 1 8 3 7 5 1 2 2 4 3 6 7 6 1 1 0 7 6
                  8 5 9 9 2 8 4 2 3 3 5 4 0 8 5 7 0 2 2 1 7 0 8 6 9 9 3 3 4 4 6
                  5 1 9 4 8 6 7 1 3 3 2 7 1 8 0 9 4 5 5 0 6 2 1 9 9 5 8 0 7 2 4
                  4 3 7 2 8 5 6 6 1 0 3 3 8 2 9 9 6 8 1 7 3 5 5 4 7 6 0 0 2 1 4
                  5 7 4 8 3 9 9 0 8 2 7 4 6 6 0 1 1 3 2 5 2 1 3 2 4 3 5 4 6 5 0
                  6 1 0 7 7 8 8 9 9 4 2 5 3 6 4 0 5 1 6 2 0 3 1 8 9 9 7 7 8 6 3
                  0 4 1 5 2 6 3 0 4 1 5 2 9 8 7 9 8 7};

               else if m = 22 & l = 2 then
                  fac = suen(shape({1 4 1 4 4 1 4 4 4 1 1 2 5 3 2 4 3 3 5 1 3 1
                  1 1 5 5 4 3 5 2 1 5 6 2 6 4 6 1 1 6 1 2 1 3 3 3 4 1 2 3 6 3 4
                  2 6 4 1 3 4 5 2 4 3 1 2 4 4 1 5 5 2 1 3 1 1 6 5 4 2 1 6 1 4 6
                  2 1 1 3 1 3 5 2 6 6 2 5 1 5 1 6 3 6 5 1 4 2 1 5 4 1 2 3 1 5 5
                  3 3 1 3 1 4 2 2 4 6 3 3 1 6 3 6 5 6 4 2 4 4 1 2 3 5 3 4 3 5 5
                  2 2 5 1 5 1 3 4 1 3 3 6 5 2 1 1 6 3 5 6 4 3 3 6 3 4 1 6 6 4 6
                  4 1 2 5 3 6 1 4 1 6 2 3 4 2 5 4 3 5 1 5 6 5 6 5 4 4 4 6 1 3 3
                  2 3 2 4 6 6 1 4 4 3 6 5 6 1 4 4 3 3 5 4 3 2 3 6 5 3 1 6 6 6 5
                  1 6 1 4 1 6 2 5 3 6 1 1 1 6 3 5 1 4 3 6 1 6 4 2 3 3 3 2 4 1 6
                  1 2 2 3 3 3 2 2 5 3 5 6 4 5 6 2 2 6 4 2 5 3 6 5 2 3 5 3 4 3 5
                  6 4 4 3 2 4 1 5 3 5 5 6 6 4 3 4 4 1 2 2 3 1 5 3 3 5 3 4 3 3 3
                  6 3 5 6 1 2 4 6 5 6 1 6 6 6 4 2 2 5 2 1 3 4 4 6 2 6 6 3 1 6 2
                  3 4 1 6 3 4 4 1 1 3 2 5 4 4 2 4 3 4 4 4 5 6 1 3 1 5 4 6 1 2 6
                  2 2 3 6 4 6 6 2 5 3 2 1 2 2 5 3 2 4 2 3 2 2 3 5 2 1 2 4 1 2 4
                  3 2 3 3 6 4 2 1 5 6 1 2 2 5 4 5 5 3 1 1 6 2 5 6 1 3 3 3 2 5 5
                  6 4 2 1 5 5 4 6 2 3 2 2 2 4 4 1 5 3 3 4 5 1 5 3 3 2 1 4 2 4 5
                  5 3 6 6 3 1 6 6 6 1 4 3 3 2 3 3 5 4 3 4 1 5 5 2 2 4 6 3 6 1 2
                  4 5 5 6 4 1 3}, n / 4));

               else if m = 16 & l = 2 then
                  fac = expand({ 17 3 3 3 1 1 4 3 1 1 2 2 1 1 2 2 1 1 2 2 2 2 1
                  1 1 1 2 1 1 1 1 2 1 1 1 4 1 1 1 2 1 3 1 1 4 3 2 1 5 1 1 5 1 1
                  2 2 7 6 2 4 1 1 1 4 1 1 1 1 4 1 1 1 1 3 1 1 1 1 1 1 1 1 1 3 2
                  1 1 2 2 3 1 3 2 2 5 3 2 4 2 2 2 1 2 2 1 2 1 1 1 1 3 4 1 3 5 1
                  2 2 4 5 2 1 1 2 1 1 1 2 2 2 1 1 3 3 1 1 2 1 1 2 1 1 2 1 1 2 2
                  1 1 2 2 2 2 2 1 3 1 1 1 1 2 2 2 3  18 3 3 3 1 1 4 3 1 1 2 2 1
                  1 2 2 1 1 2 2 2 2 1 1 1 1 2 1 1 1 1 2 1 1 1 4 1 1 1 2 1 3 1 1
                  4 3 2 1 5 1 1 5 1 1 2 2 7 6 2 4 1 1 1 4 1 1 1 1 4 1 1 1 1 3 1
                  1 1 1 1 1 1 1 1 3 2 1 1 2 2 3 1 3 2 2 5 3 2 4 2 2 2 1 2 2 1 2
                  1 1 1 1 3 4 1 3 5 1 2 2 4 5 2 1 1 2 1 1 1 2 2 2 1 1 3 3 1 1 2
                  1 1 2 1 1 2 1 1 2 2 1 1 2 2 2 2 2 1 3 1 1 1 1 2 2 2 3  18 3 3
                  3 1 1 4 3 1 1 2 2 1 1 2 2 1 1 2 2 2 2 1 1 1 1 2 1 1 1 1 2 1 1
                  1 4 1 1 1 2 1 3 1 1 4 3 2 1 5 1 1 5 1 1 2 2 7 6 2 4 1 1 1 4 1
                  1 1 1 4 1 1 1 1 3 1 1 1 1 1 1 1 1 1 3 2 1 1 2 2 3 1 3 2 2 5 3
                  2 4 2 2 2 1 2 2 1 2 1 1 1 1 3 4 1 3 5 1 2 2 4 5 2 1 1 2 1 1 1
                  2 2 2 1 1 3 3 1 1 2 1 1 2 1 1 2 1 1 2 2 1 1 2 2 2 2 2 1 3 1 1
                  1 1 2 2 2 3  18 3 3 3 1 1 4 3 1 1 2 2 1 1 2 2 1 1 2 2 2 2 1 1
                  1 1 2 1 1 1 1 2 1 1 1 4 1 1 1 2 1 3 1 1 4 3 2 1 5 1 1 5 1 1 2
                  2 7 6 2 4 1 1 1 4 1 1 1 1 4 1 1 1 1 3 1 1 1 1 1 1 1 1 1 3 2 1
                  1 2 2 3 1 3 2 2 5 3 2 4 2 2 2 1 2 2 1 2 1 1 1 1 3 4 1 3 5 1 2
                  2 4 5 2 1 1 2 1 1 1 2 2 2 1 1 3 3 1 1 2 1 1 2 1 1 2 1 1 2 2 1
                  1 2 2 2 2 2 1 3 1 1 1 1 2 2 2 3  18 3 3 3 1 1 4 3 1 1 2 2 1 1
                  2 2 1 1 2 2 2 2 1 1 1 1 2 1 1 1 1 2 1 1 1 4 1 1 1 2 1 3 1 1 4
                  3 2 1 5 1 1 5 1 1 2 2 7 6 2 4 1 1 1 4 1 1 1 1 4 1 1 1 1 3 1 1
                  1 1 1 1 1 1 1 3 2 1 1 2 2 3 1 3 2 2 5 3 2 4 2 2 2 1 2 2 1 2 1
                  1 1 1 3 4 1 3 5 1 2 2 4 5 2 1 1 2 1 1 1 2 2 2 1 1 3 3 1 1 2 1
                  1 2 1 1 2 1 1 2 2 1 1 2 2 2 2 2 1 3 1 1 1 1 2 2 2 3 1}) + 1;
               end;

            else if n = 96 then do;
               if m = 1 & l = 0 then
                  fac = {[5]0[5]1[5]2[5]3[2]0[2]1 0 1 0 1 0[2]1[4]0 1 0[4]1 0[6]
                  1[5]0 1[3]0 1 2[3]0[5]1[5]0[5]3[5]2[2]0[3]1 0 1 0 1[2]0 1[3]0
                  1 0[4]1 0[6]1[5]0 1[3]0[2]1[3]0[5]2[5]3[5]0[7]1[3]0 1 0[2]1[2]
                  0 1[3]0 1 0[4]1 0[6]1[5]0 1[3]0[2]2[3]0[5]3[5]2[5]1[5]0[2]1[2]
                  0 1 0 1[2]0[2]1[4]0 1 0[4]1 0[6]1[5]0 1[3]0 2 1[4]0 1[2]3 2 1
                  0[2]2 3 2 3[2]1 0 3 2[2]0 1[2]0[2]1 0 1 0[2]1[2]0[4]1[2]0[2]1
                  0 1[2]0[2]1[5]0 1 0[3]1 0[2]1 2 0[3]1 0[2]2 3 0 1[2]3 2 3 2[2]
                  0 1 2 3[2]1[3]0[3]1 0 1[2]0[2]1 0[3]1[2]0[2]1 0 1[2]0[2]1[5]0
                  1 0[3]1 0[3]1 0[2]1 2 3[2]1 0 3 2[2]0 1 0 1[2]3 2 1 0[2]2 3[2]
                  1[3]0 1 0 1 0[2]1 0[3]1[2]0[2]1 0 1[2]0[2]1[5]0 1 0[3]1 0 1[2]
                  2 0[2]1 3 2[2]0 1 2 3[2]1 0 1 0[2]2 3 0 1[2]3 2[2]1[2]0 1 0 1
                  0 1[2]0[4]1[2]0[2]1 0 1[2]0[2]1[5]0 1 0[3]1 0 1 2 1 0[2]1[5]0
                  [5]2[5]3[5]1 0[2]1[3]0[2]1 0 1 0 1 0[5]1 0 1[4]0 1 0[2]1 0[2]
                  1 0 1[2]0 1[2]3 1 0[2]2[5]1[5]3[5]2[6]0[2]1 0[2]1[2]0 1 0 1[2]
                  0[5]1 0 1[4]0 1 0[2]1 0[2]1 0 1[2]0 1[2]3 2 0[7]2[5]0[5]1[5]3
                  1[2]0 1[2]0[3]1 0 1[2]0[5]1 0 1[4]0 1 0[2]1 0[2]1 0 1[2]0 1 3
                  0 1 0[2]2[5]3[5]1[5]0[5]2 1[2]0[3]1[3]0 1 0 1 0[5]1 0 1[4]0 1
                  0[2]1 0[2]1 0 1[2]0 1 3 0 2 0[2]2 0 1[2]3[2]2 3[2]1 0 3 2[2]0
                  [2]1 0[2]2 3 0[2]1[3]0[3]1 0 1 0 1[3]0[2]1 0[3]1[2]0[2]1[3]0
                  1 0[3]1 0 1 2 3 1 0[2]3 1 0[2]2[2]3 2[2]0 1 2 3[2]1[2]0 1[2]3
                  2 0[2]1 0[2]1[3]0 1 0[2]1[3]0[2]1 0[3]1[2]0[2]1[3]0 1 0[3]1 0
                  1 2 3 2 0[2]3 2 3[2]1[2]0 1[2]3 2 1 0[2]2[2]3 2[2]0[2]1[2]0 1
                  [2]0[2]1 0 1 0[2]1[3]0[2]1 0[3]1[2]0[2]1[3]0 1 0[3]1 0 1 2 0
                  1 0[3]3 2[2]0[2]1 0[2]2 3 0 1[2]3[2]2 3[2]1 0 1[2]0[3]1[2]0 1
                  0 1 0 1[3]0[2]1 0[3]1[2]0[2]1[3]0 1 0[3]1 0 1 2 0 2 0[2]3[5]0
                  [5]3[5]1[5]2 0 1 0 1 0[2]1[3]0[2]1 0[5]1 0 1[3]0 1 0 1[2]0 1
                  [2]0 1 0[3]1 0 2 3 0 2 4[5]1[5]2[5]0[5]3 0 1 0[2]1[2]0[3]1[3]
                  0[5]1 0 1[3]0 1 0 1[2]0 1[2]0 1 0[3]1 0 2[2]0 2 4[5]2[5]1[5]3
                  [5]0 1 0 1[2]0[2]1 0[2]1[3]0[5]1 0 1[3]0 1 0 1[2]0 1[2]0 1 0
                  [3]1 0 1 3 0 2 4[5]3[5]0[5]2[6]1 0 1 0 1[2]0 1[2]0[2]1 0[5]1
                  0 1[3]0 1 0 1[2]0 1[2]0 1 0[3]1 0 1[2]0 2 4 0 1[2]3 2 3 2[2]0
                  [2]1 0[2]2 3 2 3[2]1[2]0 1 0 1 0[2]1 0[2]1[2]0 1[3]0[2]1 0[3]
                  1 0 1[2]0[3]1 0 1[3]0[3]1 2 3 0 3 5 1 0[2]2 3 2 3[2]1[2]0 1[2]
                  3 2 3 2[2]0 1 0 1 0[2]1[2]0 1[2]0[3]1[3]0[2]1 0[3]1 0 1[2]0[3]
                  1 0 1[3]0[3]1 2[2]0 3 5 2 3[2]1 0 1 0[2]2[2]3 2[2]0 1 0 1[2]3
                  2 1 0 1[2]0[2]1[3]0[3]1[3]0[2]1 0[3]1 0 1[2]0[3]1 0 1[3]0[4]1
                  3 0 3 5 3 2[2]0 1 0 1[2]3[2]2 3[2]1 0 1 0[2]2 3 1 0 1 0 1[2]0
                  [3]1[2]0 1[3]0[2]1 0[3]1 0 1[2]0[3]1 0 1[3]0[4]1[2]0 3 5 0[2]
                  2 3[2]0[2]2 3[2]0[2]2 3[2]0[2]2 3 0[8]1[7]0 1 0[4]1 0 1[5]0[5]
                  1 0 1 0[3]3[2]0 6 1[2]3 2[2]1[2]3 2[2]1[2]3 2[2]1[2]3 2[5]1[4]
                  0[4]1[3]0 1 0[4]1 0 1[5]0[5]1 0 1 0[2]3[3]0 6 2[2]0 1[2]2[2]0
                  1[2]2[2]0 1[2]2[2]0 1 2[4]0[8]1[3]0 1 0[4]1 0 1[5]0[5]1 0 1 0
                  3 0 3[2]0 6 3[2]1 0[2]3[2]1 0[2]3[2]1 0[2]3[2]1 0 3[15]0 1 0
                  [4]1 0 1[5]0[5]1 0 1 0 3[4]0 6 0 3 1 0 2 0 3 1 0 2 0 3 1 0 2
                  0 3 1 0 2[15]1[2]0[2]1 0 1[4]0[5]1 0 1[4]0 2[2]3 0 1 7 1 2 0
                  1 3 1 2 0 1 3 1 2 0 1 3 1 2 0 1 3[4]1[8]0[3]1[2]0[2]1 0 1[4]0
                  [5]1 0 1[4]0 2 3[2]0 1 7 2 1 3 2 0 2 1 3 2 0 2 1 3 2 0 2 1 3
                  2[5]0[4]1[4]0[3]1[2]0[2]1 0 1[4]0[5]1 0 1[4]0 2 0 3 0 1 7 3 0
                  2 3 1 3 0 2 3 1 3 0 2 3 1 3 0 2 3 1[8]0[7]1[2]0[2]1 0 1[4]0[5]
                  1 0 1[4]0 2[3]0 1 7 0[2]2 3 0[2]3 0[2]2[2]3 0[2]2[2]3 0[2]2 1
                  [3]0 1[30]0 2[3]0 1 4 8 1[2]3 2 1[2]2 1[2]3[2]2 1[2]3[2]2 1[2]
                  3 1[4]0[7]1[23]0 2[2]0 3 1 4 8 2[2]0 1 2[2]1 2[2]0[2]1 2[2]0
                  [2]1 2[3]0[4]1[3]0[4]1[23]0 2 0 3 0 1 4 8 3[2]1 0 3[2]0 3[2]1
                  [2]0 3[2]1[2]0 3[2]1 0[3]1 0[3]1[27]0 2 0[2]3 1 4 8 0 3 1 0 2
                  3 2 3 1 0 3 2 3 1 0 3 2 3 1 0 1[3]0 1[3]0[4]1 0[2]1[4]0[4]1[3]
                  0[2]1[3]0[4]1 3[3]0 1 5 9 1 2 0 1 3 2 3 2 0 1 2 3 2 0 1 2 3 2
                  0[2]1[4]0[3]1[5]0[2]1[4]0[4]1[3]0[2]1[3]0[4]1 3[2]0 3 1 5 9 2
                  1 3 2 0 1 0 1 3 2 1 0 1 3 2 1 0 1 3 2 0[4]1[8]0[2]1[4]0[4]1[3]
                  0[2]1[3]0[4]1 3 0 3 0 1 5 9 3 0 2 3 1 0 1 0 2 3 0 1 0 2 3 0 1
                  0 2 3 0[3]1 0[7]1 0[2]1[4]0[4]1[3]0[2]1[3]0[4]1 3 0[2]3 1 5 9
                  0[2]2 3 0 2 0 3 0 3 2 0 3 0 3 2 0 3 0 3[5]1[4]0[5]1 0[2]1[3]0
                  1 0 1[2]0 1 0[2]1 0 1 0[2]1 0 4 1 3 0 2 6 10 1[2]3 2 1 3 1 2
                  1 2 3 1 2 1 2 3 1 2 1 2[4]1 0[4]1[3]0[2]1 0[2]1[3]0 1 0 1[2]0
                  1 0[2]1 0 1 0[2]1 0 4 1[2]3 2 6 10 2[2]0 1 2 0 2 1 2 1 0 2 1
                  2 1 0 2 1 2 1[4]0 1[3]0 1[3]0[2]1 0[2]1[3]0 1 0 1[2]0 1 0[2]1
                  0 1 0[2]1 0 4 1[2]0 2 6 10 3[2]1 0 3 1 3 0 3 0 1 3 0 3 0 1 3
                  0 3[6]0[3]1 0[5]1 0[2]1[3]0 1 0 1[2]0 1 0[2]1 0 1 0[2]1 0 4 1
                  0 3 2 6 10 0 3 1 0[2]2 1 0 3 1 2 1 0 3 1 2 1 0 3[6]1[3]0 1[4]
                  0 1[2]0[3]1[2]0[2]1[2]0[2]1[2]0[2]1[2]0[2]1 5[2]3 0 2 7 11 1
                  2 0 1[2]3 0 1 2 0 3 0 1 2 0 3 0 1 2 0[4]1 0[3]1 0[3]1 0 1[2]0
                  [3]1[2]0[2]1[2]0[2]1[2]0[2]1[2]0[2]1 5[3]3 2 7 11 2 1 3 2[2]0
                  3 2 1 3 0 3 2 1 3 0 3 2 1 3[4]0 1[4]0[3]1 0 1[2]0[3]1[2]0[2]1
                  [2]0[2]1[2]0[2]1[2]0[2]1 5 3[2]0 2 7 11 3 0 2 3[2]1 2 3 0 2 1
                  2 3 0 2 1 2 3 0 2[5]0[4]1[4]0 1[2]0[3]1[2]0[2]1[2]0[2]1[2]0[2]
                  1[2]0[2]1 5 3 0 3 2 7 11[2]0 3[2]1 3[2]1 0 3[2]0 3[3]1 3 0 3
                  0 1 0 1 0 1 0[2]1[3]0 1[11]0[12]1 2 3[3]1 4 12[2]1 2[2]0 2[2]
                  0 1 2[2]1 2[3]0 2 1 2[2]1 0 1[2]0 1[2]0[3]1[12]0[12]1 2 3 1 2
                  1 4 12[2]2 1[2]3 1[2]3 2 1[2]2 1[3]3 1 2 1 2 0 1 0[2]1 0[5]1
                  [12]0[12]1 2 3 2[2]1 4 12[2]3 0[2]2 0[2]2 3 0[2]3 0[3]2 0 3 0
                  3 0 1 0 1 0 1[5]0 1[11]0[12]1 2 3[2]2 1 4 12 0 1 0 2[2]3 0 2
                  3 1 0 1 0 2 3 1 2 3 0 2 1 0 1 0 1 0[5]1[2]0[2]1[4]0[7]1[2]0[3]
                  1[4]0[2]3[3]1 5 13 1 0 1 3[2]2 1 3 2 0 1 0 1 3 2 0 3 2 1 3 1
                  0 1[2]0 1[5]0 1 0[2]1[4]0[7]1[2]0[3]1[4]0[2]3 1 2 1 5 13 2 3
                  2 0[2]1 2 0 1 3 2 3 2 0 1 3 0 1 2[2]0 1 0[2]1 0[2]1[3]0 1 0[2]
                  1[4]0[7]1[2]0[3]1[4]0[2]3 2[2]1 5 13 3 2 3 1[2]0 3 1 0 2 3 2
                  3 1 0 2 1 0 3 1 0 1 0 1 0 1[2]0[3]1[2]0[2]1[4]0[7]1[2]0[3]1[4]
                  0[2]3[2]2 1 5 13[2]0 3[2]1[2]0 3[3]1 3 0 3 0 3[2]1 0 3[2]1[2]
                  0[3]1[3]0 1 0[2]1 0 1[2]0[2]1 0 1 0 1 0 1 0 1 0 1[2]0 1 0 1[3]
                  2 3 1 8 14[2]1 2[2]0[2]1 2[3]0 2 1 2 1 2[2]0 1 2[2]1[5]0[3]1
                  0[3]1 0 1[2]0[2]1 0 1 0 1 0 1 0 1 0 1[2]0 1 0 1[3]2 0 1 8 14
                  [2]2 1[2]3[2]2 1[3]3 1 2 1 2 1[2]3 2 1[2]0[5]1 0[2]1 0[3]1 0
                  1[2]0[2]1 0 1 0 1 0 1 0 1 0 1[2]0 1 0 1[2]2 1 3 1 8 14[2]3 0
                  [2]2[2]3 0[3]2 0 3 0 3 0[2]2 3[3]0[2]1[3]0 1[2]0 1 0[2]1 0 1
                  [2]0[2]1 0 1 0 1 0 1 0 1 0 1[2]0 1 0 1[2]2 1 0 1 8 14 0 1 0 2
                  3 0 1 0 2 3 1 2 3 0 2 3 0 2 3[3]1[2]0[3]1 0[2]1 0[2]1 0[2]1 0
                  1[3]0[3]1 0[2]1[3]0[3]1[2]0 3 1 2 3 1 9 15 1 0 1 3 2 1 0 1 3
                  2 0 3 2 1 3 2 1 3 2 0[2]1[5]0 1[2]0 1 0 1 0[2]1 0 1[3]0[3]1 0
                  [2]1[3]0[3]1[2]0 3 1 2 0 1 9 15 2 3 2 0 1 2 3 2 0 1 3 0 1 2 0
                  1 2 0 1 3[2]0[5]1[3]0 1 0 1 0[2]1 0 1[3]0[3]1 0[2]1[3]0[3]1[2]
                  0 3[2]1 3 1 9 15 3 2 3 1 0 3 2 3 1 0 2 1 0 3 1 0 3 1 0 2[2]0
                  [2]1[3]0[3]1 0[2]1 0[2]1 0 1[3]0[3]1 0[2]1[3]0[3]1[2]0 3[2]1
                  0 1 9 15[2]0 3[3]1 3 0 3 0 3[2]1 0 3[2]0 3[3]1[2]0[3]1 0 1 0
                  1[2]0[2]1 0 1[2]0[2]1 0 1[2]0 1 0 1 0 1 0[2]1 0 1 0 2 1 0 2 1
                  8 16[2]1 2[3]0 2 1 2 1 2[2]0 1 2[2]1 2[2]0 1[2]0 1[2]0 1 0 1
                  0[4]1 0 1[2]0[2]1 0 1[2]0 1 0 1 0 1 0[2]1 0 1 0 2 1 0[2]1 8
                  16[2]2 1[3]3 1 2 1 2 1[2]3 2 1[2]2 1[2]3 0[2]1 0[2]1 0[2]1
                  0[4]1 0 1[2]0[2]1 0 1[2]0 1 0 1 0 1 0[2]1 0 1 0 2 1 3 2 1 8
                  16[2]3 0[3]2 0 3 0 3 0[2]2 3 0[2]3 0[2]2 0[2]1[3]0 1[2]0 1[2]
                  0[2]1 0 1[2]0[2]1 0 1[2]0 1 0 1 0 1 0[2]1 0 1 0 2 1 3[2]1 8
                  16 0 1 0 2 3 1 2 3 0 2 3 0 2 3 1 0 1 0 2 3 1[2]0[3]1 0[2]1 0
                  [3]1 0[2]1 0 1[3]0[2]1 0 1[2]0[3]1[3]0[2]1 3 2 0 2 1 9 17 1 0
                  1 3 2 0 3 2 1 3 2 1 3 2 0 1 0 1 3 2 1[2]0 1[2]0 1[2]0 1[2]0 1
                  0[2]1 0 1[3]0[2]1 0 1[2]0[3]1[3]0[2]1 3 2 0[2]1 9 17 2 3 2 0
                  1 3 0 1 2 0 1 2 0 1 3 2 3 2 0 1 0[2]1 0[2]1 0 1 0 1[2]0 1 0[2]
                  1 0 1[3]0[2]1 0 1[2]0[3]1[3]0[2]1 3 2 3 2 1 9 17 3 2 3 1 0 2
                  1 0 3 1 0 3 1 0 2 3 2 3 1[2]0[2]1[3]0 1 0 1 0[3]1 0[2]1 0 1[3]
                  0[2]1 0 1[2]0[3]1[3]0[2]1 3 2 3[2]1 9 17 0 2 1 2 1 2[2]0 1 2
                  0 2 1 2[3]1 2[5]0 1 0 1[3]0 1 0 1[2]0[3]1 0 1 0[2]1[2]0[3]1 0
                  1[3]0 1 0 1 4[2]0[2]2 10 18 1 3 0 3 0 3[2]1 0 3 1 3 0 3[3]0 3
                  [2]1[3]0[2]1 0[3]1 0 1[3]0[3]1 0 1 0[2]1[2]0[3]1 0 1[3]0 1 0
                  1 4[2]0 1 2 10 18 2 0 3 0 3 0[2]2 3 0 2 0 3 0[3]3 0[2]2[3]1[2]
                  0 1[2]0 1 0 1[3]0[3]1 0 1 0[2]1[2]0[3]1 0 1[3]0 1 0 1 4 0 3[2]
                  2 10 18 3 1 2 1 2 1[2]3 2 1 3 1 2 1[3]2 1[2]3[3]1 0 1 0[2]1 0
                  1 0 1[2]0[3]1 0 1 0[2]1[2]0[3]1 0 1[3]0 1 0 1 4 0 3 1 2 10 18
                  0 3 2 1 3 2 1 3 2[2]0 3 2 1 3 1 0 1 3 2[3]0 1 0 1[2]0 1 0 1 0
                  1 0 1 0 1 0[2]1[2]0 1 0[2]1 0 1[2]0[2]1[2]0 1 5 1 0[2]2 11 19
                  1 2 3 0 2 3 0 2 3[2]1 2 3 0 2 0 1 0 2 3[3]0[2]1 0[2]1 0 1 0[2]
                  1 0 1 0 1 0[2]1[2]0 1 0[2]1 0 1[2]0[2]1[2]0 1 5 1 0 1 2 11 19
                  2 1 0 3 1 0 3 1 0[2]2 1 0 3 1 3 2 3 1 0[3]1[2]0 1[3]0 1 0[2]1
                  0 1 0 1 0[2]1[2]0 1 0[2]1 0 1[2]0[2]1[2]0 1 5 1 3[2]2 11 19 3
                  0 1 2 0 1 2 0 1[2]3 0 1 2 0 2 3 2 0[4]1 0 1 0[3]1 0 1 0 1 0 1
                  0 1 0[2]1[2]0 1 0[2]1 0 1[2]0[2]1[2]0 1 5 1 3 1 2 11 19 0 2 1
                  2[3]1 2[2]0 2[2]0 1 2 0 2 1 2 1 0 1[4]0 1[2]0[2]1 0[2]1 0[2]1
                  [3]0 1 0[3]1 0 1[2]0 1 0 1[2]0 1 4[2]2 1 2 6 20 1 3 0 3[3]0 3
                  [2]1 3[2]1 0 3 1 3 0 3[2]0 1[2]0[2]1 0[2]1[2]0[3]1 0[2]1[3]0
                  1 0[3]1 0 1[2]0 1 0 1[2]0 1 4[4]2 6 20 2 0 3 0[3]3 0[2]2 0[2]
                  2 3 0 2 0 3 0 3 1 0[2]1[2]0 1 0 1[2]0[3]1 0[2]1[3]0 1 0[3]1 0
                  1[2]0 1 0 1[2]0 1 4 2[2]1 2 6 20 3 1 2 1[3]2 1[2]3 1[2]3 2 1
                  3 1 2 1 2 1 0[4]1 0 1 0[2]1 0[2]1 0[2]1[3]0 1 0[3]1 0 1[2]0 1
                  0 1[2]0 1 4 2 1[2]2 6 20 0 3 2 1 3 1 0 1 3[2]2 1 3 2[2]0 3 2
                  1 3 0 1[4]0 1 0 1[2]0 1 0 1[2]0[3]1[2]0[4]1[2]0[2]1[2]0[2]1[2]
                  0 5 0 2 1 2 7 21 1 2 3 0 2 0 1 0 2[2]3 0 2 3[2]1 2 3 0 2 0 1
                  [2]0[2]1 0 1 0[2]1[2]0 1[2]0[3]1[2]0[4]1[2]0[2]1[2]0[2]1[2]0
                  5 0[3]2 7 21 2 1 0 3 1 3 2 3 1[2]0 3 1 0[2]2 1 0 3[2]1 0[2]1
                  [2]0 1[2]0[2]1[2]0 1[2]0[3]1[2]0[4]1[2]0[2]1[2]0[2]1[2]0 5 0
                  [2]1 2 7 21 3 0 1 2 0 2 3 2 0[2]1 2 0 1[2]3 0 1 2 0 1 0[4]1 0
                  [2]1[2]0 1 0 1[2]0[3]1[2]0[4]1[2]0[2]1[2]0[2]1[2]0 5 0 1[2]2
                  7 21 0 2 1 2 1 0 2 1 2[3]1 2[2]0 2[2]0 1 2[2]0 1[4]0 1[2]0[2]
                  1[2]0[3]1 0 1 0[2]1 0 1[3]0 1 0[3]1 0 1 0 4 3 1 0 2 10 22 1 3
                  0 3 0 1 3 0 3[3]0 3[2]1 3[2]1 0 3[2]0 1 0[3]1 0[2]1[4]0[3]1 0
                  1 0[2]1 0 1[3]0 1 0[3]1 0 1 0 4 3 1 3 2 10 22 2 0 3 0 3 2 0 3
                  0[3]3 0[2]2 0[2]2 3 0[2]1 0 1[3]0[3]1[4]0[3]1 0 1 0[2]1 0 1[3]
                  0 1 0[3]1 0 1 0 4 3 2 0 2 10 22 3 1 2 1 2 3 1 2 1[3]2 1[2]3 1
                  [2]3 2[3]1 0[4]1[3]0[2]1[2]0[3]1 0 1 0[2]1 0 1[3]0 1 0[3]1 0
                  1 0 4 3 2 3 2 10 22 0 3 2 1 3 0 3 2 1 3 1 0 1 3[2]2 1 3 2[3]0
                  1[4]0[3]1[2]0 1 0 1 0 1 0[2]1[2]0[2]1[2]0 1 0[2]1[2]0[2]1 0 5
                  2 1 0 2 11 23 1 2 3 0 2 1 2 3 0 2 0 1 0 2[2]3 0 2 3 1[2]0 1 0
                  [3]1[3]0[3]1 0 1 0 1 0[2]1[2]0[2]1[2]0 1 0[2]1[2]0[2]1 0 5 2
                  1 3 2 11 23 2 1 0 3 1 2 1 0 3 1 3 2 3 1[2]0 3 1 0 2[2]1 0 1[3]
                  0 1[2]0[3]1 0 1 0 1 0[2]1[2]0[2]1[2]0 1 0[2]1[2]0[2]1 0 5[2]2
                  0 2 11 23 3 0 1 2 0 3 0 1 2 0 2 3 2 0[2]1 2 0 1 3[2]1 0[4]1 0
                  [2]1[2]0 1 0 1 0 1 0[2]1[2]0[2]1[2]0 1 0[2]1[2]0[2]1 0 5[2]2
                  3 2 11 23} + 1;
               end;

            else if n = 92 then do;
               if m = 21 & l = 2 then
                  fac = suen(shape({2 4 4 5 1 2 1 4 3 4 6 1 5 5 1 4 2 2 3 2 3 6
                  6 4 3 3 2 5 2 3 4 3 4 4 5 5 6 5 5 2 5 4 2 4 3 3 5 2 3 2 6 6 3
                  5 5 5 6 2 1 4 3 3 2 5 4 3 3 6 2 6 4 6 5 2 6 1 4 1 6 6 2 1 6 1
                  3 3 6 3 2 4 6 2 5 4 4 5 2 6 2 2 5 2 5 2 4 3 2 2 4 5 2 6 5 1 1
                  5 3 3 1 5 1 1 5 5 2 3 6 4 6 6 6 3 4 3 5 1 1 2 1 5 5 3 3 1 1 3
                  4 5 3 4 5 5 1 5 3 6 5 1 1 2 5 1 4 5 3 2 3 6 3 2 3 4 3 1 2 5 2
                  1 3 5 5 6 4 5 6 3 6 1 1 6 2 6 1 3 5 2 3 2 2 2 6 6 5 4 1 6 5 2
                  4 6 2 4 6 4 3 4 1 4 5 2 3 1 2 4 4 2 3 3 2 1 3 2 1 2 3 4 4 4 1
                  3 4 1 3 6 5 3 3 3 6 6 3 1 5 5 1 3 6 6 4 5 2 4 6 6 4 3 4 2 1 4
                  1 6 5 3 4 4 5 5 6 3 4 4 1 3 1 3 6 6 1 2 2 5 6 6 3 1 6 4 2 2 3
                  5 2 2 1 2 4 5 5 3 1 3 4 1 5 5 4 3 6 6 4 5 3 3 3 1 2 4 6 1 1 2
                  4 3 5 1 3 2 4 3 1 2 1 5 6 5 5 5 4 5 6 2 2 3 6 4 4 6 5 2 3 6 1
                  6 2 6 1 1 5 6 4 6 1 5 1 2 1 1 4 6 2 4 3 2 4 3 1 3 2 6 3 5 3 6
                  2 2 4 1 4 1 3 1 3 6 3 6 6 2 2 5 5 1 5 5 4 1 1 4 6 2 1 3 3 1 3
                  6 5 1 2 1 2 6 6 3 4 6 6 3 5 5 5 4 6 1 2 4 2 2 6 1 5 4 5 2 5 6
                  1 4 4 3}, n / 4));
              end;

            else if n = 84 then do;
               if m = 20 & l = 2 then
                  fac = suen(shape({1 4 1 4 4 2 4 5 5 5 3 6 6 2 2 6 2 3 3 1 3 6
                  4 3 1 4 3 4 1 4 3 6 6 1 2 4 6 3 1 4 1 6 1 4 3 1 3 3 3 6 2 3 4
                  5 1 5 6 4 4 1 1 6 3 4 6 4 3 1 6 6 4 6 3 2 5 3 5 2 4 3 1 2 5 2
                  5 3 4 1 1 2 5 4 1 6 6 3 1 2 4 5 1 3 4 5 3 1 2 3 4 3 5 1 5 4 4
                  2 3 2 5 3 4 5 1 1 6 6 3 2 1 5 3 3 2 6 6 4 6 1 3 4 3 3 6 1 4 3
                  2 4 3 2 5 3 2 1 5 2 6 3 4 2 3 1 3 5 1 6 2 6 4 6 1 4 4 2 2 6 4
                  2 4 3 4 1 3 2 1 1 6 4 2 2 3 5 4 2 2 2 2 3 5 2 1 1 5 1 4 6 1 1
                  2 3 1 2 3 6 6 6 4 2 2 4 2 5 6 6 2 2 3 2 5 3 3 3 3 4 1 4 4 5 5
                  1 6 5 5 1 5 5 6 1 6 5 2 5 3 1 5 2 3 1 3 1 5 4 6 2 4 1 2 3 6 1
                  2 1 1 5 1 5 4 5 3 1 4 3 4 5 2 3 6 6 3 2 3 6 5 1 4 4 1 4 5 2 3
                  2 1 6 6 6 6 1 4 3 3 1 6 4 4 4 2 6 1 6 1 3 2 4 5 5 1 6 5 3 5 6
                  2 5 4 3 6 6 3 5 4 6 2 1 5 5 5 3 1 1 5 5 1 3 5 4 1 3 4 1 2 3 5
                  3 4 3 3 1 6 3 6 4 4 4 6 4 4 6 5 4 6 2 2 6 4 6 1 1 5 3 6 6 4 6
                  1 1 3 4 3}, n / 4));

               else if m = 7 & l = 2 then
                  fac = expand({7 2 1 1 3 1 3 2 2 1 1 3 1 2 1 1 2 2 1 1 1 3 4 6
                  1 1 3 1 3 1 5 1 2 1 2 1 1 1 1 2 3 2 3 1 1 2 2 2 1 1 1 4 2 1 2
                  2 2 1 1 1 1 4 1 2 1 1 1 3 2 1 1 2 1 1 5 4 2 1 1 1 2 1 1 2 1 3
                  1 1 1 5 2 2 2 4 1 1 4 1 1 3 1 3 2 2 1 4 3 1 2 2 3 2 1 1 2 1 1
                  3 2 1 5 1 4 2 1 3 2 1 2 1 2 1 2 3 2 1 2 4 1 5 2 2 4 1 1 1 1 2
                  2 1 1 5 5 3 2 1 4 1 2 1 1 2 5 4 1 1 2 4 1 2 2 8 2 1 5 1 2 1 2
                  2 1 2 4 5 1 2 1 1 1 2 2 2 3 2 1 1 1 7 1 1 1 1 1 3 1 1 1 1 2 1
                  3 3 1 7 4 1 1 1 1 1 2 2 2 3 2 4 1 3 1 1 5 1 1 2 1 3 1 1 1 1 1
                  2 2 1 3 1 3 2 1 2 1 3 2 3 2 2 2 1 1 2 1 2 2 1 3 3 3 2 5 1 2 1
                  2 1 1 1 2 3 1 1 3 1 3 3 1 2 7 2 2 1 1 1 1 1 1 1 1 1 2}) + 1;

               else if m = 6 & l = 2 then
                  fac = expand({1 2 1 1 2 1 3 3 1 1 5 5 2 3 1 1 2 6 4 1 2 1 1 2
                  5 2 1 1 1 1 1 2 1 2 2 2 1 2 5 2 3 1 1 2 1 2 2 6 2 1 3 1 1 1 1
                  1 2 1 3 2 1 1 1 4 1 1 2 6 4 1 2 2 2 2 2 2 1 6 9 2 2 1 3 1 1 1
                  1 1 1 5 2 6 2 1 3 1 1 1 1 1 2 1 3 3 7 1 3 4 1 1 1 1 1 4 6 3 1
                  1 2 3 1 1 1 5 5 1 2 1 1 3 1 1 1 5 1 1 3 5 1 2 2 1 1 2 2 2 5 5
                  4 1 1 1 2 4 2 1 1 3 1 1 3 1 1 2 2 1 1 1 2 2 4 2 1 4 1 3 4 1 1
                  1 1 2 3 6 3 1 1 2 3 1 1 1 5 5 1 2 1 1 3 1 1 1 5 1 1 3 5 1 2 1
                  2 2 1 2 2 5 4 3 1 1 1 1 1 2 2 1 1 2 1 1 3 1 1 1 1 1 1 1 2 2 1
                  1 1 2 2 3}) + 1;

               else if m = 5 & l = 2 then
                  fac = expand({2 1 1 3 1 1 1 1 1 3 6 1 1 3 1 1 1 1 1 1 1 3 3 1
                  1 3 1 2 3 3 4 2 4 1 1 2 1 6 4 3 3 1 5 1 4 4 2 2 1 1 2 1 3 7 1
                  3 1 1 2 1 1 1 1 1 1 2 1 6 4 2 2 1 2 1 4 2 1 1 3 1 2 1 1 1 1 4
                  1 1 2 4 1 1 1 4 3 2 1 1 1 1 6 1 1 3 1 2 2 1 3 1 1 2 2 1 2 4 3
                  1 1 2 3 1 4 1 3 1 2 1 2 7 1 1 3 2 1 3 1 1 1 2 1 1 1 2 3 4 2 4
                  3 1 1 1 1 3 1 1 2 1 1 2 2 1 2 3 5 2 1 2 1 3 1 8 3 1 2 3 2 1 1
                  1 1 2 3 1 1 2 2 1 2 1 3 3 2 2 1 2 1 1 5 3 1 2 1 2 1 1 5 1 2 1
                  3}) + 1;

               else if m = 2 & l = 2 then
                  fac = expand({2 2 2 1 2 1 1 2 5 1 1 1 1 2 2 2 1 1 1 1 1 1 1 6
                  1 1 4 1 2 3 1 3 2 2 3 3 3 3 4 1 2 2 2 3 1 1 1 2 1 1 5 4 1 2 1
                  1 1 1 1 1 1 1 1 1 1 2 1 2 1 2 2 2 3 2 2 4 2 2 1 2 3 1 2 3 4 1
                  1 3 3}) + 1;

               else if m = 2.1 & l = 2 then
                  fac = expand({2 3 1 1 3 5 4 1 4 3 3 4 1 1 1 1 3 1 1 1 1 1 1 2
                  2 4 1 1 3 5 2 1 3 1 1 3 3 1 1 1 1 3 3 2 1 2 4 1 1 1 2 4 4 5 4
                  3 2 3 1 2 6 1 1 3 1 2 1 1 1 2 3 1 2 1 1 1 7 2 1 3 1 1 1 2 5 1
                  5 1 1 4 2 1 1 1 2 1 1 3 2 2 1 3 3 1 1 1 1 1 1 1 1 2 1 7 2 2 1
                  1 1 1 1 4 1 1 2}) + 1;

               else if m = 1 & l = 2 then
                  fac = { 1 2 1 2 2 1 2 1 2 1 1 2 2 1 2 2 1 1 2 2 2 1 1 1 1 1 2
                  2 2 1 2 1 1 1 2 2 1 2 1 1 2 2 2 1 2 1 1 2 1 1 1 1 2 1 2 2 2 1
                  1 1 2 1 2 2 2 2 2 1 1 2 1 2 1 2 2 1 2 1 2 1 2 1 2 1}`;
               end;

            else if n = 76 then do;
               if m = 19 & l = 2 then
                  fac = suen(shape({4 4 4 4 4 4 4 1 4 1 2 5 5 2 3 6 3 5 5 4 4 4
                  2 4 3 4 1 6 2 6 4 1 3 3 6 4 3 2 5 4 5 3 6 5 6 4 2 1 3 3 6 4 2
                  1 3 6 4 2 5 1 6 2 3 2 6 3 1 5 1 1 5 6 5 6 3 4 4 2 6 2 6 6 3 3
                  1 2 1 4 4 5 2 5 2 5 3 6 3 3 3 1 1 5 5 4 1 1 3 2 3 2 5 1 2 6 1
                  4 4 5 5 3 1 2 2 5 6 4 2 2 5 1 1 5 3 1 6 6 2 3 2 3 4 3 4 1 3 6
                  5 2 3 5 1 2 1 5 2 6 5 6 5 5 6 6 3 1 3 3 6 1 6 5 3 2 3 1 4 3 3
                  5 4 3 5 4 2 1 4 5 1 5 1 4 5 6 1 3 4 3 5 3 2 1 1 5 3 4 6 4 3 5
                  3 4 3 5 4 2 2 6 1 3 5 2 1 2 6 6 3 1 6 1 5 5 6 3 1 1 5 5 1 6 2
                  3 3 1 3 3 3 4 2 1 6 2 2 2 5 1 1 2 3 5 5 1 2 1 4 2 6 6 3 2 1 4
                  3 3 5 5 1 2 4 3 2 4 6 4 1 4 5 4 5 2 6 2 3 1 6 3 5 4 3 6 2 5 4
                  6 1 6 3 1 3 6 6 3 6 1 4 2 2 4 6 6 4 2 2 1 3 4 3 2 3 5 5 5 3 2
                  4 5 3 3 4 5 4 2 1 6}, n / 4));
               end;

            else if n = 72 then do;

               if m = 1 & l = 0 then
                  fac = { 4 3 2 2 3 3 2 2[4]3 1 1 3 3[4]1 2 2 1 1 2 1[4]2 1 2
                  [4]1[3]2 1 2 1 4 2 2 3 3 2 2 3 3 1 1 3 3 1 1 3 3 2 2[4]1 2
                  1[4]2 1[5]2[3]1 2 1 4 4 3 3 2 2 3 3 2 2 3 3 1 1 3 3 1 1 2 2 1
                  1[5]2 1 2[5]1 2 2 1 1 2[3]1[3]3 2 2 3 3 2 2 1 1 3 3 1 1 3 3 1
                  1[7]2 1 2[3]1 2 2 1 1 2 2 1 2 2 3 6 2 2[4]3 2 2 1 1[4]2 1 1 3
                  3 1 1 2 2 1 1 2 2 1 2[3]1 2 1[3]2[3]1 2 5 2 2[4]3[4]2[4]1 2 2
                  1 1 3 3 2 2 1 1 2 2 1 2 1 1 2 1 2[3]1[3]2 3 5 3 3[4]2 3 3 1
                  1[4]2[4]1 3 3[3]1 2 2 1[3]2 1 1 2 2 1 1[3]2 1 2 6 3 3[4]2 3 3
                  2 2[4]1 2 2 3 3[5]1 2 2 1[3]2 1 2 1 1 2 2[3]1 2 1 3 3 1 3 1 1
                  2 1 2 3 1 3 1 1 2 1 2 3 1 3[20]1 4 4 3 1 3 1 1 2 1 2 1 3 1 3
                  2 1 2 1 1 3 1 3[10]1[10]2 4 1 3 1 3 2 1 2 1 3 1 3 1 1 2 1 2 1
                  3 1 3 2 1 2 1 1 2 2 1 2[3]1 2 1 2 2 1 2 1 3 3 1 3 1 3 2 1 2 1
                  1 3 1 3 2 1 2 1 3 1 3 1 2 1 2 1 1 2 2 1 2 1 2 2 1 2 1 1 2 1 2
                  1 6 3 1 3 1 2 1 2 1 1 2 2 1 3 1 1 3 1 2 2 1 2 2[4]1[4]2[3]1 2
                  1[4]2 4 5 3 1 3 1 2 1 2 1 2 1 1 2 1 3 3 1 2 1 1[3]2[4]1[7]2 1
                  2[4]1 3 5 1 3 1 3 1 2 1 2 1 2 2 1 3 1 1 3 2 1 1 2 1 2 2 1
                  2[3]1 2 2 1 1[3]2 1 1[3]2 6 1 3 1 3 1 2 1 2 2 1 1 2 1 3 3 1 1
                  2 2 1 1 2 2 1 2[3]1[4]2[3]1 2 2 1 1 2 5 1 2 2 1 3 1 1 3 3 1 3
                  1 2 1 2 1 1 2 2 1 2 1 2 2[3]1 2 1 2 1 2 1 2 2 1 2 2 1 3 6 1 2
                  2 1 3 1 1 3 1 3 1 3 1 2 1 2 2 1 1 2 2 1 2 2[3]1 2 1 2 2 1 2 1
                  1 2 1 1 2 1 6 2 1 1 2 1 3 3 1 3 1 3 1 2 1 2 1 2 1 1 2 1[3]2 1
                  2 1 2 2 1 1 2 2 1 2 1 2 1 2 4 5 2 1 1 2 1 3 3 1 1 3 1 3 1 2 1
                  2 1 2 2 1 1[3]2 1 2 1 2 2 1 2 1 1 2 1 2 1 2 1 4 4 1 2 2 1 1 3
                  3 1 1 2 2 1 1 3 3 1 3 1 3 1 2 1 1[3]2 1 1 2 2 1 2[5]1 2 2 1 3
                  1 2 2 1 1 3 3 1 2 1 1 2 3 1 1 3 1 3 1 3 2 1 1[3]2 1 1[3]2
                  1[5]2 1 1 2 3 2 1 1 2 3 1 1 3 1 2 2 1 1 3 3 1 1 3 1 3 1 2 1 2
                  1 2 2 1 1 2 1[3]2 1 2 1 1 2 3 4 2 1 1 2 3 1 1 3 2 1 1 2 3 1 1
                  3 3 1 3 1 1 2 1 2 1 2 2 1 1 2 2[3]1 2 1 2 2 1 4 1 3 3 1 1 3
                  3[4]1 2 2 1 1[4]2 3 3 1 1 2 1[4]2 1 2[4]1[3]2 1 2 1 2 3 3 1 1
                  3 3 1 1 2 2 1 1 2 2 1 1 3 3 2 2 1 1 2 1[4]2 1[5]2[3]1 2 1 4 2
                  1 1 3 3 1 1 3 3 1 1 2 2 1 1 2 2 3 3[7]2 1 2[5]1 2 2 1 1 2[6]1
                  3 3 1 1 3 3 2 2 1 1 2 2 1 1 2 2 3 3[5]2 1 2[3]1 2 2 1 1 2 2 1
                  2 2 3 4 3 3[4]1 3 3 2 2[4]3 2 2 1 1[4]2 1 1 2 2 1 2[3]1 2
                  1[3]2[3]1 2[3]3[4]1[4]3[4]2 3 3 2 2 1 1 2 2 1 1 2 2 1 2 1 1 2
                  1 2[3]1[3]2 3 3 1 1[4]3 1 1 2 2[4]3[4]2[5]1 2 2 1[3]2 1 1 2 2
                  1 1[3]2 1 2 4 1 1[4]3 1 1 3 3[4]2 3 3 1 1 2 2[3]1 2 2 1[3]2 1
                  2 1 1 2 2[3]1 2[3]1 2 1 2 2 3 2 3 1 2 1 2 2 3 2 3 1 2 1
                  2[19]1 4 2 1 2 1 2 2 3 2 3 2 1 2 1 3 2 3 2 2 1 2[11]1[12]2 1
                  2 1 3 2 3 2 1 2 1 2 2 3 2 3 2 1 2 1 2 1 2 1 1 2 2 1 2[3]1 2 1
                  2 2 1 2 1 3 1 2 1 2 1 3 2 3 2 2 1 2 1 3 2 3 2 1 2 1 2 2 1 2 1
                  1 2 2 1 2 1 2 2 1 2 1 1 2 1 2 1 4 1 2 1 2 3 2 3 2 2 3 3 2 1 2
                  2 1 2 3 3[3]2[4]1[4]2[3]1 2 1[4]2 4 3 1 2 1 2 3 2 3 2 3 2 2 3
                  2 1 1 2 3 2 2 3 2 2[4]1[7]2 1 2[4]1 3 3 2 1 2 1 2 3 2 3 2 3 3
                  2 1 2 2 1 3 2 2 3 1 2 2 1 2[3]1 2 2 1 1[3]2 1 1[3]2 4 2 1 2 1
                  2 3 2 3 3 2 2 3 2 1 1 2 2 3 3 2 1 2 2 1 2[3]1[4]2[3]1 2 2 1 1
                  2 3 2 3 3 2 1 2 2 1 1 2 1 2 3 2 3 2 2 3 3 2 2 1 2 2[3]1 2 1 2
                  1 2 1 2 2 1 2 2 1 3 4 2 3 3 2 1 2 2 1 2 1 2 1 2 3 2 3 3 2 2 3
                  2 1 2 2[3]1 2 1 2 2 1 2 1 1 2 1 1 2 1 4 3 2 2 3 2 1 1 2 1 2 1
                  2 3 2 3 2 3 2 2 3 1[3]2 1 2 1 2 2 1 1 2 2 1 2 1 2 1 2 4 3 3 2
                  2 3 2 1 1 2 2 1 2 1 2 3 2 3 2 3 3 2 1[3]2 1 2 1 2 2 1 2 1 1 2
                  1 2 1 2 1 4 2 2 3 3 2 2 1 1 2 2 3 3 2 2 1 1 2 1 2 1 2 2 1
                  1[3]2 1 1 2 2 1 2[5]1 2 2 1 1 2 3 3 2 2 1 1 2 3 2 2 3 1 2 2 1
                  2 1 2 1 2 1 1[3]2 1 1[3]2 1[5]2 1 1 2 1 3 2 2 3 1 2 2 1 2 3 3
                  2 2 1 1 2 2 1 2 1 1 2 1 2 1 2 2 1 1 2 1[3]2 1 2 1 1 2 3 2 3 2
                  2 3 1 2 2 1 3 2 2 3 1 2 2 1 1 2 1 2 1 2 1 2 1 2 2 1 1 2 2[3]1
                  2 1 2 2 1 4 5 1 1 2 2 1 1[4]2 3 3 2 2[4]3[4]1 2 1[4]2 1
                  2[4]1[3]2 1 2 1 6 1 1 2 2 1 1 2 2 3 3 2 2 3 3 2 2 1 1 3 3 1 1
                  2 1[4]2 1[5]2[3]1 2 1 4 6 2 2 1 1 2 2 1 1 2 2 3 3 2 2 3 3 1 1
                  3 3[5]2 1 2[5]1 2 2 1 1 2[3]1 5 2 2 1 1 2 2 1 1 3 3 2 2 3 3 2
                  2 3 3 1 1[5]2 1 2[3]1 2 2 1 1 2 2 1 2 2 3 2 1 1[4]2 1 1 3
                  3[4]1 3 3 2 2 3 3 2 2 1 1 2 2 1 2[3]1 2 1[3]2[3]1
                  2[3]1[4]2[4]1[4]3 1 1 3 3[4]2 1 1 2 2 1 2 1 1 2 1 2[3]1[3]2 3
                  1 2 2[4]1 2 2 3 3[4]1[4]3 2 2[3]1 2 2 1[3]2 1 1 2 2 1 1[3]2
                  1[4]2[4]1 2 2 1 1[4]3 1 1 2 2 3 3[3]1 2 2 1[3]2 1 2 1 1 2
                  2[3]1 2 1 5 2 3 2 3 3 1 3 1 2 3 2 3 3 1 3 1 2 3 2 3[19]1 4 6
                  2 3 2 3 3 1 3 1 3 2 3 2 1 3 1 3 3 2 3 2[10]1[10]2 6 3 2 3 2 1
                  3 1 3 2 3 2 3 3 1 3 1 3 2 3 2 2 1 2 1 1 2 2 1 2[3]1 2 1 2 2 1
                  2 1 3 5 3 2 3 2 1 3 1 3 3 2 3 2 1 3 1 3 2 3 2 3 2 1 2 1 1 2 2
                  1 2 1 2 2 1 2 1 1 2 1 2 1 2 2 3 2 3 1 3 1 3 3 1 1 3 2 3 3 2 3
                  1 1 3 2 2[4]1[4]2[3]1 2 1[4]2 4 1 2 3 2 3 1 3 1 3 1 3 3 1 3 2
                  2 3 1 3 3 1 2 2[4]1[7]2 1 2[4]1 3 1 3 2 3 2 3 1 3 1 3 1 1 3 2
                  3 3 2 1 3 3 1 1 2 2 1 2[3]1 2 2 1 1[3]2 1 1[4]2 3 2 3 2 3 1 3
                  1 1 3 3 1 3 2 2 3 3 1 1 3 1 2 2 1 2[3]1[4]2[3]1 2 2 1 1 2 1 3
                  1 1 3 2 3 3 2 2 3 2 3 1 3 1 3 3 1 1 3 2 1 2 2[3]1 2 1 2 1 2 1
                  2 2 1 2 2 1 3 2 3 1 1 3 2 3 3 2 3 2 3 2 3 1 3 1 1 3 3 1 2 1 2
                  2[3]1 2 1 2 2 1 2 1 1 2 1 1 2 1 2 1 3 3 1 3 2 2 3 2 3 2 3 1 3
                  1 3 1 3 3 1 1[3]2 1 2 1 2 2 1 1 2 2 1 2 1 2 1 2 4 1 1 3 3 1 3
                  2 2 3 3 2 3 2 3 1 3 1 3 1 1 3 1[3]2 1 2 1 2 2 1 2 1 1 2 1 2 1
                  2 1 4 6 3 1 1 3 3 2 2 3 3 1 1 3 3 2 2 3 2 3 2 3 2 1 1[3]2 1 1
                  2 2 1 2[5]1 2 2 1 5 3 1 1 3 3 2 2 3 1 3 3 1 2 3 3 2 3 2 3 2 2
                  1 1[3]2 1 1[3]2 1[5]2 1 1 2 5 1 3 3 1 2 3 3 2 3 1 1 3 3 2 2 3
                  3 2 3 2 1 2 1 2 1 2 2 1 1 2 1[3]2 1 2 1 1 2 3 6 1 3 3 1 2 3 3
                  2 1 3 3 1 2 3 3 2 2 3 2 3 1 2 1 2 1 2 2 1 1 2 2[3]1 2 1 2 2 1
                  };

               else if m = 2 & l = 0 then
                  fac = { 4 5 3 2 2 3 3 2 2[4]3 1 1 3 3[4]1 2 1[4]2 1 2[3]1[3]2
                  1 2 1 6 4 2 2 3 3 2 2 3 3 1 1 3 3 1 1 3 3 1 1 2 1[4]2
                  1[4]2[3]1 2 1 4 6 4 3 3 2 2 3 3 2 2 3 3 1 1 3 3 1 1[5]2 1
                  2[5]1 2 1 1 2[3]1 5[3]3 2 2 3 3 2 2 1 1 3 3 1 1 3 3[5]2 1
                  2[3]1 2 2 1 2 2 1 2 2 3 3 6 2 2[4]3 2 2 1 1[4]2 1 1 2 2 1 1 2
                  2 1 2[3]1[4]2[3]1 2 4 5 2 2[4]3[4]2[4]1[4]2 1 1 2 2 1 2 1 1
                  2[4]1[3]2 3 4 5 3 3[4]2 3 3 1 1[4]2[5]1 2 2 1[3]2 1 1 2 1
                  1[3]2 1 2 3 6 3 3[4]2 3 3 2 2[4]1 2 2[3]1 2 2 1[3]2 1 2 1 2
                  2[3]1 2 1[3]3 1 3 1 1 2 1 2 3 1 3 1 1 2 1 2[18]1[3]4 3 1 3 1
                  1 2 1 2 1 3 1 3 2 1 2[11]1[9]2 4 4 1 3 1 3 2 1 2 1 3 1 3 1 1
                  2 1 2 2 1 2 1 1 2 2 1 2[4]1 2 2 1 2 1[3]3 1 3 1 3 2 1 2 1 1 3
                  1 3 2 1 2 1 2 1 2 1 1 2 2 1 2 1[3]2 1 1 2 1 2 1 5 6 3 1 3 1 2
                  1 2 1 1 2 2 1 3 1 1 3 2 2[4]1[4]2 1 1 2 1[4]2 4 6 5 3 1 3 1 2
                  1 2 1 2 1 1 2 1 3 3 1 2 2[4]1[6]2 1 2[4]1 3 6 5 1 3 1 3 1 2 1
                  2 1 2 2 1 3 1 1 3 1 2 2 1 2[3]1 2 2 1 1 2 2 1 1[3]2 5 6 1 3 1
                  3 1 2 1 2 2 1 1 2 1 3 3 1 1 2 2 1 2[3]1[4]2 1 1 2 2 1 1 2 5 5
                  1 2 2 1 3 1 1 3 3 1 3 1 2 1 2 1 2 1 2 2[3]1 2 1 2 1[3]2 1 2 2
                  1 3 6 6 1 2 2 1 3 1 1 3 1 3 1 3 1 2 1 2 2 1 2 2[3]1 2 1 2
                  2[3]1 2 1 1 2 1 6 6 2 1 1 2 1 3 3 1 3 1 3 1 2 1 2 1 1[3]2 1 2
                  1 2 2 1 1 2 1 2 1 2 1 2 4 5 5 2 1 1 2 1 3 3 1 1 3 1 3 1 2 1 2
                  1[3]2 1 2 1 2 2 1 2 1 2 1 2 1 2 1 4 3 4 1 2 2 1 1 3 3 1 1 2 2
                  1 1 3 3 1 2 1 1[3]2 1 1 2 2 1 2[4]1 2 2 1 4 3 1 2 2 1 1 3 3 1
                  2 1 1 2 3 1 1 3 2 1 1[3]2 1 1[3]2 1[4]2 1 1 2 4 3 2 1 1 2 3 1
                  1 3 1 2 2 1 1 3 3 1 1 2 1 2 1 2 2 1 1 2 1 2 2 1 2 1 1 2 3 3 4
                  2 1 1 2 3 1 1 3 2 1 1 2 3 1 1 3 1 2 1 2 1 2 2 1 1 2 2 1 1 2 1
                  2 2 1 4 1 1 3 3 1 1 3 3[4]1 2 2 1 1 2 2 1 1 2 1[4]2 1
                  2[3]1[3]2 1 2 1 2 2 3 3 1 1 3 3 1 1 2 2 1 1 2 2[4]1 2 1[4]2
                  1[4]2[3]1 2 1 4 2 2 1 1 3 3 1 1 3 3 1 1 2 2 1 1[7]2 1 2[5]1 2
                  1 1 2[7]1 3 3 1 1 3 3 2 2 1 1 2 2 1 1[5]2 1 2[3]1 2 2 1 2 2 1
                  2 2 3 5 4 3 3[4]1 3 3 2 2[4]3[4]2 1 1 2 2 1 2[3]1[4]2[3]1 2
                  6[3]3[4]1[4]3[4]2 3 3 2 2 1 1 2 2 1 2 1 1 2[4]1[3]2 3 6 3 1
                  1[4]3 1 1 2 2[4]3 2 2[3]1 2 2 1[3]2 1 1 2 1 1[3]2 1 2 5 4 1
                  1[4]3 1 1 3 3[4]2 3 3[3]1 2 2 1[3]2 1 2 1 2 2[3]1 2 1 5 1 1 2
                  1 2 2 3 2 3 1 2 1 2 2 3 2 3[18]1 4 6 2 1 2 1 2 2 3 2 3 2 1 2
                  1 3 2 3 2[10]1[9]2 6 2 2 1 2 1 3 2 3 2 1 2 1 2 2 3 2 3 2 1 2
                  1 1 2 2 1 2[4]1 2 2 1 2 1 3 5 1 2 1 2 1 3 2 3 2 2 1 2 1 3 2 3
                  2 2 1 2 1 1 2 2 1 2 1[3]2 1 1 2 1 2 1 1 4 1 2 1 2 3 2 3 2 2 3
                  3 2 1 2 2 1 2 2[4]1[4]2 1 1 2 1[4]2 4 2 3 1 2 1 2 3 2 3 2 3 2
                  2 3 2 1 1[3]2[4]1[6]2 1 2[4]1 3 2 3 2 1 2 1 2 3 2 3 2 3 3 2 1
                  2 2 1 1 2 2 1 2[3]1 2 2 1 1 2 2 1 1[3]2 1 4 2 1 2 1 2 3 2 3 3
                  2 2 3 2 1 1 2 1 2 2 1 2[3]1[4]2 1 1 2 2 1 1 2 1 3 2 3 3 2 1 2
                  2 1 1 2 1 2 3 2 3 2 2 1 2 2[3]1 2 1 2 1[3]2 1 2 2 1 3 2 4 2 3
                  3 2 1 2 2 1 2 1 2 1 2 3 2 3 2 1 2 2[3]1 2 1 2 2[3]1 2 1 1 2 1
                  2 4 3 2 2 3 2 1 1 2 1 2 1 2 3 2 3 2 1[3]2 1 2 1 2 2 1 1 2 1 2
                  1 2 1 2 4 1 3 3 2 2 3 2 1 1 2 2 1 2 1 2 3 2 3 1[3]2 1 2 1 2 2
                  1 2 1 2 1 2 1 2 1 4 5 2 2 3 3 2 2 1 1 2 2 3 3 2 2 1 1 2 2 1
                  1[3]2 1 1 2 2 1 2[4]1 2 2 1 6 1 2 3 3 2 2 1 1 2 3 2 2 3 1 2 2
                  1 2 1 1[3]2 1 1[3]2 1[4]2 1 1 2 6 1 3 2 2 3 1 2 2 1 2 3 3 2 2
                  1 1 2 1 2 1 2 1 2 2 1 1 2 1 2 2 1 2 1 1 2 3 5 2 3 2 2 3 1 2 2
                  1 3 2 2 3 1 2 2 1 1 2 1 2 1 2 2 1 1 2 2 1 1 2 1 2 2 1 4 3 5 1
                  1 2 2 1 1[4]2 3 3 2 2 3 3 1 1 2 1[4]2 1 2[3]1[3]2 1 2 1 4 6 1
                  1 2 2 1 1 2 2 3 3 2 2 3 3 2 2 1 1 2 1[4]2 1[4]2[3]1 2 1 4 4 6
                  2 2 1 1 2 2 1 1 2 2 3 3 2 2 3 3[5]2 1 2[5]1 2 1 1 2[3]1 3 5 2
                  2 1 1 2 2 1 1 3 3 2 2 3 3[7]2 1 2[3]1 2 2 1 2 2 1 2 2 3 1 2 1
                  1[4]2 1 1 3 3[4]1 3 3 2 2 1 1 2 2 1 2[3]1[4]2[3]1 2
                  2[3]1[4]2[4]1[4]3 1 1 2 2 1 1 2 2 1 2 1 1 2[4]1[3]2 3 2 1 2
                  2[4]1 2 2 3 3[4]1 3 3[3]1 2 2 1[3]2 1 1 2 1 1[3]2 1 2
                  1[3]2[4]1 2 2 1 1[4]3[5]1 2 2 1[3]2 1 2 1 2 2[3]1 2 1 1 5 2 3
                  2 3 3 1 3 1 2 3 2 3 3 1 3[19]1 4 2 6 2 3 2 3 3 1 3 1 3 2 3 2
                  1 3 1 3[10]1[10]2 6 3 2 3 2 1 3 1 3 2 3 2 3 3 1 3 1 2 1 2 1 1
                  2 2 1 2[4]1 2 2 1 2 1 3 1 5 3 2 3 2 1 3 1 3 3 2 3 2 1 3 1 3 2
                  1 2 1 1 2 2 1 2 1[3]2 1 1 2 1 2 1 3 2 2 3 2 3 1 3 1 3 3 1 1 3
                  2 3 3[3]2[4]1[4]2 1 1 2 1[4]2 4 4 1 2 3 2 3 1 3 1 3 1 3 3 1 3
                  2 2 3 2 2[4]1[6]2 1 2[4]1 3 4 1 3 2 3 2 3 1 3 1 3 1 1 3 2 3 3
                  2 1 2 2 1 2[3]1 2 2 1 1 2 2 1 1[3]2 3 2 3 2 3 2 3 1 3 1 1 3 3
                  1 3 2 2 3 1 2 2 1 2[3]1[4]2 1 1 2 2 1 1 2 3 1 3 1 1 3 2 3 3 2
                  2 3 2 3 1 3 1 3 2 1 2 2[3]1 2 1 2 1[3]2 1 2 2 1 3 4 2 3 1 1 3
                  2 3 3 2 3 2 3 2 3 1 3 1 2 1 2 2[3]1 2 1 2 2[3]1 2 1 1 2 1 4 2
                  1 3 3 1 3 2 2 3 2 3 2 3 1 3 1 3 1[3]2 1 2 1 2 2 1 1 2 1 2 1 2
                  1 2 4 3 1 1 3 3 1 3 2 2 3 3 2 3 2 3 1 3 1 1[3]2 1 2 1 2 2 1 2
                  1 2 1 2 1 2 1 4 1 6 3 1 1 3 3 2 2 3 3 1 1 3 3 2 2 3 2 1 1[3]2
                  1 1 2 2 1 2[4]1 2 2 1 2 5 3 1 1 3 3 2 2 3 1 3 3 1 2 3 3 2 2 1
                  1[3]2 1 1[3]2 1[4]2 1 1 2 2 5 1 3 3 1 2 3 3 2 3 1 1 3 3 2 2 3
                  1 2 1 2 1 2 2 1 1 2 1 2 2 1 2 1 1 2 3 1 6 1 3 3 1 2 3 3 2 1 3
                  3 1 2 3 3 2 1 2 1 2 1 2 2 1 1 2 2 1 1 2 1 2 2 1};

               else if m = 3 & l = 0 then
                  fac = { 4 3 5 3 2 2 3 3 2 2[4]3[4]1 2 1[4]2 1 2 1 1[3]2 1 2 1
                  4 6 4 2 2 3 3 2 2 3 3 1 1 3 3 1 1 2 1[4]2 1[3]2[3]1 2 1 4 3 6
                  4 3 3 2 2 3 3 2 2 3 3 1 1[5]2 1 2[4]1 2 1 1 2[3]1 4 5[3]3 2 2
                  3 3 2 2 1 1 3 3[5]2 1 2[3]1 2 1 2 2 1 2 2 3 6 3 6 2 2[4]3 2 2
                  1 1[4]2 1 1 2 2 1 2[3]1[3]2[3]1 2 5 4 5 2 2[4]3[4]2 1 1 2 2 1
                  1 2 2 1 2 1 1 2[3]1[3]2 3 6 4 5 3 3[4]2 3 3 1 1 2 2[3]1 2 2
                  1[3]2[4]1[3]2 1 2 5 3 6 3 3[4]2 3 3 2 2[5]1 2 2 1[3]2
                  1[3]2[3]1 2 1 5[3]3 1 3 1 1 2 1 2 3 1 3[18]1 4 6 4 4 3 1 3 1
                  1 2 1 2 1 3 1 3[10]1[8]2 5 4 4 1 3 1 3 2 1 2 1 3 1 3 1 2 1 2
                  1 1 2 2 1 2[3]1 2 2 1 2 1 3 6 3 3 1 3 1 3 2 1 2 1 1 3 1 3 2 1
                  2 1 1 2 2 1 2 1 2 2 1 1 2 1 2 1 3 5 6 3 1 3 1 2 1 2 1 1 2 2 1
                  2 2[4]1[4]2 1 2 1[4]2 4 4 6 5 3 1 3 1 2 1 2 1 2 1
                  1[3]2[4]1[5]2 1 2[4]1 3 3 6 5 1 3 1 3 1 2 1 2 1 2 2 1 1 2 2 1
                  2[3]1 2 2 1 2 2 1 1[3]2 4 5 6 1 3 1 3 1 2 1 2 2 1 1 2 1 2 2 1
                  2[3]1[3]2 1 1 2 2 1 1 2 6 5 5 1 2 2 1 3 1 1 3 3 1 3 1 2 1 2
                  2[3]1 2 1 2 1 2 2 1 2 2 1 3 5 6 6 1 2 2 1 3 1 1 3 1 3 1 3 2 1
                  2 2[3]1 2 1 2 2 1 1 2 1 1 2 1[3]6 2 1 1 2 1 3 3 1 3 1 3 1
                  1[3]2 1 2 1 2 2[3]1 2 1 2 1 2 4[3]5 2 1 1 2 1 3 3 1 1 3 1 3
                  1[3]2 1 2 1 2 2 1 2 2 1 2 1 2 1 4 4 3 4 1 2 2 1 1 3 3 1 1 2 2
                  1 2 1 1[3]2 1 1 2 2[5]1 2 2 1 3 4 3 1 2 2 1 1 3 3 1 2 1 1 2 2
                  1 1[3]2 1 1[7]2 1 1 2 4 4 3 2 1 1 2 3 1 1 3 1 2 2 1 1 2 1 2 1
                  2 2 1 1 2 1 2 1 2 1 1 2[3]3 4 2 1 1 2 3 1 1 3 2 1 1 2 1 2 1 2
                  1 2 2 1 1 2 2 1 2 1 2 2 1 4 5 1 1 3 3 1 1 3 3[4]1 2 2 1 1 2
                  1[4]2 1 2 1 1[3]2 1 2 1 6 2 2 3 3 1 1 3 3 1 1 2 2[4]1 2 1[4]2
                  1[3]2[3]1 2 1 4 5 2 2 1 1 3 3 1 1 3 3 1 1[7]2 1 2[4]1 2 1 1
                  2[3]1 6[4]1 3 3 1 1 3 3 2 2 1 1[5]2 1 2[3]1 2 1 2 2 1 2 2 3 2
                  5 4 3 3[4]1 3 3 2 2 3 3 2 2 1 1 2 2 1 2[3]1[3]2[3]1 2 1
                  6[3]3[4]1[4]3[4]2 1 1 2 2 1 2 1 1 2[3]1[3]2 3 2 6 3 1 1[4]3 1
                  1 2 2 3 3[3]1 2 2 1[3]2[4]1[3]2 1 2 1 5 4 1 1[4]3 1 1 3 3 2
                  2[3]1 2 2 1[3]2 1[3]2[3]1 2 1 1 5 1 1 2 1 2 2 3 2 3 1 2 1
                  2[17]1 4 2 6 2 1 2 1 2 2 3 2 3 2 1 2[11]1[8]2 1 6 2 2 1 2 1 3
                  2 3 2 1 2 1 2 2 1 2 1 1 2 2 1 2[3]1 2 2 1 2 1 3 2 5 1 2 1 2 1
                  3 2 3 2 2 1 2 1 2 1 2 1 1 2 2 1 2 1 2 2 1 1 2 1 2 1 5 1 4 1 2
                  1 2 3 2 3 2 2 3 3[3]2[4]1[4]2 1 2 1[4]2 4 6 2 3 1 2 1 2 3 2 3
                  2 3 2 2 3 2 2[4]1[5]2 1 2[4]1 3 5 2 3 2 1 2 1 2 3 2 3 2 3 3 2
                  1 2 2 1 2[3]1 2 2 1 2 2 1 1[3]2 6 1 4 2 1 2 1 2 3 2 3 3 2 2 3
                  1 2 2 1 2[3]1[3]2 1 1 2 2 1 1 2 2 1 3 2 3 3 2 1 2 2 1 1 2 1 2
                  2 1 2 2[3]1 2 1 2 1 2 2 1 2 2 1 3 1 2 4 2 3 3 2 1 2 2 1 2 1 2
                  1 2 1 2 2[3]1 2 1 2 2 1 1 2 1 1 2 1 2 2 4 3 2 2 3 2 1 1 2 1 2
                  1 2 1[3]2 1 2 1 2 2[3]1 2 1 2 1 2 4 1 1 3 3 2 2 3 2 1 1 2 2 1
                  2 1 1[3]2 1 2 1 2 2 1 2 2 1 2 1 2 1 4 6 5 2 2 3 3 2 2 1 1 2 2
                  3 3 2 2 1 1[3]2 1 1 2 2[5]1 2 2 1 5 6 1 2 3 3 2 2 1 1 2 3 2 2
                  3 2 1 1[3]2 1 1[7]2 1 1 2 6 6 1 3 2 2 3 1 2 2 1 2 3 3 2 1 2 1
                  2 1 2 2 1 1 2 1 2 1 2 1 1 2 3 5 5 2 3 2 2 3 1 2 2 1 3 2 2 3 1
                  2 1 2 1 2 2 1 1 2 2 1 2 1 2 2 1 4 1 3 5 1 1 2 2 1 1[4]2 3 3 1
                  1 2 1[4]2 1 2 1 1[3]2 1 2 1 2 4 6 1 1 2 2 1 1 2 2 3 3 2 2 1 1
                  2 1[4]2 1[3]2[3]1 2 1 4 1 4 6 2 2 1 1 2 2 1 1 2 2 3 3[5]2 1
                  2[4]1 2 1 1 2[3]1 2 3 5 2 2 1 1 2 2 1 1 3 3[7]2 1 2[3]1 2 1 2
                  2 1 2 2 3 4 1 2 1 1[4]2 1 1 3 3 1 1 2 2 1 1 2 2 1
                  2[3]1[3]2[3]1 2 3 2[3]1[4]2[4]1 3 3 2 2 1 1 2 2 1 2 1 1
                  2[3]1[3]2 3 4 2 1 2 2[4]1 2 2 3 3[5]1 2 2 1[3]2[4]1[3]2 1 2 3
                  1[3]2[4]1 2 2 1 1 3 3[3]1 2 2 1[3]2 1[3]2[3]1 2 1 3 1 5 2 3 2
                  3 3 1 3 1 2 3 2 3[17]1 4 4 2 6 2 3 2 3 3 1 3 1 3 2 3
                  2[10]1[8]2 3 2 6 3 2 3 2 1 3 1 3 2 3 2 3 2 1 2 1 1 2 2 1
                  2[3]1 2 2 1 2 1 3 4 1 5 3 2 3 2 1 3 1 3 3 2 3 2 2 1 2 1 1 2 2
                  1 2 1 2 2 1 1 2 1 2 1 1 3 2 2 3 2 3 1 3 1 3 3 1 1 3 2
                  2[4]1[4]2 1 2 1[4]2 4 2 4 1 2 3 2 3 1 3 1 3 1 3 3 1 2
                  2[4]1[5]2 1 2[4]1 3 1 4 1 3 2 3 2 3 1 3 1 3 1 1 3 1 2 2 1
                  2[3]1 2 2 1 2 2 1 1[4]2 3 2 3 2 3 2 3 1 3 1 1 3 3 1 1 2 2 1
                  2[3]1[3]2 1 1 2 2 1 1 2 4 3 1 3 1 1 3 2 3 3 2 2 3 2 3 2 1 2
                  2[3]1 2 1 2 1 2 2 1 2 2 1 3 3 4 2 3 1 1 3 2 3 3 2 3 2 3 2 2 1
                  2 2[3]1 2 1 2 2 1 1 2 1 1 2 1 4 4 2 1 3 3 1 3 2 2 3 2 3 2 3
                  1[3]2 1 2 1 2 2[3]1 2 1 2 1 2 4 3 3 1 1 3 3 1 3 2 2 3 3 2 3 2
                  1[3]2 1 2 1 2 2 1 2 2 1 2 1 2 1 4 2 1 6 3 1 1 3 3 2 2 3 3 1 1
                  3 2 1 1[3]2 1 1 2 2[5]1 2 2 1 1 2 5 3 1 1 3 3 2 2 3 1 3 3 1 2
                  1 1[3]2 1 1[7]2 1 1[3]2 5 1 3 3 1 2 3 3 2 3 1 1 3 1 2 1 2 1 2
                  2 1 1 2 1 2 1 2 1 1 2 3 1 1 6 1 3 3 1 2 3 3 2 1 3 3 1 1 2 1 2
                  1 2 2 1 1 2 2 1 2 1 2 2 1};

               else if m = 5 & l = 0 then
                  fac = { 4 3 1 2 2 3 3 2 2[4]3 1 1 3 3[4]1 2 2 1 1 2[4]1[3]2 1
                  2 1 4 1 2 2 3 3 2 2 3 3 1 1 3 3 1 1 3 3 2 2[4]1[5]2[3]1 2 1 4
                  4 1 3 3 2 2 3 3 2 2 3 3 1 1 3 3 1 1 2 2 1 1[3]2 1 1 2 2 1 1
                  2[3]1 3 1 3 3 2 2 3 3 2 2 1 1 3 3 1 1 3 3 1 1[7]2 1 1 2 2 1 2
                  2 3 6 1 2 2[4]3 2 2 1 1[4]2 1 1 3 3 1 1 2 2 1 1 2 1[3]2[3]1 2
                  5 1 2 2[4]3[4]2[4]1 2 2 1 1 3 3 2 2 1 2 1 2[3]1[3]2 3 5 1 3
                  3[4]2 3 3 1 1[4]2[4]1 3 3[4]1 2 2 1 1[3]2 1 2 6 1 3 3[4]2 3 3
                  2 2[4]1 2 2 3 3[5]1 2 1 1 2 2[3]1 2 1 3 2 3 1 3 1 1 2 1 2 3 1
                  3 1 1 2 1 2 3 1 3[13]1 4 4 2 3 1 3 1 1 2 1 2 1 3 1 3 2 1 2 1
                  1 3 1 3[3]1[10]2 4 2 1 3 1 3 2 1 2 1 3 1 3 1 1 2 1 2 1 3 1 3
                  2 1 2 1 1 2 1 2 2 1 2 1 3 3 2 1 3 1 3 2 1 2 1 1 3 1 3 2 1 2 1
                  3 1 3 1 2 1[3]2 1 2 1 1 2 1 2 1 6 2 3 1 3 1 2 1 2 1 1 2 2 1 3
                  1 1 3 1 2 2 1 2 2[4]1 2 1[4]2 4 5 2 3 1 3 1 2 1 2 1 2 1 1 2 1
                  3 3 1 2 1 1[3]2 1[3]2 1 2[4]1 3 5 2 1 3 1 3 1 2 1 2 1 2 2 1 3
                  1 1 3 2 1 1 2 1 2 2 1 1[3]2 1 1[3]2 6 2 1 3 1 3 1 2 1 2 2 1 1
                  2 1 3 3 1 1 2 2 1 1[4]2[3]1 2 2 1 1 2 5 3 1 2 2 1 3 1 1 3 3 1
                  3 1 2 1 2 1 1 2 2 1 2 1 2 1 2 1 2 2 1 2 2 1 3 6 3 1 2 2 1 3 1
                  1 3 1 3 1 3 1 2 1 2 2 1 1 2 2 1 2 2 1 2 1 1 2 1 1 2 1 6 3 2 1
                  1 2 1 3 3 1 3 1 3 1 2 1 2 1 2 1 1 2 1 2 2 1 2 2 1 2 1 2 1 2 4
                  5 3 2 1 1 2 1 3 3 1 1 3 1 3 1 2 1 2 1 2 2 1 1[3]2 1 1 2 1 2 1
                  2 1 4 4 3 1 2 2 1 1 3 3 1 1 2 2 1 1 3 3 1 3 1 3 1 2[3]1 2[5]1
                  2 2 1 3 3 1 2 2 1 1 3 3 1 2 1 1 2 3 1 1 3 1 3 1 3 2 1 1 2
                  1[5]2 1 1 2 3 3 2 1 1 2 3 1 1 3 1 2 2 1 1 3 3 1 1 3 1 3 1 2 1
                  1[3]2 1 2 1 1 2 3 4 3 2 1 1 2 3 1 1 3 2 1 1 2 3 1 1 3 3 1 3 1
                  1 2 1 2[3]1 2 1 2 2 1 4 1 1 3 3 1 1 3 3[4]1 2 2 1 1[4]2 3 3 1
                  1 2[4]1[3]2 1 2 1 2 1 3 3 1 1 3 3 1 1 2 2 1 1 2 2 1 1 3 3 2 2
                  1 1[5]2[3]1 2 1 4 2[3]1 3 3 1 1 3 3 1 1 2 2 1 1 2 2 3 3[5]2 1
                  1 2 2 1 1 2[7]1 3 3 1 1 3 3 2 2 1 1 2 2 1 1 2 2 3 3[5]2 1 1 2
                  2 1 2 2 3 4 1 3 3[4]1 3 3 2 2[4]3 2 2 1 1[4]2 1 1 2 1[3]2[3]1
                  2 3 1 3 3[4]1[4]3[4]2 3 3 2 2 1 1 2 2 1 2 1 2[3]1[3]2 3
                  3[3]1[4]3 1 1 2 2[4]3[4]2[6]1 2 2 1 1[3]2 1 2 4[3]1[4]3 1 1 3
                  3[4]2 3 3 1 1 2 2[3]1 2 1 1 2 2[3]1 2 1 1 2 1 2 1 2 2 3 2 3 1
                  2 1 2 2 3 2 3 1 2 1 2[12]1 4 2 2 1 2 1 2 2 3 2 3 2 1 2 1 3 2
                  3 2 2 1 2[4]1[13]2 1 2 1 3 2 3 2 1 2 1 2 2 3 2 3 2 1 2 1 2 1
                  2 1 1 2 1 2 2 1 2 1 3 1 2 2 1 2 1 3 2 3 2 2 1 2 1 3 2 3 2 1 2
                  1 2 2 1[3]2 1 2 1 1 2 1 2 1 4 2 1 2 1 2 3 2 3 2 2 3 3 2 1 2 2
                  1 2 3 3[3]2[4]1 2 1[4]2 4 3 2 1 2 1 2 3 2 3 2 3 2 2 3 2 1 1 2
                  3 2 2 3 2 2 1[3]2 1 2[4]1 3 3 2 2 1 2 1 2 3 2 3 2 3 3 2 1 2 2
                  1 3 2 2 3 1 2 2 1 1[3]2 1 1[3]2 4 2 2 1 2 1 2 3 2 3 3 2 2 3 2
                  1 1 2 2 3 3 2 1[4]2[3]1 2 2 1 1 2 3 3 2 3 3 2 1 2 2 1 1 2 1 2
                  3 2 3 2 2 3 3 2 2 1 2 1 2 1 2 2 1 2 2 1 3 4 3 2 3 3 2 1 2 2 1
                  2 1 2 1 2 3 2 3 3 2 2 3 2 1 2 2 1 2 1 1 2 1 1 2 1 4 3 3 2 2 3
                  2 1 1 2 1 2 1 2 3 2 3 2 3 2 2 3 1 2 2 1 2 2 1 2 1 2 1 2 4[3]3
                  2 2 3 2 1 1 2 2 1 2 1 2 3 2 3 2 3 3 2 1[3]2 1 1 2 1 2 1 2 1 4
                  2 3 2 3 3 2 2 1 1 2 2 3 3 2 2 1 1 2 1 2 1 2 2[3]1 2[5]1 2 2 1
                  1 3 2 3 3 2 2 1 1 2 3 2 2 3 1 2 2 1 2 1 2 1 2 1 1 2 1[5]2 1 1
                  2 1 3 3 2 2 3 1 2 2 1 2 3 3 2 2 1 1 2 2 1 2 1 1 2 1 1[3]2 1 2
                  1 1 2 3 2 3 3 2 2 3 1 2 2 1 3 2 2 3 1 2 2 1 1 2 1 2 1 2 1
                  2[3]1 2 1 2 2 1 4 5[3]1 2 2 1 1[4]2 3 3 2 2[4]3[4]1 2[4]1[3]2
                  1 2 1 6[3]1 2 2 1 1 2 2 3 3 2 2 3 3 2 2 1 1 3 3 1 1[5]2[3]1 2
                  1 4 6 1 2 2 1 1 2 2 1 1 2 2 3 3 2 2 3 3 1 1 3 3[3]2 1 1 2 2 1
                  1 2[3]1 5 1 2 2 1 1 2 2 1 1 3 3 2 2 3 3 2 2 3 3 1 1[5]2 1 1 2
                  2 1 2 2 3 2[3]1[4]2 1 1 3 3[4]1 3 3 2 2 3 3 2 2 1 1 2
                  1[3]2[3]1 2[4]1[4]2[4]1[4]3 1 1 3 3[4]2 1 2 1 2[3]1[3]2 3 1 1
                  2 2[4]1 2 2 3 3[4]1[4]3 2 2[4]1 2 2 1 1[3]2 1 2 2 1 2 2[4]1 2
                  2 1 1[4]3 1 1 2 2 3 3[3]1 2 1 1 2 2[3]1 2 1 5 2 2 3 2 3 3 1 3
                  1 2 3 2 3 3 1 3 1 2 3 2 3[12]1 4 6 2 2 3 2 3 3 1 3 1 3 2 3 2
                  1 3 1 3 3 2 3 2[3]1[10]2 6 2 3 2 3 2 1 3 1 3 2 3 2 3 3 1 3 1
                  3 2 3 2 2 1 2 1 1 2 1 2 2 1 2 1 3 5 2 3 2 3 2 1 3 1 3 3 2 3 2
                  1 3 1 3 2 3 2 3 2 1[3]2 1 2 1 1 2 1 2 1[3]2 3 2 3 1 3 1 3 3 1
                  1 3 2 3 3 2 3 1 1 3 2 2[4]1 2 1[4]2 4 1 2 2 3 2 3 1 3 1 3 1 3
                  3 1 3 2 2 3 1 3 3 1 2 2 1[3]2 1 2[4]1 3 1 2 3 2 3 2 3 1 3 1 3
                  1 1 3 2 3 3 2 1 3 3 1 1 2 2 1 1[3]2 1 1[5]2 3 2 3 2 3 1 3 1 1
                  3 3 1 3 2 2 3 3 1 1 3 1[4]2[3]1 2 2 1 1 2 1 3 3 1 1 3 2 3 3 2
                  2 3 2 3 1 3 1 3 3 1 1 3 2 1 2 1 2 1 2 2 1 2 2 1 3 2 3 3 1 1 3
                  2 3 3 2 3 2 3 2 3 1 3 1 1 3 3 1 2 1 2 2 1 2 1 1 2 1 1 2 1 2 3
                  1 3 3 1 3 2 2 3 2 3 2 3 1 3 1 3 1 3 3 1 1 2 2 1 2 2 1 2 1 2 1
                  2 4 1 3 1 3 3 1 3 2 2 3 3 2 3 2 3 1 3 1 3 1 1 3 1[3]2 1 1 2 1
                  2 1 2 1 4 6 3 3 1 1 3 3 2 2 3 3 1 1 3 3 2 2 3 2 3 2 3 2[3]1
                  2[5]1 2 2 1 5 3 3 1 1 3 3 2 2 3 1 3 3 1 2 3 3 2 3 2 3 2 2 1 1
                  2 1[5]2 1 1 2 5 3 1 3 3 1 2 3 3 2 3 1 1 3 3 2 2 3 3 2 3 2 1 2
                  1 1[3]2 1 2 1 1 2 3 6 3 1 3 3 1 2 3 3 2 1 3 3 1 2 3 3 2 2 3 2
                  3 1 2 1 2[3]1 2 1 2 2 1};

               else if m = 6 & l = 0 then
                  fac = { 1 3 2 2 3 3 2 2[4]3 1 1 3 3[4]1 2 2[4]1[5]2 1 2 1 4 2
                  2 3 3 2 2 3 3 1 1 3 3 1 1 3 3 2 2 1 1[4]2[5]1 2 1 2 4 3 3 2 2
                  3 3 2 2 3 3 1 1 3 3 1 1 2 2[4]1 2 2 1 2 2 1 2 1 1 2[3]3 2 2 3
                  3 2 2 1 1 3 3 1 1 3 3 1 1[4]2 1 1 2 1 1 2 1 2 2 3 6 2 2[4]3 2
                  2 1 1[4]2 1 1 3 3[3]1 2 1[3]2 1 2[3]1 3 5 2 2[4]3[4]2[4]1 2 2
                  1 1 3 3 2 1 2[3]1 2 1[3]2 4 5 3 3[4]2 3 3 1 1[4]2[4]1 3 3 1 2
                  2 1 1 2 1[3]2 1 4 6 3 3[4]2 3 3 2 2[4]1 2 2 3 3 1 1 2 1 1 2 2
                  1 2[3]1 2 5 3 3 1 3 1 1 2 1 2 3 1 3 1 1 2 1 2 3 1 3[12]1 5 4
                  3 1 3 1 1 2 1 2 1 3 1 3 2 1 2 1 1 3 1 3[11]2 6 4 1 3 1 3 2 1
                  2 1 3 1 3 1 1 2 1 2 1 3 1 3 1 1 2 1 2 1 2 2 1 2 1 6 3 1 3 1 3
                  2 1 2 1 1 3 1 3 2 1 2 1 3 1 3 1 2 2 1 2 1 2 1 1 2 1 2 7 6 3 1
                  3 1 2 1 2 1 1 2 2 1 3 1 1 3 1 2 2[4]1 2[3]1[4]2 7 5 3 1 3 1 2
                  1 2 1 2 1 1 2 1 3 3 1 2 1 1[4]2 1[3]2[4]1 8 5 1 3 1 3 1 2 1 2
                  1 2 2 1 3 1 1 3 2 1 1 2 1 1[4]2[3]1 2 2 8 6 1 3 1 3 1 2 1 2 2
                  1 1 2 1 3 3 1 1 2 2 1 2 2[4]1[3]2 1 1 9 5 1 2 2 1 3 1 1 3 3 1
                  3 1 2 1 2 1 1 2 2 1 1 2 1 2 2 1 2 1 2 2 1 9 6 1 2 2 1 3 1 1 3
                  1 3 1 3 1 2 1 2 2 1 1 2 2 1 2 1 1 2 1 2 1 1 2 10 6 2 1 1 2 1
                  3 3 1 3 1 3 1 2 1 2 1 2 1 1 2 1 2 2 1 2[3]1 2 1 2 10 5 2 1 1
                  2 1 3 3 1 1 3 1 3 1 2 1 2 1 2 2 1 2 1 1 2 1[3]2 1 2 1 11 4 1
                  2 2 1 1 3 3 1 1 2 2 1 1 3 3 1 3 1 3 1 1 2[3]1 2 2 1 1 2 2 11
                  3 1 2 2 1 1 3 3 1 2 1 1 2 3 1 1 3 1 3 1 3 2 1[3]2 1 1 2 2 1 1
                  12 3 2 1 1 2 3 1 1 3 1 2 2 1 1 3 3 1 1 3 1 3 1[3]2 1 1 2 2 1
                  1 2 12 4 2 1 1 2 3 1 1 3 2 1 1 2 3 1 1 3 3 1 3 1 2[3]1 2 2 1
                  1 2 2[3]1 3 3 1 1 3 3[4]1 2 2 1 1[4]2 3 3[4]1[5]2 1 2 1 2 3 3
                  1 1 3 3 1 1 2 2 1 1 2 2 1 1 3 3[6]2[5]1 2 1 2 2 1 1 3 3 1 1 3
                  3 1 1 2 2 1 1 2 2 3 3 2 2 1 1 2 2 1 2 2 1 2 1 1 2[3]1 3 3 1 1
                  3 3 2 2 1 1 2 2 1 1 2 2 3 3 2 2 1 1 2 1 1 2 1 2 2 3 4 3 3[4]1
                  3 3 2 2[4]3 2 2 1 1 2 2 1 2 1[3]2 1 2[3]1[4]3[4]1[4]3[4]2 3 3
                  2 2 1 1 2 1 2[3]1 2 1[3]2 4 3 1 1[4]3 1 1 2 2[4]3[4]2[3]1 2 2
                  1 1 2 1[3]2 1 4 4 1 1[4]3 1 1 3 3[4]2 3 3 1 1[3]2 1 1 2 2 1
                  2[3]1 2 5 1 1 2 1 2 2 3 2 3 1 2 1 2 2 3 2 3 1 2 1 2[11]1 5 2
                  1 2 1 2 2 3 2 3 2 1 2 1 3 2 3 2 2 1 2 1[11]2 6 2 2 1 2 1 3 2
                  3 2 1 2 1 2 2 3 2 3 2 1 2[3]1 2 1 2 1 2 2 1 2 1 6 1 2 1 2 1 3
                  2 3 2 2 1 2 1 3 2 3 2 1 2 1[3]2 1 2 1 2 1 1 2 1 2 7 4 1 2 1 2
                  3 2 3 2 2 3 3 2 1 2 2 1 2 3 3 2[3]1 2[3]1[4]2 7 3 1 2 1 2 3 2
                  3 2 3 2 2 3 2 1 1 2 3 2 2 3[3]2 1[3]2[4]1 8 3 2 1 2 1 2 3 2 3
                  2 3 3 2 1 2 2 1 3 2 2 3 1 1[4]2[3]1 2 2 8 4 2 1 2 1 2 3 2 3 3
                  2 2 3 2 1 1 2 2 3 3[3]2[4]1[3]2 1 1 9 3 2 3 3 2 1 2 2 1 1 2 1
                  2 3 2 3 2 2 3 3 2 1 2 1 2 2 1 2 1 2 2 1 9 4 2 3 3 2 1 2 2 1 2
                  1 2 1 2 3 2 3 3 2 2 3 2 1 2 1 1 2 1 2 1 1 2 10 4 3 2 2 3 2 1
                  1 2 1 2 1 2 3 2 3 2 3 2 2 3 1 2 2 1 2[3]1 2 1 2 10 3 3 2 2 3
                  2 1 1 2 2 1 2 1 2 3 2 3 2 3 3 2 2 1 1 2 1[3]2 1 2 1 11 2 2 3
                  3 2 2 1 1 2 2 3 3 2 2 1 1 2 1 2 1 2 1 2[3]1 2 2 1 1 2 2 11 1
                  2 3 3 2 2 1 1 2 3 2 2 3 1 2 2 1 2 1 2 1 2 1[3]2 1 1 2 2 1 1
                  12 1 3 2 2 3 1 2 2 1 2 3 3 2 2 1 1 2 2 1 2 1 1[3]2 1 1 2 2 1
                  1 2 12 2 3 2 2 3 1 2 2 1 3 2 2 3 1 2 2 1 1 2 1 2 2[3]1 2 2 1
                  1 2 2 1 1 5 1 1 2 2 1 1[4]2 3 3 2 2[4]3[6]1[5]2 1 2 1 6 1 1 2
                  2 1 1 2 2 3 3 2 2 3 3 2 2 1 1 3 3[4]2[5]1 2 1 2 6 2 2 1 1 2 2
                  1 1 2 2 3 3 2 2 3 3 1 1 3 3 1 1 2 2 1 2 2 1 2 1 1 2 5 2 2 1 1
                  2 2 1 1 3 3 2 2 3 3 2 2 3 3 1 1 2 2 1 1 2 1 1 2 1 2 2 3 2 1
                  1[4]2 1 1 3 3[4]1 3 3 2 2 3 3 1 2 1[3]2 1 2[3]1
                  3[3]1[4]2[4]1[4]3 1 1 3 3[3]2 1 2[3]1 2 1[3]2 4 1 2 2[4]1 2 2
                  3 3[4]1[4]3 2 2 1 2 2 1 1 2 1[3]2 1 4[3]2[4]1 2 2 1 1[4]3 1 1
                  2 2 3 3 2 1 1 2 2 1 2[3]1 2 5 5 2 3 2 3 3 1 3 1 2 3 2 3 3 1 3
                  1 2 3 2 3[11]1 5 6 2 3 2 3 3 1 3 1 3 2 3 2 1 3 1 3 3 2 3[12]2
                  6 6 3 2 3 2 1 3 1 3 2 3 2 3 3 1 3 1 3 2 3 2 1 1 2 1 2 1 2 2 1
                  2 1 6 5 3 2 3 2 1 3 1 3 3 2 3 2 1 3 1 3 2 3 2 3 2 2 1 2 1 2 1
                  1 2 1 2 7 2 2 3 2 3 1 3 1 3 3 1 1 3 2 3 3 2 3 1 1 3[3]1
                  2[3]1[4]2 7 1 2 3 2 3 1 3 1 3 1 3 3 1 3 2 2 3 1 3 3 1[3]2
                  1[3]2[4]1 8 1 3 2 3 2 3 1 3 1 3 1 1 3 2 3 3 2 1 3
                  3[3]1[4]2[3]1 2 2 8 2 3 2 3 2 3 1 3 1 1 3 3 1 3 2 2 3 3 1 1 3
                  2 2[4]1[3]2 1 1 9 1 3 1 1 3 2 3 3 2 2 3 2 3 1 3 1 3 3 1 1 3 1
                  2 1 2 2 1 2 1 2 2 1 9 2 3 1 1 3 2 3 3 2 3 2 3 2 3 1 3 1 1 3 3
                  1 2 1 2 1 1 2 1 2 1 1 2 10 2 1 3 3 1 3 2 2 3 2 3 2 3 1 3 1 3
                  1 3 3 1 1 2 2 1 2[3]1 2 1 2 10 1 1 3 3 1 3 2 2 3 3 2 3 2 3 1
                  3 1 3 1 1 3 2 1 1 2 1[3]2 1 2 1 11 6 3 1 1 3 3 2 2 3 3 1 1 3
                  3 2 2 3 2 3 2 3 1 2[3]1 2 2 1 1 2 2 11 5 3 1 1 3 3 2 2 3 1 3
                  3 1 2 3 3 2 3 2 3 2 2 1[3]2 1 1 2 2 1 1 12 5 1 3 3 1 2 3 3 2
                  3 1 1 3 3 2 2 3 3 2 3 2 1[3]2 1 1 2 2 1 1 2 12 6 1 3 3 1 2 3
                  3 2 1 3 3 1 2 3 3 2 2 3 2 3 2[3]1 2 2 1 1 2 2 1};

               else if m = 7 & l = 0 then
                  fac = { 4 5 3 1 2 2 3 3 2 2[4]3 1 1 3 3[4]1 2[3]1[3]2 1 2 1 6
                  4 1 2 2 3 3 2 2 3 3 1 1 3 3 1 1 3 3 1 1[4]2[3]1 2 1 4 6 4 1 3
                  3 2 2 3 3 2 2 3 3 1 1 3 3 1 1[3]2 1 1 2 1 1 2[3]1 5 3 1 3 3 2
                  2 3 3 2 2 1 1 3 3 1 1 3 3[5]2 1 2 2 1 2 2 3 3 6 1 2 2[4]3 2 2
                  1 1[4]2 1 1 2 2 1 1[4]2[3]1 2 4 5 1 2 2[4]3[4]2[4]1[4]2 1
                  2[4]1[3]2 3 4 5 1 3 3[4]2 3 3 1 1[4]2[6]1 2 1 1[3]2 1 2 3 6 1
                  3 3[4]2 3 3 2 2[4]1 2 2[3]1 2 1 2 2[3]1 2 1 3 3 2 3 1 3 1 1 2
                  1 2 3 1 3 1 1 2 1 2[11]1[3]4 2 3 1 3 1 1 2 1 2 1 3 1 3 2 1
                  2[4]1[9]2 4 4 2 1 3 1 3 2 1 2 1 3 1 3 1 1 2 1 2 2 1 2[3]1 2 2
                  1 2 1[3]3 2 1 3 1 3 2 1 2 1 1 3 1 3 2 1 2 1 2 1[4]2 1 1 2 1 2
                  1 5 6 2 3 1 3 1 2 1 2 1 1 2 2 1 3 1 1 3 2 2[3]1 2 1[4]2 4 6 5
                  2 3 1 3 1 2 1 2 1 2 1 1 2 1 3 3 1 2 2 1 2 2 1 2[4]1 3 6 5 2 1
                  3 1 3 1 2 1 2 1 2 2 1 3 1 1 3 1 2 2 1 1 2 2 1 1[3]2 5 6 2 1 3
                  1 3 1 2 1 2 2 1 1 2 1 3 3 1 1[4]2 1 1 2 2 1 1 2 5 5 3 1 2 2 1
                  3 1 1 3 3 1 3 1 2 1 2 1 2 1 2 1[3]2 1 2 2 1 3 6 6 3 1 2 2 1 3
                  1 1 3 1 3 1 3 1 2 1 2 2 1 2 2[3]1 2 1 1 2 1 6 6 3 2 1 1 2 1 3
                  3 1 3 1 3 1 2 1 2 1 1 2 2 1 2 1 2 1 2 1 2 4 5 5 3 2 1 1 2 1 3
                  3 1 1 3 1 3 1 2 1 2 1[3]2 1 2 1 2 1 2 1 4 3 4 3 1 2 2 1 1 3 3
                  1 1 2 2 1 1 3 3 1 2[3]1 2[4]1 2 2 1 4 3 3 1 2 2 1 1 3 3 1 2 1
                  1 2 3 1 1 3 2 1 1 2 1[4]2 1 1 2 4 3 3 2 1 1 2 3 1 1 3 1 2 2 1
                  1 3 3 1 1 2 1 1 2 2 1 2 1 1 2 3 3 4 3 2 1 1 2 3 1 1 3 2 1 1 2
                  3 1 1 3 1 2 1 2 1 1 2 1 2 2 1 4[3]1 3 3 1 1 3 3[4]1 2 2 1 1 2
                  2 1 1 2[3]1[3]2 1 2 1 2 2 1 3 3 1 1 3 3 1 1 2 2 1 1 2
                  2[4]1[4]2[3]1 2 1 4 2 2[3]1 3 3 1 1 3 3 1 1 2 2 1 1[5]2 1 1 2
                  1 1 2[8]1 3 3 1 1 3 3 2 2 1 1 2 2 1 1[5]2 1 2 2 1 2 2 3 5 4 1
                  3 3[4]1 3 3 2 2[4]3[4]2 1 1[4]2[3]1 2 6 3 1 3 3[4]1[4]3[4]2 3
                  3 2 2 1 2[4]1[3]2 3 6 3[3]1[4]3 1 1 2 2[4]3 2 2[4]1 2 1 1[3]2
                  1 2 5 4[3]1[4]3 1 1 3 3[4]2 3 3[3]1 2 1 2 2[3]1 2 1 5 1 2 1 2
                  1 2 2 3 2 3 1 2 1 2 2 3 2 3[11]1 4 6 2 2 1 2 1 2 2 3 2 3 2 1
                  2 1 3 2 3 2[3]1[9]2 6[3]2 1 2 1 3 2 3 2 1 2 1 2 2 3 2 3 2 1
                  2[3]1 2 2 1 2 1 3 5 1 2 2 1 2 1 3 2 3 2 2 1 2 1 3 2 3 2 2
                  1[4]2 1 1 2 1 2 1 1 4 2 1 2 1 2 3 2 3 2 2 3 3 2 1 2 2 1 2
                  2[3]1 2 1[4]2 4 2 3 2 1 2 1 2 3 2 3 2 3 2 2 3 2 1 1[3]2 1 2 2
                  1 2[4]1 3 2 3 2 2 1 2 1 2 3 2 3 2 3 3 2 1 2 2 1 1 2 2 1 1 2 2
                  1 1[3]2 1 4 2 2 1 2 1 2 3 2 3 3 2 2 3 2 1 1 2 1[4]2 1 1 2 2 1
                  1 2 1 3 3 2 3 3 2 1 2 2 1 1 2 1 2 3 2 3 2 2 1 2 1[3]2 1 2 2 1
                  3 2 4 3 2 3 3 2 1 2 2 1 2 1 2 1 2 3 2 3 2 1 2 2[3]1 2 1 1 2 1
                  2 4 3 3 2 2 3 2 1 1 2 1 2 1 2 3 2 3 2 1 2 2 1 2 1 2 1 2 1 2 4
                  1[3]3 2 2 3 2 1 1 2 2 1 2 1 2 3 2 3 1[3]2 1 2 1 2 1 2 1 4 5 2
                  3 2 3 3 2 2 1 1 2 2 3 3 2 2 1 1 2 2[3]1 2[4]1 2 2 1 6 1 3 2 3
                  3 2 2 1 1 2 3 2 2 3 1 2 2 1 2 1 1 2 1[4]2 1 1 2 6 1 3 3 2 2 3
                  1 2 2 1 2 3 3 2 2 1 1 2 1 2 1 1 2 2 1 2 1 1 2 3 5 2 3 3 2 2 3
                  1 2 2 1 3 2 2 3 1 2 2 1 1 2 1 2 1 1 2 1 2 2 1 4 3 5[3]1 2 2 1
                  1[4]2 3 3 2 2 3 3 1 1 2[3]1[3]2 1 2 1 4 6[3]1 2 2 1 1 2 2 3 3
                  2 2 3 3 2 2 1 1[4]2[3]1 2 1 4 4 6 1 2 2 1 1 2 2 1 1 2 2 3 3 2
                  2 3 3[3]2 1 1 2 1 1 2[3]1 3 5 1 2 2 1 1 2 2 1 1 3 3 2 2 3
                  3[7]2 1 2 2 1 2 2 3 1 2[3]1[4]2 1 1 3 3[4]1 3 3 2 2 1
                  1[4]2[3]1 2 2[4]1[4]2[4]1[4]3 1 1 2 2 1 2[4]1[3]2 3 2 1 1 2
                  2[4]1 2 2 3 3[4]1 3 3[4]1 2 1 1[3]2 1 2 1 2 1 2 2[4]1 2 2 1
                  1[4]3[5]1 2 1 2 2[3]1 2 1 1 5 2 2 3 2 3 3 1 3 1 2 3 2 3 3 1
                  3[12]1 4 2 6 2 2 3 2 3 3 1 3 1 3 2 3 2 1 3 1 3[3]1[10]2 6 2 3
                  2 3 2 1 3 1 3 2 3 2 3 3 1 3 1 2 1 2[3]1 2 2 1 2 1 3 1 5 2 3 2
                  3 2 1 3 1 3 3 2 3 2 1 3 1 3 2 1[4]2 1 1 2 1 2 1 3[3]2 3 2 3 1
                  3 1 3 3 1 1 3 2 3 3[3]2[3]1 2 1[4]2 4 4 1 2 2 3 2 3 1 3 1 3 1
                  3 3 1 3 2 2 3 2 2 1 2 2 1 2[4]1 3 4 1 2 3 2 3 2 3 1 3 1 3 1 1
                  3 2 3 3 2 1 2 2 1 1 2 2 1 1[3]2 3 2 2 3 2 3 2 3 1 3 1 1 3 3 1
                  3 2 2 3 1[4]2 1 1 2 2 1 1 2 3 1 3 3 1 1 3 2 3 3 2 2 3 2 3 1 3
                  1 3 2 1 2 1[3]2 1 2 2 1 3 4 2 3 3 1 1 3 2 3 3 2 3 2 3 2 3 1 3
                  1 2 1 2 2[3]1 2 1 1 2 1 4 2 3 1 3 3 1 3 2 2 3 2 3 2 3 1 3 1 3
                  1 2 2 1 2 1 2 1 2 1 2 4 3 1 3 1 3 3 1 3 2 2 3 3 2 3 2 3 1 3 1
                  1[3]2 1 2 1 2 1 2 1 4 1 6 3 3 1 1 3 3 2 2 3 3 1 1 3 3 2 2 3
                  2[3]1 2[4]1 2 2 1 2 5 3 3 1 1 3 3 2 2 3 1 3 3 1 2 3 3 2 2 1 1
                  2 1[4]2 1 1 2 2 5 3 1 3 3 1 2 3 3 2 3 1 1 3 3 2 2 3 1 2 1 1 2
                  2 1 2 1 1 2 3 1 6 3 1 3 3 1 2 3 3 2 1 3 3 1 2 3 3 2 1 2 1 2 1
                  1 2 1 2 2 1};

               else if m = 8 & l = 0 then
                  fac = { 4 1 3 2 2 3 3 2 2[4]3 1 1 3 3[4]1 2 2[5]1[3]2 1 2 1 1
                  4 2 2 3 3 2 2 3 3 1 1 3 3 1 1 3 3 2 2[3]1[4]2[3]1 2 1 4 2 4 3
                  3 2 2 3 3 2 2 3 3 1 1 3 3 1 1 2 2 1 1 2 1 1 2 2 1 1 2[3]1
                  2[3]3 2 2 3 3 2 2 1 1 3 3 1 1 3 3 1 1[5]2 1 1 2 2 1 2 2 3 1 6
                  2 2[4]3 2 2 1 1[4]2 1 1 3 3 1 1 2 1 2 1[3]2[3]1 2 1 5 2
                  2[4]3[4]2[4]1 2 2 1 1 3 3 2 2 1 2[3]1[3]2 3 2 5 3 3[4]2 3 3 1
                  1[4]2[4]1 3 3 1 1 2 2 1 1[3]2 1 2 2 6 3 3[4]2 3 3 2 2[4]1 2 2
                  3 3[3]1 2 1 1 2 2[3]1 2 1[3]3 1 3 1 1 2 1 2 3 1 3 1 1 2 1 2 3
                  1 3[11]1 4 3 4 3 1 3 1 1 2 1 2 1 3 1 3 2 1 2 1 1 3 1 3 1[10]2
                  3 4 1 3 1 3 2 1 2 1 3 1 3 1 1 2 1 2 1 3 1 3 2 1 1 2 1 2 2 1 2
                  1[3]3 1 3 1 3 2 1 2 1 1 3 1 3 2 1 2 1 3 1 3 1[3]2 1 2 1 1 2 1
                  2 1 4 6 3 1 3 1 2 1 2 1 1 2 2 1 3 1 1 3 1 2 2 1 2[3]1 2 1[4]2
                  4 4 5 3 1 3 1 2 1 2 1 2 1 1 2 1 3 3 1 2 1 1[5]2 1 2[4]1 3 4 5
                  1 3 1 3 1 2 1 2 1 2 2 1 3 1 1 3 2 1 1 2[3]1[3]2 1 1[3]2 4 6 1
                  3 1 3 1 2 1 2 2 1 1 2 1 3 3 1 1 2 2 1 1 2 2[3]1 2 2 1 1 2 5 5
                  1 2 2 1 3 1 1 3 3 1 3 1 2 1 2 1 1 2 2 1 2 1 2 1 2 2 1 2 2 1 3
                  5 6 1 2 2 1 3 1 1 3 1 3 1 3 1 2 1 2 2 1 1[3]2 1 2 1 1 2 1 1 2
                  1 5 6 2 1 1 2 1 3 3 1 3 1 3 1 2 1 2 1 2 1 1 2 1 1 2 2 1 2 1 2
                  1 2 4 5 5 2 1 1 2 1 3 3 1 1 3 1 3 1 2 1 2 1 2 2 1 1 2 1 1 2 1
                  2 1 2 1 4 6 4 1 2 2 1 1 3 3 1 1 2 2 1 1 3 3 1 3 1 3 1 2 1
                  2[5]1 2 2 1 6 3 1 2 2 1 1 3 3 1 2 1 1 2 3 1 1 3 1 3 1 3 2 2
                  1[5]2 1 1 2 6 3 2 1 1 2 3 1 1 3 1 2 2 1 1 3 3 1 1 3 1 3 1
                  1[3]2 1 2 1 1 2 3 6 4 2 1 1 2 3 1 1 3 2 1 1 2 3 1 1 3 3 1 3 1
                  1 2[3]1 2 1 2 2 1 4 1 1 3 3 1 1 3 3[4]1 2 2 1 1[4]2 3
                  3[5]1[3]2 1 2 1 1 2 3 3 1 1 3 3 1 1 2 2 1 1 2 2 1 1 3 3 2 2
                  1[4]2[3]1 2 1 4 2 2 1 1 3 3 1 1 3 3 1 1 2 2 1 1 2 2 3 3[3]2 1
                  1 2 2 1 1 2[3]1 2[3]1 3 3 1 1 3 3 2 2 1 1 2 2 1 1 2 2 3 3[3]2
                  1 1 2 2 1 2 2 3 1 4 3 3[4]1 3 3 2 2[4]3 2 2 1 1[3]2 1 2
                  1[3]2[3]1 2 1[3]3[4]1[4]3[4]2 3 3 2 2 1 1 2 2 1 2[3]1[3]2 3 2
                  3 1 1[4]3 1 1 2 2[4]3[4]2[4]1 2 2 1 1[3]2 1 2 2 4 1 1[4]3 1 1
                  3 3[4]2 3 3 1 1 2 2 1 2 1 1 2 2[3]1 2 1 3 1 1 2 1 2 2 3 2 3 1
                  2 1 2 2 3 2 3 1 2 1 2[10]1 4 3 2 1 2 1 2 2 3 2 3 2 1 2 1 3 2
                  3 2 2 1 2 1 1[10]2 3 2 2 1 2 1 3 2 3 2 1 2 1 2 2 3 2 3 2 1 2
                  1 2 1 1 2 1 2 2 1 2 1 3 3 1 2 1 2 1 3 2 3 2 2 1 2 1 3 2 3 2 1
                  2 1[4]2 1 2 1 1 2 1 2 1 4 4 1 2 1 2 3 2 3 2 2 3 3 2 1 2 2 1 2
                  3 3 2 2[3]1 2 1[4]2 4 4 3 1 2 1 2 3 2 3 2 3 2 2 3 2 1 1 2 3 2
                  2 3[4]2 1 2[4]1 3 4 3 2 1 2 1 2 3 2 3 2 3 3 2 1 2 2 1 3 2 2
                  3[3]1[3]2 1 1[3]2 4 4 2 1 2 1 2 3 2 3 3 2 2 3 2 1 1 2 2 3 3 2
                  1 2 2[3]1 2 2 1 1 2 5 3 2 3 3 2 1 2 2 1 1 2 1 2 3 2 3 2 2 3 3
                  2 2 1 2 1 2 2 1 2 2 1 3 5 4 2 3 3 2 1 2 2 1 2 1 2 1 2 3 2 3 3
                  2 2 3 2 2 1 2 1 1 2 1 1 2 1 5 4 3 2 2 3 2 1 1 2 1 2 1 2 3 2 3
                  2 3 2 2 3 1 1 2 2 1 2 1 2 1 2 4 5 3 3 2 2 3 2 1 1 2 2 1 2 1 2
                  3 2 3 2 3 3 2 1 2 1 1 2 1 2 1 2 1 4 6 2 2 3 3 2 2 1 1 2 2 3 3
                  2 2 1 1 2 1 2 1 2 2 1 2[5]1 2 2 1 6 1 2 3 3 2 2 1 1 2 3 2 2 3
                  1 2 2 1 2 1 2 1 2 2 1[5]2 1 1 2 6 1 3 2 2 3 1 2 2 1 2 3 3 2 2
                  1 1 2 2 1 2[3]1[3]2 1 2 1 1 2 3 6 2 3 2 2 3 1 2 2 1 3 2 2 3 1
                  2 2 1 1 2 1 2 1 2[3]1 2 1 2 2 1 4 1 5 1 1 2 2 1 1[4]2 3 3 2
                  2[4]3[7]1[3]2 1 2 1 1 6 1 1 2 2 1 1 2 2 3 3 2 2 3 3 2 2 1 1 3
                  3 1[4]2[3]1 2 1 4 2 6 2 2 1 1 2 2 1 1 2 2 3 3 2 2 3 3 1 1 3 3
                  2 1 1 2 2 1 1 2[3]1 2 5 2 2 1 1 2 2 1 1 3 3 2 2 3 3 2 2 3 3 1
                  1[3]2 1 1 2 2 1 2 2 3 1 2 1 1[4]2 1 1 3 3[4]1 3 3 2 2 3 3 2 1
                  2 1[3]2[3]1 2[4]1[4]2[4]1[4]3 1 1 3 3[4]2 1 2[3]1[3]2 3 2 1 2
                  2[4]1 2 2 3 3[4]1[4]3 2 2 1 1 2 2 1 1[3]2 1[5]2[4]1 2 2 1
                  1[4]3 1 1 2 2 3 3 1 2 1 1 2 2[3]1 2 1 3 5 2 3 2 3 3 1 3 1 2 3
                  2 3 3 1 3 1 2 3 2 3[10]1 4 3 6 2 3 2 3 3 1 3 1 3 2 3 2 1 3 1
                  3 3 2 3 2 1[10]2 3 6 3 2 3 2 1 3 1 3 2 3 2 3 3 1 3 1 3 2 3 2
                  2 1 1 2 1 2 2 1 2 1 3 3 5 3 2 3 2 1 3 1 3 3 2 3 2 1 3 1 3 2 3
                  2 3[3]2 1 2 1 1 2 1 2 1 4 2 2 3 2 3 1 3 1 3 3 1 1 3 2 3 3 2 3
                  1 1 3 2[3]1 2 1[4]2 4 4 1 2 3 2 3 1 3 1 3 1 3 3 1 3 2 2 3 1 3
                  3 1[4]2 1 2[4]1 3 4 1 3 2 3 2 3 1 3 1 3 1 1 3 2 3 3 2 1 3
                  3[4]1[3]2 1 1[3]2 4 2 3 2 3 2 3 1 3 1 1 3 3 1 3 2 2 3 3 1 1 3
                  1 2 2[3]1 2 2 1 1 2 5 1 3 1 1 3 2 3 3 2 2 3 2 3 1 3 1 3 3 1 1
                  3 2 1 2 1 2 2 1 2 2 1 3 5 2 3 1 1 3 2 3 3 2 3 2 3 2 3 1 3 1 1
                  3 3 1 2 2 1 2 1 1 2 1 1 2 1 5 2 1 3 3 1 3 2 2 3 2 3 2 3 1 3 1
                  3 1 3 3[3]1 2 2 1 2 1 2 1 2 4 5 1 1 3 3 1 3 2 2 3 3 2 3 2 3 1
                  3 1 3 1 1 3 1 2 1 1 2 1 2 1 2 1 4 6 6 3 1 1 3 3 2 2 3 3 1 1 3
                  3 2 2 3 2 3 2 3 2 1 2[5]1 2 2 1 6 5 3 1 1 3 3 2 2 3 1 3 3 1 2
                  3 3 2 3 2 3[3]2 1[5]2 1 1 2 6 5 1 3 3 1 2 3 3 2 3 1 1 3 3 2 2
                  3 3 2 3 2 1 1[3]2 1 2 1 1 2 3 6 6 1 3 3 1 2 3 3 2 1 3 3 1 2 3
                  3 2 2 3 2 3 1 2[3]1 2 1 2 2 1};

               else if m = 9 & l = 0 then
                  fac = { 1 5 3 2 2 3 3 2 2[4]3 1 1 3 3[5]1[5]2 1 2 1 6 4 2 2 3
                  3 2 2 3 3 1 1 3 3 1 1 3 3[3]2[5]1 2 1 2 6 4 3 3 2 2 3 3 2 2 3
                  3 1 1 3 3[4]1 2 1 2 2 1 2 1 1 2 5[3]3 2 2 3 3 2 2 1 1 3 3 1 1
                  3 3 2 2 1 2 1 1 2 1 2 2 3 3 6 2 2[4]3 2 2 1 1[4]2[3]1[4]2 1
                  2[3]1 3 4 5 2 2[4]3[4]2[4]1[3]2[4]1 2 1[3]2 4 4 5 3 3[4]2 3 3
                  1 1[4]2[3]1 2 1 1 2 1[3]2 1 4 3 6 3 3[4]2 3 3 2 2[4]1[3]2 1 2
                  2 1 2[3]1 2 5[3]3 1 3 1 1 2 1 2 3 1 3 1 1 2 1 2[10]1 5 4 4 3
                  1 3 1 1 2 1 2 1 3 1 3 2 1 2 1[10]2 6 4 4 1 3 1 3 2 1 2 1 3 1
                  3 1 1 2 1 2[3]1 2 1 2 2 1 2 1 6 3 3 1 3 1 3 2 1 2 1 1 3 1 3 2
                  1 2 1[3]2 1 2 1 1 2 1 2 7 5 6 3 1 3 1 2 1 2 1 1 2 2 1 3 1 1 3
                  1 1 2[3]1[4]2 7 6 5 3 1 3 1 2 1 2 1 2 1 1 2 1 3 3 1 2 2
                  1[3]2[4]1 8 6 5 1 3 1 3 1 2 1 2 1 2 2 1 3 1 1 3 1 1[3]2[3]1 2
                  2 8 5 6 1 3 1 3 1 2 1 2 2 1 1 2 1 3 3 1 2 2[3]1[3]2 1 1 9 5 5
                  1 2 2 1 3 1 1 3 3 1 3 1 2 1 2 1 1[3]2 1 2 1 2 2 1 9 6 6 1 2 2
                  1 3 1 1 3 1 3 1 3 1 2 1 2 2[3]1 2 1 2 1 1 2 10 6 6 2 1 1 2 1
                  3 3 1 3 1 3 1 2 1 2 1 1 2 1 2[3]1 2 1 2 10 5 5 2 1 1 2 1 3 3
                  1 1 3 1 3 1 2 1 2 2 1 2 1[3]2 1 2 1 11 3 4 1 2 2 1 1 3 3 1 1
                  2 2 1 1 3 3 1 1 2 1 1 2 2 1 1 2 2 11 4 3 1 2 2 1 1 3 3 1 2 1
                  1 2 3 1 1 3 2 1 2 2 1 1 2 2 1 1 12 4 3 2 1 1 2 3 1 1 3 1 2 2
                  1 1 3 3 1 1 2 2 1 1 2 2 1 1 2 12 3 4 2 1 1 2 3 1 1 3 2 1 1 2
                  3 1 1 3 2 1 1 2 2 1 1 2 2[4]1 3 3 1 1 3 3[4]1 2 2 1 1 2
                  2[3]1[5]2 1 2 1 2 2 3 3 1 1 3 3 1 1 2 2 1 1 2 2 1 1[3]2[5]1 2
                  1[3]2 1 1 3 3 1 1 3 3 1 1 2 2 1 1 2 2 1 1 2 1 2 2 1 2 1 1
                  2[4]1 3 3 1 1 3 3 2 2 1 1 2 2 1 1 2 2 1 2 1 1 2 1 2 2 3 5 4 3
                  3[4]1 3 3 2 2[4]3 2 2 1[4]2 1 2[3]1 3 6[3]3[4]1[4]3[4]2 3 3
                  2[4]1 2 1[3]2 4 6 3 1 1[4]3 1 1 2 2[4]3 2 2 1 2 1 1 2 1[3]2 1
                  4 5 4 1 1[4]3 1 1 3 3[4]2 3 3 2 1 2 2 1 2[3]1 2 5 5 1 1 2 1 2
                  2 3 2 3 1 2 1 2 2 3 2 3[10]1 5 6 2 1 2 1 2 2 3 2 3 2 1 2 1 3
                  2 3[11]2 6 6 2 2 1 2 1 3 2 3 2 1 2 1 2 2 3 2 3[3]1 2 1 2 2 1
                  2 1 6 5 1 2 1 2 1 3 2 3 2 2 1 2 1 3 2 3[4]2 1 2 1 1 2 1 2 7 1
                  4 1 2 1 2 3 2 3 2 2 3 3 2 1 2 2[3]1 2[3]1[4]2 7 2 3 1 2 1 2 3
                  2 3 2 3 2 2 3 2 1 1[3]2 1[3]2[4]1 8 2 3 2 1 2 1 2 3 2 3 2 3 3
                  2 1 2 2[3]1[3]2[3]1 2 2 8 1 4 2 1 2 1 2 3 2 3 3 2 2 3 2 1
                  1[3]2[3]1[3]2 1 1 9 1 3 2 3 3 2 1 2 2 1 1 2 1 2 3 2 3 2 1[3]2
                  1 2 1 2 2 1 9 2 4 2 3 3 2 1 2 2 1 2 1 2 1 2 3 2 3 2[3]1 2 1 2
                  1 1 2 10 2 4 3 2 2 3 2 1 1 2 1 2 1 2 3 2 3 2 1 2 1 2[3]1 2 1
                  2 10 1 3 3 2 2 3 2 1 1 2 2 1 2 1 2 3 2 3 2 1 2 1[3]2 1 2 1 11
                  5 2 2 3 3 2 2 1 1 2 2 3 3 2 2 1 1 2 1 2 1 1 2 2 1 1 2 2 11 6
                  1 2 3 3 2 2 1 1 2 3 2 2 3 1 2 2 1 2 1 2 2 1 1 2 2 1 1 12 6 1
                  3 2 2 3 1 2 2 1 2 3 3 2 2 1 1 2 1 2 2 1 1 2 2 1 1 2 12 5 2 3
                  2 2 3 1 2 2 1 3 2 2 3 1 2 2 1 2 1 1 2 2 1 1 2 2 1 1 3 5 1 1 2
                  2 1 1[4]2 3 3 2 2 3 3[3]1[5]2 1 2 1 4 6 1 1 2 2 1 1 2 2 3 3 2
                  2 3 3[5]2[5]1 2 1 2 4 6 2 2 1 1 2 2 1 1 2 2 3 3 2 2 3 3 1 1 2
                  1 2 2 1 2 1 1 2 3 5 2 2 1 1 2 2 1 1 3 3 2 2 3 3[4]2 1 2 1 1 2
                  1 2 2 3 1 2 1 1[4]2 1 1 3 3[4]1 3 3 1[4]2 1 2[3]1 3
                  2[3]1[4]2[4]1[4]3 1 1 2[4]1 2 1[3]2 4 2 1 2 2[4]1 2 2 3 3[4]1
                  3 3 1 2 1 1 2 1[3]2 1 4 1[3]2[4]1 2 2 1 1[4]3 1 1 2 1 2 2 1
                  2[3]1 2 5 1 5 2 3 2 3 3 1 3 1 2 3 2 3 3 1 3[11]1 5 2 6 2 3 2
                  3 3 1 3 1 3 2 3 2 1 3 1 3[10]2 6 2 6 3 2 3 2 1 3 1 3 2 3 2 3
                  3 1 3[4]1 2 1 2 2 1 2 1 6 1 5 3 2 3 2 1 3 1 3 3 2 3 2 1 3 1
                  3[3]2 1 2 1 1 2 1 2 7 3 2 2 3 2 3 1 3 1 3 3 1 1 3 2 3 3 2 1 1
                  2[3]1[4]2 7 4 1 2 3 2 3 1 3 1 3 1 3 3 1 3 2 2 3 2 2 1[3]2[4]1
                  8 4 1 3 2 3 2 3 1 3 1 3 1 1 3 2 3 3 2 1 1[3]2[3]1 2 2 8 3 2 3
                  2 3 2 3 1 3 1 1 3 3 1 3 2 2 3 2 2[3]1[3]2 1 1 9 3 1 3 1 1 3 2
                  3 3 2 2 3 2 3 1 3 1 3 1[3]2 1 2 1 2 2 1 9 4 2 3 1 1 3 2 3 3 2
                  3 2 3 2 3 1 3 1 2[3]1 2 1 2 1 1 2 10 4 2 1 3 3 1 3 2 2 3 2 3
                  2 3 1 3 1 3 1 2 1 2[3]1 2 1 2 10 3 1 1 3 3 1 3 2 2 3 3 2 3 2
                  3 1 3 1 2 1 2 1[3]2 1 2 1 11 1 6 3 1 1 3 3 2 2 3 3 1 1 3 3 2
                  2 3 1 2 1 1 2 2 1 1 2 2 11 2 5 3 1 1 3 3 2 2 3 1 3 3 1 2 3 3
                  2 2 1 2 2 1 1 2 2 1 1 12 2 5 1 3 3 1 2 3 3 2 3 1 1 3 3 2 2 3
                  1 2 2 1 1 2 2 1 1 2 12 1 6 1 3 3 1 2 3 3 2 1 3 3 1 2 3 3 2 2
                  1 1 2 2 1 1 2 2 1};

               else if m = 10 & l = 0 then
                  fac = { 4 3 5 3 1 2 2 3 3 2 2[4]3[4]1 2 1 1[3]2 1 2 1 4 6 4 1
                  2 2 3 3 2 2 3 3 1 1 3 3 1 1[3]2[3]1 2 1 4 3 6 4 1 3 3 2 2 3 3
                  2 2 3 3 1 1[3]2 1 2 1 1 2[3]1 4 5 3 1 3 3 2 2 3 3 2 2 1 1 3
                  3[4]2 1 2 2 1 2 2 3 6 3 6 1 2 2[4]3 2 2 1 1[4]2 1 1[3]2[3]1 2
                  5 4 5 1 2 2[4]3[4]2 1 1 2 2 1 2[3]1[3]2 3 6 4 5 1 3 3[4]2 3 3
                  1 1 2 2[6]1[3]2 1 2 5 3 6 1 3 3[4]2 3 3 2 2[5]1[3]2[3]1 2 1 5
                  3 3 2 3 1 3 1 1 2 1 2 3 1 3[11]1 4 6 4 4 2 3 1 3 1 1 2 1 2 1
                  3 1 3[3]1[8]2 5 4 4 2 1 3 1 3 2 1 2 1 3 1 3 1 2 1 2 1 1 2 2 1
                  2 1 3 6 3 3 2 1 3 1 3 2 1 2 1 1 3 1 3 2 1[3]2 1 1 2 1 2 1 3 5
                  6 2 3 1 3 1 2 1 2 1 1 2 2 1 2 2 1 1 2 1[4]2 4 4 6 5 2 3 1 3 1
                  2 1 2 1 2 1 1[3]2 1 2 1 2[4]1 3 3 6 5 2 1 3 1 3 1 2 1 2 1 2 2
                  1 1 2 2 1 2 2 1 1[3]2 4 5 6 2 1 3 1 3 1 2 1 2 2 1 1 2 1[3]2 1
                  1 2 2 1 1 2 6 5 5 3 1 2 2 1 3 1 1 3 3 1 3 1 2 1 2 1 2 2 1 2 2
                  1 3 5 6 6 3 1 2 2 1 3 1 1 3 1 3 1 3 2 1 2 2 1 1 2 1 1 2 1[3]6
                  3 2 1 1 2 1 3 3 1 3 1 3 1 1 2 2 1 1 2 1 2 1 2 4[3]5 3 2 1 1 2
                  1 3 3 1 1 3 1 3 1[4]2 1 2 1 2 1 4 4 3 4 3 1 2 2 1 1 3 3 1 1 2
                  2 1 2[7]1 2 2 1 3 4 3 3 1 2 2 1 1 3 3 1 2 1 1 2 2 1 1[5]2 1 1
                  2 4 4 3 3 2 1 1 2 3 1 1 3 1 2 2 1 1 2 1 1 2 1 2 1 1 2[3]3 4 3
                  2 1 1 2 3 1 1 3 2 1 1 2 1 2 1 2 1 2 1 2 2 1 4 5[3]1 3 3 1 1 3
                  3[4]1 2 2 1 1 2 1 1[3]2 1 2 1 6 2 2 1 3 3 1 1 3 3 1 1 2
                  2[4]1[3]2[3]1 2 1 4 5 2 2[3]1 3 3 1 1 3 3 1 1[5]2 1 2 1 1
                  2[3]1 6[5]1 3 3 1 1 3 3 2 2 1 1[4]2 1 2 2 1 2 2 3 2 5 4 1 3
                  3[4]1 3 3 2 2 3 3 2 2 1 1[3]2[3]1 2 1 6 3 1 3 3[4]1[4]3[4]2 1
                  2[3]1[3]2 3 2 6 3[3]1[4]3 1 1 2 2 3 3[6]1[3]2 1 2 1 5
                  4[3]1[4]3 1 1 3 3 2 2[3]1[3]2[3]1 2 1 1 5 1 2 1 2 1 2 2 3 2 3
                  1 2 1 2[10]1 4 2 6 2 2 1 2 1 2 2 3 2 3 2 1 2[4]1[8]2 1 6[3]2
                  1 2 1 3 2 3 2 1 2 1 2 2 1 2 1 1 2 2 1 2 1 3 2 5 1 2 2 1 2 1 3
                  2 3 2 2 1 2 1 2 1[3]2 1 1 2 1 2 1 5 1 4 2 1 2 1 2 3 2 3 2 2 3
                  3[3]2 1 1 2 1[4]2 4 6 2 3 2 1 2 1 2 3 2 3 2 3 2 2 3 2 2 1 2 1
                  2[4]1 3 5 2 3 2 2 1 2 1 2 3 2 3 2 3 3 2 1 2 2 1 2 2 1 1[3]2 6
                  1 4 2 2 1 2 1 2 3 2 3 3 2 2 3 1[3]2 1 1 2 2 1 1 2 2 1 3 3 2 3
                  3 2 1 2 2 1 1 2 1 2 2 1 2 1 2 2 1 2 2 1 3 1 2 4 3 2 3 3 2 1 2
                  2 1 2 1 2 1 2 1 2 2 1 1 2 1 1 2 1 2 2 4 3 3 2 2 3 2 1 1 2 1 2
                  1 2 1 2 2 1 1 2 1 2 1 2 4 1 1[3]3 2 2 3 2 1 1 2 2 1 2 1 1[4]2
                  1 2 1 2 1 4 6 5 2 3 2 3 3 2 2 1 1 2 2 3 3 2 2[7]1 2 2 1 5 6 1
                  3 2 3 3 2 2 1 1 2 3 2 2 3 2 1 1[5]2 1 1 2 6 6 1 3 3 2 2 3 1 2
                  2 1 2 3 3 2 1 2 1 1 2 1 2 1 1 2 3 5 5 2 3 3 2 2 3 1 2 2 1 3 2
                  2 3 1 2 1 2 1 2 1 2 2 1 4 1 3 5[3]1 2 2 1 1[4]2 3 3 1 1 2 1
                  1[3]2 1 2 1 2 4 6[3]1 2 2 1 1 2 2 3 3 2 2 1 1[3]2[3]1 2 1 4 1
                  4 6 1 2 2 1 1 2 2 1 1 2 2 3 3[3]2 1 2 1 1 2[3]1 2 3 5 1 2 2 1
                  1 2 2 1 1 3 3[6]2 1 2 2 1 2 2 3 4 1 2[3]1[4]2 1 1 3 3 1 1 2 2
                  1 1[3]2[3]1 2 3 2[4]1[4]2[4]1 3 3 2 2 1 2[3]1[3]2 3 4 2 1 1 2
                  2[4]1 2 2 3 3[8]1[3]2 1 2 3 1 2 1 2 2[4]1 2 2 1 1 3
                  3[3]1[3]2[3]1 2 1 3 1 5 2 2 3 2 3 3 1 3 1 2 3 2 3[10]1 4 4 2
                  6 2 2 3 2 3 3 1 3 1 3 2 3 2[3]1[8]2 3 2 6 2 3 2 3 2 1 3 1 3 2
                  3 2 3 2 1 2 1 1 2 2 1 2 1 3 4 1 5 2 3 2 3 2 1 3 1 3 3 2 3 2 2
                  1[3]2 1 1 2 1 2 1 1 3[3]2 3 2 3 1 3 1 3 3 1 1 3 2 2 1 1 2
                  1[4]2 4 2 4 1 2 2 3 2 3 1 3 1 3 1 3 3 1 2 2 1 2 1 2[4]1 3 1 4
                  1 2 3 2 3 2 3 1 3 1 3 1 1 3 1 2 2 1 2 2 1 1[4]2 3 2 2 3 2 3 2
                  3 1 3 1 1 3 3 1 1[3]2 1 1 2 2 1 1 2 4 3 1 3 3 1 1 3 2 3 3 2 2
                  3 2 3 2 1 2 1 2 2 1 2 2 1 3 3 4 2 3 3 1 1 3 2 3 3 2 3 2 3 2 2
                  1 2 2 1 1 2 1 1 2 1 4 4 2 3 1 3 3 1 3 2 2 3 2 3 2 3 1 2 2 1 1
                  2 1 2 1 2 4 3 3 1 3 1 3 3 1 3 2 2 3 3 2 3 2 1[4]2 1 2 1 2 1 4
                  2 1 6 3 3 1 1 3 3 2 2 3 3 1 1 3 2[7]1 2 2 1 1 2 5 3 3 1 1 3 3
                  2 2 3 1 3 3 1 2 1 1[5]2 1 1[3]2 5 3 1 3 3 1 2 3 3 2 3 1 1 3 1
                  2 1 1 2 1 2 1 1 2 3 1 1 6 3 1 3 3 1 2 3 3 2 1 3 3 1 1 2 1 2 1
                  2 1 2 2 1};

               else if m = 11 & l = 0 then
                  fac = { 4 1 5 3 2 2 3 3 2 2[4]3 1 1 3 3[6]1[3]2 1 2 1 1 6 4 2
                  2 3 3 2 2 3 3 1 1 3 3 1 1 3 3 1[3]2[3]1 2 1 4 2 6 4 3 3 2 2 3
                  3 2 2 3 3 1 1 3 3 1 1 2 1 1 2 1 1 2[3]1 2 5[3]3 2 2 3 3 2 2 1
                  1 3 3 1 1 3 3[3]2 1 2 2 1 2 2 3 1 3 6 2 2[4]3 2 2 1 1[4]2 1 1
                  2 1[4]2[3]1 2 1 4 5 2 2[4]3[4]2[4]1[4]2[4]1[3]2 3 2 4 5 3
                  3[4]2 3 3 1 1[4]2[4]1 2 1 1[3]2 1 2 2 3 6 3 3[4]2 3 3 2 2[4]1
                  2 2 1 2 1 2 2[3]1 2 1[4]3 1 3 1 1 2 1 2 3 1 3 1 1 2 1 2[9]1 4
                  3 4 4 3 1 3 1 1 2 1 2 1 3 1 3 2 1 2 1 1[9]2 3 4 4 1 3 1 3 2 1
                  2 1 3 1 3 1 1 2 1 2 2[3]1 2 2 1 2 1[4]3 1 3 1 3 2 1 2 1 1 3 1
                  3 2 1 2 1[4]2 1 1 2 1 2 1 4 5 6 3 1 3 1 2 1 2 1 1 2 2 1 3 1 1
                  3 2 1 1 2 1[4]2 4 4 6 5 3 1 3 1 2 1 2 1 2 1 1 2 1 3 3 1[3]2 1
                  2[4]1 3 4 6 5 1 3 1 3 1 2 1 2 1 2 2 1 3 1 1 3[3]1 2 2 1 1[3]2
                  4 5 6 1 3 1 3 1 2 1 2 2 1 1 2 1 3 3 1 1 2 2 1 1 2 2 1 1 2[3]5
                  1 2 2 1 3 1 1 3 3 1 3 1 2 1 2 1 2 1[3]2 1 2 2 1 3 5 6 6 1 2 2
                  1 3 1 1 3 1 3 1 3 1 2 1[3]2[3]1 2 1 1 2 1 5 6 6 2 1 1 2 1 3 3
                  1 3 1 3 1 2 1 2[3]1 2 1 2 1 2 1 2 4[3]5 2 1 1 2 1 3 3 1 1 3 1
                  3 1 2 1 2 1 2 1 2 1 2 1 2 1 4 6 3 4 1 2 2 1 1 3 3 1 1 2 2 1 1
                  3 3 1 2 1 2[4]1 2 2 1 6 4 3 1 2 2 1 1 3 3 1 2 1 1 2 3 1 1 3 2
                  2 1[4]2 1 1 2 6 4 3 2 1 1 2 3 1 1 3 1 2 2 1 1 3 3[3]1 2 2 1 2
                  1 1 2 3 6 3 4 2 1 1 2 3 1 1 3 2 1 1 2 3 1 1 3 1 2 1 1 2 1 2 2
                  1 4[3]1 3 3 1 1 3 3[4]1 2 2 1 1 2 2[4]1[3]2 1 2 1 1 2 2 3 3 1
                  1 3 3 1 1 2 2 1 1 2 2[3]1[3]2[3]1 2 1 4[3]2 1 1 3 3 1 1 3 3 1
                  1 2 2 1 1[3]2 1 1 2 1 1 2[3]1 2[4]1 3 3 1 1 3 3 2 2 1 1 2 2 1
                  1[3]2 1 2 2 1 2 2 3 1 5 4 3 3[4]1 3 3 2 2[4]3[3]2 1[4]2[3]1 2
                  1 6[3]3[4]1[4]3[4]2 3 3 2 2[4]1[3]2 3 2 6 3 1 1[4]3 1 1 2
                  2[4]3 2 2 1 1 2 1 1[3]2 1 2 2 5 4 1 1[4]3 1 1 3 3[4]2 3 3 1 2
                  1 2 2[3]1 2 1 3 5 1 1 2 1 2 2 3 2 3 1 2 1 2 2 3 2 3[9]1 4 3 6
                  2 1 2 1 2 2 3 2 3 2 1 2 1 3 2 3 2 1[9]2 3 6 2 2 1 2 1 3 2 3 2
                  1 2 1 2 2 3 2 3 2[3]1 2 2 1 2 1 3 3 5 1 2 1 2 1 3 2 3 2 2 1 2
                  1 3 2 3[5]2 1 1 2 1 2 1 4 1 4 1 2 1 2 3 2 3 2 2 3 3 2 1 2 2 1
                  2 1 1 2 1[4]2 4 4 2 3 1 2 1 2 3 2 3 2 3 2 2 3 2 1 1[4]2 1
                  2[4]1 3 4 2 3 2 1 2 1 2 3 2 3 2 3 3 2 1 2 2[4]1 2 2 1 1[3]2 4
                  1 4 2 1 2 1 2 3 2 3 3 2 2 3 2 1 1 2 1 2 2 1 1 2 2 1 1 2 5 1 3
                  2 3 3 2 1 2 2 1 1 2 1 2 3 2 3 2 2 1[3]2 1 2 2 1 3 5 2 4 2 3 3
                  2 1 2 2 1 2 1 2 1 2 3 2 3 2 2[3]1 2 1 1 2 1 5 2 4 3 2 2 3 2 1
                  1 2 1 2 1 2 3 2 3 2 1 1 2 1 2 1 2 1 2 4 5 1 3 3 2 2 3 2 1 1 2
                  2 1 2 1 2 3 2 3 1 2 1 2 1 2 1 2 1 4 6 5 2 2 3 3 2 2 1 1 2 2 3
                  3 2 2 1 1 2 2 1 2[4]1 2 2 1 6 6 1 2 3 3 2 2 1 1 2 3 2 2 3 1 2
                  2 1 2 2 1[4]2 1 1 2 6 6 1 3 2 2 3 1 2 2 1 2 3 3 2 2 1 1 2 1 1
                  2 2 1 2 1 1 2 3 6 5 2 3 2 2 3 1 2 2 1 3 2 2 3 1 2 2 1 1 2 1 1
                  2 1 2 2 1 4 1 3 5 1 1 2 2 1 1[4]2 3 3 2 2 3 3[4]1[3]2 1 2 1 1
                  4 6 1 1 2 2 1 1 2 2 3 3 2 2 3 3 2 2 1[3]2[3]1 2 1 4 2 4 6 2 2
                  1 1 2 2 1 1 2 2 3 3 2 2 3 3 2 1 1 2 1 1 2[3]1 2 3 5 2 2 1 1 2
                  2 1 1 3 3 2 2 3 3[5]2 1 2 2 1 2 2 3 1 1 2 1 1[4]2 1 1 3 3[4]1
                  3 3 2 1[4]2[3]1 2 1 2[3]1[4]2[4]1[4]3 1 1 2 2[4]1[3]2 3 2 2 1
                  2 2[4]1 2 2 3 3[4]1 3 3 1 1 2 1 1[3]2 1 2 2 1[3]2[4]1 2 2 1
                  1[4]3[3]1 2 1 2 2[3]1 2 1 3 1 5 2 3 2 3 3 1 3 1 2 3 2 3 3 1
                  3[10]1 4 3 2 6 2 3 2 3 3 1 3 1 3 2 3 2 1 3 1 3 1[9]2 3 2 6 3
                  2 3 2 1 3 1 3 2 3 2 3 3 1 3 1 2[3]1 2 2 1 2 1 3 3 1 5 3 2 3 2
                  1 3 1 3 3 2 3 2 1 3 1 3[4]2 1 1 2 1 2 1 4 3 2 2 3 2 3 1 3 1 3
                  3 1 1 3 2 3 3 2 2 1 1 2 1[4]2[3]4 1 2 3 2 3 1 3 1 3 1 3 3 1 3
                  2 2 3[3]2 1 2[4]1 3 4 4 1 3 2 3 2 3 1 3 1 3 1 1 3 2 3 3 2[3]1
                  2 2 1 1[3]2 4 3 2 3 2 3 2 3 1 3 1 1 3 3 1 3 2 2 3 1 2 2 1 1 2
                  2 1 1 2 5 3 1 3 1 1 3 2 3 3 2 2 3 2 3 1 3 1 3 2 1[3]2 1 2 2 1
                  3 5 4 2 3 1 1 3 2 3 3 2 3 2 3 2 3 1 3 1 2 2[3]1 2 1 1 2 1 5 4
                  2 1 3 3 1 3 2 2 3 2 3 2 3 1 3 1 3 1 1 2 1 2 1 2 1 2 4 5 3 1 1
                  3 3 1 3 2 2 3 3 2 3 2 3 1 3 1 1 2 1 2 1 2 1 2 1 4 6 1 6 3 1 1
                  3 3 2 2 3 3 1 1 3 3 2 2 3 2 1 2[4]1 2 2 1 6 2 5 3 1 1 3 3 2 2
                  3 1 3 3 1 2 3 3[3]2 1[4]2 1 1 2 6 2 5 1 3 3 1 2 3 3 2 3 1 1 3
                  3 2 2 3 1 1 2 2 1 2 1 1 2 3 6 1 6 1 3 3 1 2 3 3 2 1 3 3 1 2 3
                  3 2 1 2 1 1 2 1 2 2 1};

               else if m = 12 & l = 0 then
                  fac = { 1 3 5 3 2 2 3 3 2 2[4]3[4]1[5]2 1 2 1 4 6 4 2 2 3 3 2
                  2 3 3 1 1 3 3 2 2[5]1 2 1 2 3 6 4 3 3 2 2 3 3 2 2 3 3[3]1 2 1
                  2 2 1 2 1 1 2 4 5[3]3 2 2 3 3 2 2 1 1 3 3 2 1 2 1 1 2 1 2 2 3
                  6 3 6 2 2[4]3 2 2 1 1 2 2 1[3]2 1 2[3]1 3 5 4 5 2 2[4]3[4]2 1
                  1 2[3]1 2 1[3]2 4 6 4 5 3 3[4]2 3 3 1 1 2 2[3]1 2 1[3]2 1 4 5
                  3 6 3 3[4]2 3 3 2 2 1 1[3]2 1 2[3]1 2 5 5[3]3 1 3 1 1 2 1 2 3
                  1 3[10]1 5 6 4 4 3 1 3 1 1 2 1 2 1 3 1 3[9]2 6 5 4 4 1 3 1 3
                  2 1 2 1 3 1 3[3]1 2 1 2 2 1 2 1 6 6 3 3 1 3 1 3 2 1 2 1 1 3 1
                  3 2 2 1 2 1 1 2 1 2 7 3 5 6 3 1 3 1 2 1 2 1 1 2 2 1 1
                  2[3]1[4]2 7 4 6 5 3 1 3 1 2 1 2 1 2 1 1 2 2 1[3]2[4]1 8 3 6 5
                  1 3 1 3 1 2 1 2 1 2 2 1 1[3]2[3]1 2 2 8 4 5 6 1 3 1 3 1 2 1 2
                  2 1 1 2 2[3]1[3]2 1 1 9 6 5 5 1 2 2 1 3 1 1 3 3 1 3 1 1 2 2 1
                  2 1 2 2 1 9 5 6 6 1 2 2 1 3 1 1 3 1 3 1 3 2 1 1 2 1 2 1 1 2
                  10[3]6 2 1 1 2 1 3 3 1 3 1 3[3]1 2[3]1 2 1 2 10[3]5 2 1 1 2 1
                  3 3 1 1 3 1 3 2 2 1[3]2 1 2 1 11 4 3 4 1 2 2 1 1 3 3 1 1 2
                  2[4]1 2 2 1 1 2 2 11 3 4 3 1 2 2 1 1 3 3 1 2 1 1[4]2 1 1 2 2
                  1 1 12 4 4 3 2 1 1 2 3 1 1 3 1 2 2 1 1 2 1 1 2 2 1 1 2 12 3 3
                  4 2 1 1 2 3 1 1 3 2 1 1 2 2 1 2 2 1 1 2 2 1 1 5 1 1 3 3 1 1 3
                  3[4]1 2 2 1 1[5]2 1 2 1 6 2 2 3 3 1 1 3 3 1 1 2 2 1 1 2 2[5]1
                  2 1 2 5 2 2 1 1 3 3 1 1 3 3 1 1 2 2 1 2 1 2 2 1 2 1 1 2 6[4]1
                  3 3 1 1 3 3 2 2 1 1 2 1 2 1 1 2 1 2 2 3 2 5 4 3 3[4]1 3 3 2 2
                  3 3 1[3]2 1 2[3]1 3 1 6[3]3[4]1[4]3[3]2[3]1 2 1[3]2 4 2 6 3 1
                  1[4]3 1 1 2 2 3 3[3]1 2 1[3]2 1 4 1 5 4 1 1[4]3 1 1 3 3[5]2 1
                  2[3]1 2 5 1 5 1 1 2 1 2 2 3 2 3 1 2 1 2[9]1 5 2 6 2 1 2 1 2 2
                  3 2 3 2 1 2 1[9]2 6 1 6 2 2 1 2 1 3 2 3 2 1 2 1 2 1 1 2 1 2 2
                  1 2 1 6 2 5 1 2 1 2 1 3 2 3 2 2 1 2 1 2 2 1 2 1 1 2 1 2 7 5 1
                  4 1 2 1 2 3 2 3 2 2 3 3 2 1 2[3]1[4]2 7 6 2 3 1 2 1 2 3 2 3 2
                  3 2 2 3 2 1[3]2[4]1 8 5 2 3 2 1 2 1 2 3 2 3 2 3 3 2 1[3]2[3]1
                  2 2 8 6 1 4 2 1 2 1 2 3 2 3 3 2 2 3 2[3]1[3]2 1 1 9 2 1 3 2 3
                  3 2 1 2 2 1 1 2 1 2 1 2 2 1 2 1 2 2 1 9 1 2 4 2 3 3 2 1 2 2 1
                  2 1 2 1 2 1 1 2 1 2 1 1 2 10 2 2 4 3 2 2 3 2 1 1 2 1 2 1 2 1
                  1 2[3]1 2 1 2 10 1 1 3 3 2 2 3 2 1 1 2 2 1 2 1 2 2 1[3]2 1 2
                  1 11 6 5 2 2 3 3 2 2 1 1 2 2 3 3 2[3]1 2 2 1 1 2 2 11 5 6 1 2
                  3 3 2 2 1 1 2 3 2 2 3[3]2 1 1 2 2 1 1 12 6 6 1 3 2 2 3 1 2 2
                  1 2 3 3 2 1 2 1 1 2 2 1 1 2 12 5 5 2 3 2 2 3 1 2 2 1 3 2 2 3
                  2 1 2 2 1 1 2 2[3]1 3 5 1 1 2 2 1 1[4]2 3 3 1 1[5]2 1 2 1 2 4
                  6 1 1 2 2 1 1 2 2 3 3[4]2[5]1 2 1 2 1 4 6 2 2 1 1 2 2 1 1 2 2
                  3 3 1 2 1 2 2 1 2 1 1 2 2 3 5 2 2 1 1 2 2 1 1 3 3[3]2 1 2 1 1
                  2 1 2 2 3 4 1 2 1 1[4]2 1 1 3 3[3]1[3]2 1 2[3]1 3 3
                  2[3]1[4]2[4]1 3 3 2[3]1 2 1[3]2 4 4 2 1 2 2[4]1 2 2 3 3[5]1 2
                  1[3]2 1 4 3 1[3]2[4]1 2 2 1 1 3 3[3]2 1 2[3]1 2 5 3 1 5 2 3 2
                  3 3 1 3 1 2 3 2 3[9]1 5 4 2 6 2 3 2 3 3 1 3 1 3 2 3[10]2 6 3
                  2 6 3 2 3 2 1 3 1 3 2 3 2 3 1 1 2 1 2 2 1 2 1 6 4 1 5 3 2 3 2
                  1 3 1 3 3 2 3[3]2 1 2 1 1 2 1 2 7 1 3 2 2 3 2 3 1 3 1 3 3 1 1
                  3 1 2[3]1[4]2 7 2 4 1 2 3 2 3 1 3 1 3 1 3 3 1 2 1[3]2[4]1 8 1
                  4 1 3 2 3 2 3 1 3 1 3 1 1 3 1[3]2[3]1 2 2 8 2 3 2 3 2 3 2 3 1
                  3 1 1 3 3 1 2[3]1[3]2 1 1 9 4 3 1 3 1 1 3 2 3 3 2 2 3 2 3 1 2
                  2 1 2 1 2 2 1 9 3 4 2 3 1 1 3 2 3 3 2 3 2 3 2 2 1 1 2 1 2 1 1
                  2 10 4 4 2 1 3 3 1 3 2 2 3 2 3 2 3 1 1 2[3]1 2 1 2 10 3 3 1 1
                  3 3 1 3 2 2 3 3 2 3[3]2 1[3]2 1 2 1 11 2 1 6 3 1 1 3 3 2 2 3
                  3 1 1 3[3]1 2 2 1 1 2 2 11 1 2 5 3 1 1 3 3 2 2 3 1 3 3 1[3]2
                  1 1 2 2 1 1 12 2 2 5 1 3 3 1 2 3 3 2 3 1 1 3 1 2 1 1 2 2 1 1
                  2 12 1 1 6 1 3 3 1 2 3 3 2 1 3 3 1 2 1 2 2 1 1 2 2 1};

               else if m = 14 & l = 0 then
                  fac = { 4 1 3 5 3 2 2 3 3 2 2[4]3[5]1[3]2 1 2 1 1 4 6 4 2 2 3
                  3 2 2 3 3 1 1 3 3 1 2 2[3]1 2 1 4 2 3 6 4 3 3 2 2 3 3 2 2 3 3
                  1 1 2 1 2 1 1 2[3]1 2 4 5[3]3 2 2 3 3 2 2 1 1 3 3 2 2 1 2 2 1
                  2 2 3 1 6 3 6 2 2[4]3 2 2 1 1[3]2 1[3]2[3]1 2 1 5 4 5 2
                  2[4]3[4]2 1 1 2 2[3]1[3]2 3 2 6 4 5 3 3[4]2 3 3 1 1 2
                  2[4]1[3]2 1 2 2 5 3 6 3 3[4]2 3 3 2 2[3]1[3]2[3]1 2 1 3 5[3]3
                  1 3 1 1 2 1 2 3 1 3[9]1 4 3 6 4 4 3 1 3 1 1 2 1 2 1 3 1 3
                  1[8]2 3 5 4 4 1 3 1 3 2 1 2 1 3 1 3 1 2 1 1 2 2 1 2 1 3 3 6 3
                  3 1 3 1 3 2 1 2 1 1 3 1 3[3]2 1 1 2 1 2 1 4 3 5 6 3 1 3 1 2 1
                  2 1 1 2 2 1 2 1 2 1[4]2[3]4 6 5 3 1 3 1 2 1 2 1 2 1 1[3]2 1
                  2[4]1 3 4 3 6 5 1 3 1 3 1 2 1 2 1 2 2[3]1 2 2 1 1[3]2 4 4 5 6
                  1 3 1 3 1 2 1 2 2 1 1 2 1 2 1 1 2 2 1 1 2 5 6 5 5 1 2 2 1 3 1
                  1 3 3 1 3 1 2 1 2 2 1 2 2 1 3 5 5 6 6 1 2 2 1 3 1 1 3 1 3 1 3
                  2 2 1 1 2 1 1 2 1 5[3]6 2 1 1 2 1 3 3 1 3 1 3[4]1 2 1 2 1 2
                  4[4]5 2 1 1 2 1 3 3 1 1 3 1 3 1 2 2 1 2 1 2 1 4 6 4 3 4 1 2 2
                  1 1 3 3 1 1 2 2 1 2[5]1 2 2 1 6 3 4 3 1 2 2 1 1 3 3 1 2 1
                  1[7]2 1 1 2 6 4 4 3 2 1 1 2 3 1 1 3 1 2 2[3]1 2 1 2 1 1 2 3 6
                  3 3 4 2 1 1 2 3 1 1 3 2 1 1 2 1 2 1 2 1 2 2 1 4 1 5 1 1 3 3 1
                  1 3 3[4]1 2 2[3]1[3]2 1 2 1 1 6 2 2 3 3 1 1 3 3 1 1 2 2[3]1 2
                  2[3]1 2 1 4 2 5 2 2 1 1 3 3 1 1 3 3 1 1[3]2 1 2 1 1 2[3]1 2
                  6[4]1 3 3 1 1 3 3 2 2 1 1 2 2 1 2 2 1 2 2 3 1 2 5 4 3 3[4]1 3
                  3 2 2 3 3 2 1[3]2[3]1 2 1 1 6[3]3[4]1[4]3[4]2[3]1[3]2 3 2 2 6
                  3 1 1[4]3 1 1 2 2 3 3[4]1[3]2 1 2 2 1 5 4 1 1[4]3 1 1 3 3 2 2
                  1[3]2[3]1 2 1 3 1 5 1 1 2 1 2 2 3 2 3 1 2 1 2[8]1 4 3 2 6 2 1
                  2 1 2 2 3 2 3 2 1 2 1 1[8]2 3 1 6 2 2 1 2 1 3 2 3 2 1 2 1 2 2
                  1 1 2 2 1 2 1 3 3 2 5 1 2 1 2 1 3 2 3 2 2 1 2 1[3]2 1 1 2 1 2
                  1 4 5 1 4 1 2 1 2 3 2 3 2 2 3 3 2 2 1 2 1[4]2 4 4 6 2 3 1 2 1
                  2 3 2 3 2 3 2 2 3 2 2 1 2[4]1 3 4 5 2 3 2 1 2 1 2 3 2 3 2 3 3
                  2 1 1 2 2 1 1[3]2 4 6 1 4 2 1 2 1 2 3 2 3 3 2 2 3 1 2 1 1 2 2
                  1 1 2 5 2 1 3 2 3 3 2 1 2 2 1 1 2 1 2 2 1 2 2 1 2 2 1 3 5 1 2
                  4 2 3 3 2 1 2 2 1 2 1 2 1 2 2 1 1 2 1 1 2 1 5 2 2 4 3 2 2 3 2
                  1 1 2 1 2 1 2[3]1 2 1 2 1 2 4 5 1 1 3 3 2 2 3 2 1 1 2 2 1 2 1
                  1 2 2 1 2 1 2 1 4 6 6 5 2 2 3 3 2 2 1 1 2 2 3 3 2 2[5]1 2 2 1
                  6 5 6 1 2 3 3 2 2 1 1 2 3 2 2 3[6]2 1 1 2[3]6 1 3 2 2 3 1 2 2
                  1 2 3 3 2 1 1 2 1 2 1 1 2 3 6 5 5 2 3 2 2 3 1 2 2 1 3 2 2 3 1
                  2 1 2 1 2 2 1 4 1 1 3 5 1 1 2 2 1 1[4]2 3 3[3]1[3]2 1 2 1 1 2
                  4 6 1 1 2 2 1 1 2 2 3 3 2 2 1 2 2[3]1 2 1 4 2 1 4 6 2 2 1 1 2
                  2 1 1 2 2 3 3 2 1 2 1 1 2[3]1 2 2 3 5 2 2 1 1 2 2 1 1 3 3[4]2
                  1 2 2 1 2 2 3 1 4 1 2 1 1[4]2 1 1 3 3 1 1 2 1[3]2[3]1 2 1 3
                  2[3]1[4]2[4]1 3 3 2 2[3]1[3]2 3 2 4 2 1 2 2[4]1 2 2 3
                  3[6]1[3]2 1 2 2 3 1[3]2[4]1 2 2 1 1 3 3 1[3]2[3]1 2 1 3 3 1 5
                  2 3 2 3 3 1 3 1 2 3 2 3[8]1 4 3 4 2 6 2 3 2 3 3 1 3 1 3 2 3 2
                  1[8]2 3 3 2 6 3 2 3 2 1 3 1 3 2 3 2 3 2 1 1 2 2 1 2 1 3 3 4 1
                  5 3 2 3 2 1 3 1 3 3 2 3[4]2 1 1 2 1 2 1 4 1 3 2 2 3 2 3 1 3 1
                  3 3 1 1 3 2 1 2 1[4]2 4 4 2 4 1 2 3 2 3 1 3 1 3 1 3 3 1 2 2 1
                  2[4]1 3 4 1 4 1 3 2 3 2 3 1 3 1 3 1 1 3 1 1 2 2 1 1[3]2 4 2 3
                  2 3 2 3 2 3 1 3 1 1 3 3 1 1 2 1 1 2 2 1 1 2 5 4 3 1 3 1 1 3 2
                  3 3 2 2 3 2 3 2 1 2 2 1 2 2 1 3 5 3 4 2 3 1 1 3 2 3 3 2 3 2
                  3[3]2 1 1 2 1 1 2 1 5 4 4 2 1 3 3 1 3 2 2 3 2 3 2 3[3]1 2 1 2
                  1 2 4 5 3 3 1 1 3 3 1 3 2 2 3 3 2 3 2 1 2 2 1 2 1 2 1 4 6 2 1
                  6 3 1 1 3 3 2 2 3 3 1 1 3 2[5]1 2 2 1 6 1 2 5 3 1 1 3 3 2 2 3
                  1 3 3 1[6]2 1 1 2 6 2 2 5 1 3 3 1 2 3 3 2 3 1 1 3 1 1 2 1 2 1
                  1 2 3 6 1 1 6 1 3 3 1 2 3 3 2 1 3 3 1 1 2 1 2 1 2 2 1};

               else if m = 15 & l = 0 then
                  fac = {[42]0[27]1[13]0 3 0[20]2[22]0[10]3[10]0[7]1[13]0 3
                  0[10]4[10]1 [22]0[10]5[10]2[7]1[13]0 3[2]0[2]2 0 5 1 2 4 0 5
                  0[3]1 0 2 0[2]1 2 0[2]1 0[6]1[3]0 1 0 1 0[2]1 0 1 0 1[2]3 1 0
                  2 3 5 1 0 1[3]2 1 0 1[2]2 0 1[2]0 1[3]0[3]1[3]0 1 0 1 0[2]1 0
                  2 0 2[2]4 2 1 3 4 0 2 1 2[3]0 2 1 2[2]0 1 0[2]1 0[6]1[3]0 1 0
                  1 0[2]1 0 1 0 3[2]5 3 2 4 5 1 3 2 0[3]1 0 2 0[2]1 2 1[2]0 1[3]
                  0[3]1[3]0 1 0 1 0[2]1 0 2 0 4 [2]0 4 3 5 0 2 4 3 1[3]2 1 0 1
                  [2]2[2]0[2]1 0[6]1[3]0 1 0 1 0[2]1 0 1 0 5[2]1 5 4 0 1 3 5 4 2
                  [3]0 2 1 2[2]0[2]1[2]0 1[3]0[3]1[3]0 1 0 1 0[2]1 0 2[2]0 2 3
                  1 4 5 4[2]1 4[3]0[2]2[3]0[2]2[2]1[2]0 1 0 1 0 1 0[4]1[3]0 1 2
                  1 3 0 1 3 4 2 5 0 5[2]2 5[3]1[2]0[3]1[4]0 [2]1 0 1[2]0 1 0[4]
                  1[3]0 1 2 1[2]0 2 4 5 3 0 1 0[2]3 0[3]2[2]1 [3]2[4]1[2]0 1 0 1
                  0 1 0[4]1[3]0 1 2 1 3 0 3 5 0 4 1 2 1[2]4 1[3]0 [2]2[3]0[2]2
                  [2]0[2]1 0 1[2]0 1 0[4]1[3]0 1 2 1[2]0 4 0 1 5 2 3 2 [2]5 2[3]
                  1[2]0[3]1[2]0[2]1[2]0 1 0 1 0 1 0[4]1[3]0 1 2 1 3 0 5 1 2 0 3
                  4 3[2]0 3[3]2[2]1[3]2[2]1[2]0[2]1 0 1[2]0 1 0[4]1[3]0 1 2 1[3]
                  0 3 0[2]2[2]3 2 4 0 2 1 2 0 2 1 2[2]1 2 1 0[3]1 0[2]1 0 [3]1
                  0 1[2]0 1 0 3 1 2 0 1 4 1[2]3[2]4 3 5 1 0 2 0 1 0 2 0[2]2[2]0
                  1[3]0 1 0 1 0[3]1 0 1[2]0 1 0 3[2]1 0 2 5 2[2]4[2]5 4 0 2 1 0
                  1 2 1 0 1[2]0[2]1 0[3]1 0[2]1 0[3]1 0 1[2]0 1 0 3 1 2 0 3 0
                  3[2]5 [2]0 5 1 3 2 1 2 0 2 1 2[2]1 2 0 1[3]0 1 0 1 0[3]1 0
                  1[2]0 1 0 3 [2]1 0 4 1 4[2]0[2]1 0 2 4 0 2 0 1 0 2 0[2]2 0 1
                  0[3]1 0[2]1 0[3]1 0 1[2]0 1 0 3 1 2 0 5 2 5[2]1[2]2 1 3 5 1 0
                  1 2 1 0 1[2]0 1 0 1[3]0 1 0 1 0[3]1 0 1[2]0 1 0 3[2]1[3]0 1 3
                  4 2 0 1 5[3]2 1 0[2]1 0[2]2 1[2]0[4]1[4]0 1[2]0[5]1 4 2 0[3]1
                  2 4 5 3 1 2 0 3[2]0 2 1[2]2 1 [2]0 2[2]1[4]0 1[3]0 1[2]0[5]1
                  4 2 3 1[2]2 3 5 0 4 2 3 1 4[2]1 0 2[2]0 2[2]1[3]0[4]1[4]0 1[2]
                  0[5]1 4 2 0 1[2]3 4 0 1 5 3 4 2 5 [2]2 1 0[2]1 0[2]2[3]1[4]0
                  1[3]0 1[2]0[5]1 4 2 3 1[2]4 5 1 2 0 4 5 3[3]0 2 1[2]2 1[2]0
                  2[2]0[4]1[4]0 1[2]0[5]1 4 2 0 1[2]5 0 2 3 1 5 0 4[3]1 0 2[2]0
                  2[2]1 0[2]1[4]0 1[3]0 1[2]0[5]1 4 2 3 1 0 3 5 4 5 4 3 5 2[2]1
                  2 0 2 0 2 1 0[2]1 0 1 0[2]1[2]0[2]1[2]0 1[2]0 1 0[2]1 5 2[3]1
                  4 0 5 0 5 4 0 3[2]2 0 1 0 1 0 2 1[2]2 1 0 1[2]0 [4]1[2]0 1[2]
                  0 1 0[2]1 5[2]2 1 2 5 1 0 1 0 5 1 4 3 0 1 2 1 2 1 0 2 [3]0 1
                  0[2]1[2]0[2]1[2]0 1[2]0 1 0[2]1 5 2[2]1 3 0 2 1 2 1 0 2 5 4 1
                  2 0 2 0 2 1 0[3]1 0 1[2]0[4]1[2]0 1[2]0 1 0[2]1 5[2]2 1 4 1 3
                  2 3 2 1 3 0 5 2 0 1 0 1 0 2 1[2]2 0 1 0[2]1[2]0[2]1[2]0 1[2]0
                  1 0[2]1 5 2[2]1 5 2 4 3 4 3 2 4 1[2]0 1 2 1 2 1 0 2[2]0 1 0 1
                  [2]0 [4]1[2]0 1[2]0 1 0[2]1 5[2]2 1 0 5 1 2 0 4 1 5 4 2 1 0 2
                  1[2]0[2]2 0[4]1[2]0 1[2]0[2]1[2]0[3]1 0 1 0 6[2]3[2]1 0 2 3 1
                  5 2 0 5 3 2 1 0 2[2]1[2]0 1 2[3]0[2]1 0 1 0[2]1[2]0[3]1 0 1 0
                  6 3 0 1 2 1 3 4 2 0 3 1 0 4 0 2 1 0[2]2[2]1 2 0[3]1[2]0 1[2]0
                  [2]1[2]0[3]1 0 1 0 6[2]3 1 3 2 4 5 3 1 4 2 1 5 1 0 2 1[2]0[2]
                  2 0 1[3]0[2]1 0 1 0[2]1 [2]0[3]1 0 1 0 6 3 0 1 4 3 5 0 4 2 5
                  3 2 0 2 1 0 2[2]1[2]0 1 2[3]1 [2]0 1[2]0[2]1[2]0[3]1 0 1 0 6
                  [2]3 1 5 4 0 1 5 3 0 4 3 1 0 2 1 0 [2]2[2]1 2[4]0[2]1 0 1 0
                  [2]1[2]0[3]1 0 1 0 6 3 0 1 0 5 4 1 3 0 1 0 3 5[2]2 1[2]2[2]1
                  2 1 0 1[3]0[2]1 0 1 0[2]1 0 1 0 1[2]0 1 7 3 2 [2]1 0 5 2 4 1
                  2 1 4[3]0 2[2]0[2]2 0 2 1 0[3]1[2]0[2]1 0[2]1 0 1 0 1[2]0 1 7
                  3[2]1 2 1 0 3 5 2 3 2 5[3]1 0[2]1[2]0 1 0 2 1[3]0[2]1 0 1 0[2]
                  1 0 1 0 1[2]0 1 7 3 2 1 3 2 1 4 0 3 4 3 0[3]2 1[2]2[2]1 2 1
                  [2]0[3]1[2]0[2]1 0[2]1 0 1 0 1[2]0 1 7 3[2]1 4 3 2 5 1 4 5 4
                  1 3 [2]0 2[2]0[2]2 0 2[2]1[3]0[2]1 0 1 0[2]1 0 1 0 1[2]0 1 7
                  3 2 1 5 4 3 0 2 5 0 5 2 4[2]1 0[2]1[2]0 1 0 2 0[3]1[2]0[2]1 0
                  [2]1 0 1 0 1 [2]0 1 7 3[2]1 0 1[2]5 1 3 5 3[3]1 2[2]0 1 2 1 2
                  0 2[2]0 1[3]0 1 [2]0 1 0[2]1[2]0[3]1 8 4 0 2 1 2[2]0 2 4 0 4
                  [3]2 0[2]1 2 0 2 0 1 0 [2]1 0[3]1[3]0 1 0[2]1[2]0[3]1 8 4 3[2]
                  2 3[2]1 3 5 1 5[2]3 0 1 [2]2 0 1 0 1 2 1[2]0 1[3]0 1[2]0 1 0
                  [2]1[2]0[3]1 8 4 0 2 3 4[2]2 4 0 2 0[2]4 1 2[2]0 1 2 1 2 0 2
                  [2]1 0[3]1[3]0 1 0[2]1[2]0[3]1 8 4 3 2 4 5[2]3 5 1 3 1[2]5 2
                  0[2]1 2 0 2 0 1[3]0 1[3]0 1[2]0 1 0 [2]1[2]0[3]1 8 4 0 2 5 0
                  [2]4 0 2 4 2[3]0 1[2]2 0 1 0 1 2[3]1 0 [3]1[3]0 1 0[2]1[2]0[3]
                  1 8 4 3 2 0 1 4 3 2[2]5 2 3 4 1 0 2[3]1 2 0 2[2]0 1 0 1 0[4]1
                  0 1 0 1[2]0[2]1 0 9 4 1 2 1 2 5 4 3[2]0 3 4 5 2 1 0[3]2 0 1 0
                  [2]1 0 1 0 1[2]0[2]1 0 1 0 1[2]0[2]1 0 9 4[3]2 3 0 5 4 [2]1 4
                  5[2]0 2 1[3]0 1 2 1 2 0 1 0 1 0[4]1 0 1 0 1[2]0[2]1 0 9 4 1 2
                  3 4 1 0 5[2]2 5 0[2]1 0 2[3]1 2 0 2 0 1 0 1 0 1[2]0[2]1 0 1 0
                  1 [2]0[2]1 0 9 4[2]2 4 5 2 1 0[2]3 0 1[2]2 1 0[3]2 0 1 0 1 0
                  1 0 1 0 [4]1 0 1 0 1[2]0[2]1 0 9 4 1 2 5 0 3 2 1[2]4 1 2 3 0
                  2 1[3]0 1 2 1 2 1 0 1 0 1[2]0[2]1 0 1 0 1[2]0[2]1 0 9 4[2]2 0
                  4 2 5 3 2[2]4 5 3 0 [2]1 2 1[2]2 1[2]0[4]1[4]0[4]1[2]0[2]1[2]0
                  10 5 3 2 1 5 3 0 4 3 [2]5 0 4 1[2]2 0 2[2]0 2[2]1[4]0[3]1 0
                  [4]1[2]0[2]1[2]0 10 5 0 [2]2 0 4 1 5 4[2]0 1 5 2[2]0 1 0[2]1
                  0[2]2[4]1[4]0[4]1[2]0[2]1 [2]0 10 5 3 2 3 1 5 2 0 5[2]1 2[2]0
                  [2]1 2 1[2]2 1[6]0[3]1 0[4]1 [2]0[2]1[2]0 10 5 0 2 4 2 0 3 1
                  0[2]2 3[2]1[2]2 0 2[2]0 2[6]1 [4]0[4]1[2]0[2]1[2]0 10 5 3 2 5
                  3 1 4 2 1[2]3 4[2]2[2]0 1 0[2]1 0[2]2[4]0[3]1 0[4]1[2]0[2]1[2]
                  0 10 5 0 2 0 4 3 4[2]1 2 3 2 3 2 1 2 1 2 0[2]1 2[2]1[2]0 1 0
                  [3]1[3]0[5]1[2]0 11 5[2]2 1 5 4 5[2]2 3 4 3 4 0 2 0 2 0 1[2]2
                  0 2 0[2]1 0 1[2]0 1[3]0[5]1[2]0 11 5 1 [2]2 0 5 0[2]3 4 5 4 5
                  1 0 1 0 1 2[2]0 1 0 1[2]0 1 0[3]1[3]0[5]1 [2]0 11 5[2]2 3 1 0
                  1[2]4 5 0 5 0 2 1 2 1 2 0[2]1 2 1 0[2]1 0 1[2]0 1[3]0[5]1[2]0
                  11 5 1 2 4 2 1 2[2]5 0 1 0 1 0 2 0 2 0 1[2]2 0 2 1 [2]0 1 0[3]
                  1[3]0[5]1[2]0 11 5[2]2 5 3 2 3[2]0 1 2 1 2 1 0 1 0 1 2 [2]0 1
                  [2]0[2]1 0 1[2]0 1[3]0[5]1[2]0 11 5 1 2} + 1;
               end;

            else if n = 68 then do;
               if m = 18 & l = 2 then
                  fac = suen(shape({4 1 2 3 6 3 3 2 2 2 3 5 2 6 6 5 6 4 5 5 6 1
                  1 1 1 4 3 6 4 2 6 1 2 6 4 5 5 2 2 5 4 3 1 1 3 5 2 6 4 5 3 3 1
                  3 6 1 3 1 3 4 1 4 1 1 6 4 1 2 2 2 1 1 3 6 6 5 5 2 1 6 6 6 5 5
                  2 1 1 1 1 4 4 2 5 3 4 3 3 1 3 2 2 3 3 4 3 4 1 1 2 4 6 4 2 1 6
                  3 3 4 2 6 3 6 4 6 3 4 6 2 2 1 1 1 2 2 6 5 6 5 4 2 6 1 2 1 6 2
                  1 6 4 2 5 3 3 5 2 2 4 2 5 6 4 6 6 2 4 4 1 6 3 5 1 3 4 4 3 4 1
                  1 3 6 2 5 5 2 5 4 3 3 5 2 1 3 3 2 4 1 5 4 6 3 3 1 3 1 1 3 5 5
                  5 5 3 4 3 6 2 4 2 4 2 2 6 5 6 3 1 4 1 4 1 5 5 4 3 2 2 1 5 5 3
                  3 6 5 5 4 2 3 2 5 5 5 4 3 5 1 4 5 4 3 6 6 5 2 2 1 3 5 2 5 3 4
                  1 2 3 4 6 3 4 4 5 2 4 5 5 3 6 2 5 4 6}, n / 4));
               end;

            else if n = 64 then do;
               if m = 1 then
                  fac = shape({ 1 1 1 1 2 3 1 3 2 2 4 4 4 2 1 4 2 3 2 4 3 1 1 1
                  1 1 2 4 2 2 3 3 2 1 3 3 3 4 3 4 1 1 4 4 1 2 2 2 1 1 4 3 3 1 1
                  4 4 1 4 2 1 1 1 1 1 3 1 2 2 2 1 2 3 2 2 4 3 1 3 4 2 2 4 2 2 4
                  2 2 2 1 1 2 2 1 3 1 4 3 1 2 1 4 1 1 1 3 4 2 4 2 2 1 1 2 2 2 4
                  4 1 2 3 3 2 1 3 1 4 4 3 3 3 3 2 2 2 1 1 3 2 1 1 4 4 2 1 3 4 3
                  2 1 3 3 2 4 2 2 2 1 1 4 1 4 4 1 2 3 2 2 2 3 3 2 4 2 1 1 1 1 2
                  2 2 3 3 4 4 4 1 1 4 3 3 1 3 1 2 3 1 1 1 1 2 2 2 4 4 1 1 1 3 4
                  3 2 1 1 2 2 1 2 2 4 1 2 1 1 1 1 2 4 1 3 4 1 4 4 2 3 4 3 1 2 3
                  3 1 2 1 1 1 2 1 1 4 2 2 4 3 1 4 3 1 4 2 3 4 2 2 1 2 1 2 1 1 2
                  2 1 4 3 1 1 3 4 4 1 4 1 2 2 2 1 2 1 2 2 2 3 3 4 2 2 2 4 1 4 1
                  2 3 4 1 3 2 2 1 2 1 3 4 2 3 2 1 3 1 2 2 2 3 3 3 4 4 4 2 2 1 2
                  1 4 3 3 2 3 3 2 2 3 4 2 2 4 4 1 3 1 1 1 1 2 1 3 1 3 1 4 3 1 1
                  1 4 4 3 2 4 2 4 3 1 1 1 2 1 4 2 2 4 1 1 4 2 4 2 4 2 1 3 3 3 2
                  1 2 2 1 2 1 4 3 4 3 2 1 1 2 1 2 4 4 3 3 2 1 1 2 2 1 2 2 3 2 1
                  2 4 4 2 3 3 2 1 3 4 2 1 4 2 1 1 1 1 1 3 1 3 1 2 3 4 3 4 1 4 2
                  2 4 3 4 2 1 1 1 1 2 4 4 2 4 4 2 3 2 2 1 1 1 1 1 4 1 2 2 2 2 2
                  3 2 1 2 2 3 3 4 4 1 3 3 4 1 1 1 2 2 2 2 2 2 4 1 4 3 3 1 2 3 1
                  3 3 2 3 2 4 2 3 1 1 2 1 1 3 3 4 3 2 2 4 1 4 2 1 2 4 4 1 2 3 1
                  1 2 1 1 4 4 1 2 3 4 1 2 1 4 1 3 3 3 4 1 2 1 2 1 2 2 1 2 4 2 1
                  3 4 1 3 3 3 1 2 3 4 4 1 1 2 1 2 2 2 1 1 3 4 1 1 2 2 1 3 4 1 4
                  1 3 4 2 1 2 2 1 1 1 2 1 3 3 2 4 2 2 4 1 4 2 3 1 4 2 1 2 2 1 2
                  2 3 4 2 1 3 3 3 4 4 4 3 1 2 2 1 2 2 1 1 2 3 4 2 4 4 2 2 4 1 3
                  2 2 2 1 2 3 2 2 2 1 1 2 4 3 3 1 1 4 3 3 4 1 2 3 1 2 3 4 3 1 1
                  2 2 2 3 2 4 3 3 4 3 3 2 4 2 1 2 4 3 3 2 1 1 2 2 2 4 1 1 2 2 2
                  2 4 3 2 2 4 1 3 2 4 3 1 2 1 1 1 1 3 4 2 4 1 3 3 1 1 4 2 4 3 2
                  1 4 1 2 1 1 1 2 4 1 3 1 3 2 4 4 3 4 3 3 4 3 2 1 2 1 2 1 2 1 4
                  2 1 2 1 1 2 4 4 3 2 2 2 1 4 1 2 1 2 1 2 2 3 3 4 3 3 4 1 1 2 3
                  3 1 1 4 3 4 2 2 1 2 1 3 1 2 4 1 4 1 2 3 1 1 1 4 1 4 2 3 2 2 1
                  2 1 4 2 3 1 4 2 4 1 2 3 1 4 3 2 1 1 2 1 1 1 1 2 3 4 3 1 1 1 2
                  3 3 2 3 4 4 4 4 1 2 1 1 1 1 2 4 3 2 4 4 3 3 4 2 4 3 1 3 3 1 2
                  3 1 2 2 2 1 1 1 3 4 2 4 2 3 4 3 1 3 2 3 1 3 4 1 2 2 2 1 2 2 2
                  1 3 2 3 4 1 1 1 2 1 4 4 4 1 2 1 1 2 2 1 2 1 3 4 4 4 2 1 2 2 3
                  4 2 2 2 1 2 1 1 2 2 2 1 4 2 1 2 1 1 4 4 2 2 3 1 3 1 4 2 2 2 1
                  1 3 3 1 2 3 1 4 2 2 3 4 4 2 1 3 4 3 2 2 2 1 1 4 4 4 3 2 3 1 1
                  3 1 4 1 1 2 2 3 2 1 1 2 1 1 3 2 4 4 1 3 2 2 1 1 2 4 3 2 1 4 4
                  1 1 2 1 1 4 1 1 1 4 1 3 1 4 3 2 1 4 1 4 3 1 1 2 1 2 2 1 3 4 1
                  2 2 2 2 2 4 4 3 1 1 4 2 2 1 2 1 2 2 2 4 1 4 3 4 3 1 3 2 4 2 2
                  2 1 1 3 2 1 2 2 1 1 4 2 2 4 2 4 3 3 1 3 3 3 4 3 3 3 2 1 2 2 1
                  2 3 3 3 1 4 1 4 2 3 3 2 4 3 2 4 2 2 2 1 1 2 3 1 2 3 3 3 4 3 4
                  4 1 4 1 3 2 1 1 2 2 1 1 2 4 2 3 2 2 1 1 4 1 2 1 1 2 4 3 2 4 1
                  1 1 2 1 3 4 3 2 3 2 3 2 4 3 3 1 1 2 2 2 4 1 1 1 2 1 4 3 2 3 2
                  4 2 1 1 1 3 4 2 1 3 1 1 1 2 2 1 2 1 1 3 3 4 3 3 2 3 2 1 2 3 1
                  3 4 2 1 2 2 1 2 2 2 2 2 1 1 2 1 2 4 1 3 4 2 2 3 3 2 1 1 1 1 1
                  2 1 4 2 3 1 3 2 3 2 2 1 4 4 1 3 2 1 1 1 1 2 1 4 1 3 1 4 4 3 1
                  2 3 2 3 1 2 2 2 2 2 2 2 3 3 1 1 1 2 1 3 1 2 4 1 3 3 1 3 1 2 2
                  2 2 2 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4}, 64) ||
                  (1:8)` @ j(8, 1, 1);

               else if m = 2 then
                  fac = shape({ 2 1 1 2 2 1 4 2 4 3 2 3 3 3 4 6 2 2 2 2 1 1 3 4
                  1 2 2 1 2 3 1 2 1 8 2 1 2 2 1 4 2 1 3 4 3 1 4 3 3 7 1 2 2 1 1
                  2 2 2 2 1 1 4 4 4 1 1 4 7 1 1 2 1 1 1 1 3 1 4 1 4 2 4 4 3 5 1
                  2 1 2 2 3 1 4 3 1 2 1 2 2 2 8 3 1 1 1 1 2 4 3 4 2 3 4 2 1 4 3
                  2 6 1 2 2 2 1 2 3 3 4 2 3 3 1 2 1 5 4 2 2 1 2 1 2 2 4 2 4 2 1
                  1 4 1 1 5 2 1 2 1 2 4 2 3 4 1 1 4 1 2 3 6 3 2 2 2 2 2 3 4 3 1
                  3 3 3 2 4 2 4 6 2 1 1 1 1 1 4 4 3 2 4 2 2 2 4 7 4 1 2 2 1 2 2
                  3 1 3 3 1 2 4 3 1 8 2 1 1 1 2 1 4 3 2 1 2 2 3 4 1 3 3 8 1 2 1
                  1 1 3 1 2 4 4 4 4 3 3 2 5 1 1 1 2 2 2 1 1 1 2 1 3 1 3 1 4 2 7
                  2 2 2 1 1 2 1 2 3 3 3 4 2 2 3 2 8 2 1 1 2 2 4 1 1 1 2 4 1 2 4
                  1 5 2 2 2 1 1 2 3 3 1 4 4 2 2 1 2 4 3 7 2 1 2 2 1 1 3 2 2 1 1
                  3 1 4 2 8 1 1 2 1 2 2 2 4 3 2 4 4 3 3 1 3 7 3 1 1 2 1 1 4 4 4
                  4 1 3 2 3 3 1 4 5 1 2 2 2 1 3 2 4 1 3 1 1 4 1 4 6 4 1 1 1 1 2
                  1 2 3 3 2 2 4 4 3 2 1 6 2 1 2 1 2 1 3 4 1 4 3 2 4 1 2 5 3 2 2
                  1 2 1 3 3 3 3 1 4 3 4 3 4 2 5 2 1 1 1 1 4 1 3 2 3 2 4 3 1 1 8
                  4 2 2 2 2 2 2 1 4 4 2 1 1 3 3 3 3 6 1 1 1 2 1 1 2 1 4 3 4 1 1
                  2 2 4 8 1 2 2 1 2 3 2 2 2 2 3 4 1 4 4 7 2 1 1 2 2 2 4 4 2 3 4
                  1 3 2 2 1 1 7 1 2 1 1 1 2 4 1 1 1 2 2 2 4 3 6 1 1 1 2 1 1 3 4
                  3 1 2 4 1 1 3 3 8 7 1 2 1 2 2 1 4 4 3 3 3 4 1 1 1 3 1 1 1 1 1
                  2 2 2 4 2 1 1 3 2 3 4 5 8 1 2 2 2 1 4 2 3 4 4 2 2 2 1 2 2 2 2
                  1 1 2 2 3 1 2 4 1 3 2 4 4 3 1 4 2 2 2 1 1 1 1 1 2 4 4 3 4 2 1
                  6 6 2 1 2 2 1 2 3 1 3 2 2 4 3 4 4 4 3 2 2 1 1 2 4 3 2 1 3 1 1
                  3 2 2 7 5 1 2 2 1 2 4 2 1 3 1 4 3 3 4 2 3 4 1 1 1 2 1 2 2 2 1
                  4 3 2 3 2 4 8 6 1 2 1 1 1 1 4 2 4 2 1 1 4 4 1 2 3 1 1 2 2 2 3
                  4 1 2 3 2 4 4 2 3 5 5 2 2 1 2 1 4 3 4 2 2 3 4 2 3 2 6 7 2 1 2
                  1 2 2 3 3 4 3 4 1 2 1 4 1 1 2 2 2 2 2 1 1 3 1 1 2 2 1 3 1 7 8
                  2 1 1 1 1 3 1 4 3 4 1 3 1 1 3 4 2 1 2 1 2 2 4 1 3 2 2 1 2 4 2
                  4 4 1 1 1 2 1 1 2 1 4 4 3 2 3 4 4 2 7 7 1 2 2 2 1 1 3 4 1 1 4
                  4 3 2 3 1 2 1 1 1 1 2 3 3 3 3 4 3 1 3 4 1 6 8 2 2 2 1 1 4 4 2
                  3 1 2 1 1 1 4 5 6 2 1 1 2 2 2 4 1 1 4 1 4 1 3 2 2 4 2 2 1 1 2
                  1 2 1 4 2 3 3 2 1 3 8 5 2 1 2 2 1 3 2 2 2 3 4 2 2 3 1 3 3 1 1
                  1 2 1 3 3 1 4 1 1 4 2 1 1 7 6 1 2 2 1 2 1 3 2 2 4 2 1 2 3 3 4
                  4 1 1 2 2 2 2 1 2 3 2 4 2 1 1 2 6 5 1 2 1 1 1 4 1 1 1 3 3 3 1
                  3 4 1 3 2 1 2 1 2 3 2 4 1 2 2 3 3 2 1 2 1 2 2 1 2 1 1 2 3 3 3
                  1 2 3 4 3 5 7 2 1 1 1 1 2 4 3 2 1 3 1 4 2 2 3 2 2 2 2 2 2 4 4
                  4 4 4 4 4 4 4 4 8 8}, 64)
                  || (1:8)` @ j(8, 1, 1) || j(8, 1, 1) @ (1:8)`;

               else if m = 3 then
                  fac = shape({ 1 4 4 4 4 3 3 1 3 3 3 3 1 1 4 1 2 1 3 3 4 2 4 1
                  2 3 4 4 2 5 2 1 2 2 3 3 4 4 3 1 2 1 3 3 8 2 3 3 1 4 4 1 1 1 4
                  2 2 2 4 1 4 1 4 3 1 2 3 2 2 1 1 4 2 4 6 4 3 1 4 2 1 2 3 4 4 1
                  3 3 3 3 3 4 2 1 2 2 4 3 2 3 4 2 4 2 2 3 2 3 2 1 1 1 2 4 2 4 1
                  1 1 7 1 1 4 1 3 2 2 2 4 4 4 1 3 2 1 1 3 1 2 4 1 3 3 2 1 4 2 2
                  1 8 2 4 2 3 4 2 1 3 4 2 1 3 1 4 5 2 2 3 4 3 1 4 2 2 3 1 4 4 3
                  4 4 4 4 2 2 3 2 1 1 2 2 2 4 3 7 4 2 1 1 1 4 3 4 3 3 2 1 1 4 2
                  3 1 2 4 1 3 1 4 1 4 3 4 2 1 3 3 3 3 3 2 4 4 1 3 1 3 3 3 2 6 3
                  4 1 1 3 1 4 4 1 1 1 3 2 4 7 3 2 4 2 4 2 1 1 3 4 1 4 3 3 2 4 1
                  3 3 4 1 3 1 1 3 4 1 4 2 3 4 3 2 4 3 2 2 4 3 2 4 2 1 1 6 2 1 1
                  2 2 4 4 3 4 3 3 4 1 1 1 2 3 4 1 1 3 1 2 2 2 3 3 4 2 8 1 4 3 4
                  1 4 3 2 4 1 2 2 3 3 5 1 2 2 3 2 3 2 3 2 4 2 1 2 4 4 3 1 1 4 4
                  4 1 3 2 2 2 1 4 3 6 3 3 4 3 3 3 4 2 4 3 2 2 1 4 3 4 4 3 2 3 4
                  2 2 2 4 3 3 2 1 2 4 2 2 1 4 3 3 3 4 1 3 4 3 2 7 2 4 1 3 1 1 1
                  4 3 4 4 2 3 2 4 2 2 4 4 2 2 4 1 1 1 4 1 2 1 5 1 1 3 1 2 1 2 1
                  3 2 1 4 1 4 8 1 3 2 2 1 2 3 4 1 3 1 3 4 3 1 4 3 2 2 1 1 4 1 2
                  4 2 4 1 2 5 4 1 3 1 2 2 1 4 4 1 2 3 4 1 4 3 2 4 4 2 1 3 4 2 2
                  3 2 3 4 1 3 4 1 3 1 2 2 1 4 3 3 1 2 3 8 1 2 2 1 4 4 4 2 3 2 4
                  3 2 3 3 1 4 3 2 3 3 1 3 1 3 4 4 3 4 6 2 3 4 3 3 4 3 3 3 4 1 1
                  4 1 7 2 1 1 4 4 3 2 2 1 1 1 2 1 2 2 4 2 2 3 2 4 1 2 1 3 1 2 3
                  1 8 4 4 3 4 1 3 4 3 3 2 1 1 2 2 1 3 3 4 1 1 4 2 3 1 1 4 4 1 3
                  4 3 1 1 2 2 3 3 2 3 4 4 3 4 4 5 1 3 2 4 3 1 1 1 4 1 3 1 4 4 2
                  1 1 3 3 4 2 4 4 2 4 3 2 1 3 7 2 2 4 2 4 1 2 4 4 3 2 3 2 2 6 2
                  4 1 1 3 2 3 1 2 2 2 4 3 1 3 2 3 3 3 2 3 3 4 4 2 4 4 2 3 2 2 1
                  2 4 1 4 2 1 2 3 4 3 3 4 7 1 2 1 1 1 3 4 1 4 4 1 2 4 1 6 1 4 4
                  2 2 4 1 4 2 1 1 1 1 2 3 3 2 3 4 3 2 3 3 1 4 2 3 1 2 8 3 4 2 3
                  4 1 2 2 3 1 2 4 4 1 1 4 3 1 2 4 2 4 2 1 2 3 1 3 4 4 4 1 4 1 3
                  1 1 3 3 3 3 2 2 3 5 2 2 3 2 1 2 2 3 3 1 3 2 4 4 3 2 4 2 1 2 1
                  3 2 1 4 3 1 1 3 6 1 3 1 4 2 2 1 2 3 3 2 4 2 2 7 1 1 4 3 1 1 4
                  3 1 2 2 3 3 1 2 3 3 3 1 4 3 2 4 2 3 1 1 3 1 5 3 1 2 2 3 4 3 1
                  4 2 1 2 2 2 4 4 2 1 3 3 3 1 1 2 1 4 3 1 3 1 4 4 4 4 4 4 4 4 4
                  4 4 4 4 4 8}, 64)
                  || (1:8)` @ j(8, 1, 1) || j(8, 1, 1) @ (1:8)`;

               else if m = 4 then
                  fac = shape({ 2 1 4 3 1 4 4 2 4 3 7 2 4 3 2 3 2 1 8 1 7 4 3 1
                  2 2 4 1 3 7 3 8 3 3 4 1 3 2 3 2 1 2 4 8 4 3 3 1 1 3 3 6 8 1 5
                  4 2 4 4 3 1 2 4 5 5 2 1 3 1 4 4 2 4 3 7 6 1 1 2 2 1 2 4 1 5 6
                  2 6 1 1 2 4 2 2 3 4 1 1 8 1 4 1 1 4 4 2 6 4 5 3 4 1 4 1 3 3 4
                  5 2 6 4 4 4 3 4 1 1 1 3 3 2 7 3 3 1 2 2 1 4 8 5 3 6 3 2 2 3 4
                  3 1 2 8 7 1 2 3 3 3 3 4 3 1 6 8 2 2 2 4 2 1 2 2 7 7 4 5 4 2 1
                  3 3 3 3 8 7 2 3 4 3 2 2 1 1 2 2 6 6 8 1 2 3 2 2 2 4 1 8 5 7 1
                  3 4 3 4 4 1 7 5 1 4 2 4 2 1 3 4 4 4 3 4 1 2 1 1 4 1 2 1 6 2 8
                  6 3 4 4 4 2 1 3 5 4 7 5 3 1 3 1 4 3 2 3 1 3 2 3 2 3 4 4 1 4 6
                  6 4 4 3 3 4 1 2 3 1 4 7 8 7 2 2 1 1 1 4 3 3 5 7 8 2 3 2 4 3 2
                  2 5 8 3 3 1 4 4 2 4 2 3 2 2 2 2 1 1 3 3 2 4 2 8 3 6 5 4 4 2 3
                  1 3 4 7 1 5 6 4 1 1 2 3 1 1 1 4 1 1 2 4 1 1 2 1 1 7 8 6 2 2 1
                  2 4 4 3 4 1 5 2 5 3 4 3 4 3 4 2 2 7 1 6 3 1 4 1 1 2 3 8 6 5 1
                  4 2 2 3 2 2 2 3 4 8 4 4 3 1 2 4 4 3 5 1 4 7 1 2 4 2 3 3 1 6 3
                  3 8 1 3 3 3 1 1 4 4 2 7 3 1 4 3 2 1 3 2 5 5 8 1 1 1 4 3 3 1 3
                  3 8 4 6 4 4 1 3 4 2 1 4 6 3 5 4 1 2 2 2 4 4 6 7 7 2 3 2 4 4 1
                  4 1 1 1 6 3 3 3 3 1 3 2 4 7 4 2 8 2 2 2 1 4 1 2 8 2 1 7 2 3 1
                  4 2 3 3 2 3 5 4 4 3 4 1 4 2 2 1 3 7 6 4 2 3 4 2 4 3 7 2 3 1 1
                  3 2 4 1 3 1 8 4 4 2 1 2 1 1 3 1 4 2 1 8 5 2 1 3 3 4 1 1 5 7 5
                  8 2 4 4 2 2 3 4 3 6 1 3 3 1 1 2 1 4 2 4 8 2 4 3 4 2 3 3 2 3 6
                  5 6 7 3 3 2 2 3 4 1 3 2 5 5 3 2 1 3 1 2 4 5 3 1 2 2 3 4 3 2 1
                  2 6 1 2 1 2 2 3 2 4 3 3 4 4 6 6 1 1 1 4 3 3 2 7 6 7 7 1 4 2 1
                  1 1 3 1 7 3 4 4 1 3 1 2 2 1 2 5 4 3 4 4 4 4 4 4 4 8 8 8 8},
                  64) || (1:8)` @ j(8, 1, 1) || j(8, 1, 1) @ (1:8)`;
               end;

            else if n = 60 then do;
               if m = 30 & l = 2 then
                  fac = { 1 2 2 1 2 2[3]1 2 1 2 2 1[3]2 1 2[4]1 2 2 1 2
                  1 2 1[4]2[3]1 2 1 2 2 1 2 1[3]2[3]1 2 2 1 2 1 2 2[6]1 2 1 2 2
                  1 2[3]1 2 1 2[4]1 2 1[3]2 1 2 2 1[4]2 1 1 2 1 1 2 2 1 2 2 1 2
                  2[3]1 2 2 1 2 1 2 2[5]1 2 1 2 2 1 1 2 2 1 1 2 2[3]1 2 2 1 2 2
                  1[3]2 1 2[3]1 2[4]1 2 1 2 2[5]1 2 1 2 2[3]1 2 1[3]2 1 2 2
                  1[4]2 1 2 2[3]1 2 1[3]2 1 1 2 1[4]2 1 2[3]1 2 1 2 1 2
                  1[7]2[3]1 2 2 1 2[4]1 2 1[3]2 1[4]2 1 2 2[4]1[4]2 1 2 1 2[5]1
                  2 2 1 2 1 2 2[3]1 2 2 1 1 2 1 1 2[5]1[3]2 1[3]2 1 2[4]1[3]2 1
                  2 1[4]2[3]1[3]2 1 1 2 1[4]2 1 2[3]1 2[6]1[3]2[3]1 2 1 2 1 2 1
                  1[4]2 1 2 2 1[3]2 1 2 1 1 2[4]1 2 2[3]1 2 1 2 2 1 1 2 2 1
                  2[4]1 2 2 1[5]2 1 2 2 1 2 2 1 2[3]1 2 2 1 2[4]1 2 1[3]2 1 2 1
                  2 1 2 1 1[6]2 1 2 2 1 2 1 2 2 1 2 1 2 2[4]1 2 2[3]1 2 2[3]1 2
                  1 2 1 2[4]1 2 2 1 2 2 1[3]2 1 2 1 1 2[4]1 2 1[5]2[4]1 2 2 1 2
                  2[4]1 2 1 2 2 1 2 1 2 1 2 1 2[3]1 2 1 2 2 1[3]2 1 2[4]1 2 2 1
                  2 1 2 2 1 2 2 1 1 2 1 2 2 1 1[3]2 1 1 2 1 1 2[3]1 2 1[3]2 1 1
                  2 1 2 1 1 2 2 1 2 1 2 2 1 2 1 1 2 1 2 1 2 2[4]1 2 1 2[31]1 2 1
                  2 1 1 2 1 2 1 2 1 1 2 1 2 2 1[3]2 1 1[4]2 1 2 1 2 1 1 2 2 1 1
                  2 1[3]2 1[4]2 1 1 2 1 1 2[3]1 2 1[12]2 1 1 2 2 1 1 2 1 2[3]1 2
                  1 1 2[5]1[3]2[5]1[3]2[8]1 2 1[4]2 1 1 2 2[3]1[3]2 1 2 2 1[3]2
                  1 2 2[4]1 2 1 1 2[3]1 2 1 2 2 1 2 1 1 2 2[4]1 2 1 1 2 1 2 1
                  1[3]2 1 1[4]2[3]1 2[3]1 2 2 1[3]2 1 2 1 1 2 2 1 1 2 1[4]2 1 1
                  2 2 1 2[3]1[4]2 1 2 2 1 2[5]1 2 2[4]1 2 1[5]2 1 2 1 2 1[4]2 1
                  2 1 1 2[3]1 2 1[4]2 1 1[3]2 1 2 1[3]2 1 1 2[4]1 2 1 2 1[5]2 1
                  1[3]2 1 1[3]2[3]1[3]2 1 2 1 1 2 1[5]2 1 1[4]2 1 2[5]1 2 2 1
                  1[3]2 1[5]2[3]1 2[3]1 2 1[4]2 1 1 2 1 1 2 1 2 1 2 1[5]2 1 2 2
                  1[3]2 1 2[3]1 2 1 1 2 1 2[4]1 2[6]1[4]2 1 2 2 1 1[4]2 1 2 1 1
                  2 1 2 1 2 1[3]2 1[3]2 1[6]2 1 2[3]1 2 1 1 2 1 1 2 2 1 2 2[4]1
                  2 1 1 2 1 1[3]2 1 1[3]2 1 1[5]2 1[3]2[3]1 2 2[3]1[8]2 1 2[4]1
                  2 2[3]1 2 2 1 2 2 1 2 2 1 2 1 1 2[3]1 2 1[4]2 1 1[3]2
                  1[5]2[4]1 2 1 1 2 1 2[3]1[3]2 1 1[3]2 1 1[5]2 1 2 2 1 2
                  1[3]2[3]1 2 1[3]2 1[3]2 1 2[4]1[3]2[7]1 2 2 1[3]2[3]1[3]2 1 2
                  2 1[3]2[3]1 2 2 1 2 1 2 2 1 1 2[5]1[3]2 1 1[3]2 1 1[3]2[3]1
                  2[3]1 2 2 1 1 2 1[3]2 1 2 1 1[4]2 1 2[3]1 2[5]1 2[4]1 2 2 1
                  1[3]2 1 1 2 1 1 2[3]1 2 1[3]2 1 1 2 2 1 2 1 2 1 2[3]1 2 1 2 1
                  2 1[3]2 1 2[4]1 2 1 2 2[3]1 2 1[3]2[5]1 2 1 1[4]2 1 1[3]2 1 1
                  2 2 1 1 2 1[3]2 1 2[4]1 2 2 1[3]2 1 1[4]2 1 2[4]1[4]2 1 1[3]2
                  1 1 2 2 1 2 2 1 2[3]1 2 1 1 2 1 1[8]2[27]1 2 1 1[3]2 1 2 2[3]1
                  2 2 1 2 2 1[3]2 1 2[5]1 2[3]1 2 1 2 2 1 2[3]1[4]2[4]1 2 1[3]2
                  1 2 1 2 2 1 2 1 2 1 2 2 1[5]2[3]1 2 2 1 2 1 2 2[3]1[3]2[5]1 2
                  2 1 1 2 2 1 1[3]2[3]1 2 1 1 2[3]1[4]2 1 2 2 1[3]2 1 1[3]2 1
                  2[3]1 2[3]1 2 1[4]2 1[7]2[3]1 2 2 1 1 2 2 1 2 1[3]2 1 2 2
                  1[4]2 1 1[4]2 1 2[3]1[3]2 1 2 1 1 2 1[3]2 1 2 2 1 2 2 1 2 1 1
                  2 1 2 2 1 2 1[4]2 1 1 2 1 2 1 2 2[4]1 2 1 2 2 1 1 2 2 1 2 1 2
                  2 1[3]2 1 2 1 2 1 2[3]1 2 1[4]2 1[3]2[3]1 2[4]1[5]2 1 2 2[4]1
                  2 2 1 2 1 2 2 1};

               else if m = 17 & l = 0 then do; /* 2 ** 13 6 ** 1 10 ** 1 */
                  f1 = facx(24, 13, 0);
                  f2 = facx(36, 13, 2) || facx(36, 2, 6);
                  fac = (f1[,1:10] // f2[, 1:10]) || facx(60, 5, 2) ||
                        (f1[,13] // f2[,14]) || (f1[,12] // f2[,15] + 4);
                  end;

               else if m = 17 & l = 2 then
                  fac = suen(shape({4 2 5 2 2 3 2 6 4 1 4 2 1 3 4 5 4 1 1 4 3 2
                  1 3 5 3 6 2 2 3 3 6 2 2 2 4 5 4 4 1 3 6 1 6 2 3 5 4 2 1 2 2 1
                  3 4 4 1 2 3 1 5 2 4 6 3 1 5 6 4 2 5 1 4 3 1 5 6 6 1 6 6 1 4 2
                  2 3 5 6 5 4 4 4 6 4 1 2 1 5 2 1 6 4 4 3 5 5 6 2 6 3 2 3 1 5 2
                  4 2 1 3 2 6 1 3 6 2 5 6 1 4 3 5 3 5 3 2 3 3 2 1 2 3 1 3 1 5 4
                  3 5 1 3 5 4 2 2 1 6 3 3 1 6 3 5 5 6 1 2 1 2 4 6 4 4 4 5 5 3 1
                  2 2 1 3 2 3 3 3 5 3 2 5 5 4 3 1 4 2 3 6 1 1 6 1 3 3 6 5 6 3 5
                  4 4 1 1 4 3 6 3 1 2 4 2 6 4 2 2 5 4 4 2 4 6 3 2 3 4 2 1 5 1 5
                  4}, n / 4));

               else if m = 7 & l = 2 then
                  fac = expand({2 6 1 4 1 1 2 1 2 1 1 2 1 2 2 2 1 1 1 1 5 4 2 1
                  2 2 1 1 1 1 1 1 2 1 2 1 1 2 5 1 1 2 2 1 3 6 2 2 2 2 3 2 2 1 1
                  5 1 3 1 1 2 2 4 1 3 1 1 1 1 1 3 3 2 2 2 2 3 1 1 1 1 1 1 1 3 1
                  1 1 2 3 4 1 1 5 1 1 2 1 1 5 1 3 2 1 1 2 4 1 2 1 2 1 1 3 1 5 4
                  1 1 3 3 1 3 2 3 1 7 1 1 1 4 2 1 1 1 2 1 1 1 2 1 2 1 4 1 3 1 1
                  1 4 2 3 1 2 4 1 2 1 1 3 3 1 1 3 1 1 4 3 1 2 1 5 1 1 1 1 1 1 4
                  1 1 1 5 5 1 2 4 5 2 1 2 1 2 1 1 1 3 1 2 1 1 1 1 3 5 2 3 1 3 2
                  1 3 2 4 1 1}) + 1;

               else if m = 3 & l = 2 then
                  fac = expand({0 2 4 3 2 1 2 1 1 3 3 1 2 1 1 1 1 3 1 1 1 3 2 3
                  5 1 1 1 3 2 1 1 2 3 2 1 1 2 1 2 1 1 1 2 4 1 2 1 4  10 1 1 1 3
                  2 2 2 1 1 1 3 1 4 1 1 2 1 3 1 1 1 1 1 1 2 1 3 1 1 3 3 1 1 4 2
                  4 2 1 1 1 2 1 2 1 1 2 3 1}) + 1;

               else if m = 2 & l = 2 then
                  fac = expand({2 1 1 1 2 1 1 1 2 3 3 2 1 3 1 1 2 1 1 3 2 1 2 3
                  2 1 2 2 1 2 1 1 5 2 1 1 2 2 1 1 1 2 1 1 2 3 2 5 4 1 2 1 2 2 6
                  3 1 2 1 5 1 2 2}) + 1;
               end;

            else if n = 52 then do;
               if m = 16 & l = 2 then
                  fac = suen(shape({1 4 4 1 2 2 3 3 2 3 6 5 6 4 5 3 4 1 4 2 2 3
                  6 3 6 6 5 3 2 3 1 1 1 6 2 2 3 5 3 4 3 2 5 2 2 5 4 1 5 1 3 6 5
                  5 3 5 6 1 1 3 2 2 2 1 4 4 4 3 1 1 5 3 5 1 5 4 1 1 2 2 3 1 2 1
                  1 3 3 3 6 6 5 2 3 2 6 3 6 4 6 5 2 5 4 4 3 3 5 6 4 4 5 2 6 5 3
                  1 5 5 3 6 5 2 5 1 6 1 1 5 4 6 1 6 4 3 6 3 3 4 6 3 6 5 4 6 5 3
                  2 5 2 5 4 2 3 3 3 6 6 6 4 2 4 2 1 6 4 2 2 4 4 1 6 4 1 5 2 3 5
                  2 3 3 3 1 4 2 6 3 2 1 5 4 4 2 2 2 1}, n / 4));
               end;

            else if n = 48 then do;
               if m = 33 & l = 0 then fac = facx(n, m, l);
               end;

            else if n = 44 then do;
               if m = 15 & l = 2 then
                  fac = suen(shape({2 1 2 4 4 2 3 6 3 5 1 6 4 2 4 1 4 6 5 3 3 4
                  3 2 4 2 2 1 2 3 5 6 5 3 5 3 2 1 6 4 3 5 6 4 3 4 6 3 6 6 2 5 1
                  2 6 4 5 5 5 1 4 3 4 2 6 4 6 4 1 6 2 5 1 6 5 6 1 6 2 6 5 3 1 2
                  2 2 5 6 3 5 6 1 3 3 4 1 4 3 1 3 3 2 1 6 2 3 2 6 5 4 5 5 6 4 2
                  3 5 5 2 3 3 2 3 1 1 3 5 5 6 5 6 5 3 1 6 5 5 5 4 3 3 6 2 1 1 5
                  5 4 6 2 5 1 6 3}, n / 4));
               end;

            else if n = 36 then do;

               if m = 27 & l = 2 then do;
                  fac = ks((0:2)`, gh(3, 3), 3) || j(3, 1, 1) @ (0:2)`;
                  fac = fac[,{1 4 3}] @ j(4, 1, 4) +
                        j(9, 1, 1) @ (gh(4, 4)[2:4,] + 1)`;
                  f1  = had(12, 9);
                  fac = f1[fac[,1],] || f1[fac[,2],] || f1[fac[,3],];
                  end;

               else if m = 20 & l = 2 then do;
                  fac = ks((0:2)`, gh(3, 3), 3) || j(3, 1, 1) @ (0:2)`;
                  fac = fac[,{1 4 3}] @ j(4, 1, 4) +
                        j(9, 1, 1) @ (gh(4, 4)[2:4,] + 1)`;
                  f1  = had(12, 9);
                  fac = f1[fac[,1],] || f1[fac[,2],] ||
                        facx(12, 3, 3)[fac[,3],1:2];
                  end;

               else if m = 15 & l = 0 then do;
                  fac = { 0 0 2 1 1 2 2 2 2 4 2 2 0 0 4 4 0 0 0 1 1 1 1 3 1 1 2
                  2 3 3 5 2 2 0 0 5 5 1 0 0 1 1 0 1 4 2 2 1 1 2 3 0 0 0 2 2 4 5
                  2 1 1 0 0 1 0 1 2 2 1 1 3 2 3 0 0 2 2 5 4 5 1 1 0 0 2 4 2 1 2
                  1 2 4 0 4 2 0 2 0 0 2 0 0 1 0 1 3 5 4 1 2 1 2 5 1 0 2 0 2 0 1
                  3 2 0 1 0 1 2 5 3 2 1 2 1 4 1 5 0 2 0 2 0 3 1 1 0 1 0 3 4 1 2
                  1 2 1 5 0 3 0 2 0 2 1 2 5 1 0 1 0 4 2 1 1 2 2 1 0 4 3 2 0 0 2
                  2 0 5 0 1 1 0 5 3 4 1 2 2 1 1 5 0 2 0 0 2 3 1 2 0 1 1 0 4 3 3
                  2 1 1 2 0 5 5 0 2 2 0 2 1 1 1 0 0 1 5 2 2 2 1 1 2 1 4 4 0 2 2
                  0 3 0 0 1 0 0 1};
                  end;

               else if m = 16 & l = 2 then do;
                  fac = suen(shape({1 1 2 1 5 1 2 2 3 2 5 1 4 3 5 1 1 2 2 3 4 1
                  2 4 2 2 4 6 2 2 2 4 5 5 5 4 4 1 2 2 6 6 4 4 2 2 6 1 4 6 1 3 5
                  6 6 1 3 1 4 3 1 4 2 2 5 6 4 5 2 5 4 5 1 1 6 4 2 5 4 3 1 3 3 3
                  6 6 4 4 1 3 4 4 4 1 2 3 1 4 5 1 6 3 2 2 6 2 5 3 6 2 6 2 6 2 4
                  2 3 6 3 5 4 1 2 6 2 4 3 5 1 2 5 1 1 5 2}, n / 4));
                  end;

               else if m = 13 & l = 2 then fac = facx(n, m, l);

               else if m = 13.1 & l = 2 then
                  fac = expand({0  13 2 2 1 2 3 1 1 2 3 1 1 1 9 2 2 4 2 2 3 1 2
                  2 1 2 1 2 4 1 3 1 1 1 1 2 4 1 2 1 2 2 1 1 1 1 1 2 3 2 1 1 2 2
                  2 3 1 1 1 3 3 1 1 3 1 1 1 2 1 1 1 2 7 4 1 1 1 3 3 1 2 1 1 1 1
                  1 1 2 4 1 1 3 1 9 6 2 1 2 2 3 1 1 3 1 1 1 4 2 3 2 2 3 2 2 1 1
                  1 2 1 1 1 1 2 1 5 4 2 1 1 1 1 2 4 2 1 2 1 2 1 1 1 1 1 1 1 1 1
                  3 1 1 2 2 1 2 3 1 1 3 4 2 1 1 3 1 1 1 2 3 4 2 2 2 1 5 3 1 2 3
                  3 2 4 1 4  10 2 1 3 2 1 2 2 1 2 3 1 1 2 2 1 1 2 1 1 1 1 3 7 1
                  1 1 1 2 1 1 2 3 5 5 1 1 1 1 1 2 4 1 1 1 3 2 1 1 1 1 1 5})
                  + 1;

               else if m = 10 & l = 2 then do;
                  fac = { 1 1 2 2 2 1 2 2 1 2 2 2 2 1 2 2 1 1 2 1 2 2 2 1 2 2 2
                  1 1 1 1 1 2 2 2 1 1 2 2 1 1 1 1 2 2 1 1 1 1 2 1 2 1 1 2 2 2 2
                  1 1 1 1 1 1 1 1 1 1 1 2 1 1 2 1 1 2 1 2 1 1 1 1 2 2 2 2 1 1 2
                  1 1 2 2 2 2 2 2 1 1 1 2 2 2 2 2 1 2 1 2 1 1 1 1 2 1 2 1 2 2 2
                  2 2 2 2 1 2 1 1 1 2 2 2 1 1 2 2 1 1 2 2 1 2 2 1 2 2 2 1 2 2 2
                  1 1 1 2 2 2 1 1 1 2 1 2 1 2 2 1 2 2 1 1 1 1 2 1 2 1 1 2 1 2 1
                  2 1 1 2 2 1 2 1 1 1 1 2 1 1 1 2 1 2 1 2 2 2 1 1 2 1 1 2 2 1 2
                  2 1 1 1 1 1 2 1 2 1 1 1 2 1 1 2 2 1 2 1 2 1 2 1 2 2 2 2 1 2 2
                  1 2 2 2 2 1 2 2};
                  fac = (1:2)` @ j(18, 1, 1) || {1 1}` @ (1:2)` @ j(9, 1, 1) ||
                        {[4]1[5]2[5]1[4]2[5]1[4]2[4]1[5]2}` || shape(fac, 36);
                  end;

               else if m = 10.1 & l = 2 then do;
                  fac = { 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2
                  2 1 1 2 2 2 2 2 1 1 2 2 2 2 2 2 2 2 1 1 1 2 2 2 2 1 1 1 2 2 2
                  2 1 1 1 2 1 2 2 2 1 2 1 1 2 2 2 1 2 1 1 2 2 2 1 2 1 2 2 1 1 2
                  2 2 2 2 1 1 2 2 2 2 2 1 1 2 2 2 2 1 1 2 2 1 1 2 1 1 2 2 1 1 2
                  1 1 2 2 1 1 2 1 1 2 1 2 2 2 1 1 2 1 2 2 2 1 1 2 1 2 2 2 2 2 2
                  2 1 1 2 2 2 2 2 1 1 2 2 2 2 2 1 1 1 2 1 1 2 2 1 1 2 1 1 2 2 1
                  1 2 1 1 2 2 1 1 1 2 1 2 1 2 1 1 2 1 2 1 2 1 1 2 1 2 1 2 1 2 1
                  2 1 1 2 1 2 1 2 1 1 2 1 2 1 2 1 1 2 2 1 2 1 1 2 1 2 1 2 1 1 2
                  1 2 1 2 1 1 2 1};
                  fac = (1:2)` @ j(18, 1, 1) || {1 1}` @ (1:2)` @ j(9, 1, 1) ||
                        {[3]1[6]2[6]1[3]2[6]1[3]2[3]1[6]2}` || shape(fac, 36);
                  end;

               else if m = 9 & l = 2 then do;
                  fac = { 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 2 2 1 1 2
                  1 2 2 1 1 2 1 2 2 2 2 2 2 1 1 2 2 2 2 1 1 2 2 2 2 1 1 1 2 2 2
                  2 2 1 2 2 2 2 2 1 2 2 2 2 2 2 1 1 2 2 1 2 1 1 2 2 1 2 1 1 2 2
                  1 2 2 1 1 1 2 2 2 1 1 1 2 2 2 1 1 1 2 1 2 1 2 1 2 1 2 1 2 1 2
                  1 2 1 2 1 2 2 2 2 1 2 1 2 2 2 1 2 1 2 2 2 1 2 1 2 1 1 2 2 2 2
                  1 1 2 2 2 2 1 1 2 2 2 2 1 2 1 1 2 2 1 2 1 1 2 2 1 2 1 1 2 1 1
                  2 2 1 1 1 1 2 2 1 1 1 1 2 2 1 1 1 2 1 1 2 1 1 2 1 1 2 1 1 2 1
                  1 2 1};
                  fac = (1:2)` @ j(18, 1, 1) || {1 1}` @ (1:2)` @ j(9, 1, 1) ||
                        {[3]1[6]2[6]1[3]2[6]1[3]2[3]1[6]2}` || shape(fac, 36);
                  end;

               else if m = 8 & l = 2 then do;
                  fac = { 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 1 2 1 2 2 1 2 1 2 2 1
                  2 1 2 2 2 1 2 1 2 2 1 2 1 2 2 1 2 1 1 2 1 2 2 1 2 1 2 2 1 2 1
                  2 2 1 1 2 2 2 1 1 2 2 2 1 1 2 2 2 2 2 2 1 1 2 2 2 1 1 2 2 2 1
                  1 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 1 1 2 2 1 1 1 2 2 1 1 1 2 2 1
                  1 2 1 1 2 1 2 1 1 2 1 2 1 1 2 1 2 2 1 1 1 2 2 1 1 1 2 2 1 1 2
                  1 1 2 1 2 1 1 2 1 2 1 1 2 1 2 1 1 1 2 2 1 1 1 2 2 1 1 1 2};
                  fac = (1:2)` @ j(18, 1, 1) || {1 1}` @ (1:2)` @ j(9, 1, 1) ||
                        {[6]1[3]2[3]1[6]2[3]1[6]2[6]1[3]2}` || shape(fac, 36);
                  end;

               else if m = 4 & l = 2 then
                  fac = (1:2)` @ j(18, 1, 1) || {1 1}` @ (1:2)` @ j(9, 1, 1) ||
                        {[6]1[3]2[3]1[6]2[3]1[6]2[6]1[3]2}` ||
                        { 1 1 1 2 2 2 2 2 2 1 1 2 1 1 1 1 2 2
                          1 1 2 1 1 1 1 2 2 1 1 2 2 2 2 1 2 2}`;

               else if m = 3 & l = 2 then
                  fac = (1:2)` @ j(18, 1, 1) || {1 1}` @ (1:2)` @ j(9, 1, 1) ||
                        {[6]1[3]2[3]1[6]2[3]1[6]2[6]1[3]2}`;

               else if m = 3.1 & l = 2 then
                  fac = (1:2)` @ j(18, 1, 1) || {1 1}` @ (1:2)` @ j(9, 1, 1) ||
                        {[3]1[6]2[6]1[3]2[6]1[3]2[3]1[6]2}`;

               else if m = 2 & l = 2 then
                  fac = (1:2)` @ j(18, 1, 1) || {1 1}` @ (1:2)` @ j(9, 1, 1);

               else if m = 1 & l = 2 then
                  fac = { 1 1 1 2 2 2 1 1 1 2 2 2 1 2 2 2 1 2
                          1 1 1 1 2 2 2 2 1 1 2 2 1 2 1 2 1 1};

               else if m = 1.1 & l = 2 then
                  fac = { 0 0 0 2 2 1 0 1 1 0 0 0 1 0 2 1 2 0
                          2 1 2 1 0 2 0 1 1 2 1 2 2 2 1 1 2 0};

               else if m = 9 & l = 3 then
                  fac = { 1 3 3 1 1 3 3 1 1 1 1 1 2 2 1 1 2 2 1 2 2 3 3 2 2 3 3
                  1 1 1 3 3 3 3 1 1 2 3 1 3 1 1 2 1 2 1 2 2 1 1 1 1 2 2 2 1 2 1
                  2 2 3 2 3 2 2 3 2 3 3 1 3 1 1 3 3 2 2 2 2 3 3 2 1 3 1 3 1 2 1
                  2 3 1 2 2 1 3 1 1 3 3 2 3 3 2 1 2 2 1 2 2 1 2 1 2 3 2 3 2 3 2
                  3 2 3 1 3 1 3 3 1 1 3 2 3 3 2 3 2 1 1 2 3 1 1 3 3 3 2 2 3 1 2
                  2 1 3 1 3 3 1 2 3 3 2 3 3 2 2 3 2 1 1 2 2 2 3 2 3 1 3 1 3 2 3
                  1 3 1 2 1 2 1 3 1 3 3 1 3 2 2 3 3 2 1 1 2 1 3 3 1 2 1 2 1 2 3
                  2 3 2 3 2 3 3 2 2 1 1 2 3 3 1 1 3 3 2 2 3 3 1 2 2 1 1 3 3 1 1
                  1 1 2 2 2 2 1 1 1 2 2 3 3 3 3 2 2 1 3 3 1 1 1 1 3 3 1 2 2 1 1
                  2 2 1 1 2 3 2 3 2 1 3 1 3 2 1 3 1 3 2 1 2 1 1 3 3 2 2 3 3 2 2
                  2 2 1 2 1 3 2 3 2 1 1 1 3 3 1 1 3 3};

               else if m = 8 & l = 3 then
                  fac = { 2 2 3 3 2 2 3 3 3 3 1 1 3 3 1 1 1 1 2 2 1 1 2 2 2 3 2
                  3 3 1 3 1 3 1 3 1 1 2 1 2 1 2 1 2 2 3 2 3 3 3 2 2 2 2 3 3 1 1
                  3 3 3 3 1 1 2 2 1 1 1 1 2 2 3 2 3 2 3 1 3 1 1 3 1 3 1 2 1 2 2
                  1 2 1 2 3 2 3 2 3 3 2 1 2 2 1 3 1 1 3 2 3 3 2 1 2 2 1 3 1 1 3
                  3 2 2 3 1 2 2 1 1 3 3 1 2 3 3 2 2 1 1 2 3 1 1 3 2 3 2 3 1 3 1
                  3 3 1 3 1 2 1 2 1 1 2 1 2 3 2 3 2 3 2 2 3 2 1 1 2 1 3 3 1 3 2
                  2 3 2 1 1 2 1 3 3 1 2 3 3 2 2 1 1 2 3 1 1 3 3 2 2 3 1 2 2 1 1
                  3 3 1 2 2 3 3 3 3 2 2 3 3 1 1 1 1 3 3 1 1 2 2 2 2 1 1 3 2 3 2
                  1 3 1 3 1 3 1 3 2 1 2 1 2 1 2 1 3 2 3 2 3 3 2 2 3 3 2 2 1 1 3
                  3 1 1 3 3 2 2 1 1 2 2 1 1};

               else if m = 7 & l = 3 then
                  fac = { 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 1 1 3 2 2 3 3 2 2 1 3
                  3 3 3 3 3 3 1 1 2 2 3 3 2 3 3 3 2 1 2 3 3 3 1 2 3 3 1 1 3 3 3
                  3 2 2 1 2 1 1 2 3 3 2 1 2 1 1 1 2 3 1 2 3 2 1 1 2 2 3 3 2 3 2
                  2 1 2 3 2 1 3 3 3 2 1 3 3 1 3 1 1 2 1 1 2 1 3 3 2 1 2 1 2 2 1
                  1 3 2 3 1 2 2 2 2 3 1 1 2 3 1 1 3 2 3 2 3 1 2 1 3 1 3 2 2 3 2
                  2 2 1 1 3 3 1 1 2 3 1 3 1 2 3 3 1 2 1 3 2 3 2 3 2 3 1 2 1 1 1
                  1 3 1 3 2 2 3 3 1 3 1 1 3 2 1 2 2 3 3 3 2 3 2 2 1 2 1 1 1 3 3
                  1 2 3 2 3 2 3 1 3 2 1 1 2 2 2 1 3 1 2 3 1 3 3 1 2 2 1 2 2 2 1
                  3 3 1 1 3 1 2 3};

               else if m = 5 & l = 3 then
                  fac = { 1 2 2 3 3 2 2 3 2 3 3 3 2 2 3 1 3 3 1 1 2 3 1 3 1 3 1
                  3 3 1 1 1 1 2 2 2 1 2 1 2 3 2 1 1 2 1 3 3 2 2 2 2 3 2 3 3 2 3
                  3 2 1 1 1 3 3 2 3 1 3 1 3 3 1 1 3 1 2 2 1 1 2 1 2 1 2 3 1 2 2
                  1 1 2 2 3 3 2 3 2 3 2 3 2 3 3 2 1 3 3 1 1 2 1 3 1 3 3 3 1 1 3
                  1 1 1 2 2 2 2 1 2 1 3 1 2 2 1 1 3 3 2 2 2 3 2 3 2 3 3 2 2 3 1
                  1 1 3 3 2 1 3 1 3 3 1 3 3 1 1 2 2 1 1 2 2 1 2 1 3 2 1 1 2};

               else if m = 4 & l = 3 then
                  fac = { 2 2 3 3 3 3 1 1 1 1 2 2 2 3 2 3 3 1 3 1 1 2 1 2 3 2 2
                  3 1 3 3 1 2 1 1 2 2 3 2 3 3 1 3 1 1 2 1 2 2 3 3 2 3 1 1 3 1 2
                  2 1 3 3 2 2 1 1 3 3 2 2 1 1 2 2 3 3 3 3 1 1 1 1 2 2 3 2 3 2 1
                  3 1 3 2 1 2 1 2 3 3 2 3 1 1 3 1 2 2 1 3 2 3 2 1 3 1 3 2 1 2 1
                  3 2 2 3 1 3 3 1 2 1 1 2 3 3 2 2 1 1 3 3 2 2 1 1};

               else if m = 3 & l = 3 then
                  fac = (1:3)` @ j(12, 1, 1) || j(3, 1, 1) @ (1:3)` @ j(4, 1, 1)
                        || { 1 1 3 3 1 1 2 2 2 2 3 3 2 2 3 3 1 1
                             3 3 1 1 2 2 1 1 2 2 2 2 3 3 1 1 3 3}`;

               else if m = 2 & l = 3 then do;
                  fac = ks((0:2)`, gh(3, 3), 3) || j(3, 1, 1) @ (0:2)`;
                  fac = fac[,{1 4 3}] @ j(4, 1, 4) +
                        j(9, 1, 1) @ (gh(4, 4)[2:4,] + 1)`;
                  fac = (ks((0:2)`, (0:2)`, 3) + 1) @ j(4, 1, 1) ||
                        facx(12, 3, 3)[fac[,3],3];
                  end;

               else if m = 2.1 & l = 3 then
                  fac = { 1 1 2 3 3 2 2 1 3 2 1 3 2 3 1 2 2 2 3 3 1 1 3 2 1 2 2
                  3 3 1 3 3 2 1 1 1 3 3 1 1 2 1 3 1 1 3 3 1 1 2 2 2 3 3 1 2 2 1
                  2 2 2 3 3 2 1 3 2 2 3 1 1 3};

               else if m = 1 & l = 3 then
                  fac = (ks((0:2)`, (0:2)`, 3) + 1) @ j(4, 1, 1);

               else if m = 1.1 & l = 3 then
                  fac = { 3 2 1 1 3 3 2 1 3 2 2 1 2 3 1 3 2 1 3 1 2 1 2 1 3 2 2
                  2 1 3 3 1 2 3 3 1};

               else if m = 1.2 & l = 3 then
                  fac = { 3 2 1 3 2 3 1 1 2 2 1 3 2 1 2 3 3 1 1 2 3 2 3 1 1 3 2
                  3 1 1 2 1 2 3 3 2};

               else if m = 1.3 & l = 3 then
                  fac = {1 3 2 1 3 2 3 2 1 3 2 1 2 1 3 2 1 3}`@ {[2]1}`;

               else if m = 3 & l = 6 then
                  fac = { 1 1 6 5 2 6 3 4 3 2 3 3 2 5 2 3 6 2 4 4 4 6 3 1 1 6 5
                  5 5 5 4 1 1 6 2 4 6 1 2 4 2 2 1 5 4 5 6 1 5 3 4 1 4 1 2 1 5 3
                  2 5 4 6 3 6 5 3 2 4 6 3 3 6 4 3 5 6 4 5 2 6 4 3 5 1 1 2 3 5 1
                  3 4 5 6 6 6 6 2 2 1 3 1 4 1 3 2 5 4 2};

               else if m = 3.1 & l = 6 then
                  fac = { 1 4 3 2 1 1 2 3 3 4 2 5 4 4 4 5 3 6 1 5 5 3 2 2 6 1 2
                  1 6 2 3 1 4 6 5 6 2 6 5 3 4 6 4 5 1 5 2 4 5 6 3 6 3 1 3 6 1 5
                  5 2 6 6 4 1 1 6 2 5 4 3 3 5 4 6 6 5 4 1 6 2 3 1 2 1 2 2 6 3 5
                  3 4 3 2 5 1 5 6 4 5 1 3 4 2 4 2 4 1 3};

               else if m = 3.2 & l = 6 then
                  fac = (1:6)` @ j(6, 1, 1) || j(6, 1, 1) @ (1:6)` ||
                        { 1 2 3 4 5 6 2 3 1 6 4 5 3 1 2 5 6 4
                          4 5 6 1 2 3 5 6 4 3 1 2 6 4 5 2 3 1}`;

               else if m = 3.3 & l = 6 then
                  fac = { 1 4 2 2 2 3 2 6 6 3 5 4 4 1 3 5 3 1 4 4 6 5 2 5 6 6 4
                  4 5 1 6 1 5 1 3 5 1 2 4 2 3 2 3 1 1 5 5 6 3 6 2 6 4 3 3 4 5 6
                  5 2 1 6 1 1 1 6 3 3 3 4 3 4 6 2 1 2 5 5 5 1 2 5 4 4 5 6 3 2 1
                  4 3 2 6 4 2 2 6 3 6 4 6 5 1 5 3 2 4 1};

               else if m = 3.4 & l = 6 then
                  fac = { 2 1 1 3 3 3 1 5 5 5 2 2 6 4 4 4 6 6 5 3 6 6 5 2 4 1 4
                  3 4 6 1 6 2 2 2 4 4 5 3 5 1 5 6 3 1 2 6 3 3 2 5 1 4 1 2 3 5 3
                  5 1 1 1 3 3 6 4 1 2 6 2 4 2 4 4 5 5 6 1 6 2 3 5 5 4 6 1 6 4 3
                  2 4 2 1 5 4 3 6 6 5 3 1 2 1 3 4 2 5 6};

               else if m = 2 & l = 6 then fac = facx(n, m, l);

               else if m = 2.1 & l = 6 then
                  fac = { 1 1 4 6 6 3 3 3 6 2 2 5 5 5 2 4 4 1 1 2 3 5 6 4 3 4 5
                  1 2 6 5 6 1 3 4 2 2 2 3 6 5 3 4 4 5 2 1 5 6 6 1 4 3 1 2 1 4 5
                  5 4 4 3 6 1 1 6 6 5 2 3 3 2};

               else if m = 2.2 & l = 6 then
                  fac = { 1 1 3 3 5 5 4 6 6 2 2 4 6 3 2 5 4 1 3 5 5 1 1 3 6 4 2
                  6 4 2 1 2 3 4 5 6 2 2 4 4 6 6 3 6 5 2 1 4 5 3 1 5 3 1 4 5 6 1
                  2 3 5 4 1 6 3 2 2 1 4 3 6 5};

               else if m = 2.3 & l = 6 then
                  fac = { 3 3 1 4 4 1 6 5 2 5 2 1 1 5 5 6 3 4 6 1 3 2 1 3 5 2 2
                  6 4 2 4 6 5 4 6 3 5 1 6 6 4 4 5 3 2 2 1 2 4 3 6 2 3 6 1 6 3 5
                  2 4 4 5 6 4 2 3 5 5 3 1 1 1}`;

               else if m = 1 & l = 6 then
                  fac = { 5 2 6 2 4 3 5 3 6 1 4 1 6 1 4 1 5 2 6 2 4 3 5 3 4 3 5
                  3 6 1 4 1 5 2 6 2};

               else if m = 1.1 & l = 6 then
                  fac = { 3 5 1 4 1 6 3 5 2 2 1 3 4 6 5 2 4 6 3 6 2 5 1 4 4 6 2
                  6 2 4 5 5 1 1 3 3};

               else if m = 1.2 & l = 6 then
                  fac = { 1 3 5 5 1 3 2 4 6 6 2 4 3 5 1 4 6 2 6 2 4 3 5 1 4 6 2
                  2 4 6 5 1 3 1 3 5};
               end;

            else if n = 28 then
               fac = { [18]1[12]2[6]1[12]2 1 2 1 1 2 2 1 1[3]2 1 1[3]2 1 1 2
               2[3]1 2 2 1 1 2 2 1[3]2 1 2 1 2 1 2 1 1 2[3]1 2 1 2 1 1[5]2 1 1
               2 2 1 1 2 2 1 2 1 1 2 1 1[4]2[5]1[3]2 1 1 2 1 1 2 1 2 2 1 2 1 1
               2 1[3]2[4]1[3]2[4]1 2 2 1 2 1 1 2 1[4]2 1 1[4]2[3]1 2[3]1 2 2 1
               2 1 1 2 2[3]1[3]2 1 2 2[4]1 2 2 1 2 1 2 1 2 1 1 2 2 1 2 2[3]1 2
               1 2 2 1 2 2[3]1 2 2 1 1[3]2 1 1 2 1 1 2 2 1 2 1 2 1 1 2 1 2 1 2
               1 2 1 1[3]2 1 2 1 2 1 2 2 1 2 1 1 2 2 1 2 2[3]1 2 2 1 2 1 2 1 2
               1[3]2 1 2[4]1 2 1 2 2 1[3]2 1[3]2 1 1 2 1 2[3]1 2 1 2 1 2 2[4]1
               2 1 2 2};

            else if n = 24 & m = 13 then fac = facx(n, m, l);

            else if n = 20 then do;
               fac = { 1 1 2 2 1 2 1 2 1 2 2 1 2 1 1 2 2 1 1 2 1 2 2 1 2 2 1 1
               2 1 2[5]1 2 1 2 2 1[6]2 1 2 1 2 2[4]1 2 2[3]1 2 1 2[3]1[5]2 1
               2[3]1 2 1 1 2 1 2[3]1[7]2[6]1 2 2[4]1 2 2 1 1 2 2 1 2 1 2 1 2 2
               1[3]2 1 2 1 1 2 2[9]1[3]2 1 1[3]2 1 2 1[3]2 1 2 2 1[3]2 1 2};
               end;

            else if n = 16 then do;
               if m = 12 & l = 2 then
                  fac = twolev((h2 @ h2 @ h2 @ h2)[,{2 3 4 6 7 8 10 11 12
                        14 15}]) || {2 1 1 2 1 2 2 1 1 2 2 1 2 1 1 2}`;
               else if m = 1 & l = 2 then
                  fac = { 2 1 2 1 1 1 2 2 1 1 2 2 2 2 1 1};
               end;

            else if n = 12 then do;
               if m = 3 & l = 3 then
                  fac = { 0 0 0 2 2 1 0 1 1 0 0 0 1 0 2 1 2 0 2 1 2 1 0 2 0 1 1
                  2 1 2 2 2 1 1 2 0};

               else if m = 7 & l = 3 then
                  fac = { 0 0 0 0 0 0 0 0 2 1 2 1 1 2 0 1 1 0 2 2 1 1 0 0 2 2 0
                  1 2 0 2 1 0 1 1 0 1 2 1 2 0 2 1 2 0 1 1 2 1 1 0 2 0 1 1 2 1 1
                  1 2 0 2 0 2 2 0 0 0 2 2 2 2 1 1 2 1 0 2 1 2 2 1 0 0};

               else if m = 4 & l = 2 then
                  fac = { [5]1 2 1 2 2 1 2 1[4]2 1 1 2 2 1 2 2 1 2[3]1 2 2 1 2
                          1 1 2 2 1 2 1 1 2 1 1[4]2 1};

               else if m = 6 & l = 2 then
                  fac = { 0 0 0 0 0 0 0 0 1 1 1 1 0 1 0 1 1 1 1 1 0 0 0 1 0 1 1
                  0 1 0 0 0 0 1 0 0 1 0 1 1 0 1 1 1 0 1 1 0 1 1 1 1 0 0 1 0 1 0
                  1 0 1 0 0 0 1 1 0 1 1 0 0 1};

               else if m = 8 & l = 2 then
                  fac = { 0 0 0 0 0 0 0 1 0 1 1 1 0 1 0 0 1 1 1 1 0 1 1 0 1 0 1
                  0 0 1 1 0 1 0 0 0 0 0 1 0 0 1 1 0 1 0 0 1 1 0 1 0 1 1 1 1 0 1
                  1 0 0 1 1 1 1 0 0 1 0 1 1 0 0 0 1 1 0 1 1 1 1 0 0 0};

               else if m = 3 & l = 0 then fac = facx(n, m, l);
               end;

            free f1 f2;
            return(shape(fac, n));
            finish;

         *------if IML allowed recursion, I would not need this function------;
         start facx(n, m, l) global(h2);

            if n = 60 & m = 5 & l = 2 then
               fac = expand({0 4 1 2 1 1 3 1 3 1 2 1 2 4 3 1 3 1 1 5 4 4 2 1 3 1
               1 3 1 3 1 3 3 1 1 2 4 1 3 3 1 1 2 2 2 1 2 4 1 1 2 2 1 1 2 1 3 1 1
               3 5 1 1 2 1 4 1 3 4 3 1 2 2 2 1 4 1 3 1 1 6 2 3 1 3 1 4 1 1 1 1 1
               1 1 3 1 2 1 4 1 1 2 1 1 2 2 2 1 1 1 1 1 3 2 1 1 2 1 1 2 2 4 1 1 2
               3 1 2 7 1 1 2 1 4 1 1 1 11 1 1 3 2 3 1 1 3 2 2 1 2 2}) + 1;

            else if n = 48 & m = 33 & l = 0 then do; /* 2 ** 31 6 ** 1 8 ** 1 */
               f1  = twolev((h2 @ h2)[,2:4] @ sortmat(had(12, 1), 2))
                     || j(4, 1, 1) @ twolev({1 -1}` @ j(3, 1, 1) @ h2);
               f2  = j(1, 38, 1); f2[{1 2 13 14 25 26 37}] = 0;
               fac = f1[, loc(f2)] || j(8, 1, 1) @ (1:6)` ||
                     (1:8)` @ j(6, 1, 1);
               end;

            else if n = 36 & m = 13 & l = 2 then do;
               fac = { 1 1 1 1 1 1 2 1 2 2 1 2 1 2 2 1 1 2 1 1 2 1 1 1 2 1 2 2 1
               1 2 2 1 2 1 2 1 2 1 2 2 2 2 2 2 2 2 1 2 1 1 1 1 2 2 2 1 2 1 2 1 2
               2 1 1 1 1 1 2 1 2 1 2 2 1 2 2 2 2 1 2 2 1 1 1 1 2 2 2 2 1 1 2 2 2
               1 2 2 2 2 1 2 1 1 2 2 2 1 2 2 2 1 2 1 1 2 1 1 1 1 2 1 2 2 2 1 1 1
               1 1 1 1 2 2 1 1 2 1 1 2 1 2 1 2 1 1 1 1 2 1 1 2 2 1 2 2 2 1 1 2 2
               1 1 1 2 2 1 2 2 1 2 2 2 1 1 2 1 2 1 2 1 2 2 1 1 2 2 2 1 1 1 2 2 2
               1 2 1 1 1 2 2 1 2 2 1 1 1 2 2 2 2 2 2 1 2 1 1 1 2 2 1 1 1 2 1 2 2
               1 1 1 1 1 1 2 2 2 1 1 2 1 1 2 2 1 2 1 2 2 1 1 2 1 1 1 2 1 2 1 1 2
               2 1 2 1 2 2 1 1 2 2 1 1 1 1 1 1 1 2 1 2 1 1 2 2 2 2 2 2 2 2 1 2 1
               1 2 2 1 2 2 1 2 1 1 1 1 2 2 1 1 1 2 2 1 2 1 1 2 1 2 1 1 1 2 1 1 1
               1 2 2 1 2 2 1 2 1 2 2 2 2 2 2 2 1 2 2 1 1 1 1 2 2 2 2 2 2 1 2 2 1
               1};
               fac = (1:2)` @ j(18, 1, 1) || {1 1}` @ (1:2)` @ j(9, 1, 1) ||
                     {[5]1[4]2[4]1[5]2[4]1[5]2[5]1[4]2}` || shape(fac, 36);
               end;

            else if n = 36 & m = 2 & l = 6 then do;
               fac = { 1 6 6 2 4 3 5 5 5 6 3 4 3 2 6 3 2 1 2 5 3 3 2 2 1 1 5 4 4
               5 6 1 1 4 4 6 2 4 1 3 3 1 4 4 4 1 2 6 1 5 5 2 6 5 6 6 4 2 5 1 3 5
               6 4 5 3 1 2 2 3 3 6};
               end;

            else if n = 24 & m = 13 & l = 0 then do;
               f1  = sortmat(had(12, 1), 1:2);
               f2  = (1:2)` @ j(6, 1, 1) ||
                     (j(3, 1, 1) @ (1:2)` // j(3, 1, 1) @ {2, 1});
               fac = (twolev({-1 1}` @ f1) ||
                      j(2, 1, 1) @ f2[,1:2])[,(3:12) || 14] ||
                     (1:4)` @ j(6, 1, 1) || j(4, 1, 1) @ (1:6)`;
               end;

            else if n = 12 & m = 3 & l = 3 then
               fac = { 1 1 1 1 1 1 1 2 2 1 2 2 2 2 1 2 2 1
                       1 2 3 1 1 3 2 1 2 2 1 2 2 1 3 2 2 3};

            else if n = 9 & m = 9 & l = 9 then
               run fieldop(9, add, sub, fac, inv);

            free f1 f2;
            return(shape(fac, n));
            finish;

         *------Hadamard order 428, Kharaghania and Tayfeh-Rezaiea (2004)------;
         start hadk(n);
            b4 = shape(expand({ 0 1 1 5 1 1 2 4 1 3 1 2 2 3 1 1 2 1 3 4 1 1 5 2
            2 1 1 3 2 1 1 1 1 3 1 4 1 4 4 2 1 1 1 1 5 4 1 2 1 4 4 1 2 2 3 4 1 4
            2 2 2 1 2 1 1 1 2 5 1 2 1 4 1 3 1 1 1 1 2 3 1 1 2 2 5 1 1 3 1 3 1 2
            1 1 3 2 2 1 3 1 4 2 1 1 5 1 2 1 4 4 1 2 1 4 5 1 1 1 1 2 4 3 1 1 4 1
            3 1 1 1 1 2 3 1 1 2 2 5 1 1 4 3 1 2 1 1 3 2 2 1 3 1 4 2 1 1 5 1 1 1
            2 3 4 1 4 2 2 2 1 2 1 1 1 2 5 1 1 1 1 4 1 3 1 1 1 1 2 3 1 1 2 2 5 1
            1 3 1 3 1 2 1 1 3 2 2 1 3 1 4 2 1 1 5 1 1}), 4) # 2 - 1;
            b1 = shape(b4[1,], 107, 107); b2 = shape(b4[2,], 107, 107);
            b3 = shape(b4[3,], 107, 107); b4 = shape(b4[4,], 107, 107);
            b5 = 107 || (1:106);
            do ii = 2 to 107;
               b1[ii,] = b1[ii - 1, b5]; b2[ii,] = b2[ii - 1, b5];
               b3[ii,] = b3[ii - 1, b5]; b4[ii,] = b4[ii - 1, b5];
               end;
            b5 = j(107, 107, 0);
            do ii = 1 to 107; b5[ii, 108 - ii] = 1; end;
            h = ( b1      ||  b2  * b5 ||  b3  * b5 ||  b4  * b5) //
                (-b2 * b5 ||  b1       ||  b4` * b5 || -b3` * b5) //
                (-b3 * b5 || -b4` * b5 ||  b1       ||  b2` * b5) //
                (-b4 * b5 ||  b3` * b5 || -b2` * b5 ||  b1      );
            free b1 b2 b3 b4 b5;
            h = h[,1] # h; h = h # h[1,];
            return(h);
            finish;

         *------Williamson construction of a Hadamard------;
         start hadw(n, had);
            %if &flow or &flowinit %then %do;
               file log;
               put "** Making Hadamard &hadsize, Williamson construction.";
               file print;
               %end;

            if n = 28 then ha = {0 0 0} // i(3);
            else if n = 36 then ha = i(4);
            else if n = 52 then
               ha = { 0 0 0 0 1 0, 1 0 1 0 0 1, 0 0 0 1 1 1, 0 1 0 0 1 1};
            else if n = 92 then
               ha = { 0 1 1 0 1 1 0 0 1 1 1, 1 0 0 0 1 0 0 0 1 0 1,
                      1 1 0 0 0 1 1 0 1 0 1, 1 1 0 1 1 1 0 1 0 0 0};
            else if n = 100 then
               ha = { 1 0 1 1 1 1 0 0 1 0 1 1, 0 0 0 0 1 1 1 1 0 0 1 0,
                      0 1 0 1 1 0 0 0 1 0 0 1, 0 0 0 1 0 1 0 0 1 1 1 1};
            else if n = 108 then
               ha = { 0 0 1 0 0 0 1 1 1 0 0 0 0, 0 1 1 0 1 0 0 0 0 1 1 0 0,
                      0 1 0 1 1 1 0 0 1 0 0 0 1, 1 1 1 0 1 0 0 0 1 0 0 1 0};
            else if n = 116 then
               ha = { 1 1 1 0 1 1 0 1 0 0 0 1 1 1, 1 0 0 1 0 0 1 0 1 1 1 0 1 1,
                      1 1 0 0 0 1 1 0 0 1 0 1 0 0, 0 1 0 0 0 1 1 0 0 1 0 1 1 1};
            else if n = 124 then
               ha = { 1 0 1 0 0 0 1 1 1 1 1 0 1 0 1,
                      1 1 1 1 0 0 1 1 1 0 1 0 0 1 0,
                      1 0 1 1 0 0 1 0 0 0 0 1 1 0 0,
                      0 0 0 1 1 1 0 1 1 0 1 0 1 0 0};
            else if n = 132 then
               ha = { 1 1 0 0 1 0 0 0 1 0 0 1 0 0 0 0,
                      1 1 0 0 0 0 1 0 1 0 1 0 0 1 1 0,
                      0 1 1 0 0 1 0 1 0 0 0 1 1 1 1 0,
                      1 0 0 1 0 0 0 0 1 1 0 1 0 1 1 1};
            else if n = 148 then
               ha = { 0 0 1 1 1 0 1 0 0 0 0 0 1 0 0 0 0 1,
                      1 1 1 1 0 1 0 0 0 0 0 1 1 0 0 0 0 1,
                      0 0 0 1 1 0 1 1 0 0 1 0 1 0 1 1 0 0,
                      0 0 1 0 1 0 1 0 1 1 0 0 0 1 0 0 1 1};
            else if n = 156 then
               ha = { 1 1 0 0 1 0 1 0 0 0 0 0 1 0 0 1 1 0 0,
                      1 1 0 0 1 1 0 1 0 0 0 1 0 1 0 0 1 0 0,
                      1 1 1 0 0 0 1 0 0 1 1 0 0 0 0 1 0 1 0,
                      0 0 0 1 1 0 1 0 1 0 0 0 0 0 1 1 1 0 1};
            else if n = 164 then
               ha = { 0 1 0 0 0 1 1 1 0 1 1 0 1 0 0 1 1 1 1 1,
                      0 1 0 0 0 1 1 1 0 1 1 0 1 0 0 1 1 1 1 1,
                      1 0 0 0 0 1 0 1 1 1 0 1 1 0 1 1 0 0 0 1,
                      0 1 1 1 1 0 1 0 0 0 1 0 0 1 0 0 1 1 1 0};
            else if n = 172 then
               ha = { 0 0 0 1 1 0 0 1 1 1 1 0 1 0 1 1 1 0 1 1 0,
                      1 0 1 1 1 1 1 1 0 0 0 0 1 0 1 0 0 1 1 0 1,
                      1 1 0 1 0 1 1 0 0 1 0 1 0 1 1 1 1 0 1 0 0,
                      1 0 0 0 1 1 1 1 0 1 0 0 1 0 0 1 1 0 0 0 0};
            else if n = 180 then
               ha = { 1 1 0 0 1 1 1 0 1 1 0 1 0 1 1 1 1 1 0 0 0 0,
                      0 0 1 1 0 0 0 1 0 0 1 0 1 0 0 0 0 0 1 1 1 1,
                      1 0 0 1 1 1 1 0 1 0 1 0 1 1 1 0 0 1 0 0 1 0,
                      1 0 0 1 1 1 1 0 1 0 1 0 1 1 1 0 0 1 0 0 1 0};
            else if n = 204 then
               ha = { 1 1 1 0 0 1 0 0 1 1 1 0 1 1 0 1 0 0 0 1 1 1 1 1 0,
                      0 0 1 0 1 1 0 0 1 1 0 1 1 1 0 0 0 0 1 0 1 1 1 1 1,
                      1 0 1 0 1 0 1 0 0 0 0 1 1 0 0 1 0 1 0 0 1 0 1 1 0,
                      1 0 0 1 1 1 0 1 1 1 0 1 0 1 1 1 1 0 0 0 0 1 0 0 0};
            else if n = 220 then
               ha = { 0 1 0 0 1 0 1 0 1 1 0 0 1 0 1 1 1 1 1 0 1 1 1 0 0 1 1,
                      1 0 1 1 0 1 0 1 0 0 1 1 0 1 0 0 0 0 0 1 0 0 0 1 1 0 0,
                      1 1 0 0 0 0 1 1 0 1 1 0 0 1 1 0 0 0 0 1 0 1 0 1 1 1 1,
                      1 1 0 0 0 0 1 1 0 1 1 0 0 1 1 0 0 0 0 1 0 1 0 1 1 1 1};
            else if n = 228 then
               ha = { 0 0 0 1 1 0 1 0 0 1 1 1 1 0 1 1 1 0 1 1 0 0 0 1 0 1 1 1,
                      1 1 1 0 0 1 0 1 1 0 0 0 0 1 0 0 0 1 0 0 1 1 1 0 1 0 0 0,
                      1 0 1 0 1 0 0 0 1 1 0 1 1 0 1 0 0 1 1 1 0 0 0 0 0 1 0 0,
                      1 0 1 0 1 0 0 0 1 1 0 1 1 0 1 0 0 1 1 1 0 0 0 0 0 1 0 0};
            else if n = 244 then
               ha = {1 0 0 1 0 0 1 1 0 0 1 0 1 0 1 1 1 1 0 0 1 0 0 0 0 0 1 0 0 0
               , 1 0 0 1 0 0 1 1 0 0 1 0 1 0 1 1 1 1 0 0 1 0 0 0 0 0 1 0 0 0 , 0
               0 0 1 0 1 0 1 1 1 1 0 0 0 1 1 0 0 1 0 1 1 0 1 0 0 0 1 1 1 , 1 1 1
               0 1 0 1 0 0 0 0 1 1 1 0 0 1 1 0 1 0 0 1 0 1 1 1 0 0 0};
            else if n = 252 then
               ha = {1 1 1 0 1 1 0 1 0 1 1 1 1 0 1 0 0 0 1 0 0 0 1 1 1 0 0 0 1 1
               1 , 1 1 1 0 1 1 0 1 0 1 1 1 1 0 1 0 0 0 1 0 0 0 1 1 1 0 0 0 1 1 1
               , 1 0 1 1 1 0 0 1 1 0 1 1 0 0 1 0 0 1 0 1 1 0 1 0 1 1 1 0 0 0 0 ,
               0 1 0 0 0 1 1 0 0 1 0 0 1 1 0 1 1 0 1 0 0 1 0 1 0 0 0 1 1 1 1};

            a = ha[1,]; b = ha[2,]; c = ha[3,]; d = ha[4,]; mm = ncol(ha);
            a = 1 || a || a[,mm : 1];   a = toeplitz(2 # a - 1);
            b = 1 || b || b[,mm : 1];   b = toeplitz(2 # b - 1);
            c = 1 || c || c[,mm : 1];   c = toeplitz(2 # c - 1);
            d = 1 || d || d[,mm : 1];   d = toeplitz(2 # d - 1);
            h = (  a ||  b ||  c ||  d ) // ( -b ||  a || -d ||  c ) //
                ( -c ||  d ||  a || -b ) // ( -d || -c ||  b ||  a );
            h = h[,1] # h; h = h # h[1,];
            if had = 0 then h = twolev(h[,2:ncol(h)]);
            return(h);
            finish;

         *------intermediate matrices for 100 factors, projection approach------;
         start proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
            l4  = sortmat(twolev((h2 @ h2)[,2:4]) - 1, 1:2)[,{2 3 1}];
            b1  = {1 4 2 3 5 8 6 7 9 12 10 11 13 16 14 15 17 20 18 19};
            l20 = (fac(20, 8, 2) || (1:5)` @ j(4, 1, 1))[b1,];
            l25 = sortmat(ks((0:4)`, gh(5, 5), 5) || j(5, 1, 1) @ (0:4)`, 1:2);
            b1  = {3 1 2 4 7 5 6 10 8 15 11 12 9 16 13 14 19 17 18 20};
            h20 = had(20, 0)[b1,];
            b1  = j(5, 1, 1) @ (0:4)`;    i25 = i(25);
            b4  = {0 0 1 1}`;             i4  = i(4);
            b3  = j(5, 1, 1) @ (1:20)`;
            s1  = i25[sortlike(b1, l25[,1]),] @ i4[sortlike(b4, l4[,1]) ,] * b3;
            s2  = i25[sortlike(b1, l25[,2]),] @ i4[sortlike(b4, l4[,2]) ,] * b3;
            s3  = i25[sortlike(b1, l25[,3]),] @ i4[sortlike(b4, l4[,3]) ,] * b3;
            s4  = i25[sortlike(b1, l25[,4]),] @ i4 * b3;
            s5  = i25[sortlike(b1, l25[,5]),] @ i4 * b3;
            s6  = i25[sortlike(b1, l25[,6]),] @ i4 * b3;
            b2  = b1 @ j(4, 1, 1) + 1;
            finish;

         *------intermediate matrices for 108 factors, projection approach------;
         start proj108(m1,m2,m3,m4,s1,s2,s3,s4,o3);
            t1 = i(3)[{1 3 2},];   t2 = i(3);
            t3 = i(3)[{3 1 2},];   t4 = t3;
            q1 = i(3)[{3 2 1},];   q2 = q1;
            q3 = i(3)[{2 1 3},];   q4 = q3;
            s = i(12)[{1 3 11 5 2 6 9 8 12 7 4 10},];
            t = j(12, 12, 0);   t[1:3,1:3] = t1;      t[4:6,4:6] = t2;
            t[7:9,7:9] = t3;    t[10:12,10:12] = t4;
            q = j(12, 12, 0);   q[1:3,1:3] = q1;      q[4:6,4:6] = q2;
            q[7:9,7:9] = q3;    q[10:12,10:12] = q4;
            m1 = i(12);   m2 = s * t * inv(s);   m3 = s * q * inv(s);
            m4 = i(12)[{1 5 2 3 9 7 4 8 6 12 11 10},];
            n3 = i(3)[,1]*i(3)[2,] + i(3)[,2]*i(3)[3,] + i(3)[,3]*i(3)[1,];
            k33 = j(9,9,0);
            do i = 1 to 3; do j = 1 to 3;
               k33 = k33 + (i(3)[,i] * i(3)[j,]) @ (i(3)[,j] * i(3)[i,]);
               end; end;
            s1 = i(9); s2 = k33; s3 = j(9,9,0);
            s3[1:3,1:3] = i(3);   s3[4:6,4:6] = n3;       s3[7:9,7:9] = n3 * n3;
            s3 = k33 * s3 * k33`; s4 = j(9,9,0);
            s4[1:3,1:3] = i(3);   s4[4:6,4:6] = n3 * n3;  s4[7:9,7:9] = n3;
            s4 = k33 * s4 * k33`; o3 = j(3, 1, 1);
            free t1 t2 t3 t4 q1 q2 q3 q4 s t q n3 k33 i j;
            finish;

         *------Kronecker sum in Abelian group------;
         start ks(a, b, f) global(summat);
            pa = nrow(a); qa = ncol(a);
            pb = nrow(b); qb = ncol(b);
            r  = j(pa # pb, qa # qb, 0);

            if mod(f, 4) = 0 | mod(f, 9) = 0 then do;
               if f = ncol(summat) then t = summat;
               else do;
                  run fieldop(f, t, sub, mul, inv); free sub mul inv;
                  summat = t;
                  end;
               end;

            if ncol(t) then do; /* f contains a square: 4 8 9 12 16 18 20 24 */
               aa = a + 1; bb = b + 1;
               do i = 1 to pa;       row = (i - 1) # pb;
                  do j = 1 to qa;    col = (j - 1) # qb;
                     c = aa[i, j];
                     do k = 1 to pb;
                        rk = row + k;
                        do l = 1 to qb; r[rk, col + l] = t[bb[k, l], c]; end;
                        end;
                     end;
                  end;
               end;

            else do; /* prime f and composites, no square, up to 23 */
               do i = 1 to pa;
                  row = ((i - 1) # pb + 1) : (i # pb);
                  do j = 1 to qa;
                     col = ((j - 1) # qb + 1) : (j # qb);
                     r[row, col] = mod(b + a[i, j], f);
                     end;
                  end;
               end;

            return(r);
            finish;

         *------Kronecker sum of a matrix with itself------;
         start ks2(a, f); b = a; return(ks(a, b, f)); finish;

         *------concatenate matrices with different numbers of columns------;
         start cat(d1, d2);
           m = min(ncol(d1), ncol(d2));
           return(d1[,1:m] // d2[,1:m]);
           finish;

         *------permutation matrix------;
         start q(l, c);
            n = nrow(l);
            v = (1:n);
            h = l;
            do i = 1 to n;
               if h[i,1] ^= h[i,c] then do;
                  do j = i + 1 to n;
                     if h[j,1] = h[i,c] then do;
                        t = h[j,1]; h[j,1] = h[i,1]; h[i,1] = t;
                        t = v[j]; v[j] = v[i]; v[i] = t;
                        end;
                     end;
                  end;
               end;
            return(i(n)[v,]);
            finish;

         *------Zhang et all permutation matrix, 96------;
         start kpq(p, q);
            k = j(p # q, p # q, 0);
            do i = 1 to p; do j = 1 to q;
               k = k + (i(p)[,i] * i(q)[j,]) @ (i(q)[,j] * i(p)[i,]);
               end; end;
            return(k);
            finish;

         *------field 3 addition, complementary pairs------;
         start addcomp(a, b);
            x = j(nrow(a), ncol(a), 0);
            do i = 1 to nrow(a);
               do j = 1 to ncol(a);
                  if a[i, j] ^= . then x[i, j] = mod(a[i, j] + b, 3);
                  end;
               end;
            return(x);
            finish;

         *------certain D(r, r/2, 3) difference schemes, some smaller------;
         start halfhad(n, a1, a2, ds);
            q = (nrow(ds) ^= ncol(ds) & mod(ncol(ds), 2)) + 1;
            do i = 1 to n / 7; free r;
               do j = 0 to (floor(ncol(ds) / 2) - 1);
                  r = r || (addcomp(a1, ds[i, j*2 + q]) +
                            addcomp(a2, ds[i, j*2 + q + 1]));
                  end;
               d = d // r;
               end;
            if q = 2 then d = j(n, 1, 1) || d;
            return(d);
            finish;

         *------field addition of matrices------;
         start add(a, b, f);
            if f ^= 4 then x = mod(a + b, f);
            else do;
               run fieldop(4, add, sub, mul, inv);
               k = b + 1;
               x = a # 0;
               do i = 1 to nrow(a);
                  do j = 1 to ncol(a);
                     x[i, j] = add[a[i, j] + 1, k];
                     end;
                  end;
               end;
            return(x);
            finish;

         *------row operations on difference schemes------;
         start rowop(k, t, f);
            n = nrow(k);
            m = ncol(t);
            fm1 = f - 1;
            do i = 1 to n;
               do c = 1 to fm1;
                  if all(add(k[i,1:m], c, f) = t[i,])
                     then k[i,] = add(k[i,], c, f);
                  end;
               end;
            do i = 1 to n;
               if any(k[i,1:m] ^= t[i,]) then do;
                  do j = 1 to n;
                     do c = 0 to fm1;
                        if any(k[j,1:m] ^= t[j,]) &
                           all(add(k[j,1:m], c, f) = t[i,]) then do;
                           k[j,] = add(k[j,], c, f);
                           h = k[i,]; k[i,] = k[j,]; k[j,] = h;
                           end;
                        end;
                     end;
                  end;
               end;
            return(k);
            finish;

         *------make difference scheme first row and column all zero------;
         start nrmz(d, f);
            run fieldop(f, add, sub, mul, inv);
            do i = 1 to nrow(d);
               do j = 2 to ncol(d);
                  d[i,j] = sub[d[i,j] + 1, d[i,1] + 1];
                  end;
               d[i,1] = sub[d[i,1] + 1, d[i,1] + 1];
               end;
            do j = 1 to ncol(d);
               do i = 2 to nrow(d);
                  d[i,j] = sub[d[i,j] + 1, d[1,j] + 1];
                  end;
               d[1,j] = sub[d[1,j] + 1, d[1,j] + 1];
               end;
            d = sortmat(d, 1:min(ncol(d), 8));
            return(d);
            finish;

         *------make generalized Hadamards D(r(r+1), r(r+1), p)------;
         start ghr(d1, d2);
            n1 = ncol(d1); n2 = ncol(d2); n0 = n1 - 1; n = n1 # n2;
            a  = d2[2:n2,2:n2]; f = max(d2) + 1; z = j(n1, 1, 0);
            l  = ks((0:n0)`, d1, n1) || (1+z) @ (0:n0)`;
            t  = ks(a, z, f); d = j(n, n, 0); d[,1:n1] = ks(d2[,2:n2], z, f);
            do i = 1 to n1;
               v1 = (i#n1+1) : ((i+1)#n1); v2 = ((i-1)#n1+1) : (i#n1);
               d[,v1] = (((t`)[,v2] // (q(l, i + 1) * t)));
               end;
            return(d);
            finish;

         *------make some Generalized Hadamards and difference schemes------;
         start gh(n, f);

            d = 0; dnz = 1;

            if f = 3 then do;

               d3 = mod((0:2)` * (0:2), 3);
               d2 = {0 0, 0 1};
               a   = d3[2:3,2:3];
               h00 = d3[,1]; h01 = d3[,2]; h02 = d3[,3];
               z2  = {0, 0};
               l4  = sortmat(ks(d2, d2 + 0, 2)[,2:4], 1:3);
               q1  = q(l4, 2); q2 = q(l4, 3);
               free l4;
               if n = 3 then d = d3;

               e00 = h01 || h02;
               d6  = ks(a, z2, 3);
               h10 = ks(e00, z2, 3);
               h11 = (d6`)[,1:2] // (q1 * d6);
               h12 = (d6`)[,3:4] // (q2 * d6);
               d6  = h10 || h11 || h12;
               if n = 6 then d = d6;
               free e00;

               if any(n = {42 84 105 114 147 165 168}) then do;
                  a2 = {2 1 1 . 1 . . . . . 2 . 0 1 , . 2 1 1 . 1 . 1 . . . 2 .
                  0 , . . 2 1 1 . 1 0 1 . . . 2 . , 1 . . 2 1 1 . . 0 1 . . . 2
                  , . 1 . . 2 1 1 2 . 0 1 . . . , 1 . 1 . . 2 1 . 2 . 0 1 . . ,
                  1 1 . 1 . . 2 . . 2 . 0 1 .};
                  a1 = a2[,1:7]; a2 = a2[,8:14];
                  end;

               if any(n = {9 27 51 57 81 87 129 153 165 171})
                  then do;
                  d9 = ks2(d3, 3);
                  if n = 9 then d = d9;
                  end;

               if any(n = {12 24 36 36 48 51 66 69 72 78 84 96 102 108 111 114
                           120 129 132 138 141 144 153 156 159 165 168})
                  then do;
                  e10 = h11 || h12;
                  e11 = rowop(ks(h00, z2, 3) || e10, ks(h01, z2, 3), 3)[,2:5];
                  e12 = rowop(ks(h00, z2, 3) || e10, ks(h02, z2, 3), 3)[,2:5];
                  h20 = ks(e10, z2, 3);
                  h21 = i(3) @ q1 * ks(e11, z2, 3);
                  h22 = i(3) @ q2 * ks(e12, z2, 3);
                  d12 = h20 || h21 || h22;
                  if n = 12 then d = d12;
                  free e10 e11 e12;
                  end;

               if any(n = {15 45 69 75 105 135}) then do;
                  d15 = j(15, 9, 0);
                  d15[2:15, 2:9] = shape( { 0 0 2 1 0 1 2 0 0 1 1 2 1 0 2 1 0 1
                  2 0 2 2 1 1 0 2 0 0 2 1 2 1 1 0 0 2 2 0 1 2 1 0 1 2 2 2 2 0 1
                  1 2 2 0 1 1 2 1 2 2 1 1 0 0 0 1 2 1 0 1 2 1 0 2 0 2 0 1 1 0 2
                  2 1 1 1 2 2 0 2 2 1 0 1 1 0 2 2 2 2 0 2 0 2 0 1 2 2 1 1 0 1 1
                  1}, 14);
                  if n = 15 then d = d15;
                  end;

               if any(n = {18 54 102 162}) then do;
                  d18 = ks(d3, nrmz(d6, 3), 3);
                  if n = 18 then do; d = d18; dnz = 0; end;
                  end;

               if any(n = {21 63 93 111 123 126 141 147 159}) then do;
                  d21 = j(21, 12, 0);
                  d21[2:21, 2:12] = shape({ 0 0 2 0 1 0 2 1 2 1 1 0 0 2 2 1 2 2
                  2 1 2 1 0 1 0 1 2 2 1 1 2 2 0 0 1 2 1 0 1 2 2 1 0 0 0 2 0 0 2
                  0 1 2 1 1 1 0 2 1 1 1 2 0 1 0 1 2 1 0 1 0 0 2 1 2 2 1 2 1 0 1
                  2 2 1 2 1 0 0 0 1 1 0 2 1 1 1 1 1 0 2 1 1 1 0 1 2 0 0 1 2 0 1
                  2 2 0 0 2 1 0 0 0 2 1 2 2 1 2 1 2 0 0 1 1 1 2 2 2 0 0 0 1 2 2
                  0 2 0 1 1 2 0 0 2 1 0 2 2 0 1 2 1 1 1 0 2 2 1 2 1 0 1 0 2 1 2
                  0 2 1 2 1 0 2 0 0 2 1 2 1 2 2 1 1 2 2 0 0 0 0 1 1 2 2 0 0 2 1
                  2 0 1 2 0 2 2 2 1 1 1 0 2 2 0 2}, 20);
                  if n = 21 then d = d21;
                  end;

               if any(n = {24 48 51 69 72 78 96 102 111 120 129
                           138 141 144 153 156 159 168}) then do;
                  e20 = h21 || h22;
                  e21 = rowop(ks(h10, z2, 3) || e20, ks(h11, z2, 3), 3)[,3:10];
                  e22 = rowop(ks(h10, z2, 3) || e20, ks(h12, z2, 3), 3)[,3:10];
                  h30 = ks(e20, z2, 3);
                  h31 = i(6) @ q1 * ks(e21, z2, 3);
                  h32 = i(6) @ q2 * ks(e22, z2, 3);
                  d24 = h30 || h31 || h32;
                  if n = 24 then d = d24;
                  free e20 e21 e22;
                  end;

               if any(n = {27 51 57 81 87 129 153 165 171}) then do;
                  d27 = ks(d3, d9, 3);
                  if n = 27 then d = d27;
                  end;

               if any(n = {30 57 60 66 75 78 87 90 93 114 123 138 141 171})
                  then do;
                  d30 = ghr(mod((0:4)` * (0:4), 5), nrmz(d6, 3));
                  if n = 30 then d = d30;
                  end;

               if any(n = {33 99}) then do;
                  d33 = j(33, 13, 0);
                  d33[2:33, 2:13] = shape({ 0 0 0 1 2 0 0 1 0 1 2 2 0 0 2 2 1 1
                  1 0 0 2 1 2 0 1 0 0 2 2 0 2 1 0 2 1 0 1 1 2 1 0 1 0 1 1 2 1 0
                  1 2 0 1 1 2 2 1 1 0 1 0 2 0 2 0 1 1 0 1 2 0 1 0 2 0 2 2 1 2 2
                  0 1 2 2 0 2 1 0 1 2 0 1 0 0 0 2 0 2 1 1 0 2 0 2 2 1 1 1 0 2 2
                  2 1 2 2 1 1 1 2 0 1 0 1 1 2 1 2 0 1 0 2 2 1 0 1 1 2 2 1 1 0 2
                  2 1 1 0 2 2 0 1 0 2 2 0 2 0 1 1 0 1 1 0 2 0 0 0 1 0 1 1 0 1 1
                  0 2 1 2 2 0 0 1 1 0 2 0 1 1 2 2 0 1 2 1 1 2 0 0 0 2 2 0 1 1 1
                  1 1 2 2 2 2 1 1 2 1 1 1 1 2 1 2 1 0 0 2 2 0 2 0 1 2 2 0 0 1 0
                  1 0 2 1 0 1 2 2 1 0 2 2 0 0 0 0 1 2 0 0 0 0 0 1 1 2 1 2 2 2 0
                  0 2 1 1 0 1 2 2 0 1 2 0 1 0 0 0 2 1 1 2 1 1 2 0 1 0 2 1 2 0 2
                  1 1 0 2 0 1 1 1 2 2 2 2 1 0 2 2 1 0 1 2 2 1 0 2 2 2 0 2 1 1 2
                  0 2 0 0 1 2 1 2 2 1 2 0 2 0 0 0 1 0 0 2 2 2 1 0 2 0 1 2 1 2 1
                  0 2 2 2 1 1 2 1 2 1 2 0 0 2 2 2 1 2 1 1 1 0 0 0 2}, 32);
                  if n = 33 then d = d33;
                  end;

               if any(n = {36 66 108 156}) then do;
                  d36 = ks(d3, nrmz(d12, 3), 3);
                  if n = 36 then do; d = d36; dnz = 0; end;
                  end;

               if any(n = {36.6}) then do;
                  d36 = ks2(nrmz(d6, 3), 3);
                  if n = 36.6 then do; d = d36; dnz = 0; end;
                  end;

               if any(n = {39 117}) then do;
                  d39 = j(39, 13, 0);
                  d39[2:39,2:13] = shape( { 0 0 1 1 0 1 0 0 1 2 1 1 0 0 2 2 2 2
                  1 1 0 0 0 1 0 1 0 1 0 1 1 2 2 0 0 1 0 1 1 1 2 1 2 2 0 2 1 2 0
                  1 2 1 0 0 1 2 2 2 2 2 0 1 2 2 1 0 0 1 2 0 0 0 0 2 0 0 1 2 2 1
                  1 1 2 1 0 2 0 2 1 1 0 1 1 2 1 2 0 2 1 0 2 0 1 0 2 0 2 0 0 2 1
                  1 1 2 2 2 2 0 1 0 0 2 1 2 1 1 2 1 0 2 0 1 0 2 2 0 2 1 1 1 1 1
                  2 2 1 0 1 2 0 0 2 1 1 1 1 0 1 0 1 2 0 1 1 0 1 1 2 2 1 0 1 2 1
                  0 0 2 2 2 1 0 1 0 2 1 1 2 1 0 1 2 0 0 1 1 0 0 0 0 2 0 2 0 1 2
                  1 1 0 0 1 2 0 0 1 1 0 1 1 1 0 1 2 0 0 1 0 0 2 2 1 1 0 2 2 0 2
                  0 1 2 2 1 1 1 2 0 0 1 1 2 2 1 1 0 1 1 2 1 1 2 0 1 0 1 0 2 1 2
                  0 0 2 2 1 1 0 2 2 0 1 2 1 0 2 1 2 0 2 2 0 1 1 2 2 1 0 0 0 2 0
                  2 2 1 2 0 0 0 0 2 1 0 0 2 1 2 2 0 0 1 2 1 0 0 2 1 2 2 2 0 0 2
                  2 1 2 1 0 0 1 0 2 0 1 2 1 0 1 2 0 1 2 2 2 0 2 0 1 0 2 2 0 1 0
                  2 2 0 2 1 2 1 2 1 2 1 1 0 2 1 1 0 1 2 2 2 0 0 2 2 2 1 1 1 2 2
                  1 2 1 0 0 1 2 1 1 2 2 2 0 0 2 1 0 1 2 2 0 2 0 2 0 2 2 1 0 0 2
                  2 2 0 0 2 0 0 1 0 1 0 2 2 2 1 1 1 1 2 1 0 1 1 2 2 2 2 0 0 2 1
                  1 2 2 1}, 38);
                  if n = 39 then d = d39;
                  end;

               if any(n = {42}) then do;
                  d42 = halfhad(42, a1, a2, d6);
                  if n = 42 then d = d42;
                  end;

               if any(n = {45 69 75 135}) then do;
                  d45 = ks(d3, d15, 3);
                  if n = 45 then d = d45;
                  end;

               if any(n = {48 78 96 102 111 120 129
                           138 141 156 159}) then do;
                  e30 = h31 || h32;
                  e31 = rowop(ks(h20, z2, 3) || e30, ks(h21, z2, 3), 3)[,5:20];
                  e32 = rowop(ks(h20, z2, 3) || e30, ks(h22, z2, 3), 3)[,5:20];
                  h40 = ks(e30, z2, 3);
                  h41 = i(12) @ q1 * ks(e31, z2, 3);
                  h42 = i(12) @ q2 * ks(e32, z2, 3);
                  d48 = h40 || h41 || h42;
                  if n = 48 then d = d48;
                  free e30 e31 e32;
                  end;

               if any(n = {51 153}) then do;
                  d51 = cat(d24, d27);
                  if n = 51 then d = d51;
                  end;

               if any(n = {54 162 102}) then do;
                  d54 = ks(d3, d18, 3);
                  if n = 54 then d = d54;
                  end;

               if any(n = {57 171}) then do;
                  d57 = cat(d27, d30);
                  if n = 57 then d = d57;
                  end;

               if any(n = {60 87}) then do;
                  d60 = d30 // d30;
                  if n = 60 then d = d60;
                  end;

               if any(n = {63 93 111 123 141 159}) then do;
                  d63 = ks(d3, d21, 3);
                  if n = 63 then d = d63;
                  end;

               if any(n = {66}) then do;
                  d66 = cat(d30, d36);
                  if n = 66 then d = d66;
                  end;

               if any(n = {69}) then do;
                  d69 = cat(d24, d45);
                  if n = 69 then d = d69;
                  end;

               if any(n = {72 120}) then do;
                  d72 = ks(d3, nrmz(d24, 3), 3);
                  if n = 72 then do; d = d72; dnz = 0; end;
                  end;

               if any(n = {75}) then do;
                  d75 = cat(d30, d45);
                  if n = 75 then d = d75;
                  end;

               if any(n = {78}) then do;
                  d78 = cat(d30, d48);
                  if n = 78 then d = d78;
                  end;

               if any(n = {81 129 165}) then do;
                  d81 = ks(d3, d27, 3);
                  if n = 81 then d = d81;
                  end;

               if any(n = {84 114 165}) then do;
                  d84 = halfhad(84, a1, a2, d12);
                  if n = 84 then d = d84;
                  end;

               if any(n = {87}) then do;
                  d87 = cat(d27, d60);
                  if n = 87 then d = d87;
                  end;

               if any(n = {90 138}) then do;
                  d90 = ks(d3, nrmz(d30, 3), 3);
                  if n = 90 then do; d = d90; dnz = 0; end;
                  end;

               if any(n = {93 123}) then do;
                  d93 = cat(d30, d63);
                  if n = 93 then d = d93;
                  end;

               if any(n = {96}) then do;
                  e40 = h41 || h42;
                  e41 = rowop(ks(h30, z2, 3) || e40, ks(h31, z2, 3), 3)[,9:40];
                  e42 = rowop(ks(h30, z2, 3) || e40, ks(h32, z2, 3), 3)[,9:40];
                  h50 = ks(e40, z2, 3);
                  h51 = i(24) @ q1 * ks(e41, z2, 3);
                  h52 = i(24) @ q2 * ks(e42, z2, 3);
                  d96 = h50 || h51 || h52;
                  if n = 96 then d = d96;
                  free e40 e41 e42;
                  end;

               if any(n = {99}) then do;
                  d99 = ks(d3, d33, 3);
                  if n = 99 then d = d99;
                  end;

               if any(n = {102}) then do;
                  d102 = cat(d48, d54);
                  if n = 102 then d = d102;
                  end;

               if any(n = {105}) then do;
                  d105 = halfhad(105, a1, a2, d15);
                  if n = 105 then d = d105;
                  end;

               if any(n = {108 156}) then do;
                  d108 = ks(d3, d36, 3);
                  if n = 108 then d = d108;
                  end;

               if any(n = {111 141 159}) then do;
                  d111 = cat(d48, d63);
                  if n = 111 then d = d111;
                  end;

               if any(n = {114}) then do;
                  d114 = cat(d30, d84);
                  if n = 114 then d = d114;
                  end;

               if any(n = {117}) then do;
                  d117 = ks(d3, d39, 3);
                  if n = 117 then d = d117;
                  end;

               if any(n = {120}) then do;
                  d120 = cat(d48, d72);
                  if n = 120 then d = d120;
                  end;

               if any(n = {123}) then do;
                  d123 = cat(d30, d93);
                  if n = 123 then d = d123;
                  end;

               if any(n = {126}) then do;
                  d126 = ks(nrmz(d6, 3), d21, 3);
                  if n = 126 then do; d = d126; dnz = 0; end;
                  end;

               if any(n = {129}) then do;
                  d129 = cat(d48, d81);
                  if n = 129 then d = d129;
                  end;

               if any(n = {132}) then do;
                  d132 = ghr(mod((0:10)` * (0:10), 11), nrmz(d12, 3));
                  if n = 132 then d = d132;
                  end;

               if any(n = {135}) then do;
                  d135 = ks(d3, d45, 3);
                  if n = 135 then d = d135;
                  end;

               if any(n = {138}) then do;
                  d138 = cat(d48, d90);
                  if n = 138 then d = d138;
                  end;

               if any(n = {141}) then do;
                  d141 = cat(d30, d111);
                  if n = 141 then d = d141;
                  end;

               if any(n = {144}) then do;
                  d144 = ks(nrmz(d6, 3), d24, 3);
                  if n = 144 then do; d = d144; dnz = 0; end;
                  end;

               if any(n = {147}) then do;
                  d147 = halfhad(147, a1, a2, d21);
                  if n = 147 then d = d147;
                  end;

               if any(n = {150}) then do;
                  t1 = sqrsym({27 18 39 39 9 69 9 69 27 18 69 27 18 39 9 7 57 37
                  67 47 27 67 47 7 57 37 39 69 57 37 67 47 7 69 18 9 47 7 57 37
                  67 18 9 27 39 37 67 47 7 57 9 27 39 69 18 17 58 8 48 68 27 9
                  18 69 39 17 8 48 68 17 58 39 27 9 18 69 58 8 68 17 58 8 48 69
                  39 27 9 18 8 48 68 58 8 48 68 17 18 69 39 27 9 48 68 17 58 48
                  68 17 58 8 9 18 69 39 27 68 17 58 8 48 27 69 9 39 18 17 48 58
                  68 8 7 37 47 57 67 17 18 27 69 9 39 68 8 17 48 58 47 57 67 7
                  37 48 58 39 18 27 69 9 48 58 68 8 17 67 7 37 47 57 58 68 8 9
                  39 18 27 69 8 17 48 58 68 37 47 57 67 7 68 8 17 48 69 9 39 18
                  27 58 68 8 17 48 57 67 7 37 47 8 17 48 58 68});
                  t2 = shape({2 3 4 5 6 2 3 4 5 6 2 3 4 5 6 2 3 4 5 6 62 63 64
                  65 66 52 53 54 55 56 42 43 44 45 46 32 33 34 35 36 52 53 54 55
                  56 32 33 34 35 36 62 63 64 65 66 42 43 44 45 46 42 43 44 45 46
                  62 63 64 65 66 32 33 34 35 36 52 53 54 55 56 32 33 34 35 36 42
                  43 44 45 46 52 53 54 55 56 62 63 64 65 66}, 5);
                  t3 = {17 8 68 58 48}` @ j(1,5,1) || {17 68 48 8 58}` @
                  j(1,5,1) || {27 9 18 69 39}` @ j(1,5,1) || {27 69 9 39 18}` @
                  j(1,5,1);
                  t4 = shape({2 6 5 4 3 3 2 6 5 4 4 3 2 6 5 5 4 3 2 6 6 5 4 3 2
                  2 5 3 6 4 3 6 4 2 5 4 2 5 3 6 5 3 6 4 2 6 4 2 5 3 2 4 6 3 5 3
                  5 2 4 6 4 6 3 5 2 5 2 4 6 3 6 3 5 2 4 2 3 4 5 6 3 4 5 6 2 4 5
                  6 2 3 5 6 2 3 4 6 2 3 4 5}, 20);
                  t1 = (sqrsym({1 11 21 11 1 21 21 11 21 11 21 21 11 1 11}) ||
                       j(5,1,1) @ (2:6) || t2) // ((2:6)` @ j(1,5,-1) ||
                       j(5,5,7) || t3) // (-t4 || t3` || t1);
                  t2 = {[6]0 1 1 2 2 1 2 0 1 2 1 0 2 2 1 2 1 2 1 0 2 2 1 0 1}`
                  @ j(1, 5, 1) // shape({ 0 1 1 2 2 1 2 0 1 2 1 0 2 2 1 2 1 2 1
                  0 2 2 1 0 1 0 0 2 1 2 1 2 2 0 0 2 0 1 2 0 2 1 0 0 2 0 2 0 2 1
                  0 2 1 1 0 1 0 1 0 2 1 1 0 2 0 2 0 0 1 1 0 1 2 0 1}, 15);
                  d150 = j(150, 150, .);
                  do i = 1 to 30;
                     r = (i - 1) # 5 + (1:5);
                     do j = 1 to 30;
                        k = t1[i,j];
                        if k ^= . then do;
                           t3 = t2[(mod(abs(k), 10) - 1) # 5 + (1:5),];
                           if      k <  0 then t3 = t3`;
                           else if k < 10 then ;
                           else if k < 20 then t3 = mod(t3 + 1, 3);
                           else if k < 30 then t3 = mod(t3 + 2, 3);
                           else if k < 40 then t3 = t3[{2 3 4 5 1},];
                           else if k < 50 then t3 = t3[{3 4 5 1 2},];
                           else if k < 60 then t3 = t3[{4 5 1 2 3},];
                           else if k < 70 then t3 = t3[{5 1 2 3 4},];
                           d150[r, (j - 1) # 5 + (1:5)] = t3;
                           if k = t1[j,i] & i ^= j then do;
                              d150[(j - 1) # 5 + (1:5), r] = t3;
                              t1[j,i] = .;
                              end;
                           end;
                        end;
                     end;
                  free i j k r t1 t2 t3 t4;
                  if n = 150 then d = d150;
                  end;

               if any(n = {153}) then do;
                  d153 = ks(d3, d51, 3);
                  if n = 153 then d = d153;
                  end;

               if any(n = {156}) then do;
                  d156 = cat(d48, d108);
                  if n = 156 then d = d156;
                  end;

               if any(n = {159}) then do;
                  d159 = cat(d48, d111);
                  if n = 159 then d = d159;
                  end;

               if any(n = {162}) then do;
                  d162 = ks(d3, d54, 3);
                  if n = 162 then d = d162;
                  end;

               if any(n = {165}) then do;
                  d165 = cat(d81, d84);
                  if n = 165 then d = d165;
                  end;

               if any(n = {168}) then do;
                  d168 = halfhad(168, a1, a2, d24);
                  if n = 168 then d = d168;
                  end;

               if any(n = {171}) then do;
                  d171 = ks(d3, d57, 3);
                  if n = 171 then d = d171;
                  end;

               end;

            else if f = 4 then do;

               run fieldop(4, add, sub, d4, inv); free add sub inv;
               d3 = mod({0 1 2}` * {0 1 2}, 3);
               a   = d4[2:4,2:4];
               h00 = d4[,1]; h01 = d4[,2]; h02 = d4[,3]; h03 = d4[,4];
               z3  = {0 0 0}`;
               l9  = ks((0:2)`, d3, 3) || {1 1 1}` @ (0:2)`;
               q1  = q(l9, 2); q2  = q(l9, 3); q3  = q(l9, 4);
               free l9;
               if n = 4 then d = d4;

               if any(n = {8 32 44 56 68 88 92 96 104 112 120 128}) then do;
                  run fieldop(8, add, sub, d8, inv); free add sub inv;
                  d8 = mod(d8, 4);
                  if n = 8 then d = d8;
                  end;

               if any(n = {12 28 36 44 48 52 60 68 72 76 84 92 96 100
                           104 108 116 124}) then do;
                  d12 = j(12, 12, 0);
                  e00 = h01 || h02 || h03;
                  d12 = ks(a, z3, 4);
                  h10 = ks(e00, z3, 4);
                  h11 = (d12`)[,1:3] // (q1 * d12);
                  h12 = (d12`)[,4:6] // (q2 * d12);
                  h13 = (d12`)[,7:9] // (q3 * d12);
                  d12  = h10 || h11 || h12 || h13;
                  free e00;
                  if n = 12 then d = d12;
                  end;

               if any(n = {16 28 40 52 64 76 100 120 124})
                  then do;
                  d16 = ks2(d4, 4);
                  if n = 16 then d = d16;
                  end;

               if any(n = {20 80 116}) then do;
                  d20 = j(20, 10, 0);
                  d20[2:20, 2:10] = shape({ 0 0 3 1 1 2 1 2 2 0 1 2 1 3 0 2 0 1
                  0 2 3 3 1 1 2 3 3 0 3 2 0 1 3 0 1 2 1 0 1 0 3 2 2 2 3 1 0 1 2
                  0 3 2 1 0 1 2 0 3 2 3 0 2 0 1 2 2 3 3 1 3 3 1 1 3 1 3 2 0 1 0
                  2 2 0 3 2 3 2 3 1 2 2 1 0 2 0 0 1 3 3 2 1 1 1 2 1 3 3 2 2 2 0
                  0 2 3 1 2 1 2 3 3 1 2 1 0 0 3 3 1 2 2 1 1 1 2 0 3 1 2 3 3 2 0
                  1 0 3 2 0 1 0 0 3 1 1 3 3 1 2 1 3 3 0 3 3 3 3 0 0 2 2 3 1},
                  19);
                  if n = 20 then d = d20;
                  end;

               if any(n = {24 40 60 76 124}) then do;
                  d24 = j(24, 20, 0);
                  d24[2:24, 2:20] = shape({ 0 0 0 0 2 2 2 2 2 3 3 3 3 3 1 1 1 1
                  1 0 0 0 0 3 3 3 3 3 1 1 1 1 1 2 2 2 2 2 0 3 1 1 0 2 1 2 1 1 3
                  3 2 1 3 3 0 2 2 0 3 1 1 1 1 2 0 0 3 1 2 1 2 0 2 2 3 0 0 3 1 1
                  2 0 0 1 2 2 2 1 3 3 2 0 3 0 3 1 0 2 3 0 3 2 1 3 1 2 0 1 3 3 2
                  3 1 0 1 0 2 3 1 0 1 3 2 3 0 1 2 0 0 3 1 0 2 1 0 2 3 2 1 3 2 0
                  2 3 2 0 1 2 1 0 3 1 1 3 3 2 0 1 3 3 2 0 1 3 3 2 0 1 3 3 2 1 3
                  3 2 2 3 1 1 0 3 2 0 0 1 1 0 2 2 3 1 3 3 2 3 2 0 0 1 1 0 2 2 3
                  2 3 1 1 0 2 1 2 1 0 0 3 1 1 0 2 2 3 0 3 1 2 1 2 2 1 2 1 1 3 0
                  3 0 3 3 0 2 2 1 3 3 2 1 2 1 2 1 3 1 1 0 3 2 0 3 0 3 2 2 1 3 3
                  2 2 3 0 1 3 3 2 1 2 0 0 1 2 3 1 1 0 3 2 2 3 0 2 2 1 3 3 1 1 2
                  0 0 0 0 3 1 1 2 2 3 0 3 1 2 1 2 0 2 1 2 1 1 3 0 3 0 3 1 0 2 1
                  2 0 1 3 2 1 3 2 0 3 0 2 3 1 3 1 0 2 2 3 2 0 1 1 0 1 3 2 0 1 0
                  2 3 3 1 0 2 3 0 1 2 0 0 3 2 1 3 1 2 3 0 2 3 2 1 3 0 1 0 2 3 0
                  3 1 0 2 3 0 1 2 0 3 2 1 3 1 2 3 0 2 3 2 3 1 0 1 2 0 1 3 3 2 1
                  3 3 0 2 3 1 2 1 0 3 1 2 3 2 0 1}, 23);
                  if n = 24 then d = d24;
                  end;

               if any(n = {28}) then do;
                  d28 = cat(d12, d16);
                  if n = 28 then d = d28;
                  end;

               if any(n = {32 44 68 88 128}) then do;
                  d32 = ks(d4, d8, 4);
                  if n = 32 then d = d32;
                  end;

               if any(n = {36 52 60 68 72 76 84 92 100 104 108 116 124})
                  then do;
                  e10 = h11 || h12 || h13;
                  e11 = rowop(ks(h00, z3, 4) || e10, ks(h01, z3, 4), 4)[,2:10];
                  e12 = rowop(ks(h00, z3, 4) || e10, ks(h02, z3, 4), 4)[,2:10];
                  e13 = rowop(ks(h00, z3, 4) || e10, ks(h03, z3, 4), 4)[,2:10];
                  h20 = ks(e10, z3, 4);
                  h21 = i(4) @ q1 * ks(e11, z3, 4);
                  h22 = i(4) @ q2 * ks(e12, z3, 4);
                  h23 = i(4) @ q3 * ks(e13, z3, 4);
                  d36 = h20 || h21 || h22 || h23;
                  free e10 e11 e12 e13;
                  if n = 36 then d = d36;
                  end;

               if any(n = {40}) then do;
                  d40 = cat(d16, d24);
                  if n = 40 then d = d40;
                  end;

               if any(n = {44}) then do;
                  d44 = cat(d12, d32);
                  if n = 44 then d = d44;
                  end;

               if any(n = {48 84 104}) then do;
                  d48 = ks(nrmz(d12, 4), d4, 4);
                  if n = 48 then do; d = d48; dnz = 0; end;
                  end;

               if any(n = {52}) then do;
                  d52 = cat(d16, d36);
                  if n = 52 then d = d52;
                  end;

               if any(n = {56 88 92 104 112 120}) then do;
                  d56 = ghr(mod((0:6)` * (0:6), 7), d8);
                  if n = 56 then d = d56;
                  end;

               if any(n = {60 76}) then do;
                  d60 = cat(d24, d36);
                  if n = 60 then d = d60;
                  end;

               if any(n = {64 100 120 124}) then do;
                  d64 = ks(d4, d16, 4);
                  if n = 64 then d = d64;
                  end;

               if any(n = {68}) then do;
                  d68 = cat(d32, d36);
                  if n = 68 then d = d68;
                  end;

               if any(n = {72}) then do;
                  d72 = d36 // d36;
                  if n = 72 then d = d72;
                  end;

               if any(n = {76}) then do;
                  d76 = cat(d16, d60);
                  if n = 76 then d = d76;
                  end;

               if any(n = {80 116}) then do;
                  d80 = ks(d4, d20, 4);
                  if n = 80 then d = d80;
                  end;

               if any(n = {84}) then do;
                  d84 = cat(d36, d48);
                  if n = 84 then d = d84;
                  end;

               if any(n = {88}) then do;
                  d88 = cat(d32, d56);
                  if n = 88 then d = d88;
                  end;

               if any(n = {92}) then do;
                  d92 = cat(d36, d56);
                  if n = 92 then d = d92;
                  end;

               if any(n = {96}) then do;
                  d96 = ks(d8, nrmz(d12, 4), 4);
                  if n = 96 then do; d = d96; dnz = 0; end;
                  end;

               if any(n = {100 124}) then do;
                  d100 = cat(d36, d64);
                  if n = 100 then d = d100;
                  end;

               if any(n = {104}) then do;
                  d104 = cat(d48, d56);
                  if n = 104 then d = d104;
                  end;

               if any(n = {108}) then do;
                  e20 = h21 || h22 || h23;
                  e21 = rowop(ks(h10, z3, 4) || e20, ks(h11, z3, 4), 4)[,4:30];
                  e22 = rowop(ks(h10, z3, 4) || e20, ks(h12, z3, 4), 4)[,4:30];
                  e23 = rowop(ks(h10, z3, 4) || e20, ks(h13, z3, 4), 4)[,4:30];
                  h30 = ks(e20, z3, 4);
                  h31 = i(12) @ q1 * ks(e21, z3, 4);
                  h32 = i(12) @ q2 * ks(e22, z3, 4);
                  h33 = i(12) @ q3 * ks(e23, z3, 4);
                  d108 = h30 || h31 || h32 || h33;
                  free e20 e21 e22 e23;
                  if n = 108 then d = d108;
                  end;

               if any(n = {112}) then do;
                  d112 = d56 // d56;
                  if n = 112 then d = d112;
                  end;

               if any(n = {116}) then do;
                  d116 = cat(d36, d80);
                  if n = 116 then d = d116;
                  end;

               if any(n = {120}) then do;
                  d120 = cat(d56, d64);
                  if n = 120 then d = d120;
                  end;

               if any(n = {124}) then do;
                  d124 = cat(d24, d100);
                  if n = 124 then d = d124;
                  end;

               if any(n = {128}) then do;
                  d128 = ks(d4, d32, 4);
                  if n = 128 then d = d128;
                  end;

               end;

            else if f = 5 then do;

               d5 = mod((0:4)` * (0:4), 5);
               d4 = { 0 0 0 0, 0 1 2 3, 0 2 3 1, 0 3 1 2};
               a   = d5[2:5,2:5];
               h00 = d5[,1]; h01 = d5[,2]; h02 = d5[,3];
               h03 = d5[,4]; h04 = d5[,5];
               z   = j(4, 1, 0);
               l16 = ks((0:3)`, d4, 4) || (1+z) @ (0:3)`;
               q1  = q(l16, 2); q2 = q(l16, 3); q3 = q(l16, 4); q4 = q(l16, 5);
               free l16;
               if n = 5 then d = d5;

               if any(n = {10 30 35 50 55 70 90}) then do;
                  d10 = doubledif(5);
                  if n = 10 then do; d = d10; dnz = 0; end;
                  end;

               if any(n = {15 75 95}) then do;
                  d15 = j(15, 8, 0);
                  d15[2:15, 2:8] = shape( { 0 0 4 3 2 2 3 0 3 2 2 4 3 4 1 2 2 3
                  3 0 1 1 3 4 0 1 4 2 1 4 0 1 4 2 2 2 1 3 2 3 2 1 2 2 4 1 0 1 4
                  2 4 2 4 1 3 0 3 1 1 4 4 0 3 3 2 3 1 2 4 4 3 4 1 4 3 1 0 4 0 3
                  3 1 1 3 4 1 0 2 0 4 1 4 3 1 0 2 3 2}, 14);
                  if n = 15 then d = d15;
                  end;

               if any(n = {20 40 45 60 65 70 80 85 90 95 100}) then do;
                  e00 = h01 || h02 || h03 || h04;
                  d20 = ks(a, z, 5);
                  h10 = ks(e00, z, 5);
                  h11 = (d20`)[, 1: 4] // (q1 * d20);
                  h12 = (d20`)[, 5: 8] // (q2 * d20);
                  h13 = (d20`)[, 9:12] // (q3 * d20);
                  h14 = (d20`)[,13:16] // (q4 * d20);
                  d20  = h10 || h11 || h12 || h13 || h14;
                  free e00;
                  if n = 20 then d = d20;
                  end;

               if any(n = {25 35 45 55 65 85}) then do;
                  d25 = ks2(d5, 5);
                  if n = 25 then d = d25;
                  end;

               if any(n = {30 55}) then do;
                  d30 = d10[,1:9] @ {[3]1}` || shape( { 0 0 1 4 3 2 1 2 3 3 4 0
                  0 3 2 1 4 4 4 0 1 2 3 3 3 0 0 2 1 4 1 1 4 3 2 0 2 4 3 1 0 2 4
                  1 0 3 2 4 0 4 2 1 3 2 4 0 2 3 1 1}, 30);
                  if n = 30 then d = d30;
                  end;

               if any(n = {35}) then do;
                  d35 = cat(d10, d25);
                  if n = 35 then d = d35;
                  end;

               if any(n = {40 60 80}) then do;
                  d40 = d20 // d20;
                  if n = 40 then d = d40;
                  end;

               if any(n = {45 65 85}) then do;
                  d45 = cat(d20, d25);
                  if n = 45 then d = d45;
                  end;

               if any(n = {50 70 90}) then do;
                  d50 = ks(d5, d10, 5);
                  if n = 50 then d = d50;
                  end;

               if any(n = {55}) then do;
                  d55 = cat(d30, d25);
                  if n = 55 then d = d55;
                  end;

               if any(n = {60}) then do;
                  d60 = cat(d20, d40);
                  if n = 60 then d = d60;
                  end;

               if any(n = {65 85}) then do;
                  d65 = cat(d20, d45);
                  if n = 65 then d = d65;
                  end;

               if any(n = {70 90}) then do;
                  d70 = cat(d20, d50);
                  if n = 70 then d = d70;
                  end;

               if any(n = {75 95}) then do;
                  d75 = ks(d5, d15, 5);
                  if n = 75 then d = d75;
                  end;

               if any(n = {80}) then do;
                  e10 = h11 || h12 || h13 || h14;
                  e11 = rowop(ks(h00, z, 5) || e10, ks(h01, z, 5), 5);
                  e11 = rowop(ks(h00, z, 5) || e10, ks(h01, z, 5), 5)[,2:17];
                  e12 = rowop(ks(h00, z, 5) || e10, ks(h02, z, 5), 5)[,2:17];
                  e13 = rowop(ks(h00, z, 5) || e10, ks(h03, z, 5), 5)[,2:17];
                  e14 = rowop(ks(h00, z, 5) || e10, ks(h04, z, 5), 5)[,2:17];
                  h20 = ks(e10, z, 5);
                  h21 = i(5) @ q1 * ks(e11, z, 5);
                  h22 = i(5) @ q2 * ks(e12, z, 5);
                  h23 = i(5) @ q3 * ks(e13, z, 5);
                  h24 = i(5) @ q4 * ks(e14, z, 5);
                  d80 = h20 || h21 || h22 || h23 || h24;
                  free e10 e11 e12 e13 e14;
                  if n = 80 then d = d80;
                  end;

               if any(n = {85}) then do;
                  d85 = cat(d20, d65);
                  if n = 85 then d = d85;
                  end;

               if any(n = {90}) then do;
                  d90 = ghr(facx(9, 9, 9), nrmz(d10, 5));
                  if n = 90 then d = d90;
                  end;

               if any(n = {95}) then do;
                  d95 = cat(d20, d75);
                  if n = 95 then d = d95;
                  end;

               if any(n = {100}) then do;
                  d100 = ks(d5, nrmz(d20, 5), 5);
                  if n = 100 then do; d = d100; dnz = 0; end;
                  end;

               end;

            else if f = 6 then do;

               if any(n = {6 18 30 42 54 66 72 78}) then do;
                  d6 = j(6, 1, 0) || (0:5)`;
                  if n = 6 then d = d6;
                  end;

               if any(n = {12 24 36 60 72}) then do;
                  d12 = j(12, 6, 0);
                  d12[,2] = j(2, 1, 1) @ (0:5)`;
                  d12[2:12, 3:6] = shape( { 3 2 4 0 0 1 5 2 1 5 4 2 3 5 2 1 5
                  3 1 1 2 3 2 3 2 4 0 5 5 2 3 4 4 1 1 4 1 0 3 5 4 4 5 3}, 11);
                  if n = 12 then d = d12;
                  end;

               if any(n = {18}) then do;
                  d18 = j(3, 1, 1) @ d6;
                  if n = 18 then d = d18;
                  end;

               if any(n = {24}) then do;
                  d24 = j(2, 1, 1) @ d12;
                  if n = 24 then d = d24;
                  end;

               if any(n = {30}) then do;
                  d30 = j(5, 1, 1) @ d6;
                  if n = 30 then d = d30;
                  end;

               if any(n = {36}) then do;
                  d36 = j(36, 7, 0);
                  d36[,2] = (0:5)` @ j(6, 1, 1);
                  d36[2:36, 3:7] = shape( { 1 5 2 3 1 2 1 3 5 2 3 0 1 5 2 3 5 2
                  3 4 5 3 5 4 3 0 0 5 2 3 1 2 4 2 0 2 2 0 1 3 3 5 1 1 1 5 1 3 3
                  2 5 4 1 0 4 1 3 5 1 2 1 4 2 4 5 3 0 3 4 3 4 2 0 3 0 4 4 3 1 5
                  5 0 5 0 0 0 5 3 0 4 1 0 0 5 5 1 3 4 3 2 2 5 5 5 1 3 4 1 0 1 4
                  2 1 4 0 0 2 3 5 2 0 4 2 1 3 2 1 2 4 4 4 1 5 2 3 4 3 0 2 1 5 5
                  4 2 5 0 1 1 5 5 2 2 2 3 5 2 3 4 2 4 3 4 4 1 0 4 1 0 0 4 5 3 4
                  4 1}, 35);
                  if n = 36 then d = d36;
                  end;

               if any(n = {42}) then do;
                  d42 = j(7, 1, 1) @ d6;
                  if n = 42 then d = d42;
                  end;

               if any(n = {48}) then do;
                  d48 = j(48, 10, 0);
                  d48[,2] = (0:5)` @ j(8, 1, 1);
                  d48[4:48, 3:10] = shape({ 4 2 4 1 3 3 1 2 5 2 5 1 3 3 1 2 5 2
                  5 1 4 4 1 2 5 3 5 1 4 4 1 3 5 3 5 2 4 4 2 3 1 0 3 3 2 4 4 2 1
                  0 3 3 2 5 5 2 1 0 4 4 2 5 5 2 1 0 4 4 3 5 5 3 2 3 4 0 1 2 3 4
                  2 3 5 0 1 2 3 5 2 4 5 0 1 2 4 5 3 4 5 0 1 3 4 5 2 0 4 4 3 5 5
                  3 3 3 2 4 0 4 1 1 3 3 2 5 0 5 1 1 3 4 5 0 2 3 4 5 4 1 3 2 4 1
                  3 0 4 4 2 5 0 5 1 1 5 1 3 2 5 1 3 0 5 1 4 2 5 1 4 0 0 1 1 4 2
                  3 2 4 0 1 1 5 2 3 2 5 1 4 2 1 4 2 0 3 1 5 2 1 5 2 0 3 4 4 3 5
                  0 5 1 1 4 4 3 5 0 5 2 2 5 1 4 3 5 1 4 0 5 2 4 3 5 2 4 0 0 1 1
                  5 2 4 2 5 0 1 1 5 3 4 3 5 0 2 2 5 3 4 3 5 1 5 2 1 5 2 0 4 1 5
                  3 1 5 3 0 4 2 4 0 3 3 1 2 1 2 5 3 2 5 3 0 4 3 2 1 2 1 0 4 3 2
                  5 0 3 3 1 2 1 2 5 0 4 4 1 2 1 3 2 1 2 1 0 5 3 3 5 0 4 4 1 3 1
                  3 5 0 4 4 2 3 2 4 2 1 2 1 0 5 4 4 3 1 3 1 0 5 4 4 3 2 3 2 0 5
                  4}, 45);
                  if n = 48 then d = d48;
                  end;

               if any(n = {54}) then do;
                  d54 = j(9, 1, 1) @ d6;
                  if n = 54 then d = d54;
                  end;

               if any(n = {60}) then do;
                  d60 = j(60, 8, 0);
                  d60[,2] = (0:5)` @ j(10, 1, 1);
                  d60[2:60, 3:8] = shape( { 0 1 2 3 3 1 0 2 5 4 3 5 0 3 2 5 3 0
                  1 5 1 5 0 3 2 2 5 3 1 2 3 2 2 3 4 3 3 4 3 4 1 0 4 4 2 1 2 4 5
                  5 2 3 1 1 0 0 1 5 5 2 0 3 4 1 3 3 1 3 2 0 2 5 2 0 3 5 4 4 2 1
                  0 0 0 4 2 1 4 1 2 4 3 5 4 3 1 2 4 1 0 2 4 4 5 1 4 1 4 0 5 4 4
                  2 1 4 0 0 3 2 1 0 2 4 0 3 2 0 2 5 0 2 5 4 3 0 1 1 0 1 3 0 4 4
                  2 4 3 5 0 3 5 3 4 3 1 0 4 2 4 3 3 4 5 1 4 4 5 0 0 1 5 5 3 5 1
                  1 0 1 1 5 2 3 0 4 3 5 4 5 1 1 5 1 5 2 1 5 2 2 4 3 2 3 1 4 4 1
                  2 4 2 2 2 3 3 2 4 5 0 1 4 2 1 1 3 5 5 0 1 1 1 1 5 1 3 4 5 2 0
                  5 0 4 3 5 1 3 1 0 2 0 1 4 0 4 0 0 1 4 4 3 0 5 1 5 1 4 2 5 2 3
                  3 5 2 5 2 3 5 3 3 2 3 2 2 1 5 0 5 3 5 0 0 4 5 4 4 0 3 3 1 1 3
                  3 0 2 1 2 5 0 5 1 3 0 5 1 4 4 3 1 5 2 1 0 4 0 3 2 4 5 4 2 0 5
                  3 2 4 2 4 0 1 5 4 5 0 4 5 3 5 0 5 2 3 2 5 2 2 2 5 3}, 59);
                  if n = 60 then d = d60;
                  end;

               if any(n = {66}) then do;
                  d66 = j(11, 1, 1) @ d6;
                  if n = 66 then d = d66;
                  end;

               if any(n = {72}) then do;
                  d72 = ks(d6, d12, 6);
                  if n = 72 then d = d72;
                  end;

               if any(n = {78}) then do;
                  d78 = j(13, 1, 1) @ d6;
                  if n = 78 then d = d78;
                  end;

               if any(n = {84}) then do;
                  d84 = j(84, 8, 0);
                  d84[,2] = (0:5)` @ j(14, 1, 1);
                  d84[2:84, 3:8] = shape( { 0 3 2 4 3 0 1 4 5 0 4 2 1 4 5 1 0 0
                  1 5 2 0 2 5 2 2 5 3 3 4 3 0 4 5 1 5 3 0 5 5 4 1 3 5 3 5 5 4 3
                  5 5 5 2 2 5 0 0 1 4 2 5 3 5 2 4 3 5 4 1 4 4 3 5 5 5 1 3 3 0 1
                  4 4 3 3 0 2 3 2 3 3 0 4 2 2 2 0 1 0 4 3 1 0 1 2 3 5 2 4 2 3 4
                  1 0 4 2 3 4 1 5 1 2 4 5 3 2 0 2 5 3 3 2 3 3 3 3 4 4 0 3 3 4 0
                  3 1 4 1 5 4 0 2 4 2 0 0 1 4 5 4 1 0 4 5 0 2 3 0 1 1 0 2 4 5 4
                  5 0 5 1 1 5 0 1 3 2 4 1 4 2 5 2 5 2 2 3 1 4 3 1 1 3 2 2 3 1 1
                  4 0 2 4 5 4 4 1 3 3 4 5 4 3 3 2 1 3 4 4 2 4 0 3 4 4 3 0 5 2 5
                  1 0 3 3 0 5 2 1 1 1 5 0 0 5 0 0 4 0 3 5 3 1 3 0 4 0 5 3 2 1 0
                  3 2 3 4 1 2 1 4 3 2 1 5 1 0 3 1 2 1 1 1 0 1 2 3 0 5 5 4 2 3 1
                  3 5 1 3 1 1 2 3 0 3 4 3 1 1 2 4 1 1 5 0 3 5 0 5 3 4 5 5 1 0 0
                  2 1 0 2 3 4 5 5 1 0 2 5 2 5 1 2 4 2 5 3 2 0 0 5 3 5 2 0 5 4 1
                  2 2 1 1 2 2 4 2 1 5 1 0 4 3 5 2 0 2 4 4 2 2 4 0 0 4 3 4 1 5 5
                  4 5 3 2 5 2 4 5 4 2 0 1 5 5 1 3 2 0 5 5 4 2 4 5 0 0 4 5 5 1 0
                  5 0 3 4 2 1 1 3 0 1 3 1 2 0 1 3 2 1 4 0 1 2 4 2 4 0 2 0 1 3 0
                  1 5 5 5 3 1 4 4 1 0 3 3 0 4 0 3 4 2 2 1 4 3 4 4 1 2 0 5 5 1 2
                  2 5 2 5 3 0 3 4 0 5 4 2 0 2 1}, 83);
                  if n = 84 then d = d84;
                  end;

               end;

            else if f = 7 then do;

               if any(n = {7 21 35 49 63}) then do;
                  d7 = mod((0:6)` * (0:6), 7);
                  if n =  7 then d = d7;
                  end;

               if any(n = {14 42 56 63}) then do;
                  d14 = doubledif(7);
                  if n = 14 then d = d14;
                  end;

               if any(n = {21}) then do;
                  d21 = d7[,1:5 || 7] @ {[3]1}` || shape( { 0 0 0 5 4 5 4 5 4 3
                  5 2 2 3 5 5 2 3 3 1 0 0 3 1 1 0 3 1 6 5 6 5 1 5 1 6 4 4 6 6 3
                  3 3 6 4 1 1 4 2 4 1 4 2 2 0 2 6 2 6 0 6 0 2}, 21);
                  if n = 21 then d = d21;
                  end;

               if any(n = {28 42 56 70}) then do;
                  d28 = gh4(7, {1 1 1 3});
                  if n = 28 then d = d28;
                  end;

               if any(n = {35}) then do;
                  d35 = d7[,1:5 || 7] @ {[5]1}` || shape( { 0 0 0 0 6 6 4 3 5 3
                  6 4 1 4 5 1 4 5 2 6 4 5 2 2 6 3 5 4 5 4 3 6 2 1 4 1 3 2 0 5 3
                  6 1 2 0 1 0 3 2 2 5 0 1 3 2 4 4 0 3 6 1 4 3 1 0 1 1 0 5 0 0 2
                  2 5 5 5 6 6 6 4 0 4 3 3 5 2 4 0 4 6 1 2 6 3 5 6 3 5 6 5 3 0 4
                  4 4 4 1 5 5 3 6 1 1 2 3 0 2 1 2 3 1 6 4 5 3 2 0 2 6 1 6 3 0 5
                  2 1 2 0 1 6}, 35);
                  if n = 35 then d = d35;
                  end;

               if any(n = {42 70}) then do;
                  d42 = j(42, 18, 0);
                  d42[,2] = (0:6)` @ j(6, 1, 1);
                  d42[2:42, 3:18] = shape({ 0 0 0 0 1 1 1 1 1 1 2 2 2 2 2 2 0 0
                  0 0 2 2 2 2 2 2 4 4 4 4 4 4 0 0 0 0 3 3 3 3 3 3 6 6 6 6 6 6 0
                  0 0 0 4 4 4 4 4 4 1 1 1 1 1 1 0 0 0 0 5 5 5 5 5 5 3 3 3 3 3 3
                  2 3 4 5 6 0 1 2 3 4 5 6 0 1 2 3 2 3 4 5 0 2 4 6 1 3 6 1 3 5 0
                  2 2 3 4 5 1 4 0 3 6 2 0 3 6 2 5 1 2 3 4 5 2 6 3 0 4 1 1 5 2 6
                  3 0 2 3 4 5 3 1 6 4 2 0 2 0 5 3 1 6 2 3 4 5 4 3 2 1 0 6 3 2 1
                  0 6 5 4 6 1 3 5 6 0 1 2 3 2 1 0 6 5 4 4 6 1 3 6 1 3 5 0 2 4 5
                  6 0 1 2 4 6 1 3 0 3 6 2 5 1 5 0 2 4 6 1 4 6 1 3 1 5 2 6 3 0 6
                  2 5 1 4 0 4 6 1 3 2 0 5 3 1 6 0 4 1 5 2 6 4 6 1 3 3 2 1 0 6 5
                  1 6 4 2 0 5 6 2 5 1 4 5 6 0 1 2 0 5 3 1 6 4 6 2 5 1 5 0 2 4 6
                  1 1 0 6 5 4 3 6 2 5 1 6 2 5 1 4 0 3 4 5 6 0 1 6 2 5 1 0 4 1 5
                  2 6 4 6 1 3 5 0 6 2 5 1 1 6 4 2 0 5 5 1 4 0 3 6 6 2 5 1 2 1 0
                  6 5 4 6 3 0 4 1 5 1 5 2 6 3 4 5 6 0 1 5 2 6 3 0 4 1 5 2 6 4 6
                  1 3 5 0 6 4 2 0 5 3 1 5 2 6 5 1 4 0 3 6 0 6 5 4 3 2 1 5 2 6 6
                  3 0 4 1 5 2 3 4 5 6 0 1 5 2 6 0 5 3 1 6 4 3 5 0 2 4 6 1 5 2 6
                  1 0 6 5 4 3 4 0 3 6 2 5 3 1 6 4 2 3 4 5 6 0 3 6 2 5 1 4 3 1 6
                  4 3 5 0 2 4 6 4 1 5 2 6 3 3 1 6 4 4 0 3 6 2 5 5 3 1 6 4 2 3 1
                  6 4 5 2 6 3 0 4 6 5 4 3 2 1 3 1 6 4 6 4 2 0 5 3 1 2 3 4 5 6 3
                  1 6 4 0 6 5 4 3 2 2 4 6 1 3 5 5 4 3 2 1 2 3 4 5 6 1 3 5 0 2 4
                  5 4 3 2 2 4 6 1 3 5 2 5 1 4 0 3 5 4 3 2 3 6 2 5 1 4 3 0 4 1 5
                  2 5 4 3 2 4 1 5 2 6 3 4 2 0 5 3 1 5 4 3 2 5 3 1 6 4 2 5 4 3 2
                  1 0 5 4 3 2 6 5 4 3 2 1 0 1 2 3 4 5}, 41);
                  if n = 42 then d = d42;
                  end;

               if any(n = {49 63}) then do;
                  d49 = ks2(d7, 7);
                  if n = 49 then d = d49;
                  end;

               if any(n = {56}) then do;
                  d56 = d28 // d28;
                  if n = 56 then d = d56;
                  end;

               if any(n = {63}) then do;
                  d63 = cat(d14, d49);
                  if n = 63 then d = d63;
                  end;

               if any(n = {70}) then do;
                  d70 = cat(d28, d42);
                  if n = 70 then d = d70;
                  end;

               end;

            else if f = 8 then do;

               if any(n = {8 24 40 56 64}) then do;
                  run fieldop(8, add, sub, mul, inv); free add sub inv;
                  d8 = mul;
                  if n = 8 then d = d8;
                  end;

               if any(n = {16 24 48}) then do;
                  run fieldop(16, add, sub, d16, inv); free add sub inv;
                  d16 = mod(d16, 8);
                  if n = 16 then d = d16;
                  end;

               if any(n = {24}) then do;
                  d24 = cat(d8, d16);
                  if n = 24 then d = d24;
                  end;

               if any(n = {32 48}) then do;
                  d32 = j(32, 32, 0);
                  d32[2:32, 2:32] = sqrsym( { 1 2 4 3 6 5 4 0 4 0 5 2 7 4 1 6 4
                  2 0 6 4 7 6 1 4 3 2 5 0 0 0 1 1 1 1 2 1 2 3 5 4 7 6 2 3 2 4 6
                  1 3 5 7 2 0 6 3 6 5 5 6 3 0 2 1 4 7 4 0 4 1 5 1 5 3 7 3 7 2 5
                  2 7 5 0 7 2 3 6 1 4 6 3 6 4 2 1 7 5 3 3 5 7 1 2 4 6 7 6 1 5 2
                  3 4 3 4 5 2 6 1 0 7 0 1 1 2 2 3 3 5 5 4 4 7 7 6 6 2 1 3 2 6 7
                  5 4 5 4 6 7 3 2 0 1 2 3 2 5 7 2 0 7 5 5 7 0 2 7 5 2 0 3 1 6 3
                  7 4 6 5 1 2 5 6 2 1 3 0 4 7 3 0 4 7 4 1 5 2 6 3 7 4 0 5 1 6 2
                  7 3 0 4 1 5 2 5 3 6 6 3 5 0 4 1 7 2 2 7 1 4 0 5 3 6 6 3 6 5 3
                  2 4 7 1 4 2 1 7 6 0 3 5 1 7 4 2 3 5 6 7 7 0 6 1 1 6 4 3 3 4 2
                  5 5 2 1 6 6 1 7 0 0 7 0 1 1 3 3 2 2 7 7 6 6 4 4 5 5 7 7 6 6 4
                  4 5 5 0 1 3 2 7 6 4 5 7 6 4 5 0 1 3 2 7 6 4 5 0 1 3 2 0 1 2 5
                  7 3 1 6 4 7 5 2 0 4 6 1 3 6 4 3 1 5 7 0 2 1 3 4 3 7 4 7 4 0 3
                  7 4 0 3 0 3 7 4 6 5 1 2 1 2 6 5 1 2 6 5 4 1 5 3 7 2 6 6 2 7 3
                  5 1 4 0 5 1 4 0 6 2 7 3 3 7 2 6 0 5 3 6 7 2 4 1 6 3 5 0 1 4 2
                  7 5 0 6 3 2 7 1 4 3 6 0 5 4 1 6 5 3 3 5 6 0 6 0 3 5 5 3 0 6 4
                  2 1 7 7 1 2 4 2 4 7 1 1 7 4 7 7 0 7 0 0 7 6 1 1 6 1 6 6 1 4 3
                  3 4 3 4 4 3 2 5 5 2 5 2 2 5});
                  if n = 32 then d = d32;
                  end;

               if any(n = {40}) then do;
                  d40 = d8[,{1 2 3 4 6 7 8}] @ {[5]1}` || shape({ 0 0 0 1 1 4 4
                  6 6 6 5 5 7 2 1 0 2 4 2 6 0 3 1 2 4 5 5 5 4 1 1 0 5 2 5 4 3 7
                  1 4 3 0 5 4 3 0 3 1 1 1 7 5 0 4 6 4 5 7 7 0 0 6 7 1 7 2 3 2 6
                  6 5 1 7 1 3 2 1 3 3 5 6 4 3 2 5 6 7 7 2 5 2 4 2 3 7 4 4 0 7 5
                  2 6 6 3 3 0 7 0 1 4 6 2 0 2 6 6 3 7 3 7}, 40);
                  if n = 40 then d = d40;
                  end;

               if any(n = {48}) then do;
                  d48 = cat(d16, d32);
                  if n = 48 then d = d48;
                  end;

               if any(n = {56}) then do;
                  d56 = ghr(mod((0:6)` * (0:6), 7), d8);
                  if n = 56 then d = d56;
                  end;

               if any(n = {64}) then do;
                  d64 = ks2(d8, 8);
                  if n = 64 then d = d64;
                  end;

               end;

            else if f = 9 then do;

               if any(n = {9 18 36 45}) then do;
                  run fieldop(9, add, sub, mul, inv); free add sub inv;
                  d9 = mul;
                  if n = 9 then d = d9;
                  end;

               if any(n = {18 36 45}) then do;
                  d18 = ddif2(9);
                  if n = 18 then d = d18;
                  end;

               if any(n = {18 27 45 54}) then do;
                  run fieldop(27, add, sub, d27, inv); free add sub inv;
                  d27 = mod(d27, 9);
                  if n = 27 then d = d27;
                  end;

               if n = 36 then do;
                  alpha = {0 0 0 0, 2 3 5 4, 1 4 8 7, 4 6 2 5};
                  beta  = {1 1 1 1, 1 6 5 4, 1 8 8 5, 1 7 3 4};
                  gamma = {0 7 4 8, 0 4 3 5, 0 3 7 6, 0 1 2 7};
                  d36 = makehad4(f, alpha + 1, beta + 1, gamma + 1);
                  free alpha beta gamma;
                  if n = 36 then d = d36;
                  end;

               if n = 45 then do;
                  d45 = cat(d18, d27);
                  if n = 45 then d = d45;
                  end;

               if n = 54 then do;
                  d54 = mod(ddif2(27), 9);
                  if n = 54 then d = d54;
                  end;

               end;

            else if f = 10 then do;

               if any(n = {10 30 50}) then do;
                  d10 = j(10, 1, 0) || (0:9)`;
                  if n = 10 then d = d10;
                  end;

               if any(n = {20 40}) then do;
                  d20 = j(20, 5, 0);
                  d20[2:20,2:5] = shape({ 0 2 1 5 1 4 5 8 1 7 2 3 2 4 8 6 2 6 4
                  1 3 0 3 9 3 8 1 6 4 3 8 4 4 5 9 2 5 3 0 7 5 8 7 1 6 1 9 0 6 2
                  4 3 7 1 3 8 7 5 6 5 8 7 7 9 8 9 5 7 9 6 2 2 9 9 6 4}, 19);
                  if n = 20 then d = d20;
                  end;

               if any(n = {30}) then do;
                  d30 = j(3, 1, 1) @ d10;
                  if n = 30 then d = d30;
                  end;

               if any(n = {40}) then do;
                  d40 = j(40, 6, 0); d40[,2] = (0:9)` @ j(4, 1, 1);
                  d40[2:40,3:6] = shape({ 6 5 9 6 9 1 6 8 9 7 2 6 1 0 3 3 4 3 8
                  8 4 9 1 2 7 8 2 8 0 7 7 3 3 7 6 1 5 1 3 0 9 4 5 0 5 3 1 6 7 4
                  0 3 8 1 7 6 8 1 8 5 1 6 0 7 1 9 9 1 4 7 4 9 5 5 9 2 2 6 4 4 3
                  8 4 2 6 3 8 9 7 9 6 9 0 0 7 8 1 2 3 7 4 0 2 5 8 5 4 5 2 3 1 7
                  3 6 9 7 6 4 7 1 7 9 5 1 0 4 5 3 2 8 6 4 6 2 1 0 9 8 0 9 2 2 3
                  5 3 5 8 2 5 6 5 4 8 2 2 4}, 39);
                  if n = 40 then d = d40;
                  end;

               if any(n = {50}) then do;
                  d50 = j(5, 1, 1) @ d10;
                  if n = 50 then d = d50;
                  end;

               end;

            else if f = 11 then do;

               if any(n = {11 33}) then do;
                  d11 = mod((0:10)` * (0:10), 11);
                  if n = 11 then d = d11;
                  end;

               else if any(n = {22 44}) then do;
                  d22 = doubledif(11);
                  if n = 22 then d = d22;
                  end;

               if any(n = {33}) then do;
                  d33 = j(3, 1, 1) @ d11;
                  if n = 33 then d = d33;
                  end;

               else if any(n = {44}) then do;
                  d44 = gh4(11, {1 1 3 6});
                  if n = 44 then d = d44;
                  end;

               end;

            else if f = 12 then do;

               if any(n = {12 24 36}) then do;
                  d12 = j(12, 6, 0);
                  d12[2:12, 2:6] = shape( { 1 6 3 8 4 2 8 6 1 11 3 2 1 11 10 4
                  7 9 5 2 5 1 11 9 7 6 9 2 3 8 7 11 8 10 6 8 4 5 2 9 9 10 4 7 1
                  10 5 7 6 3 11 3 10 4 5}, 11);
                  if n = 12 then d = d12;
                  end;

               if any(n = {24}) then do;
                  d24 = d12 // d12;
                  if n = 24 then d = d24;
                  end;

               if any(n = {36}) then do;
                  d36 = j(3, 1, 1) @ d12;
                  if n = 36 then d = d36;
                  end;

               end;

            else if f = 13 then do;

               if n = 13 then d = mod((0:12)` * (0:12), 13);

               else if n = 26 then d = doubledif(13);

               end;

            else if f = 14 then do;

               if n = 14 then do;
                  d14 = j(14, 1, 0) || (0:13)`;
                  if n = 14 then d = d14;
                  end;

               if n = 28 then do;
                  d28 = j(28, 5, 0); d28[,2] = (0:13)` @ j(2, 1, 1);
                  d28[2:28, 3:5] = shape({ 9 1 7 3 1 6 4 0 2 6 11 8 12 4 7 10
                  6 13 11 8 4 2 10 3 5 8 12 2 13 11 11 2 1 10 3 0 13 12 4 4 12
                  10 13 6 9 6 10 5 8 3 8 8 7 2 12 13 13 1 4 12 9 11 5 5 9 10 7
                  7 6 3 5 9 7 5 1 0 9 11 1 2 3}, 27);
                  if n = 28 then d = d28;
                  end;

               end;

            else if f = 15 then do;

               if any(n = {15 30}) then do;
                  d15 = j(15, 5, 0);
                  d15[2:15, 2:5] = shape({ 1 10 4 7 2 12 1 11 3 6 14 8 4 9 10 14
                  5 1 13 4 6 3 8 10 7 5 11 3 8 14 6 5 9 8 3 1 10 11 7 12 11 13
                  12 9 12 4 2 13 13 2 5 6 14 7 9 2}, 14);
                  if n = 15 then d = d15;
                  end;

               if any(n = {30}) then do;
                  d30 = d15 // d15;
                  if n = 30 then d = d30;
                  end;

               end;

            else if f = 16 then do;

               if n = 16 then do;
                  run fieldop(16, summat, sub, mul, inv); free sub inv;
                  d16 = mul;
                  if n = 16 then d = d16;
                  end;

               if n = 32 then do;
                  d32 = j(32, 32, 0);
                  d32[2:32, 2:32] = shape({ 15 14 13 12 11 10 9 8 7 6 5 4 3 2
                  1 0 15 14 13 12 11 10 9 8 7 6 5 4 3 2 1 7 5 3 1 15 13 11 9 7
                  5 3 1 15 13 11 9 14 12 10 8 6 4 2 0 14 12 10 8 6 4 2 8 11 14
                  13 4 7 2 1 0 3 6 5 12 15 10 9 1 2 7 4 13 14 11 8 9 10 15 12
                  5 6 3 7 3 15 11 7 3 15 11 14 10 6 2 14 10 6 2 5 1 13 9 5 1
                  13 9 12 8 4 0 12 8 4 8 13 2 7 12 9 6 3 9 12 3 6 13 8 7 2 10
                  15 0 5 14 11 4 1 11 14 1 4 15 10 5 0 6 12 10 8 14 4 2 9 15 5
                  3 1 7 13 11 11 13 7 1 3 5 15 9 2 4 14 8 10 12 6 15 8 1 6 3 4
                  13 10 14 9 0 7 2 5 12 11 4 3 10 13 8 15 6 1 5 2 11 12 9 14 7
                  14 6 14 6 7 15 7 15 12 4 12 4 5 13 5 13 3 11 3 11 10 2 10 2
                  1 9 1 9 8 0 8 1 8 3 10 12 5 14 7 11 2 9 0 6 15 4 13 12 5 14
                  7 1 8 3 10 6 15 4 13 11 2 9 9 3 13 7 8 2 12 6 11 1 15 5 10 0
                  14 4 13 7 9 3 12 6 8 2 15 5 11 1 14 4 10 6 13 0 11 3 8 5 14
                  12 7 10 1 9 2 15 4 2 9 4 15 7 12 1 10 8 3 14 5 13 6 11 9 5 1
                  13 0 12 8 4 2 14 10 6 11 7 3 15 6 10 14 2 15 3 7 11 13 1 5 9
                  4 8 12 6 11 12 1 11 6 1 12 5 8 15 2 8 5 2 15 9 4 3 14 4 9 14
                  3 10 7 0 13 7 10 13 14 0 2 12 15 1 3 13 5 11 9 7 4 10 8 6 8
                  6 4 10 9 7 5 11 3 13 15 1 2 12 14 1 14 15 0 4 11 10 5 2 13
                  12 3 7 8 9 6 7 8 9 6 2 13 12 3 4 11 10 5 1 14 15 12 12 5 5
                  14 14 7 7 1 1 8 8 3 3 10 10 6 6 15 15 4 4 13 13 11 11 2 2 9
                  9 0 3 2 8 9 5 4 14 15 6 7 13 12 0 1 11 10 9 8 2 3 15 14 4 5
                  12 13 7 6 10 11 1 11 9 6 4 1 3 12 14 6 4 11 9 12 14 1 3 8 10
                  5 7 2 0 15 13 5 7 8 10 15 13 2 4 7 11 8 10 9 5 6 1 2 14 13
                  15 12 0 3 7 4 8 11 9 10 6 5 2 1 13 14 12 15 3 11 15 10 14 9
                  13 8 12 15 11 14 10 13 9 12 8 3 7 2 6 1 5 0 4 7 3 6 2 5 1 4
                  4 1 7 2 2 7 1 4 8 13 11 14 14 11 13 8 12 9 15 10 10 15 9 12
                  0 5 3 6 6 3 5 12 10 9 15 6 0 3 5 8 14 13 11 2 4 7 1 13 11 8
                  14 7 1 2 4 9 15 12 10 3 5 6 3 4 4 3 13 10 10 13 15 8 8 15 1
                  6 6 1 2 5 5 2 12 11 11 12 14 9 9 14 0 7 7 2 10 11 3 9 1 0 8
                  13 5 4 12 6 14 15 7 5 13 12 4 14 6 7 15 10 2 3 11 1 9 8 13 4
                  6 15 2 11 9 0 10 3 1 8 5 12 14 7 10 3 1 8 5 12 14 7 13 4 6
                  15 2 11 9 5 15 8 2 6 12 11 1 10 0 7 13 9 3 4 14 11 1 6 12 8
                  2 5 15 4 14 9 3 7 13 10 10 1 5 14 13 6 2 9 13 6 2 9 10 1 5
                  14 4 15 11 0 3 8 12 7 3 8 12 7 4 15 11 5 9 4 8 14 2 15 3 3
                  15 2 14 8 4 9 5 0 12 1 13 11 7 10 6 6 10 7 11 13 1 12 10 7 9
                  4 5 8 6 11 4 9 7 10 11 6 8 5 15 2 12 1 0 13 3 14 1 12 2 15
                  14 3 13 2 12 7 9 1 15 4 10 4 10 1 15 7 9 2 12 14 0 11 5 13 3
                  8 6 8 6 13 3 11 5 14 13 2 10 5 10 5 13 2 3 12 4 11 4 11 3 12
                  1 14 6 9 6 9 1 14 15 0 8 7 8 7 15}, 31);
                  if n = 32 then d = d32;
                  end;
               end;

            else if f = 17 & n = 17 then d = mod((0:16)` * (0:16), 17);

            else if f = 18 & n = 18 then d = j(18, 1, 0) || (0:17)`;

            else if f = 19 & n = 19 then d = mod((0:18)` * (0:18), 19);

            else if f = 20 then do;

               if n = 20 then do;
                  d = j(20, 1, 0) || (0:19)` || shape({ 0 0 13 9 4 19 9 16 18 7
                  2 1 17 14 6 18 14 2 19 11 7 17 15 13 5 6 1 12 12 15 16 10 11
                  5 10 3 3 8 8 4}, 20);
                  end;

               end;

            else if f = 21 then do;

               if n = 21 then do;
                  d21 = j(21, 6, 0);
                  d21[2:21, 2:6] = shape({ 8 9 11 14 18 16 5 1 19 17 20 4 10 3
                  12 2 15 6 13 7 9 11 14 18 8 5 1 19 17 16 4 10 3 12 20 15 6 13
                  7 2 11 14 18 8 9 1 19 17 16 5 10 3 12 20 4 6 13 7 2 15 14 18 8
                  9 11 19 17 16 5 1 3 12 20 4 10 13 7 2 15 6 18 8 9 11 14 17 16
                  5 1 19 12 20 4 10 3 7 2 15 6 13 }, 20);
                  if n = 21 then d = d21;
                  end;

               end;

            else if f = 22 then do;

               if n = 22 then d = j(22, 1, 0) || (0:21)`;

               end;

            else if f = 24 then do;

               if n = 24 then do;
                  d24 = j(24, 7, 0);
                  d24[,2:7] = shape({ 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16
                  17 18 19 20 21 22 23 0 2 12 14 8 10 18 16 3 1 11 9 20 22 4 6
                  13 15 23 21 19 17 7 5 0 3 15 12 21 22 10 9 6 5 18 17 11 8 23
                  20 19 16 4 7 14 13 1 2 0 12 1 13 2 14 15 3 18 6 7 19 5 17 16
                  4 21 9 8 20 11 23 10 22 0 13 5 16 19 6 12 1 2 15 3 14 23 10 22
                  11 8 21 7 18 17 4 20 9 0 14 17 7 10 20 9 23 22 8 19 5 6 16 3
                  13 12 2 15 1 4 18 21 11 }, 6)`;
                  if n = 24 then d = d24;
                  end;

               end;

            if dnz & any(f = {3 4 5}) then d = nrmz(d, f);

            return(d);
            finish;

         *------make difference scheme D(2f, 2f, f), for prime f------;
         start doubledif(f);
            s     = (0:(f - 1));
            alpha = { . . 2 . 2 . 5 . . . 2 . 2 . . . 5}[f];
            u     = mod(s` * s, f);
            ssqu  = shape(mod(s # s, f), f, f);
            inv   = (u = 1)[<:>,2:f] - 1;
            beta  = inv[2];
            gamma = mod((alpha - 1) # inv[mod(2 * alpha, f)], f);
            delta = mod(alpha # inv[2], f);
            eps   = mod((alpha - 1) # inv[2], f);
            v     = mod(u +  beta # ssqu , f);
            w     = mod(u + gamma # ssqu`, f);
            x     = mod(alpha * u + delta # ssqu + eps # ssqu`, f);
            return( (u || w) // (v || x) );
            finish;

         *------make difference scheme D(2f, 2f, f), for f a prime power------;
         start ddif2(f);
            run fieldop(f, add, sub, mul, inv);
            alpha = 4;
            ssqu  = shape(vecdiag(mul), f, f) + 1;
            beta  = inv[2]+1;
            gamma = mul[sub[alpha,2]+1, inv[mul[3,alpha]]+1]+1;
            delta = mul[alpha,beta]+1;
            eps   = mul[sub[alpha,2]+1,beta]+1;
            v     = j(f, f, 0);   w = v;   x = v;
            do i = 1 to f; do j = 1 to f;
               v[i,j] = add[mul[i,j]+1, mul[beta , ssqu[i,j]]+1];
               w[i,j] = add[mul[i,j]+1, mul[gamma, ssqu[j,i]]+1];
               x[i,j] =                 mul[alpha,  mul[i,j ]+1];
               x[i,j] = add[x[i,j]+1,   mul[delta, ssqu[i,j]]+1];
               x[i,j] = add[x[i,j]+1,   mul[eps  , ssqu[j,i]]+1];
               end; end;
            return((mul || w) // (v || x));
            finish;

         *------sort matrix on a few columns------;
         start sortmat(x, j);
            %if %sysevalf(&sysver >= 9) %then %do; call sort(x, j); %end;
            %else %do;
               x0 = x;
               x[rank((x[,j] + 2) * (10 ## ((ncol(j)-1):0)`)),] = x0;
               %end;
            return(x);
            finish;

         *============ End Orthogonal Array Construction Functions ============;

         %if &renderdif %then %do;
            options ls=165;
            file log;
            start rend(r);
               if r < 10 then put r 1. @;
               else if r < 100 then put r 2. @;
               else if r < 1e3 then put r 3. @;
               else if r < 1e4 then put r 4. @;
               else if r < 1e5 then put r 5. @;
               else if r < 1e6 then put r 6. @;
               else put r best7. @;
               return(r);
               finish;

            do f = 3 to 22;
               do r = f to 513 / f by f;
                  x = gh(r, f);
                  if any(x ^= 0) then do;
                     x = nrmz(x, f);
                     put / 'D(' @; t = rend(r); put ',' @; c = rend(ncol(x));
                     put ',' @; t = rend(f); put ')';
                     do i = 1 to r;
                        do j = 1 to c;
                           if f < 11 then put (x[i,j]) 1. @;
                           else           put (x[i,j]) 2. @;
                           end;
                        put;
                        end;
                     end;
                  end;
               end;
            file print;
            %end;

         *================== Most of the OA Work Starts Here ==================;

         start inittab(facts) global(balanced, lineage, h2, ha
                                     %do j = 2 %to &maxlevneed;
                                        , %if &j > 9 %then d&j; %else d0&j;
                                        %end;
                                    );

            h2    = {1 1, 1 -1};
            any00 = 0;
            %do i =  1 %to   9;     any0&i = sum(facts = &i); %end;
            %do i = 10 %to &bigkid; any&i  = sum(facts = &i); %end;

            *------ 4 - &bigkid run designs ------;
            %if &tabsize ge 4 and &tabsize le &bigkid and &tabsize ne 128
               %then %do;
               %if &tabsize lt 10 %then %let jj = 0&tabsize;
               %else                    %let jj = &tabsize;
               d&jj = (1:&jj)`;
               %end;

            %*------ 128 runs ------;
            %else %if &tabsize eq 128 %then %do;

               if index(lineage, '4 ** 32 32 ** 1') then do;
                  d04 = ks((0:3)`, gh(32, 4), 4) + 1;
                  d32 = j(4, 1, 1) @ (1:32)`;
                  end;

               else if index(lineage, '8 ** 16 16 ** 1') then do;
                  d08 = ks((0:7)`, gh(16, 8), 8) + 1;
                  d16 = j(8, 1, 1) @ (1:16)`;
                  end;

               %end;

            %*------ 147 runs ------;
            %else %if &tabsize eq 147 %then %do;

               if index(lineage, '7 ** 9 21 ** 1') then do;
                  d07 = ks((0:6)`, gh(21, 7), 7) + 1;
                  d21 = j(7, 1, 1) @ (1:21)`;
                  end;

               %end;

            %*------ 150 runs ------;
            %else %if &tabsize eq 150 %then %do;

               if index(lineage, '5 ** 11 30 ** 1') then do;
                  d05 = ks((0:4)`, gh(30, 5), 5) + 1;
                  d30 = j(5, 1, 1) @ (1:30)`;
                  end;

               %end;

            %*------ 152 runs ------;
            %else %if &tabsize eq 152 %then %do;

               if index(lineage, '2 ** 76 76 ** 1') then do;
                  if any02 then d02 = twolev({1 -1}` @ hadp2(76, 1));
                  d76 = j(2, 1, 1) @ (1:76)`;
                  end;

               %end;

            %*------ 153 runs ------;
            %else %if &tabsize eq 153 %then %do;

               if index(lineage, '3 ** 25 17 ** 1') then do;
                  d03 = ks((0:2)`, gh(51, 3), 3) + 1 ||
                        j(3, 1, 1) @ (1:3)` @ j(17, 1, 1);
                  d17 = j(9, 1, 1) @ (1:17)`;
                  end;

               %end;

            %*------ 160 runs ------;
            %else %if &tabsize eq 160 %then %do;

               if index(lineage, '2 ** 133 5 ** 1 16 ** 1') then do;
                  b1  = twolev(h2 @ h2 @ h2)[,2:8];
                  b2  = j(8, 1, 1) @ sortmat(fac(20, 8, 2) ||
                        {5 2 4 3 1}` @ j(4, 1, 1), 1);
                  d02 = ks(b1, sortmat(had(20, 0), 1:2)[,2:19], 2) + 1 ||
                        b2[,2:8];
                  d05 = b2[,9];
                  d16 = (1:16)` @ j(10, 1, 1);
                  end;

               else if index(lineage, '2 ** 127 10 ** 1 16 ** 1') then do;
                  b1  = twolev(h2 @ h2 @ h2)[,2:8];
                  d02 = ks(b1, sortmat(had(20, 0), 1:2)[,2:19], 2) + 1 ||
                        j(8, 1, 1) @ twolev({1 -1}` @ j(5, 1, 1) @ {1 -1}`);
                  d10 = j(16, 1, 1) @ (1:10)`;
                  d16 = (1:16)` @ j(10, 1, 1);
                  end;

               else if index(lineage, '2 ** 80 80 ** 1') then do;
                  d02 = twolev({1 -1}` @ h2 @ h2 @ had(20, 1));
                  d80 = j(2, 1, 1) @ (1:80)`;
                  end;

               else if index(lineage, '4 ** 16 40 ** 1') then do;
                  d04 = ks((0:3)`, gh(40, 4), 4) + 1;
                  d40 = j(4, 1, 1) @ (1:40)`;
                  end;

               %end;

            %*------ 162 runs ------;
            %else %if &tabsize eq 162 %then %do;

               if index(lineage, '3 ** 65 6 ** 1 27 ** 1') then do;
                  d03 = ks(gh(3, 3), (0:2)`, 3) || (0:2)` @ j(3, 1, 1);
                  d03 = ks(gh(3, 3), d03, 3) || (0:2)` @ j(9, 1, 1);
                  d03 = ks(gh(6, 3)[,2:6], d03, 3) + 1;
                  d06 = (1:6)` @ j(27, 1, 1);
                  d27 = j(6, 1, 1) @ (1:27)`;
                  end;

               else if index(lineage, '3 ** 54 54 ** 1') then do;
                  d03 = ks((0:2)`, gh(54, 3), 3) + 1;
                  d54 = j(3, 1, 1) @ (1:54)`;
                  end;

               else if index(lineage, '9 ** 18 18 ** 1') then do;
                  d09 = ks((0:8)`, gh(18, 9), 9) + 1;
                  d18 = j(9, 1, 1) @ (1:18)`;
                  end;

               %end;

            %*------ 168 runs ------;
            %else %if &tabsize eq 168 %then %do;

               if index(lineage, '2 ** 84 84 ** 1') then do;
                  d02 = twolev({1 -1}` @ hadp1(84, 1));
                  d84 = j(2, 1, 1) @ (1:84)`;
                  end;

               %end;

            %*------ 169 runs ------;
            %else %if &tabsize eq 169 %then %do;

               if index(lineage, '13 ** 14') then do;
                  d13 = (ks((0:12)`, gh(13, 13), 13) + 1) ||
                        j(13, 1, 1) @ (1:13)`;
                  end;

               %end;

            %*------ 171 runs ------;
            %else %if &tabsize eq 171 %then %do;

               if index(lineage, '3 ** 28 19 ** 1') then do;
                  d03 = ks((0:2)`, gh(57, 3), 3) + 1 ||
                        j(3, 1, 1) @ (1:3)` @ j(19, 1, 1);
                  d19 = j(9, 1, 1) @ (1:19)`;
                  end;

               %end;

            %*------ 175 runs ------;
            %else %if &tabsize eq 175 %then %do;

               if index(lineage, '5 ** 10 35 ** 1') then do;
                  d05 = ks((0:4)`, gh(35, 5), 5) + 1;
                  d35 = j(5, 1, 1) @ (1:35)`;
                  end;

               %end;

            %*------ 176 runs ------;
            %else %if &tabsize eq 176 %then %do;

               if index(lineage, '2 ** 88 88 ** 1') then do;
                  d02 = twolev({1 -1}` @ h2 @ hadp1(44, 1));
                  d88 = j(2, 1, 1) @ (1:88)`;
                  end;

               else if index(lineage, '4 ** 12 44 ** 1') then do;
                  d04 = ks((0:3)`, gh(44, 4), 4) + 1;
                  d44 = j(4, 1, 1) @ (1:44)`;
                  end;

               %end;

            %*------ 180 runs ------;
            %else %if &tabsize eq 180 %then %do;

               if index(lineage, '3 ** 30 60 ** 1') then do;
                  d03 = ks((0:2)`, gh(60, 3), 3) + 1;
                  d60 = j(3, 1, 1) @ (1:60)`;
                  end;

               else if index(lineage, '6 ** 2 30 ** 1') then do;
                  d06 = ks((0:5)`, gh(30, 6), 6) + 1;
                  d30 = j(6, 1, 1) @ (1:30)`;
                  end;

               %end;

            %*------ 184 runs ------;
            %else %if &tabsize eq 184 %then %do;

               if index(lineage, '2 ** 92 92 ** 1') then do;
                  d02 = twolev({1 -1}` @ hadw(92, 1));
                  d92 = j(2, 1, 1) @ (1:92)`;
                  end;

               %end;

            %*------ 189 runs ------;
            %else %if &tabsize eq 189 %then %do;

               if index(lineage, '3 ** 36 63 ** 1') then do;
                  d03 = ks((0:2)`, gh(63, 3), 3) + 1;
                  d63 = j(3, 1, 1) @ (1:63)`;
                  end;

               %end;

            %*------ 192 runs ------;
            %else %if &tabsize eq 192 %then %do;

               if index(lineage, '2 ** 96 96 ** 1') then do;
                  d02 = twolev({1 -1}` @ h2 @ h2 @ h2 @ had(12, 1));
                  d96 = j(2, 1, 1) @ (1:96)`;
                  end;

               else if index(lineage, '4 ** 48 48 ** 1') then do;
                  d04 = ks((0:3)`, gh(48, 4), 4) + 1;
                  d48 = j(4, 1, 1) @ (1:48)`;
                  end;

               else if index(lineage, '8 ** 8 24 ** 1') then do;
                  d08 = ks((0:7)`, gh(24, 8), 8) + 1;
                  d24 = j(8, 1, 1) @ (1:24)`;
                  end;

               %end;

            %*------ 196 runs ------;
            %else %if &tabsize eq 196 %then %do;

               if index(lineage, '7 ** 28 28 ** 1') then do;
                  d07 = ks((0:6)`, gh(28, 7), 7) + 1;
                  d28 = j(7, 1, 1) @ (1:28)`;
                  end;

               else if index(lineage, '14 ** 5') then do;
                  d14 = (1:14)` @ j(14, 1, 1) || j(14, 1, 1) @ (1:14)` ||
                        fac(196, 3, 14);
                  end;

               %end;

            %*------ 198 runs ------;
            %else %if &tabsize eq 198 %then %do;

               if index(lineage, '3 ** 30 66 ** 1') then do;
                  d03 = ks((0:2)`, gh(66, 3), 3) + 1;
                  d66 = j(3, 1, 1) @ (1:66)`;
                  end;

               %end;

            %*------ 200 runs ------;
            %else %if &tabsize eq 200 %then %do;

               if index(lineage, '2 ** 100 100 ** 1') then do;
                  d02  = twolev({1 -1}` @ hadw(100, 1));
                  d100 = j(2, 1, 1) @ (1:100)`;
                  end;

               else if index(lineage, '5 ** 20 40 ** 1') then do;
                  d05 = ks((0:4)`, gh(40, 5), 5) + 1;
                  d40 = j(5, 1, 1) @ (1:40)`;
                  end;

               else if index(lineage, '10 ** 5 20 ** 1') then do;
                  d10 = ks((0:9)`, gh(20, 10), 10) + 1;
                  d20 = j(10, 1, 1) @ (1:20)`;
                  end;

               %end;

            %*------ 207 runs ------;
            %else %if &tabsize eq 207 %then %do;

               if index(lineage, '3 ** 25 23 ** 1') then do;
                  d03 = ks((0:2)`, gh(69, 3), 3) + 1 ||
                        j(3, 1, 1) @ (1:3)` @ j(23, 1, 1);
                  d23 = j(9, 1, 1) @ (1:23)`;
                  end;

               %end;

            %*------ 208 runs ------;
            %else %if &tabsize eq 208 %then %do;

               if index(lineage, '2 ** 104 104 ** 1') then do;
                  d02  = twolev({1 -1}` @ h2 @ hadw(52, 1));
                  d104 = j(2, 1, 1) @ (1:104)`;
                  end;

               else if index(lineage, '4 ** 16 52 ** 1') then do;
                  d04 = ks((0:3)`, gh(52, 4), 4) + 1;
                  d52 = j(4, 1, 1) @ (1:52)`;
                  end;

               %end;

            %*------ 216 runs ------;
            %else %if &tabsize eq 216 %then %do;

               if index(lineage, '2 ** 108 108 ** 1') then do;
                  d02  = twolev({1 -1}` @ hadw(108, 1));
                  d108 = j(2, 1, 1) @ (1:108)`;
                  end;

               else if index(lineage, '2 ** 11 3 ** 77 12 ** 1 18 ** 1')
                  then do;
                  b1  = ks((0:2)`, gh(6, 3), 3) + 1 ||
                        j(3, 1, 1) @ (0:2)` @ {1 1}` ||  j(9, 1, 1) @ (0:1)`;
                  d02 = ks(had(12, 0) - 1, b1[,8], 2) + 1;
                  d03 = ks(gh(12, 3)[,2:12], b1[,1:7], 3) + 1;
                  d12 = (1:12)` @ j(18, 1, 1);
                  d18 = j(12, 1, 1) @ (1:18)`;
                  end;

               else if index(lineage, '3 ** 72 72 ** 1') then do;
                  d03 = ks((0:2)`, gh(72, 3), 3) + 1;
                  d72 = j(3, 1, 1) @ (1:72)`;
                  end;

               else if index(lineage, '3 ** 66 6 ** 5 12 ** 1 18 ** 1') then do;
                  d03 = ks(gh(12, 3)[,2:12], ks(gh(6, 3), (0:2)`, 3), 3) + 1;
                  d06 = ks(gh(12, 6)[,2: 6], (0:5)` @ j(3, 1, 1), 6) + 1;
                  d12 = (1:12)` @ j(18, 1, 1);
                  d18 = j(12, 1, 1) @ (1:18)`;
                  end;

               else if index(lineage, '6 ** 7 36 ** 1') then do;
                  d06 = ks((0:5)`, gh(36, 6), 6) + 1;
                  d36 = j(6, 1, 1) @ (1:36)`;
                  end;

               %end;

            %*------ 224 runs ------;
            %else %if &tabsize eq 224 %then %do;

               if index(lineage, '2 ** 193 7 ** 1 16 ** 1') then do;
                  b1  = twolev(h2 @ h2 @ h2)[,2:8];
                  b2  = j(8, 1, 1) @ sortmat(fac(28, 12, 2) ||
                        (1:7)` @ j(4, 1, 1), 1);
                  d02 = ks(b1, sortmat(hadw(28, 0), 1:2)[,2:27], 2) + 1 ||
                        b2[,2:12];
                  d07 = b2[,13];
                  d16 = (1:16)` @ j(14, 1, 1);
                  end;

               else if index(lineage, '2 ** 183 14 ** 1 16 ** 1') then do;
                  b1  = twolev(h2 @ h2 @ h2)[,2:8];
                  d02 = ks(b1, sortmat(hadw(28, 0), 1:2)[,2:27], 2) + 1 ||
                        j(8, 1, 1) @ twolev({1 -1}` @ j(7, 1, 1) @ {1 -1}`);
                  d14 = j(16, 1, 1) @ (1:14)`;
                  d16 = (1:16)` @ j(14, 1, 1);
                  end;

               else if index(lineage, '2 ** 112 112 ** 1') then do;
                  d02  = twolev(({1 -1}` @ h2 @ h2) @ hadw(28, 1));
                  d112 = j(2, 1, 1) @ (1:112)`;
                  end;

               else if index(lineage, '4 ** 56 56 ** 1') then do;
                  d04 = ks((0:3)`, gh(56, 4), 4) + 1;
                  d56 = j(4, 1, 1) @ (1:56)`;
                  end;

               %end;

            %*------ 225 runs ------;
            %else %if &tabsize eq 225 %then %do;

               if index(lineage, '3 ** 27 75 ** 1') then do;
                  d03 = ks((0:2)`, gh(75, 3), 3) + 1;
                  d75 = j(3, 1, 1) @ (1:75)`;
                  end;

               else if index(lineage, '5 ** 20 45 ** 1') then do;
                  d05 = ks((0:4)`, gh(45, 5), 5) + 1;
                  d45 = j(5, 1, 1) @ (1:45)`;
                  end;

               else if index(lineage, '15 ** 6') then do;
                  d15 = ks((0:14)`, gh(15, 15), 15) + 1 ||
                        j(15, 1, 1) @ (1:15)`;
                  end;

               %end;

            %*------ 232 runs ------;
            %else %if &tabsize eq 232 %then %do;

               if index(lineage, '2 ** 116 116 ** 1') then do;
                  d02  = twolev({1 -1}` @ hadw(116, 1));
                  d116 = j(2, 1, 1) @ (1:116)`;
                  end;

               %end;

            %*------ 234 runs ------;
            %else %if &tabsize eq 234 %then %do;

               if index(lineage, '3 ** 30 78 ** 1') then do;
                  d03 = ks((0:2)`, gh(78, 3), 3) + 1;
                  d78 = j(3, 1, 1) @ (1:78)`;
                  end;

               %end;

            %*------ 240 runs ------;
            %else %if &tabsize eq 240 %then %do;

               if index(lineage, '2 ** 205 5 ** 1 24 ** 1') then do;
                  b1  = j(12, 1, 1) @ sortmat(fac(20, 8, 2) ||
                        {5 2 4 3 1}` @ j(4, 1, 1), 1:2);
                  d02 = ks(had(12, 0), sortmat(had(20, 0), 1:2)[,2:19], 2) + 1
                        || b1[,2:8];
                  d05 = b1[,9];
                  d24 = (1:24)` @ j(10, 1, 1);
                  end;

               else if index(lineage, '2 ** 199 10 ** 1 24 ** 1') then do;
                  d02 = ks(had(12, 0), sortmat(had(20, 0), 1:2)[,2:19], 2) + 1
                        || j(12, 1, 1) @ twolev({1 -1}` @ j(5, 1, 1) @ {1 -1}`);
                  d10 = j(24, 1, 1) @ (1:10)`;
                  d24 = (1:24)` @ j(10, 1, 1);
                  end;

               else if index(lineage, '2 ** 120 120 ** 1') then do;
                  d02  = twolev({1 -1}` @ h2 @ hadp1(60, 1));
                  d120 = j(2, 1, 1) @ (1:120)`;
                  end;

               else if index(lineage, '4 ** 20 60 ** 1') then do;
                  d04 = ks((0:3)`, gh(60, 4), 4) + 1;
                  d60 = j(4, 1, 1) @ (1:60)`;
                  end;

               %end;

            %*------ 242 runs ------;
            %else %if &tabsize eq 242 %then %do;

               if index(lineage, '11 ** 22 22 ** 1') then do;
                  d11 = ks((0:10)`, gh(22, 11), 11) + 1;
                  d22 = j(11, 1, 1) @ (1:22)`;
                  end;

               %end;

            %*------ 243 runs ------;
            %else %if &tabsize eq 243 %then %do;

               if index(lineage, '3 ** 81 81 ** 1') then do;
                  d03 = ks((0:2)`, gh(81, 3), 3) + 1;
                  d81 = j(3, 1, 1) @ (1:81)`;
                  end;

               else if index(lineage, '9 ** 27 27 ** 1') then do;
                  d09 = ks((0:8)`, gh(27, 9), 9) + 1;
                  d27 = j(9, 1, 1) @ (1:27)`;
                  end;

               %end;

            %*------ 245 runs ------;
            %else %if &tabsize eq 245 %then %do;

               if index(lineage, '7 ** 10 35 ** 1') then do;
                  d07 = ks((0:6)`, gh(35, 7), 7) + 1;
                  d35 = j(7, 1, 1) @ (1:35)`;
                  end;

               %end;

            %*------ 248 runs ------;
            %else %if &tabsize eq 248 %then %do;

               if index(lineage, '2 ** 124 124 ** 1') then do;
                  d02  = twolev({1 -1}` @ hadw(124, 1));
                  d124 = j(2, 1, 1) @ (1:124)`;
                  end;

               %end;

            %*------ 250 runs ------;
            %else %if &tabsize eq 250 %then %do;

               if index(lineage, '5 ** 50 50 ** 1') then do;
                  d05 = ks((0:4)`, gh(50, 5), 5) + 1;
                  d50 = j(5, 1, 1) @ (1:50)`;
                  end;

               %end;

            %*------ 252 runs ------;
            %else %if &tabsize eq 252 %then %do;

               if index(lineage, '3 ** 42 84 ** 1') then do;
                  d03 = ks((0:2)`, gh(84, 3), 3) + 1;
                  d84 = j(3, 1, 1) @ (1:84)`;
                  end;

               else if index(lineage, '6 ** 2 42 ** 1') then do;
                  d06 = ks((0:5)`, gh(42, 6), 6) + 1;
                  d42 = j(6, 1, 1) @ (1:42)`;
                  end;

               %end;

            %*------ 256 runs ------;
            %else %if &tabsize eq 256 %then %do;

               if index(lineage, '8 ** 32 32 ** 1') then do;
                  d08 = ks((0:7)`, gh(32, 8), 8) + 1;
                  d32 = j(8, 1, 1) @ (1:32)`;
                  end;

               if index(lineage, '16 ** 17') then do;
                  d16 = ks((0:15)`, gh(16, 16), 16) + 1 ||
                        j(16, 1, 1) @ (1:16)`;
                  end;

               %end;

            %*------ 261 runs ------;
            %else %if &tabsize eq 261 %then %do;

               if index(lineage, '3 ** 27 87 ** 1') then do;
                  d03 = ks((0:2)`, gh(87, 3), 3) + 1;
                  d87 = j(3, 1, 1) @ (1:87)`;
                  end;

               %end;

            %*------ 264 runs ------;
            %else %if &tabsize eq 264 %then %do;

               if index(lineage, '2 ** 132 132 ** 1') then do;
                  d02  = twolev({1 -1}` @ hadw(132, 1));
                  d132 = j(2, 1, 1) @ (1:132)`;
                  end;

               %end;

            %*------ 270 runs ------;
            %else %if &tabsize eq 270 %then %do;

               if index(lineage, '3 ** 90 90 ** 1') then do;
                  d03 = ks((0:2)`, gh(90, 3), 3) + 1;
                  d90 = j(3, 1, 1) @ (1:90)`;
                  end;

               %end;

            %*------ 272 runs ------;
            %else %if &tabsize eq 272 %then %do;

               if index(lineage, '2 ** 136 136 ** 1') then do;
                  d02  = twolev({1 -1}` @ h2 @ hadp1(68, 1));
                  d136 = j(2, 1, 1) @ (1:136)`;
                  end;

               else if index(lineage, '4 ** 32 68 ** 1') then do;
                  d04 = ks((0:3)`, gh(68, 4), 4) + 1;
                  d68 = j(4, 1, 1) @ (1:68)`;
                  end;

               %end;

            %*------ 275 runs ------;
            %else %if &tabsize eq 275 %then %do;

               if index(lineage, '5 ** 11 55 ** 1') then do;
                  d05 = ks((0:4)`, gh(55, 5), 5) + 1;
                  d55 = j(5, 1, 1) @ (1:55)`;
                  end;

               %end;

            %*------ 279 runs ------;
            %else %if &tabsize eq 279 %then %do;

               if index(lineage, '3 ** 30 93 ** 1') then do;
                  d03 = ks((0:2)`, gh(93, 3), 3) + 1;
                  d93 = j(3, 1, 1) @ (1:93)`;
                  end;

               %end;

            %*------ 280 runs ------;
            %else %if &tabsize eq 280 %then %do;

               if index(lineage, '2 ** 140 140 ** 1') then do;
                  d02  = twolev({1 -1}` @ hadp1(140, 1));
                  d140 = j(2, 1, 1) @ (1:140)`;
                  end;

               %end;

            %*------ 288 runs ------;
            %else %if &tabsize eq 288 %then %do;

               if index(lineage, '2 ** 253 9 ** 1 16 ** 1') then do;
                  b1  = j(8, 1, 1) @ sortmat(fac(36, 16, 2) ||
                        (1:9)` @ j(4, 1, 1), 1);
                  d02 = twolev((h2 @ h2 @ h2)[,2:8] @
                        sortmat(hadw(36, 1), 1:2)[,3:36]) || b1[,2:16];
                  d09 = b1[,17];
                  d16 = (1:16)` @ j(18, 1, 1);
                  end;

               else if index(lineage, '2 ** 239 16 ** 1 18 ** 1') then do;
                  d02 = twolev((h2 @ h2 @ h2)[,2:8] @
                        sortmat(hadw(36, 1), 1:2)[,3:36]) ||
                        j(8, 1, 1) @ twolev({1 -1}` @ j(9, 1, 1) @ {1 -1}`);
                  d18 = j(16, 1, 1) @ (1:18)`;
                  d16 = (1:16)` @ j(18, 1, 1);
                  end;

               else if index(lineage, '2 ** 144 144 ** 1') then do;
                  d02  = twolev({1 -1}` @ h2 @ h2 @ hadw(36, 1));
                  d144 = j(2, 1, 1) @ (1:144)`;
                  end;

               else if index(lineage, '3 ** 96 96 ** 1') then do;
                  d03 = ks((0:2)`, gh(96, 3), 3) + 1;
                  d96 = j(3, 1, 1) @ (1:96)`;
                  end;

               else if index(lineage, '4 ** 36 72 ** 1') then do;
                  d04 = ks((0:3)`, gh(72, 4), 4) + 1;
                  d72 = j(4, 1, 1) @ (1:72)`;
                  end;

               else if index(lineage, '12 ** 6 24 ** 1') then do;
                  d12 = ks((0:11)`, gh(24, 12), 12) + 1;
                  d24 = j(12, 1, 1) @ (1:24)`;
                  end;

               else if index(lineage, '6 ** 10 48 ** 1') then do;
                  d06 = ks((0:5)`, gh(48, 6), 6) + 1;
                  d48 = j(6, 1, 1) @ (1:48)`;
                  end;

               %end;

            %*------ 289 runs ------;
            %else %if &tabsize eq 289 %then %do;

               if index(lineage, '17 ** 18') then do;
                  d17 = (ks((0:16)`, gh(17, 17), 17) + 1) ||
                        j(17, 1, 1) @ (1:17)`;
                  end;

               %end;

            %*------ 294 runs ------;
            %else %if &tabsize eq 294 %then %do;

               if index(lineage, '7 ** 18 42 ** 1') then do;
                  d07 = ks((0:6)`, gh(42, 7), 7) + 1;
                  d42 = j(7, 1, 1) @ (1:42)`;
                  end;

               %end;

            %*------ 297 runs ------;
            %else %if &tabsize eq 297 %then %do;

               if index(lineage, '3 ** 39 99 ** 1') then do;
                  d03 = ks((0:2)`, gh(99, 3), 3) + 1;
                  d99 = j(3, 1, 1) @ (1:99)`;
                  end;

               %end;

            %*------ 300 runs ------;
            %else %if &tabsize eq 300 %then %do;

               if index(lineage, '5 ** 20 60 ** 1') then do;
                  d05 = ks((0:4)`, gh(60, 5), 5) + 1;
                  d60 = j(5, 1, 1) @ (1:60)`;
                  end;

               else if index(lineage, '10 ** 2 30 ** 1') then do;
                  d10 = ks((0:9)`, gh(30, 10), 10) + 1;
                  d30 = j(10, 1, 1) @ (1:30)`;
                  end;

               %end;

            %*------ 304 runs ------;
            %else %if &tabsize eq 304 %then %do;

               if index(lineage, '2 ** 228 76 ** 1') then do;
                  if any02 then d02 = twolev((h2 @ h2)[,2:4] @ hadp2(76, 1));
                  d76 = j(4, 1, 1) @ (1:76)`;
                  end;

               else if index(lineage, '4 ** 16 76 ** 1') then do;
                  d04 = ks((0:3)`, gh(76, 4), 4) + 1;
                  d76 = j(4, 1, 1) @ (1:76)`;
                  end;

               %end;

            %*------ 306 runs ------;
            %else %if &tabsize eq 306 %then %do;

               if index(lineage, '3 ** 48 102 ** 1') then do;
                  d03  = ks((0:2)`, gh(102, 3), 3) + 1;
                  d102 = j(3, 1, 1) @ (1:102)`;
                  end;

               %end;

            %*------ 315 runs ------;
            %else %if &tabsize eq 315 %then %do;

               if index(lineage, '3 ** 29 105 ** 1') then do;
                  d03  = ks((0:2)`, gh(105, 3), 3) + 1;
                  d105 = j(3, 1, 1) @ (1:105)`;
                  end;

               %end;

            %*------ 320 runs ------;
            %else %if &tabsize eq 320 %then %do;

               if index(lineage, '2 ** 240 80 ** 1') then do;
                  d02 = twolev((h2 @ h2)[,2:4] @ h2 @ h2 @ had(20, 1));
                  d80 = j(4, 1, 1) @ (1:80)`;
                  end;

               else if index(lineage, '4 ** 40 80 ** 1') then do;
                  d04 = ks((0:3)`, gh(80, 4), 4) + 1;
                  d80 = j(4, 1, 1) @ (1:80)`;
                  end;

               else if index(lineage, '8 ** 10 40 ** 1') then do;
                  d08 = ks((0:7)`, gh(40, 8), 8) + 1;
                  d40 = j(8, 1, 1) @ (1:40)`;
                  end;

               %end;

            %*------ 324 runs ------;
            %else %if &tabsize eq 324 %then %do;

               if index(lineage, '2 ** 323') then do;
                  d02 = twolev(hadc(324))[,2:324];
                  end;

               else if index(lineage, '3 ** 108 108 ** 1') then do;
                  d03  = ks((0:2)`, gh(108, 3), 3) + 1;
                  d108 = j(3, 1, 1) @ (1:108)`;
                  end;

               else if index(lineage, '3 ** 143 12 ** 1 27 ** 1') then do;
                  d03 = ks(gh(3, 3), (0:2)`, 3) || (0:2)` @ j(3, 1, 1);
                  d03 = ks(gh(3, 3), d03, 3) || (0:2)` @ j(9, 1, 1);
                  d03 = ks(gh(12, 3)[,2:12], d03, 3) + 1;
                  d12 = (1:12)` @ j(27, 1, 1);
                  d27 = j(12, 1, 1) @ (1:27)`;
                  end;

               else if index(lineage, '6 ** 2 54 ** 1') then do;
                  d06 = ks((0:5)`, gh(54, 6), 6) + 1;
                  d54 = j(6, 1, 1) @ (1:54)`;
                  end;

               else if index(lineage, '9 ** 36 36 ** 1') then do;
                  d09 = ks((0:8)`, gh(36, 9), 9) + 1;
                  d36 = j(9, 1, 1) @ (1:36)`;
                  end;

               else if index(lineage, '18 ** 3') then do;
                  d18 = ks((0:17)`, gh(18, 18), 18) + 1 ||
                        j(18, 1, 1) @ (1:18)`;
                  end;

               %end;

            %*------ 325 runs ------;
            %else %if &tabsize eq 325 %then %do;

               if index(lineage, '5 ** 20 65 ** 1') then do;
                  d05 = ks((0:4)`, gh(65, 5), 5) + 1;
                  d65 = j(5, 1, 1) @ (1:65)`;
                  end;

               %end;

            %*------ 333 runs ------;
            %else %if &tabsize eq 333 %then %do;

               if index(lineage, '3 ** 36 111 ** 1') then do;
                  d03  = ks((0:2)`, gh(111, 3), 3) + 1;
                  d111 = j(3, 1, 1) @ (1:111)`;
                  end;

               %end;

            %*------ 336 runs ------;
            %else %if &tabsize eq 336 %then %do;

               if index(lineage, '2 ** 297 7 ** 1 24 ** 1') then do;
                  b1  = had(12, 0);
                  b2  = j(12, 1, 1) @ sortmat(fac(28, 12, 2) ||
                        (1:7)` @ j(4, 1, 1), 1);
                  d02 = ks(b1, sortmat(hadw(28, 0), 1:2)[,2:27], 2) + 1 ||
                        b2[,2:12];
                  d07 = b2[,13];
                  d24 = (1:24)` @ j(14, 1, 1);
                  end;

               else if index(lineage, '2 ** 287 14 ** 1 24 ** 1') then do;
                  d02 = twolev(had(12, 1)[,2:12] @
                        sortmat(hadw(28, 1), 1:2)[,3:28]) ||
                        j(12, 1, 1) @ twolev({1 -1}` @ j(7, 1, 1) @ {1 -1}`);
                  d14 = j(24, 1, 1) @ (1:14)`;
                  d24 = (1:24)` @ j(14, 1, 1);
                  end;

               else if index(lineage, '2 ** 252 84 ** 1') then do;
                  d02 = twolev((h2 @ h2)[,2:4] @ hadp1(84, 1));
                  d84 = j(4, 1, 1) @ (1:84)`;
                  end;

               else if index(lineage, '4 ** 36 84 ** 1') then do;
                  d04 = ks((0:3)`, gh(84, 4), 4) + 1;
                  d84 = j(4, 1, 1) @ (1:84)`;
                  end;

               %end;

            %*------ 338 runs ------;
            %else %if &tabsize eq 338 %then %do;

               if index(lineage, '13 ** 26 26 ** 1') then do;
                  d13 = ks((0:12)`, gh(26, 13), 13) + 1;
                  d26 = j(13, 1, 1) @ (1:26)`;
                  end;

               %end;

            %*------ 342 runs ------;
            %else %if &tabsize eq 342 %then %do;

               if index(lineage, '3 ** 30 114 ** 1') then do;
                  d03  = ks((0:2)`, gh(114, 3), 3) + 1;
                  d114 = j(3, 1, 1) @ (1:114)`;
                  end;

               %end;

            %*------ 343 runs ------;
            %else %if &tabsize eq 343 %then %do;

               if index(lineage, '7 ** 49 49 ** 1') then do;
                  d07 = ks((0:6)`, gh(49, 7), 7) + 1;
                  d49 = j(7, 1, 1) @ (1:49)`;
                  end;

               %end;

            %*------ 350 runs ------;
            %else %if &tabsize eq 350 %then %do;

               if index(lineage, '5 ** 20 70 ** 1') then do;
                  d05 = ks((0:4)`, gh(70, 5), 5) + 1;
                  d70 = j(5, 1, 1) @ (1:70)`;
                  end;

               %end;

            %*------ 351 runs ------;
            %else %if &tabsize eq 351 %then %do;

               if index(lineage, '3 ** 39 117 ** 1') then do;
                  d03  = ks((0:2)`, gh(117, 3), 3) + 1;
                  d117 = j(3, 1, 1) @ (1:117)`;
                  end;

               %end;

            %*------ 352 runs ------;
            %else %if &tabsize eq 352 %then %do;

               if index(lineage, '2 ** 264 88 ** 1') then do;
                  d02 = twolev((h2 @ h2)[,2:4] @ h2 @ hadp1(44, 1));
                  d88 = j(4, 1, 1) @ (1:88)`;
                  end;

               else if index(lineage, '4 ** 32 88 ** 1') then do;
                  d04 = ks((0:3)`, gh(88, 4), 4) + 1;
                  d88 = j(4, 1, 1) @ (1:88)`;
                  end;

               %end;

            %*------ 360 runs ------;
            %else %if &tabsize eq 360 %then %do;

               if index(lineage, '3 ** 48 120 ** 1') then do;
                  d03  = ks((0:2)`, gh(120, 3), 3) + 1;
                  d120 = j(3, 1, 1) @ (1:120)`;
                  end;

               else if index(lineage, '6 ** 8 60 ** 1') then do;
                  d06 = ks((0:5)`, gh(60, 6), 6) + 1;
                  d60 = j(6, 1, 1) @ (1:60)`;
                  end;

               %end;

            %*------ 361 runs ------;
            %else %if &tabsize eq 361 %then %do;

               if index(lineage, '19 ** 20') then do;
                  d19 = (ks((0:18)`, gh(19, 19), 19) + 1) ||
                        j(19, 1, 1) @ (1:19)`;
                  end;

               %end;

            %*------ 363 runs ------;
            %else %if &tabsize eq 363 %then %do;

               if index(lineage, '11 ** 11 33 ** 1') then do;
                  d11 = ks((0:10)`, gh(33, 11), 11) + 1;
                  d33 = j(11, 1, 1) @ (1:33)`;
                  end;

               %end;

            %*------ 368 runs ------;
            %else %if &tabsize eq 368 %then %do;

               if index(lineage, '2 ** 276 92 ** 1') then do;
                  d02 = twolev((h2 @ h2)[,2:4] @ hadw(92, 1));
                  d92 = j(4, 1, 1) @ (1:92)`;
                  end;

               else if index(lineage, '4 ** 36 92 ** 1') then do;
                  d04 = ks((0:3)`, gh(92, 4), 4) + 1;
                  d92 = j(4, 1, 1) @ (1:92)`;
                  end;

               %end;

            %*------ 369 runs ------;
            %else %if &tabsize eq 369 %then %do;

               if index(lineage, '3 ** 30 123 ** 1') then do;
                  d03  = ks((0:2)`, gh(123, 3), 3) + 1;
                  d123 = j(3, 1, 1) @ (1:123)`;
                  end;

               %end;

            %*------ 375 runs ------;
            %else %if &tabsize eq 375 %then %do;

               if index(lineage, '5 ** 40 75 ** 1') then do;
                  d05 = ks((0:4)`, gh(75, 5), 5) + 1;
                  d75 = j(5, 1, 1) @ (1:75)`;
                  end;

               %end;

            %*------ 378 runs ------;
            %else %if &tabsize eq 378 %then %do;

               if index(lineage, '3 ** 72 126 ** 1') then do;
                  d03  = ks((0:2)`, gh(126, 3), 3) + 1;
                  d126 = j(3, 1, 1) @ (1:126)`;
                  end;

               %end;

            %*------ 384 runs ------;
            %else %if &tabsize eq 384 %then %do;

               if index(lineage, '4 ** 96 96 ** 1') then do;
                  d04 = ks((0:3)`, gh(96, 4), 4) + 1;
                  d96 = j(4, 1, 1) @ (1:96)`;
                  end;

               else if index(lineage, '8 ** 16 48 ** 1') then do;
                  d08 = ks((0:7)`, gh(48, 8), 8) + 1;
                  d48 = j(8, 1, 1) @ (1:48)`;
                  end;

               %end;

            %*------ 387 runs ------;
            %else %if &tabsize eq 387 %then %do;

               if index(lineage, '3 ** 48 129 ** 1') then do;
                  d03  = ks((0:2)`, gh(129, 3), 3) + 1;
                  d129 = j(3, 1, 1) @ (1:129)`;
                  end;

               %end;

            %*------ 392 runs ------;
            %else %if &tabsize eq 392 %then %do;

               if index(lineage, '2 ** 196 7 ** 28 28 ** 1') then do;
                  if any02 then d02 = twolev({1 -1}` @ hadp2(196, 1));
                  d07 = {1 1}` @ ks((0:6)`, gh(28, 7), 7) + 1;
                  d28 = j(14, 1, 1) @ (1:28)`;
                  end;

               else if index(lineage, '7 ** 28 56 ** 1') then do;
                  d07 = ks((0:6)`, gh(56, 7), 7) + 1;
                  d56 = j(7, 1, 1) @ (1:56)`;
                  end;

               else if index(lineage, '14 ** 5 28 ** 1') then do;
                  d14 = ks((0:13)`, gh(28, 14), 14) + 1;
                  d28 = j(14, 1, 1) @ (1:28)`;
                  end;

               %end;

            %*------ 396 runs ------;
            %else %if &tabsize eq 396 %then %do;

               if index(lineage, '3 ** 132 132 ** 1') then do;
                  d03  = ks((0:2)`, gh(132, 3), 3) + 1;
                  d132 = j(3, 1, 1) @ (1:132)`;
                  end;

               else if index(lineage, '6 ** 2 66 ** 1') then do;
                  d06 = ks((0:5)`, gh(66, 6), 6) + 1;
                  d66 = j(6, 1, 1) @ (1:66)`;
                  end;

               %end;

            %*------ 400 runs ------;
            %else %if &tabsize eq 400 %then %do;

               if index(lineage, '2 ** 300 100 ** 1') then do;
                  d02  = twolev((h2 @ h2)[,2:4] @ hadw(100, 1));
                  d100 = j(4, 1, 1) @ (1:100)`;
                  end;

               else if index(lineage, '4 ** 36 100 ** 1') then do;
                  d04  = ks((0:3)`, gh(100, 4), 4) + 1;
                  d100 = j(4, 1, 1) @ (1:100)`;
                  end;

               else if index(lineage, '5 ** 80 80 ** 1') then do;
                  d05 = ks((0:4)`, gh(80, 5), 5) + 1;
                  d80 = j(5, 1, 1) @ (1:80)`;
                  end;

               else if index(lineage, '10 ** 6 40 ** 1') then do;
                  d10 = ks((0:9)`, gh(40, 10), 10) + 1;
                  d40 = j(10, 1, 1) @ (1:40)`;
                  end;

               else if index(lineage, '20 ** 5') then do;
                  d20 = ks((0:19)`, gh(20, 20), 20) + 1 ||
                        j(20, 1, 1) @ (1:20)`;
                  end;

               %end;

            %*------ 405 runs ------;
            %else %if &tabsize eq 405 %then %do;

               if index(lineage, '3 ** 81 135 ** 1') then do;
                  d03  = ks((0:2)`, gh(135, 3), 3) + 1;
                  d135 = j(3, 1, 1) @ (1:135)`;
                  end;

               else if index(lineage, '9 ** 18 45 ** 1') then do;
                  d09 = ks((0:8)`, gh(45, 9), 9) + 1;
                  d45 = j(9, 1, 1) @ (1:45)`;
                  end;

               %end;

            %*------ 414 runs ------;
            %else %if &tabsize eq 414 %then %do;

               if index(lineage, '3 ** 48 138 ** 1') then do;
                  d03  = ks((0:2)`, gh(138, 3), 3) + 1;
                  d138 = j(3, 1, 1) @ (1:138)`;
                  end;

               %end;

            %*------ 416 runs ------;
            %else %if &tabsize eq 416 %then %do;

               if index(lineage, '2 ** 312 104 ** 1') then do;
                  d02  = twolev((h2 @ h2 @ h2)[,3:8] @ hadw(52, 1));
                  d104 = j(4, 1, 1) @ (1:104)`;
                  end;

               else if index(lineage, '4 ** 48 104 ** 1') then do;
                  d04  = ks((0:3)`, gh(104, 4), 4) + 1;
                  d104 = j(4, 1, 1) @ (1:104)`;
                  end;

               %end;

            %*------ 423 runs ------;
            %else %if &tabsize eq 423 %then %do;

               if index(lineage, '3 ** 30 141 ** 1') then do;
                  d03  = ks((0:2)`, gh(141, 3), 3) + 1;
                  d141 = j(3, 1, 1) @ (1:141)`;
                  end;

               %end;

            %*------ 425 runs ------;
            %else %if &tabsize eq 425 %then %do;

               if index(lineage, '5 ** 20 85 ** 1') then do;
                  d05 = ks((0:4)`, gh(85, 5), 5) + 1;
                  d85 = j(5, 1, 1) @ (1:85)`;
                  end;

               %end;

            %*------ 432 runs ------;
            %else %if &tabsize eq 432 %then %do;

               if index(lineage, '2 ** 389 9 ** 1 24 ** 1') then do;
                  b1  = had(12, 0);
                  b2  = j(12, 1, 1) @ sortmat(fac(36, 16, 2) ||
                        (1:9)` @ j(4, 1, 1), 1);
                  d02 = ks(b1, sortmat(hadw(36, 0), 1:2)[,2:35], 2) + 1 ||
                        b2[,2:16];
                  d09 = b2[,17];
                  d24 = (1:24)` @ j(18, 1, 1);
                  end;

               else if index(lineage, '2 ** 375 18 ** 1 24 ** 1') then do;
                  d02 = twolev(had(12, 1)[,2:12] @
                        sortmat(hadw(36, 1), 1:2)[,3:36]) ||
                        j(12, 1, 1) @ twolev({1 -1}` @ j(9, 1, 1) @ {1 -1}`);
                  d18 = j(24, 1, 1) @ (1:18)`;
                  d24 = (1:24)` @ j(18, 1, 1);
                  end;

               else if index(lineage, '3 ** 144 144 ** 1') then do;
                  d03  = ks((0:2)`, gh(144, 3), 3) + 1;
                  d144 = j(3, 1, 1) @ (1:144)`;
                  end;

               else if index(lineage, '4 ** 108 108 ** 1') then do;
                  d04  = ks((0:3)`, gh(108, 4), 4) + 1;
                  d108 = j(4, 1, 1) @ (1:108)`;
                  end;

               else if index(lineage, '6 ** 12 72 ** 1') then do;
                  d06 = ks((0:5)`, gh(72, 6), 6) + 1;
                  d72 = j(6, 1, 1) @ (1:72)`;
                  end;

               else if index(lineage, '12 ** 6 36 ** 1') then do;
                  d12 = ks((0:11)`, gh(36, 12), 12) + 1;
                  d36 = j(12, 1, 1) @ (1:36)`;
                  end;

               %end;

            %*------ 441 runs ------;
            %else %if &tabsize eq 441 %then %do;

               if index(lineage, '3 ** 42 7 ** 7 21 ** 1') then do;
                  d03 = ks((0:2)`, gh(147, 3), 3) + 1;
                  d07 = j(3, 1, 1) @ ks((0:6)`, gh(21, 7), 7) + 1;
                  d21 = j(21, 1, 1) @ (1:21)`;
                  end;

               else if index(lineage, '7 ** 14 63 ** 1') then do;
                  d07 = ks((0:6)`, gh(63, 7), 7) + 1;
                  d63 = j(7, 1, 1) @ (1:63)`;
                  end;

               else if index(lineage, '21 ** 7') then do;
                  d21 = ks((0:20)`, gh(21, 21), 21) + 1 ||
                        j(21, 1, 1) @ (1:21)`;
                  end;

               %end;

            %*------ 448 runs ------;
            %else %if &tabsize eq 448 %then %do;

               if index(lineage, '2 ** 336 112 ** 1') then do;
                  d02  = twolev((h2 @ h2)[,2:4] @ h2 @ h2 @ hadw(28, 1));
                  d112 = j(4, 1, 1) @ (1:112)`;
                  end;

               else if index(lineage, '4 ** 56 112 ** 1') then do;
                  d04  = ks((0:3)`, gh(112, 4), 4) + 1;
                  d112 = j(4, 1, 1) @ (1:112)`;
                  end;

               else if index(lineage, '8 ** 56 56 ** 1') then do;
                  d08 = ks((0:7)`, gh(56, 8), 8) + 1;
                  d56 = j(8, 1, 1) @ (1:56)`;
                  end;

               %end;

            %*------ 450 runs ------;
            %else %if &tabsize eq 450 %then %do;

               if index(lineage, '3 ** 150 5 ** 11 30 ** 1') then do;
                  d03 = ks((0:2)`, gh(150, 3), 3) + 1;
                  d05 = j(3, 1, 1) @ (ks((0:4)`, gh(30, 5), 5) + 1);
                  d30 = j(3, 1, 1) @ j(5, 1, 1) @ (1:30)`;
                  end;

               else if index(lineage, '5 ** 90 90 ** 1') then do;
                  d05 = ks((0:4)`, gh(90, 5), 5) + 1;
                  d90 = j(5, 1, 1) @ (1:90)`;
                  end;

               else if index(lineage, '15 ** 5 30 ** 1') then do;
                  d15 = ks((0:14)`, gh(30, 15), 15) + 1;
                  d30 = j(15, 1, 1) @ (1:30)`;
                  end;

               %end;

            %*------ 459 runs ------;
            %else %if &tabsize eq 459 %then %do;

               if index(lineage, '3 ** 72 9 ** 1 17 ** 1') then do;
                  d03 = ks((0:2)`, gh(153, 3), 3) + 1;
                  d09 = j(3, 1, 1) @ (1:9)` @ j(17, 1, 1);
                  d17 = j(27, 1, 1) @ (1:17)`;
                  end;

               %end;

            %*------ 464 runs ------;
            %else %if &tabsize eq 464 %then %do;

               if index(lineage, '2 ** 348 116 ** 1') then do;
                  d02  = twolev((h2 @ h2)[,2:4] @ hadw(116, 1));
                  d116 = j(4, 1, 1) @ (1:116)`;
                  end;

               else if index(lineage, '4 ** 36 116 ** 1') then do;
                  d04  = ks((0:3)`, gh(116, 4), 4) + 1;
                  d116 = j(4, 1, 1) @ (1:116)`;
                  end;

               %end;

            %*------ 468 runs ------;
            %else %if &tabsize eq 468 %then %do;

               if index(lineage, '3 ** 49 52 ** 1') then do;
                  d03 = ks((0:2)`, gh(156, 3), 3) + 1 ||
                        j(3, 1, 1) @ (1:3)` @ j(52, 1, 1);
                  d52 = j(9, 1, 1) @ (1:52)`;
                  end;

               else if index(lineage, '6 ** 2 78 ** 1') then do;
                  d06 = ks((0:5)`, gh(78, 6), 6) + 1;
                  d78 = j(6, 1, 1) @ (1:78)`;
                  end;

               %end;

            %*------ 475 runs ------;
            %else %if &tabsize eq 475 %then %do;

               if index(lineage, '5 ** 20 95 ** 1') then do;
                  d05 = ks((0:4)`, gh(95, 5), 5) + 1;
                  d95 = j(5, 1, 1) @ (1:95)`;
                  end;

               %end;

            %*------ 477 runs ------;
            %else %if &tabsize eq 477 %then %do;

               if index(lineage, '3 ** 37 53 ** 1') then do;
                  d03 = ks((0:2)`, gh(159, 3), 3) + 1 ||
                        j(3, 1, 1) @ (1:3)` @ j(53, 1, 1);
                  d53 = j(9, 1, 1) @ (1:53)`;
                  end;

               %end;

            %*------ 480 runs ------;
            %else %if &tabsize eq 480 %then %do;

               if index(lineage, '2 ** 360 120 ** 1') then do;
                  d02  = twolev((h2 @ h2)[,2:4] @ h2 @ hadp1(60, 1));
                  d120 = j(4, 1, 1) @ (1:120)`;
                  end;

               if index(lineage, '4 ** 56 120 ** 1') then do;
                  d04  = ks((0:3)`, gh(120, 4), 4) + 1;
                  d120 = j(4, 1, 1) @ (1:120)`;
                  end;

               %end;

            %*------ 484 runs ------;
            %else %if &tabsize eq 484 %then %do;

               if index(lineage, '11 ** 44 44 ** 1') then do;
                  d11 = ks((0:10)`, gh(44, 11), 11) + 1;
                  d44 = j(11, 1, 1) @ (1:44)`;
                  end;

               else if index(lineage, '22 ** 3') then do;
                  d22 = ks((0:21)`, gh(22, 22), 22) + 1 ||
                        j(22, 1, 1) @ (1:22)`;
                  end;

               %end;

            %*------ 486 runs ------;
            %else %if &tabsize eq 486 %then %do;

               if index(lineage, '9 ** 54 54 ** 1') then do;
                  d09 = ks((0:8)`, gh(54, 9), 9) + 1;
                  d54 = j(9, 1, 1) @ (1:54)`;
                  end;

               %end;

            %*------ 490 runs ------;
            %else %if &tabsize eq 490 %then %do;

               if index(lineage, '7 ** 18 70 ** 1') then do;
                  d07 = ks((0:6)`, gh(70, 7), 7) + 1;
                  d70 = j(7, 1, 1) @ (1:70)`;
                  end;

               %end;

            %*------ 495 runs ------;
            %else %if &tabsize eq 495 %then %do;

               if index(lineage, '3 ** 42 5 ** 1 33 ** 1') then do;
                  d03 = ks((0:2)`, gh(165, 3), 3) + 1;
                  d05 = j(3, 1, 1) @ (1:5)` @ j(33, 1, 1);
                  d33 = j(15, 1, 1) @ (1:33)`;
                  end;

               %end;

            %*------ 496 runs ------;
            %else %if &tabsize eq 496 %then %do;

               if index(lineage, '2 ** 372 124 ** 1') then do;
                  d02  = twolev((h2 @ h2)[,2:4] @ hadw(124, 1));
                  d124 = j(4, 1, 1) @ (1:124)`;
                  end;

               else if index(lineage, '4 ** 18 124 ** 1') then do;
                  d04  = ks((0:3)`, gh(124, 4), 4) + 1;
                  d124 = j(4, 1, 1) @ (1:124)`;
                  end;

               %end;

            %*------ 500 runs ------;
            %else %if &tabsize eq 500 %then %do;

               if index(lineage, '5 ** 100 100 ** 1') then do;
                  d05  = ks((0:4)`, gh(100, 5), 5) + 1;
                  d100 = j(5, 1, 1) @ (1:100)`;
                  end;

               else if index(lineage, '10 ** 2 50 ** 1') then do;
                  d10 = ks((0:9)`, gh(50, 10), 10) + 1;
                  d50 = j(10, 1, 1) @ (1:50)`;
                  end;

               %end;

            %*------ 504 runs ------;
            %else %if &tabsize eq 504 %then %do;

               if index(lineage, '2 ** 84 3 ** 84 84 ** 1') then do;
                  d03 = ks((0:2)`, gh(168, 3), 3) + 1;
                  d02 = j(3, 1, 1) @ twolev({1 -1}` @ hadp1(84, 1));
                  d84 = j(6, 1, 1) @ (1:84)`;
                  end;

               else if index(lineage, '6 ** 8 84 ** 1') then do;
                  d06 = ks((0:5)`, gh(84, 6), 6) + 1;
                  d84 = j(6, 1, 1) @ (1:84)`;
                  end;

               %end;

            %*------ 512 runs ------;
            %else %if &tabsize eq 512 %then %do;

               if index(lineage, '8 ** 64 64 ** 1') then do;
                  d08 = ks((0:7)`, gh(64, 8), 8) + 1;
                  d64 = j(8, 1, 1) @ (1:64)`;
                  end;

               else if index(lineage, '4 ** 160 32 ** 1') then do;
                  d04 = (ks((0:3)`, gh(128, 4), 4) ||
                        j(4, 1, 1) @ ks((0:3)`, gh(32, 4), 4)) + 1;
                  d32 = j(16, 1, 1) @ (1:32)`;
                  end;

               else if index(lineage, '16 ** 32 32 ** 1') then do;
                  d16 = ks((0:15)`, gh(32, 16), 16) + 1;
                  d32 = j(16, 1, 1) @ (1:32)`;
                  end;

               %end;

            %*------ 513 runs ------;
            %else %if &tabsize eq 513 %then %do;

               if index(lineage, '3 ** 81 9 ** 1 19 ** 1') then do;
                  d03 = ks((0:2)`, gh(171, 3), 3) + 1;
                  d09 = j(3, 1, 1) @ (1:9)` @ j(19, 1, 1);
                  d19 = j(27, 1, 1) @ (1:19)`;
                  end;

               %end;

            %*------ 576 runs ------;
            %else %if &tabsize eq 576 %then %do;

               if index(lineage, '24 ** 8') then do;
                  d24 = ks((0:23)`, gh(24, 24), 24) + 1 || j(24, 1, 1) @ (1:24)`;
                  end;

               %end;

            %*------ Ad hoc larger Hadamard constructions ------;
            %let adhochad = ;

            * Need to be here to handle larger sizes, e.g. 2 * 188 as well.;
            %if &hadsize eq 188 or &hadsize eq 236 or &hadsize eq 260 or
                &hadsize eq 268 or &hadsize eq 292 or &hadsize eq 356 or
                &hadsize eq 404 or &hadsize eq 436 or &hadsize eq 596
                %then %let adhochad = hadthw;

            %else %if &hadsize eq 324 and &tabsize gt 324 %then
               %let adhochad = hadc;

            %else %if &hadsize eq 340 or &hadsize eq 580 or &hadsize eq 724
               or &hadsize eq 1060 %then %let adhochad = hadp2pp;

            %else %if &hadsize eq 372 or &hadsize eq 612 or &hadsize eq 732
               or &hadsize eq 756 %then %let adhochad = hadd;

            %else %if &hadsize eq 412 or &hadsize eq 604
               %then %let adhochad = hady;

            %else %if &hadsize eq 428 %then %let adhochad = hadk;

            %if &adhochad ne %then %do;
               %if &flow or &flowinit %then %do;
                  file log;
                  put "** Making Hadamard &hadsize, &hadsize construction.";
                  file print;
                  %end;
               if ncol(d02) = 0 then h = &adhochad(&hadsize);
               %end;

            *------read full-factorial or FACTEX designs------;
            %if &factexnlevs or &full %then %do;
               free d&n;
               %if &flow or &flowinit %then %do;
                  file log;
                  put "** Reading PLAN or FACTEX design.";
                  file print;
                  %end;
               %do j = 2 %to &maxlevneed;
                  %if &j > 9 %then %let jj = &j; %else %let jj = 0&j;
                  %if %index(&tableflag._, _&jj._) %then %do;
                     use __tmpxcand(keep=x&jj._:);
                     read all into d&jj;
                     close __tmpxcand;
                     %end;
                  %end;
               %end;

            free b1 b2;

            *============ Begin Recoding High-Level to Lower-Level ============;

            *------144 runs, recode 144-level factors------;
            avail = max(0, ncol(d144) - any144);
            do while(avail);

               if index(lineage, '144 ** 1 > 2 ** 136 8 ** 1') then do;
                  b1  = twolev((h2 @ h2) @ sortmat(hadw(36, 1), 1:2));
                  b2  = j(1, 144, 1); b2[{1 2 37 38 73 74 109 110}] = 0;
                  d02 = d02 || b1[d144[,1],loc(b2)];
                  d08 = d08 || ((1:8)` @ j(18, 1, 1))[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 2 ** 117 8 ** 1 9 ** 1')
                  then do;
                  b1  = j(4, 1, 1) @ sortmat(fac(36, 16, 2) ||
                        (1:9)` @ j(4, 1, 1), 1);
                  d02 = d02 || (twolev((h2 @ h2)[,2:4] @
                               sortmat(hadw(36, 1), 1:2)[,3:36]) ||
                               b1[,2:16])[d144[,1],];
                  d08 = d08 || ((1:8)` @ j(18, 1, 1))[d144[,1],];
                  d09 = d09 || (b1[,17])[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 2 ** 113 3 ** 1 24 ** 1')
                  then do;
                  b1  = had(12, 0) - 1;
                  b2  = j(12, 1, 1) @ sortmat((fac(12, 4, 2) ||
                        (1:3)` @ j(4, 1, 1)), 1);
                  d02 = d02 || (ks(b1, sortmat(b1, 1:2)[,2:11], 2)  + 1 ||
                               b2[,2:4])[d144[,1],];
                  d03 = d03 || (b2[,5])[d144[,1],];
                  d24 = d24 || ((1:24)` @ j(6, 1, 1))[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 2 ** 111 6 ** 1 24 ** 1')
                  then do;
                  b1  = had(12, 0) - 1;
                  d02 = d02 || (ks(b1, sortmat(b1, 1:2)[,2:11], 2)  + 1 ||
                               j(12, 1, 1) @ twolev({1 -1}` @ j(3, 1, 1) @
                               {1 -1}`))[d144[,1],];
                  d06 = d06 || (j(24, 1, 1) @ (1:6)`)[d144[,1],];
                  d24 = d24 || ((1:24)` @ j(6, 1, 1))[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 2 ** 103 8 ** 1 18 ** 1')
                  then do;
                  b1  = 1;
                  d02 = d02 || (twolev((h2 @ h2)[,2:4] @
                               sortmat(hadw(36, 1), 1:2)[,3:36]) ||
                               j(4, 1, 1) @ twolev({1 -1}` @ j(9, 1, 1) @
                               {1 -1}`))[d144[,1],];
                  d18 = d18 || (j(8, 1, 1) @ (1:18)`)[d144[,1],];
                  d08 = d08 || ((1:8)` @ j(18, 1, 1))[d144[,1],];
                  end;

               else if index(lineage,
                  '144 ** 1 > 2 ** 76 3 ** 7 4 ** 1 6 ** 5 12 ** 1') then do;
                  b1  = ks(j(3, 1, 1) @ (0:1)`, fac(12, 8, 2), 2) + 1 ||
                        ks(j(2, 1, 1) @ (0:2)`, fac(12, 7, 3), 3) + 1 ||
                        ks((0:5)`, gh(12, 6)[,{1 3 4 5 6}], 6) + 1 ||
                        j(6, 1, 1) @ (1:12)`;
                  b1  = j(2, 1, 1) @ sortmat(b1, 1);
                  b2  = (twolev({1 -1}` @ h2 @ hadw(36, 1)))[,(2:36)||(38:72)];
                  d02 = d02 || b2[d144[,1],] || b1[d144[,1],2:7];
                  d03 = d03 || b1[d144[,1], 8:14];
                  d06 = d06 || b1[d144[,1],15:19];
                  d12 = d12 || b1[d144[,1],20];
                  d04 = d04 || ((1:4)` @ j(36, 1, 1))[d144[,1],];
                  end;

               else if index(lineage,
                  '144 ** 1 > 2 ** 76 3 ** 12 6 ** 4 8 ** 1') then do;
                  b1  = j(2, 1, 1) @ sortmat(fac(72, 14, 0), 1);
                  b2  = (twolev({1 -1}` @ h2 @
                                sortmat(hadw(36, 1), 1:2)))[,(3:36)||(39:72)];
                  d02 = d02 || b2[d144[,1],] || b1[d144[,1],18:25];
                  d03 = d03 || b1[d144[,1],6:17];
                  d06 = d06 || b1[d144[,1],2: 5];
                  d08 = d08 || ((1:8)` @ j(18, 1, 1))[d144[,1],];
                  end;

               else if index(lineage,
                  '144 ** 1 > 2 ** 75 3 ** 3 4 ** 1 6 ** 6 12 ** 1') then do;
                  b1  = twolev({1 -1}` @ h2 @ hadw(36, 1));
                  b2  = ks(j(3, 1, 1) @ (0:1)`, fac(12, 6, 2), 2) + 1 ||
                        ks(j(2, 1, 1) @ (0:2)`, fac(12, 3, 3), 3) + 1 ||
                        ks((0:5)`, gh(12, 6), 6) + 1 || (j(6, 1, 1) @ (1:12)`);
                  b2  = sortmat(b2, 1);
                  d02 = d02 || b1[d144[,1],(2:36) || (38:72)] ||
                               (j(2, 1, 1) @ b2[, 2: 6])[d144[,1],];
                  d03 = d03 || (j(2, 1, 1) @ b2[, 7: 9])[d144[,1],];
                  d04 = d04 || ((1:4)` @ j(36, 1, 1)   )[d144[,1],];
                  d06 = d06 || (j(2, 1, 1) @ b2[,10:15])[d144[,1],];
                  d12 = d12 || (j(2, 1, 1) @ b2[,   16])[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 2 ** 74 3 ** 4 6 ** 6 8 ** 1')
                  then do;
                  b1  = twolev({1 -1}` @ h2 @ sortmat(hadw(36, 1), 1:2));
                  b2  = j(18, 1, 1) @ (1:4)` ||
                        ks(j(3, 1, 1) @ (0:1)`, fac(12, 6, 2), 2) + 1 ||
                        ks(j(2, 1, 1) @ (0:2)`, fac(12, 3, 3), 3) + 1 ||
                        j(6, 1, 1) @ (1:3)` @ j(4, 1, 1) ||
                        ks((0:5)`, gh(12, 6), 6) + 1;
                  b2  = sortmat(b2, 1);
                  d02 = d02 || b1[d144[,1],(3:36) || (39:72)] ||
                               (j(2, 1, 1) @ b2)[d144[,1], 2: 7];
                  d03 = d03 || (j(2, 1, 1) @ b2)[d144[,1], 8:11];
                  d06 = d06 || (j(2, 1, 1) @ b2)[d144[,1],12:17];
                  d08 = d08 || ((1:8)` @ j(18, 1, 1))[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 2 ** 72 72 ** 1') then do;
                  d02 = d02 || (twolev({1 -1}` @ h2 @ hadw(36, 1)))[d144[,1],];
                  d72 = d72 || (j(2, 1, 1) @ (1:72)`)[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 2 ** 44 3 ** 11 12 ** 2')
                  then do;
                  b1  = fac(12, 4, 2) || (1:3)` @ j(4, 1, 1);
                  d02 = d02 || (ks(had(12, 0) - 1, b1[,1:4] - 1, 2) + 1)
                               [d144[,1],];
                  d03 = d03 || (ks(gh(12, 3)[,2:12], b1[,5], 3) + 1)[d144[,1],];
                  d12 = d12 || (j(12, 1, 1) @ (1:12)` || (1:12)` @ j(12, 1, 1))
                               [d144[,1],];
                  end;

               else if index(lineage,
                  '144 ** 1 > 2 ** 16 3 ** 3 6 ** 6 24 ** 1') then do;
                  b1  = j(6, 1, 1) @ (1:12)`;
                  b2  = ks((0:5)`, gh(12, 6), 6) + 1;
                  b3  = ks(j(2, 1, 1) @ (0:2)`, fac(12, 3, 3), 3) + 1;
                  b4  = ks(j(3, 1, 1) @ (0:1)`, fac(12, 6, 2), 2) + 1;
                  d02 = d02 || (b4 // b4 || fac(144, 10, 2))[d144[,1],];
                  d03 = d03 || (b3 // b3)[d144[,1],];
                  d06 = d06 || (b2 // b2)[d144[,1],];
                  d24 = d24 || (b1 // b1 + 12)[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 3 ** 48 48 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (ks((0:2)`, gh(48, 3), 3) + 1)[d144[,1],];
                  d48 = d48 || (j(3, 1, 1) @ (1:48)`)[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 4 ** 36 36 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (ks((0:3)`, gh(36, 4), 4) + 1)[d144[,1],];
                  d36 = d36 || (j(4, 1, 1) @ (1:36)`)[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 4 ** 11 12 ** 2') then do;
                  b1  = ks((0:3)`, gh(12, 4), 4) + 1;
                  d04 = d04 || (j(4, 1, 1) @ b1)[d144[,1],2:12];
                  d12 = d12 || (j(12, 1, 1) @ (1:12)` ||
                               (1:12)` @ j(12, 1, 1))[d144[,1],];
                  end;

               else if index(lineage, '144 ** 1 > 12 ** 7') then do;
                  b1  = 1;
                  d12 = d12 || (ks((0:11)`, gh(12, 12), 12) + 1 ||
                               j(12, 1, 1) @ (1:12)`)[d144[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d144) = 1 then free d144;
                  else do; file log; put 'ERROR: 144 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------143 runs, recode 143-level factors------;
            avail = max(0, ncol(d143) - any143);
            do while(avail);

               if index(lineage, '143 ** 1 > 11 ** 1 13 ** 1') then do;
                  b1  = 1;
                  d11 = d11 || (j(13, 1, 1) @ (1:11)`)[d143[,1],];
                  d13 = d13 || ((1:13)` @ j(11, 1, 1))[d143[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d143) = 1 then free d143;
                  else do; file log; put 'ERROR: 143 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------142 runs, recode 142-level factors------;
            avail = max(0, ncol(d142) - any142);
            do while(avail);

               if index(lineage, '142 ** 1 > 2 ** 1 71 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(71, 1, 1) @ (1:2)`)[d142[,1],];
                  d71 = d71 || ((1:71)` @ j(2, 1, 1))[d142[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d142) = 1 then free d142;
                  else do; file log; put 'ERROR: 142 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------141 runs, recode 141-level factors------;
            avail = max(0, ncol(d141) - any141);
            do while(avail);

               if index(lineage, '141 ** 1 > 3 ** 1 47 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(47, 1, 1) @ (1:3)`)[d141[,1],];
                  d47 = d47 || ((1:47)` @ j(3, 1, 1))[d141[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d141) = 1 then free d141;
                  else do; file log; put 'ERROR: 141 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------140 runs, recode 140-level factors------;
            avail = max(0, ncol(d140) - any140);
            do while(avail);

               if index(lineage, '140 ** 1 > 2 ** 139') then do;
                  b1  = 1;
                  d02 = d02 || hadp1(140, 0)[d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 2 ** 38 7 ** 1') then do;
                  run proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
                  b5  = twolev(h2 @ had(20, 1));
                  d02 = d02 || ((h20[s1,3:19] || h20[s2,3:19] || h20[s3,3:6]) //
                               b5[,2:39])[d140[,1],];
                  d07 = d07 || (b2[s4,] // (b5[,40] + 5))[d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 2 ** 36 10 ** 1') then do;
                  b1  = twolev(sortmat(hadw(28, 1), 1:2));
                  d02 = d02 || (j(5, 1, 1) @ b1[,3:27] ||
                                fac(140, 10, 2))[d140[,1],];
                  d10 = d10 || ((1:10)` @ j(14, 1, 1))[d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 2 ** 34 14 ** 1') then do;
                  run proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
                  b5  = twolev(h2 @ had(20, 1));
                  d02 = d02 || ((h20[s1,3:19] || h20[s2,3:19]) //
                               b5[,(3:20) || (23:38)])[d140[,1],];
                  d14 = d14 || (((b2[s3,] - 1) # 2 + h20[s2,1]) //
                               ((b5[,2 ] - 1) # 2 + b5[,21] + 10))[d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 2 ** 27 5 ** 1 7 ** 1')
                  then do;
                  run proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
                  b1  = h20[s1,3:19] || l20[s2,3:9] || l20[s3,3:9];
                  b5  = sortmat(had(20, 0), 1);
                  b6  = sortmat(fac(20, 8, 2) || (1:5)` @ j(4, 1, 1), 1);
                  b5  = b5  // reverse(b5);
                  b6  = (b6 // b6) || twolev((h2 @ h2)[,2:4]) @ j(10, 1, 1);
                  d02 = d02 || ((b5[,2:19] || b6[,2:8] || b6[,10:11]) //
                               b1[,(1:23) || (25:28)])[d140[,1],];
                  d05 = d05 || (b6[, 9] // b2[s4,])[d140[,1],];
                  d07 = d07 || (b6[,12] // (b2[s5,] + 2))[d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 2 ** 25 5 ** 1 14 ** 1')
                  then do;
                  run proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
                  b5  = sortmat(had(20, 0), 1);
                  b6  = sortmat(fac(20, 8, 2) || (1:5)` @ j(4, 1, 1), 1);
                  b5  = b5  // reverse(b5);
                  b6  = (b6 // b6) || (1:4)` @ j(10, 1, 1);
                  d02 = d02 || ((b5[,2:19] || b6[,2:8]) //
                               (h20[s1,3:19] || h20[s2,3:10]))[d140[,1],];
                  d05 = d05 || (b6[, 9] // b2[s4,])[d140[,1],];
                  d14 = d14 || (b6[,10] // ((b2[s3,] - 1) # 2 + h20[s2,1] + 4))
                               [d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 2 ** 22 35 ** 1') then do;
                  b1  = h2 @ h2;
                  d02 = d02 || (twolev((b1[,2:4] @ b1 @ h2)[,1:22]) //
                                fac(108, 22, 2))[d140[,1],];
                  d35 = d35 || (j(4, 1, 1) @ (1:8)` //
                                (9:35)` @ j(4, 1, 1))[d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 2 ** 21 7 ** 1 10 ** 1')
                  then do;
                  run proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
                  b5  = sortmat(had(20, 0), 1);
                  b6  = twolev({1 -1}` @ j(5, 1, 1) @ h2) || {1 1}` @ (1:10)`;
                  b5  = b5  // reverse(b5);
                  b6  = (b6 // b6) || twolev((h2 @ h2)[,2:4]) @ j(10, 1, 1);
                  d02 = d02 || ((b5[,2:19] || b6[,2 || (4:5)]) //
                               (h20[s1,3:19] || h20[s2,3:6]))[d140[,1],];
                  d10 = d10 || (b6[,3] // ((b2[s3,] - 1) # 2 + h20[s2,1]))
                               [d140[,1],];
                  d07 = d07 || (b6[,6] // (b2[s4,] + 2))[d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 2 ** 17 10 ** 1 14 ** 1')
                  then do;
                  b1  = sortmat(had(20, 0), 1);
                  b2  = twolev({1 -1}` @ j(5, 1, 1) @ h2) || {1 1}` @ (1:10)`;
                  b1  = b1  // reverse(b1);
                  b2  = (b2 // b2) || (1:4)` @ j(10, 1, 1);
                  b3  = sortmat(had(20, 0), 1:2) @ j(5, 1, 1);
                  b4  = j(4, 1, 1) @ sortmat(ks((0:4)`, gh(5, 5), 5) + 1 ||
                        j(5, 1, 1) @ (1:5)`, 1:3);
                  d02 = d02 || (b1[,2:18] // b3[,3:19])[d140[,1],];
                  d10 = d10 || (b2[,3] // (b4[,2] - 1) # 2 + b3[,1])[d140[,1],];
                  d14 = d14 || (b2[,4] // (b4[,3] - 1) # 2 + b3[,2] + 4)
                               [d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 2 ** 2 70 ** 1')
                  then %mktn22(140);

               else if index(lineage, '140 ** 1 > 4 ** 1 35 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(35, 1, 1) @ (1:4)`)[d140[,1],];
                  d35 = d35 || ((1:35)` @ j(4, 1, 1))[d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 5 ** 1 28 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(28, 1, 1) @ (1:5)`)[d140[,1],];
                  d28 = d28 || ((1:28)` @ j(5, 1, 1))[d140[,1],];
                  end;

               else if index(lineage, '140 ** 1 > 7 ** 1 20 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(20, 1, 1) @ (1:7)`)[d140[,1],];
                  d20 = d20 || ((1:20)` @ j(7, 1, 1))[d140[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d140) = 1 then free d140;
                  else do; file log; put 'ERROR: 140 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------138 runs, recode 138-level factors------;
            avail = max(0, ncol(d138) - any138);
            do while(avail);

               if index(lineage, '138 ** 1 > 2 ** 1 69 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(69, 1, 1) @ (1:2)`)[d138[,1],];
                  d69 = d69 || ((1:69)` @ j(2, 1, 1))[d138[,1],];
                  end;

               else if index(lineage, '138 ** 1 > 3 ** 1 46 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(46, 1, 1) @ (1:3)`)[d138[,1],];
                  d46 = d46 || ((1:46)` @ j(3, 1, 1))[d138[,1],];
                  end;

               else if index(lineage, '138 ** 1 > 6 ** 1 23 ** 1') then do;
                  b1  = 1;
                  d06 = d06 || (j(23, 1, 1) @ (1:6)`)[d138[,1],];
                  d23 = d23 || ((1:23)` @ j(6, 1, 1))[d138[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d138) = 1 then free d138;
                  else do; file log; put 'ERROR: 138 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------136 runs, recode 136-level factors------;
            avail = max(0, ncol(d136) - any136);
            do while(avail);

               if index(lineage, '136 ** 1 > 2 ** 132 4 ** 1') then do;
                  b1  = twolev(h2 @ sortmat(hadp1(68, 1), 1:2));
                  d02 = d02 || b1[d136[,1], (3:68) || (71:136)];
                  d04 = d04 || ((1:4)` @ j(34, 1, 1))[d136[,1],];
                  end;

               else if index(lineage, '136 ** 1 > 2 ** 83 4 ** 1 17 ** 1')
                  then do;
                  b1  = fac(68, 18, 2) || (1:17)` @ j(4, 1, 1);
                  b1  = j(2, 1, 1) @ sortmat(b1, 1:2);
                  b2  = sortmat(hadp1(68, 1), 1:2);
                  d02 = d02 || (twolev({1 -1}` @ b2[,3:68]) ||
                               b1[,2:18])[d136[,1],];
                  d04 = d04 || ((1:4)` @ j(34, 1, 1))[d136[,1],];
                  d17 = d17 || b1[d136[,1],19];
                  end;

               else if index(lineage, '136 ** 1 > 2 ** 68 68 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || twolev({1 -1}` @ hadp1(68, 1))[d136[,1],];
                  d68 = d68 || (j(2, 1, 1) @ (1:68)`)[d136[,1],];
                  end;

               else if index(lineage, '136 ** 1 > 2 ** 67 4 ** 1 34 ** 1')
                  then do;
                  b1  = sortmat(twolev({1 -1}` @ j(17, 1, 1) @ h2) ||
                                j(2, 1, 1) @ (1:34)`, 1);
                  b2  = sortmat(hadp1(68, 1), 1:2);
                  d02 = d02 || (twolev({1 -1}` @ b2[,3:68]) || {1 1}` @ b1[,2])
                               [d136[,1],];
                  d04 = d04 || ((1:4)` @ j(34, 1, 1))[d136[,1],];
                  d34 = d34 || (j(4, 1, 1) @ (1:34)`)[d136[,1],];
                  end;

               else if index(lineage, '136 ** 1 > 8 ** 1 17 ** 1') then do;
                  b1  = 1;
                  d08 = d08 || (j(17, 1, 1) @ (1:8)`)[d136[,1],];
                  d17 = d17 || ((1:17)` @ j(8, 1, 1))[d136[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d136) = 1 then free d136;
                  else do; file log; put 'ERROR: 136 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------135 runs, recode 135-level factors------;
            avail = max(0, ncol(d135) - any135);
            do while(avail);

               if index(lineage, '135 ** 1 > 3 ** 27 45 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (ks((0:2)`, gh(45, 3), 3) + 1)[d135[,1],];
                  d45 = d45 || (j(3, 1, 1) @ (1:45)`)[d135[,1],];
                  end;

               else if index(lineage, '135 ** 1 > 3 ** 32 9 ** 1 15 ** 1')
                  then do;
                  b1  = gh(15, 3);
                  b2  = (1:36); b2[{1 10 19 28}] = 0;
                  d03 = d03 || (ks((0:2)`, ks(gh(3, 3), b1, 3), 3) + 1 ||
                        j(3, 1, 1) @ (ks((0:2)`, b1, 3) + 1))[d135[,1],loc(b2)];
                  d09 = d09 || ((1:9)` @ j(15, 1, 1))[d135[,1],];
                  d15 = d15 || (j(9, 1, 1) @ (1:15)`)[d135[,1],];
                  end;

               else if index(lineage, '135 ** 1 > 5 ** 1 27 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(27, 1, 1) @ (1:5)`)[d135[,1],];
                  d27 = d27 || ((1:27)` @ j(5, 1, 1))[d135[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d135) = 1 then free d135;
                  else do; file log; put 'ERROR: 135 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------134 runs, recode 134-level factors------;
            avail = max(0, ncol(d134) - any134);
            do while(avail);

               if index(lineage, '134 ** 1 > 2 ** 1 67 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(67, 1, 1) @ (1:2)`)[d134[,1],];
                  d67 = d67 || ((1:67)` @ j(2, 1, 1))[d134[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d134) = 1 then free d134;
                  else do; file log; put 'ERROR: 134 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------133 runs, recode 133-level factors------;
            avail = max(0, ncol(d133) - any133);
            do while(avail);

               if index(lineage, '133 ** 1 > 7 ** 1 19 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(19, 1, 1) @ (1:7)`)[d133[,1],];
                  d19 = d19 || ((1:19)` @ j(7, 1, 1))[d133[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d133) = 1 then free d133;
                  else do; file log; put 'ERROR: 133 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------132 runs, recode 132-level factors------;
            avail = max(0, ncol(d132) - any132);
            do while(avail);

               if index(lineage, '132 ** 1 > 2 ** 131') then do;
                  b1  = 1;
                  d02 = d02 || hadp1(132, 0)[d132[,1],];
                  end;

               else if index(lineage, '132 ** 1 > 2 ** 42 6 ** 1') then do;
                  b1  = hadp1(44, 1);
                  b2  = twolev(h2 @ b1);
                  b1  = twolev(b1);
                  d02 = d02 || (b1[,3:44] // b2[,3:44])[d132[,1],];
                  d06 = d06 || (b1[,2] //
                               ((b2[,2] - 1) # 2 + b2[,45] + 2))[d132[,1],];
                  end;

               else if index(lineage, '132 ** 1 > 2 ** 27 11 ** 1') then do;
                  b1  = sortmat(had(12, 1), 1:3);
                  b1  = twolev((h2 @ h2)[,2:4] @ h2 @ b1 ||
                               j(4, 1, 1) @ {1 -1}` @ b1);
                  b2  = (b1[,1] - 1) # 4 + (b1[,24] - 1) # 2 + b1[,48];
                  d02 = d02 || (fac(36, 27, 2) //
                               b1[,(2:23) || (26:30)])[d132[,1],];
                  d11 = d11 || (fac(36, 1, 3) // (b2 + 3))[d132[,1],];
                  end;

               else if index(lineage, '132 ** 1 > 2 ** 22 33 ** 1') then do;
                  b1  = h2 @ h2;
                  d02 = d02 || (twolev((b1[,2:4] @ b1 @ h2)[,1:22]) //
                                fac(100, 22, 2))[d132[,1],];
                  d33 = d33 || (j(4, 1, 1) @ (1:8)` //
                                (9:33)` @ j(4, 1, 1))[d132[,1],];
                  end;

               else if index(lineage, '132 ** 1 > 2 ** 18 3 ** 1 22 ** 1')
                  then do;
                  b1  = 1;
                  d02 = d02 || (twolev({1 -1}` @ h2 @ had(12, 1))[,1:18] //
                               twolev((h2 @ h2)[,2:4] @
                               sortmat(had(12, 1), 2))[,(3:12) || (15:22)] //
                               fac(36, 20, 2)[,1:18])[d132[,1],];
                  d03 = d03 || (j(2, 1, 1) @ (1:3)` @ j(8, 1, 1) //
                               j(16, 1, 1) @ (1:3)` // fac(36, 2, 3)[,1])
                               [d132[,1],];
                  d22 = d22 || (j(6, 1, 1) @ (1:8)` // (9:16)` @ j(6, 1, 1) //
                               (fac(36, 1, 6) + 16))[d132[,1],];
                  end;

               else if index(lineage, '132 ** 1 > 2 ** 18 6 ** 1 11 ** 1')
                  then do;
                  b1  = 1;
                  d02 = d02 || (twolev({1 -1}` @ h2 @ had(12, 1))[,1:18] //
                              twolev((h2 @ h2)[,2:4] @
                              sortmat(had(12, 1), 2))[,(3:12) || (15:22)] //
                              fac(36, 20, 2)[,1:18])[d132[,1],];
                  d06 = d06 || (j(2, 1, 1) @ j(4, 1, 1) @ (1:6)` //
                               j(8, 1, 1) @ (1:6)` // fac(36, 1, 6))[d132[,1],];
                  d11 = d11 || (j(2, 1, 1) @ (1:4)` @ j(6, 1, 1) //
                               (5:8)` @ j(12, 1, 1) //
                               (fac(36, 2, 3)[,1] + 8))[d132[,1],];
                  end;

               else if index(lineage, '132 ** 1 > 2 ** 15 6 ** 1 22 ** 1')
                  then do;
                  b1  = sortmat(fac(36, 1, 3) || fac(36, 27, 2), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(ks((0:2)`, gh(12, 3), 3) + 1 ||
                                j(3, 1, 1) @ (1:12)`, 1);
                  b2  = b2 // b2;
                  b3  = fac(60, 17, 0);
                  d02 = d02 || cat(b1[,2:28], b3[,1:15])[d132[,1],];
                  d06 = d06 || ((1:6)` @ j(12, 1, 1) // b3[,16])[d132[,1],];
                  d22 = d22 || (b2[,13] // b3[,17] + 12 )[d132[,1],];
                  end;

               else if index(lineage, '132 ** 1 > 2 ** 2 66 ** 1')
                  then %mktn22(132);

               else if index(lineage, '132 ** 1 > 3 ** 1 44 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(44, 1, 1) @ (1:3)`)[d132[,1],];
                  d44 = d44 || ((1:44)` @ j(3, 1, 1))[d132[,1],];
                  end;

               else if index(lineage, '132 ** 1 > 4 ** 1 33 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(33, 1, 1) @ (1:4)`)[d132[,1],];
                  d33 = d33 || ((1:33)` @ j(4, 1, 1))[d132[,1],];
                  end;

               else if index(lineage, '132 ** 1 > 11 ** 1 12 ** 1') then do;
                  b1  = 1;
                  d11 = d11 || (j(12, 1, 1) @ (1:11)`)[d132[,1],];
                  d12 = d12 || ((1:12)` @ j(11, 1, 1))[d132[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d132) = 1 then free d132;
                  else do; file log; put 'ERROR: 132 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------130 runs, recode 130-level factors------;
            avail = max(0, ncol(d130) - any130);
            do while(avail);

               if index(lineage, '130 ** 1 > 2 ** 1 65 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(65, 1, 1) @ (1:2)`)[d130[,1],];
                  d65 = d65 || ((1:65)` @ j(2, 1, 1))[d130[,1],];
                  end;

               else if index(lineage, '130 ** 1 > 5 ** 1 26 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(26, 1, 1) @ (1:5)`)[d130[,1],];
                  d26 = d26 || ((1:26)` @ j(5, 1, 1))[d130[,1],];
                  end;

               else if index(lineage, '130 ** 1 > 10 ** 1 13 ** 1') then do;
                  b1  = 1;
                  d10 = d10 || (j(13, 1, 1) @ (1:10)`)[d130[,1],];
                  d13 = d13 || ((1:13)` @ j(10, 1, 1))[d130[,1],];
                  end;

               else if index(lineage, '130 ** 1 > 11 ** 1 12 ** 1') then do;
                  b1  = 1;
                  d11 = d11 || (j(12, 1, 1) @ (1:11)`)[d130[,1],];
                  d12 = d12 || ((1:12)` @ j(11, 1, 1))[d130[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d130) = 1 then free d130;
                  else do; file log; put 'ERROR: 130 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------129 runs, recode 129-level factors------;
            avail = max(0, ncol(d129) - any129);
            do while(avail);

               if index(lineage, '129 ** 1 > 3 ** 1 43 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(43, 1, 1) @ (1:3)`)[d129[,1],];
                  d43 = d43 || ((1:43)` @ j(3, 1, 1))[d129[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d129) = 1 then free d129;
                  else do; file log; put 'ERROR: 129 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------126 runs, recode 126-level factors------;
            avail = max(0, ncol(d126) - any126);
            do while(avail);

               if index(lineage, '126 ** 1 > 2 ** 1 63 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(63, 1, 1) @ (1:2)`)[d126[,1],];
                  d63 = d63 || ((1:63)` @ j(2, 1, 1))[d126[,1],];
                  end;

               else if index(lineage, '126 ** 1 > 3 ** 24 14 ** 1') then do;
                  b1  = j(3, 1, 1) @ (0:2)` || ks((0:2)`, gh(3, 3), 3);
                  d03 = d03 || ((ks(b1, gh(6, 3), 3) //
                                 ks((0:2)`, gh(24, 3), 3)) + 1)[d126[,1],];
                  d14 = d14 || ((j(9, 1, 1) @ (1:6)`) //
                                j(9, 1, 1) @ (7:14)`)[d126[,1],];
                  end;

               else if index(lineage, '126 ** 1 > 3 ** 23 6 ** 1 7 ** 1')
                  then do;
                  b1  = j(3, 1, 1) @ (0:2)` || ks((0:2)`, gh(3, 3), 3);
                  b1  = ks(b1, gh(6, 3), 3);
                  d03 = d03 || ((b1 // ks((0:2)`, gh(24, 3), 3))[,1:23] + 1)
                               [d126[,1],];
                  d06 = d06 || (j(9, 1, 1) @ (1:6)` //
                                j(3, 1, 1) @ j(4, 1, 1) @ (1:6)`)[d126[,1],];
                  d07 = d07 || (b1[,24] + 1 //
                                j(3, 1, 1) @ (4:7)` @ j(6, 1, 1))[d126[,1],];
                  end;

               else if index(lineage, '126 ** 1 > 3 ** 21 42 ** 1') then do;
                  d03 = d03 || (ks((0:2)`, gh(42, 3), 3) + 1)[d126[,1],];
                  d42 = d42 || (j(3, 1, 1) @ (1:42)`)[d126[,1],];
                  end;

               else if index(lineage, '126 ** 1 > 3 ** 20 6 ** 1 21 ** 1')
                  then do;
                  b1  = j(3, 1, 1) @ (0:2)` || ks((0:2)`, gh(3, 3), 3);
                  b1  = ks(b1, gh(6, 3),  3);
                  b2  = fac(72, 6, 0);
                  d03 = d03 || (b1[, (2:6) || (8:12) || (14:18) || (20:24)] + 1
                                // b2[, 3:22])[d126[,1],];
                  d06 = d06 || (j(9, 1, 1) @ (1:6)` // b2[,2])[d126[,1],];
                  d21 = d21 || ((1:9)` @ j(6, 1, 1) // b2[,1] + 9)[d126[,1],];
                  end;

               else if index(lineage, '126 ** 1 > 7 ** 1 18 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(18, 1, 1) @ (1:7)`)[d126[,1],];
                  d18 = d18 || ((1:18)` @ j(7, 1, 1))[d126[,1],];
                  end;

               else if index(lineage, '126 ** 1 > 9 ** 1 14 ** 1') then do;
                  b1  = 1;
                  d09 = d09 || (j(14, 1, 1) @ (1:9)`)[d126[,1],];
                  d14 = d14 || ((1:14)` @ j(9, 1, 1))[d126[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d126) = 1 then free d126;
                  else do; file log; put 'ERROR: 126 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------125 runs, recode 125-level factors------;
            avail = max(0, ncol(d125) - any125);
            do while(avail);

               if index(lineage, '125 ** 1 > 5 ** 25 25 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (ks((0:4)`, gh(25, 5), 5) + 1)[d125[,1],];
                  d25 = d25 || (j(5, 1, 1) @ (1:25)`)[d125[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d125) = 1 then free d125;
                  else do; file log; put 'ERROR: 125 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------124 runs, recode 124-level factors------;
            avail = max(0, ncol(d124) - any124);
            do while(avail);

               if index(lineage, '124 ** 1 > 2 ** 123') then do;
                  b1  = 1;
                  d02 = d02 || hadw(124, 0)[d124[,1],];
                  end;

               else if index(lineage, '124 ** 1 > 2 ** 22 31 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(124, 22, 2)[d124[,1],];
                  d31 = d31 || ((1:31)` @ j(4, 1, 1))[d124[,1],];
                  end;

               else if index(lineage, '124 ** 1 > 2 ** 2 62 ** 1')
                  then %mktn22(124);

               else if index(lineage, '124 ** 1 > 4 ** 1 31 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(31, 1, 1) @ (1:4)`)[d124[,1],];
                  d31 = d31 || ((1:31)` @ j(4, 1, 1))[d124[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d124) = 1 then free d124;
                  else do; file log; put 'ERROR: 124 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------123 runs, recode 123-level factors------;
            avail = max(0, ncol(d123) - any123);
            do while(avail);

               if index(lineage, '123 ** 1 > 3 ** 1 41 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(41, 1, 1) @ (1:3)`)[d123[,1],];
                  d41 = d41 || ((1:41)` @ j(3, 1, 1))[d123[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d123) = 1 then free d123;
                  else do; file log; put 'ERROR: 123 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------122 runs, recode 122-level factors------;
            avail = max(0, ncol(d122) - any122);
            do while(avail);

               if index(lineage, '122 ** 1 > 2 ** 1 61 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(61, 1, 1) @ (1:2)`)[d122[,1],];
                  d61 = d61 || ((1:61)` @ j(2, 1, 1))[d122[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d122) = 1 then free d122;
                  else do; file log; put 'ERROR: 122 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------121 runs, recode 121-level factors------;
            avail = max(0, ncol(d121) - any121);
            do while(avail);

               if index(lineage, '121 ** 1 > 11 ** 12') then do;
                  b1  = 1;
                  d11 = d11 || ((ks((0:10)`, gh(11, 11), 11) + 1) ||
                               j(11, 1, 1) @ (1:11)`)[d121[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d121) = 1 then free d121;
                  else do; file log; put 'ERROR: 121 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------120 runs, recode 120-level factors------;
            avail = max(0, ncol(d120) - any120);
            do while(avail);

               if index(lineage, '120 ** 1 > 2 ** 116 4 ** 1') then do;
                  b1  = twolev(h2 @ sortmat(hadp1(60, 1), 1:2));
                  d02 = d02 || b1[d120[,1], (3:60) || (63:120)];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 87 3 ** 1 4 ** 1')
                  then do;
                  b1  = fac(60, 30, 2) || (1:3)` @ j(20, 1, 1);
                  b1  = j(2, 1, 1) @ sortmat(b1, 1:2);
                  b2  = twolev({1 -1}` @ sortmat(hadp1(60, 1), 1:2));
                  d02 = d02 || (b2[,3:60] || b1[,2:30])[d120[,1],];
                  d03 = d03 || b1[d120[,1],31];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 79 4 ** 1 5 ** 1')
                  then do;
                  b1  = twolev((h2 @ had(12, 1))[,2:24]);
                  b1  = (b1[,1:22] // fac(36, 27, 2)[,1:22]) ||
                        (b1[,23] // fac(36, 1, 3) + 2);
                  b1  = j(2, 1, 1) @ sortmat(b1, 1:2);
                  b2  = twolev({1 -1}` @ sortmat(hadp1(60, 1), 1:2));
                  d02 = d02 || (b2[,3:60] || b1[,2:22])[d120[,1],];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  d05 = d05 || b1[d120[,1],23];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 75 4 ** 1 6 ** 1')
                  then do;
                  b1  = had(20, 0);
                  b2  = sortmat(b1, 1);
                  b1  = (b1 // b2 // reverse(b2))[,2:19] ||
                        (b1[,1] // (3:6)` @ j(10, 1, 1));
                  b1  = j(2, 1, 1) @ sortmat(b1, 1:2);
                  b2  = twolev({1 -1}` @ sortmat(hadp1(60, 1), 1:2));
                  d02 = d02 || (b2[,3:60] || b1[,2:18])[d120[,1],];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  d06 = d06 || b1[d120[,1],19];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 75 4 ** 1 10 ** 1')
                  then do;
                  b1  = twolev((h2 @ had(12, 1))[,2:24]);
                  b2  = b1[,1] # 2 + b1[,12] - 2;
                  b1  = (b1[,(2:11) || (14:21)] // fac(36, 20, 2)[,1:18]) ||
                        (b2 // fac(36, 1, 6) + 4);
                  b1  = j(2, 1, 1) @ sortmat(b1, 1:2);
                  b2  = twolev({1 -1}` @ sortmat(hadp1(60, 1), 1:2));
                  d02 = d02 || (b2[,3:60] || b1[,2:18])[d120[,1],];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  d10 = d10 || b1[d120[,1],19];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 74 4 ** 1 15 ** 1')
                  then do;
                  b1  = fac(60, 17, 2) || (1:15)` @ j(4, 1, 1);
                  b1  = j(2, 1, 1) @ sortmat(b1, 1:2);
                  b2  = twolev({1 -1}` @ sortmat(hadp1(60, 1), 1:2));
                  d02 = d02 || (b2[,3:60] || b1[,2:17])[d120[,1],];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  d15 = d15 || b1[d120[,1],18];
                  end;

               else if index(lineage,
                  '120 ** 1 > 2 ** 70 3 ** 1 4 ** 1 10 ** 1') then do;
                  b1  = sortmat(had(12, 1), 1:2);
                  b2  = sortmat(fac(12, 4, 2) || (1:3)` @ j(4, 1, 1), 1);
                  b3  = (j(2, 1, 1) @ b2[,5]);
                  b2  = twolev({-1 1}` @ b1) || j(2, 1, 1) @ b2[,1:4];
                  b1  = (b2[,(3:12) || (14:16)] // fac(36, 20, 2)[,1:13]) ||
                        (b3 // fac(36, 2, 3)[,1]) ||
                        ((1:4)` @ j(6, 1, 1) // fac(36, 1, 6) + 4);
                  b1  = j(2, 1, 1) @ sortmat(b1, 1:2);
                  b2  = twolev({1 -1}` @ sortmat(hadp1(60, 1), 1:2));
                  d02 = d02 || (b2[,3:60] || b1[,2:13])[d120[,1],];
                  d03 = d03 || b1[d120[,1],14];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  d10 = d10 || b1[d120[,1],15];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 70 4 ** 1 5 ** 1 6 ** 1')
                  then do;
                  b1  = (twolev({1 -1}` @ had(12, 1)) ||
                         j(2, 1, 1) @ twolev({1 -1}` @ j(3, 1, 1) @ h2)) //
                        fac(36, 20, 2)[,1:14];
                  b1  = b1[,1:13] || (b1[1:24,14] // fac(36, 2, 3)[,1] + 2) ||
                        (j(4, 1, 1) @ (1:6)` // fac(36, 1, 6));
                  b1  = j(2, 1, 1) @ sortmat(b1, 1:2);
                  b2  = twolev({1 -1}` @ sortmat(hadp1(60, 1), 1:2));
                  d02 = d02 || (b2[,3:60] || b1[,2:13])[d120[,1],];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  d05 = d05 || b1[d120[,1],14];
                  d06 = d06 || b1[d120[,1],15];
                  end;

               else if index(lineage,
                  '120 ** 1 > 2 ** 68 4 ** 1 6 ** 1 10 ** 1') then do;
                  b1  = sortmat(had(12, 1), 1:2);
                  b2  = (1:2)` @ j(6, 1, 1) ||
                        (j(3, 1, 1) @ (1:2)` // j(3, 1, 1) @ {2, 1});
                  b2  = (twolev({-1 1}` @ b1) || {1 1}` @ b2)[,(3:12) || 14];
                  b1  = (b2 // fac(36, 13, 2)[,1:11]) ||
                        ((j(4, 1, 1) @ (1:6)`) // fac(36, 2, 6)[,1]) ||
                        ((1:4)` @ j(6, 1, 1) // fac(36, 2, 6)[,2] + 4);
                  b1  = j(2, 1, 1) @ sortmat(b1, 1:2);
                  b2  = twolev({1 -1}` @ sortmat(hadp1(60, 1), 1:2));
                  d02 = d02 || (b2[,3:60] || b1[,2:11])[d120[,1],];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  d06 = d06 || b1[d120[,1],12];
                  d10 = d10 || b1[d120[,1],13];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 60 60 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (twolev({1 -1}` @ hadp1(60, 1)))[d120[,1],];
                  d60 = d60 || (j(2, 1, 1) @ (1:60)`)[d120[,1],];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 59 4 ** 1 30 ** 1')
                  then do;
                  b1  = twolev({1 -1}` @ sortmat(hadp1(60, 1), 1:2));
                  b2  = j(2, 1, 1) @ twolev({1 -1}` @ j(15, 1, 1) @ h2);
                  d02 = d02 || (b1[,3:60] || b2[,2])[d120[,1],];
                  d04 = d04 || ((1:4)` @ j(30, 1, 1))[d120[,1],];
                  d30 = d30 || (j(4, 1, 1) @ (1:30)`)[d120[,1],];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 28 10 ** 1 12 ** 1')
                  then do;
                  b1  = ks((0:3)`, gh(12, 4), 4) + 1;
                  b2  = twolev((h2 @ h2)[,2:4]);
                  b3  = sortmat(fac(36, 1, 3) || fac(36, 27, 2), 1);
                  b3  = b3 // reverse(b3);
                  b4  = sortmat(ks((0:2)`, gh(12, 3), 3) + 1 ||
                                j(3, 1, 1) @ (1:12)`, 1);
                  b4  = b4 // b4;   free b5;
                  do i = 2 to 10; b5 = b5 || b2[b1[,i],]; end;
                  d02 = d02 || ((b5[,1:26] // b3[,2:27]) ||
                                fac(120, 2, 2))[d120[,1],];
                  d10 = d10 || (b1[,1] // (5:10)` @ j(12, 1, 1))[d120[,1],];
                  d12 = d12 || (j(4, 1, 1) @ (1:12)` // b4[,13])[d120[,1],];
                  end;

               else if index(lineage, '120 ** 1 > 2 ** 30 6 ** 1 20 ** 1')
                  then do;
                  b1  = fac(60, 30, 2);
                  d02 = d02 || (b1 // reverse(b1))[d120[,1],];
                  d06 = d06 || ((1:6)` @ j(20, 1, 1))[d120[,1],];
                  d20 = d20 || (j(6, 1, 1) @ (1:20)`)[d120[,1],];
                  end;

               else if index(lineage, '120 ** 1 > 3 ** 1 40 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(40, 1, 1) @ (1:3)`)[d120[,1],];
                  d40 = d40 || ((1:40)` @ j(3, 1, 1))[d120[,1],];
                  end;

               else if index(lineage, '120 ** 1 > 5 ** 1 24 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(24, 1, 1) @ (1:5)`)[d120[,1],];
                  d24 = d24 || ((1:24)` @ j(5, 1, 1))[d120[,1],];
                  end;

               else if index(lineage, '120 ** 1 > 8 ** 1 15 ** 1') then do;
                  b1  = 1;
                  d08 = d08 || (j(15, 1, 1) @ (1:8)`)[d120[,1],];
                  d15 = d15 || ((1:15)` @ j(8, 1, 1))[d120[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d120) = 1 then free d120;
                  else do; file log; put 'ERROR: 120 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------119 runs, recode 119-level factors------;
            avail = max(0, ncol(d119) - any119);
            do while(avail);

               if index(lineage, '119 ** 1 > 7 ** 1 17 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(17, 1, 1) @ (1:7)`)[d119[,1],];
                  d17 = d17 || ((1:17)` @ j(7, 1, 1))[d119[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d119) = 1 then free d119;
                  else do; file log; put 'ERROR: 119 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------118 runs, recode 118-level factors------;
            avail = max(0, ncol(d118) - any118);
            do while(avail);

               if index(lineage, '118 ** 1 > 2 ** 1 59 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(59, 1, 1) @ (1:2)`)[d118[,1],];
                  d59 = d59 || ((1:59)` @ j(2, 1, 1))[d118[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d118) = 1 then free d118;
                  else do; file log; put 'ERROR: 118 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------117 runs, recode 117-level factors------;
            avail = max(0, ncol(d117) - any117);
            do while(avail);

               if index(lineage, '117 ** 1 > 3 ** 13 39 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (ks((0:2)`, gh(39, 3), 3) + 1)[d117[,1],];
                  d39 = d39 || (j(3, 1, 1) @ (1:39)`)[d117[,1],];
                  end;

               else if index(lineage, '117 ** 1 > 9 ** 1 13 ** 1') then do;
                  b1  = 1;
                  d09 = d09 || (j(13, 1, 1) @ (1:9)`)[d117[,1],];
                  d13 = d13 || ((1:13)` @ j(9, 1, 1))[d117[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d117) = 1 then free d117;
                  else do; file log; put 'ERROR: 117 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------116 runs, recode 116-level factors------;
            avail = max(0, ncol(d116) - any116);
            do while(avail);

               if index(lineage, '116 ** 1 > 2 ** 115') then do;
                  b1  = 1;
                  d02 = d02 || hadw(116, 0)[d116[,1],];
                  end;

               else if index(lineage, '116 ** 1 > 2 ** 23 29 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(116, 23, 2)[d116[,1],];
                  d29 = d29 || ((1:29)` @ j(4, 1, 1))[d116[,1],];
                  end;

               else if index(lineage, '116 ** 1 > 2 ** 2 58 ** 1')
                  then %mktn22(116);

               else if index(lineage, '116 ** 1 > 4 ** 1 29 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(29, 1, 1) @ (1:4)`)[d116[,1],];
                  d29 = d29 || ((1:29)` @ j(4, 1, 1))[d116[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d116) = 1 then free d116;
                  else do; file log; put 'ERROR: 116 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------115 runs, recode 115-level factors------;
            avail = max(0, ncol(d115) - any115);
            do while(avail);

               if index(lineage, '115 ** 1 > 5 ** 1 23 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(23, 1, 1) @ (1:5)`)[d115[,1],];
                  d23 = d23 || ((1:23)` @ j(5, 1, 1))[d115[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d115) = 1 then free d115;
                  else do; file log; put 'ERROR: 115 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------114 runs, recode 114-level factors------;
            avail = max(0, ncol(d114) - any114);
            do while(avail);

               if index(lineage, '114 ** 1 > 2 ** 1 57 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(57, 1, 1) @ (1:2)`)[d114[,1],];
                  d57 = d57 || ((1:57)` @ j(2, 1, 1))[d114[,1],];
                  end;

               else if index(lineage, '114 ** 1 > 3 ** 1 38 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(38, 1, 1) @ (1:3)`)[d114[,1],];
                  d38 = d38 || ((1:38)` @ j(3, 1, 1))[d114[,1],];
                  end;

               else if index(lineage, '114 ** 1 > 6 ** 1 19 ** 1') then do;
                  b1  = 1;
                  d06 = d06 || (j(19, 1, 1) @ (1:6)`)[d114[,1],];
                  d19 = d19 || ((1:19)` @ j(6, 1, 1))[d114[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d114) = 1 then free d114;
                  else do; file log; put 'ERROR: 114 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------112 runs, recode 112-level factors------;
            avail = max(0, ncol(d112) - any112);
            do while(avail);

               if index(lineage, '112 ** 1 > 2 ** 104 8 ** 1') then do;
                  b1  = twolev(h2 @ h2 @ sortmat(hadw(28, 1), 1:2));
                  b2  = j(1, 112, 1); b2[{1 2 29 30 57 58 85 86}] = 0;
                  d02 = d02 || (b1[,loc(b2)])[d112[,1],];
                  d08 = d08 || ((1:8)` @ j(14, 1, 1))[d112[,1],];
                  end;

               else if index(lineage, '112 ** 1 > 2 ** 89 7 ** 1 8 ** 1')
                  then do;
                  b1  = sortmat(fac(28, 12, 2) || (1:7)` @ j(4, 1, 1), 1);
                  d02 = d02 || (twolev((h2 @ h2)[,2:4] @
                               sortmat(hadw(28, 1), 1:2)[,3:28]) ||
                               j(4, 1, 1) @ b1[,2:12])[d112[,1],];
                  d07 = d07 || (j(4, 1, 1) @ b1[,13])[d112[,1],];
                  d08 = d08 || ((1:8)` @ j(14, 1, 1))[d112[,1],];
                  end;

               else if index(lineage, '112 ** 1 > 2 ** 79 8 ** 1 14 ** 1')
                  then do;
                  b1  = 1;
                  d02 = d02 || (twolev((h2 @ h2)[,2:4] @
                               sortmat(hadw(28, 1), 1:2)[,3:28]) ||
                               j(4, 1, 1) @ twolev({1 -1}` @
                               j(7, 1, 1) @ {1 -1}`))[d112[,1],];
                  d14 = d14 || (j(8, 1, 1) @ (1:14)`)[d112[,1],];
                  d08 = d08 || ((1:8)` @ j(14, 1, 1))[d112[,1],];
                  end;

               else if index(lineage, '112 ** 1 > 2 ** 75 4 ** 3 28 ** 1')
                  then do;
                  b1  = hadw(28, 1);
                  b2  = j(84, 1, 1); b2[{1 3 4 29 30 32 57 58 59}] = 0;
                  d02 = d02 || (twolev((h2 @ h2)[,2:4] @ b1)[,loc(b2)])
                               [d112[,1],];
                  d04 = d04 || (((({1 3 -1 -3}` @ b1[,2] || {3 -3 -1 1}` @
                               b1[,3] || {3 -1 1 -3}` @ b1[,4]) + 5) / 2))
                               [d112[,1],];
                  d28 = d28 || (j(4, 1, 1) @ (1:28)`)[d112[,1],];
                  end;

               else if index(lineage, '112 ** 1 > 2 ** 56 56 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (twolev({1 -1}` @ h2 @ hadw(28, 1)))[d112[,1],];
                  d56 = d56 || (j(2, 1, 1) @ (1:56)`)[d112[,1],];
                  end;

               else if index(lineage, '112 ** 1 > 4 ** 12 28 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (ks((0:3)`, gh(28, 4), 4) + 1)[d112[,1],];
                  d28 = d28 || (j(4, 1, 1) @ (1:28)`)[d112[,1],];
                  end;

               else if index(lineage, '112 ** 1 > 7 ** 1 16 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(16, 1, 1) @ (1:7)`)[d112[,1],];
                  d16 = d16 || ((1:16)` @ j(7, 1, 1))[d112[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d112) = 1 then free d112;
                  else do; file log; put 'ERROR: 112 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------111 runs, recode 111-level factors------;
            avail = max(0, ncol(d111) - any111);
            do while(avail);

               if index(lineage, '111 ** 1 > 3 ** 1 37 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(37, 1, 1) @ (1:3)`)[d111[,1],];
                  d37 = d37 || ((1:37)` @ j(3, 1, 1))[d111[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d111) = 1 then free d111;
                  else do; file log; put 'ERROR: 111 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------110 runs, recode 110-level factors------;
            avail = max(0, ncol(d110) - any110);
            do while(avail);

               if index(lineage, '110 ** 1 > 2 ** 1 55 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(55, 1, 1) @ (1:2)`)[d110[,1],];
                  d55 = d55 || ((1:55)` @ j(2, 1, 1))[d110[,1],];
                  end;

               else if index(lineage, '110 ** 1 > 5 ** 1 22 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(22, 1, 1) @ (1:5)`)[d110[,1],];
                  d22 = d22 || ((1:22)` @ j(5, 1, 1))[d110[,1],];
                  end;

               else if index(lineage, '110 ** 1 > 10 ** 1 11 ** 1') then do;
                  b1  = 1;
                  d10 = d10 || (j(11, 1, 1) @ (1:10)`)[d110[,1],];
                  d11 = d11 || ((1:11)` @ j(10, 1, 1))[d110[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d110) = 1 then free d110;
                  else do; file log; put 'ERROR: 110 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------108 runs, recode 108-level factors------;
            avail = max(0, ncol(d108) - any108);
            do while(avail);

               if index(lineage, '108 ** 1 > 2 ** 107') then do;
                  b1  = 1;
                  d02 = d02 || hadp1(108, 0)[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 34 3 ** 29 6 ** 1')
                  then do;
                  b1  = ks((0:2)`, sortmat(gh(36, 3), 1:3), 3) + 1;
                  b2  = j(3, 1, 1) @ sortmat(twolev(hadw(36, 1)), 1:2);
                  b3  = (1:36)`; b3[{1 2 3 5 6 7 8}] = 0;
                  d02 = d02 || b2[d108[,1],3:36];
                  d03 = d03 || b1[d108[,1],loc(b3)];
                  d06 = d06 || ((1:6)` @ j(18, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 3 ** 44 9 ** 1 12 ** 1')
                  then do;
                  b1  = ks(gh(3, 3), (0:2)`, 3) || (0:2)` @ j(3, 1, 1);
                  d03 = d03 || (ks(gh(12, 3)[,2:12],
                                   b1, 3) + 1)[d108[,1],];
                  d09 = d09 || (j(12, 1, 1) @ (1:9)`)[d108[,1],];
                  d12 = d12 || ((1:12)` @ j(9, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 3 ** 36 36 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (ks((0:2)`, gh(36, 3), 3) + 1)[d108[,1],];
                  d36 = d36 || (j(3, 1, 1) @ (1:36)`)[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 40 6 ** 1') then do;
                  b1  = twolev(hadw(36, 1));
                  d02 = d02 || (j(3, 1, 1) @ b1[,3:35] ||
                                fac(108, 7, 2))[d108[,1],];
                  d06 = d06 || (b1[,2] // b1[,2] + 2 // b1[,2] + 4)[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 27 3 ** 33 9 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 1, 3) || fac(36, 27, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],2:28];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)])[d108[,1],];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 22 27 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(108, 22, 2)[d108[,1],];
                  d27 = d27 || ((1:27)` @ j(4, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 20 3 ** 34 9 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 2, 3) || fac(36, 20, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],3:22];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)] ||
                                b1[,2])[d108[,1],];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 18 3 ** 33 6 ** 1 9 ** 1') then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 2, 3) || fac(36, 1, 6) || fac(36, 20, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],4:21];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)])[d108[,1],];
                  d06 = d06 || b1[d108[,1],3];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 18 3 ** 31 18 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 1, 6) || fac(36, 2, 3) || fac(36, 20, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],4:21];
                  d03 = d03 || (b2[,loc(b3)] || b1[,2])[d108[,1],];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 21 3 ** 1 6 ** 2')
                  then do;
                  b1  = fac(36, 20, 2)[,1:18];
                  b2  = b1[,1];
                  b3  = j(3, 1, 1);
                  d02 = d02 || (b3 @ b1[,2:17] || fac(108, 5, 2))[d108[,1],];
                  d03 = d03 || (b3 @ fac(36, 2, 3)[,1])[d108[,1],];
                  d06 = d06 || (b3 @ fac(36, 1, 6) ||
                               (b2 // b2 + 2 // b2 + 4))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 17 3 ** 29 6 ** 2')
                  then do;
                  b1  = ks((0:2)`, sortmat(gh(36, 3), 1:3), 3) + 1;
                  b2  = sortmat(fac(36, 20, 2)[,1:18] ||
                                fac(36, 2, 3)[,1] || fac(36, 1, 6), 1:2);
                  b3  = (1:36)`; b3[{1 2 3 5 6 7 8}] = 0;

                  d02 = d02 || (j(3, 1, 1) @ b2[,2:18])[d108[,1],];
                  d03 = d03 || b1[d108[,1], loc(b3)];
                  d06 = d06 || ((1:6)` @ j(18, 1, 1) ||
                                j(3, 1, 1) @ b2[,20])[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 15 6 ** 1 18 ** 1')
                  then do;
                  b1  = fac(48, 33, 0);
                  b2  = fac(60, 17, 0);
                  d02 = d02 || (b1[, 1:15] // b2[, 1:15])[d108[,1],];
                  d06 = d06 || (b1[, 32] // b2[, 16])[d108[,1],];
                  d18 = d18 || (b1[, 33] // b2[, 17] + 8)[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 13 3 ** 30 6 ** 1 18 ** 1') then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 2, 6) || fac(36, 13, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],3:15];
                  d03 = d03 || b2[d108[,1],loc(b3)];
                  d06 = d06 || b1[d108[,1],2];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 13 6 ** 3') then do;
                  b1  = fac(36, 13, 2);
                  b2  = b1[,1];
                  b3  = j(3, 1, 1);
                  d02 = d02 || (b3 @ b1[,2:13] || fac(108, 1, 2))[d108[,1],];
                  d06 = d06 || (b3 @ fac(36, 2, 6) ||
                               (b2 // b2 + 2 // b2 + 4))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 12 3 ** 29 6 ** 3')
                  then do;
                  b1  = ks((0:2)`, sortmat(gh(36, 3), 1:3), 3) + 1;
                  b2  = sortmat(fac(36, 13, 2) || fac(36, 2, 6), 1:2);
                  b3  = (1:36)`; b3[{1 2 3 5 6 7 8}] = 0;
                  d02 = d02 || (j(3, 1, 1) @ b2[,2:13])[d108[,1],];
                  d03 = d03 || b1[d108[,1], loc(b3)];
                  d06 = d06 || ((1:6)` @ j(18, 1, 1) ||
                                j(3, 1, 1) @ b2[,14:15])[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 10 3 ** 40 6 ** 1 9 ** 1') then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 8, 3) || fac(36, 1.2, 6) || fac(36, 10.1, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],10:19];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)] ||
                                b1[,2:8])[d108[,1],];
                  d06 = d06 || b1[d108[,1],9];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 10 3 ** 33 6 ** 2 9 ** 1') then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 1.2, 3) || fac(36, 2.3, 6) || fac(36, 10, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],4:13];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)])[d108[,1],];
                  d06 = d06 || b1[d108[,1],2:3];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 10 3 ** 31 6 ** 1 18 ** 1') then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 2.3, 6) || fac(36, 1.2, 3) || fac(36, 10, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],4:13];
                  d03 = d03 || (b2[,loc(b3)] || b1[,3])[d108[,1],];
                  d06 = d06 || b1[d108[,1],2];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 9 3 ** 36 6 ** 2 9 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 4, 3) || fac(36, 2.2, 6) || fac(36, 9, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],7:15];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)] ||
                                b1[,2:4])[d108[,1],];
                  d06 = d06 || b1[d108[,1],5:6];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 9 3 ** 34 6 ** 1 18 ** 1') then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 2.2, 6) || fac(36, 4, 3) || fac(36, 9, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],7:15];
                  d03 = d03 || (b2[,loc(b3)] || b1[,3:6])[d108[,1],];
                  d06 = d06 || b1[d108[,1],2];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 8 3 ** 30 6 ** 2 18 ** 1') then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 3.4, 6) || fac(36, 8, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],4:11];
                  d03 = d03 || b2[d108[,1],loc(b3)];
                  d06 = d06 || b1[d108[,1],2:3];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 4 3 ** 33 6 ** 3 9 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 1.1, 3) || fac(36, 3.3, 6) || fac(36, 4, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],5:8];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)])[d108[,1],];
                  d06 = d06 || b1[d108[,1],2:4];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 4 3 ** 31 6 ** 2 18 ** 1') then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 3.3, 6) || fac(36, 1.1, 3) || fac(36, 4, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],5:8];
                  d03 = d03 || (b2[,loc(b3)] || b1[,4])[d108[,1],];
                  d06 = d06 || b1[d108[,1],2:3];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 3 3 ** 41 6 ** 1 9 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 9, 3) || fac(36, 1.1, 6) || fac(36, 3.1, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],11:13];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)] ||
                                b1[,2:9])[d108[,1],];
                  d06 = d06 || b1[d108[,1],10];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 3 3 ** 39 18 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 1.1, 6) || fac(36, 9, 3) || fac(36, 3.1, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],11:13];
                  d03 = d03 || (b2[,loc(b3)] || b1[,2:10])[d108[,1],];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 3 3 ** 34 6 ** 3 9 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 2.1, 3) || fac(36, 3.1, 6) || fac(36, 3, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],6:8];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)] ||
                                b1[,2])[d108[,1],];
                  d06 = d06 || b1[d108[,1],3:5];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 3 3 ** 32 6 ** 2 18 ** 1') then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 3.1, 6) || fac(36, 2.1, 3) || fac(36, 3, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],6:8];
                  d03 = d03 || (b2[,loc(b3)] || b1[,4:5])[d108[,1],];
                  d06 = d06 || b1[d108[,1],2:3];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 3 3 ** 16 6 ** 8') then do;
                  run proj108(m1,m2,m3,m4,s1,s2,s3,s4,o3);
                  l36 = fac(36, 15, 0);
                  l36 = l36 || mod(l36[,3], 2);
                  b1  = ((m1 @ s1) * (l36[,{1 2 4 5 6 7 8}] @ o3) ||
                         (m2 @ s2) * (l36[,{1 2 4 5 6 7 8}] @ o3) ||
                         (m3 @ s3) * (l36[,{1 2 4 5 6 7 8}] @ o3) ||
                         (m4 @ s4) * (l36[,{1 2 4 5 6 7  }] @ o3)) + 1;
                  d02 = d02 || b1[d108[,1],{7 14 21}];
                  d03 = d03 || b1[d108[,1],{3 4 5 6 10 11 12 13
                                            17 18 19 20 24 25 26 27}];
                  d06 = d06 || b1[d108[,1], {1 2 8 9 15 16 22 23}];
                  free m1 m2 m3 m4 s1 s2 s3 s4 l36 o3;
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 2 3 ** 42 18 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = j(6, 1, 1) @ (1:6)` || ks((0:2)`, gh(12, 3), 3) + 1 ||
                        j(3, 1, 1) @ twolev({1 -1}` @ j(3, 1, 1) @ h2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],14:15];
                  d03 = d03 || (b2[,loc(b3)] || b1[,2:13])[d108[,1],];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 2 3 ** 37 6 ** 2 9 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 5, 3) || fac(36, 2.1, 6) || fac(36, 2, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],8:9];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)] ||
                                b1[,2:5])[d108[,1],];
                  d06 = d06 || b1[d108[,1],6:7];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 2 3 ** 35 6 ** 1 18 ** 1') then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 2.1, 6) || fac(36, 5, 3) || fac(36, 2, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],8:9];
                  d03 = d03 || (b2[,loc(b3)] || b1[,3:7])[d108[,1],];
                  d06 = d06 || b1[d108[,1],2];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 2 ** 2 54 ** 1')
                  then %mktn22(108);

               else if index(lineage, '108 ** 1 > 2 ** 1 3 ** 35 6 ** 3 9 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 3, 3) || fac(36, 3, 6) || fac(36, 1, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d02 = d02 || b1[d108[,1],7];
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)] ||
                                b1[,2:3])[d108[,1],];
                  d06 = d06 || b1[d108[,1],4:6];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage,
                  '108 ** 1 > 2 ** 1 3 ** 33 6 ** 2 18 ** 1') then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 3, 6) || fac(36, 3, 3) || fac(36, 1, 2);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d02 = d02 || b1[d108[,1],7];
                  d03 = d03 || (b2[,loc(b3)] || b1[,4:6])[d108[,1],];
                  d06 = d06 || b1[d108[,1],2:3];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 3 ** 39 6 ** 3 9 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36, 3), 3) + 1;
                  b1  = fac(36, 7, 3) || fac(36, 3.2, 6);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  d03 = d03 || (b2[,(2:12) || (14:24) || (26:36)] ||
                                b1[,2:7])[d108[,1],];
                  d06 = d06 || b1[d108[,1],8:10];
                  d09 = d09 || ((1:9)` @ j(12, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 3 ** 37 6 ** 2 18 ** 1')
                  then do;
                  b2  = ks((0:2)`, gh(36.6, 3), 3) + 1;
                  b1  = fac(36, 3.2, 6) || fac(36, 7, 3);
                  b1  = j(3, 1, 1) @ sortmat(b1, 1);
                  b3  = j(1, 36, 1); b3[{1 7 13 19 25 31}] = 0;
                  d03 = d03 || (b2[,loc(b3)] || b1[,4:10])[d108[,1],];
                  d06 = d06 || b1[d108[,1],2:3];
                  d18 = d18 || ((1:18)` @ j(6, 1, 1))[d108[,1],];
                  end;

               else if index(lineage, '108 ** 1 > 3 ** 4 6 ** 11') then do;
                  run proj108(m1,m2,m3,m4,s1,s2,s3,s4,o3);
                  l36 = fac(36, 15, 0);
                  b1  = ((m1 @ s1) * (l36[,{1 2 3}] @ o3) ||
                         (m2 @ s2) * (l36[,{1 2 3}] @ o3) ||
                         (m3 @ s3) * (l36[,{1 2 3}] @ o3) ||
                         (m4 @ s4) * (l36[,{1 2 4 5 6 7}] @ o3)) + 1;
                  d03 = d03 || b1[d108[,1],12:15];
                  d06 = d06 || b1[d108[,1], 1:11];
                  free m1 m2 m3 m4 s1 s2 s3 s4 l36 o3;
                  end;

               else if index(lineage, '108 ** 1 > 4 ** 1 27 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(27, 1, 1) @ (1:4)`)[d108[,1],];
                  d27 = d27 || ((1:27)` @ j(4, 1, 1))[d108[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d108) = 1 then free d108;
                  else do; file log; put 'ERROR: 108 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------106 runs, recode 106-level factors------;
            avail = max(0, ncol(d106) - any106);
            do while(avail);

               if index(lineage, '106 ** 1 > 2 ** 1 53 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(53, 1, 1) @ (1:2)`)[d106[,1],];
                  d53 = d53 || ((1:53)` @ j(2, 1, 1))[d106[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d106) = 1 then free d106;
                  else do; file log; put 'ERROR: 106 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------105 runs, recode 105-level factors------;
            avail = max(0, ncol(d105) - any105);
            do while(avail);

               if index(lineage, '105 ** 1 > 3 ** 1 35 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(35, 1, 1) @ (1:3)`)[d105[,1],];
                  d35 = d35 || ((1:35)` @ j(3, 1, 1))[d105[,1],];
                  end;

               else if index(lineage, '105 ** 1 > 5 ** 1 21 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(21, 1, 1) @ (1:5)`)[d105[,1],];
                  d21 = d21 || ((1:21)` @ j(5, 1, 1))[d105[,1],];
                  end;

               else if index(lineage, '105 ** 1 > 7 ** 1 15 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(15, 1, 1) @ (1:7)`)[d105[,1],];
                  d15 = d15 || ((1:15)` @ j(7, 1, 1))[d105[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d105) = 1 then free d105;
                  else do; file log; put 'ERROR: 105 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------104 runs, recode 104-level factors------;
            avail = max(0, ncol(d104) - any104);
            do while(avail);

               if index(lineage, '104 ** 1 > 2 ** 100 4 ** 1') then do;
                  b1  = twolev(h2 @ sortmat(hadw(52, 1), 1:2));
                  d02 = d02 || b1[d104[,1], (3:52) || (55:104)];
                  d04 = d04 || ((1:4)` @ j(26, 1, 1))[d104[,1],];
                  end;

               else if index(lineage, '104 ** 1 > 2 ** 65 4 ** 1 13 ** 1')
                  then do;
                  b1  = fac(52, 16, 2) || (1:13)` @ j(4, 1, 1);
                  b1  = j(2, 1, 1) @ sortmat(b1, 1);
                  b2  = twolev({1 -1}` @ sortmat(hadw(52, 1), 1:2));
                  d02 = d02 || (b2[,3:52] || b1[,2:16])[d104[,1],];
                  d04 = d04 || ((1:4)` @ j(26, 1, 1))[d104[,1],];
                  d13 = d13 || b1[d104[,1],17];
                  end;

               else if index(lineage, '104 ** 1 > 2 ** 52 52 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || twolev({1 -1}` @ hadw(52, 1))[d104[,1],];
                  d52 = d52 || (j(2, 1, 1) @ (1:52)`)[d104[,1],];
                  end;

               else if index(lineage, '104 ** 1 > 2 ** 51 4 ** 1 26 ** 1')
                  then do;
                  b1  = j(2, 1, 1) @ twolev({1 -1}` @ j(13, 1, 1) @ h2);
                  b2  = twolev({1 -1}` @ sortmat(hadw(52, 1), 1:2));
                  d02 = d02 || (b2[,3:52] || b1[,2])[d104[,1],];
                  d04 = d04 || ((1:4)` @ j(26, 1, 1))[d104[,1],];
                  d26 = d26 || (j(4, 1, 1) @ (1:26)`)[d104[,1],];
                  end;

               else if index(lineage, '104 ** 1 > 8 ** 1 13 ** 1') then do;
                  b1  = 1;
                  d08 = d08 || ((1:8)` @ j(13, 1, 1))[d104[,1],];
                  d13 = d13 || (j(8, 1, 1) @ (1:13)`)[d104[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d104) = 1 then free d104;
                  else do; file log; put 'ERROR: 104 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------102 runs, recode 102-level factors------;
            avail = max(0, ncol(d102) - any102);
            do while(avail);

               if index(lineage, '102 ** 1 > 2 ** 1 51 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(51, 1, 1) @ (1:2)`)[d102[,1],];
                  d51 = d51 || ((1:51)` @ j(2, 1, 1))[d102[,1],];
                  end;

               else if index(lineage, '102 ** 1 > 3 ** 1 34 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(34, 1, 1) @ (1:3)`)[d102[,1],];
                  d34 = d34 || ((1:34)` @ j(3, 1, 1))[d102[,1],];
                  end;

               else if index(lineage, '102 ** 1 > 6 ** 1 17 ** 1') then do;
                  b1  = 1;
                  d06 = d06 || (j(17, 1, 1) @ (1:6)`)[d102[,1],];
                  d17 = d17 || ((1:17)` @ j(6, 1, 1))[d102[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d102) = 1 then free d102;
                  else do; file log; put 'ERROR: 102 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------100 runs, recode 100-level factors------;
            avail = max(0, ncol(d100) - any100);
            do while(avail);

               if index(lineage, '100 ** 1 > 2 ** 99') then do;
                  b1  = 1;
                  d02 = d02 ||  hadw(100, 0)[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 51 5 ** 3') then do;
                  run proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
                  d02 = d02 || (h20[s1,3:19] || h20[s2,3:19] ||
                                h20[s3,3:19])[d100[,1],];
                  d05 = d05 || (b2[s4,] || b2[s5,] || b2[s6,])[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 40 5 ** 4') then do;
                  run proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
                  b1  = h20[s1,3:19] || h20[s2,3:19] || l20[s3,3:9];
                  d05 = d05 || (b2[s4,] || b2[s5,] || b2[s6,] ||
                                b1[,41])[d100[,1],];
                  d02 = d02 || b1[d100[,1],1:40];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 34 5 ** 3 10 ** 1')
                  then do;
                  run proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
                  d02 = d02 || (h20[s1,3:19] || h20[s2,3:19])[d100[,1],];
                  d05 = d05 || (b2[s4,] || b2[s5,] || b2[s6,])[d100[,1],];
                  d10 = d10 || ((b2[s3,] - 1) # 2 + h20[s2,1])[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 29 5 ** 5') then do;
                  run proj100(b1,b2,b3,b4,h2,h20,l20,l25,l4,s1,s2,s3,s4,s5,s6);
                  b1  = h20[s1,3:19] || l20[s2,3:9] || l20[s3,3:9];
                  d05 = d05 || (b2[s4,] || b2[s5,] || b2[s6,] ||
                                b1[,{24 31}])[d100[,1],];
                  d02 = d02 || b1[d100[,1],(1:23) || (25:30)];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 22 25 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(100, 22, 2)[d100[,1],];
                  d25 = d25 || ((1:25)` @ j(4, 1, 1))[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 18 5 ** 9 10 ** 1')
                  then do;
                  b1  = j(2, 1, 1) @ sortmat(j(5, 1, 1) @ (0:9)` ||
                        ks((0:4)`, gh(10, 5), 5) + 1, 1:2);
                  b2  = sortmat(had(20, 0), 1) @ j(5, 1, 1);
                  d02 = d02 || b2[d100[,1],2:19];
                  d05 = d05 || b1[d100[,1],3:11];
                  d10 = d10 || ((b1[,2] - 1) # 2 + b2[,1])[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 16 5 ** 3 10 ** 3')
                  then do;
                  b1  = 1;
                  d02 = d02 || fac(100, 16, 2)[d100[,1],];
                  d05 = d05 || fac(100, 4, 5)[d100[,1],1:3];
                  d10 = d10 || fac(100, 3, 10)[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 7 5 ** 10 10 ** 1')
                  then do;
                  b1  = j(2, 1, 1) @ sortmat(j(5, 1, 1) @ (0:9)` ||
                        ks((0:4)`, gh(10, 5), 5) + 1, 1:2);
                  b2  = sortmat(fac(20, 8, 2) ||
                        {5 2 4 3 1}` @ j(4, 1, 1), 1) @ j(5, 1, 1);
                  d02 = d02 || b2[d100[,1],2:8];
                  d05 = d05 || (b1[,3:11] || b2[,9])[d100[,1],];
                  d10 = d10 || ((b1[,2] - 1) # 2 + b2[,1])[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 5 5 ** 4 10 ** 3')
                  then do;
                  b1  = 1;
                  d02 = d02 || fac(100, 16, 2)[d100[,1],1:5];
                  d05 = d05 || fac(100, 4, 5)[d100[,1],];
                  d10 = d10 || fac(100, 3, 10)[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 4 10 ** 4') then do;
                  b1  = j(10, 1, 1);
                  d10 = d10 || (((0:9)` @ b1 || b1 @ (0:9)` ||
                               fac(100, 4, 10)) + 1)[d100[,1],];
                  d02 = d02 || fac(100, 4, 2)[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 2 ** 2 50 ** 1')
                  then %mktn22(100);

               else if index(lineage, '100 ** 1 > 4 ** 1 25 ** 1') then do;
                  b1  = j(25, 1, 1) @ (1:4)`;
                  b2  = (1:25)` @ j(4, 1, 1);
                  d04 = d04 || b1[d100[,1],];
                  d25 = d25 || b2[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 5 ** 20 20 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (ks((0:4)`, gh(20, 5), 5) + 1)[d100[,1],];
                  d20 = d20 || (j(5, 1, 1) @ (1:20)`)[d100[,1],];
                  end;

               else if index(lineage, '100 ** 1 > 5 ** 8 10 ** 3') then do;
                  b1  = gh(10, 5);
                  b1[6:10,2] = b1[6:10,2] + 5;
                  d05 = d05 || (ks(((0:4) ||
                                    (0:4))`, b1[,3:10],  5) + 1)[d100[,1],];
                  d10 = d10 || (ks((0:9)`, b1[,1:2], 10) + 1 ||
                                j(10, 1, 1) @ (1:10)`)[d100[,1],];
                  end;

               free l4 l25 l20 h20 s1 s2 s3 s4 s5 s6 i4 i25 b4;

               if ncol(b1) then do;
                  if ncol(d100) = 1 then free d100;
                  else do; file log; put 'ERROR: 100 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------99 runs, recode 99-level factors------;
            avail = max(0, ncol(d99) - any99);
            do while(avail);

               if index(lineage, '99 ** 1 > 3 ** 13 33 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (ks((0:2)`, gh(33, 3), 3) + 1)[d99[,1],];
                  d33 = d33 || (j(3, 1, 1) @ (1:33)`)[d99[,1],];
                  end;

               else if index(lineage, '99 ** 1 > 9 ** 1 11 ** 1') then do;
                  b1  = 1;
                  d09 = d09 || ((1:9)` @ j(11, 1, 1))[d99[,1],];
                  d11 = d11 || (j(9, 1, 1) @ (1:11)`)[d99[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d99) = 1 then free d99;
                  else do; file log; put 'ERROR: 99 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------98 runs, recode 98-level factors------;
            avail = max(0, ncol(d98) - any98);
            do while(avail);

               if index(lineage, '98 ** 1 > 2 ** 1 49 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || ((1:2)` @ j(49, 1, 1))[d98[,1],];
                  d49 = d49 || (j(2, 1, 1) @ (1:49)`)[d98[,1],];
                  end;

               else if index(lineage, '98 ** 1 > 7 ** 14 14 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (ks((0:6)`, gh(14, 7), 7) + 1)[d98[,1],];
                  d14 = d14 || (j(7, 1, 1) @ (1:14)`)[d98[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d98) = 1 then free d98;
                  else do; file log; put 'ERROR: 98 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------96 runs, recode 96-level factors------;
            avail = max(0, ncol(d96) - any96);
            do while(avail);

               if index(lineage, '96 ** 1 > 2 ** 80 16 ** 1') then do;
                  b1  = h2 @ had(12, 1);
                  b2  = j(96, 1, 1);
                  b2[{1 2 13 14 25 26 37 38 49 50 61 62 73 74 85 86}] = 0;
                  b3  = ks(gh(4, 4), (b1[,2] # 2 + b1[,13] + 3) / 2, 4);
                  d02 = d02 || twolev((h2 @ h2 @ b1)[,loc(b2)])[d96[,1],];
                  d16 = d16 || (b3[,1] # 4 + b3[,2] + 1)[d96[,1],];
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 73 3 ** 1 16 ** 1')
                  then do;
                  b1  = sortmat(had(12, 1), 1:2);
                  b2  = h2 @ b1;
                  b3  = ks(gh(4, 4), (b2[,2] # 2 + b2[,13]  + 3) / 2, 4);
                  d16 = d16 || (b3[,1] # 4 + b3[,2] + 1)[d96[,1],];
                  b3  = j(72, 1, 1);b3[{1 2 13 14 25 26 37 38 49 50 61 62}] = 0;
                  b3  = twolev((h2 @ h2)[,2:4] @ b2)[,loc(b3)];
                  b2  = sortmat(fac(12, 4, 2) || (1:3)` @ j(4, 1, 1), 1);
                  d03 = d03 || (j(8, 1, 1) @ b2[,5])[d96[,1],];
                  b2  = twolev({-1 1}` @ b1) || j(2, 1, 1) @ b2[,1:4];
                  b3  = b3 || j(4, 1, 1) @ b2[,(3:12) || (14:16)];
                  d02 = d02 || b3[d96[,1],];
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 71 6 ** 1 16 ** 1')
                  then do;
                  b1  = sortmat(had(12, 1), 1:2);
                  b2  = h2 @ b1;
                  b3  = ks(gh(4, 4), (b2[,2] # 2 + b2[,13]  + 3) / 2, 4);
                  d16 = d16 || (b3[,1] # 4 + b3[,2] + 1)[d96[,1],];
                  b3  = j(72, 1, 1);b3[{1 2 13 14 25 26 37 38 49 50 61 62}] = 0;
                  b3  = twolev((h2 @ h2)[,2:4] @ b2)[,loc(b3)];
                  b2  = (1:2)` @ j(6, 1, 1) ||
                        (j(3, 1, 1) @ (1:2)` // j(3, 1, 1) @ {2, 1});
                  b2  = twolev({-1 1}` @ b1) || j(2, 1, 1) @ b2;
                  d02 = d02 || (b3 || j(4, 1, 1) @ b2[,(3:12) || 14])[d96[,1],];
                  d06 = d06 || (j(16, 1, 1) @ (1:6)`)[d96[,1],];
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 48 48 ** 1') then do;
                  d48 = d48 || (j(2, 1, 1) @ (1:48)`)[d96[,1],];
                  b1  = twolev({1 -1}` @ h2 @ h2 @ had(12, 1));
                  d02 = d02 || b1[d96[,1],];
                  end;

               else if index(lineage,
                  '96 ** 1 > 2 ** 44 4 ** 11 8 ** 1 12 ** 1') then do;
                  b1  = sortmat(had(12, 1), 1:3);
                  b1  = twolev(h2 @ h2 @ b1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(ks((0:3)`, gh(12, 4), 4) + 1 ||
                                j(4, 1, 1) @ (1:12)`, 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d96[,1],(2:12) || (14:24) ||
                                          (26:36) || (38:48)];
                  d04 = d04 || b2[d96[,1],2:12];
                  d08 = d08 || ((1:8)` @ j(12, 1, 1))[d96[,1],];
                  d12 = d12 || b2[d96[,1],13];
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 43 4 ** 15 8 ** 1')
                  then do;
                  b1  = (had(12, 1) = -1);
                  b2  = ks(((h2 @ h2 @ h2)[,{2 6 4 8}] = -1), b1, 2) ||
                        ks(j(8, 1, 0), b1[,2:12], 2);
                  b1  = ks((0:3)` @ j(2, 1, 1), gh(12, 4), 4) ||
                        2 # b2[, 1] + b2[,50] || 2 # b2[,13] + b2[,51] ||
                        2 # b2[,25] + b2[,52] || 2 # b2[,37] + b2[,53];
                  d08 = d08 || (2 # b1[,1]  + b2[,2] + 1)[d96[,1],];
                  b3  = j(59, 1, 1);
                  b3[{1 2 3 13 14 16 25 26 29 37 38 42 50 51 52 53}] = 0;
                  d02 = d02 || b2[d96[,1], loc(b3)] + 1;
                  d04 = d04 || b1[d96[,1], 2:16] + 1;
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 43 4 ** 12 6 ** 1 8 ** 1')
                  then do;
                  b1  = sortmat((had(12, 1) = -1)
                                [,{1 2 3 4 7 5 6 8 9 10 11 12}], (1:3));
                  b2  = sortmat(twolev({1 -1}` @ j(3, 1, 1) @ h2) ||
                                j(2, 1, 1) @ (1:6)`, (1:2));
                  d06 = d06 || (j(8, 1, 1) @ b2[,3])[d96[,1],];
                  b2  = b2[,(1:2)];
                  b2  = ks(((h2 @ h2 @ h2)[,{2 6 4 8}] = -1), b1, 2) ||
                        ks(j(8, 1, 0), b2, 2);
                  b1  = ks((0:3)` @ j(2, 1, 1), gh(12, 4), 4) ||
                        2 # b2[, 1] + b2[,50];
                  d08 = d08 || (2 # b1[,1]  + b2[,2] + 1)[d96[,1],];
                  b3  = j(50, 1, 1); b3[{1 2 3 14 26 38 50}] = 0;
                  d02 = d02 || b2[d96[,1], loc(b3)] + 1;
                  d04 = d04 || b1[d96[,1], 2:13] + 1;
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 39 3 ** 1 4 ** 14 8 ** 1')
                  then do;
                  b1  = sortmat((had(12, 1) = -1)
                                [,{1 2 3 4 7 5 6 8 9 10 11 12}], (1:5));
                  b2  = sortmat(fac(12, 4, 2) - 1 ||
                                (0:2)` @ j(4, 1, 1), (1:4));
                  d03 = d03 || (j(8, 1, 1) @ b2[,5] + 1)[d96[,1],];
                  b2  = b2[,(1:4)];
                  b2  = ks(((h2 @ h2 @ h2)[,{2 6 4 8}] = -1), b1, 2) ||
                        ks(j(8, 1, 0), b2, 2);
                  b1  = ks((0:3)` @ j(2, 1, 1), gh(12, 4), 4) ||
                        2 # b2[, 1] + b2[,50] || 2 # b2[,13] + b2[,51] ||
                        2 # b2[,25] + b2[,52];
                  d08 = d08 || (2 # b1[,1] + b2[,2] + 1)[d96[,1],];
                  b3  = j(52, 1, 1);
                  b3[{1 2 3 13 14 16 25 26 29 38 50 51 52}] = 0;
                  d02 = d02 || b2[d96[,1], loc(b3)] + 1;
                  d04 = d04 || b1[d96[,1], 2:15] + 1;
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 26 4 ** 23')
                  then do;
                  b1 = fac(96, 1, 0);
                  d02 = d02 || b1[d96[,1], {21 24 25} || (28:32) ||
                                           (34:43) || (46:49) || (52:55)];
                  d04 = d04 || b1[d96[,1], (1:20) || (57:59)];
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 19 3 ** 1 4 ** 23')
                  then do;
                  b1 = fac(96, 1, 0);
                  d02 = d02 || b1[d96[,1], {21 24 25} || (28:32) ||
                                           (34:36) || (46:49) || (52:55)];
                  d03 = d03 || b1[d96[,1], 60];
                  d04 = d04 || b1[d96[,1], (1:20) || (57:59)];
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 18 4 ** 22 12 ** 1')
                  then do;
                  b1 = fac(96, 1, 0);
                  d02 = d02 || b1[d96[,1], {21 24 25} || (28:32) || (46:55)];
                  d04 = d04 || b1[d96[,1], (1:20) || {58 59}];
                  d12 = d12 || b1[d96[,1], 61];
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 17 4 ** 23 6 ** 1')
                  then do;
                  b1 = fac(96, 1, 0);
                  d02 = d02 || b1[d96[,1], {21 24 25} || (28:32) ||
                                           34 || (46:49) || (52:55)];
                  d04 = d04 || b1[d96[,1], (1:20) || (57:59)];
                  d06 = d06 || b1[d96[,1], 56];
                  end;

               else if index(lineage, '96 ** 1 > 2 ** 12 4 ** 20 24 ** 1')
                  then do;
                  n2 = i(2)[,1] * i(2)[2,] + i(2)[,2] * i(2)[1,];
                  n3 = i(3)[,1] * i(3)[2,] + i(3)[,2] * i(3)[3,] +
                       i(3)[,3] * i(3)[1,];
                  s0 = i(4); s1 = i(2) @ n2; s2 = n2 @ i(2); s3 = n2 @ n2;
                  m0 = kpq(4,3) @ i(8); m1 = j(96, 96, 0);
                  m2 = j(96, 96, 0);    m3 = j(96, 96, 0);
                  m1[ 1: 8, 1: 8] = i(2) @ s1; m1[ 9:16, 9:16] = i(2) @ s2;
                  m1[17:24,17:24] = i(2) @ s3;
                  m1[25:96,25:96] = kpq(3, 3) @ i(8); m1 = m1 * m0;
                  m2[ 1: 8, 1: 8] = i(2) @ s2; m2[ 9:16, 9:16] = i(2) @ s3;
                  m2[17:24,17:24] = i(2) @ s1; t = j(9, 9, 0);
                  t[1:3,1:3] = i(3); t[4:6,4:6] = n3; t[7:9,7:9] = n3 * n3;
                  m2[25:96,25:96] = t * kpq(3,3) @ i(8); m2 = m2 * m0;
                  m3[ 1: 8, 1: 8] = i(2) @ s3; m3[ 9:16, 9:16] = i(2) @ s1;
                  m3[17:24,17:24] = i(2) @ s2; t = j(9, 9, 0);
                  t[1:3,1:3] = i(3); t[4:6,4:6] = n3 * n3;
                  t[7:9,7:9] = n3; m3[25:96,25:96] = t * kpq(3,3) @ i(8);
                  m3 = m3 * m0; d = gh(8, 4);
                  d = shape({0 0 0 0 0 0 1 3 3 2 1 3 3 2 1 1 2 0 1 3 2 2 1 3 3
                       2 3 2 0 1 3 1 2 1 2 3 0 1 2 0}, 8);
                  z = {1, 1}; t = {-1, 1};
                  l = z @ t @ z @ t @ z || t @ z @ z @ z @ t ||
                      t @ t @ t @ t @ t;
                  l = ((l + 1) / 2) || ks(d, (0:3)`, 4); l = {1, 1, 1} @ l;
                  l = (m0 * l || m1 * l || m2 * l || m3 * l) + 1;
                  d02 = d02 || l[d96[,1],{1 2 3 9 10 11 17 18 19 25 26 27}];
                  d04 = d04 || l[d96[,1],(4:8) || (12:16)||(20:24)||(28:32)];
                  d24 = d24 || ((1:24)` @ j(4, 1, 1))[d96[,1],];
                  free n2 n3 s0 s1 s2 s3 m0 m1 m2 m3 t d z l;
                  end;

               else if index(lineage, '96 ** 1 > 3 ** 1 32 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(32, 1, 1) @ (1:3)`)[d96[,1],];
                  d32 = d32 || ((1:32)` @ j(3, 1, 1))[d96[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d96) = 1 then free d96;
                  else do; file log; put 'ERROR: 96 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------95 runs, recode 95-level factors------;
            avail = max(0, ncol(d95) - any95);
            do while(avail);

               if index(lineage, '95 ** 1 > 5 ** 1 19 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(19, 1, 1) @ (1:5)`)[d95[,1],];
                  d19 = d19 || ((1:19)` @ j(5, 1, 1))[d95[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d95) = 1 then free d95;
                  else do; file log; put 'ERROR: 95 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------94 runs, recode 94-level factors------;
            avail = max(0, ncol(d94) - any94);
            do while(avail);

               if index(lineage, '94 ** 1 > 2 ** 1 47 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(47, 1, 1) @ (1:2)`)[d94[,1],];
                  d47 = d47 || ((1:47)` @ j(2, 1, 1))[d94[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d94) = 1 then free d94;
                  else do; file log; put 'ERROR: 94 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------93 runs, recode 93-level factors------;
            avail = max(0, ncol(d93) - any93);
            do while(avail);

               if index(lineage, '93 ** 1 > 3 ** 1 31 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(31, 1, 1) @ (1:3)`)[d93[,1],];
                  d31 = d31 || ((1:31)` @ j(3, 1, 1))[d93[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d93) = 1 then free d93;
                  else do; file log; put 'ERROR: 93 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------92 runs, recode 92-level factors------;
            avail = max(0, ncol(d92) - any92);
            do while(avail);

               if index(lineage, '92 ** 1 > 2 ** 91') then do;
                  b1  = 1;
                  d02 = d02 || hadw(92, 0)[d92[,1],];
                  end;

               else if index(lineage, '92 ** 1 > 2 ** 2 46 ** 1')
                  then %mktn22(92);

               else if index(lineage, '92 ** 1 > 2 ** 21 23 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(92, 21, 2)[d92[,1],];
                  d23 = d23 || ((1:23)` @ j(4, 1, 1))[d92[,1],];
                  end;

               else if index(lineage, '92 ** 1 > 4 ** 1 23 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(23, 1, 1) @ (1:4)`)[d92[,1],];
                  d23 = d23 || ((1:23)` @ j(4, 1, 1))[d92[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d92) = 1 then free d92;
                  else do; file log; put 'ERROR: 92 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------91 runs, recode 91-level factors------;
            avail = max(0, ncol(d91) - any91);
            do while(avail);

               if index(lineage, '91 ** 1 > 7 ** 1 13 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(13, 1, 1) @ (1:7)`)[d91[,1],];
                  d13 = d13 || ((1:13)` @ j(7, 1, 1))[d91[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d91) = 1 then free d91;
                  else do; file log; put 'ERROR: 91 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------90 runs, recode 90-level factors------;
            avail = max(0, ncol(d90) - any90);
            do while(avail);

               if index(lineage, '90 ** 1 > 2 ** 1 45 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || ((1:2)` @ j(45, 1, 1))[d90[,1],];
                  d45 = d45 || (j(2, 1, 1) @ (1:45)`)[d90[,1],];
                  end;

               else if index(lineage, '90 ** 1 > 3 ** 30 30 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (ks((0:2)`, gh(30, 3), 3) + 1)[d90[,1],];
                  d30 = d30 || (j(3, 1, 1) @ (1:30)`)[d90[,1],];
                  end;

               else if index(lineage, '90 ** 1 > 3 ** 26 6 ** 1 15 ** 1')
                  then do;
                  b1  = ks((0:2)`, sortmat(gh(30, 3), 1:2), 3) + 1;
                  d03 = d03 || b1[d90[,1],5:30];
                  d06 = d06 || (b1[,2] # 2 +
                                j(3, 1, 1) @ (-1:0)` @ j(15, 1, 1))[d90[,1],];
                  d15 = d15 || (j(6, 1, 1) @ (1:15)`)[d90[,1],];
                  end;

               else if index(lineage, '90 ** 1 > 5 ** 1 18 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || ((1:5)` @ j(18, 1, 1))[d90[,1],];
                  d18 = d18 || (j(5, 1, 1) @ (1:18)`)[d90[,1],];
                  end;

               else if index(lineage, '90 ** 1 > 9 ** 1 10 ** 1') then do;
                  b1  = 1;
                  d09 = d09 || ((1:9)` @ j(10, 1, 1))[d90[,1],];
                  d10 = d10 || (j(9, 1, 1) @ (1:10)`)[d90[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d90) = 1 then free d90;
                  else do; file log; put 'ERROR: 90 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------88 runs, recode 88-level factors------;
            avail = max(0, ncol(d88) - any88);
            do while(avail);

               if index(lineage, '88 ** 1 > 2 ** 84 4 ** 1') then do;
                  b1  = twolev(h2 @ sortmat(hadp1(44, 1), 1:2));
                  d02 = d02 || b1[d88[,1], (3:44) || (47:88)];
                  d04 = d04 || ((1:4)` @ j(22, 1, 1))[d88[,1],];
                  end;

               else if index(lineage, '88 ** 1 > 2 ** 56 4 ** 1 11 ** 1')
                  then do;
                  b1  = sortmat(hadp1(44, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(44, 15, 2) || (1:11)` @ j(4, 1, 1), 1);
                  b2  = b2 // b2;
                  d02 = d02 || (b1[,2:43] || b2[,2:15])[d88[,1],];
                  d11 = d11 || b2[d88[,1],16];
                  d04 = d04 || ((1:4)` @ j(22, 1, 1))[d88[,1],];
                  end;

               else if index(lineage, '88 ** 1 > 2 ** 44 44 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (twolev({1 -1}` @ hadp1(44, 1)))[d88[,1],];
                  d44 = d44 || (j(2, 1, 1) @ (1:44)`)[d88[,1],];
                  end;

               else if index(lineage, '88 ** 1 > 2 ** 43 4 ** 1 22 ** 1')
                  then do;
                  b1  = sortmat(hadp1(44, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(twolev({1 -1}` @ j(11, 1, 1) @ h2) ||
                                j(2, 1, 1) @ (1:22)`, 1);
                  b2  = b2 // b2;
                  d02 = d02 || (b1[,2:43] || b2[,2])[d88[,1],];
                  d22 = d22 || b2[d88[,1],3];
                  d04 = d04 || ((1:4)` @ j(22, 1, 1))[d88[,1],];
                  end;

               else if index(lineage, '88 ** 1 > 8 ** 1 11 ** 1') then do;
                  d08 = d08 || ((1:8)` @ j(11, 1, 1))[d88[,1],];
                  d11 = d11 || (j(8, 1, 1) @ (1:11)`)[d88[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d88) = 1 then free d88;
                  else do; file log; put 'ERROR: 88 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------87 runs, recode 87-level factors------;
            avail = max(0, ncol(d87) - any87);
            do while(avail);

               if index(lineage, '87 ** 1 > 3 ** 1 29 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(29, 1, 1) @ (1:3)`)[d87[,1],];
                  d29 = d29 || ((1:29)` @ j(3, 1, 1))[d87[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d87) = 1 then free d87;
                  else do; file log; put 'ERROR: 87 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------86 runs, recode 86-level factors------;
            avail = max(0, ncol(d86) - any86);
            do while(avail);

               if index(lineage, '86 ** 1 > 2 ** 1 43 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(43, 1, 1) @ (1:2)`)[d86[,1],];
                  d43 = d43 || ((1:43)` @ j(2, 1, 1))[d86[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d86) = 1 then free d86;
                  else do; file log; put 'ERROR: 86 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------85 runs, recode 85-level factors------;
            avail = max(0, ncol(d85) - any85);
            do while(avail);

               if index(lineage, '85 ** 1 > 5 ** 1 17 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(17, 1, 1) @ (1:5)`)[d85[,1],];
                  d17 = d17 || ((1:17)` @ j(5, 1, 1))[d85[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d85) = 1 then free d85;
                  else do; file log; put 'ERROR: 85 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------84 runs, recode 84-level factors------;
            avail = max(0, ncol(d84) - any84);
            do while(avail);

               if index(lineage, '84 ** 1 > 2 ** 83') then do;
                  b1  = hadp1(84, 0);
                  d02 = d02 || b1[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 2 ** 33 3 ** 1') then do;
                  b1  = j(3, 1, 1) @ twolev(hadw(28, 1))[,2:27] ||
                        fac(84, 7, 2);
                  d02 = d02 || b1[d84[,1],];
                  d03 = d03 || ((1:3)` @ j(28, 1, 1))[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 2 ** 28 7 ** 1') then do;
                  b1  = twolev(h2 @ sortmat(had(12, 1), 1:2))
                              [,1 || (3:12) || (15:24) || {2 13 14}];
                  d02 = d02 || (cat(fac(36, 27, 2), {1 1}` @ b1[,(3:24)]) ||
                                fac(84, 6, 2))[d84[,1],];
                  d07 = d07 || (fac(36, 1, 3) //
                                (b1[,2] + 3 // b1[,2] + 5))[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 2 ** 27 6 ** 1') then do;
                  b1  = hadw(28, 1);
                  b2  = twolev(h2 @ b1);
                  b1  = twolev(b1);
                  d02 = d02 || ((b2[,3:27] // b1[,2:26]) || fac(84, 2, 2))
                               [d84[,1],];
                  d06 = d06 || (((b2[,2] - 1) # 2 + b2[,29]) //
                                (b1[,28] + 4))[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 2 ** 20 3 ** 1 14 ** 1')
                  then do;
                  b1  = twolev((h2 @ h2)[,2:4] @
                        sortmat(had(12, 1), 2))[,(3:12) || (15:22)] //
                        fac(36, 20, 2)[,1:18];
                  d02 = d02 || (b1[,1:17] || fac(84, 2.1, 2))[d84[,1],];
                  d03 = d03 || (j(16, 1, 1) @ (1:3)` //
                                fac(36, 2, 3)[,1])[d84[,1],];
                  d14 = d14 || ((1:8)` @ j(6, 1, 1) //
                                (fac(36, 1, 6) + 8))[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 2 ** 22 6 ** 1 7 ** 1')
                  then do;
                  b1  = (twolev((h2 @ h2)[,2:4] @
                        sortmat(had(12, 1), 2))[,(3:12) || (15:21)] //
                        fac(36, 20, 2)[,1:17]) || fac(84, 5, 2);
                  d02 = d02 || b1[d84[,1],];
                  d06 = d06 || ((j(8, 1, 1) @ (1:6)`) //
                                fac(36, 1, 6))[d84[,1],];
                  d07 = d07 || ((1:4)` @ j(12, 1, 1) //
                                (fac(36, 2, 3)[,1] + 4))[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 2 ** 20 21 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(84, 20, 2)[d84[,1],];
                  d21 = d21 || ((1:21)` @ j(4, 1, 1))[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 2 ** 14 6 ** 1 14 ** 1')
                  then do;
                  b1  = fac(36, 2, 6);
                  b2  = (twolev((h2 @ h2)[,2:4] @
                        sortmat(had(12, 1), 2))[,(3:12) || (15:17)] //
                        fac(36, 13, 2)) || fac(84, 1, 2);
                  d02 = d02 || b2[d84[,1],];
                  d06 = d06 || (j(8, 1, 1) @ (1:6)` // b1[,1])[d84[,1],];
                  d14 = d14 || ((1:8)` @ j(6, 1, 1) // (b1[,2] + 8))[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 2 ** 2 42 ** 1')
                  then %mktn22(84);

               else if index(lineage, '84 ** 1 > 4 ** 1 21 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(21, 1, 1) @ (1:4)`)[d84[,1],];
                  d21 = d21 || ((1:21)` @ j(4, 1, 1))[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 3 ** 1 28 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(28, 1, 1) @ (1:3)`)[d84[,1],];
                  d28 = d28 || ((1:28)` @ j(3, 1, 1))[d84[,1],];
                  end;

               else if index(lineage, '84 ** 1 > 7 ** 1 12 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(12, 1, 1) @ (1:7)`)[d84[,1],];
                  d12 = d12 || ((1:12)` @ j(7, 1, 1))[d84[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d84) = 1 then free d84;
                  else do; file log; put 'ERROR: 84 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------82 runs, recode 82-level factors------;
            avail = max(0, ncol(d82) - any82);
            do while(avail);

               if index(lineage, '82 ** 1 > 2 ** 1 41 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(41, 1, 1) @ (1:2)`)[d82[,1],];
                  d41 = d41 || ((1:41)` @ j(2, 1, 1))[d82[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d82) = 1 then free d82;
                  else do; file log; put 'ERROR: 82 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------81 runs, recode 81-level factors------;
            avail = max(0, ncol(d81) - any81);
            do while(avail);

               if index(lineage, '81 ** 1 > 3 ** 27 27 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (ks((0:2)`, gh(27, 3), 3) + 1)[d81[,1],];
                  d27 = d27 || (j(3, 1, 1) @ (1:27)`)[d81[,1],];
                  end;

               else if index(lineage, '81 ** 1 > 9 ** 10') then do;
                  b1  = 1;
                  d09 = d09 || ((ks((0:8)`, gh(9, 9), 9) + 1) ||
                               j(9, 1, 1) @ (1:9)`)[d81[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d81) = 1 then free d81;
                  else do; file log; put 'ERROR: 81 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------80 runs, recode 80-level factors------;
            avail = max(0, ncol(d80) - any80);
            do while(avail);

               if index(lineage, '80 ** 1 > 2 ** 72 8 ** 1') then do;
                  b1 = h2 @ h2 @ sortmat(had(20, 1), 1:2);
                  b2 = j(1, 80, 1);
                  b2[{1 2 21 22 41 42 61 62}] = 0;
                  d02 = twolev(b1[d80[,1], loc(b2)]);
                  d08 = ((1:8)` @ j(10, 1, 1))[d80[,1],];
                  end;

               else if index(lineage, '80 ** 1 > 2 ** 55 8 ** 1 10 ** 1')
                  then do;
                  b1  = sortmat(had(20, 1), 1:2);
                  b2  = twolev((h2 @ h2)[,2:4] @ b1) ||
                        j(4, 1, 1) @ ((1:2)` @ j(10, 1, 1) ||
                        (j(5, 1, 1) @ (1:2)` // j(5, 1, 1) @ {2, 1}));
                  d08 = d08 || ({1 3 -1 -3}` @ b1[,2] + b2[,61] + 3)[d80[,1],];
                  d10 = d10 || (j(8, 1, 1) @ (1:10)`)[d80[,1],];
                  b3  = j(62, 1, 1); b3[{1 2 21 22 41 42 61}] = 0;
                  d02 = d02 || b2[d80[,1], loc(b3)];
                  end;

               else if index(lineage, '80 ** 1 > 2 ** 61 5 ** 1 8 ** 1')
                  then do;
                  b2  = ((h2 @ h2)[,{3 2 4}] = -1);
                  b2  = ks(b2, sortmat((had(20, 1) = -1), 1:2), 2);
                  b1  = j(4, 1, 1) @ sortmat(fac(20, 8, 2) - 1 ||
                                             (0:4)` @ j(4, 1, 1), 1:2);
                  d05 = d05 || b1[d80[,1],9] + 1;
                  b2  = b2  || b1[,1:8];
                  d08 = d08 || (b2[,1]#4 + b2[,21]#2 + b2[,2] + 1)[d80[,1],];
                  b1  = j(68, 1, 1); b1[{1 2 21 22 41 42 61}] = 0;
                  d02 = d02 || b2[d80[,1], loc(b1)] + 1;
                  end;

               else if index(lineage, '80 ** 1 > 2 ** 51 4 ** 3 20 ** 1')
                  then do;
                  b1  = sortmat(had(20, 1), 1:4);
                  b2  = j(60, 1, 1); b2[{1 3 4 21 22 24 41 42 43}] = 0;
                  d02 = d02 || twolev((h2 @ h2)[,2:4] @ b1)[d80[,1],loc(b2)];
                  d04 = d04 || ((({1 3 -1 -3}` @ b1[,2] || {3 -3 -1 1}` @
                        b1[,3] || {3 -1 1 -3}` @ b1[,4]) + 5) / 2)[d80[,1],];
                  d20 = d20 || (j(4, 1, 1) @ (1:20)`)[d80[,1],];
                  end;

               else if index(lineage, '80 ** 1 > 2 ** 40 40 ** 1') then do;
                  b1  = twolev({1 -1}` @ h2 @ had(20, 1));
                  d02 = d02 || b1[d80[,1],];
                  d40 = d40 || (j(2, 1, 1) @ (1:40)`)[d80[,1],];
                  end;

               else if index(lineage, '80 ** 1 > 4 ** 10 20 ** 1') then do;
                  b1  = ks((0:3)`, gh(20, 4), 4) + 1;
                  d04 = d04 || b1[d80[,1],];
                  d20 = d20 || (j(4, 1, 1) @ (1:20)`)[d80[,1],];
                  end;

               else if index(lineage, '80 ** 1 > 5 ** 1 16 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(16, 1, 1) @ (1:5)`)[d80[,1],];
                  d16 = d16 || ((1:16)` @ j(5, 1, 1))[d80[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d80) = 1 then free d80;
                  else do; file log; put 'ERROR: 80 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------78 runs, recode 78-level factors------;
            avail = max(0, ncol(d78) - any78);
            do while(avail);

               if index(lineage, '78 ** 1 > 2 ** 1 39 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(39, 1, 1) @ (1:2)`)[d78[,1],];
                  d39 = d39 || ((1:39)` @ j(2, 1, 1))[d78[,1],];
                  end;

               else if index(lineage, '78 ** 1 > 3 ** 1 26 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(26, 1, 1) @ (1:3)`)[d78[,1],];
                  d26 = d26 || ((1:26)` @ j(3, 1, 1))[d78[,1],];
                  end;

               else if index(lineage, '78 ** 1 > 6 ** 1 13 ** 1') then do;
                  b1  = 1;
                  d06 = d06 || (j(13, 1, 1) @ (1:6)`)[d78[,1],];
                  d13 = d13 || ((1:13)` @ j(6, 1, 1))[d78[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d78) = 1 then free d78;
                  else do; file log; put 'ERROR: 78 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------77 runs, recode 77-level factors------;
            avail = max(0, ncol(d77) - any77);
            do while(avail);

               if index(lineage, '77 ** 1 > 7 ** 1 11 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(11, 1, 1) @ (1:7)`)[d77[,1],];
                  d11 = d11 || ((1:11)` @ j(7, 1, 1))[d77[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d77) = 1 then free d77;
                  else do; file log; put 'ERROR: 77 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------76 runs, recode 76-level factors------;
            avail = max(0, ncol(d76) - any76);
            do while(avail);

               if index(lineage, '76 ** 1 > 2 ** 75') then do;
                  b1  = hadp2(76, 0);
                  d02 = d02 || b1[d76[,1],];
                  end;

               else if index(lineage, '76 ** 1 > 2 ** 2 38 ** 1')
                  then %mktn22(76);

               else if index(lineage, '76 ** 1 > 4 ** 1 19 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || ((1:4)` @ j(19, 1, 1))[d76[,1],];
                  d19 = d19 || (j(4, 1, 1) @ (1:19)`)[d76[,1],];
                  end;

               else if index(lineage, '76 ** 1 > 2 ** 19 19 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(76, 19, 2)[d76[,1],];
                  d19 = d19 || ((1:19)` @ j(4, 1, 1))[d76[,1],];
                  end;

               else if index(lineage, '76 ** 1 > 4 ** 1 19 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(19, 1, 1) @ (1:4)`)[d76[,1],];
                  d19 = d19 || ((1:19)` @ j(4, 1, 1))[d76[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d76) = 1 then free d76;
                  else do; file log; put 'ERROR: 76 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------75 runs, recode 75-level factors------;
            avail = max(0, ncol(d75) - any75);
            do while(avail);

               if index(lineage, '75 ** 1 > 3 ** 1 25 ** 1') then do;
                  d03 = d03 || ((1:3)` @ j(25, 1, 1))[d75[,1],];
                  d25 = d25 || (j(3, 1, 1) @ (1:25)`)[d75[,1],];
                  end;

               else if index(lineage, '75 ** 1 > 5 ** 8 15 ** 1') then do;
                  d05 = d05 || (ks((0:4)`, gh(15, 5), 5) + 1)[d75[,1],];
                  d15 = d15 || (j(5, 1, 1) @ (1:15)`)[d75[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d75) = 1 then free d75;
                  else do; file log; put 'ERROR: 75 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------74 runs, recode 74-level factors------;
            avail = max(0, ncol(d74) - any74);
            do while(avail);

               if index(lineage, '74 ** 1 > 2 ** 1 37 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(37, 1, 1) @ (1:2)`)[d74[,1],];
                  d37 = d37 || ((1:37)` @ j(2, 1, 1))[d74[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d74) = 1 then free d74;
                  else do; file log; put 'ERROR: 74 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------72 runs, recode 72-level factors------;
            avail = max(0, ncol(d72) - any72);
            do while(avail);

               if index(lineage, '72 ** 1 > 2 ** 68 4 ** 1') then do;
                  b1  = twolev(h2 @ sortmat(hadw(36, 1), 1:2));
                  d02 = d02 || b1[d72[,1],(3:36) || (39:72)];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 60 3 ** 1 4 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 27, 2) || fac(36, 1, 3), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:27];
                  d03 = d03 || b2[d72[,1],28];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 53 3 ** 2 4 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 20, 2) || fac(36, 2, 3), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:20];
                  d03 = d03 || b2[d72[,1],21:22];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 51 3 ** 1 4 ** 1 6 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 20, 2) || fac(36, 2, 3) ||
                                fac(36, 1, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:18];
                  d03 = d03 || b2[d72[,1],21];
                  d06 = d06 || b2[d72[,1],23];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 46 3 ** 2 4 ** 1 6 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1:2);
                  b1  = b1 // reverse(b1);
                  b3  = sortmat(fac(36, 13.1, 2) || {[6]1}` @ (1:3)` @ {[2]1}`
                             || fac(36, 1.3, 3) || (1:6)` @ j(6, 1, 1), 1:5);
                  d02 = d02 || ({[2]1}` @ b3[,2:13] || b1[,2:35])[d72[,1],];
                  d03 = d03 || ({[2]1}` @ b3)[d72[,1],14:15];
                  d06 = d06 || ({[2]1}` @ b3)[d72[,1],16];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 46 4 ** 1 6 ** 2')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = fac(36, 13, 2) || fac(36, 2, 6);
                  b2  = j(2, 1, 1) @ sortmat(b2, 1:2);
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:13];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  d06 = d06 || b2[d72[,1],14:15];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 49 4 ** 1 9 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 16, 2) || (1:9)` @ j(4, 1, 1), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:16];
                  d09 = d09 || b2[d72[,1],17];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 44 3 ** 12 4 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b2  = sortmat(j(3, 1, 1) @ had(12, 0) ||
                                (ks((0:2)`, gh(12, 3), 3) + 1), 1);
                  b1  = (b1 // reverse(b1)) || (b2 // b2);
                  d02 = d02 || b1[d72[,1], (2:35) || (37:46)];
                  d03 = d03 || b1[d72[,1], 47:58];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 43 3 ** 8 4 ** 1 6 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 10.1, 2) || fac(36, 8, 3) ||
                                fac(36, 1.2, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:10];
                  d03 = d03 || b2[d72[,1],11:18];
                  d06 = d06 || b2[d72[,1],19];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 43 3 ** 1 4 ** 1 6 ** 2')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 10, 2) || fac(36, 1.2, 3) ||
                                fac(36, 2.3, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:10];
                  d03 = d03 || b2[d72[,1],11];
                  d06 = d06 || b2[d72[,1],12:13];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 42 3 ** 4 4 ** 1 6 ** 2')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 9, 2) || fac(36, 4, 3) ||
                                fac(36, 2.2, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:9];
                  d03 = d03 || b2[d72[,1],10:13];
                  d06 = d06 || b2[d72[,1],14:15];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 41 4 ** 1 6 ** 3')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 8, 2) || fac(36, 3.4, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:8];
                  d06 = d06 || b2[d72[,1],9:11];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 37 3 ** 13 4 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b2  = sortmat(j(3, 1, 1) @ (fac(12, 4, 2) ||
                                             (1:3)` @ j(4, 1, 1)) ||
                                (ks((0:2)`, gh(12, 3), 3) + 1), 1);
                  b1  = (b1 // reverse(b1)) || (b2 // b2);
                  d02 = d02 || b1[d72[,1], (2:35) || (37:39)];
                  d03 = d03 || b1[d72[,1], 40:52];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 37 3 ** 1 4 ** 1 6 ** 3')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 4, 2) || fac(36, 1.1, 3) ||
                                fac(36, 3.3, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:4];
                  d03 = d03 || b2[d72[,1],5];
                  d06 = d06 || b2[d72[,1],6:8];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 36 3 ** 9 4 ** 1 6 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 3.1, 2) || fac(36, 9, 3) ||
                                fac(36, 1.1, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:3];
                  d03 = d03 || b2[d72[,1],4:12];
                  d06 = d06 || b2[d72[,1],13];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 36 3 ** 2 4 ** 1 6 ** 3')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 3, 2) || fac(36, 2.1, 3) ||
                                fac(36, 3.1, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2:3];
                  d03 = d03 || b2[d72[,1],4:5];
                  d06 = d06 || b2[d72[,1],6:8];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 36 36 ** 1') then do;
                  b1  = twolev({1 -1}` @ hadw(36, 1));
                  d02 = d02 || b1[d72[,1],];
                  d36 = d36 || (j(2, 1, 1) @ (1:36)`)[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 35 3 ** 12 4 ** 1 6 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b2  = sortmat(j(3, 1, 1) @ (twolev({1 -1}` @ j(3, 1, 1) @ h2)
                        || {1 1}` @ (1:6)`) || ks((0:2)`, gh(12, 3), 3) + 1, 1);
                  b1  = (b1 // reverse(b1)) || (b2 // b2);
                  d02 = d02 || b1[d72[,1], (2:35) || 37];
                  d03 = d03 || b1[d72[,1], 39:50];
                  d06 = d06 || b1[d72[,1], 38];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 35 3 ** 5 4 ** 1 6 ** 2')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 2, 2) || fac(36, 5, 3) ||
                                fac(36, 2.1, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2];
                  d03 = d03 || b2[d72[,1],3:7];
                  d06 = d06 || b2[d72[,1],8:9];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 35 4 ** 1 18 ** 1')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(twolev({1 -1}` @ j(9, 1, 1) @ h2) ||
                                j(2, 1, 1) @ (1:18)`, 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35] || b2[d72[,1],2];
                  d18 = d18 || b2[d72[,1],3];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 34 3 ** 8 4 ** 1 6 ** 2')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 3.2, 6) || fac(36, 7, 3), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35];
                  d03 = d03 || (j(4, 1, 1) @ (1:3)` @ j(6, 1, 1))[d72[,1],] ||
                               b2[d72[,1],4:10];
                  d06 = d06 || b2[d72[,1],2:3];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 34 3 ** 3 4 ** 1 6 ** 3')
                  then do;
                  b1  = sortmat(hadw(36, 0), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 1, 2) || fac(36, 3, 6) ||
                                fac(36, 3, 3), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:35];
                  d03 = d03 || b2[d72[,1],5:7];
                  d06 = d06 || b2[d72[,1],2:4];
                  d04 = d04 || ((1:4)` @ j(18, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 31 6 ** 4') then do;
                  b1  = sortmat(fac(36, 1, 3) || fac(36, 27, 2), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 1.1, 3) || fac(36, 4, 2) ||
                                fac(36, 3.3, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:28] || b2[d72[,1],2:5];
                  d06 = d06 || b2[d72[,1],6:8] ||
                               ((1:6)` @ j(12, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 30 3 ** 1 6 ** 4')
                  then do;
                  b1  = sortmat(fac(36, 1, 3) || fac(36, 27, 2), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 2.1, 3) || fac(36, 3, 2) ||
                                fac(36, 3.1, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:28] || b2[d72[,1],3:5];
                  d03 = d03 || b2[d72[,1],2];
                  d06 = d06 || b2[d72[,1],6:8] ||
                               ((1:6)` @ j(12, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 28 3 ** 2 6 ** 4')
                  then do;
                  b1  = sortmat(fac(36, 1, 3) || fac(36, 27, 2), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 3, 3) || fac(36, 1, 2) ||
                                fac(36, 3, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:28] || b2[d72[,1],4];
                  d03 = d03 || b2[d72[,1],2:3];
                  d06 = d06 || b2[d72[,1],5:7] ||
                               ((1:6)` @ j(12, 1, 1))[d72[,1],];
                  end;

               else if index(lineage,
                  '72 ** 1 > 2 ** 27 3 ** 11 6 ** 1 12 ** 1') then do;
                  b1  = sortmat(fac(36, 1, 3) || fac(36, 27, 2), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(ks((0:2)`, gh(12, 3), 3) + 1 ||
                                j(3, 1, 1) @ (1:12)`, 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:28];
                  d03 = d03 || b2[d72[,1],2:12];
                  d12 = d12 || b2[d72[,1],13];
                  d06 = d06 || ((1:6)` @ j(12, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 27 3 ** 6 6 ** 4')
                  then do;
                  b1  = sortmat(fac(36, 1, 3) || fac(36, 27, 2), 1);
                  b1  = b1 // reverse(b1);
                  b2  = sortmat(fac(36, 7, 3) || fac(36, 3.2, 6), 1);
                  b2  = b2 // b2;
                  d02 = d02 || b1[d72[,1],2:28];
                  d03 = d03 || b2[d72[,1],2:7];
                  d06 = d06 || b2[d72[,1],8:10] ||
                               ((1:6)` @ j(12, 1, 1))[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 19 3 ** 20 4 ** 1 6 ** 1')
                  then do;
                  b1  = fac(72, 1, 0);
                  d04 = d04 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2];
                  d03 = d03 || b1[d72[,1], 3:22];
                  d02 = d02 || b1[d72[,1],23:41];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 18 3 ** 16 4 ** 1 6 ** 2')
                  then do;
                  b1  = fac(72, 2, 0);
                  d04 = d04 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2:3];
                  d03 = d03 || b1[d72[,1], 4:19];
                  d02 = d02 || b1[d72[,1],20:37];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 17 3 ** 12 4 ** 1 6 ** 3')
                  then do;
                  b1  = fac(72, 3, 0);
                  d04 = d04 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2:4];
                  d03 = d03 || b1[d72[,1], 5:16];
                  d02 = d02 || b1[d72[,1],17:33];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 15 3 ** 7 4 ** 1 6 ** 5')
                  then do;
                  b1  = fac(72, 15, 0);
                  d06 = d06 || b1[d72[,1], {1 7 8 9 10}];
                  d04 = d04 || b1[d72[,1], 41];
                  d03 = d03 || b1[d72[,1], 14:20];
                  d02 = d02 || b1[d72[,1], (23:27) || (29:38)];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 14 3 ** 3 4 ** 1 6 ** 6')
                  then do;
                  b1  = fac(72, 15, 0);
                  d06 = d06 || b1[d72[,1], (1:6)];
                  d04 = d04 || b1[d72[,1], 41];
                  d03 = d03 || b1[d72[,1], 11:13];
                  d02 = d02 || b1[d72[,1], (23:26) || (29:38)];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 12 3 ** 21 4 ** 1 6 ** 1')
                  then do;
                  b1  = fac(72, 5, 0);
                  d04 = d04 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2];
                  d03 = d03 || b1[d72[,1], 3:23];
                  d02 = d02 || b1[d72[,1],24:35];
                  end;

               else if index(lineage,
                  '72 ** 1 > 2 ** 11 3 ** 20 6 ** 1 12 ** 1') then do;
                  b1  = fac(72, 6, 0);
                  d12 = d12 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2];
                  d03 = d03 || b1[d72[,1], 3:22];
                  d02 = d02 || b1[d72[,1],23:33];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 11 3 ** 17 4 ** 1 6 ** 2')
                  then do;
                  b1  = fac(72, 7, 0);
                  d04 = d04 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2:3];
                  d03 = d03 || b1[d72[,1], 4:20];
                  d02 = d02 || b1[d72[,1],21:31];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 10 3 ** 20 4 ** 1 6 ** 2')
                  then do;
                  b1  = fac(72, 8, 0);
                  d04 = d04 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2:3];
                  d03 = d03 || b1[d72[,1], 4:23];
                  d02 = d02 || b1[d72[,1],24:33];
                  end;

               else if index(lineage,
                  '72 ** 1 > 2 ** 10 3 ** 16 6 ** 2 12 ** 1') then do;
                  b1  = fac(72, 9, 0);
                  d12 = d12 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2:3];
                  d03 = d03 || b1[d72[,1], 4:19];
                  d02 = d02 || b1[d72[,1],20:29];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 10 3 ** 13 4 ** 1 6 ** 3')
                  then do;
                  b1  = fac(72, 10, 0);
                  d04 = d04 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2:4];
                  d03 = d03 || b1[d72[,1], 5:17];
                  d02 = d02 || b1[d72[,1],18:27];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 9 3 ** 16 4 ** 1 6 ** 3')
                  then do;
                  b1  = fac(72, 11, 0);
                  d04 = d04 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2:4];
                  d03 = d03 || b1[d72[,1], 5:20];
                  d02 = d02 || b1[d72[,1],21:29];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 9 3 ** 12 6 ** 3 12 ** 1')
                  then do;
                  b1  = fac(72, 12, 0);
                  d12 = d12 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2:4];
                  d03 = d03 || b1[d72[,1], 5:16];
                  d02 = d02 || b1[d72[,1],17:25];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 8 3 ** 12 4 ** 1 6 ** 4')
                  then do;
                  b1  = fac(72, 14, 0);
                  d04 = d04 || b1[d72[,1], 1];
                  d06 = d06 || b1[d72[,1], 2:5];
                  d03 = d03 || b1[d72[,1], 6:17];
                  d02 = d02 || b1[d72[,1],18:25];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 8 3 ** 8 4 ** 1 6 ** 5')
                  then do;
                  b1  = fac(72, 15, 0);
                  d06 = d06 || b1[d72[,1], {1 7 8 9 10}];
                  d04 = d04 || b1[d72[,1], 41];
                  d03 = d03 || b1[d72[,1], 42 || (14:20)];
                  d02 = d02 || b1[d72[,1], {23 24 25 26 27 29 30 31}];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 7 3 ** 7 6 ** 5 12 ** 1')
                  then do;
                  b1  = j(6, 1, 1) @ (1:12)`;
                  d12 = d12 || b1[d72[,1],];
                  b1  = ks((0:5)`, gh(12, 6)[,{1 3 4 5 6}], 6) + 1;
                  d06 = d06 || b1[d72[,1],];
                  b1  = ks(j(2, 1, 1) @ (0:2)`, fac(12, 7, 3), 3) + 1;
                  d03 = d03 || b1[d72[,1],];
                  b1  = ks(j(3, 1, 1) @ (0:1)`, fac(12, 8, 2), 2) + 1;
                  d02 = d02 || b1[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 7 3 ** 4 4 ** 1 6 ** 6')
                  then do;
                  b1  = fac(72, 15, 0);
                  d06 = d06 || b1[d72[,1], (1:6)];
                  d04 = d04 || b1[d72[,1], 41];
                  d03 = d03 || b1[d72[,1], 42 || (11:13)];
                  d02 = d02 || b1[d72[,1], (23:26) || (29:31)];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 6 3 ** 7 4 ** 1 6 ** 6')
                  then do;
                  b1  = fac(72, 15, 0);
                  d06 = d06 || b1[d72[,1], {40 1 7 8 9 10}];
                  d04 = d04 || b1[d72[,1], 41];
                  d03 = d03 || b1[d72[,1], 14:20];
                  d02 = d02 || b1[d72[,1], {23 24 25 26 27 29}];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 6 3 ** 3 6 ** 6 12 ** 1')
                  then do;
                  b1  = j(6, 1, 1) @ (1:12)`;
                  d12 = d12 || b1[d72[,1],];
                  b1  = ks((0:5)`, gh(12, 6), 6) + 1;
                  d06 = d06 || b1[d72[,1],];
                  b1  = ks(j(2, 1, 1) @ (0:2)`, fac(12, 3, 3), 3) + 1;
                  d03 = d03 || b1[d72[,1],];
                  b1  = ks(j(3, 1, 1) @ (0:1)`, fac(12, 6, 2), 2) + 1;
                  d02 = d02 || b1[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 2 ** 5 3 ** 3 4 ** 1 6 ** 7')
                  then do;
                  b1  = fac(72, 15, 0);
                  d06 = d06 || b1[d72[,1], 40 || (1:6)];
                  d04 = d04 || b1[d72[,1], 41];
                  d03 = d03 || b1[d72[,1], 11:13];
                  d02 = d02 || b1[d72[,1], (23:26) || 29];
                  end;

               else if index(lineage, '72 ** 1 > 3 ** 24 24 ** 1') then do;
                  b1  = ks((0:2)`, gh(24, 3), 3) + 1;
                  d03 = d03 || b1[d72[,1],];
                  b1  = j(3, 1, 1) @ (1:24)`;
                  d24 = d24 || b1[d72[,1],];
                  end;

               else if index(lineage, '72 ** 1 > 8 ** 1 9 ** 1') then do;
                  b1  = 1;
                  d08 = d08 || (j(9, 1, 1) @ (1:8)`)[d72[,1],];
                  d09 = d09 || ((1:9)` @ j(8, 1, 1))[d72[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d72) = 1 then free d72;
                  else do; file log; put 'ERROR: 72 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------70 runs, recode 70-level factors------;
            avail = max(0, ncol(d70) - any70);
            do while(avail);

               if index(lineage, '70 ** 1 > 2 ** 1 35 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(35, 1, 1) @ (1:2)`)[d70[,1],];
                  d35 = d35 || ((1:35)` @ j(2, 1, 1))[d70[,1],];
                  end;

               else if index(lineage, '70 ** 1 > 5 ** 1 14 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(14, 1, 1) @ (1:5)`)[d70[,1],];
                  d14 = d14 || ((1:14)` @ j(5, 1, 1))[d70[,1],];
                  end;

               else if index(lineage, '70 ** 1 > 7 ** 1 10 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(10, 1, 1) @ (1:7)`)[d70[,1],];
                  d10 = d10 || ((1:10)` @ j(7, 1, 1))[d70[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d70) = 1 then free d70;
                  else do; file log; put 'ERROR: 70 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------69 runs, recode 69-level factors------;
            avail = max(0, ncol(d69) - any69);
            do while(avail);

               if index(lineage, '69 ** 1 > 3 ** 1 23 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(23, 1, 1) @ (1:3)`)[d69[,1],];
                  d23 = d23 || ((1:23)` @ j(3, 1, 1))[d69[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d69) = 1 then free d69;
                  else do; file log; put 'ERROR: 69 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------68 runs, recode 68-level factors------;
            avail = max(0, ncol(d68) - any68);
            do while(avail);

               if index(lineage, '68 ** 1 > 2 ** 67') then do;
                  b1  = hadp1(68, 0);
                  d02 = d02 || b1[d68[,1],];
                  end;

               else if index(lineage, '68 ** 1 > 2 ** 2 34 ** 1')
                   then %mktn22(68);

               else if index(lineage, '68 ** 1 > 2 ** 18 17 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(68, 18, 2)[d68[,1],];
                  d17 = d17 || ((1:17)` @ j(4, 1, 1))[d68[,1],];
                  end;

               else if index(lineage, '68 ** 1 > 4 ** 1 17 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(17, 1, 1) @ (1:4)`)[d68[,1],];
                  d17 = d17 || ((1:17)` @ j(4, 1, 1))[d68[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d68) = 1 then free d68;
                  else do; file log; put 'ERROR: 68 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------66 runs, recode 66-level factors------;
            avail = max(0, ncol(d66) - any66);
            do while(avail);

               if index(lineage, '66 ** 1 > 2 ** 1 33 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(33, 1, 1) @ (1:2)`)[d66[,1],];
                  d33 = d33 || ((1:33)` @ j(2, 1, 1))[d66[,1],];
                  end;

               else if index(lineage, '66 ** 1 > 3 ** 1 22 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(22, 1, 1) @ (1:3)`)[d66[,1],];
                  d22 = d22 || ((1:22)` @ j(3, 1, 1))[d66[,1],];
                  end;

               else if index(lineage, '66 ** 1 > 6 ** 1 11 ** 1') then do;
                  b1  = 1;
                  d06 = d06 || (j(11, 1, 1) @ (1:6)`)[d66[,1],];
                  d11 = d11 || ((1:11)` @ j(6, 1, 1))[d66[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d66) = 1 then free d66;
                  else do; file log; put 'ERROR: 66 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------65 runs, recode 65-level factors------;
            avail = max(0, ncol(d65) - any65);
            do while(avail);

               if index(lineage, '65 ** 1 > 5 ** 1 13 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(13, 1, 1) @ (1:5)`)[d65[,1],];
                  d13 = d13 || ((1:13)` @ j(5, 1, 1))[d65[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d65) = 1 then free d65;
                  else do; file log; put 'ERROR: 65 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------63 runs, recode 63-level factors------;
            avail = max(0, ncol(d63) - any63);
            do while(avail);

               if index(lineage, '63 ** 1 > 3 ** 12 21 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (ks((0:2)`, gh(21, 3), 3) + 1)[d63[,1],];
                  d21 = d21 || (j(3, 1, 1) @ (1:21)`)[d63[,1],];
                  end;

               else if index(lineage, '63 ** 1 > 7 ** 1 9 ** 1') then do;
                  b1  = 1;
                  d07 = d07 || (j(9, 1, 1) @ (1:7)`)[d63[,1],];
                  d09 = d09 || ((1:9)` @ j(7, 1, 1))[d63[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d63) = 1 then free d63;
                  else do; file log; put 'ERROR: 63 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------62 runs, recode 62-level factors------;
            avail = max(0, ncol(d62) - any62);
            do while(avail);

               if index(lineage, '62 ** 1 > 2 ** 1 31 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(31, 1, 1) @ (1:2)`)[d62[,1],];
                  d31 = d31 || ((1:31)` @ j(2, 1, 1))[d62[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d62) = 1 then free d62;
                  else do; file log; put 'ERROR: 62 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------60 runs, recode 60-level factors------;
            avail = max(0, ncol(d60) - any60);
            do while(avail);

               if index(lineage, '60 ** 1 > 2 ** 59') then do;
                  b1  = 1;
                  d02 = d02 || hadp1(60, 0)[d60[,1],];
                  end;

               else if index(lineage, '60 ** 1 > 2 ** 30 3 ** 1') then do;
                  b1  = fac(60, 30, 2);
                  d02 = d02 || b1[d60[,1],];
                  d03 = d03 || ((1:3)` @ j(20, 1, 1))[d60[,1],];
                  end;

               else if index(lineage, '60 ** 1 > 2 ** 23 5 ** 1') then do;
                  b1  = twolev(h2 @ had(12, 1));
                  b2  = (b1[,2:22] // fac(36, 27, 2)[,1:21]) || fac(60, 2, 2);
                  d02 = d02 || b2[d60[,1],];
                  d05 = d05 || (b1[,24] // fac(36, 1, 3) + 2)[d60[,1],];
                  end;

              else if index(lineage, '60 ** 1 > 2 ** 24 6 ** 1') then do;
                  b1  = had(20, 0);
                  b2  = sortmat(b1, 1);
                  d02 = d02 || ((b1 // b2 // reverse(b2)) || fac(60, 7, 2))
                               [d60[,1],(2:18) || (20:26)];
                  d06 = d06 || (b1[,1] // (3:6)` @ j(10, 1, 1))[d60[,1],];
                  end;

               else if index(lineage, '60 ** 1 > 2 ** 21 10 ** 1') then do;
                  b1  = twolev((h2 @ had(12, 1))[,2:24]);
                  b2  = b1[,1] # 2 + b1[,12] - 2;
                  d02 = d02 || ((b1[,(2:11) || (14:21)] //
                        fac(36, 20, 2)[,1:18]) || fac(60, 3, 2))[d60[,1],];
                  d10 = d10 || (b2 // fac(36, 1, 6) + 4)[d60[,1],];
                  end;

               else if index(lineage, '60 ** 1 > 2 ** 17 15 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(60, 17, 2)[d60[,1],];
                  d15 = d15 || ((1:15)` @ j(4, 1, 1))[d60[,1],];
                  end;

               else if index(lineage, '60 ** 1 > 2 ** 15 6 ** 1 10 ** 1')
                  then do;
                  b1  = fac(60, 17, 0);
                  d02 = d02 || b1[d60[,1],1:15];
                  d06 = d06 || b1[d60[,1],16];
                  d10 = d10 || b1[d60[,1],17];
                  end;

               else if index(lineage, '60 ** 1 > 2 ** 2 30 ** 1') then
                  %mktn22(60);

               else if index(lineage, '60 ** 1 > 4 ** 1 15 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(15, 1, 1) @ (1:4)`)[d60[,1],];
                  d15 = d15 || ((1:15)` @ j(4, 1, 1))[d60[,1],];
                  end;

               else if index(lineage, '60 ** 1 > 5 ** 1 12 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(12, 1, 1) @ (1:5)`)[d60[,1],];
                  d12 = d12 || ((1:12)` @ j(5, 1, 1))[d60[,1],];
                  end;

               else if index(lineage, '60 ** 1 > 3 ** 1 20 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(20, 1, 1) @ (1:3)`)[d60[,1],];
                  d20 = d20 || ((1:20)` @ j(3, 1, 1))[d60[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d60) = 1 then free d60;
                  else do; file log; put 'ERROR: 60 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------58 runs, recode 58-level factors------;
            avail = max(0, ncol(d58) - any58);
            do while(avail);

               if index(lineage, '58 ** 1 > 2 ** 1 29 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(29, 1, 1) @ (1:2)`)[d58[,1],];
                  d29 = d29 || ((1:29)` @ j(2, 1, 1))[d58[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d58) = 1 then free d58;
                  else do; file log; put 'ERROR: 58 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------57 runs, recode 57-level factors------;
            avail = max(0, ncol(d57) - any57);
            do while(avail);

               if index(lineage, '57 ** 1 > 3 ** 1 19 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(19, 1, 1) @ (1:3)`)[d57[,1],];
                  d19 = d19 || ((1:19)` @ j(3, 1, 1))[d57[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d57) = 1 then free d57;
                  else do; file log; put 'ERROR: 57 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------56 runs, recode 56-level factors------;
            avail = max(0, ncol(d56) - any56);
            do while(avail);

               if index(lineage, '56 ** 1 > 2 ** 52 4 ** 1') then do;
                  b1  = twolev(h2 @ sortmat(hadw(28, 1), 1:2));
                  d02 = d02 || b1[d56[,1],(3:28) || (31:56)];
                  d04 = d04 || ((1:4)` @ j(14, 1, 1))[d56[,1],];
                  end;

               else if index(lineage, '56 ** 1 > 2 ** 28 28 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (twolev({1 -1}` @ hadw(28, 1)))[d56[,1],];
                  d28 = d28 || (j(2, 1, 1) @ (1:28)`)[d56[,1],];
                  end;

               else if index(lineage, '56 ** 1 > 2 ** 37 4 ** 1 7 ** 1')
                  then do;
                  b1  = fac(28, 12, 2);
                  b1  = j(2, 1, 1) @ sortmat(b1 || (1:7)` @ j(4, 1, 1), 1:2);
                  b2  = twolev({1 -1}` @ sortmat(hadw(28, 1), 1:2));
                  d02 = d02 || (b2[,3:28] || b1[,2:12])[d56[,1],];
                  d04 = d04 || ((1:4)` @ j(14, 1, 1)  )[d56[,1],];
                  d07 = d07 || b1[d56[,1],13];
                  end;

               else if index(lineage, '56 ** 1 > 2 ** 27 4 ** 1 14 ** 1')
                  then do;
                  b1  = j(2, 1, 1) @ twolev({1 -1}` @ j(7, 1, 1) @ h2);
                  b2  = (twolev({1 -1}` @ sortmat(hadw(28, 1), 1:2)));
                  d02 = d02 || (b2[,3:28] || b1[,2] )[d56[,1],];
                  d04 = d04 || ((1:4)` @ j(14, 1, 1))[d56[,1],];
                  d14 = d14 || (j(4, 1, 1) @ (1:14)`)[d56[,1],];
                  end;

               else if index(lineage, '56 ** 1 > 7 ** 1 8 ** 1') then do;
                  b1  = j(8, 1, 1) @ (1:7)`;
                  b2  = (1:8)` @ j(7, 1, 1);
                  d07 = d07 || b1[d56[,1],];
                  d08 = d08 || b2[d56[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d56) = 1 then free d56;
                  else do; file log; put 'ERROR: 56 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------55 runs, recode 55-level factors------;
            avail = max(0, ncol(d55) - any55);
            do while(avail);

               if index(lineage, '55 ** 1 > 5 ** 1 11 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(11, 1, 1) @ (1:5)`)[d55[,1],];
                  d11 = d11 || ((1:11)` @ j(5, 1, 1))[d55[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d55) = 1 then free d55;
                  else do; file log; put 'ERROR: 55 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------54 runs, recode 54-level factors------;
            avail = max(0, ncol(d54) - any54);
            do while(avail);

               if index(lineage, '54 ** 1 > 3 ** 18 18 ** 1') then do;
                  b1  = ks((0:2)`, gh(18, 3), 3) + 1;
                  d03 = d03 || b1[d54[,1],];
                  d18 = d18 || (j(3, 1, 1) @ (1:18)`)[d54[,1],];
                  end;

               else if index(lineage, '54 ** 1 > 2 ** 1 27 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(27, 1, 1) @ (1:2)`)[d54[,1],];
                  d27 = d27 || ((1:27)` @ j(2, 1, 1))[d54[,1],];
                  end;

               else if index(lineage, '54 ** 1 > 3 ** 20 6 ** 1 9 ** 1')
                  then do;
                  b1  = j(3, 1, 1) @ (0:2)` || ks((0:2)`, gh(3, 3), 3);
                  b1  = ks(b1, gh(6, 3), 3);
                  d03 = d03 || b1[d54[,1], (2:6) || (8:12) ||
                                           (14:18) || (20:24)] + 1;
                  d06 = d06 || (j(9, 1, 1) @ (1:6)`)[d54[,1],];
                  d09 = d09 || ((1:9)` @ j(6, 1, 1))[d54[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d54) = 1 then free d54;
                  else do; file log; put 'ERROR: 54 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------ 52 runs, recode 52-level factors ------;
            avail = max(0, ncol(d52) - any52);
            do while(avail);

               if index(lineage, '52 ** 1 > 2 ** 51') then do;
                  b1  = 1;
                  d02 = d02 || hadw(52, 0)[d52[,1],];
                  end;

               else if index(lineage, '52 ** 1 > 2 ** 2 26 ** 1')
                  then %mktn22(52);

               else if index(lineage, '52 ** 1 > 2 ** 16 13 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(52, 16, 2)[d52[,1],];
                  d13 = d13 || ((1:13)` @ j(4, 1, 1))[d52[,1],];
                  end;

               else if index(lineage, '52 ** 1 > 4 ** 1 13 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(13, 1, 1) @ (1:4)`)[d52[,1],];
                  d13 = d13 || ((1:13)` @ j(4, 1, 1))[d52[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d52) = 1 then free d52;
                  else do; file log; put 'ERROR: 52 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------51 runs, recode 51-level factors------;
            avail = max(0, ncol(d51) - any51);
            do while(avail);

               if index(lineage, '51 ** 1 > 3 ** 1 17 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(17, 1, 1) @ (1:3)`)[d51[,1],];
                  d17 = d17 || ((1:17)` @ j(3, 1, 1))[d51[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d51) = 1 then free d51;
                  else do; file log; put 'ERROR: 51 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------50 runs, recode 50-level factors------;
            avail = max(0, ncol(d50) - any50);
            do while(avail);

               if index(lineage, '50 ** 1 > 5 ** 10 10 ** 1') then do;
                  b1  = ks((0:4)`, gh(10, 5), 5) + 1;
                  d05 = d05 || b1[d50[,1],];
                  b1  = j(5, 1, 1) @ (1:10)`;
                  d10 = d10 || b1[d50[,1],];
                  end;

               else if index(lineage, '50 ** 1 > 2 ** 1 25 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || (j(25, 1, 1) @ (1:2)`)[d50[,1],];
                  d25 = d25 || ((1:25)` @ j(2, 1, 1))[d50[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d50) = 1 then free d50;
                  else do; file log; put 'ERROR: 50 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------49 runs, recode 49-level factors------;
            avail = max(0, ncol(d49) - any49);
            do while(avail);

               b1  = ks((0:6)`, gh(7, 7), 7) + 1 || j(7, 1, 1) @ (1:7)`;
               d07 = d07 || b1[d49[,1],];

               if ncol(b1) then do;
                  if ncol(d49) = 1 then free d49;
                  else do; file log; put 'ERROR: 49 ** 2.'; file print; end;
                  free b1;
                  end;

               free b1;
               avail = avail - 1;
               end;

            *------48 runs, recode 48-level factors------;
            avail = max(0, ncol(d48) - any48);
            do while(avail);

               if index(lineage, '48 ** 1 > 2 ** 40 8 ** 1') then do;
                  b1  = twolev((h2 @ h2) @ sortmat(had(12, 1), 1:2));
                  b2  = j(1, 48, 1); b2[{1 2 13 14 25 26 37 38}] = 0;
                  d02 = d02 || b1[d48[,1],loc(b2)];
                  d08 = d08 || ((1:8)` @ j(6, 1, 1))[d48[,1],];
                  end;

               else if index(lineage, '48 ** 1 > 2 ** 33 3 ** 1 8 ** 1')
                  then do;
                  b1  = twolev((h2 @ h2)[,2:4] @ sortmat(had(12, 1), 2)) ||
                        j(4, 1, 1) @ sortmat(fac(12, 4, 2) ||
                                             (1:3)` @ j(4, 1, 1), 1);
                  d03 = d03 || b1[d48[,1],41];
                  d08 = d08 || ((1:8)` @ j(6, 1, 1))[d48[,1],];
                  b2  = j(1, 41, 1); b2[{1 2 13 14 25 26 37 41}] = 0;
                  d02 = d02 || b1[d48[,1], loc(b2)];
                  end;

               else if index(lineage, '48 ** 1 > 2 ** 31 6 ** 1 8 ** 1')
                  then do;
                  b1  = fac(48, 33, 0);
                  d02 = d02 || b1[d48[,1], 1:31];
                  d06 = d06 || b1[d48[,1], 32];
                  d08 = d08 || b1[d48[,1], 33];
                  end;

               else if index(lineage, '48 ** 1 > 2 ** 24 24 ** 1') then do;
                  b1  = twolev({1 -1}` @ h2 @ had(12, 1));
                  d02 = d02 || b1[d48[,1],];
                  b1  = j(2, 1, 1) @ (1:24)`;
                  d24 = d24 || b1[d48[,1],];
                  end;

               else if index(lineage, '48 ** 1 > 4 ** 12 12 ** 1') then do;
                  b1  = ks((0:3)`, gh(12, 4), 4) + 1;
                  d04 = d04 || b1[d48[,1],];
                  b1  = j(4, 1, 1) @ (1:12)`;
                  d12 = d12 || b1[d48[,1],];
                  end;

               else if index(lineage, '48 ** 1 > 3 ** 1 16 ** 1') then do;
                  b1  = j(16, 1, 1) @ (1:3)`;
                  d03 = d03 || b1[d48[,1],];
                  b1  = j(3, 1, 1) @ (1:16)`;
                  d16 = d16 || b1[d48[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d48) = 1 then free d48;
                  else do; file log; put 'ERROR: 48 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------46 runs, recode 46-level factors------;
            avail = max(0, ncol(d46) - any46);
            do while(avail);

               b1  = j(23, 1, 1) @ (1:2)`;
               b2  = (1:23)` @ j(2, 1, 1);
               d02 = d02 || b1[d46[,1],];
               d23 = d23 || b2[d46[,1],];

               if ncol(d46) = 1 then free d46;
               else d46 = d46[,2:ncol(d46)];
               avail = avail - 1;
               end;

            *------45 runs, recode 45-level factors------;
            avail = max(0, ncol(d45) - any45);
            do while(avail);

               if index(lineage, '45 ** 1 > 3 ** 9 15 ** 1') then do;
                  b1  = ks((0:2)`, gh(15, 3), 3) + 1;
                  d03 = d03 || b1[d45[,1],];
                  b1  = j(3, 1, 1) @ (1:15)`;
                  d15 = d15 || b1[d45[,1],];
                  end;

               else if index(lineage, '45 ** 1 > 5 ** 1 9 ** 1') then do;
                  b1  = (1:5)` @ j(9, 1, 1);
                  d05 = d05 || b1[d45[,1],];
                  b1  = j(5, 1, 1) @ (1:9)`;
                  d09 = d09 || b1[d45[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d45) = 1 then free d45;
                  else do; file log; put 'ERROR: 45 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------44 runs, recode 44-level factors------;
            avail = max(0, ncol(d44) - any44);
            do while(avail);

               if index(lineage, '44 ** 1 > 2 ** 2 22 ** 1') then %mktn22(44);

               else if index(lineage, '44 ** 1 > 2 ** 15 11 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(44, 15, 2)[d44[,1],];
                  d11 = d11 || ((1:11)` @ j(4, 1, 1))[d44[,1],];
                  end;

               else if index(lineage, '44 ** 1 > 2 ** 43') then do;
                  b1  = hadp1(44, 0);
                  d02 = d02 || b1[d44[,1],];
                  end;

               else if index(lineage, '44 ** 1 > 4 ** 1 11 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(11, 1, 1) @ (1:4)`)[d44[,1],];
                  d11 = d11 || ((1:11)` @ j(4, 1, 1))[d44[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d44) = 1 then free d44;
                  else do; file log; put 'ERROR: 44 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------42 runs, recode 42-level factors------;
            avail = max(0, ncol(d42) - any42);
            do while(avail);

               if index(lineage, '42 ** 1 > 2 ** 1 21 ** 1') then do;
                  b1  = j(21, 1, 1) @ (1:2)`;
                  b2  = (1:21)` @ j(2, 1, 1);
                  d02 = d02 || b1[d42[,1],];
                  d21 = d21 || b2[d42[,1],];
                  end;

               else if index(lineage, '42 ** 1 > 6 ** 1 7 ** 1') then do;
                  b1  = j(7, 1, 1) @ (1:6)`;
                  b2  = (1:7)` @ j(6, 1, 1);
                  d06 = d06 || b1[d42[,1],];
                  d07 = d07 || b2[d42[,1],];
                  end;

               else if index(lineage, '42 ** 1 > 3 ** 1 14 ** 1') then do;
                  b1  = 1;
                  d03 = d03 || (j(14, 1, 1) @ (1:3)`)[d42[,1],];
                  d14 = d14 || ((1:14)` @ j(3, 1, 1))[d42[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d42) = 1 then free d42;
                  else do; file log; put 'ERROR: 42 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------40 runs, recode 40-level factors------;
            avail = max(0, ncol(d40) - any40);
            do while(avail);

               if index(lineage, '40 ** 1 > 2 ** 36 4 ** 1') then do;
                  b1  = twolev(h2 @ sortmat(had(20, 1), 1:2));
                  d02 = d02 || b1[d40[,1],(3:20) || (23:40)];
                  d04 = d04 || ((1:4)` @ j(10, 1, 1))[d40[,1],];
                  end;

               else if index(lineage, '40 ** 1 > 2 ** 25 4 ** 1 5 ** 1')
                  then do;
                  b1  = sortmat(had(20, 0), 1);
                  b2  = sortmat(fac(20, 8, 2) || (1:5)` @ j(4, 1, 1), 1);
                  b1  = b1  // reverse(b1);
                  b2  = (b2 // b2) || (1:4)` @ j(10, 1, 1);
                  d02 = d02 || (b1[,2:19] || b2[,2:8])[d40[,1],];
                  d04 = d04 || b2[d40[,1],10];
                  d05 = d05 || b2[d40[,1], 9];
                  end;

               else if index(lineage, '40 ** 1 > 2 ** 20 20 ** 1') then do;
                  b1  = twolev({1 -1}` @ had(20, 1));
                  d02 = d02 || b1[d40[,1],];
                  b1  = j(2, 1, 1) @ (1:20)`;
                  d20 = d20 || b1[d40[,1],];
                  end;

               else if index(lineage, '40 ** 1 > 2 ** 19 4 ** 1 10 ** 1')
                  then do;
                  b1  = sortmat(had(20, 0), 1);
                  b2  = twolev({1 -1}` @ j(5, 1, 1) @ h2) || {1 1}` @ (1:10)`;
                  b1  = b1  // reverse(b1);
                  b2  = (b2 // b2) || (1:4)` @ j(10, 1, 1);
                  d02 = d02 || (b1[,2:19] || b2[,2])[d40[,1],];
                  d04 = d04 || b2[d40[,1],4];
                  d10 = d10 || b2[d40[,1],3];
                  end;

               else if index(lineage, '40 ** 1 > 5 ** 1 8 ** 1') then do;
                  b1  = 1;
                  d05 = d05 || (j(8, 1, 1) @ (1:5)`)[d40[,1],];
                  d08 = d08 || ((1:8)` @ j(5, 1, 1))[d40[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d40) = 1 then free d40;
                  else do; file log; put 'ERROR: 40 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------39 runs, recode 39-level factors------;
            avail = max(0, ncol(d39) - any39);
            do while(avail);

               b1  = j(13, 1, 1) @ (1:3)`;
               b2  = (1:13)` @ j(3, 1, 1);
               d03 = d03 || b1[d39[,1],];
               d13 = d13 || b2[d39[,1],];

               if ncol(d39) = 1 then free d39;
               else d39 = d39[,2:ncol(d39)];
               avail = avail - 1;
               end;

            *------38 runs, recode 38-level factors------;
            avail = max(0, ncol(d38) - any38);
            do while(avail);

               b1  = j(19, 1, 1) @ (1:2)`;
               b2  = (1:19)` @ j(2, 1, 1);
               d02 = d02 || b1[d38[,1],];
               d19 = d19 || b2[d38[,1],];

               if ncol(d38) = 1 then free d38;
               else d38 = d38[,2:ncol(d38)];
               avail = avail - 1;
               end;

            *------36 runs, recode 36-level factors------;
            avail = max(0, ncol(d36) - any36);
            do while(avail);

               if index(lineage, '36 ** 1 > 2 ** 35') then do;
                  b1  = hadw(36, 0);
                  d02 = d02 || b1[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 27 3 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 27, 2)[d36[,1],];
                  d03 = d03 || fac(36, 1, 3)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 20 3 ** 2') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 20, 2)[d36[,1],];
                  d03 = d03 || fac(36, 2, 3)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 18 3 ** 1 6 ** 1')
                  then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 20, 2)[d36[,1],1:18];
                  d03 = d03 || fac(36, 2, 3)[d36[,1],1];
                  d06 = d06 || fac(36, 1, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 16 9 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 16, 2)[d36[,1],];
                  d09 = d09 || ((1:9)` @ j(4, 1, 1))[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 13 3 ** 2 6 ** 1')
                  then do;
                  b1  = {[6]1}` @ (1:3)` @ {[2]1}` ;
                  d02 = d02 || fac(36, 13.1, 2)[d36[,1],];
                  d03 = d03 || (b1 || fac(36, 1.3, 3))[d36[,1],];
                  d06 = d06 || ((1:6)` @ j(6, 1, 1))[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 13 6 ** 2') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 13, 2)[d36[,1],];
                  d06 = d06 || fac(36, 2, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 10 3 ** 8 6 ** 1')
                  then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 10.1, 2)[d36[,1],];
                  d03 = d03 || fac(36, 8, 3)[d36[,1],];
                  d06 = d06 || fac(36, 1.2, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 10 3 ** 1 6 ** 2')
                  then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 10, 2)[d36[,1],];
                  d03 = d03 || fac(36, 1.2, 3)[d36[,1],];
                  d06 = d06 || fac(36, 2.3, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 9 3 ** 4 6 ** 2') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 9, 2)[d36[,1],];
                  d03 = d03 || fac(36, 4, 3)[d36[,1],];
                  d06 = d06 || fac(36, 2.2, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 8 6 ** 3') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 8, 2)[d36[,1],];
                  d06 = d06 || fac(36, 3.4, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 4 3 ** 1 6 ** 3') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 4, 2)[d36[,1],];
                  d03 = d03 || fac(36, 1.1, 3)[d36[,1],];
                  d06 = d06 || fac(36, 3.3, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 3 3 ** 9 6 ** 1') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 3.1, 2)[d36[,1],];
                  d03 = d03 || fac(36, 9, 3)[d36[,1],];
                  d06 = d06 || fac(36, 1.1, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 3 3 ** 2 6 ** 3') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 3, 2)[d36[,1],];
                  d03 = d03 || fac(36, 2.1, 3)[d36[,1],];
                  d06 = d06 || fac(36, 3.1, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 2 18 ** 1')
                  then %mktn22(36);

               else if index(lineage, '36 ** 1 > 2 ** 2 3 ** 5 6 ** 2') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 2, 2)[d36[,1],];
                  d03 = d03 || fac(36, 5, 3)[d36[,1],];
                  d06 = d06 || fac(36, 2.1, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 2 ** 1 3 ** 3 6 ** 3') then do;
                  b1  = 1;
                  d02 = d02 || fac(36, 1, 2)[d36[,1],];
                  d03 = d03 || fac(36, 3, 3)[d36[,1],];
                  d06 = d06 || fac(36, 3, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 3 ** 12 12 ** 1') then do;
                  b1  = ks((0:2)`, gh(12, 3), 3) + 1;
                  b2  = j(3, 1, 1) @ (1:12)`;
                  d03 = d03 || b1[d36[,1],];
                  d12 = d12 || b2[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 3 ** 7 6 ** 3') then do;
                  b1  = 1;
                  d03 = d03 || fac(36, 7, 3)[d36[,1],];
                  d06 = d06 || fac(36, 3.2, 6)[d36[,1],];
                  end;

               else if index(lineage, '36 ** 1 > 4 ** 1 9 ** 1') then do;
                  b1  = 1;
                  d04 = d04 || (j(9, 1, 1) @ (1:4)`)[d36[,1],];
                  d09 = d09 || ((1:9)` @ j(4, 1, 1))[d36[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d36) = 1 then free d36;
                  else do; file log; put 'ERROR: 36 ** 2.'; file print; end;
                  free b1 b2;
                  end;

               avail = avail - 1;
               end;

            *------35 runs, recode 35-level factors------;
            avail = max(0, ncol(d35) - any35);
            do while(avail);

               b1  = j(7, 1, 1) @ (1:5)`;
               b2  = (1:7)` @ j(5, 1, 1);
               d05 = d05 || b1[d35[,1],];
               d07 = d07 || b2[d35[,1],];

               if ncol(d35) = 1 then free d35;
               else d35 = d35[,2:ncol(d35)];
               avail = avail - 1;
               end;

            *------34 runs, recode 34-level factors------;
            avail = max(0, ncol(d34) - any34);
            do while(avail);

               b1  = j(17, 1, 1) @ (1:2)`;
               b2  = (1:17)` @ j(2, 1, 1);
               d02 = d02 || b1[d34[,1],];
               d17 = d17 || b2[d34[,1],];

               if ncol(d34) = 1 then free d34;
               else d34 = d34[,2:ncol(d34)];
               avail = avail - 1;
               end;

            *------33 runs, recode 33-level factors------;
            avail = max(0, ncol(d33) - any33);
            do while(avail);

               b1  = j(11, 1, 1) @ (1:3)`;
               b2  = (1:11)` @ j(3, 1, 1);
               d03 = d03 || b1[d33[,1],];
               d11 = d11 || b2[d33[,1],];

               if ncol(d33) = 1 then free d33;
               else d33 = d33[,2:ncol(d33)];
               avail = avail - 1;
               end;

            *------30 runs, recode 30-level factors------;
            avail = max(0, ncol(d30) - any30);
            do while(avail);

               if index(lineage, '30 ** 1 > 3 ** 1 10 ** 1') then do;
                  b1  = j(10, 1, 1) @ (1:3)`;
                  d03 = d03 || b1[d30[,1],];
                  b1  = (1:10)` @ j(3, 1, 1);
                  d10 = d10 || b1[d30[,1],];
                  end;

               else if index(lineage, '30 ** 1 > 5 ** 1 6 ** 1') then do;
                  b1  = j(6, 1, 1) @ (1:5)`;
                  d05 = d05 || b1[d30[,1],];
                  b1  = (1:6)` @ j(5, 1, 1);
                  d06 = d06 || b1[d30[,1],];
                  end;

               else if index(lineage, '30 ** 1 > 2 ** 1 15 ** 1') then do;
                  b1  = j(15, 1, 1) @ (1:2)`;
                  d02 = d02 || b1[d30[,1],];
                  b1  = (1:15)` @ j(2, 1, 1);
                  d15 = d15 || b1[d30[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d30) = 1 then free d30;
                  else do; file log; put 'ERROR: 30 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------28 runs, recode 28-level factors------;
            avail = max(0, ncol(d28) - any28);
            do while(avail);

               if index(lineage, '28 ** 1 > 2 ** 12 7 ** 1') then do;
                  b1  = fac(28, 12, 2);
                  d02 = d02 || b1[d28[,1],];
                  b1  = (1:7)` @ j(4, 1, 1);
                  d07 = d07 || b1[d28[,1],];
                  end;

               else if index(lineage, '28 ** 1 > 2 ** 2 14 ** 1')
                  then %mktn22(28);

               else if index(lineage, '28 ** 1 > 4 ** 1 7 ** 1') then do;
                  b1  = j(7, 1, 1) @ (1:4)`;
                  d04 = d04 || b1[d28[,1],];
                  b1  = (1:7)` @ j(4, 1, 1);
                  d07 = d07 || b1[d28[,1],];
                  end;

               else if index(lineage, '28 ** 1 > 2 ** 27') then do;
                  b1  = hadw(28, 0);
                  d02 = d02 || b1[d28[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d28) = 1 then free d28;
                  else do; file log; put 'ERROR: 28 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------27 runs, recode 27-level factors------;
            avail = max(0, ncol(d27) - any27);
            do while(avail);

               b1  = ks((0:2)`, gh(9, 3), 3) + 1;
               b2  = j(3, 1, 1) @ (1:9)`;
               d03 = d03 || b1[d27[,1],];
               d09 = d09 || b2[d27[,1],];

               if ncol(b1) then do;
                  if ncol(d27) = 1 then free d27;
                  else do; file log; put 'ERROR: 27 ** 2.'; file print; end;
                  free b1;
                  end;

               avail = avail - 1;
               end;

            *------26 runs, recode 26-level factors------;
            avail = max(0, ncol(d26) - any26);
            do while(avail);

               b1  = j(13, 1, 1) @ (1:2)`;
               b2  = (1:13)` @ j(2, 1, 1);
               d02 = d02 || b1[d26[,1],];
               d13 = d13 || b2[d26[,1],];

               if ncol(d26) = 1 then free d26;
               else d26 = d26[,2:ncol(d26)];
               avail = avail - 1;
               end;

            *------25 runs, recode 25-level factors------;
            avail = max(0, ncol(d25) - any25);
            do while(avail);

               b1  = ks((0:4)`, gh(5, 5), 5) + 1 || j(5, 1, 1) @ (1:5)`;
               d05 = d05 || b1[d25[,1],];

               if ncol(b1) then do;
                  if ncol(d25) = 1 then free d25;
                  else do; file log; put 'ERROR: 25 ** 2.'; file print; end;
                  free b1;
                  end;

               free b1;
               avail = avail - 1;
               end;

            *------24 runs, recode 24-level factors------;
            avail = max(0, ncol(d24) - any24);
            do while(avail);

               if index(lineage, '24 ** 1 > 2 ** 20 4 ** 1') then do;
                  b1  = twolev(h2 @ sortmat(had(12, 1), 1:2));
                  d02 = d02 || b1[d24[,1],(3:12) || (15:24)];
                  d04 = d04 || ((1:4)` @ j(6, 1, 1))[d24[,1],];
                  end;

               else if index(lineage, '24 ** 1 > 2 ** 13 3 ** 1 4 ** 1')
                  then do;
                  b1  = sortmat(had(12, 1), 1:2);
                  b2  = sortmat(fac(12, 4, 2) || (1:3)` @ j(4, 1, 1), 1);
                  d03 = d03 || (j(2, 1, 1) @ b2[,5])[d24[,1],];
                  b2  = twolev({-1 1}` @ b1) || j(2, 1, 1) @ b2[,1:4];
                  d04 = d04 || ((1:4)` @ j(6, 1, 1))[d24[,1],];
                  d02 = d02 || b2[d24[,1],(3:12) || (14:16)];
                  end;

               else if index(lineage, '24 ** 1 > 2 ** 12 12 ** 1') then do;
                  b1  = twolev({1 -1}` @ had(12, 1));
                  b2  = j(2, 1, 1) @ (1:12)`;
                  d02 = d02 || b1[d24[,1],];
                  d12 = d12 || b2[d24[,1],];
                  end;

               else if index(lineage, '24 ** 1 > 2 ** 11 4 ** 1 6 ** 1')
                  then do;
                  b1  = fac(24, 13, 0);
                  d02 = d02 || b1[d24[,1], 1:11];
                  d04 = d04 || b1[d24[,1], 12];
                  d06 = d06 || b1[d24[,1], 13];
                  end;

               else if index(lineage, '24 ** 1 > 3 ** 1 8 ** 1') then do;
                  b1  = (1:3)` @ j(8, 1, 1);
                  d03 = d03 || b1[d24[,1],];
                  b1  = j(3, 1, 1) @ (1:8)`;
                  d08 = d08 || b1[d24[,1],];
                  end;

               if ncol(d24) = 1 then free d24;
               else d24 = d24[,2:ncol(d24)];
               avail = avail - 1;
               end;

            *------22 runs, recode 22-level factors------;
            avail = max(0, ncol(d22) - any22);
            do while(avail);

               b1  = j(11, 1, 1) @ (1:2)`;
               b2  = (1:11)` @ j(2, 1, 1);
               d02 = d02 || b1[d22[,1],];
               d11 = d11 || b2[d22[,1],];

               if ncol(d22) = 1 then free d22;
               else d22 = d22[,2:ncol(d22)];
               avail = avail - 1;
               end;

            *------21 runs, recode 21-level factors------;
            avail = max(0, ncol(d21) - any21);
            do while(avail);

               b1  = j(7, 1, 1) @ (1:3)`;
               b2  = (1:7)` @ j(3, 1, 1);
               d03 = d03 || b1[d21[,1],];
               d07 = d07 || b2[d21[,1],];

               if ncol(d21) = 1 then free d21;
               else d21 = d21[,2:ncol(d21)];
               avail = avail - 1;
               end;

            *------20 runs, recode 20-level factors------;
            avail = max(0, ncol(d20) - any20);
            do while(avail);

               if index(lineage, '20 ** 1 > 2 ** 2 10 ** 1') then %mktn22(20);

               else if index(lineage, '20 ** 1 > 4 ** 1 5 ** 1') then do;
                  b1  = j(5, 1, 1) @ (1:4)`;
                  b2  = (1:5)` @ j(4, 1, 1);
                  d04 = d04 || b1[d20[,1],];
                  d05 = d05 || b2[d20[,1],];
                  end;

               else if index(lineage, '20 ** 1 > 2 ** 8 5 ** 1') then do;
                  b1  = fac(20, 8, 2);
                  b2  = {5 2 4 3 1}` @ j(4, 1, 1);
                  d02 = d02 || b1[d20[,1],];
                  d05 = d05 || b2[d20[,1],];
                  end;

               else if index(lineage, '20 ** 1 > 2 ** 19') then do;
                  b1  = had(20, 0);
                  d02 = d02 || b1[d20[,1],];
                  end;

               if ncol(d20) = 1 then free d20;
               else d20 = d20[,2:ncol(d20)];
               avail = avail - 1;
               end;

            *------18 runs, recode 18-level factors------;
            avail = max(0, ncol(d18) - any18);
            do while(avail);

               if index(lineage, '18 ** 1 > 3 ** 6 6 ** 1') then do;
                  b1  = ks((0:2)`, gh(6, 3), 3) + 1;
                  b2  = j(3, 1, 1) @ (1:6)`;
                  d03 = d03 || b1[d18[,1],];
                  d06 = d06 || b2[d18[,1],];
                  end;

               else if index(lineage, '18 ** 1 > 2 ** 1 9 ** 1') then do;
                  b1  = j(9, 1, 1) @ (1:2)`;
                  b2  = (1:9)` @ j(2, 1, 1);
                  d02 = d02 || b1[d18[,1],];
                  d09 = d09 || b2[d18[,1],];
                  end;

               if ncol(d18) = 1 then free d18 ;
               else d18 = d18[,2:ncol(d18)];
               avail = avail - 1;
               end;

            *------15 runs, recode 15-level factors------;
            avail = max(0, ncol(d15) - any15);
            do while(avail);

               b1  = j(5, 1, 1) @ (1:3)`;
               b2  = (1:5)` @ j(3, 1, 1);
               d03 = d03 || b1[d15[,1],];
               d05 = d05 || b2[d15[,1],];

               if ncol(d15) = 1 then free d15;
               else d15 = d15[,2:ncol(d15)];
               avail = avail - 1;
               end;

            *------14 runs, recode 14-level factors------;
            avail = max(0, ncol(d14) - any14);
            do while(avail);

               b1  = j(7, 1, 1) @ (1:2)`;
               b2  = (1:7)` @ j(2, 1, 1);
               d02 = d02 || b1[d14[,1],];
               d07 = d07 || b2[d14[,1],];

               if ncol(d14) = 1 then free d14;
               else d14 = d14[,2:ncol(d14)];
               avail = avail - 1;
               end;

            *------12 runs, recode 12-level factors------;
            avail = max(0, ncol(d12) - any12);
            do while(avail);
               free b1;

               i = index(lineage, '12 ** 1 > 2 ** 2 6 ** 1');
               j = index(lineage, '12 ** 1 > 3 ** 1 4 ** 1');
               k = index(lineage, '12 ** 1 > 2 ** 4 3 ** 1');
               l = index(lineage, '12 ** 1 > 2 ** 11');

               if i then %mktn22(12);

               else if j then do;
                  b1  = (1:3)` @ j(4, 1, 1);
                  b2  = j(3, 1, 1) @ (1:4)`;
                  d03 = d03 || b1[d12[,1],];
                  d04 = d04 || b2[d12[,1],];
                  end;

               else if k then do;
                  b1  = (1:3)` @ j(4, 1, 1);
                  b2  = fac(12, 4, 2);
                  d03 = d03 || b1[d12[,1],];
                  d02 = d02 || b2[d12[,1],];
                  end;

               else if l then do;
                  b1  = twolev(had(12, 1)[,2:12]);
                  d02 = d02 || b1[d12[,1],];
                  end;

               if ncol(b1) then do;
                  if ncol(d12) = 1 then free d12;
                  else d12 = d12[,2:ncol(d12)];
                  if      i then do; b1 = i - 1; b2 = i + 26; end;
                  else if j then do; b1 = j - 1; b2 = j + 26; end;
                  else if k then do; b1 = k - 1; b2 = k + 26; end;
                  else if l then do; b1 = l - 1; b2 = l + 20; end;
                  lineage = concat(substr(lineage, 1, b1), substr(lineage, b2));
                  free b1 b2;
                  end;

               avail = avail - 1;
               end;

            *------10 runs, recode 10-level factors------;
            avail = max(0, ncol(d10) - any10);
            do while(avail);

               b1  = (1:2)` @ j(5, 1, 1);
               b2  = j(2, 1, 1) @ (1:5)`;
               d02 = d02 || b1[d10[,1],];
               d05 = d05 || b2[d10[,1],];

               if ncol(d10) = 1 then free d10;
               else d10 = d10[,2:ncol(d10)];
               avail = avail - 1;
               end;

            *------9 runs, recode 9-level factors------;
            avail = max(0, ncol(d09) - any09);
            do while(avail);
               b1  = (ks((0:2)`, gh(3, 3), 3) || j(3, 1, 1) @ (0:2)`) + 1;
               d03 = d03 || b1[d09[,1],];
               if ncol(d09) = 1 then free d09;
               else d09 = d09[,2:ncol(d09)];
               avail = avail - 1;
               end;

            *------6 runs, recode 6-level factors------;
            avail = (any02 > ncol(d02) | any03 > ncol(d03)) #
                     max(0, ncol(d06) - any06);
            do while(avail);

               b1  = j(3, 1, 1) @ (1:2)`;
               b2  = (1:3)` @ j(2, 1, 1);
               d02 = d02 || b1[d06[,1],];
               d03 = d03 || b2[d06[,1],];

               if ncol(d06) = 1 then free d06;
               else d06 = d06[,2:ncol(d06)];
               avail = (any02 > ncol(d02) | any03 > ncol(d03)) #
                        max(0, ncol(d06) - any06);
               end;

            *------recode 128, 64, 32, 16, 8, 4-level factors------;
            start two(lineage, d128, d64, d32, d16, d08, d04, d02,
                      any128, any64, any32, any16, any08, any04, any02, h2);

               *------128 runs, recode 128-level factors------;
               avail = max(0, ncol(d128) - any128);
               do while(avail);

                  if index(lineage, '128 ** 1 > 2 ** 64 64 ** 1') then do;
                     b1  = j(2, 1, 1) @ (1:64)`;
                     d64 = d64 || b1[d128[,1],];
                     b1  = twolev({1 -1}` @ h2 @ h2 @ h2 @ h2 @ h2 @ h2);
                     d02 = d02 || b1[d128[,1],];
                     end;

                  else if index(lineage, '128 ** 1 > 4 ** 32 32 ** 1') then do;
                     b1  = 1;
                     d04 = d04 || (ks((0:3)`, gh(32, 4), 4) + 1)[d128[,1],];
                     d32 = d04 || (j(4, 1, 1) @ (1:32)`)[d128[,1],];
                     end;

                  else if index(lineage, '128 ** 1 > 8 ** 16 16 ** 1') then do;
                     b1  = 1;
                     d08 = d04 || (ks((0:7)`, gh(16, 8), 8) + 1)[d128[,1],];
                     d16 = d04 || (j(8, 1, 1) @ (1:16)`)[d128[,1],];
                     end;

                  if ncol(b1) then do;
                     if ncol(d128) = 1 then free d128;
                     else do; file log; put 'ERROR: 128 ** 2.'; file print; end;
                     free b1;
                     end;

                  avail = avail - 1;
                  end;

               *------64 runs, recode 64-level factors------;
               avail = max(0, ncol(d64) - any64);
               do while(avail);

                  if index(lineage, '64 ** 1 > 2 ** 32 32 ** 1') then do;
                     b1  = j(2, 1, 1) @ (1:32)`;
                     d32 = d32 || b1[d64[,1],];
                     b1  = twolev({1 -1}` @ h2 @ h2 @ h2 @ h2 @ h2);
                     d02 = d02 || b1[d64[,1],];
                     end;

                  else if index(lineage, '64 ** 1 > 2 ** 5 4 ** 17 8 ** 1')
                     then do;
                     b1  = fac(64, 1, 0);
                     d02 = d02 || b1[d64[,1],1:5];
                     d04 = d04 || b1[d64[,1],6:22];
                     d08 = d08 || b1[d64[,1],23];
                     end;

                  else if index(lineage, '64 ** 1 > 2 ** 5 4 ** 10 8 ** 4')
                     then do;
                     b1  = fac(64, 2, 0);
                     d02 = d02 || b1[d64[,1],1:5];
                     d04 = d04 || b1[d64[,1],6:15];
                     d08 = d08 || b1[d64[,1],16:19];
                     end;

                  else if index(lineage, '64 ** 1 > 4 ** 16 16 ** 1') then do;
                     b1  = j(4, 1, 1) @ (1:16)`;
                     d16 = d16 || b1[d64[,1],];
                     b1  = ks((0:3)`, gh(16, 4), 4) + 1;
                     d04 = d04 || b1[d64[,1],];
                     end;

                  else if index(lineage, '64 ** 1 > 4 ** 14 8 ** 3') then do;
                     b1  = fac(64, 3, 0);
                     d04 = d04 || b1[d64[,1],1:14];
                     d08 = d08 || b1[d64[,1],15:17];
                     end;

                  else if index(lineage, '64 ** 1 > 4 ** 7 8 ** 6') then do;
                     b1  = fac(64, 4, 0);
                     d04 = d04 || b1[d64[,1],1:7];
                     d08 = d08 || b1[d64[,1],8:13];
                     end;

                  else if index(lineage, '64 ** 1 > 8 ** 9') then do;
                     b1  = ks((0:7)`, gh(8, 8), 8) + 1 || j(8, 1, 1) @ (1:8)`;
                     d08 = d08 || b1[d64[,1],];
                     end;

                  if ncol(b1) then do;
                     if ncol(d64) = 1 then free d64;
                     else do; file log; put 'ERROR: 64 ** 2.'; file print; end;
                     free b1;
                     end;

                  avail = avail - 1;
                  end;

               *------32 runs, recode 32-level factors------;
               avail = max(0, ncol(d32) - any32);
               do while(avail);

                  if index(lineage, '32 ** 1 > 2 ** 16 16 ** 1') then do;
                     b1  = j(2, 1, 1) @ (1:16)`;
                     d16 = d16 || b1[d32[,1],];
                     b1  = twolev({1 -1}` @ h2 @ h2 @ h2 @ h2);
                     d02 = d02 || b1[d32[,1],];
                     end;

                  else if index(lineage, '32 ** 1 > 4 ** 8 8 ** 1') then do;
                     b1  = j(4, 1, 1) @ (1:8)`;
                     d08 = d08 || b1[d32[,1],];
                     b1  = ks((0:3)`, gh(8, 4), 4) + 1;
                     d04 = d04 || b1[d32[,1],];
                     end;

                  if ncol(b1) then do;
                     if ncol(d32) = 1 then free d32;
                     else do; file log; put 'ERROR: 32 ** 2.'; file print; end;
                     free b1;
                     end;

                  avail = avail - 1;
                  end;

               *------16 runs, recode 16-level factors------;
               avail = max(0, ncol(d16) - any16);
               do while(avail);

                  i = index(lineage, '16 ** 1 > 2 ** 8 8 ** 1');
                  j = index(lineage, '16 ** 1 > 4 ** 5');

                  if i then do;
                     b1  = j(2, 1, 1) @ (1:8)`;
                     d08 = d08 || b1[d16[,1],];
                     b1  = twolev({1 -1}` @ h2 @ h2 @ h2);
                     d02 = d02 || b1[d16[,1],];
                     end;

                  else if j then do;
                     b1  = j(4, 1, 1) @ (1:4)`;
                     d04 = d04 || b1[d16[,1],];
                     b1  = ks((0:3)`, gh(4, 4), 4) + 1;
                     d04 = d04 || b1[d16[,1],];
                     end;

                  if ncol(b1) then do;
                     if ncol(d16) = 1 then free d16;
                     else d16 = d16[,2:ncol(d16)];
                     if      i then do; b1 = i - 1; b2 = i + 26; end;
                     else if j then do; b1 = j - 1; b2 = j + 19; end;
                     lineage = concat(substr(lineage, 1, b1),
                                      substr(lineage, b2));
                     free b1 b2;
                     end;

                  avail = avail - 1;
                  end;

               *------8 runs, recode 8-level factors------;
               avail = max(0, ncol(d08) - any08);
               do while(avail);
                  b1  = twolev((h2 @ h2 @ h2)[,{2 4 6 8}]);
                  b2  = (1:4)` @ j(2, 1, 1);
                  d02 = d02 || b1[d08[,1],];
                  d04 = d04 || b2[d08[,1],];
                  if ncol(d08) = 1 then free d08;
                  else d08 = d08[,2:ncol(d08)];
                  avail = avail - 1;
                  end;

               *------4 runs, recode 4-level factors------;
               avail = (any02 > ncol(d02)) # max(0, ncol(d04) - any04);
               do while(avail);
                  b1  = twolev((h2 @ h2)[,2:4]);
                  d02 = d02 || b1[d04[,1],];

                  if ncol(d04) = 1 then free d04;
                  else d04 = d04[,2:ncol(d04)];
                  avail = (any02 > ncol(d02)) # max(0, ncol(d04) - any04);
                  end;

               finish;

            *------recode 128, 64, 32, 16, 8, 4-level factors------;
            call two(lineage, d128, d64, d32, d16, d08, d04, d02,
                     any128, any64, any32, any16, any08, any04, any02, h2);

            free b1 b2 b3 b4 b5;

            *================ Done Expanding, Work on Hadamards ===============;
            need = (any02 | any04 | any08 | any16 | any32 | any64 | any128);

            if need then do;
               need = 0;
               %do j = 2 %to &maxlevneed;
                  need = need + ncol(%if &j > 9 %then d&j;%else d0&j;);
                  %end;
               need = ^need;
               end;

            *------Paley method Hadamards------;
            %if &hadflag ne %then %do;
               if need & ncol(h) = 0 then h =
                  %if &hadflag eq hadamard1 %then hadp1(&hadsize, 1);
                                            %else hadp2(&hadsize, 1);;
               %end;

            *------Williamson Hadamard construction------;
            if &hadsize & need & ncol(h) = 0 then h = hadw(&hadsize, 1);

            *------make the larger Hadamard sizes (2**n) * hadsize------;
            if ncol(h) then do;
               v = j(1, &tabsize, 1);  v[1] = 0;

               %if &flow %then %do;
                  file log;
                  put "** Hadamard &hadsize special code.";
                  file print;
                  %end;

               if nrow(h) < &tabsize then do;
                  h = sortmat(h, 1:2);
                  do nh = 0 to 100 while(nrow(h) < &tabsize); h = h2 @ h; end;

                  %if &flow or &flowinit %then %do;
                     file log;
                     put "** Hadamard &hadsize -> &tabsize, "
                         (compress(char(nh))) " doublings.";
                     file print;
                     %end;

                  *------extract 4, 8, 16, 32, 64, 128------;
                  if nh >= 6 & (index(lineage, '128 ** 1') |
                     any08 | any16 | any32 | any64 | any128) then do;
                     d128 = (1:128)` @ j(&tabsize / 128, 1, 1);
                     v[&hadsize # (0:63) @ {1 1} + j(1, 64, 1) @ {1 2}] = 0;
                     end;
                  else if nh = 5 & (index(lineage, '64 ** 1') |
                     any08 | any16 | any32 | any64) then do;
                     d64 = (1:64)` @ j(&tabsize / 64, 1, 1);
                     v[&hadsize # (0:31) @ {1 1} + j(1, 32, 1) @ {1 2}] = 0;
                     end;
                  else if nh = 4 & (index(lineage, '32 ** 1') |
                     any08 | any16 | any32) then do;
                     d32 = (1:32)` @ j(&tabsize / 32, 1, 1);
                     v[&hadsize # (0:15) @ {1 1} + j(1, 16, 1) @ {1 2}] = 0;
                     end;
                  else if nh = 3 & (index(lineage, '16 ** 1') |
                     any08 | any16) then do;
                     d16 = (1:16)` @ j(&tabsize / 16, 1, 1);
                     v[&hadsize # (0:7) @ {1 1} + j(1, 8, 1) @ {1 2}] = 0;
                     end;
                  else if nh = 2 & (index(lineage, ' 8 ** 1') |
                     any08) then do;
                     d08 = (1:8)` @ j(&tabsize / 8, 1, 1);
                     v[&hadsize # (0:3) @ {1 1} + j(1, 4, 1) @ {1 2}] = 0;
                     end;
                  else if nh & (any04 | any03) then do;
                     need = min(max(0, any04 + any03,
                                max(0, floor((ncol(h) - 1 - any02) / 3))));
                     do i = 1 to min(2 ## nh - 1, &hadsize - 1, need);
                        d04 = d04 || (2 # (h[,i+1] = 1) +
                                     (h[,i # &hadsize + 1] = 1) + 1);
                        v[(i+1) || (i#&hadsize+1) || (i#&hadsize+i+1)] = 0;
                        end;
                     end;
                  h = h # h[1,];
                  end;

               *------make the two-level factors------;
               d02 = twolev(h[,loc(v)]);
               free h;

               *------recode 128, 64, 32, 16, 8, 4-level factors------;
               call two(lineage, d128, d64, d32, d16, d08, d04, d02,
                        any128, any64, any32, any16, any08, any04, any02, h2);
               end;

            *------flag the imbalanced initializations-------;
            balanced = j(1, &maxlevneed, 0);
            %do j = 1 %to &maxlevneed;
               %if &j > 9 %then %let jj = &j; %else %let jj = 0&j;
               %if &debugextra %then %do;
                  if any&jj ^= ncol(d&jj) then do;
                     print (ncol(d&jj) - any&jj) "Extra &j-Level Col" any&jj;
                     show d&jj;
                     end;
                  %end;
               balanced[&j] = ncol(d&jj);
               %if &nrunsintab ne &n %then %do; balanced[&j] = 0; %end;
               %else %if &factexnlevs and &powerof %then %do;
                  logj = log(&j) / log(&powerof);
                  if abs(logj - round(logj)) >= 1e-8 then balanced[&j] = 0;
                  %end;
               %end;

            *------try to do some coding down: 6->5, 6->4, etc.------;
            start mymod(x, to);
               from = max(x);
               res = mod(x - 1, to) + 1;
               if from = 6 & to = 4 | from = 5 & to = 3 then
                  res = res # (x ^= from) + to # (x = from);
               return(res);
               finish;

            %do j = 6 %to 3 %by -1;
               %do jj = %eval(&j-1) %to 2 %by -1;
                  *------make &jj-level factors from &j-levels------;
                  want  = max(0, any0&jj - ncol(d0&jj));
                  avail = min(want, max(0, ncol(d0&j) - any0&j));
                  if avail then do;
                     firstavail = ncol(d0&j) - avail + 1;
                     vec        = firstavail : ncol(d0&j);
                     d0&jj      = d0&jj || mymod(d0&j[,vec], &jj);
                     if avail >= ncol(d0&j) then free d0&j;
                     else d0&j = d0&j[, 1 : (firstavail - 1)];
                  end;
                  %end;
               %end;

            %if &flow %then %do;
               file log;
               reset nocenter;
               %do j = 2 %to &maxlevneed;
                  %if &j > 9 %then %let jj = &j; %else %let jj = 0&j;
                  if ncol(d&jj) then put "** d&jj: " (nrow(d&jj)) " rows,"
                                                     (ncol(d&jj)) " cols.";
                  %end;
               reset center;
               file print;
               %end;

            finish;

         *================ Rest of the IML Function Definitions ===============;

         *-----get OA initialization for the relevant factors------;
         start getinit(xmat) global(raninit, m, nints, maxint, int, lineage,
                                    facts, nt, inittype, balanced, seed
                                    %do j = 2 %to &maxlevneed;
                                       , %if &j > 9 %then d&j; %else d0&j;
                                       %end;
                                    );

            *------figure out how many init cols we have available------;
            ini = j(1, &maxlevneed, 0); /* n factors for each of the levels */
            inn = j(1, &maxlevneed, 0); /* n factors actually available     */

            %do j = 2 %to &maxlevneed;
               %if &j > 9 %then %let jj = &j; %else %let jj = 0&j;
               inn[&j] = ncol(d&jj);
               if &unbalanced = 0.0 |
                  nt > &unbalanced %if &pass = 1 %then + &canditer; then
                  inn[&j] = min(inn[&j], balanced[&j]);
               %end;

            *------figure out what rows to extract, fill in------;
            %let k = ;
            %let i = ;
            %if &n gt &nrunsintab %then
               %let i = %eval(&n + 1 - &nrunsintab) : &n;
            %else %if &nrunsintab gt &n %then %do;
               rowvec = j(1, &n, 0);
               call ranuni(seed, rowvec);
               rowvec = rank(rowvec)[1:&n];
               %let k = rowvec;
               %end;

            *------initialize columns of xmat, flag them as not random------;
            %if &nooadups %then %do;
               reverse = 0; dupseed = &seed;
tryagain:      if reverse > 1 then do;
                  %do j = 2 %to &maxlevneed;
                     %if &j > 9 %then %let jj = &j; %else %let jj = 0&j;
                     if inn[&j] then do;
                        j = j(ncol(d&jj), 1, 0);
                        call ranuni(dupseed, j);
                        d&jj = d&jj[,rank(j)];
                        end;
                     %end;
                  end;
               %end;

            do j = 1 to m;
               %do j = 2 %to &maxlevneed;
                  %if &j > 9 %then %let jj = &j; %else %let jj = 0&j;
                  %if &j > 2 %then else;
                  if facts[j] = &j & ini[&j] < inn[&j] then do;
                     ini[&j] = ini[&j] + 1;
                     %if &nooadups %then %do;
                        if reverse then
                           xmat[&i, j] = d&jj[&k, inn[&j] + 1 - ini[&j]];
                        else
                        %end;
                     xmat[&i, j] = d&jj[&k, ini[&j]];
                     if ini[&j] <= balanced[&j] then raninit[j] = 0;
                     else do;
                        inittype = 'Tab,Unb';
                        raninit[j] = -1;
                        end;
                     end;
                  %end;
               end;

            %if &nooadups %then %do;
               j = 1:m;
               call sortndx(i, xmat, j);
               j = uniqueby(xmat, j, i);
               if reverse <= 10 then do;
                  reverse = reverse + 1;
                  if nrow(i) ^= nrow(j) then goto tryagain;
                  else if reverse > 1 then put '*** It Worked *** ' reverse;
                  end;
               else if reverse then do;
                  if nrow(i) ^= nrow(j) then put '*** Duplicates ***';
                  end;
               %end;

            *------initialization: Unbalanced, Balanced, Random------;
            %if &flow or &flowinit %then %do;
               if nt = ^(&initdata) then do;
                  file log;
                  reset nocenter;
                  put '** lineage: ' lineage '.';
                  reset center;
                  file print;
                  end;
               file log;
               reset nocenter;
               inistart = 1; inistop = 500;
               do until(inistr = ' ');
                  inistr = ' ';
                  if inistart <= ncol(raninit) then do;
                     inivec = inistart : min(inistop, ncol(raninit));
                     inistr = rowcatc({'u' 'b' 'r'}[raninit[inivec] + 2]`);
                     end;
                  if inistr ^= ' ' then put '** ' inistr;
                  inistart = inistart + 500; inistop = inistop + 500;
                  end;
               free inistart inistop inistr inivec;
               reset center;
               file print;
               %end;

            *------initial design part OA, part random?------;
            if "&destype" = 'random' then inittype = 'Ran';
            else if any(raninit = 1) then inittype = concat(inittype, ',Ran');
            j = loc(raninit = -1);
            if ncol(j) then raninit[j] = 1;

            *------let factors that have interactions change------;
            do i = 1 to nints; do j = 1 to maxint;
              k = int[i, j];
              if k then raninit[k] = 1;
              end; end;
            finish;
         %end;

         *------build coded main-effects design matrix------;
         start codedes;
            d = j(n, 1, 1);     /* design matrix, starting with an intercept */
            c = 2;
            do j = 1 to m;
               t = orthogcode(xmat[,j], facts[j]);
               s1[j] = c;       /* index into coded design matrix */
               c = c + ncol(t);
               s2[j] = c - 1;   /* index into coded design matrix */
               d = d || t;
               end;

            *------add any interactions to design matrix------;
            do j = 1 to nints;
               jj = int[j, 1];
               t = d[,s1[jj] : s2[jj]];
               do k = 2 to maxint while(int[j, k] > 0);
                  jj = int[j, k];
                  t = hdir(t, d[,s1[jj] : s2[jj]]);
                  end;
               d = d || t;
               s1[j+m] = c;
               c = c + ncol(t);
               s2[j+m] = c - 1;
               end;

            npars = ncol(d);
            finish;

      *------report on intermediate progress------;
      start report;
         /* nstep = 1 - step = Initial
                    2          Start
                    3          End
                    4          point no violations realized
                    5          restriction report
                    0          intermediate results */
         scaleddet = 100 # (maxdet ## (1 / npars));
         if scaleddet > 101 then do;
            put "ERROR: Extreme numerical problems detected.";
            if &refreshat > 1 then
               put "NOTE: A smaller REFRESHAT=&refreshat value may be needed.";
            if &ridge ^= 1e-7 then
               put "NOTE: A different RIDGE=&ridge value may be needed.";
            call symput('abort', '1');
            scaleddet = .;
            end;
         str2      = '              ';

         if improvement /* same or better */
            %if &restrictflag or &dupflag %then & ^ restricting; then do;
            if improvement = 2 /* better */ then targetdet = scaleddet;
            if nstep ^= 3 then str2 = char(scaleddet, 14, 4);
            end;

         %if &history %then %do;
            if nstep = 1 | nstep = 2 | nstep = 4 | nstep = 5
               then str2 = concat(str2, '  ', inittype);
            if nstep ^= 5 & col = ' ' then col = concat('    ', step);
            else                           col = concat(step, col);

            %if &restrictflag or &dupflag %then %do;
               if (try - lastprob) <= n & str2 = ' ' & nstep = 3 then
                  str2 = '                Violations';
               if ^restricting | str2 ^= ' ' | nstep then do;
               %end;

            *------iteration history header------;
            if lnum = 0 | lnum >= &ps | (nstep = 2 & lnum >= (&ps - 3)) then do;
               print /
                   %if &pass = 1 %then       'Algorithm Search History';
                   %else %if &pass = 2 %then 'Design Search History';
                   %else %if &pass = 3 %then 'Design Refinement History';;
               put /
                   +&ind '                        Current          Best' /
                   +&ind 'Design    Row,Col  D-Efficiency  D-Efficiency  Notes'/
                   +&ind
                   '----------------------------------------------------------';

               lnum = 5;
               end;

            put +&ind nt 6. col scaleddet 14.4 str2;
            lnum = lnum + 1;
            if (nstep = 1 | nstep = 3) & lnum < (&ps - 2)
               then do; put; lnum = lnum + 1; end;
            %if &restrictflag or &dupflag %then %do; end; %end;
            %end;
         finish;

      *========= Start Main Body of the Computations, DO Enables GOTO =========;
      do;

      if &somenotes & &optabort = 2 then do;
         file log;
         put 'NOTE: The candidate-set initialization failed,' /
             '      but the MKTEX macro is continuing.';
         file print;
         end;

      %if &somenotes and &tableflag ne 0 %then %do;
         if symget('tablenote') = '0' then do;
            file log;
            put "NOTE: Generating the &destype design, n=&nrunsintab..";
            call symput('tablenote', '1');
            file print;
            end;
         %end;

      %if not &nointeract %then %do; int = {&interactmat}; %end;
      facts        = { &_mktlist };        n            = &n;
      m            = ncol(facts);
      i            = %if &order eq col %then m; %else 1;;
      n2           = 2 # n # i;            n3           = 3 # n # i;
      n5           = 5 # n # i;            n10          = 10 # n # i;
      n100         = 100 # n # i;          impossmess   = 0;
      nints        = nrow(int);
      maxint       = ncol(int);            rananneal    = 0;
      ei1          = j(max(facts), m, 0);  xmat         = j(n, m, 0);
      s1           = j(1, m + nints, 0);   s2           = j(1, m + nints, 0);
      sqrtninv     = 1 / sqrt(n);          ninv         = 1 / n;
      bestdet      = 0;                    force        = 0;
      badb4        = 0;                    badnow       = 0;
      indf         = facts - 2;            imat         = i(2);
      anneal       = 0;                    try          = 0;
      targetdet    = 0;                    listprob     = 0;
      improvement  = 0;                    superseed    = &seed;
      canditersleft= &canditer;            tabitersleft = &tabiter;
      bestmethod   = '   ';                ntimes       = 0;
      lnum         = 0;                    ntimesiter   = 0;
      largestop    = 0;                    maxtime      = &maxtime * 60;
      printnt      = 0;                    doitagain    = 0;
      resmet       = 0;                    seed         = 0;
      doinit       = 0;
      mutiter      = &mutiter - (&optabort > 0);
      anniter      = &anniter - (&optabort > 0);
      repeat       = %scan(&repeat, 1, %str($));
      startrepeat  = %scan(&repeat, 2, %str($));
      stoprepeat   = %scan(&repeat, 3, %str($));
      repeat2      = n2 + startrepeat;
      repeat3      = n3 + startrepeat;
      repeat5      = n5 + startrepeat;

      lineage      = symget(cshape('lineage',1,1,%length(&lineage)+7,' '));
      %if &ordcoln ne %then %do; ordcoln = j(1, &ordcoln, 0); %end;
      %if &ridgemsg %then %do; __facts = facts; %end;
      %do i = 1 %to &exchange; j&i = 1; %end;

      %if &order eq matrix %then %do;
         use &exchangedat; read all into exmat; nexs = nrow(exmat);
         %end;

      *-----store orthogonal coding for each level of each factor------;
      c = 1;
      do j = 1 to m;
         t = orthogcode((1:facts[j])`, facts[j]);
         i = ncol(t);
         essmat = essmat || shape(t, 1); /* essense matrix              */
         do k = 1 to facts[j];
            ei1[k, j] = c;               /* indices into essense matrix */
            c = c + i;
            end;
         end;

      *------set up for OA initialization------;
      %if &tableflag ne 0 %then %do;
         run inittab(facts);
         %end;

      *======= Do Iterations, Possibly with an nt=0 Initialization Pass =======;

      algortime = datetime();
      do nt = ^(&initdata) to &maxiter;

         call ranuni(superseed, seed); seed = ceil(1e7 * seed);

         *------read initial design------;
         inittype = ' ';   method = 'Ran';  raninit = j(1, m, 1);
         %if &initdata %then %do;

            if &readinit then do;
               inittype = 'Ini';
               use &init %if &rowsfixed %then (drop=&fixed);;
               read all into xmat;
               close &init;

               *------when some elements are fixed in advance------;
               %if &elefixed %then %do;
                  canchange = (xmat <= 0);
                  %if &anyneg %then %do; reinit = (xmat < 0) # abs(xmat); %end;
                  xmat = abs(xmat);
                  doinit = 1;
                  %end;

               *------when some rows are fixed in advance------;
               %if &rowsfixed %then %do;
                  use &init(keep=&fixed); read all into fixed; close &init;
                  %end;
               end;

            else do;
               if doinit then inittype = 'Ini';
               else           inittype = 'Pre';
               if ncol(bestdes) then xmat = bestdes;
               end;
            %end;

         %*------use random init design------;
         %if &initrand %then %do;
            if nt > 0 then do;
               inittype = 'Ran';
               if ncol(rand) = 0 then rand = j(n, m, 0);
               call ranuni(seed, rand);
               do j = 1 to m;
                  xmat[rank(rand[,j]), j] = shape(1:facts[j], n, 1);
                  end;

               *------use OPTEX design------;
               %if &optflag %then %do;
                  if canditersleft then do;
                     inittype = 'Can'; method = 'Can';
                     canditersleft = canditersleft - (nt > 0);
                     use __tmpxopt; read all into xmat; close __tmpxopt;
                     end;
                  %end;

               *------use OA------;
               %if &tableflag ne 0 %then %do;
                  if inittype = 'Ran' & tabitersleft then do;
                     inittype = 'Tab'; method = 'Tab';
                     tabitersleft = tabitersleft - (nt > 0);
                     run getinit(xmat);

                     *------output prematurely and quit if called for------;
                     %if &justinit %then %do;
                        vname = 'x1' : rowcatc('x' || char(m));
                        create &out from xmat[colname=vname];
                        append from xmat;
                        call symput('pass', '4');
                        goto done;
                        %end;
                     end;
                  %end;
               end;
            %end;

         *------randomly mutate the initial design, initialize------;
         %if &mutateflag or &elefixed %then %do;
            if ncol(rand) = 0 then rand = j(n, m, 0);
            if doinit then do;
               c = canchange;
               inittype = concat(inittype, ',Ran');
               end;
            else if nt >= mutiter & &mutate > 0 then do;
               if ^index(inittype, 'Mut')
                  then inittype = concat(inittype, ',Mut');
               call ranuni(seed, rand);
               c = rand < &mutate;
               end;
            if doinit | (nt >= mutiter & &mutate > 0) then do;
               call ranuni(seed, rand);
               xmat = xmat # (^c) + c # ceil(rand # facts);
               %if &anyneg %then %do;
                  if doinit then xmat = xmat # (reinit = 0) + reinit;
                  %end;
               end;
            doinit = &resetinit;
            free rand c;
            %end;

         %if &annealflag %then %do;
            if nt >= anniter & ^ index(inittype, 'Ann')
               then inittype = concat(inittype, ',Ann');
            %end;

         *------build coded design matrix------;
         run codedes;

         *------get set up for iterations------;
         changed      = 0;              lastprob  = 0;
         lastforce    = 0;              numchange = 0;
         i            = .;              updated   = 0;
         heatitup     = 0;              ridge     = &ridge # i(npars);
         partdet      = 0;

         %if &annealflag %then %do; anneal = &anneal; %end;

         *------check for restriction violations in initial design------;
         %if &restrictflag or &dupflag %then %do;
            restricting = 0; %do i = 1 %to &exchange; j&i = 1; %end;
            do i = 1 to n while(^ restricting);
               xsubi = xmat[i,];
               do j1 = 1 to m while(^ restricting);
                  %if &restrictflag %then %do;
                     if badness(xsubi) then restricting = 1;
                     %end;
                  end;
               %if &dupflag %then %do;
                  t = (shape(xsubi, n, m) = xmat);
                  t[i,] = j(1, m, 0);
                  if sum(t[,+] = m) then restricting = 1;
                  free t;
                  %end;
               end;
            restrictsmet = 0;
            %end;

         *------determinant for initial design------;
         if ncol(bestdes) = 0 then do;
            xpx        = d`;
            xpx        = (xpx * xpx`) # ninv + ridge;
            maxdet     = max(det(xpx), 0);
            maxdetplus = maxdet # &over1; maxdetminus = maxdet # &almost1;
            %if &restrictflag or &dupflag
               %then %do; if nt = 0 | ^ restricting then do; %end;

            improvement = (maxdetminus > bestdet) + (maxdetplus > bestdet);
            if improvement = 1 then do;
               if nt ^= ntimesiter then ntimes = ntimes + 1;
               ntimesiter = nt;
               end;
            if improvement = 2
               %if &restrictflag or &dupflag %then & ^restricting;
               then do;
               if nt = 0 then bestdet = maxdet;
               bestdes = xmat; bestmethod = method; ntimes  = 0;
               end;
            if nt = 0 then do;
               nstep = 1; step = 'Initial'; col = '    '; run report;
               end;
            %if &restrictflag or &dupflag %then %do; end; %end;
            end;

         *=========== Otherwise Loop Over All Rows at Most 100 Times ==========;
         n100 = 100 # n %if &order eq col %then # m;;
         if nt then do try = 0 to n100 while(((try - changed) <= n2
            %if %nrbquote(&mintry) ne n %then | try <= &mintry; )
            %if &target ne 100 %then & ((targetdet + 1e-10) <= &target);
            %if &largedesign %then & ^ largestop; );
            i = mod(try, n) + 1;      /* row number */

            *------if we think we just met restrictions------;
            %if &restrictflag or &dupflag %then %do;
               restricting = (try - %if &accept %then lastforce;
                             %else lastprob;) <= n;
               if restrictsmet = 0 & restricting = 0 then do;
                  numchange    = 0;
                  restrictsmet = 2;

                  *------it may be that restrictions cannot be met------;
                  if (try - lastprob) <= n then restrictsmet = 3;
                  end;

               %end;

            *------every once in a while, refresh determinant, inverse------;
            if mod(numchange, &refreshat) = 0 then do;
               xpx     = d`;
               xpx     = (xpx * xpx`) # ninv + ridge;
               xpxinv  = sweep(xpx, 1 : npars);
               partdet = max(det(xpx), 0);

               if try = 0
                  %if &restrictflag or &dupflag %then | restrictsmet > 1;
                  then do;

                  maxdet      = partdet;
                  maxdetplus  = maxdet # &over1;
                  maxdetminus = maxdet # &almost1;
                  improvement = (maxdetminus > bestdet) +
                                (maxdetplus  > bestdet);
                  if improvement = 2
                     %if &restrictflag or &dupflag %then & ^ restricting;
                     then do;
                     bestdet = maxdet; bestdes = xmat; bestmethod = method;
                     ntimes  = 0;
                     end;

                  if try = 0 then do;
                     nstep = 2; step = '  Start'; col = '    '; run report;
                     end;

                  %if &restrictflag or &dupflag %then %do;
                     if restrictsmet > 1 then do;
                        nstep = 4; step = char(i, 7.); col = char(1, 4, 0);
                        if restrictsmet = 2 then inittype = 'Conforms';
                        else                     inittype = 'Acceptable';
                        run report;
                        restrictsmet = 1;
                        end;
                     %end;
                  if ((inittype = 'Can' & &pass = 1) |
                      ((targetdet + 1e-10) > &target))
                     %if &restrictflag or &dupflag %then & ^ restricting;
                     %if %nrbquote(&mintry) ne n %then & try > &mintry;
                     then do;
                     nstep = 3; step = '    End'; col = '    '; run report;
                     goto enditer;
                     end;
                  end;
               end;

            *------remove the ith row from the design, determinant update------;
            dsubi = d[i,];                /* ith row        */
            q     = dsubi # sqrtninv;
            vx    = xpxinv * q`;
            q     = 1 - q * vx;
            %if &ltfr %then %do;
               if q <= 0 then do;
                  file log;
                  put 'ERROR: Determinant and inverse update error.';
                  if &somenotes then
                     put "NOTE: Try a larger value for RIDGE= than &ridge..";
                  call symput('abort', '1');
                  goto done;
                  end;
               %end;
            invstar = xpxinv + vx * vx` / q;
            detstar = partdet * q;

            *======= Change Levels in One or More Cols, Start Exchanges =======;
            %do ex = 1 %to &exchange;

               *------with restrictions, do not always start with col 1------;
               %if &order eq random %then %do;
                  %if &partinc and &ex eq 2 %then %do;
                     partinc = &partinc;
                     j2 = mod(j1 - 1, partinc) + 1;
                     xj2 = xmat[i, j2]; ok = 1;
                     do j2 = j2 + partinc to m by partinc while(ok);
                        if xmat[i, j2] ^= xj2 then ok = 0;
                        end;
                     if ^ok then do; partinc = m; j2 = j1; end;
                     else j2 = mod(j1 - 1, partinc) + 1;
                     do j2 = j2 to m by partinc;
                     %end;
                  %else %do;
                     j0 = 1;
                     j&ex.vec = j%eval(&ex-1):m; r&ex.vec = j&ex.vec;
                     call ranuni(seed, r&ex.vec);
                     r&ex.vec[rank(r&ex.vec)] = j&ex.vec;
                     do jj&ex = 1 to ncol(r&ex.vec);
                        j&ex = r&ex.vec[jj&ex];
                     %end;
                  %end;

               %else %if &order eq ranseq %then %do;
                  call ranuni(seed, j&ex);
                  %if &ex eq 1 %then %do;
                     j&ex = floor(m # j&ex);
                     do jj&ex = 1 to m;
                        j&ex = j&ex + 1;
                        if j&ex > m then j&ex = 1;
                     %end;
                  %else %do;
                     jj&ex = ceil(m # j&ex);
                     do j&ex = jj&ex to jj&ex;
                     %end;
                  %end;

               %else %if &order eq col %then %do;
                  call ranuni(seed, ordcoln);
                  do jj&ex = 1 to &ordcoln;
                     j&ex = ceil(ordcoln[jj&ex] # m);
                  %end;

               %else %if &order eq matrix %then %do;
                  %if &ex eq 1 %then %do;
                     do jj1 = 1 to nexs;
                        j1 = exmat[jj1,&ex];
                     %end;
                  %else %do;
                      jj&ex = exmat[jj1,&ex];
                      %if &missex %then %do;
                        if nmiss(jj&ex) then do;
                           call ranuni(seed, jj&ex);
                           jj&ex = ceil(m # jj&ex);
                           end;
                        %end;
                     do j&ex = jj&ex to jj&ex;
                     %end;
                  %end;

               %else %do; /* order eq sequential */
                  j0 = 0;
                  do j&ex = j%eval(&ex - 1) + 1 to m;
                  %end;

               jvec&ex = s1[j&ex] : s2[j&ex];
               esscols&ex = indf[j&ex];
               %end;

            *------loop over levels------;
            %do ex = 1 %to &exchange;
               sks&ex = 1; ske&ex = facts[j&ex]; k&ex = 1;
               %do k = 1 %to &ex - 1;
                  if j&k = j&ex then do; sks&ex = k&k; ske&ex = k&k; end;
                  %end;
               %if &tableflag ne 0 and not (&restrictflag or &dupflag)
                   %then           if   raninit[  j&ex] then;
               %if &elefixed %then if canchange[i,j&ex] then;
               do k&ex = sks&ex to ske&ex;
                  %if &exchange eq 1 and &tableflag ne 0 and
                     not (&restrictflag or &dupflag)
                     %then %do; if xmat[i, j1] ^= k1 then do; %end;
               %end;

            *------if we might exchange for reasons other than------;
            *------efficiency improvement, we need a forced   ------;
            *------exchange flag                              ------;
            %if &restrictflag or &dupflag or &annealflag
                %then %do; force = 0; %end;

            *------set up for forced exchanges------;
            %if &restrictflag or &dupflag
              %then %do; xsubi = xmat[i,]; %end;

            *------restriction violations before exchange?------;
            %if &restrictflag %then %do; badb4 = badness(xsubi); %end;

            *------duplicates before exchange?------;
            %if &dupflag %then %do;
               t = (shape(xsubi, n, m) = xmat);
               t[i,] = j(1, m, 0);
               dupsb4 = sum(t[,+] = m);
               free t;
               %end;

            *------do exchange(s)------;
            %if &restrictflag or &dupflag %then %do;
               %do ex = 1 %to &exchange;
                  xsubi[j&ex] = k&ex;
                  %end;
               %end;

            *------improvements on restrictions?------;
            %if &restrictflag %then %do;
               badnow = badness(xsubi);
               if badb4 > badnow
                  then do; force = 1; lastforce = try; end;
               if badb4 then lastprob = try;
               %end;

            *------fewer duplicates?------;
            %if &dupflag %then %do;
               t = (shape(xsubi, n, m) = xmat);
               t[i,] = j(1, m, 0);
               dupsnow = sum(t[,+] = m);
               free t;
               if dupsb4 > dupsnow then do; force = 1; lastforce = try; end;
               if dupsb4 then lastprob = try;
               %end;

            *------step back with anneal=p probability?------;
            %if &annealflag %then %do;
               if nt >= anniter
                  %if &restrictflag %then & badnow <= badb4;
                  %if &dupflag      %then & dupsb4 <= dupsnow;
                  then do;
                  call ranuni(seed, rananneal);
                  %if &restrictflag %then %do;
                     if heatitup then
                        rananneal = rananneal # 2 # anneal;
                     %end;
                  if rananneal < anneal then force = 1;
                  end;
               %if &restrictflag %then %do;
                  if heatitup > 1 & ^force then do;
                     call ranuni(seed, rananneal);
                     if rananneal < (1 / facts[j1]) then force = 1;
                     end;
                  %end;
               %end;

            *------should we consider the exchange?------;
            %if &restrictflag or &dupflag or &annealflag
                %then %do; if force | badnow <= badb4 then do; %end;

            *------build main effects for exchange------;
            t = dsubi;
            %do ex = 1 %to &exchange;
               essind     = ei1[k&ex, j&ex];
               essjk      = essmat[,essind : (essind + esscols&ex)];
               t[jvec&ex] = essjk;
               %end;

            *------build interactions for exchange------;
            do ii = 1 to nints;
               if %do ex = 1 %to &exchange;
                     any(int[ii,] = j&ex)
                     %if &ex lt &exchange %then |;
                     %end; then do;
                  do kk = 1 to maxint while(int[ii, kk] > 0);
                     f = int[ii, kk];
                     %do ex = 1 %to &exchange;
                        %if &ex gt 1 %then else;
                        if f = j&ex then lev = k&ex;
                        %end;
                     else lev = xmat[i, f];
                     essind = ei1[lev, f];
                     essind = essind : (essind + indf[f]);
                     if kk = 1 then interact = essmat[,essind];
                     else interact = hdir(interact, essmat[,essind]);
                     end;
                  jj = m + ii;
                  t[s1[jj] : s2[jj]] = interact;
                  end;
               end;

            *------current determ computed with deter update------;
            currdet = max(0, detstar # (1 + (t * invstar * t`) # ninv));

            *------is new design better or otherwise mandated?------;
            if (currdet > maxdetminus
                %if &restrictflag %then & badnow  <= badb4;
                %if &dupflag      %then & dupsnow <= dupsb4;)
               %if &restrictflag or &dupflag or &annealflag
                  %then | force;
               then do;

               *------do not report lots of ties------;
               runreport = (currdet > maxdetplus);

               *------store exchange in xmat------;
               %do ex = 1 %to &exchange; xmat[i, j&ex] = k&ex; %end;
               dsubi = t;    maxdet = currdet; updated = try;
               maxdetplus  = maxdet # &over1;
               maxdetminus = maxdet # &almost1;

               *------store results for later?------;
               improvement = (maxdetminus > bestdet) +
                             (maxdetplus  > bestdet);

               if improvement = 1 then do;
                  if nt ^= ntimesiter then ntimes = ntimes + 1;
                  ntimesiter = nt;
                  end;
               else if improvement = 2 then do;
                  %if &restrictflag or &dupflag %then %do;

                     *------do not store if there are violations------;
                     restricting = (try - %if &accept %then lastforce;
                                    %else lastprob;) <= n;
                     if ^restricting then do; %end;
                        bestdet = maxdet; bestdes = xmat;
                        bestmethod = method; ntimes = 0;
                        %if &restrictflag or &dupflag %then %do; end; %end;
                  end;

               if runreport
                  %if &restrictflag or &dupflag %then | restricting;
                  then changed = try;

               *------print iter history when det improves------;
               if improvement then do;
                  nstep = 0; step = char(i, 7.); col = char(j1, 4, 0);
                  if runreport then run report;
                  if symget('abort') = '1' then goto done;
                  end;

               end;

            %if &restrictflag or &dupflag or &annealflag
               %then %do; end; %end;

            %if &exchange eq 1 and &tableflag ne 0 and
                not (&restrictflag or &dupflag) %then %do; end; %end;
            %do ex = 1 %to &exchange; end; end; %end;

            *================== End of Looping over Exchanges =================;

            *------if something changed in this row, do rank one updates------;
            if updated = try then do;
               d[i,]   = dsubi;
               q       = dsubi # sqrtninv;
               vx      = invstar * q`;
               q       = 1 + q * vx;
               %if &ltfr %then %do;
                  if q <= 0 then do;
                     file log;
                     put 'ERROR: Determinant and inverse update error.';
                     if &somenotes then
                        put "NOTE: Try a larger value for RIDGE= than &ridge..";
                     call symput('abort', '1');
                     goto done;
                     end;
                  %end;
               xpxinv  = invstar - vx * vx` / q;
               partdet = detstar # q;
               numchange = numchange + 1;
               end;

            *------after each pass of n rows, decrease annealing prob------;
            %if &annealflag %then %do;
               if i = 1 & try > 0 then anneal = &annealfun;
               %end;

            %if &largedesign %then %do;
               if try > n
                  %if &restrictflag or &dupflag %then & ^ restricting;
                  then if (datetime() - algortime) > maxtime
                  then largestop = 1;
               %end;

            *------if we did not meet restrictions, try it again------;
            %if &restrictflag and &annealflag %then %do;
               if restricting then do;
                  %do i = 1 %to &exchange; j&i = 1; %end;
                  heatitup   = 0;
                  checkit    = badness(xmat[i,]);
                  if checkit then lastprob = try;
                  resmet     = resmet + (checkit = 0);
                  %if &resrep %then %do;
                     inittype = concat(compress(char(checkit, 5.)),
                                       ' Violations');
                     nstep = 5; step = char(i, 7, 0); col = '    '; run report;
                     %end;

                  if impossmess = 0 & try >= &mintry & resmet = 0 &
                     num(symget('outallnum')) < 3 then do;
                     file log;
                     put 'WARNING: It may be impossible'
                         ' to meet all restrictions.';
                     file print;
                     impossmess = 1;
                     end;

                  *------if some restrictions have been met, try harder------;
                  if checkit = 0 then doitagain = 0;
                  else if try >= startrepeat & try <= stoprepeat &
                     resmet then do;
                     doitagain = doitagain + 1;

                     if doitagain >= repeat * (1 + (resmet > 100) +
                        (resmet > 500) + (resmet > 1000)) then doitagain = 0;
                     else do;
                        if try >= repeat2 & try <= repeat5 &
                           mod(doitagain, 5) = 0 then
                           heatitup = 1 + (try >= repeat3);
                        lastforce = try;
                        try  = try  - 1;
                        n100 = n100 - 1;
                        end;
                     end;
                  end;
               %end;

            end; /* end do try */

         *========= Now We Are Done Looping Over Rows For this Design =========;

         *------have we seen restriction violations in the last n rows?------;
         %if &restrictflag or &dupflag %then %do;
            restricting = (try - %if &accept %then lastforce;
                           %else lastprob;) <= n;
            %end;

         *------at end of iterations for this design, report efficiency------;
         if nt then do;
            nstep = 3; step = '    End'; col = '    '; run report;
            end;

         enditer:;

         outallnum = num(symget('outallnum')) + 1;
         call symput('outallnum', char(outallnum, 12, 0));

         *------optionally store all designs------;
         %if %nrbquote(&outall) ne %then %do;
            all = j(n, 1, outallnum) || j(n, 1, scaleddet) ||
                  xmat %if &rowsfixed %then || fixed;;
            if outallnum = 1 then do;
               vname = {'Design' 'Efficiency'} ||
                       ('x1' : rowcatc('x' || char(m)));
               %if &rowsfixed %then %do; vname = vname || "&fixed"; %end;
               create &outall from all[colname=vname];
               end;
            else edit &outall;
            append from all;
            close &outall;
            free all;
            %end;

         *------store results? do not store if there are violations   ------;
         *------but store results of INIT= design on first (nt=0) pass------;
         improvement = (maxdetminus > bestdet) + (maxdetplus > bestdet);
         if improvement = 1 then do;
            if nt ^= ntimesiter then ntimes = ntimes + 1;
            ntimesiter = nt;
            end;
         else if improvement = 2 & ((nt = 0 & &pass > &firstpass)
            %if &restrictflag or &dupflag %then | ^ restricting;) then do;
            bestdes = xmat; bestdet = maxdet; bestmethod = method; ntimes = 0;
            end;

         *------stopping early due to OPTIONS=LARGEDESIGN------;
         %if &largedesign %then %do;
            if largestop then do;
               nt           = &maxiter + 1;
               tabitersleft = 0;
               file log;
               put 'NOTE: Stopping early, possibly before convergence,'
                   ' with a large design.';
               file print;
               end;
            %end;

         *------quit if we found a good enough design------;
         if ((targetdet + 1e-10) > &target
            %if &restrictflag or &dupflag %then & (try - lastprob) > n;)
            %if &maxdesigns ne %then | outallnum >= &maxdesigns;
            then do;
            nt = &maxiter + 1;
            call symput('pass', '4');
            end;

         if &pass >= 2 & &stopearly & ntimes >= &stopearly then do;
            nt = &maxiter + 1;
            if &somenotes then do;
               file log;
               put 'NOTE: Stopping since it appears'
                   ' that no improvement is possible.';
               file print;
               end;
            end;

         *------have we passed MAXTIME=?  if so we are done------;
         time = datetime() - algortime;

         %if &tableflag ne 0 %then %do;
            if nt & method ^= 'Can' & symget('pass') = '1' & tabitersleft &
               time > 0.5 # maxtime then do;
               mutiter = mutiter - tabitersleft;
               anniter = anniter - tabitersleft;
               tabitersleft = 0;
               if &somenotes then do;
                  file log;
                  put 'NOTE: Switching'
                      ' to a random initialization'
                       %if not &nostimer %then
                           ' after ' (compress(char(time / 60, 7, 2)))
                           ' minutes and ' (compress(char(nt))) ' designs';
                       '.';
                  file print;
                  end;
               if time > maxtime then time = 0;
               end;
            %end;

         if nt & method ^= 'Can' & symget('pass') ^= '4' &
            time > maxtime then do;
            if &somenotes then do;
               file log;
               put 'NOTE: Quitting the '
                   %if &pass eq 1       %then 'algorithm search';
                   %else %if &pass eq 2 %then 'design search';
                   %else                      'refinement';
                   ' step'
                   %if not &nostimer %then
                       ' after ' (compress(char(time / 60, 7, 2)))
                       ' minutes and ' (compress(char(nt))) ' designs';
                   '.';
                  file print;
                  end;
            nt = &maxiter + 1;
            end;
         end; /* end do nt */

      call ranuni(superseed, seed); seed = ceil(1e7 * seed);

      *------done creating all designs, output best one------;
      vname = 'x1' : rowcatc('x' || char(m));
      %if &rowsfixed %then %do;
         if ncol(bestdes) then do;
            vname   = vname   || "&fixed";
            xmat    = xmat    || fixed;
            bestdes = bestdes || fixed;
            end;
         %end;
      create &out from xmat[colname=vname];

      if ncol(bestdes) then do;
         append from bestdes;
         close &out;
         eff = 100 # (bestdet ## (1 / npars));
         if eff <= 99.99995 then do;
            xmat = bestdes;
            run codedes;
            xpx    = d`;
            xpx    = (xpx * xpx`) # ninv;
            maxdet = 100 # (max(det(xpx        ), 0) ## (1 / npars));
            eff    = 100 # (max(det(xpx + ridge), 0) ## (1 / npars));
            call symput('finaleff', compress(char(maxdet, 8, 4)));
            call symput('difflag' , char(abs(maxdet - eff) > 0.01, 1, 0));
            end;
         call symput('_mkteff', compress(char(eff, 8, 4)));
         end;

      %*------create empty design if restrictions are not met------;
      else do;
         call symput('_mkteff', '.');
         file log;
         put 'ERROR: The design was not created.';
         %if &restrictflag or &dupflag %then %do;
            if restricting then do;
               if &somenotes then do;
                  put 'NOTE: The problem could be due to '
                     %if &restrictflag %then
                        'restrictions not being met';
                     %if &restrictflag and &dupflag %then ' or ';
                     %if &dupflag %then
                        'duplicates not being eliminated'; '.';
                  if &initdata & (&restrictflag | &dupflag) then
                     put 'NOTE: Specifying both INIT= and restrictions'
                         ' can cause problems.';
                  end;
               end;
            %end;
         call symput('abort', '1');
         end;

      done:
      call symput('seed'  , char(superseed, 12, 0));
      call symput('method', bestmethod);
      if &pass = 1 then call symput('bestmethod', bestmethod);
      end;
      quit;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *mkttime(Done with IML);

   %if &pass ge &maxstages %then %let pass = 4;
   %end;

*========================== End of the Three Passes ===========================;
%doneiter:
*mkttime(End of the Three Passes - Final Processing);

*------if we are accepting violations, report on final design------;
%if &accept and &somenotes and (&restrictflag or &dupflag) %then %do;

   proc iml &imlopts;

      %if &restrictflag %then %do;
         %if &resflag %then %&resmac;
         %if &ridgemsg %then %do; __facts = { &_mktlist }; %end;
         start badness(x) global(xmat, i, try %if &ridgemsg %then , __facts;
                                 %do i = 1 %to &exchange; , j&i %end;
                                 %if &resflag %then , &reslist;);
            %if not &nox %then %do i = 1 %to &m; x&i = x[&i]; %end;
            bad = 0;
            %if &balance ne    %then %do; %__balance    ; %end;
            %if &partial ne    %then %do; %__partprof   ; %end;
            %if &userrestricts %then %do; %&restrictions; %end;
            return(bad
                   %if &balance ne %then + __bbad;
                   %if &partial ne %then + __pbad;);
            finish;
         %end;

      use &out;
      read all into xmat;
      close &out;
      n = nrow(xmat);
      m = ncol(xmat);

      %do i = 1 %to &exchange; j&i = 1; %end;
      try = 1;

      *------check for restriction violations in final design------;
      rb = 0; rd = 0;
      do i = 1 to n while(^ (rb | rd));
         xsubi = xmat[i,];
         do j1 = 1 to m while(^ (rb | rd));
            %if &restrictflag %then %do;
               if badness(xsubi) then rb = 1;
               %end;
            end;
         %if &dupflag %then %do;
            t = (shape(xsubi, n, m) = xmat);
            t[i,] = j(1, m, 0);
            if sum(t[,+] = m) then rd = 1;
            %end;
         end;

      file log;
      if rb then put 'NOTE: The restrictions were not met.';
      if rd then put 'NOTE: The design has duplicate runs.';
      quit;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------set final levels, randomize------;
%if  %nrbquote(&outr) ne or &levels ne 1 or &levels2 ne 1 or
   &levels3 ne 1 or &int or &first or &last or &render %then %do;
   proc iml &imlopts;
      use &out; read all into out[colname=vname]; close &out;
      seed = &ranseed;
      max  = out[<>,1:&m];
      %if %nrbquote(&outr) ne %then %do; ran = out; %end;

      %if &render %then %do;
         fr = (max >= 10) + 1 || {0};    fm  = max || {0};    fc = 0;
         pm2 = &plusmin2;                pm3 = &plusmin3;
         do j = 1 to &m;
            l = max[j];
            if      l = 2 then c = "&levels2";
            else if l = 3 then c = "&levels3";
            else               c = "&levels";
            if c = '0' then fr[j] = (l > 10) + 1;
            else if c = 'c' & mod(l, 2) = 0 then do;
               fr[j] = 4;
               if l > 20 then fr[j] = 5;
               end;
            else if c = 'c' | c = 'i' then do;
               if l < 20 then fr[j] = 2;
               else fr[j] = 3;
               if (pm2 & l = 2) | (pm3 & l = 3) then fr[j] = fr[j] - 1;
               end;
            end;
         if &int & pm2 then for = '__mktpmf1.'; else if &int then for = '1.';
         %end;

      rand = (symget('restrictions') = ' ' & symget('partial') = ' ');

      do j = 1 to &m;
         l = max[j];

         %if &first or &last %then %do;
            c = out[1,j] - 1;
            if c then out[,j] = mod(out[,j] + (l - c - 1), l) + 1;
            %end;

         %if &last %then %do;
            out[,j] = (l + 1) - out[,j];
            was = (out[,j]);
            %end;

         %if &render %then %do;
            fl = fr[j];
            fc = fc + 1;
            if fm[j] ^= fm[j+1] then do;
               if fl > 2 then w = 'best';
               else if (pm2 & l = 2) | (pm3 & l = 3) then w = '__mktpmf';
               else w = ' ';
               for = compbl(rowcatc(for || '-' || char(fc) || '*' ||
                             w || char(fl) || '.'));
               fc = 0;
               end;
            %end;

         if      l = 2 then c = "&levels2";
         else if l = 3 then c = "&levels3";
         else               c = "&levels";
         if      c = '0'                 then v = (0:(l-1));
         else if c = '1'                 then v = (1:l);
         else if c = 'i' & mod(l, 2) = 0 then v = 2 # (1:l) - (l + 1);
         else                                 v = (1:l) - (l + 1) / 2;
         out[,j] = v[out[,j]];

         *------randomize design------;
         %if %nrbquote(&outr) ne %then %do;
            if rand then do;
               r       = j(1, l, 0);
               call ranuni(seed, r);
               v       = v[rank(r)];
               ran[,j] = v[ran[,j]];
               end;
            %end;
         end;

      %if &render %then %do;
         for = compbl(translate(for, ' ', '-'));
         call symput('renfor', for);
         %end;

      if &int then do;
         out   = j(&n, 1, 1) || out;
         vname = 'x0' || vname;
         end;

      create &out from out[colname=vname]; append from out; close &out;

      *------randomly sort design and output------;
      %if %nrbquote(&outr) ne %then %do;
         out = ran;
         r = j(1, &n, 0);
         call ranuni(seed, r);
         if symget('partial') ^= ' ' then r[1] = -1;
         ran[rank(r),] = out;
         if &int then ran = j(&n, 1, 1) || ran;
         create &outr from ran[colname=vname]; append from ran; close &outr;
         %end;

      quit;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*------skip printing the final efficiency of the best design------;
%if &pass ge 3 and &pfinal and %sysevalf(&finaleff lt 1) %then %do;
   %put WARNING: The final efficiency and levels table will not be printed.;
   %put WARNING: It appears that the design has zero or near-zero efficiency.;
   %if %sysevalf(&ridge gt 0) %then
      %put WARNING: Values in the iteration history are ridged.;
   %end;

%let i = %index(%nrbquote(&varlist), interact);
%let model = model x1-x&m;
%if &i %then %let model = &model %substr(%nrbquote(&optvarlist), &i + 9);

*------print the final efficiency of the best design------;
%if (&pfinal or %nrbquote(&outeff) ne ) and
    &pass ge 3 and %sysevalf(&finaleff ge 1) %then %do;
   *mkttime(Final OPTEX);
   %if &flow %then %put &spacer Starting final OPTEX.;

   %if &showcode ne nocode %then %do;
      %let j = iter=1 keep=1 method=sequential ;
      %put proc optex data=&out%str(;);
      %put %str(   )class x1-x&m / param=orthref%str(;);
      %put %str(   )&model%str(;);
      %put %str(   )generate &j initdesign=&out%str(;);
      %if %nrbquote(&examine) ne %then %put %str(   )examine &examine%str(;);
      %put %str(   )run%str(;) quit%str(;);
      %end;

   ods select none;
   proc optex data=&out;
      ods output classlevels=__tmpxcl efficiencies=__tmpxeff
                 %if &examinevar  %then 'variance matrix'=__tmpxvar;
                 %if &examineinfo %then 'information matrix'=__tmpxinfo;;
      class x1-x&m / param=orthref;
      &model;
      generate iter=1 keep=1 method=sequential initdesign=&out;
      %if %nrbquote(&examine) ne %then %do; examine &examine; %end;
      run; quit;
   ods select all;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   data _null_;
      retain maxl ls 0;
      set __tmpxcl(drop=levels factor);
      array lev[*] $ lev:;
      m = &maxlevneed;
      l = -1;
      do i = 1 to dim(lev);
         if lev[i] eq ' ' then leave;
         l + length(lev[i]) + 1;
         end;
      if l gt maxl then do;
         if l > (&ls - 23) then l = ceil(l * 0.51);
         if l gt maxl then do;
            maxl = l;
            j = max(4, floor((&ls - max(6, l) - 15) / 2));
            k = &ls - 15 - 2 * j;
            call symput('j', compress(put(j, 12.)));
            call symput('k', compress(put(k, 12.)));
            end;
         end;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %if &pfinal %then %do;
      data _null_;
         file print;
         set __tmpxcl;
         length str $ %eval(&k + 5);
         array lev[&maxlevneed] $ lev1-lev&maxlevneed;
         if _n_ = 1 or line gt &ps then do;
            i = floor((input(getoption('linesize'), 5.) - 19) / 2);
            if _n_ ne 1 then put _page_;
            put / +i     'The OPTEX Procedure';
            put / +(i-2) 'Class Level Information' /
                / +&j    'Class  Levels  Values'   /;
            line = 7;
            end;
         put +&j factor $5. +2 levels 3. +5 @;
         do i = 1 to &maxlevneed until(lev[i] eq ' ');
            str = left(trim(str) || ' ' || lev[i]);
            if i lt &maxlevneed then do;
               if lev[i + 1] ne ' ' and
                  length(str) + length(lev[i + 1]) gt &k then do;
                  put str $ / +&j +15 @;
                  str = ' '; line + 1;
                  end;
               end;
            end;
         line + 1;
         put str $;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   data %if %nrbquote(&outeff) ne %then &outeff(drop=designnumber);
        %else _null_;;
      set __tmpxeff;
      length Method $ 20;

      if "&bestmethod" eq 'Can' then method = 'OPTEX';
      else if "&bestmethod" eq 'Tab' then method = 'Orthogonal Array';
      else method = 'Coordinate Exchange';
      Seed = &ranseed;

      %if &pfinal %then %do;
         file print ods=(template='QC.OPTEX.Efficiencies');
         put _ods_;
         %end;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %if &pfinal and (&examineinfo or &examinevar) %then %do;
      proc format; value __zer -1e-8 - 1e-8 = '  0    '; run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   %if &pfinal and &examineinfo %then %do;
      proc print data=__tmpxinfo noobs label rows=page;
         &header 'Information Matrix';
         label rowname= '00'x;
         format _numeric_ __zer7.3;
         id rowname;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      &header;
      %end;

   %if &pfinal and &examinevar %then %do;
      proc print data=__tmpxvar noobs label rows=page;
         &header 'Variance Matrix';
         label rowname= '00'x;
         format _numeric_ __zer7.3;
         id rowname;
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      &header;
      %end;
   %end;

%if &pfinal and &aliasing %then %do;

   *------make dep var, use _ as name delimiter------;
   data __tmpxvar(rename=(%do i = 1 %to &m; x&i=x&i._ %end;));
      set design;
      y = _n_;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %let inaliasing = 1;

   *------suppress transreg warnings------;
   proc printto log=__dummy; run;

   %if &syserr > 4 %then %let abort = 1;

   *------create design matrix with all terms and orthogonal coding------;
   proc transreg design data=__tmpxvar;
      model ide(y) = class(x1_ -- x&m._ x1_ %do i = 2 %to &m;|x&i._%end;
                           @&aliasing / ort);
      output replace out=__tmpxvar;
      run;

   %if &syserr > 4 %then %let abort = 1;

   proc printto log=log; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------get aliasing structure in data  set------;
   ods select none;
   proc glm data=__tmpxvar;
      ods output galiasing=__tmpxnums;
      model y = &_trgind / e aliasing;
      run; quit;
   ods select all;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   %let inaliasing = 0;

   %if &fullaliasing %then %do;

      &header 'Aliasing Structure';
      data _null_;
         length alias $ 32767;
         set __tmpxnums;
         alias = compress(alias, '_');
         length chunk $ &ls;
         ind = 1;
         file print;
         if _n_ = 1 then put;
         do until(alias eq ' ');
            blank = 1;
            l = length(alias);
            len = min(&ls - ind + 1, l);
            if l > len then do i = 1 to len;
               if substr(alias, i, 1) eq ' ' then blank = i;
               end;
            else blank = l + 1;
            chunk = substr(alias, 1, blank);
            put @ind chunk;
            alias = left(substr(alias, blank));
            ind = 4;
            end;
         put;
         if _error_ then call symput('abort', '1');
         run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   *------output each term, toss levels and coefficients------;
   data __tmpxnums(keep=t1 w v1-v5 stars);
      set __tmpxnums;
      length t1 w $ 200;
      if &mainaliasing and index(scan(alias, 1, ' '), '*') then return;
      do j = 1 to 32767 until(w = ' ');
         w = scan(alias, j, ' +-*');
         do i = 1 to 10 until(k = 0);
            k = index(w, '_');
            if k then do;
               do l = k + 1 to k + 10 until(substr(w,l,1) = 'x'); end;
               if substr(w,l,1) = 'x' then substr(w, k, l - k) = '*';
               else substr(w, k, l - k) = ' ';
               end;
            end;
         w = compress(w, ' ');
         if j = 1 then t1 = w;                 /* first term (for now)    */
         v1 = input(scan(w, 1, 'x*'), ?? 5.);  /* factor nums for sorting */
         v2 = input(scan(w, 2, 'x*'), ?? 5.);
         v3 = input(scan(w, 3, 'x*'), ?? 5.);
         v4 = input(scan(w, 4, 'x*'), ?? 5.);
         v5 = input(scan(w, 5, 'x*'), ?? 5.);
         stars = length(w) - length(compress(w, '*')); /* count stars     */
         if nmiss(input(w, ?? 32.)) and w ne ' ' and w ne '...' then output;
         end;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------toss duplicates from each (original) row------;
   proc sort nodupkeys data=__tmpxnums out=__tmpxnums;
      by t1 stars v1-v5 w;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------figure out real first term, might be different than before------;
   data __tmpxnums;
      length t2 $ 200;
      retain t2 ' ';
      set __tmpxnums;
      by t1;
      if first.t1 then t2 = w;
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------toss duplicates from each (original) row------;
   proc sort nodupkeys data=__tmpxnums out=__tmpxnums;
      by t2 stars v1-v5 w;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------array back to one row for each alias list------;
   proc transpose data=__tmpxnums out=__tmpxnums;
      var w;
      by t2;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------figure out how to sort the rows------;
   data __tmpxnums;
      set __tmpxnums;
      retain l 9;
      l = max(l, length(col1));
      call symput('ind', put(l + 3, 5. -L));
      v1 = input(scan(col1, 1, 'x*'), ?? 5.);
      v2 = input(scan(col1, 2, 'x*'), ?? 5.);
      v3 = input(scan(col1, 3, 'x*'), ?? 5.);
      v4 = input(scan(col1, 4, 'x*'), ?? 5.);
      v5 = input(scan(col1, 5, 'x*'), ?? 5.);
      stars = length(col1) - length(compress(col1, '*'));
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   *------main effects first, then two-way ... all in order------;
   proc sort data=__tmpxnums out=__tmpxnums;
      by stars v1-v5;
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

   &header 'Aliasing Scheme';

   *------nicely display the results ------;
   data _null_;
      set __tmpxnums end=eof;
      length alias $ 32767 chunk $ &ls;
      array col[*] col:;
      do i = 2 to dim(col);
         alias = trim(alias) || ' ' || col[i];
         end;
      alias = left(alias);
      file print;
      if _n_ = 1 then put 'Estimable  Aliased' / 'Effect     Effects' /;
      if &mainaliasing and index(col1, '*') then return;
      put col1 @;
      do until(alias eq ' ');
         blank = 1;
         l = length(alias);
         len = min(&ls - &ind + 1, l);
         if l > len then do i = 1 to len;
            if substr(alias, i, 1) eq ' ' then blank = i;
            end;
         else blank = l + 1;
         chunk = substr(alias, 1, blank);
         if chunk ne ' ' then put @&ind chunk;
         else put;
         alias = left(substr(alias, blank));
         end;
      put;

      if eof then put 'NOTE: Some' ' parameters' ' in' ' the'
         ' estimable' ' effects' ' are' ' aliased' ' with' ' some'
         ' parameters' ' in' ' the' ' aliased' ' effects.  For'
         ' effects' ' with' ' more' ' than' ' two' ' levels,' ' the'
         ' aliasing' ' scheme' ' displayed' ' here' ' is' ' potentially'
         ' partial.  Specify' ' EXAMINE=FULL' ' ALIASING=n' ' to'
         ' see' ' the' ' full' ' aliasing' ' structure.';

      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   &header;
   %end;

proc datasets nolist;
   delete __tmpxcand __tmpxcand2 __tmpxopt __tmpxinit __tmpxnums
          __tmpxeff __tmpxcl __tmpxcat __tmpxout __tmpxinitdat
          __tmpxvar __tmpxinfo __tmpxtf(memtype=view);
   run; quit;

%if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;

*------sort design, restore notes, issue messages, and quit------;
options &savenote no&timeopt;
%if %nrbquote(&outall) ne %then %do;
   data &outall &outallopts; set &outall; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%if %nrbquote(&outeff) ne %then %do;
   data &outeff &outeffopts; set &outeff; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%if %nrbquote(&outr) ne %then %do;
   data &outr &outropts; set &outr; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%if &dosort %then %do;
   proc sort data=&out &outopts; by x:; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%else %if %nrbquote(&out) ne __tmpxout %then %do;
   data &out &outopts; set &out; run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

%if &render %then %do;
   options &setnote;

   %if &plusmin2 | &plusmin3 %then %do;
      proc format; value __mktpmf -1 = '-' 1 = '+'; run;

      %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
      %end;

   data _null_;
      set &out;
      file &genfile;
      put (x:) (&renfor);
      if _error_ then call symput('abort', '1');
      run;

   %if &syserr > 4 %then %let abort = 1; %if &abort %then %goto endit;
   %end;

*mkttime(Done);

%endit: options &saveopts;
quit;

%if not &abort and &difflag %then %do;
   %put NOTE: The final   ridged D-efficiency criterion is &_mkteff..;
   %put NOTE: The final unridged D-efficiency criterion is &finaleff..;
   %end;

%if &abort %then %do;
   %if &syserr = 1016 or &syserr = 116 %then %put ERROR: Insufficient memory.;
   %else %if &syserr = 2000 or &syserr = 3000 %then
      %put ERROR: Syntax error.  Check your macro arguments for validity.;
   %else %if &inaliasing %then
      %put ERROR: Insufficient resources for aliasing report.;
   %put ERROR: The MKTEX macro ended abnormally.;
   %end;

%* Second condition is redundant everywhere except MVS;
%if not &nostimer and %sysfunc(getoption(stimer)) eq STIMER %then %do;
   %let time = %sysfunc(round(%sysevalf(%sysfunc(datetime()) - &time), 0.01));
   %if %sysevalf(&time >= 60) %then %do;
      %let time = %sysfunc(round(%sysevalf(&time / 60), 0.01)) minutes.;
      %end;
   %else %let time = &time seconds.;
   %put NOTE: The MKTEX macro used &time;
   %end;

%mendit: %mend;

%*------Hadmard matrix sizes defined------;
%macro mktexishad(s);
   (((&s) le 504 and not ((&s) in (452, 476))) or
    ((&s) in (580, 596, 604, 612, 724, 732, 756, 1060)))
%mend;

%*------debugging routine to dump out macro variables------;
%macro mktexdump(stuff);
   %local i word;

   %let i = 1;
   %let word = %scan(&stuff,1);
   %do %while(&word ne);
      %put &spacer &word=&&&word...;
      %let i = %eval(&i + 1);
      %let word = %scan(&stuff,&i);
      %end;

%mend;

%*------make 2 ** 2 n/2, n=n------;
%macro mktn22(n);
   %local m;
   do;
      %let m = %eval(&n / 2);
      %if &m < 10 %then %let m = 0&m;
      b1  = twolev({1 -1}` @ j(&n / 4, 1, 1) @ h2);
      b2  = j(2, 1, 1) @ (1:&m)`;
      %if &n le &bigkid %then %do;
         d02 = d02 || b1[d&n[,1],];
         d&m = d&m || b2[d&n[,1],];
         %end;
      %else %do;
         d02 = b1;
         d&m = b2;
         %end;
      end
   %mend;

%*------macro execution time------;
%macro mkttime(loc);
   %local thisstep now;
   %global _mktloc%eval(&stepnum+1) _mkttime%eval(&stepnum+1);
   %let stepnum          = %eval(&stepnum + 1);
   %let now              = %sysfunc(datetime());
   %let thisstep         = %sysfunc(round(%sysevalf(&now - &steptime), 0.001));
   %let steptime         = &now;
   %let _mktloc&stepnum  = &loc;
   %let _mkttime&stepnum = &thisstep;

   %if %nrbquote(&loc) = Done %then %do;
      data _null_;
         file log;
         put /// @23 'Timing Information' /;
         tottime = &now - &time;
         do i = 2 to &stepnum;
            step = symget('_mktloc' || put(i - 1, 2. -L));
            pct  = 100 * input(symget('_mkttime' || put(i, 2. -L)), 23.) /
                   tottime;
            sum + pct;
            put step $ 5-55 @58  pct 6.2;
            end;
         put @58 '======' / @58  sum 6.2 ///;
         run;
      %end;

   %mend;
